// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>

#include <Factory.h>

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#ifndef WIN32
#include <unistd.h>
#endif // !WIN32

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

//
// Test basic factory operations
//
void
TestFactory(DBFactory_ptr factory)
{
    //
    // Purge any records left in the database
    //
    {
        DBFactory::DBRecordKeySeq_var records = factory -> getAllKeys();
        cout << "Purging " << records -> length() << " records." << endl;
        for(CORBA::ULong i = 0 ; i < records -> length() ; i++)
        {
            try
            {
                factory -> destroyByKey(records[i]);
            }
            catch(DBFactory::NoRecord&)
            {
                cout << "Record " << records[i]
                     << " doesn't exist anymore." << endl;
            }
        }
    }

    //
    // Create a record
    //
    CORBA::Long key;
    {
        DBRecord_var record = factory -> createRecord();
        key = record -> key();
        cout << "Created record: " << key << endl;
        record -> data("This is some data");
        CORBA::String_var data(record -> data());
        cout << "Added data: " << data << endl;
    }

    //
    // Verify that getByKey works
    //
    {
        cout << "Retrieving record again." << endl;
        DBRecord_var record = factory -> getByKey(key);
        cout << "Got record: " << record -> key() << endl;
        CORBA::String_var data(record -> data());
        cout << "Read data: " << data << endl;
    }

    //
    // Destroy the record
    //
    {
        DBRecord_var record = factory -> getByKey(key);

        cout << "Destroying record." << endl;
        factory -> destroyByKey(key);

        cout << "Expecting an OBJECT_NOT_EXIST exception..." << flush;
        try
        {
            record -> key();
            cout << "FAILED!" << endl;
        }
        catch(const CORBA::OBJECT_NOT_EXIST&)
        {
            cout << "OK!" << endl;
        }

        cout << "Verifying that the record is destroyed..." << flush;
        try
        {
            DBRecord_var checkRecord = factory -> getByKey(key);
            cout << "FAILED!" << endl;
        }
        catch(DBFactory::NoRecord&)
        {
            cout << "OK!" << endl;
        }
    }
}

//
// Test a factory that uses a reaper
//
void
TestReaper(DBFactory_ptr factory)
{
    CORBA::Long key;

    //
    // Test that reaper works. This assumes that the timeout
    // and reaping periods on the server are extremely fast.
    //
    {
        cout << "Creating another record." << endl;
        DBRecord_var record = factory -> createRecord();
        key = record -> key();
        cout << "Created record: " << key << endl;
        record -> data("This is some more data");
        CORBA::String_var data(record -> data());
        cout << "Added data: " << data << endl;

        cout << "Sleeping to wait for the record to be reaped." << endl;
#if defined(WIN32)
        Sleep(1000*5);
#else
        sleep(5);
#endif

        cout << "Refetching object" << endl;
        record = factory -> getByKey(key);

        data = record -> data();
        cout << "Data is: " << data << endl;

        factory -> destroyByKey(key);
    }

    //
    // Create 10 records, then sleep to allow them all to be reaped
    //
    {
        CORBA::ULong i;

        cout << "Creating 10 records" << endl;
        for(i = 0 ; i < 10 ; i++)
            DBRecord_var rec = factory -> createRecord();

        DBFactory::DBRecordKeySeq_var records = factory -> getAllKeys();

        cout << records -> length() << " records now in database." << endl;

        //
        // We want all 10 records as _var types
        //
        DBRecord_var recs[10];
        for(i = 0 ; i < 10 ; i++)
            recs[i] = factory -> getByKey(records[i]);

        cout << "Sleeping five seconds." << endl;
#if defined(WIN32)
        Sleep(1000*5);
#else
        sleep(5);
#endif
        //
        // The records should have been reaped, so we access them
        // again to force them to be reactivated.
        //
        cout << "Setting data" << endl;
        i = 0;
        while(i < 10)
        {
            char buf[100];
            sprintf(buf, "This is record: %lu", (unsigned long)i);
            recs[i] -> data(buf);
            i++;
        }

        cout << "Sleeping five seconds." << endl;
#if defined(WIN32)
        Sleep(1000*5);
#else
        sleep(5);
#endif
        cout << "Getting data" << endl;
        i = 0;
        while(i < 10)
        {
            CORBA::String_var data = recs[i] -> data();
            cout << "Data: " << data << endl;
            i++;
        }

        //
        // Purge these records
        //
        cout << "Purging " << records -> length() << " records." << endl;
        for(i = 0 ; i < records -> length() ; i++)
        {
            try
            {
                factory -> destroyByKey(records[i]);
            }
            catch(DBFactory::NoRecord&)
            {
                cout << "Record " << records[i]
                     << " doesn't exist anymore." << endl;
            }
        }
    }
}

int
main(int argc, char* argv[], char*[])
{
    try
    {
        //
        // Create ORB
        //
        CORBA::ORB_var orb = CORBA::ORB_init(argc, argv);

        {
            //
            // Get factory object that uses a default servant
            //
            const char* refFile = "dfactory.ref";
            ifstream in; // Must use open(name) (VC++ bug)
            in.open(refFile);
            if(in.fail())
            {
                cerr << argv[0] << ": can't open `" << refFile << "': "
                     << strerror(errno) << endl;
                return EXIT_FAILURE;
            }

            char s[1000];
            in >> s;

            CORBA::Object_var obj = orb -> string_to_object(s);
            assert(!CORBA::is_nil(obj));

            DBFactory_var factory = DBFactory::_narrow(obj);
            assert(!CORBA::is_nil(factory));

            cout << "-----------------------------------------" << endl;
            cout << "Testing factory w/ default servant" << endl;
            cout << "-----------------------------------------" << endl;
            TestFactory(factory);
        }

        {
            //
            // Get factory object that uses a servant locator
            //
            const char* refFile = "lfactory.ref";
            ifstream in; // Must use open(name) (VC++ bug)
            in.open(refFile);
            if(in.fail())
            {
                cerr << argv[0] << ": can't open `" << refFile << "': "
                     << strerror(errno) << endl;
                return EXIT_FAILURE;
            }

            char s[1000];
            in >> s;

            CORBA::Object_var obj = orb -> string_to_object(s);
            assert(!CORBA::is_nil(obj));

            DBFactory_var factory = DBFactory::_narrow(obj);
            assert(!CORBA::is_nil(factory));

            cout << endl;
            cout << "-----------------------------------------" << endl;
            cout << "Testing factory w/ servant locator" << endl;
            cout << "-----------------------------------------" << endl;
            TestFactory(factory);
            TestReaper(factory);
        }
    }
    catch(CORBA::SystemException& ex)
    {
        OB::PrintException(ex);
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}
