// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>

#include <DefaultFactory_impl.h>
#include <Database.h>

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

// ----------------------------------------------------------------------
// Record servant class implemented using DSI
// ----------------------------------------------------------------------

class DSIRecord_impl : public PortableServer::DynamicImplementation
{
    CORBA::ORB_var orb_;
    Database* database_;

public:
    DSIRecord_impl(CORBA::ORB_ptr orb, Database* database);

    //
    // Standard IDL to C++ Mapping
    //
    virtual void invoke(CORBA::ServerRequest_ptr request);

    virtual CORBA::RepositoryId
    _primary_interface(const PortableServer::ObjectId&,
                       PortableServer::POA_ptr);
};

// ----------------------------------------------------------------------
// Record servant implementation
// ----------------------------------------------------------------------

DSIRecord_impl::DSIRecord_impl(CORBA::ORB_ptr orb, Database* database)
    : orb_(CORBA::ORB::_duplicate(orb)), database_(database)
{
}

void
DSIRecord_impl::invoke(CORBA::ServerRequest_ptr request)
{
    //
    // Obtain the name of the operation being invoked
    //
    CORBA::String_var name = request -> op_name();

    //
    // Resolve the POACurrent object
    //
    CORBA::Object_var obj = orb_ -> resolve_initial_references("POACurrent");
    PortableServer::Current_var current =
        PortableServer::Current::_narrow(obj);

    //
    // Get the object ID of the current invocation. The object ID
    // is our database key.
    //
    PortableServer::ObjectId_var oid = current -> get_object_id();
    CORBA::String_var id = PortableServer::ObjectId_to_string(oid);
    long key = atol(id);

    //
    // Ensure the key still refers to a valid record in the database
    //
    DBRecordData* data = database_ -> find(key);
    if(data == 0)
        throw CORBA::OBJECT_NOT_EXIST();

    //
    // Service the request
    //

    if(strcmp(name, "_get_key") == 0)
    {
        CORBA::NVList_ptr list;
        CORBA::Any* any;

        orb_ -> create_list(0, list);
        request -> params(list);

        any = new CORBA::Any;
        *any <<= key;
        request -> result(any);

        return;
    }
    else if(strcmp(name, "_get_data") == 0)
    {
        CORBA::NVList_ptr list;
        CORBA::Any* any;

        orb_ -> create_list(0, list);
        request -> params(list);

        DBRecordData* data = database_ -> find(key);
        assert(data != 0);

        any = new CORBA::Any;
        *any <<= data -> data;
        request -> result(any);

        return;
    }
    else if(strcmp(name, "_set_data") == 0)
    {
        CORBA::NVList_ptr list;
        CORBA::Any* any;

        orb_ -> create_list(0, list);
        list -> add(CORBA::ARG_IN) -> value() -> replace(CORBA::_tc_string, 0);
        request -> params(list);

        any = list -> item(0) -> value();
        char* str;
        *any >>= str;

        DBRecordData* data = database_ -> find(key);
        assert(data != 0);
        data -> data = CORBA::string_dup(str);

        return;
    }
    else
    {
        CORBA::NVList_ptr list;
        CORBA::Any* any;

        orb_ -> create_list(0, list);
        request -> params(list);

        any = new CORBA::Any;
        *any <<= new CORBA::BAD_OPERATION();
        request -> exception(any);
        return;
    }
}

CORBA::RepositoryId
DSIRecord_impl::_primary_interface(const PortableServer::ObjectId&,
                                   PortableServer::POA_ptr)
{
    return CORBA::string_dup("IDL:DBRecord:1.0");
}


// ----------------------------------------------------------------------
// DefaultFactory_impl private member implementation
// ----------------------------------------------------------------------

DBRecord_ptr
DefaultFactory_impl::makeRef(CORBA::Long key)
{
    //
    // Create the object ID for the record
    //
    char keyBuff[20];
    sprintf(keyBuff, "%ld", key);
    PortableServer::ObjectId_var oid =
        PortableServer::string_to_ObjectId(keyBuff);

    //
    // Manufacture an object reference for the record
    //
    CORBA::Object_var obj =
        poa_ -> create_reference_with_id(oid, "IDL:DBRecord:1.0");
    DBRecord_ptr result = DBRecord::_narrow(obj);
    return result;
}

// ----------------------------------------------------------------------
// DefaultFactory_impl constructor and destructor
// ----------------------------------------------------------------------
DefaultFactory_impl::DefaultFactory_impl(CORBA::ORB_ptr orb,
                                         PortableServer::POA_ptr parent)
    : nextKey_(0), database_(new Database), orb_(CORBA::ORB::_duplicate(orb))
{
    //
    // Create POA for use by factory
    //
    PortableServer::POAManager_var poaManager = parent -> the_POAManager();
    CORBA::PolicyList policies;
    policies.length(6);
    policies[0] =
        parent -> create_id_assignment_policy(PortableServer::USER_ID);
    policies[1] = parent -> create_lifespan_policy(PortableServer::PERSISTENT);
    policies[2] =
        parent -> create_servant_retention_policy(PortableServer::NON_RETAIN);
    policies[3] =
        parent -> create_id_uniqueness_policy(PortableServer::MULTIPLE_ID);
    policies[4] = parent ->
        create_request_processing_policy(PortableServer::USE_DEFAULT_SERVANT);
    policies[5] = parent ->
     create_implicit_activation_policy(PortableServer::NO_IMPLICIT_ACTIVATION);
    poa_ = parent -> create_POA("default_factory", poaManager, policies);

    //
    // Install default servant
    //
    DSIRecord_impl* defaultServant = new DSIRecord_impl(orb_, database_);
    poa_ -> set_servant(defaultServant);
}

DefaultFactory_impl::~DefaultFactory_impl()
{
    PortableServer::Servant defaultServant = poa_ -> get_servant();
    poa_ -> destroy(CORBA::TRUE, CORBA::TRUE);
    delete defaultServant;
    delete database_;
}

// ----------------------------------------------------------------------
// DefaultFactory public member implementation
// ----------------------------------------------------------------------

//
// Create a new record
//
// To create a new key we have an constantly incrementing nextKey
// member variable. For a real application this isn't good enough,
// but for this demonstration it will suffice.
//
DBRecord_ptr
DefaultFactory_impl::createRecord()
    throw(CORBA::SystemException)
{
    //
    // Add record in database
    //
    long key = nextKey_++;
    database_ -> add(key);
    return makeRef(key);
}

//
// Destroy the record with `key'. If the record isn't found
// then raise a NoRecord exception.
//
void
DefaultFactory_impl::destroyByKey(CORBA::Long key)
    throw(DBFactory::NoRecord,
          CORBA::SystemException)
{
    //
    // Ask the database to destroy the record
    //
    if(!database_ -> destroy(key))
        throw DBFactory::NoRecord();
}

//
// Retrieve the record identified by `key'. If the record isn't
// found then raise a NoRecord exception.
//
DBRecord_ptr
DefaultFactory_impl::getByKey(CORBA::Long key)
    throw(DBFactory::NoRecord,
          CORBA::SystemException)
{
    //
    // Look in the database
    //
    if(database_ -> find(key) != 0)
        return makeRef(key);

    throw DBFactory::NoRecord();
    return DBRecord::_nil(); // Some compilers need this
}

//
// Return the keys of all records in the database
//
DBFactory::DBRecordKeySeq*
DefaultFactory_impl::getAllKeys()
    throw(CORBA::SystemException)
{
    DBFactory::DBRecordKeySeq_var r =
        new DBFactory::DBRecordKeySeq(database_ -> size());
    r -> length(database_ -> size());
    for(CORBA::ULong i = 0 ; i < database_ -> size() ; i++)
        r[i] = database_ -> recordAt(i) -> key;

    return r._retn();
}
