// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <assert.h>
#include <stdlib.h>

#include <OB/CORBA.h>
#include <OB/Hashtable.h>
#include <OB/HashtableI.h>
#include <OB/Timer.h>
#include <OB/Util.h>
#include <OB/Hashers.h>

#include <LRUReaper.h>

using namespace PortableServer; // needed by MSVC for the template below

#ifndef HAVE_NO_EXPLICIT_TEMPLATES
template class OB::Hashtable< ObjectId, LRUReaper::ServantInfo* >;
#else
#ifdef HAVE_PRAGMA_DEFINE
#pragma define(OB::Hashtable< ObjectId, LRUReaper::ServantInfo* >)
#endif
#endif

// ----------------------------------------------------------------------
// LRUReaper constructor and destructor
// ----------------------------------------------------------------------

LRUReaper::ServantInfo::ServantInfo(PortableServer::Servant s)
    : servant(s), count(0)
{
    last = OB::TimerList::timeNow();
}

// ----------------------------------------------------------------------
// LRUReaper constructor and destructor
// ----------------------------------------------------------------------

LRUReaper::LRUReaper(CORBA::ULong maxIdle, ReapListener* listener)
    : maxIdle_(maxIdle), listener_(listener)
{
    servantMap_ = new ServantMap(1023);
}

LRUReaper::~LRUReaper()
{
    OB_SYNCHRONIZED

    ServantMap::Enumerator keys = servantMap_ -> keys();
    for(CORBA::ULong i = 0 ; i < keys.length() ; i++)
    {
        ServantInfo* info;
        assert(servantMap_ -> get(keys[i], info));
        servantMap_ -> remove(keys[i]);
        info -> servant -> _remove_ref();
        delete info;
    }
    delete servantMap_;
}

// ----------------------------------------------------------------------
// LRUReaper public member implementation
// ----------------------------------------------------------------------

void
LRUReaper::addServant(const PortableServer::ObjectId& oid,
                      PortableServer::Servant servant)
{
    OB_SYNCHRONIZED

    if(! servantMap_ -> contains(oid))
    {
        ServantInfo* info = new ServantInfo(servant);
        servantMap_ -> put(oid, info);
        servant -> _add_ref();
    }
}

void
LRUReaper::removeServant(const PortableServer::ObjectId& oid)
{
    OB_SYNCHRONIZED

    ServantInfo* info;
    if(servantMap_ -> get(oid, info))
    {
        servantMap_ -> remove(oid);
        info -> servant -> _remove_ref();
        delete info;
    }
}

PortableServer::Servant
LRUReaper::acquireServant(const PortableServer::ObjectId& oid)
{
    OB_SYNCHRONIZED

    ServantInfo* info;

    if(! servantMap_ -> get(oid, info))
        return 0;

    info -> last = OB::TimerList::timeNow();
    info -> count++;

    OB::MessageViewer* viewer = OB::MessageViewer::instance();
    if(viewer -> getTraceLevel() >= 2)
    {
        CORBA::String_var id = PortableServer::ObjectId_to_string(oid);
        CORBA::String_var msg = CORBA::string_dup("Touching servant ");
        msg += id;
        msg += ", count = ";
        msg += info -> count;
        viewer -> message(msg);
    }

    return info -> servant;
}

void
LRUReaper::releaseServant(const PortableServer::ObjectId& oid)
{
    OB_SYNCHRONIZED

    ServantInfo* info;

    if(! servantMap_ -> get(oid, info))
        return;

    assert(info -> count > 0);
    info -> count--;

    OB::MessageViewer* viewer = OB::MessageViewer::instance();
    if(viewer -> getTraceLevel() >= 2)
    {
        CORBA::String_var id = PortableServer::ObjectId_to_string(oid);
        CORBA::String_var msg = CORBA::string_dup("Releasing servant ");
        msg += id;
        msg += ", count = ";
        msg += info -> count;
        viewer -> message(msg);
    }
}

void
LRUReaper::reap()
{
    OB_SYNCHRONIZED

    OB::MessageViewer* viewer = OB::MessageViewer::instance();

    viewer -> trace(2, "Reaping...");

    struct timeval now = OB::TimerList::timeNow();

    ServantMap::Enumerator keys = servantMap_ -> keys();
    for(CORBA::ULong i = 0 ; i < keys.length() ; i++)
    {
        ServantInfo* info;
        assert(servantMap_ -> get(keys[i], info));

        //
        // Determine the servant's age since last use
        //
        CORBA::ULong age = now.tv_sec - info -> last.tv_sec;

        //
        // Only reap the servant if the age exceeds maxIdle and
        // the usage count == 0 (the count will be 0 when the
        // servant is not currently in use by the application).
        //
        if(age > maxIdle_ && info -> count == 0)
        {
            if(viewer -> getTraceLevel() >= 1)
            {
                CORBA::String_var id =
                    PortableServer::ObjectId_to_string(keys[i]);
                CORBA::String_var msg = CORBA::string_dup("Reaping servant ");
                msg += id;
                msg += ", age = ";
                msg += age;
                viewer -> message(msg);
            }

            servantMap_ -> remove(keys[i]);
            if(listener_)
                listener_ -> servantReaped(keys[i], info -> servant);
            info -> servant -> _remove_ref();
            delete info;
        }
    }
}
