// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef LRUReaper_h
#define LRUReaper_h

#include <Reaper.h>

//
// Concrete subclass of Reaper that implements a least-recently-used
// algorithm for determining when to reap a servant
//
class LRUReaper : public Reaper OB_COMMA_MUTEX
{
public:
    //
    // Implementation details
    //

    struct ServantInfo
    {
        ServantInfo(PortableServer::Servant s);

        PortableServer::Servant servant;
        CORBA::ULong count;
        struct timeval last;
    };

    typedef OB::Hashtable< PortableServer::ObjectId, ServantInfo* > ServantMap;

private:
    CORBA::ULong maxIdle_; // The maximum time a servant can be idle in secs
    ReapListener* listener_;
    ServantMap* servantMap_;

    LRUReaper(const LRUReaper&);
    LRUReaper& operator=(const LRUReaper&);

public:
    //
    // maxIdle defines the number of seconds a servant can be idle until
    // it becomes eligible for reaping. The listener, if supplied, will
    // be invoked when a servant is reaped.
    //
    LRUReaper(CORBA::ULong maxIdle = 10, ReapListener* listener = 0);
    virtual ~LRUReaper();

    //
    // Registers a servant with the reaper. If the reaper is currently
    // managing a servant with the same oid, nothing is done. Otherwise,
    // the reaper invokes _add_ref() once on the servant and begins
    // management of the servant.
    //
    virtual void
    addServant(const PortableServer::ObjectId& oid,
               PortableServer::Servant servant);

    //
    // Removes a servant. The reaper invokes _remove_ref() once on
    // the servant. The ReapListener (if any) is not invoked.
    //
    virtual void
    removeServant(const PortableServer::ObjectId& oid);

    //
    // Obtains the servant with the given oid and signals that the
    // servant is "in use" by the application. If found, the servant
    // is returned, otherwise 0 is returned. The servant will not be
    // eligible for reaping until releaseServant() is called an equivalent
    // number of times.
    //
    virtual PortableServer::Servant
    acquireServant(const PortableServer::ObjectId& oid);

    //
    // Indicates that the application is no longer using the servant with
    // the given oid. Each call to acquireServant() should eventually be
    // followed by a call to releaseServant().
    //
    virtual void
    releaseServant(const PortableServer::ObjectId& oid);

    //
    // Activates an immediate scan of the servants to reap any that
    // are eligible.
    //
    virtual void reap();
};

#endif
