// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_ORB_H
#define OB_ORB_H

//
// Forward declarations
// XXX Are these necessary?
//
namespace OCI
{
class Current_impl;
}

namespace OB
{

template<class K, class T> class Hashtable;
class OARoot;

}

namespace CORBA
{

class ORB;
typedef ORB* ORB_ptr;
typedef ORB_ptr ORBRef;

} // End of namespace CORBA

void OBDuplicate(CORBA::ORB_ptr);
void OBRelease(CORBA::ORB_ptr);

namespace CORBA
{

//
// The ORB class
//
typedef OB::ObjVar< ORB > ORB_var;

class ORB : public OB::RefCount
{
public: // For some compilers the following members must be public
    typedef OB::Hashtable<String_var, Object_var> InitialServiceMap;
    typedef OB::Hashtable<String_var, ValueFactoryBase_var>
        ValueFactoryBaseMap;

private:

    //
    // Hide copy-constructor and assignment operator
    //
    ORB(const ORB&);
    void operator=(const ORB&);

    //
    // The ORB singleton
    //
    static ORB_ptr orb_;

#ifdef HAVE_JTC
    //
    // JTC initialization
    //
    JTCInitialize* dummyForJTC_;
#endif

    //
    // The Object Adapter root
    //
    OB::OARoot* oaRoot_;

    //
    // The OCI connection factory registry
    //
    OCI::ConFactoryRegistry_var registry_;

    //
    // All clients
    //
    OB::ClientSeq clients_;
#ifdef HAVE_JTC
    JTCRecursiveMutex clientsMutex_;
#endif

    //
    // The OCI current object
    //
    OCI::Current_impl* ociCurrent_;
    
    //
    // Set of available initial services
    //
    InitialServiceMap* services_;

    //
    // The set of ValueFactoryBase registered.
    //
    ValueFactoryBaseMap* valueFactoryBaseMap_;
    
    //
    // Default set of policies
    //
    PolicyList policies_;

public:

    //
    // The concurrency model
    //
    enum ConcModel
    {
	//
	// Client uses blocking send/receive
	//
	ConcModelBlocking,
	
	//
	// Client uses the reactor and non-blocking send/receive
	//
	ConcModelReactive

#ifdef HAVE_JTC
	,
	//
	// Client uses separate sender and receiver threads
	//
	ConcModelThreaded
#endif
    };

private:
    
    static ConcModel concModel_;

    ULong native_cs_;
    ULong native_wcs_;

public:

    //
    // The Server Concurrency model
    //
    // TODO: The server side concurrency model does not belong in the ORB
    enum ServerConcModel
    {
	//
	// Server uses blocking send/receive
	//
	ServerConcModelBlocking,
	
	//
	// Server uses the reactor and non-blocking send/receive
	//
	ServerConcModelReactive

        //
        // Various threaded concurrency models
        //
#ifdef HAVE_JTC
	,
        //
	// Server uses separate sender and receiver threads
	//
	ServerConcModelThreaded,

        //
        // Server spawns a new thread for each client
        //
	ServerConcModelThreadPerClient,

        //
        // Server spawns a new thread for each request
        //
	ServerConcModelThreadPerRequest,

        //
        // Server uses a thread pool to dispatch method requests
        //
	ServerConcModelThreadPool
#endif
    };

public:
    
    //
    // For send_multiple_requests_deferred
    //
    typedef OB::ObjSeq< Request, int > RequestSeq;
    
private:
    
    RequestSeq deferredRequests_;
    OB::BooleanSeq deferredRequestCompletions_;

    ORB();
    ~ORB();

    //
    // ORB_init may create ORBs
    //
    friend ORB_ptr ORB_init(int&, char**, const char*);
    typedef char* OAid;

    //
    // Extract the POA parameters from the command line.
    //
    // XXX - This is called in ORB_init. Perhaps this should be
    // an _OB_ method.
    // TODO: Yes, of course it should! (ML)
    //
    void parsePOAParams(int&, char**);

public:

    //
    // Standard IDL to C++ Mapping
    //
    static inline ORB_ptr _duplicate(ORB_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline ORB_ptr _nil()
    { return 0; }

    char* object_to_string(Object_ptr);
    Object_ptr string_to_object(const char*);

    //
    // Dynamic Invocation related operations
    //
    void create_list(Long, NVList_out);
    void create_operation_list(OperationDef_ptr, NVList_out);
    void create_named_value(NamedValue_out);
    void create_exception_list(ExceptionList_out);
    void create_context_list(ContextList_out);

    void get_default_context(Context_out);
    void create_environment(Environment_out);

    void send_multiple_requests_oneway(const RequestSeq&);
    void send_multiple_requests_deferred(const RequestSeq&);
    Boolean poll_next_response();
    void get_next_response(Request_out);

    TypeCode_ptr create_struct_tc(const char*,
                                  const char*,
                                  const StructMemberSeq&);
    
    TypeCode_ptr create_union_tc(const char*,
                                 const char*,
                                 TypeCode_ptr,
                                 const UnionMemberSeq&);
    
    TypeCode_ptr create_enum_tc(const char*,
                                const char*,
                                const EnumMemberSeq&);
    
    TypeCode_ptr create_alias_tc(const char*,
                                 const char*,
                                 TypeCode_ptr);
    
    TypeCode_ptr create_exception_tc(const char*,
                                     const char*,
                                     const StructMemberSeq&);
    
    TypeCode_ptr create_interface_tc(const char*, const char*);
    
    TypeCode_ptr create_string_tc(ULong);
    
    TypeCode_ptr create_wstring_tc(ULong);
    
    TypeCode_ptr create_fixed_tc(UShort, Short);

    TypeCode_ptr create_sequence_tc(ULong, TypeCode_ptr);
    
    TypeCode_ptr create_recursive_sequence_tc(ULong, ULong);
    
    TypeCode_ptr create_array_tc(ULong, TypeCode_ptr);
    
    TypeCode_ptr create_value_tc(const char*,
                                 const char*,
                                 ValueModifier,
                                 TypeCode_ptr,
                                 const ValueMemberSeq&);
    
    TypeCode_ptr create_boxed_value_tc(const char*,
                                       const char*,
                                       TypeCode_ptr);
    
    TypeCode_ptr create_native_tc(const char*, const char*);
    
    TypeCode_ptr create_recursive_tc(const char*);
    
    TypeCode_ptr create_abstract_interface_tc(const char*, const char*);
    
    typedef char* ObjectId;
    typedef String_var ObjectId_var;
    typedef OB::StringSeq ObjectIdList;
    typedef OB::StringSeq_var ObjectIdList_var;

    struct InvalidName : public UserException
    {
	//
	// Standard IDL to C++ Mapping
	//
	InvalidName() { }
	InvalidName(const InvalidName& ex) : UserException(ex) { }
	InvalidName& operator=(const InvalidName&) { return *this; }
	
	static InvalidName* _downcast(Exception* p)
	{ return dynamic_cast<InvalidName*>(p); }
	static const InvalidName* _downcast(const Exception* p)
	{ return dynamic_cast<const InvalidName*>(p); }
	virtual void _raise() const { throw *this; }
	
	//
	// ORBacus internal functions
	// Application programs must not use these functions directly
	//
	virtual Exception* _OB_clone() const
        { return new InvalidName(*this); }
        virtual void _OB_insertInAny(Any&)
        { assert_nca(false, OB::NCANoExceptionTC); }
    };

    ObjectIdList* list_initial_services();
    Object_ptr resolve_initial_references(const char*);
    Boolean get_service_information(ServiceType, ServiceInformation_out);

    //
    // Dynamic Any related operations
    //
    struct InconsistentTypeCode : public UserException
    {
	//
	// Standard IDL to C++ Mapping
	//
	InconsistentTypeCode() { }
	InconsistentTypeCode(const InconsistentTypeCode& ex)
	    : UserException(ex) { }
	InconsistentTypeCode& operator=(const InconsistentTypeCode&)
	{ return *this; }
	
	static InconsistentTypeCode* _downcast(Exception* p)
	{ return dynamic_cast<InconsistentTypeCode*>(p); }
	static const InconsistentTypeCode* _downcast(const Exception* p)
	{ return dynamic_cast<const InconsistentTypeCode*>(p); }
	virtual void _raise() const { throw *this; }
	
	//
	// ORBacus internal functions
	// Application programs must not use these functions directly
	//
	virtual Exception* _OB_clone() const
        { return new InconsistentTypeCode(*this); }
        virtual void _OB_insertInAny(Any&)
        { assert_nca(false, OB::NCANoExceptionTC); }
    };
    
    DynAny_ptr create_dyn_any(const Any& value);
    DynAny_ptr create_basic_dyn_any(TypeCode_ptr type);
    DynStruct_ptr create_dyn_struct(TypeCode_ptr type);
    DynSequence_ptr create_dyn_sequence(TypeCode_ptr type);
    DynArray_ptr create_dyn_array(TypeCode_ptr type);
    DynUnion_ptr create_dyn_union(TypeCode_ptr type);
    DynEnum_ptr create_dyn_enum(TypeCode_ptr type);
    DynFixed_ptr create_dyn_fixed(TypeCode_ptr type);

    //
    // Thread related operations
    //
    Boolean work_pending();
    void perform_work();
    void shutdown(Boolean);
    void run();

    //
    // Policy related operations
    //
    Policy_ptr create_policy(PolicyType policy_type,
				   const Any& val);

    //
    // Value factory operations
    //
    ValueFactory register_value_factory(const char*, ValueFactory);
    void unregister_value_factory(const char*);
    ValueFactory lookup_value_factory(const char*);
                                                  
    //
    // Additional ORBacus specific functions
    //
    void init_servers();

    OCI::ConFactoryRegistry_ptr get_con_factory_registry(); 

    Object_ptr get_inet_object(const char*, UShort, const char*,
			       const char* = "");

    static ConcModel conc_model();
    static void conc_model(ConcModel);

    void add_initial_reference(const char*, Object_ptr);
    
    static void set_iiop_port_from_object(Object_ptr);
    
    // TODO: The server side concurrency model does not belong in the ORB
    static ServerConcModel server_conc_model();
    static void server_conc_model(ServerConcModel);
    static ULong conc_model_thread_pool();
    static void conc_model_thread_pool(ULong);
    OCI::AccRegistry_ptr get_acc_registry();

    //
    // Current native code sets
    //
    CORBA::ULong native_cs() const;
    void native_cs(CORBA::ULong);
    CORBA::ULong native_wcs() const;
    void native_wcs(CORBA::ULong);

    //
    // ORBacus internal functions
    // Application programs must not use these functions directly
    //
    static ORB_ptr _OB_instance();
    static ORB_ptr _OB_instance_nodup();

    void _OB_initializeRootPOA(PortableServer::POA_ptr);
    
    char* _OB_IORToString(const IOP::IOR&);
    IOP::IOR* _OB_stringToIOR(const char*);

    Object_ptr _OB_createObject(const IOP::IOR&);
    
    void _OB_changeStub(OB::StubImplBase*, const IOP::IOR&);
    
    OB::StubImplBase_ptr _OB_getStubImpl(Object_ptr);
    
    void _OB_incClientUsage(OB::Client_ptr);
    void _OB_decClientUsage(OB::Client_ptr);
    
    Boolean _OB_equivalent(const IOP::IOR&, const IOP::IOR&);
    
    ULong _OB_hash(const IOP::IOR&, ULong);

    void _OB_addInitialService(const char*, const char*, bool = true);
    void _OB_addInitialService(const char*, Object_ptr, bool = true);

    OCI::Current_impl* _OB_OCICurrent() const;
};

inline void
release(ORB_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(ORB_ptr p)
{
    return p == 0;
}

//
// ORB initialization
//
ORB_ptr ORB_init(int&, char**, const char* = "");
typedef char* ORBid;

} // End of namespace CORBA

#endif
