// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_TYPE_CODE_H
#define OB_TYPE_CODE_H

//
// Forward declarations
//
namespace OB
{
    class TypeCodeSimpleConst;
    class TypeCodeStringConst;
    class TypeCodeFixedConst;
    class TypeCodeObjectConst;
    class TypeCodeSysExConst;
    class TypeCodeConst;
}

namespace CORBA
{

class TypeCode;
typedef TypeCode* TypeCode_ptr;
typedef TypeCode_ptr TypeCodeRef;

} // End of namespace CORBA

void OBDuplicate(CORBA::TypeCode_ptr);
void OBRelease(CORBA::TypeCode_ptr);

namespace CORBA
{

//
// The TypeCode class
//
typedef OB::ObjVar< TypeCode > TypeCode_var;

class TypeCode : public OB::RefCount
{
    //
    // TypeCode kind
    //
    TCKind kind_;

    //
    // tk_objref, tk_struct, tk_union, tk_enum, tk_alias, tk_value,
    // tk_value_box, tk_native, tk_abstract_interface, tk_except
    //
    String_var id_;
    String_var name_;

    //
    // tk_struct, tk_union, tk_enum, tk_value, tk_except
    //
    OB::StringSeq memberNames_;

    //
    // tk_struct, tk_union, tk_value, tk_except
    //
    OB::TypeCodeSeq memberTypes_;

    //
    // tk_union
    //
    OB::AnySeq labels_;

    //
    // tk_union
    //
    TypeCode_var discriminatorType_;

    //
    // tk_string, tk_sequence, tk_array
    //
    ULong length_;

    //
    // tk_sequence, tk_array, tk_value_box, tk_alias
    //
    TypeCode_var contentType_;

    //
    // tk_fixed
    //
    UShort fixedDigits_;
    Short fixedScale_;

    //
    // tk_value
    //
    OB::ShortSeq memberVisibility_;
    ValueModifier typeModifier_;
    TypeCode_var concreteBaseType_;

    //
    // The number of recursive TypeCodes that are pointing to this
    // TypeCode.
    //
    // tk_struct, tk_union, tk_except, tk_value
    //
    ULong cyclic_;

    //
    // If recId_ is set, this is a placeholder recursive TypeCode that
    // was generated with create_recursive_tc(). If the placeholder
    // recursive TypeCode is already embedded, recType_ points to the
    // recursive TypeCode this placeholder delegates to.
    //
    String_var recId_;
    TypeCode_var recType_;

    //
    // Hide copy-constructor and assignment operator
    //
    TypeCode(const TypeCode&);
    void operator=(const TypeCode&);

protected:

    //
    // Only these classes may create TypeCodes
    //
    friend class ::OB::TypeCodeSimpleConst;
    friend class ::OB::TypeCodeStringConst;
    friend class ::OB::TypeCodeFixedConst;
    friend class ::OB::TypeCodeObjectConst;
    friend class ::OB::TypeCodeSysExConst;
    friend class ::OB::TypeCodeConst;
    friend class ORB;

    //
    // CDR
    //
    friend class ::OB::InputStream;
    friend class ::OB::OutputStream;
    friend class ::OB::CountStream;

    TypeCode();
    virtual ~TypeCode();

public:

    //
    // Standard IDL to C++ Mapping
    //
    static inline TypeCode_ptr _duplicate(TypeCode_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline TypeCode_ptr _nil()
    { return 0; }

    struct Bounds : public UserException
    {
	//
	// Standard IDL to C++ Mapping
	//
	Bounds() { }
	Bounds(const Bounds& ex) : UserException(ex) { }
	Bounds& operator=(const Bounds&) { return *this; }
	
	static Bounds* _downcast(Exception* p)
	{ return dynamic_cast<Bounds*>(p); }
	static const Bounds* _downcast(const Exception* p)
	{ return dynamic_cast<const Bounds*>(p); }
	virtual void _raise() const { throw *this; }
	
	//
	// ORBacus internal functions
	// Application programs must not use these functions directly
	//
	virtual Exception* _OB_clone() const
        { return new Bounds(*this); }
        virtual void _OB_insertInAny(Any&)
        { assert_nca(false, OB::NCANoExceptionTC); }
    };

    struct BadKind : public UserException
    {
	//
	// Standard IDL to C++ Mapping
	//
	BadKind() { }
	BadKind(const BadKind& ex) : UserException(ex) { }
	BadKind& operator=(const BadKind&) { return *this; }
	
	static BadKind* _downcast(Exception* p)
	{ return dynamic_cast<BadKind*>(p); }
	static const BadKind* _downcast(const Exception* p)
	{ return dynamic_cast<const BadKind*>(p); }
	virtual void _raise() const { throw *this; }
	
	//
	// ORBacus internal functions
	// Application programs must not use these functions directly
	//
	virtual Exception* _OB_clone() const
        { return new BadKind(*this); }
        virtual void _OB_insertInAny(Any&)
        { assert_nca(false, OB::NCANoExceptionTC); }
    };

    Boolean equal(TypeCode_ptr) const;
    Boolean equivalent(TypeCode_ptr) const;
    TypeCode_ptr get_compact_typecode() const;
    TCKind kind() const;

    const char* id() const;
    const char* name() const;
    
    ULong member_count() const;
    const char* member_name(ULong index) const;
    
    TypeCode_ptr member_type(ULong index) const;
    
    Any* member_label(ULong index) const;
    TypeCode_ptr discriminator_type() const;
    Long default_index() const;

    ULong length() const;

    TypeCode_ptr content_type() const;

    UShort fixed_digits() const;
    Short fixed_scale() const;

    Visibility member_visibility(ULong) const;
    ValueModifier type_modifier() const;
    TypeCode_ptr concrete_base_type() const;

    //
    // ORBacus internal functions
    // Application programs must not use these functions directly
    //
    bool _OB_isSystemException() const;

    //
    // Check whether a TypeCode is not reachable
    //
    static bool _OB_checkForNonReachableTC(TypeCode_ptr, ULong);

    //
    // Embed recursive placeholder TypeCodes
    //
    static void _OB_embedRecTC(TypeCode_ptr);
    static void _OB_embedRecTC(TypeCode_ptr, TypeCode_ptr);

    //
    // Resolve recursive TypeCodes
    //
    static void _OB_resolveRecTC(TypeCode_ptr);
    static void _OB_resolveRecTC(TypeCode_ptr, TypeCode_ptr);

    static void _OB_release(TypeCode_ptr);
};

inline void
release(TypeCode_ptr p)
{
    TypeCode::_OB_release(p); // Must call Release due to special
                              // cyclic object dependency handling
}

inline Boolean
is_nil(TypeCode_ptr p)
{
    return p == 0;
}

} // End of namespace CORBA

namespace OB
{

//
// Classes for _tc_<type> constants
//
class TypeCodeSimpleConst
{
    CORBA::TypeCode_var typeCode_;
    CORBA::TCKind kind_;

    TypeCodeSimpleConst(const TypeCodeSimpleConst&);
    void operator=(const TypeCodeSimpleConst&);

public:
    
    TypeCodeSimpleConst(CORBA::TCKind);
    operator CORBA::TypeCode_ptr();
    CORBA::TypeCode_ptr operator->() { return operator CORBA::TypeCode_ptr(); }
};

class TypeCodeStringConst
{
    CORBA::TypeCode_var typeCode_;
    bool wide_;

    TypeCodeStringConst(const TypeCodeStringConst&);
    void operator=(const TypeCodeStringConst&);

public:
    
    TypeCodeStringConst(bool);
    operator CORBA::TypeCode_ptr();
    CORBA::TypeCode_ptr operator->() { return operator CORBA::TypeCode_ptr(); }
    CORBA::TypeCode_ptr bounded(CORBA::ULong);
};

class TypeCodeFixedConst
{
    CORBA::TypeCode_var typeCode_;

    TypeCodeFixedConst(const TypeCodeFixedConst&);
    void operator=(const TypeCodeFixedConst&);

public:
    
    TypeCodeFixedConst();
    operator CORBA::TypeCode_ptr();
    CORBA::TypeCode_ptr operator->() { return operator CORBA::TypeCode_ptr(); }
    CORBA::TypeCode_ptr fixed(CORBA::UShort, CORBA::UShort);
};

class TypeCodeObjectConst
{
    CORBA::TypeCode_var typeCode_;

    TypeCodeObjectConst(const TypeCodeStringConst&);
    void operator=(const TypeCodeStringConst&);

public:
    
    TypeCodeObjectConst();
    operator CORBA::TypeCode_ptr();
    CORBA::TypeCode_ptr operator->() { return operator CORBA::TypeCode_ptr(); }
};

class TypeCodeSysExConst
{
    CORBA::TypeCode_var typeCode_;
    const char* name_;

    TypeCodeSysExConst(const TypeCodeSysExConst&);
    void operator=(const TypeCodeSysExConst&);

public:
    
    TypeCodeSysExConst(const char* name);
    operator CORBA::TypeCode_ptr();
    CORBA::TypeCode_ptr operator->() { return operator CORBA::TypeCode_ptr(); }
};

class TypeCodeConst
{
    CORBA::TypeCode_var typeCode_;
    const char* str_;

    TypeCodeConst(const TypeCodeConst&);
    void operator=(const TypeCodeConst&);

public:

    TypeCodeConst(const char* s)
	: typeCode_(CORBA::TypeCode::_nil()), str_(s) { }
    operator CORBA::TypeCode_ptr();
    CORBA::TypeCode_ptr operator->() { return operator CORBA::TypeCode_ptr(); }
};

//
// Get original type, i.e., remove aliases
//
CORBA::TypeCode_ptr GetOrigType(CORBA::TypeCode_ptr);

} // End of namespace OB


namespace CORBA
{

//
// TypeCode constants
//
OB_IMPORT extern OB::TypeCodeSimpleConst _tc_null;
OB_IMPORT extern OB::TypeCodeSimpleConst _tc_void;
OB_IMPORT extern OB::TypeCodeSimpleConst _tc_short;
OB_IMPORT extern OB::TypeCodeSimpleConst _tc_long;
OB_IMPORT extern OB::TypeCodeSimpleConst _tc_longlong;
OB_IMPORT extern OB::TypeCodeSimpleConst _tc_ushort;
OB_IMPORT extern OB::TypeCodeSimpleConst _tc_ulong;
OB_IMPORT extern OB::TypeCodeSimpleConst _tc_ulonglong;
OB_IMPORT extern OB::TypeCodeSimpleConst _tc_float;
OB_IMPORT extern OB::TypeCodeSimpleConst _tc_double;
OB_IMPORT extern OB::TypeCodeSimpleConst _tc_longdouble;
OB_IMPORT extern OB::TypeCodeSimpleConst _tc_boolean;
OB_IMPORT extern OB::TypeCodeSimpleConst _tc_char;
OB_IMPORT extern OB::TypeCodeSimpleConst _tc_wchar;
OB_IMPORT extern OB::TypeCodeSimpleConst _tc_octet;
OB_IMPORT extern OB::TypeCodeSimpleConst _tc_any;
OB_IMPORT extern OB::TypeCodeSimpleConst _tc_TypeCode;
OB_IMPORT extern OB::TypeCodeSimpleConst _tc_Principal;

OB_IMPORT extern OB::TypeCodeStringConst _tc_string;
OB_IMPORT extern OB::TypeCodeStringConst _tc_wstring;

OB_IMPORT extern OB::TypeCodeFixedConst _tc_fixed;

OB_IMPORT extern OB::TypeCodeObjectConst _tc_Object;

OB_IMPORT extern OB::TypeCodeSysExConst _tc_UNKNOWN;
OB_IMPORT extern OB::TypeCodeSysExConst _tc_BAD_PARAM;
OB_IMPORT extern OB::TypeCodeSysExConst _tc_NO_MEMORY;
OB_IMPORT extern OB::TypeCodeSysExConst _tc_IMP_LIMIT;
OB_IMPORT extern OB::TypeCodeSysExConst _tc_COMM_FAILURE;
OB_IMPORT extern OB::TypeCodeSysExConst _tc_INV_OBJREF;
OB_IMPORT extern OB::TypeCodeSysExConst _tc_NO_PERMISSION;
OB_IMPORT extern OB::TypeCodeSysExConst _tc_INTERNAL;
OB_IMPORT extern OB::TypeCodeSysExConst _tc_MARSHAL;
OB_IMPORT extern OB::TypeCodeSysExConst _tc_INITIALIZE;
OB_IMPORT extern OB::TypeCodeSysExConst _tc_NO_IMPLEMENT;
OB_IMPORT extern OB::TypeCodeSysExConst _tc_BAD_TYPECODE;
OB_IMPORT extern OB::TypeCodeSysExConst _tc_BAD_OPERATION;
OB_IMPORT extern OB::TypeCodeSysExConst _tc_NO_RESOURCES;
OB_IMPORT extern OB::TypeCodeSysExConst _tc_NO_RESPONSE;
OB_IMPORT extern OB::TypeCodeSysExConst _tc_PERSIST_STORE;
OB_IMPORT extern OB::TypeCodeSysExConst _tc_BAD_INV_ORDER;
OB_IMPORT extern OB::TypeCodeSysExConst _tc_TRANSIENT;
OB_IMPORT extern OB::TypeCodeSysExConst _tc_FREE_MEM;
OB_IMPORT extern OB::TypeCodeSysExConst _tc_INV_IDENT;
OB_IMPORT extern OB::TypeCodeSysExConst _tc_INV_FLAG;
OB_IMPORT extern OB::TypeCodeSysExConst _tc_INTF_REPOS;
OB_IMPORT extern OB::TypeCodeSysExConst _tc_BAD_CONTEXT;
OB_IMPORT extern OB::TypeCodeSysExConst _tc_OBJ_ADAPTER;
OB_IMPORT extern OB::TypeCodeSysExConst _tc_DATA_CONVERSION;
OB_IMPORT extern OB::TypeCodeSysExConst _tc_OBJECT_NOT_EXIST;
OB_IMPORT extern OB::TypeCodeSysExConst _tc_TRANSACTION_REQUIRED;
OB_IMPORT extern OB::TypeCodeSysExConst _tc_TRANSACTION_ROLLEDBACK;
OB_IMPORT extern OB::TypeCodeSysExConst _tc_INVALID_TRANSACTION;
OB_IMPORT extern OB::TypeCodeSysExConst _tc_INV_POLICY;

} // End of namespace CORBA

#endif
