// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/TemplateI.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/Fixed.h>
#include <OB/Any.h>
#include <OB/OBV.h>
#include <OB/CONV_FRAME.h>
#include <OB/TCKind.h>
#include <OB/TypeCode.h>
#include <OB/Principal.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/IFR4.h>
#include <OB/POAServantBase.h>
#include <OB/ValueBase.h>
#include <OB/AbstractBase.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Template instantiations
// ----------------------------------------------------------------------

#ifndef HAVE_NO_EXPLICIT_TEMPLATES
template class OB::VarVar< Any >;
template class OB::VarOut< Any >;
#else
#ifdef HAVE_PRAGMA_DEFINE
#pragma define(OB::VarVar< Any >)
#pragma define(OB::VarOut< Any >)
#endif
#endif

// ----------------------------------------------------------------------
// Constructor and destructor
// ----------------------------------------------------------------------

Any::Any()
    : type_(TypeCode::_duplicate(_tc_null)), value_(0), rel_(false), info_(0)
{
}

Any::Any(const Any& r)
    : type_(TypeCode::_duplicate(_tc_null)), value_(0),
      rel_(false), info_(0)
{
    *this = r;
}

Any::Any(TypeCode_ptr tc, void* val, Boolean rel)
    : type_(TypeCode::_duplicate(tc)), value_(val),
      rel_(rel), info_(0)
{
}

Any::~Any()
{
    replace(TypeCode::_nil(), 0, false);
}

// ----------------------------------------------------------------------
// Assignment
// ----------------------------------------------------------------------

Any&
Any::operator=(const Any& any)
{
    if(this != &any)
    {
	void* val = 0;
	
	TypeCode_var p = GetOrigType(any.type_);
	
	switch(p -> kind())
	{
	case tk_null:
	case tk_void:
	    break;
	    
	case tk_short:
	    val = new Short(*(const Short*)any.value_);
	    break;
	    
	case tk_long:
	    val = new Long(*(const Long*)any.value_);
	    break;

	case tk_longlong:
	    val = new LongLong(*(const LongLong*)any.value_);
	    break;
	    
	case tk_ushort:
	    val = new UShort(*(const UShort*)any.value_);
	    break;
	    
	case tk_ulong:
	case tk_enum:
	    val = new ULong(*(const ULong*)any.value_);
	    break;

	case tk_ulonglong:
	    val = new ULongLong(*(const ULongLong*)any.value_);
	    break;
    
	case tk_float:
	    val = new Float(*(const Float*)any.value_);
	    break;
	    
	case tk_double:
	    val = new Double(*(const Double*)any.value_);
	    break;
	    
	case tk_longdouble:
#if SIZEOF_DOUBLE != 0
	    val = new LongDouble(*(const LongDouble*)any.value_);
#endif
	    break;
	    
	case tk_fixed:
	    val = new Fixed(*(const Fixed*)any.value_);
	    break;
	    
	case tk_boolean:
	    val = new Boolean(*(const Boolean*)any.value_);
	    break;
	    
	case tk_char:
	    val = new Char(*(const Char*)any.value_);
	    break;
	    
	case tk_wchar:
	    val = new WChar(*(const WChar*)any.value_);
	    break;
	    
	case tk_octet:
	    val = new Octet(*(const Octet*)any.value_);
	    break;
	    
	case tk_any:
	    val = new Any(*(const Any*)any.value_);
	    break;
	    
	case tk_TypeCode:
	    val = TypeCode::_duplicate((TypeCode_ptr)any.value_);
	    break;
	    
	case tk_Principal:
	    val = Principal::_duplicate((Principal_ptr)any.value_);
	    break;
	    
	case tk_objref:
	    val = Object::_duplicate((Object_ptr)any.value_);
	    break;
	    
	case tk_struct:
	case tk_sequence:
	case tk_union:
	case tk_array:
	case tk_except:
	    
	    if(p -> _OB_isSystemException())
	    {
		Exception* ex =
		    ((SystemException*)any.value_) -> _OB_clone();
		val = SystemException::_downcast(ex);
		assert(val);
	    }
	    else if(any.info_)
	    {
		val = any.info_ -> dup(any.value_);
	    }
	    else
	    {
		const Buffer* cbufp = (const Buffer*)any.value_;
		Buffer* bufp = new Buffer(cbufp -> len);
		val = bufp;
		memcpy(bufp -> data, cbufp -> data, cbufp -> len);
	    }
	    break;

	case tk_string:
	    val = string_dup((const char*)any.value_);
	    break;
	    
	case tk_wstring:
	    val = wstring_dup((const wchar_t*)any.value_);
	    break;
	    
	case tk_value:
	case tk_value_box:
	    val = any.value_;
	    add_ref((ValueBase*)val);
	    break;

	case tk_abstract_interface:
            //
            // The only time an Any should contain an abstract interface
            // type is when the value is nil
            //
	    assert(any.value_ == 0);
	    val = 0;
	    break;
	    
	case tk_native:
	case tk_alias:
  	    assert(false);
	    break;
	}
	
	replace(any.type_, val, true, any.info_);
    }
    
    return *this;
}

// ----------------------------------------------------------------------
// Insertion into Any
// ----------------------------------------------------------------------

void
Any::operator<<=(Short v)
{
    replace(_tc_short, new Short(v), true);
}

void
Any::operator<<=(UShort v)
{
    replace(_tc_ushort, new UShort(v), true);
}

void
Any::operator<<=(Long v)
{
    replace(_tc_long, new Long(v), true);
}

void
Any::operator<<=(ULong v)
{
    replace(_tc_ulong, new ULong(v), true);
}

void
Any::operator<<=(LongLong v)
{
    replace(_tc_longlong, new LongLong(v), true);
}

void
Any::operator<<=(ULongLong v)
{
    replace(_tc_ulonglong, new ULongLong(v), true);
}

void
Any::operator<<=(Float v)
{
    replace(_tc_float, new Float(v), true);
}

void
Any::operator<<=(Double v)
{
    replace(_tc_double, new Double(v), true);
}

void
Any::operator<<=(LongDouble v)
{
#if SIZEOF_LONG_DOUBLE != 0
    replace(_tc_longdouble, new LongDouble(v), true);
#endif
}

void
Any::operator<<=(const char* v)
{
    replace(_tc_string, string_dup(v), true);
}

void
Any::operator<<=(const wchar_t* v)
{
    replace(_tc_wstring, wstring_dup(v), true);
}

void
Any::operator<<=(const Any& v)
{
    *this <<= new Any(v);
}

void
Any::operator<<=(Any* v)
{
    replace(_tc_any, v, true);
}

void
Any::operator<<=(const Exception& v)
{
    *this <<= v._OB_clone();
}

void
Any::operator<<=(Exception* v)
{
    v -> _OB_insertInAny(*this);
}

void
Any::operator<<=(TypeCode_ptr v)
{
    TypeCode_ptr val = TypeCode::_duplicate(v);
    *this <<= &val;
}

void
Any::operator<<=(TypeCode_ptr* v)
{
    replace(_tc_TypeCode, *v, true);
}

void
Any::operator<<=(Principal_ptr v)
{
    Principal_ptr val = Principal::_duplicate(v);
    *this <<= &val;
}

void
Any::operator<<=(Principal_ptr* v)
{
    replace(_tc_Principal, *v, true);
}

void
Any::operator<<=(Object_ptr v)
{
    Object_ptr val = Object::_duplicate(v);
    *this <<= &val;
}

void
Any::operator<<=(Object_ptr* v)
{
    replace(_tc_Object, *v, true);
}

void
Any::operator<<=(from_boolean v)
{
    replace(_tc_boolean, new Boolean(v.val), true);
}

void
Any::operator<<=(from_char v)
{
    replace(_tc_char, new Char(v.val), true);
}

void
Any::operator<<=(from_wchar v)
{
    replace(_tc_wchar, new WChar(v.val), true);
}

void
Any::operator<<=(from_octet v)
{
    replace(_tc_octet, new Octet(v.val), true);
}

void
Any::operator<<=(from_string v)
{
    TypeCode_var tc = _tc_string.bounded(v.bound);

    if(v.nocopy)
	replace(tc, v.val, true);
    else
	replace(tc, string_dup(v.val), true);
}

void
Any::operator<<=(from_wstring v)
{
    TypeCode_var tc = _tc_wstring.bounded(v.bound);

    if(v.nocopy)
	replace(tc, v.val, true);
    else
	replace(tc, wstring_dup(v.val), true);
}

void
Any::operator<<=(from_fixed v)
{
    TypeCode_var tc = _tc_fixed.fixed(v.digits, v.scale);
    replace(tc, new Fixed(v.val), true);
}

// ----------------------------------------------------------------------
// Extraction from Any
// ----------------------------------------------------------------------

Boolean
Any::operator>>=(Short& v) const
{
    if(check_type(_tc_short))
    {
	v = *(Short*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(UShort& v) const
{
    if(check_type(_tc_ushort))
    {
	v = *(UShort*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(Long& v) const
{
    if(check_type(_tc_long))
    {
	v = *(Long*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(ULong& v) const
{
    if(check_type(_tc_ulong))
    {
	v = *(ULong*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(LongLong& v) const
{
    if(check_type(_tc_longlong))
    {
	v = *(LongLong*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(ULongLong& v) const
{
    if(check_type(_tc_ulonglong))
    {
	v = *(ULongLong*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(Float& v) const
{
    if(check_type(_tc_float))
    {
	v = *(Float*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(Double& v) const
{
    if(check_type(_tc_double))
    {
	v = *(Double*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(LongDouble& v) const
{
#if SIZEOF_LONG_DOUBLE != 0
    if(check_type(_tc_longdouble))
    {
	v = *(LongDouble*)value_;
	return true;
    }
    else
	return false;
#else
    return false;
#endif
}

Boolean
Any::operator>>=(char*& v) const
{
    if(check_type(_tc_string))
    {
	v = (char*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(wchar_t*& v) const
{
    if(check_type(_tc_wstring))
    {
	v = (wchar_t*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(Any*& v) const
{
    if(check_type(_tc_any))
    {
	v = (Any*)value_;
	return true;
    }
    else
	return false;
}

Boolean 
Any::operator>>=(TypeCode_ptr& v) const
{
    if(check_type(_tc_TypeCode))
    {
	v = (TypeCode_ptr)value_;
	return true;
    }
    else
	return false;
}

Boolean 
Any::operator>>=(Principal_ptr& v) const
{
    if(check_type(_tc_Principal))
    {
	v = (Principal_ptr)value_;
	return true;
    }
    else
	return false;
}

Boolean 
Any::operator>>=(Object_ptr& v) const
{
    if(check_type(_tc_Object))
    {
	v = (Object_ptr)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(to_boolean v) const
{
    if(check_type(_tc_boolean))
    {
	v.val = *(Boolean*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(to_char v) const
{
    if(check_type(_tc_char))
    {
	v.val = *(Char*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(to_wchar v) const
{
    if(check_type(_tc_wchar))
    {
	v.val = *(WChar*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(to_octet v) const
{
    if(check_type(_tc_octet))
    {
	v.val = *(Octet*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(to_string v) const
{
    if(type_ -> kind() == tk_string && type_ -> length() == v.bound)
    {
	v.val = (char*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(to_wstring v) const
{
    if(type_ -> kind() == tk_wstring && type_ -> length() == v.bound)
    {
	v.val = (wchar_t*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(to_fixed v) const
{
    if(check_type(_tc_fixed) && type_ -> fixed_digits() == v.digits &&
       type_ -> fixed_scale() == v.scale)
    {
	v.val = *(Fixed*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(to_object v) const
{
    if(type_ -> kind() == tk_objref)
    {
	v.val = Object::_duplicate((Object_ptr)value_);
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(to_abstract_base v) const
{
    if(value_ == 0)
    {
        v.val = AbstractBase::_nil();
        return true;
    }

    if(type_ -> kind() == tk_objref)
    {
        Object_ptr obj = (Object_ptr)value_;
        AbstractBase_ptr base = obj -> _OB_toAbstractBase();
        if(!is_nil(base))
        {
            Object::_duplicate(obj);
            v.val = base;
            return true;
        }
    }
    else if(type_ -> kind() == tk_value)
    {
        ValueBase* vb = (ValueBase*)value_;
        AbstractBase_ptr base = vb -> _OB_toAbstractBase();
        if(!is_nil(base))
        {
            //
            // Valuetypes are not duplicated when extracted from an any
            //
            //add_ref(vb);
            v.val = base;
            return true;
        }
    }

    return false;
}

// ----------------------------------------------------------------------
// Replace type
// ----------------------------------------------------------------------

void
Any::replace(TypeCode_ptr tc, void* val, Boolean rel)
{
    replace(tc, val, rel, 0);
}

// ----------------------------------------------------------------------
// TypeCode accessor and modifier
// ----------------------------------------------------------------------

TypeCode_ptr
Any::type() const
{
    return TypeCode::_duplicate(type_);
}

void
Any::type(TypeCode_ptr tc)
{
    if(!type_ -> equivalent(tc))
        throw BAD_TYPECODE();
    else
        type_ = TypeCode::_duplicate(tc);
}

// ----------------------------------------------------------------------
// Return type and value
// ----------------------------------------------------------------------

const void*
Any::value() const
{
    return value_;
}

// ----------------------------------------------------------------------
// Additional ORBacus functions
// ----------------------------------------------------------------------

void
Any::replace(TypeCode_ptr tc, void* val, Boolean rel,
             const ConstructedInfo* info)
{
    if(rel_ && value_)
    {
	TypeCode_var p = GetOrigType(type_);
	
	switch(p -> kind())
	{
	case tk_null:
	case tk_void:
	    break;
	    
	case tk_short:
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    delete (Short*)value_;
	    break;
	    
	case tk_long:
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    delete (Long*)value_;
	    break;

	case tk_longlong:
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    delete (LongLong*)value_;
	    break;

	case tk_ushort:
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    delete (UShort*)value_;
	    break;
	    
	case tk_ulong:
	case tk_enum:
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    delete (ULong*)value_;
	    break;

	case tk_ulonglong:
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    delete (ULongLong*)value_;
	    break;

	case tk_float:
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    delete (Float*)value_;
	    break;
	    
	case tk_double:
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    delete (Double*)value_;
	    break;
	    
	case tk_longdouble:
#if SIZEOF_DOUBLE != 0
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    delete (LongDouble*)value_;
#endif
	    break;

	case tk_fixed:
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    delete (Fixed*)value_;
	    break;

	case tk_boolean:
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    delete (Boolean*)value_;
	    break;
	    
	case tk_char:
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    delete (Char*)value_;
	    break;
	    
	case tk_wchar:
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    delete (WChar*)value_;
	    break;
	    
	case tk_octet:
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    delete (Octet*)value_;
	    break;
	    
	case tk_any:
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    delete (Any*)value_;
	    break;
	    
	case tk_TypeCode:
	    release((TypeCode_ptr)value_);
	    break;
	    
	case tk_Principal:
	    release((Principal_ptr)value_);
	    break;
	    
	case tk_objref:
	    release((Object_ptr)value_);
	    break;
	    
	case tk_struct:
	case tk_sequence:
	case tk_union:
	case tk_array:
	case tk_except:

	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);

	    if(p -> _OB_isSystemException())
	    {
		delete (SystemException*)value_;
	    }
	    else if(info_)
	    {
		info_ -> free(value_);
	    }
	    else
	    {
		delete (Buffer*)value_;
	    }
	    break;
	    
	case tk_string:
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    string_free((char*)value_);
	    break;
	    
	case tk_wstring:
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    wstring_free((wchar_t*)value_);
	    break;

	case tk_value:
	case tk_value_box:
	    remove_ref((ValueBase*)value_);
	    break;

	case tk_abstract_interface:
            //
            // The only time an Any should contain an abstract interface
            // type is when the value is nil
            //
	    assert(value_ == 0);
	    break;
	    
	case tk_native:
	case tk_alias: // No alias possible here,
	    break;     // this is to keep the compiler happy
	}
    }

    type_ = TypeCode::_duplicate(tc);
    value_ = val;
    rel_ = rel;
    info_ = info;
}

Boolean
Any::check_type(TypeCode_ptr t) const
{
    return type_ -> equal(t);
}

// ----------------------------------------------------------------------
// Insertion and extraction on Any_var
// ----------------------------------------------------------------------

void
operator<<=(Any_var& any, Short val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, UShort val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, Long val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, ULong val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, LongLong val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, ULongLong val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, Float val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, Double val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, LongDouble val)
{
#if SIZEOF_DOUBLE != 0
    any.inout() <<= val;
#endif
}

void
operator<<=(Any_var& any, const char* val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, const wchar_t* val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, char* val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, wchar_t* val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, const Any& val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, Any* val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, TypeCode_ptr val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, TypeCode_ptr* val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, Principal_ptr val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, Principal_ptr* val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, Object_ptr val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, Object_ptr* val)
{
    any.inout() <<= val;
}

Boolean
operator>>=(const Any_var& any, Short& val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, UShort& val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, Long& val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, ULong& val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, LongLong& val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, ULongLong& val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, Float& val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, Double& val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, LongDouble& val)
{
#if SIZEOF_DOUBLE != 0
    return any.in() >>= val;
#endif
}

Boolean
operator>>=(const Any_var& any, char*& val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, wchar_t*& val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, Any*& val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, TypeCode_ptr& val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, Principal_ptr& val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, Object_ptr& val)
{
    return any.in() >>= val;
}

void
operator<<=(Any_var& any, Any::from_boolean val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, Any::from_octet val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, Any::from_char val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, Any::from_wchar val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, Any::from_string val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, Any::from_wstring val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, Any::from_fixed val)
{
    any.inout() <<= val;
}

Boolean
operator>>=(const Any_var& any, Any::to_boolean val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, Any::to_octet val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, Any::to_char val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, Any::to_wchar val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, Any::to_string val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, Any::to_wstring val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, Any::to_fixed val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, Any::to_object val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, Any::to_abstract_base val)
{
    return any.in() >>= val;
}

// ----------------------------------------------------------------------
// Insertion and extraction for system exceptions
// ----------------------------------------------------------------------

#define OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(T) \
\
void \
operator<<=(Any& any, T* v) \
{ \
    SystemException* ex = v; \
    any.replace(_tc_##T, ex, true); \
} \
\
void \
operator<<=(Any& any, const T& v) \
{ \
    SystemException* ex = new T(v); \
    any.replace(_tc_##T, ex, true); \
} \
\
Boolean \
operator>>=(const Any& any, T*& v) \
{ \
    if(any.check_type(_tc_##T)) \
    { \
	SystemException* ex = (SystemException*)any.value(); \
        v = T::_downcast(ex); \
        assert(v); \
        return true; \
    } \
    else \
        return false; \
}\
\
void \
T::_OB_insertInAny(Any& any) \
{ \
    any <<= this; \
}

OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(UNKNOWN)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(BAD_PARAM)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(NO_MEMORY)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(IMP_LIMIT)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(COMM_FAILURE)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(INV_OBJREF)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(NO_PERMISSION)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(INTERNAL)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(MARSHAL)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(INITIALIZE)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(NO_IMPLEMENT)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(BAD_TYPECODE)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(BAD_OPERATION)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(NO_RESOURCES)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(NO_RESPONSE)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(PERSIST_STORE)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(BAD_INV_ORDER)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(TRANSIENT)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(FREE_MEM)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(INV_IDENT)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(INV_FLAG)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(INTF_REPOS)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(BAD_CONTEXT)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(OBJ_ADAPTER)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(DATA_CONVERSION)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(OBJECT_NOT_EXIST)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(TRANSACTION_REQUIRED)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(TRANSACTION_ROLLEDBACK)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(INVALID_TRANSACTION)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(INV_POLICY)
