// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Util.h>

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#ifdef HAVE_STRSTREAM
#   include <strstream>
#else
#   if defined(HAVE_STRSTREA_H)
#       include <strstrea.h>
#   else
#       include <strstream.h>
#   endif
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

// ----------------------------------------------------------------------
// Helper function for assert_nca()
// ----------------------------------------------------------------------

void
OB::NCA(OB::NCAReason reason)
{
    //
    // Update the documentation if error messages are changed here !!!!
    //

    const char* err = 0;

    switch(reason)
    {
    case NCANotImpl:
	err =
"Application requested a feature that is not yet implemented";
	break;

    case NCADeprecated:
	err =
"Application used a deprecated feature that is not implemented anymore";
	break;

    case NCAWrongAlloc:
	err =
"Application used wrong memory allocation function";
	break;

    case NCADoubleFree:
	err =
"Memory that was already deallocated was deallocated again";
	break;

    case NCARefCountNotZero:
	err =
"Object was deleted without an object reference count of zero";
	break;

    case NCARefCountZero:
	err =
"Object was already deleted (object reference count was already zero)";
	break;

    case NCASeqLenGrMax:
	err =
"Sequence length was greater than maximum sequence length";
	break;

    case NCASeqRange:
	err =
"Index for sequence operator[]() or remove() function was out of range";
	break;

    case NCANullInit:
	err =
"Null pointer was used to initialize T_var type";
	break;

    case NCANullPtr:
	err =
"operator->() was used on null pointer or nil object reference";
	break;

    case NCANullDeref:
	err =
"Application tried to dereference a null pointer";
	break;

    case NCANullString:
	err =
"Null pointer was passed as string parameter or return value";
	break;

    case NCANullValue:
        err =
"Null value passed as parameter.";
        break;

    case NCASelfAssign:
	err =
"Self assignment caused a dangling pointer";
	break;

    case NCAAnySelfReplace:
	err =
"Replacement of Any content by its own value caused a dangling pointer";
	break;

    case NCADiscType:
	err =
"Invalid union discriminator type used";
	break;

    case NCADiscMismatch:
	err =
"Union discriminator mismatch";
	break;

    case NCAUnionInit:
	err =
"Uninitialized union used";
	break;

    case NCANoExceptionTC:
        err =
"CORBA::Any::operator<<=(Exception*) can't be used with --no-type-codes";
        break;

    case NCAUnembeddedRecTC:
	err =
"An operation on an unembedded recursive TypeCode was invoked";
	break;

    case NCAEmbeddedRecTC:
	err =
"An already embedded TypeCode was reused";
	break;
    }
    
    assert(err);

    MessageViewer* viewer = MessageViewer::instance();
    viewer -> nca(err);
}

// ----------------------------------------------------------------------
// Special ORBacus memory allocation that keeps track of the number of
// elements in a buffer
// ----------------------------------------------------------------------

static const size_t more = sizeof(CORBA::ULong) * 2;

#ifndef NDEBUG
static const CORBA::ULong magic     = 0x02468ACE;
static const CORBA::ULong magicFree = 0x0E2C4A68;
#endif

void*
OB::Alloc(size_t sz, CORBA::ULong n)
{
#ifndef NDEBUG

    char* p = (char*)malloc(more + n * sz);
    
    CORBA::ULong* np = (CORBA::ULong*)p;
    *np++ = magic;
    *np = n;

#else

    char* p = (char*)malloc(more + n * sz);

    CORBA::ULong* np = (CORBA::ULong*)p;
    *np = n;

#endif

    return p + more;
}

void
OB::Free(void* vp)
{
    char* p = (char*)vp;

#ifndef NDEBUG

    CORBA::ULong* magicp = (CORBA::ULong*)(p - more);
    assert_nca(*magicp != magicFree, NCADoubleFree);
    assert_nca(*magicp == magic, NCAWrongAlloc);
    *magicp = magicFree;

#endif

    free(p - more);
}

CORBA::ULong
OB::Allocated(void* vp)
{
    char* p = (char*)vp;

#ifndef NDEBUG

    CORBA::ULong* magicp = (CORBA::ULong*)(p - more);
    assert_nca(*magicp == magic, NCAWrongAlloc);

    CORBA::ULong* np = magicp + 1;

#else

    CORBA::ULong* np = (CORBA::ULong*)(p - more);

#endif

    return *np;
}

#ifndef NDEBUG

bool
OB::AllocCheck(const void* vp)
{
    const char* p = (const char*)vp;
    const CORBA::ULong* magicp = (const CORBA::ULong*)(p - more);
    return *magicp == magic;
}

#endif

// ----------------------------------------------------------------------
// Memory allocation functions used by string sequences
// ----------------------------------------------------------------------

char**
OB::StrSeqAllocbuf(CORBA::ULong n)
{
    char** p = (char**)Alloc(sizeof(char*), n);
    
    for(CORBA::ULong i = 0 ; i < n ; i++)
	p[i] = 0;

    return p;
}

void
OB::StrSeqFreebuf(char** p)
{
    if(p)
    {
	CORBA::ULong n = Allocated(p);
	
	for(CORBA::ULong i = 0 ; i < n ; i++)
	    CORBA::string_free(p[i]);

	Free(p);
    }
}

wchar_t**
OB::WStrSeqAllocbuf(CORBA::ULong n)
{
    wchar_t** p = (wchar_t**)Alloc(sizeof(wchar_t*), n);
    
    for(CORBA::ULong i = 0 ; i < n ; i++)
	p[i] = 0;

    return p;
}

void
OB::WStrSeqFreebuf(wchar_t** p)
{
    if(p)
    {
	CORBA::ULong n = Allocated(p);
	
	for(CORBA::ULong i = 0 ; i < n ; i++)
	    CORBA::wstring_free(p[i]);

	Free(p);
    }
}

// ----------------------------------------------------------------------
// Convert between octet sequence and ascii string
// ----------------------------------------------------------------------

char*
OB::OctetsToAscii(const CORBA::Octet* oct, CORBA::ULong len)
{
    char* str = CORBA::string_alloc(len * 2);
    char s[3];
    
    for(CORBA::ULong i = 0 ; i < len ; i++, oct++)
    {
	sprintf(s, "%02x", *oct);
	strcpy(&str[i * 2], s);
    }

    return str;
}

CORBA::Octet*
OB::AsciiToOctets(const char* str)
{
    CORBA::ULong len = strlen(str) / 2;
    CORBA::Octet* oct = new CORBA::Octet[len];

    int val;
    char hex[3];
    hex[2] = '\0';

    const char* s = str;
    for(CORBA::ULong i = 0 ; i < len ; i++)
    {
	hex[0] = *s++;
	hex[1] = *s++;
	sscanf(hex, "%x", &val);
	oct[i] = val;
    }

    return oct;
}

// ----------------------------------------------------------------------
// String memory management
// ----------------------------------------------------------------------

namespace OB
{

char* NullString = (char*)"";
wchar_t* NullWString = (wchar_t*)L"";

} // End of namespace OB

char*
CORBA::string_alloc(CORBA::ULong len)
{
    return (char*)OB::Alloc(sizeof(char), len + 1);
}

wchar_t*
CORBA::wstring_alloc(CORBA::ULong len)
{
    return (wchar_t*)OB::Alloc(sizeof(wchar_t), len + 1);
}

void
CORBA::string_free(char* p)
{
    if(p && p != OB::NullString)
	OB::Free(p);
}

void
CORBA::wstring_free(wchar_t* p)
{
    if(p && p != OB::NullWString)
	OB::Free(p);
}

char*
CORBA::string_dup(const char* p)
{
    if(p)
    {
	if(p == OB::NullString)
	    return OB::NullString;

	char* ret = CORBA::string_alloc(strlen(p));
	strcpy(ret, p);
	return ret;
    }
    else
	return 0;
}

wchar_t*
CORBA::wstring_dup(const wchar_t* p)
{
    if(p)
    {
	if(p == OB::NullWString)
	    return OB::NullWString;

	wchar_t* ret = CORBA::wstring_alloc(OB::wcslen(p));
	OB::wcscpy(ret, p);
	return ret;
    }
    else
	return 0;
}

CORBA::ULong
CORBA::string_allocated(const char* p)
{
    if(p && p != OB::NullString)
	return OB::Allocated((void*)p) - 1;
    else
	return 0;
}

CORBA::ULong
CORBA::wstring_allocated(const wchar_t* p)
{
    if(p && p != OB::NullWString)
	return OB::Allocated((void*)p) - 1;
    else
	return 0;
}

#ifndef NDEBUG

CORBA::Boolean
CORBA::string_check(const char* p)
{
    if(p && p != OB::NullString)
	return OB::AllocCheck(p);
    else
	return true;
}

CORBA::Boolean
CORBA::wstring_check(const wchar_t* p)
{
    if(p && p != OB::NullWString)
	return OB::AllocCheck(p);
    else
	return true;
}

#endif

// ----------------------------------------------------------------------
// Wide string helper functions, platform independent
// ----------------------------------------------------------------------

size_t
OB::wcslen(const wchar_t* str)
{
    size_t len = 0;
    while(*str++ != L'\0')
	len++;

    return len;
}

wchar_t*
OB::wcscpy(wchar_t* dst, const wchar_t* src)
{
    wchar_t* result = dst;
    while((*dst++ = *src++))
        /*do nothing*/;

    return result;
}

wchar_t*
OB::wcscat(wchar_t* dst, const wchar_t* src)
{
    wcscpy(dst + wcslen(dst), src);

    return dst;
}

int
OB::wcscmp(const wchar_t* ws1, const wchar_t* ws2)
{
  wchar_t wc1, wc2;
 
  do
  {
      wc1 = (wchar_t)*ws1++;
      wc2 = (wchar_t)*ws2++;
      if(!wc1)
	  return wc1 - wc2;
  }
  while(wc1 == wc2);
 
  return wc1 - wc2;
}

// ----------------------------------------------------------------------
// String_var type
// ----------------------------------------------------------------------

CORBA::String_var::String_var(char* p)
    : ptr_(p)
{
    assert_nca(p, OB::NCANullInit);
    assert_nca(string_check(p), OB::NCAWrongAlloc);
}

CORBA::String_var::String_var(const CORBA::String_var& r)
    : ptr_(string_dup(r.ptr_))
{
}

CORBA::String_var::String_var(const OB::StrForStruct& r)
    : ptr_(string_dup(r.ptr_))
{
}

CORBA::String_var::String_var(const OB::StrForSeq& r)
    : ptr_(string_dup(r))
{
}

CORBA::String_var&
CORBA::String_var::operator=(char* p)
{
    assert_nca(string_check(p), OB::NCAWrongAlloc);
    assert_nca(ptr_ == 0 || p != ptr_, OB::NCASelfAssign);
    string_free(ptr_);
    ptr_ = p;

    return *this;
}

CORBA::String_var&
CORBA::String_var::operator=(const char* p)
{
    if(p != ptr_)
    {
	string_free(ptr_);
	ptr_ = string_dup(p);
    }

    return *this;
}

CORBA::String_var&
CORBA::String_var::operator=(const CORBA::String_var& r)
{
    if(r.ptr_ != ptr_)
    {
	string_free(ptr_);
	ptr_ = string_dup(r.ptr_);
    }

    return *this;
}

CORBA::String_var&
CORBA::String_var::operator=(const OB::StrForStruct& r)
{
    return operator=((const String_var&)r);
}

CORBA::String_var&
CORBA::String_var::operator=(const OB::StrForSeq& r)
{
    if(r.in() != ptr_)
    {
	string_free(ptr_);
	ptr_ = string_dup(r);
    }

    return *this;
}

char*&
CORBA::String_var::out()
{
    string_free(ptr_);
    ptr_ = 0;
    return ptr_;
}

char*
CORBA::String_var::_retn()
{
    char* ret = ptr_;
    ptr_ = 0;
    return ret;
}

CORBA::String_var&
CORBA::String_var::operator+=(const char* p)
{
    if(p)
    {
	if(ptr_)
	{
	    ULong max = string_allocated(ptr_);
	    ULong newLen = strlen(ptr_) + strlen(p);

	    if(newLen > max)
	    {
		max = newLen > max * 2 ? newLen : max * 2;
		char* oldPtr = ptr_;
		ptr_ = string_alloc(max);
		strcpy(ptr_, oldPtr);
		string_free(oldPtr);
	    }

	    strcat(ptr_, p);
	}
	else
	    ptr_ = string_dup(p);
    }

    return *this;
}

CORBA::String_var&
CORBA::String_var::operator+=(char c)
{
    char s[2];
    s[0] = c;
    s[1] = '\0';
    return operator+=(s);
}

CORBA::String_var&
CORBA::String_var::operator+=(unsigned char c)
{
    return operator+=((char)c);
}

CORBA::String_var&
CORBA::String_var::operator+=(long n)
{
    char str[1024];
    ostrstream s(str, 1024);
    s << n << '\0';
    assert(s.good());
    return operator+=(str);
}

CORBA::String_var&
CORBA::String_var::operator+=(unsigned long n)
{
    char str[1024];
    ostrstream s(str, 1024);
    s << n << '\0';
    assert(s.good());
    return operator+=(str);
}

// ----------------------------------------------------------------------
// String type for string sequences
// ----------------------------------------------------------------------

OB::StrForSeq&
OB::StrForSeq::operator=(char* p)
{
    assert_nca(CORBA::string_check(p), NCAWrongAlloc);
    assert_nca(p != *ptr_, NCASelfAssign);

    if(rel_)
	CORBA::string_free(*ptr_);

    *ptr_ = p;

    return *this;
}

OB::StrForSeq&
OB::StrForSeq::operator=(const char* p)
{
    if(p != *ptr_)
    {
	if(rel_)
	    CORBA::string_free(*ptr_);

	*ptr_ = CORBA::string_dup(p);
    }

    return *this;
}

OB::StrForSeq&
OB::StrForSeq::operator=(const CORBA::String_var& r)
{
    if(r.in() != *ptr_)
    {
	if(rel_)
	    CORBA::string_free(*ptr_);

	*ptr_ = CORBA::string_dup(r);
    }

    return *this;
}

OB::StrForSeq&
OB::StrForSeq::operator=(const OB::StrForStruct& r)
{
    return operator=((const CORBA::String_var&)r);
}

OB::StrForSeq&
OB::StrForSeq::operator=(const OB::StrForSeq& r)
{
    if(*(r.ptr_) != *ptr_)
    {
	if(rel_)
	    CORBA::string_free(*ptr_);

	*ptr_ = CORBA::string_dup(*(r.ptr_));
    }

    return *this;
}

char*&
OB::StrForSeq::out()
{
    if(rel_)
	CORBA::string_free(*ptr_);
    *ptr_ = 0;
    return *ptr_;
}

char*
OB::StrForSeq::_retn()
{
    char* ret = *ptr_;
    *ptr_ = 0;
    return ret;
}

// ----------------------------------------------------------------------
// WString_var type
// ----------------------------------------------------------------------

CORBA::WString_var::WString_var(wchar_t* p)
    : ptr_(p)
{
    assert_nca(p, OB::NCANullInit);
    assert_nca(wstring_check(p), OB::NCAWrongAlloc);
}

CORBA::WString_var::WString_var(const WString_var& r)
    : ptr_(wstring_dup(r.ptr_))
{
}

CORBA::WString_var::WString_var(const OB::WStrForStruct& r)
    : ptr_(wstring_dup(r.ptr_))
{
}

CORBA::WString_var::WString_var(const OB::WStrForSeq& r)
    : ptr_(wstring_dup(r))
{
}

CORBA::WString_var&
CORBA::WString_var::operator=(wchar_t* p)
{
    assert_nca(wstring_check(p), OB::NCAWrongAlloc);
    assert_nca(ptr_ == 0 || p != ptr_, OB::NCASelfAssign);
    wstring_free(ptr_);
    ptr_ = p;

    return *this;
}

CORBA::WString_var&
CORBA::WString_var::operator=(const wchar_t* p)
{
    if(p != ptr_)
    {
	wstring_free(ptr_);
	ptr_ = wstring_dup(p);
    }

    return *this;
}

CORBA::WString_var&
CORBA::WString_var::operator=(const WString_var& r)
{
    if(r.ptr_ != ptr_)
    {
	wstring_free(ptr_);
	ptr_ = wstring_dup(r.ptr_);
    }

    return *this;
}

CORBA::WString_var&
CORBA::WString_var::operator=(const OB::WStrForStruct& r)
{
    return operator=((const WString_var&)r);
}

CORBA::WString_var&
CORBA::WString_var::operator=(const OB::WStrForSeq& r)
{
    if(r.in() != ptr_)
    {
	wstring_free(ptr_);
	ptr_ = wstring_dup(r);
    }

    return *this;
}

wchar_t*&
CORBA::WString_var::out()
{
    wstring_free(ptr_);
    ptr_ = 0;
    return ptr_;
}

wchar_t*
CORBA::WString_var::_retn()
{
    wchar_t* ret = ptr_;
    ptr_ = 0;
    return ret;
}

CORBA::WString_var&
CORBA::WString_var::operator+=(const wchar_t* p)
{
    if(p)
    {
	if(ptr_)
	{
	    ULong max = wstring_allocated(ptr_);
	    ULong newLen = OB::wcslen(ptr_) + OB::wcslen(p);

	    if(newLen > max)
	    {
		max = newLen > max * 2 ? newLen : max * 2;
		wchar_t* oldPtr = ptr_;
		ptr_ = wstring_alloc(max);
		OB::wcscpy(ptr_, oldPtr);
		wstring_free(oldPtr);
	    }

	    OB::wcscat(ptr_, p);
	}
	else
	    ptr_ = wstring_dup(p);
    }

    return *this;
}

CORBA::WString_var&
CORBA::WString_var::operator+=(wchar_t c)
{
    wchar_t s[2];
    s[0] = c;
    s[1] = '\0';
    return operator+=(s);
}

/*
CORBA::WString_var&
CORBA::WString_var::operator+=(long n)
{
    char str[1024];
    wchar_t wstr[1024];
    ostrstream s(str, 1024);
    s << n << '\0';
    assert(s.good());

    ULong i;
    for(i = 0 ; i < strlen(str) + 1 ; i++)
        wstr[i] = str[i];

    return operator+=(wstr);
}

CORBA::WString_var&
CORBA::WString_var::operator+=(unsigned long n)
{
    char str[1024];
    wchar_t wstr[1024];
    ostrstream s(str, 1024);
    s << n << '\0';
    assert(s.good());

    ULong i;
    for(i = 0 ; i < strlen(str) + 1 ; i++)
	wstr[i] = str[i];

    return operator+=(wstr);
}
*/

// ----------------------------------------------------------------------
// String type for string sequences
// ----------------------------------------------------------------------

OB::WStrForSeq&
OB::WStrForSeq::operator=(wchar_t* p)
{
    assert_nca(CORBA::wstring_check(p), NCAWrongAlloc);
    assert_nca(p != *ptr_, NCASelfAssign);

    if(rel_)
	CORBA::wstring_free(*ptr_);

    *ptr_ = p;

    return *this;
}

OB::WStrForSeq&
OB::WStrForSeq::operator=(const wchar_t* p)
{
    if(p != *ptr_)
    {
	if(rel_)
	    CORBA::wstring_free(*ptr_);

	*ptr_ = CORBA::wstring_dup(p);
    }

    return *this;
}

OB::WStrForSeq&
OB::WStrForSeq::operator=(const CORBA::WString_var& r)
{
    if(r.in() != *ptr_)
    {
	if(rel_)
	    CORBA::wstring_free(*ptr_);

	*ptr_ = CORBA::wstring_dup(r);
    }

    return *this;
}

OB::WStrForSeq&
OB::WStrForSeq::operator=(const OB::WStrForStruct& r)
{
    return operator=((const CORBA::WString_var&)r);
}

OB::WStrForSeq&
OB::WStrForSeq::operator=(const OB::WStrForSeq& r)
{
    if(*(r.ptr_) != *ptr_)
    {
	if(rel_)
	    CORBA::wstring_free(*ptr_);

	*ptr_ = CORBA::wstring_dup(*(r.ptr_));
    }

    return *this;
}

wchar_t*&
OB::WStrForSeq::out()
{
    if(rel_)
	CORBA::wstring_free(*ptr_);
    *ptr_ = 0;
    return *ptr_;
}

wchar_t*
OB::WStrForSeq::_retn()
{
    wchar_t* ret = *ptr_;
    *ptr_ = 0;
    return ret;
}

// ----------------------------------------------------------------------
// CORBA::String_var and OB::StrForSeq output on output stream
// ----------------------------------------------------------------------

ostream&
operator<<(ostream& out, const CORBA::String_var& s)
{
    return out << s.in();
}

ostream&
operator<<(ostream& out, const CORBA::String_out& s)
{
    return out << s;
}

ostream&
operator<<(ostream& out, const OB::StrForSeq& s)
{
    return out << s.in();
}

istream&
operator>>(istream& in, CORBA::String_var& s)
{
    s = CORBA::string_dup("");

    bool skip = true;
    while(true)
    {
	int c = in.get();
	if(!in || c == EOF)
	    break;
	if(isspace(c))
	{
	    if(!skip)
		break;
	    continue;
	}
	skip = false;
	s += (char)c;
    }
    return in;
}

istream&
operator>>(istream& in, CORBA::String_out& s)
{
    return in >> s.ptr();
}

istream&
operator>>(istream& in, OB::StrForSeq& s)
{
    return in >> s.inout();
}

// ----------------------------------------------------------------------
// CORBA::WString_var and OB::WStrForSeq output on output stream
// ----------------------------------------------------------------------

ostream&
operator<<(ostream& out, const CORBA::WString_var& s)
{
    return out << s.in();
}

ostream&
operator<<(ostream& out, const CORBA::WString_out& s)
{
    return out << s;
}

ostream&
operator<<(ostream& out, const OB::WStrForSeq& s)
{
    return out << s.in();
}

// TODO: This does currently not work with egcs 1.1.1 and Visual C++,
// since there is no operator>>(istream&, wchar_t*) available.
#ifndef HAVE_NO_WCHAR_T_SUPPORT_IN_ISTREAM

istream&
operator>>(istream& in, CORBA::WString_var& s)
{
    return in >> s.inout();
}

istream&
operator>>(istream& in, CORBA::WString_out& s)
{
    return in >> s.ptr();
}

istream&
operator>>(istream& in, OB::WStrForSeq& s)
{
    return in >> s.inout();
}

#endif
