// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/OBTypes.h>
#include <OB/CONV_FRAME.h>
#include <OB/CodeSetDatabase.h>
#include <OB/CodeSetUtil.h>
#include <OB/CodeConverter.h>

using namespace CORBA;
using namespace OB;

CodeConverterBase::CodeConverterBase(
    const CodeSetDatabase::CodeSetInfo* srcSet,
    const CodeSetDatabase::CodeSetInfo* dstSet)
    : srcSet_(srcSet), dstSet_(srcSet)
{
}

// ----------------------------------------------------------------------
// Read/write/count char/wchar (cross platform)
// ----------------------------------------------------------------------

CodeConverter::CodeConverter(
    const CodeSetDatabase::CodeSetInfo* srcSet,
    const CodeSetDatabase::CodeSetInfo* dstSet)
    : CodeConverterBase(srcSet, dstSet)
{
    CodeSetDatabase* db = CodeSetDatabase::instance();
    srcMap_ = db -> getCharMapInfo(srcSet);
    dstMap_ = db -> getCharMapInfo(dstSet);
    
//    cout << "Source code set: " << srcSet -> description << endl;
//    cout << "Destination code set: " << dstSet -> description << endl;
}

// ----------------------------------------------------------------------
// Read/write/count char/wchar
// ----------------------------------------------------------------------

Char
CodeConverter::read_char(const Octet*& in)
    throw(DATA_CONVERSION)
{
    // XXX Conversion missing
    return *in++;
}

void
CodeConverter::write_char(Octet*& out, Char value)
    throw(DATA_CONVERSION)
{
    // XXX Conversion missing
    *out++ = value;
}

ULong
CodeConverter::count_wchar(WChar value)
{
    ULong len;

    //
    // UTF-8 is a special case
    //
    if(getDest() -> rgy_value == CodeSetDatabase::UTF8)
    {
	if(value < 0x80)
	    len = 1;
	else if(value < 0x7ff)
	    len = 2;
	else if(value < 0xffff)
	    len = 3;
	else if(value < 0x1fffff)
	    len = 4;
	else if(value < 0x3ffffff)
	    len = 5;
	else
	    len = 6;
    }
    else
    {
	len = 1;

	//
	// Ignore most significant bytes as long as they are zero
	//
	for(ULong i = 1 ; i < sizeof(WChar) ; i++)
	{
	    value >>= 8;
	    if(value & 0xff)
		len = i + 1;
	}
    }

    return 1 + len;
}

WChar
CodeConverter::read_wchar(const Octet*& in)
    throw(DATA_CONVERSION)
{
    ULong len = *in++;

    //
    // Skip most significant bytes as long as they are zero
    //
    while(len > 1 && !*in)
    {
        in++;
        len--;
    }

    //
    // Remaining bytes must not exceed the local wide character size
    //
    if(!len || len > sizeof(WChar))
        throw DATA_CONVERSION();

    WChar v;

    switch(getSource() -> max_bytes)
    {

    case 1:
    {
	v = *in++;
	break;
    }

    case 2:
    {
	switch(getDest() -> max_bytes)
	{
	case 1:
	{
	    //
	    // High byte must be 0
	    //
	    if(*in++)
		throw DATA_CONVERSION();

	    v = *in++;
	    break;
	}

	case 2:
	case 4:
	{
	    v = (*in++ << 8) | *in++;
	    break;
	}

	default:
	    throw DATA_CONVERSION();
    }

    case 4:
    {
	switch(getDest() -> max_bytes)
	{

	case 1:
	{
	    //
	    // High bytes must be 0
	    //
	    if(*in++ || *in++ || *in++)
		throw DATA_CONVERSION();

	    v = *in++;
	    break;
	}

	case 2:
	{
	    //
	    // High bytes must be 0
	    //
	    if(*in++ || *in++)
		throw DATA_CONVERSION();

	    v = (*in++ << 8) | *in++;
	    break;
	}

	case 4:
	{
	    v = (*in++ << 24) | (*in++ << 16) | (*in++ < 8) | *in++;
	    break;
	}

	default:
	    throw DATA_CONVERSION();
	}
    }

    default:
	throw DATA_CONVERSION();
    }
    }

    return v;
}

void
CodeConverter::write_wchar(Octet*& out, WChar value)
    throw(DATA_CONVERSION)
{
    convert(value);

    if(getDest() -> rgy_value == CodeSetDatabase::UTF8)
	CodeSetUtil::unicodeToUtf8(out, value);
    else
    {
	ULong len = count_wchar(value) - 1;

	//
	// Check if the maximum output character length is exceeded
	//
	if(len > getDest() -> max_bytes)
	    throw DATA_CONVERSION();

	//
	// Actual wide character length
	//
	*out = (Octet)len;

	//
	// Write valid bytes
	//
	ULong i = len;
	while(i)
	{
	    out[i--] = (Octet)value;
	    value >>= 8;
	}

	out += len + 1;
    }
}

void
CodeConverter::convert(WChar& value)
{
    //
    // Convert to intermediate Unicode character and then to final character
    //
    if(value >= (WChar)srcMap_ -> lower_bound &&
       value <= (WChar)srcMap_ -> upper_bound)
    {
	value = srcMap_ -> map_values[value - srcMap_ -> lower_bound];
    }

    if(dstMap_ && getDest()-> rgy_value != CodeSetDatabase::UTF16)
    {
	for(ULong j = dstMap_ -> lower_bound ; j <= dstMap_ -> upper_bound ;
	    j++)
	{
	    if(dstMap_ -> map_values[j] == value)
	    {
		value = j;
		break;
	    }
	}
    }
}

// ----------------------------------------------------------------------
// Locally read/write char/wchar (no conversion necessary)
// ----------------------------------------------------------------------

OB::LocalCodeConverter* LocalCodeConverter::instance_ = 0;

LocalCodeConverter*
LocalCodeConverter::instance()
{
    if(!instance_)
	instance_ = new LocalCodeConverter();

    return instance_;
}

WChar
LocalCodeConverter::read_wchar(const Octet*& in)
    throw(DATA_CONVERSION)
{
    ULong len = *in++;
//    assert(len == sizeof(WChar));

    // XXX Temporarily for OB 4.0a1
    //
    // Skip most significant bytes as long as they are zero
    //
    while(len > 1 && !*in)
    {
        in++;
        len--;
    }

    WChar val = 0;
    for(ULong i = 0 ; i < len ; i++)
    {
        val <<= 8;
        val += *in++;
    }

    return val;
}

void
LocalCodeConverter::write_wchar(Octet*& out, WChar value)
    throw(DATA_CONVERSION)
{
    *out = sizeof(WChar);

    for(ULong i = sizeof(WChar) ; i > 0 ; i--)
    {
	out[i] = (Octet)value;
	value >>= 8;
    }

    out += sizeof(WChar) + 1;
}
