// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/TemplateI.h>
#include <OB/OBTypes.h>
#include <OB/CONV_FRAME.h>
#include <OB/CodeSetUtil.h>
#include <OB/CodeSetDatabase.h>
#include <OB/CodeConverter.h>

using namespace CORBA;
using namespace CONV_FRAME;
using namespace OB;

const ULong CodeSetDatabase::ISOLATIN1 = 0x00010001;
const ULong CodeSetDatabase::ISOLATIN2 = 0x00010002;
const ULong CodeSetDatabase::ISOLATIN3 = 0x00010003;
const ULong CodeSetDatabase::ISOLATIN4 = 0x00010004;
const ULong CodeSetDatabase::ISOLATIN5 = 0x00010005;
const ULong CodeSetDatabase::ISOLATIN6 = 0x00010006;
const ULong CodeSetDatabase::ISOLATIN7 = 0x00010007;
const ULong CodeSetDatabase::ISOLATIN8 = 0x00010008;
const ULong CodeSetDatabase::ISOLATIN9 = 0x00010009;
const ULong CodeSetDatabase::UTF8 = 0x05010001;
const ULong CodeSetDatabase::UTF16 = 0x00010109;

// ----------------------------------------------------------------------
// Template instantiations
// ----------------------------------------------------------------------
 
#ifndef HAVE_NO_EXPLICIT_TEMPLATES
template class OB::FixSeq< CodeConverterBase*, int >;
#else
#ifdef HAVE_PRAGMA_DEFINE
#pragma define(OB::FixSeq< CodeConverterBase*, int >)
#endif
#endif

namespace OB
{

CodeSetDatabase* CodeSetDatabase::instance_ = 0;

class CodeSetDatabaseDestroyer
{
public:

    ~CodeSetDatabaseDestroyer()
    {

	delete CodeSetDatabase::instance_;
	CodeSetDatabase::instance_ = 0;
    }
};

} // End of namespace OB

static CodeSetDatabaseDestroyer destroyer;

// ----------------------------------------------------------------------
// CodeSetDatabase public member implementation
// ----------------------------------------------------------------------

CodeSetDatabase::CodeSetDatabase()
{
    //
    // Add the supported code sets in the preferred order.
    //
    CodeSetUtil::supportedCodeSetsSeq_.append(UTF8);
    CodeSetUtil::supportedCodeSetsSeq_.append(UTF16);
    CodeSetUtil::supportedCodeSetsSeq_.append(ISOLATIN1);
    CodeSetUtil::supportedCodeSetsSeq_.append(ISOLATIN2);
    CodeSetUtil::supportedCodeSetsSeq_.append(ISOLATIN3);
    CodeSetUtil::supportedCodeSetsSeq_.append(ISOLATIN4);
    CodeSetUtil::supportedCodeSetsSeq_.append(ISOLATIN5);
    CodeSetUtil::supportedCodeSetsSeq_.append(ISOLATIN6);
    CodeSetUtil::supportedCodeSetsSeq_.append(ISOLATIN7);
    CodeSetUtil::supportedCodeSetsSeq_.append(ISOLATIN8);
    CodeSetUtil::supportedCodeSetsSeq_.append(ISOLATIN9);
}

CodeSetDatabase::~CodeSetDatabase()
{
    for(ULong i = 0 ; i < codeConverterSeq_.length() ; i++)
	delete codeConverterSeq_[0];
    codeConverterSeq_.length(0);
}

// ----------------------------------------------------------------------
// CodeSetDatabase public member implementation
// ----------------------------------------------------------------------

CodeSetDatabase*
CodeSetDatabase::instance()
{
    if(!instance_)
	instance_ = new CodeSetDatabase();

    return instance_;
}

CodeConverterBase*
CodeSetDatabase::getCharEncoder(const CodeSetComponent& dst, ULong native_cs)
{
    CodeSetComponent_var src = CodeSetUtil::createCodeSetComponent(native_cs);
    return getConverter(src, dst, UTF8);
}

CodeConverterBase*
CodeSetDatabase::getWCharEncoder(const CodeSetComponent& dst, ULong native_wcs)
{
    CodeSetComponent_var src = CodeSetUtil::createCodeSetComponent(native_wcs);
    return getConverter(src, dst, UTF16);
}

CodeConverterBase*
CodeSetDatabase::getCharDecoder(const CodeSetComponent& src, ULong native_cs)
{
    CodeSetComponent_var dst = CodeSetUtil::createCodeSetComponent(native_cs);
    return getConverter(src, dst, UTF8);
}

CodeConverterBase*
CodeSetDatabase::getWCharDecoder(const CodeSetComponent& src, ULong native_wcs)
{
    CodeSetComponent_var dst = CodeSetUtil::createCodeSetComponent(native_wcs);
    return getConverter(src, dst, UTF16);
}

CodeConverterBase*
CodeSetDatabase::getConverter(const CodeSetComponent& serverCS,
			      const CodeSetComponent& clientCS,
			      ULong fallback)
{
    ULong tcs;

    //
    // Check if native code sets are identical and we only have
    // to care for different wchar sizes
    //
    if(serverCS.native_code_set == clientCS.native_code_set)
	tcs = serverCS.native_code_set;
    else
	tcs = determineTCS(serverCS, clientCS, fallback);

    for(ULong i = 0 ; i < codeConverterSeq_.length() ; i++)
    {
	if(codeConverterSeq_[i] -> getDest() -> rgy_value == tcs &&
	   codeConverterSeq_[i] -> getSource() -> rgy_value ==
	   serverCS.native_code_set)
	    return codeConverterSeq_[i];
    }

    CodeConverter* converter =
	new CodeConverter(getCodeSetInfo(serverCS.native_code_set),
			  getCodeSetInfo(tcs));
    codeConverterSeq_.append(converter);

    return converter;
}

const CodeSetDatabase::CodeSetInfo*
CodeSetDatabase::getCodeSetInfo(ULong rgy_value)
{
    //
    // Check if code set id is listed in registry
    //
    for(ULong i = 0 ; i < codeSetInfoArraySize_ ; i++)
	if(codeSetInfoArray_[i].rgy_value == rgy_value)
	   return &codeSetInfoArray_[i];

    return 0;
}

ULong
CodeSetDatabase::determineTCS(const CodeSetComponent& serverCS,
			      const CodeSetComponent& clientCS,
			      ULong fallback)
{
    //
    // First check if the native code sets are identical
    // In case they are no conversion is required
    //
    if(clientCS.native_code_set == serverCS.native_code_set)
	return serverCS.native_code_set;

    //
    // Check if client can convert
    //
    if(checkCodeSetId(clientCS, serverCS.native_code_set))
	return serverCS.native_code_set;

    //
    // Check if server can convert
    //
    if(checkCodeSetId(serverCS, clientCS.native_code_set))
	return clientCS.native_code_set;

    //
    // Check for common code set that can be used for transmission
    // The server supported code sets have preference
    //
    for(ULong i = 0 ; i < serverCS.conversion_code_sets.length() ; i++)
    {
	if(checkCodeSetId(clientCS, serverCS.conversion_code_sets[i]))
	    return serverCS.conversion_code_sets[i];
    }

    //
    // Check compatibility by using the OSF registry,
    // use fallback code set if compatible
    //
    if(isCompatible(clientCS.native_code_set, serverCS.native_code_set))
	return fallback;

    // XXX TODO: What is CODESET_INCOMPATIBLE? Must be different exception.
//	throw CODESET_INCOMPATIBLE();
    throw DATA_CONVERSION();

    return 0;
}

// ----------------------------------------------------------------------
// Check code set compatibility
// ----------------------------------------------------------------------

bool
CodeSetDatabase::isCompatible(ULong id1, ULong id2)
{
    const CodeSetInfo* cs1 = getCodeSetInfo(id1);
    const CodeSetInfo* cs2 = getCodeSetInfo(id2);

    if(!cs1 || !cs2)
	return false;

    for(ULong i = 0 ; i < cs1 -> char_values_size ; i++)
    {
	for(ULong j = 0 ; j < cs2 -> char_values_size ; j++)
	{
	    //
	    // In order to be compatible in OFS terms two code sets must have
	    // a common character set
	    //
	    if(cs1 -> char_values[i] == cs2 -> char_values[j])
		return true;
	}
    }

    return false;
}

bool
CodeSetDatabase::checkCodeSetId(const CodeSetComponent& cs, ULong id)
{
    for(ULong i = 0 ; i < cs.conversion_code_sets.length() ; i++)
    {
	if(cs.conversion_code_sets[i] == id)
	    return true;
    }

    //
    // ID not found
    //
    return false;
}

ULong
CodeSetDatabase::nameToId(const char* name)
{
    assert(name != 0);

    //
    // Check if codeset name is listed in registry
    // Return first match so that shortcuts can be used
    //
    for(ULong i = 0 ; i < codeSetInfoArraySize_ ; i++)
    {
	if(strstr(codeSetInfoArray_[i].description, name) != 0)
	    return codeSetInfoArray_[i].rgy_value;
    }

    //
    // Name not found
    //
    return 0;
}

const char*
CodeSetDatabase::idToName(ULong rgy_value)
{
    //
    // Check if codeset ID is listed in registry
    //
    for(ULong i = 0 ; i < codeSetInfoArraySize_ ; i++)
	if(codeSetInfoArray_[i].rgy_value == rgy_value)
	    return codeSetInfoArray_[i].description;

    //
    // ID not found
    //
    return 0;
}

const CodeSetDatabase::CharMapInfo*
CodeSetDatabase::getCharMapInfo(const CodeSetInfo* codeSetInfo)
{
    const char* name = "";

    switch(codeSetInfo -> rgy_value)
    {
    case ISOLATIN1:
	name = "ISO 8859-1 (1987) to Unicode";
	break;

    case ISOLATIN2:
	name = "ISO 8859-2 (1987) to Unicode";
	break;

    case ISOLATIN3:
	name = "ISO 8859-3 (1988) to Unicode";
	break;

    case ISOLATIN4:
	name = "ISO 8859-4 (1988) to Unicode";
	break;

    case ISOLATIN5:
	name = "ISO 8859-5 (1988) to Unicode";
	break;

    case ISOLATIN6:
	name = "ISO 8859-6 (1987) to Unicode";
	break;

    case ISOLATIN7:
	name = "ISO 8859-7 (1987) to Unicode";
	break;

    case ISOLATIN8:
	name = "ISO 8859-8 (1988) to Unicode";
	break;

    case ISOLATIN9:
	name = "ISO 8859-9 (1989) to Unicode";
	break;

    default:
	break;
    }

    for(ULong i = 0 ; i < charMapInfoArraySize_ ; i++)
    {
	if(strcmp(charMapInfoArray_[i].name, name) == 0)
	    return &charMapInfoArray_[i];
    }

    return 0;
}
