// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/TemplateI.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/IOP.h>
#include <OB/IIOP.h>
#include <OB/Stream.h>
#include <OB/CONV_FRAME.h>
#include <OB/CodeSetDatabase.h>
#include <OB/CodeSetUtil.h>

#include <stdlib.h>

using namespace CORBA;
using namespace CONV_FRAME;
using namespace OB;

ULongSeq CodeSetUtil::supportedCodeSetsSeq_;

// ----------------------------------------------------------------------
// Create code set component and info for current code set settings
// ----------------------------------------------------------------------

CodeSetComponentInfo*
CodeSetUtil::createCodeSetComponentInfo(ULong id_char, ULong id_wchar)
{
    CodeSetComponentInfo* codeSetComponentInfo =
	new CodeSetComponentInfo();

    CodeSetComponent_var c = createCodeSetComponent(id_char);
    CodeSetComponent_var wc = createCodeSetComponent(id_wchar);
    codeSetComponentInfo -> ForCharData = c;
    codeSetComponentInfo -> ForWcharData = wc;

    return codeSetComponentInfo;
}

CodeSetComponent*
CodeSetUtil::createCodeSetComponent(ULong id)
{
    CodeSetDatabase::instance();

    CodeSetComponent* codeSetComponent = new CodeSetComponent();

    codeSetComponent -> native_code_set = id;

    //
    // Add conversion code sets, filter native code set
    //
    for(ULong i = 0 ; i < supportedCodeSetsSeq_.length() ; i++)
    {
	if(supportedCodeSetsSeq_[i] != id)
	    codeSetComponent -> conversion_code_sets.append(
		supportedCodeSetsSeq_[i]);
    }

    return codeSetComponent;
}

// ----------------------------------------------------------------------
// Extract code set component info from IOR
// ----------------------------------------------------------------------

CodeSetComponentInfo*
CodeSetUtil::getCodeSetComponentInfoFromIOR(const IOP::IOR* ior)
{
    for(ULong i = 0 ; i < ior -> profiles.length() ; i++)
    {
        if(ior -> profiles[i].tag == IOP::TAG_INTERNET_IOP)
        {
	    //
	    // Get the IIOP profile body
	    //
	    const Octet* coct = ior -> profiles[i].profile_data.get_buffer();
	    const Octet* coctSave = coct;
            ULong len = ior -> profiles[i].profile_data.length();
            InputStream in(coct, 0, len, false);
	    in._OB_readEndian();
	    ::IIOP::ProfileBody_1_0 body;
	    ::IIOP::ProfileBody_1_0::_OB_unmarshal(body, in);

	    //
	    // Handle IIOP 1.1 tagged components
	    //
	    if(body.iiop_version.minor > 0)
	    {
		InputStream in_1_1(coctSave, 0, len, false);
		in_1_1._OB_readEndian();
		::IIOP::ProfileBody_1_1 body_1_1;
		::IIOP::ProfileBody_1_1::_OB_unmarshal(body_1_1, in_1_1);

		for(ULong j = 0 ; j < body_1_1.components.length() ; j++)
		{
		    if(body_1_1.components[j].tag == IOP::TAG_CODE_SETS)
		    {
			const Octet* coct =
			    body_1_1.components[j].component_data.get_buffer();
			ULong len =
			    body_1_1.components[j].component_data.length();
			InputStream in(coct, 0, len, false);
			in._OB_readEndian();
			CodeSetComponentInfo* info = new CodeSetComponentInfo;
			CodeSetComponentInfo::_OB_unmarshal(*info, in);

			return info;
		    }
		}
	    }

	    //
	    // Set default code set for char if no code set was specified.
	    // XXX There is no default for wchar
	    //
	    CodeSetComponentInfo* info = new CodeSetComponentInfo;
	    info -> ForCharData.native_code_set = CodeSetDatabase::ISOLATIN1;
	    info -> ForWcharData.native_code_set = CodeSetDatabase::ISOLATIN1;

	    // XXX For debugging purposes only
/*
	    cout << "No code set information in IIOP profile, using defaults"
		 << endl;
*/

	    return info;
	}
    }

    return 0;
}

// ----------------------------------------------------------------------
// Create tagged component with code set information
// ----------------------------------------------------------------------

IOP::TaggedComponent*
CodeSetUtil::createTaggedComponentForCodeSets(ULong char_id, ULong wchar_id)
{
    IOP::TaggedComponent* comp = new IOP::TaggedComponent;
    CONV_FRAME::CodeSetComponentInfo_var info =
	createCodeSetComponentInfo(char_id, wchar_id);

    CountStream count;
    count.count_boolean(); // Endian
    CodeSetComponentInfo::_OB_marshalCount(info, count);
    comp -> component_data.length(count.count());
    comp -> tag = IOP::TAG_CODE_SETS;
    Octet* oct = comp -> component_data.get_buffer();
#ifdef OB_CLEAR_MEM
    memset(oct, 0, count.count());
#endif
    OutputStream out(oct, 0, count.count());
    out.write_boolean(Endian);
    CodeSetComponentInfo::_OB_marshal(info, out);

    return comp;
}

// ----------------------------------------------------------------------
// Convert between UTF-8 and Unicode
// ----------------------------------------------------------------------

void
CodeSetUtil::unicodeToUtf8(Octet*& out, WChar value)
    throw(DATA_CONVERSION)
{
    ULong len;

    //
    // Direct mapping (7 bits) for characters < 0x80
    //
    if(value < 0x80)
    {
	*out++ = (Octet)value;
	return;
    }
    else if(value < 0x7ff)
    {
        // 5 free bits (%110xxxxx | %vvvvv)
        *out = 0xc0;
	len = 2;
    }
    else if(value < 0xffff)
    {
        // 4 free bits (%1110xxxx | %vvvv)
        *out = 0xe0;
	len = 3;
    }
    else if(value < 0x1fffff)
    {
        // 3 free bits (%11110xxx | %vvv)
        *out = 0xf0;
	len = 4;
    }
    else if(value < 0x3ffffff)
    {
        // 2 free bits (%111110xx | %vv)
        *out = 0xf8;
	len = 5;
    }
    else if(value < 0x7fffffff)
    {
        // 1 free bit (%1111110x | %v)
        *out = 0xfc;
	len = 6;
    }
    else
	throw DATA_CONVERSION();

    //
    // Write continuation bytes (6 free bits each)
    //
    ULong cont = len - 1;
    while(cont)
    {
	out[cont] = 0x80 | (value & 0x3f);
	value >>= 6;
	cont--;
    }

    //
    // Write the rest
    //
    *out |= value;

    out += len;
}

WChar
CodeSetUtil::utf8ToUnicode(Octet*& in)
    throw(DATA_CONVERSION)
{
    Octet first = *in++;
 
    //
    // Direct mapping for characters < 0x80
    //
    if((first & 0x80) == 0)
	return first;

    int rem;
    WChar value;

    if((first & 0xe0) == 0xc0)
    {
	// 5 free bits
	rem = 1;
	value = first & 0x1f;
    }
    else if((first & 0xf0) == 0xe0)
    {
	// 4 free bits
	rem = 2;
	value = first & 0x0f;
    }
    else if((first & 0xf8) == 0xf0)
    {
	// 3 free bits;
	rem = 3;
	value = first & 0x07;
    }
    else if((first & 0xfc) == 0xf8)
    {
	// 2 free bits
	rem = 4;
	value = first & 0x03;
    }
    else if((first & 0xfe) == 0xfc)
    {
	// 1 free bit
	rem = 5;
	value = first & 0x01;
    }
    else
	throw DATA_CONVERSION();

    //
    // Check for 16 bit overflow
    //
    if(sizeof(WChar) < 4 && rem >= 3)
	throw DATA_CONVERSION();

    //
    // Get continuation bytes (6 free bits)
    //
    while(rem)
    {
	if((*in & 0xc0) != 0x80)
	    throw DATA_CONVERSION();

	value <<= 6;
	value |= *in++ & 0x3f;
	rem--;
    }

    return value;
}
