// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/Any.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/IFR1.h>
#include <OB/OBV.h>
#include <OB/Policy.h>
#include <OB/ORB.h>
#include <OB/GIOP.h>
#include <OB/OCI.h>
#include <OB/Hashtable.h>
#include <OB/HashtableI.h>
#include <OB/Hashers.h>
#include <OB/Reactor.h>
#include <OB/Stream.h>
#include <OB/CONV_FRAME.h>
#include <OB/CodeSetDatabase.h>
#include <OB/CodeConverter.h>

#include <Downcall.h>
#include <CollocatedClient.h>
#include <Upcall.h>
#include <OARoot.h>

using namespace CORBA;
using namespace OB;

//
// XXX
//
// This magic 4 offset is a little strange. The problem is that
// in a local environment the GIOP headers are not needed. However,
// some of the core code expects a non-zero offset (see DII), and
// the poll API method returns 0 to indicate a timeout.
//
// what deferred/response and poll do in an MT environment should
// depend on the server side concurrency model.
//
// This class also isn't correct, since it uses the OARoot::instance
// to dispatch method requests and NOT the Upcall object for the POA
// root defined by the object_key of the object reference.
//

namespace OB
{

//
// The CollocatedDowncall class
//
// This class is used internally to the collocated client class to
// receive downcalls. The reason we have this class, and not make
// CollocatedClient inherit from Downcall directly is that both
// Downcall and Client inherit from RefCount. I didn't want to
// make this inheritance virtual, so this separate class evolved.
//
class CollocatedDowncall : public Downcall
{
    CollocatedClient::RequestMap* replyMap_;

    void addReply(ULong, Buffer& buf, GIOP::ReplyStatusType);

public:
    CollocatedDowncall();
    virtual ~CollocatedDowncall();

    static inline CollocatedDowncall_ptr
    _duplicate(CollocatedDowncall_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline CollocatedDowncall_ptr _nil()
    { return 0; }
        
    CollocatedClient::Request* findReply(ULong);
    
    //
    // Downcall methods
    //
    void response(ULong, Buffer&);
    void systemException(ULong, Buffer&,
                         const SystemException&);
    void userException(ULong, Buffer&);
    void forward(ULong, Buffer&, Object_ptr);
};

} // End of namespace OB


// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::CollocatedDowncall_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::CollocatedDowncall_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// CollocatedDowncall private/protected member implementation
// ----------------------------------------------------------------------

void
CollocatedDowncall::addReply(ULong id, Buffer& buf,
                             GIOP::ReplyStatusType status)
{
    CollocatedClient::Request* req = new CollocatedClient::Request;
    req -> buf.consume(buf);
    req -> reqId = id;
    req -> status = GIOP::NO_EXCEPTION;

#ifdef HAVE_JTC
    JTCSynchronized sync(*replyMap_);
#endif

    replyMap_ -> put(id, req);
}

// ----------------------------------------------------------------------
// CollocatedDowncall constructor/destructor
// ----------------------------------------------------------------------

CollocatedDowncall::CollocatedDowncall()
    : replyMap_(new CollocatedClient::RequestMap(63))
{
}

CollocatedDowncall::~CollocatedDowncall()
{
    delete replyMap_;
}

// ----------------------------------------------------------------------
// CollocatedDowncall public member implementation
// ----------------------------------------------------------------------
CollocatedClient::Request*
CollocatedDowncall::findReply(ULong id)
{
#ifdef HAVE_JTC
    JTCSynchronized sync(*replyMap_);
#endif
    
    CollocatedClient::Request* r;
    if(replyMap_ -> get(id, r))
        return r;
    
    return 0;
}

void
CollocatedDowncall::response(ULong id, Buffer& buf)
{
    addReply(id, buf, GIOP::NO_EXCEPTION);
}

void
CollocatedDowncall::systemException(ULong id, Buffer& buf,
                                    const SystemException& ex)
{
    CountStream count(4);
    SystemException::_OB_marshalCount(ex, count);
    buf.alloc(count.count());
    OutputStream out(buf, 4);
    SystemException::_OB_marshal(ex, out);

    addReply(id, buf, GIOP::SYSTEM_EXCEPTION);
}

void
CollocatedDowncall::userException(ULong id, Buffer& buf)
{
    addReply(id, buf, GIOP::USER_EXCEPTION);
}

void
CollocatedDowncall::forward(ULong id, Buffer& buf, Object_ptr o)
{
    CountStream count(4);
    count.count_Object(o);
    buf.alloc(count.count());
    OutputStream out(buf, 4);
    out.write_Object(o);

    addReply(id, buf, GIOP::LOCATION_FORWARD);
}


// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::CollocatedClient_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::CollocatedClient_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// Helper operations for the hashtable
// ----------------------------------------------------------------------

ULong
OBHash(ULong l)
{
    return l;
}

bool
OBHashCompare(ULong id1, ULong id2)
{
    return id1 == id2;
}

// ----------------------------------------------------------------------
// Template instantiations
// ----------------------------------------------------------------------

#ifndef HAVE_NO_EXPLICIT_TEMPLATES
template class OB::ObjVar< CollocatedClient >;
template class OB::ObjVar< CollocatedDowncall >;
template class OB::ObjForSeq< CollocatedClient >;
template class OB::Hashtable< ULong, CollocatedClient::Request* >;
#else
#ifdef HAVE_PRAGMA_DEFINE
#pragma define(OB::ObjVar< CollocatedClient >)
#pragma define(OB::ObjForSeq< CollocatedClient >)
//XXX
#endif
#endif

// ----------------------------------------------------------------------
// CollocatedClient private/protected member implementation
// ----------------------------------------------------------------------

//
// XXX - good candiate for InterlockedIncrement under WIN32.
//
ULong
CollocatedClient::nextId()
{
#ifdef HAVE_JTC
    JTCSynchronized sync(idMutex_);
#endif
    return id_++;
}

ULong
CollocatedClient::consumeReply(Request* reply,
                               Buffer& buf,
                               bool& exc,
                               bool& fwd)
{
    GIOP::ReplyStatusType status = (GIOP::ReplyStatusType)reply -> status;
    buf.consume(reply -> buf);

    //
    //XXX we should have a cache of these to use, instead of
    //new/delete them each time.
    //
    delete reply;
    
    switch(status)
    {
    case GIOP::NO_EXCEPTION:
        exc = false;
	fwd = false;
        break;
        
    case GIOP::USER_EXCEPTION:
        exc = true;
        fwd = false;
        break;
        
    case GIOP::SYSTEM_EXCEPTION:
    {
        InputStream in(buf, 4, false);
        SystemException::_OB_unmarshalThrow(in);
        break;
    }
    
    case GIOP::LOCATION_FORWARD:
        exc = false;
        fwd = true;
        break;
    }
    
    return 4;
}

CollocatedClient::Request*
CollocatedClient::findReply(ULong id)
{
    for(;;)
    {
        Request* reply = downcall_ -> findReply(id);
        if(reply == 0 && ORB::server_conc_model() ==
           ORB::ServerConcModelReactive)
        {
            Reactor::instance() -> dispatchOneEvent(-1);
            continue;
        }
        return reply;
    }
}

// ----------------------------------------------------------------------
// CollocatedClient constructor and destructor
// ----------------------------------------------------------------------

CollocatedClient::CollocatedClient()
    : id_(0),
      downcall_(new CollocatedDowncall),
      requestMap_(new RequestMap(1023))
{
    //
    // Local converter is sufficient here
    //
    CodeConverters codeConverters;
    codeConverters.charConverter = LocalCodeConverter::instance();
    codeConverters.wcharConverter = LocalCodeConverter::instance();
    setCodeConverters(codeConverters);
}

CollocatedClient::~CollocatedClient()
{
    RequestMap::Enumerator keys = requestMap_ -> keys();
    for(ULong i = 0; i < keys.length(); ++i)
    {
	Request* req;
	requestMap_ -> get(keys[i], req);
	delete req;
    }
    requestMap_ -> clear();
    delete requestMap_;
}

// ----------------------------------------------------------------------
// CollocatedClient public member implementation
// ----------------------------------------------------------------------

//
// Destroy the client
//
void
CollocatedClient::destroy()
{
}

//
// Get the OCI Connector
//
OCI::Connector_ptr
CollocatedClient::connector()
{
    return OCI::Connector::_nil();
}

//
// Get the OCI Transport
//
OCI::Transport_ptr
CollocatedClient::transport()
{
    return OCI::Transport::_nil();
}
    
//
// Calculate offset
//
ULong
CollocatedClient::offset(StubImplBase*, const char*)
{
    return 0;
}

//
// Send request and receive response
// With explicit timeout (throws exception on timeout)
// With optional retry on communication failure
//
ULong
CollocatedClient::request(StubImplBase* p,
                          const char* op,
                          Buffer& buf,
                          bool& swap,
                          bool& exc,
                          bool& fwd,
                          Long t,
                          bool retry)
{
    swap = false;
    
    //
    // Extract the key
    //
    const OctetSeq& key = p -> _OB_POAkey();
    assert(key.length() != 0);

    ULong offsetIn = 0;
    ULong offsetOut = 4; // We'll create a fake offset

    ULong requestId = nextId();
    OARoot::instance() -> invoke(downcall_, requestId,
                                 key, op, buf, false,
                                 offsetIn, offsetOut,
                                 OCI::TransportInfo::_nil());
    Request* reply = findReply(requestId);
    if(reply == 0)
    {
        // XXX
        assert(false);
    }

    return consumeReply(reply, buf, exc, fwd);
}

//
// Send request oneway
// With implicit zero timeout (does not throw exception on timeout)
// With optional retry on communication failure
//
void
CollocatedClient::oneway(StubImplBase* p,
                         const char* op,
                         Buffer& buf,
                         bool retry)
{
    //
    // Extract the key
    //
    const OctetSeq& key = p -> _OB_POAkey();
    assert(key.length() != 0);

    ULong offsetIn = 0;
    ULong offsetOut = 0;

    OARoot::instance() -> invoke(Downcall::_nil(), nextId(),
                                 key, op, buf, false,
                                 offsetIn, offsetOut,
                                 OCI::TransportInfo::_nil());
    return;
}

//
// Send request deferred
// With implicit zero timeout (does not throw exception on timeout)
// With optional retry on communication failure
//
void
CollocatedClient::deferred(StubImplBase* p,
                           const char* op,
                           Buffer& buf,
                           ULong& reqId,
                           bool retry)
{
#ifdef HAVE_JTC
    JTCSynchronized sync(*requestMap_);
#endif
    reqId = nextId();
    
    Request* req = new Request;
    req -> stub = StubImplBase::_duplicate(p);
    req -> op = string_dup(op);
    req -> buf.alloc(buf.len);
    memcpy(req -> buf.data, buf.data, buf.len);
    req -> reqId = reqId;
    
    requestMap_ -> put(reqId, req);
}

//
// Get repsonse
// With explicit timeout (throws exception on timeout)
//
ULong
CollocatedClient::response(ULong reqId,
                           Buffer& buf,
                           bool& swap,
                           bool& exc,
                           bool& fwd,
                           Long t)
{
    Request* req;
    {
#ifdef HAVE_JTC
        JTCSynchronized sync(*requestMap_);
#endif
        if(!requestMap_ -> get(reqId, req))
            throw NO_RESPONSE();
        requestMap_ -> remove(reqId);
    }
    
    swap = false;
    
    //
    // Extract the key
    //
    //const OBFixSeq< Octet, int >& key = req -> obj -> _OB_POAkey();
    const OctetSeq& key = req -> stub -> _OB_POAkey();
    assert(key.length() != 0);

    ULong offsetIn = 0;
    ULong offsetOut = 4;

    ULong requestId = nextId();
    
    OARoot::instance() -> invoke(downcall_, requestId,
                                 key, req -> op,
                                 req -> buf,
                                 false,
                                 offsetIn, offsetOut,
                                 OCI::TransportInfo::_nil());

    Request* reply = findReply(requestId);
    if(reply == 0)
    {
        // XXX
        assert(false);
    }

    return consumeReply(reply, buf, exc, fwd);
}
    
//
// Poll response
// With implicit zero timeout (returns zero on timeout)
//
ULong
CollocatedClient::poll(ULong reqId,
                       Buffer& buf,
                       bool& swap,
                       bool& exc,
                       bool& fwd)
{
    Request* req;
    {
#ifdef HAVE_JTC
        JTCSynchronized sync(*requestMap_);
#endif
        if(!requestMap_ -> get(reqId, req))
            throw NO_RESPONSE();
        requestMap_ -> remove(reqId);
    }
    
    swap = false;
    
    //
    // Extract the key
    //
    //const OBFixSeq< Octet, int >& key = req -> obj -> _OB_POAkey();
    const OctetSeq& key = req -> stub -> _OB_POAkey();
    assert(key.length() != 0);

    ULong offsetIn = 0;
    ULong offsetOut = 4;

    ULong requestId = nextId();
    
    OARoot::instance() -> invoke(downcall_, requestId,
                                 key, req -> op,
                                 req -> buf,
                                 false,
                                 offsetIn, offsetOut,
                                 OCI::TransportInfo::_nil());

    Request* reply = findReply(requestId);
    if(reply == 0)
    {
        //XXX
        assert(false);
    }

    return consumeReply(reply, buf, exc, fwd);
}

