// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/TemplateI.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/Any.h>
#include <OB/Context.h>
#include <OB/NamedValue.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(CORBA::Context_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(CORBA::Context_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// Template instantiations
// ----------------------------------------------------------------------

#ifndef HAVE_NO_EXPLICIT_TEMPLATES
template class OB::ObjVar< Context >;
template class OB::ObjOut< Context >;
//template class OB::ObjForSeq< Context >;
//template class OB::ObjSeq< Context >;
//template class OB::SeqVar< OB::ObjSeq< Context > >;
//template class OB::SeqOut< OB::ObjSeq< Context > >;
#else
#ifdef HAVE_PRAGMA_DEFINE
#pragma define(OB::ObjVar< Context >)
#pragma define(OB::ObjOut< Context >)
//#pragma define(OB::ObjForSeq< Context >)
//#pragma define(OB::ObjSeq< Context >)
//#pragma define(OB::SeqVar< OB::ObjSeq< Context > >)
//#pragma define(OB::SeqOut< OB::ObjSeq< Context > >)
#endif
#endif

// ----------------------------------------------------------------------
// Context constructor and destructor
// ----------------------------------------------------------------------

Context::Context(const char* name)
    : name_(name),
      values_(new StringSeq)
{
}

Context::Context(const char* name, Context_ptr parent)
    : name_(name),
      parent_(Context::_duplicate(parent)),
      values_(new StringSeq)
{
}

// ----------------------------------------------------------------------
// Context public member implementation
// ----------------------------------------------------------------------

void
Context::create_child(const char* child_ctx_name, Context_out child_ctx)
{
    assert_nca(child_ctx_name, NCANullString);
    child_ctx = new Context(child_ctx_name, this);
}

void
Context::set_one_value(const char* propname, const Any& propvalue)
{
    assert_nca(propname, NCANullString);

    char* s;
    if(!(propvalue >>= s))
	throw BAD_TYPECODE();

    //
    // Search propname
    //
    for(ULong i = 0 ; i < values_ -> length() ; i += 2)
    {
	if(strcmp(propname, values_[i]) == 0)
	{
	    values_[i + 1] = string_dup(s);
	    return;
	}
    }

    //
    // Not found, add new propname
    //
    ULong len = values_ -> length();
    values_ -> length(len + 2);
    values_[len++] = propname;
    values_[len++] = string_dup(s);
}

void
Context::set_values(NVList_ptr values)
{
    //
    // Create new, empty list
    //
    values_ = new StringSeq(values -> count() * 2);
    values_ -> length(values -> count() * 2);
    
    //
    // Copy named values
    //
    for(ULong i = 0 ; i < values -> count() ; i++)
    {
	NamedValue_ptr nv = values -> item(i);
	
	char* s;
	if(!(*(nv -> value()) >>= s))
	    throw BAD_TYPECODE();
	
	if(nv -> flags() != 0)
	    throw INV_FLAG();
	

	values_[i * 2] = nv -> name();
	values_[i * 2 + 1] = string_dup(s);
    }
}

void
Context::delete_values(const char* pattern)
{
    assert_nca(pattern, NCANullString);
    
    //
    // Match the pattern
    //
    char last = 0;
    if(strlen(pattern) > 0)
	last = pattern[strlen(pattern) - 1];

    for(ULong i = 0 ; i < values_ -> length() ; i += 2)
    {
	bool match = false;

	if(last == '*')
	{
	    //
	    // Wildcard match ?
	    //
	    if(strncmp(pattern, values_[i], strlen(pattern) - 1) == 0)
		match = true;
	}
	else
	{
	    //
	    // Regular match ?
	    //
	    if(strcmp(pattern, values_[i]) == 0)
		match = true;
	}

	if(match)
	{
	    values_ -> remove(i);
	    values_ -> remove(i);
	    i -= 2;
	}
    }
}

void
Context::get_values(const char* start_scope, Flags op_flags,
		    const char* pattern, NVList_out values)
{
    assert_nca(start_scope, NCANullString);
    assert_nca(pattern, NCANullString);

    StringSeq seq;
    _OB_getValues(start_scope, op_flags, pattern, seq);

    values = new NVList;

    for(ULong i = 0 ; i < seq.length() ; i += 2)
    {
	Any* any = new Any;
	(*any) <<= seq[i + 1];
	values -> add_value_consume(seq[i]._retn(), any, 0);
    }
}

void
Context::_OB_getValues(const char* start_scope, Flags op_flags,
		       const char* pattern, StringSeq& seq)
{
    //
    // Don't do anything in this context if start_scope doesn't match name_
    //
    if(strlen(start_scope) && strcmp(start_scope, name_) != 0)
    {
	if(is_nil(parent_))
	    throw BAD_CONTEXT();
	
	parent_ -> _OB_getValues(start_scope, op_flags, pattern, seq);
	return;
    }

    //
    // If there is a parent and scope is not restricted, get values
    // from parent. Else create new value list
    //
    if(op_flags != CTX_RESTRICT_SCOPE && !is_nil(parent_))
	parent_ -> _OB_getValues("", op_flags, pattern, seq);

    //
    // Match the pattern
    //
    char last = 0;
    if(strlen(pattern) > 0)
	last = pattern[strlen(pattern) - 1];
    
    for(ULong i = 0 ; i < values_ -> length() ; i += 2)
    {
	bool match = false;
	
	if(last == '*')
	{
	    //
	    // Wildcard match ?
	    //
	    if(strncmp(pattern, values_[i], strlen(pattern) - 1) == 0)
		match = true;
	}
	else
	{
	    //
	    // Regular match ?
	    //
	    if(strcmp(pattern, values_[i]) == 0)
		match = true;
	}
	
	if(match)
	{
	    //
	    // First try to put replace value
	    //
	    ULong j;
	    for(j = 0 ; j < seq.length() ; j += 2)
		if(strcmp(seq[j], values_[i]) == 0)
		{
		    seq[j + 1] = values_[i + 1];
		    break;
		}

	    //
	    // Value not found, create new
	    //
	    if(j == seq.length())
	    {
		seq.length(j + 2);
		seq[j] = values_[i];
		seq[j + 1] = values_[i + 1];
	    }
	}
    }
}

Context*
Context::_OB_create(StringSeq* seq)
{
    Context* ctx = new Context("");
    ctx -> values_ = seq;
    return ctx;
}
