// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

//
// Author: Philippe Merle
//
// From the Laboratoire d'Informatique Fondamentale de Lille in France
// (it is a computer science laboratory of the Lille french university)
//

#ifndef OB_DYN_ANY_IMPL_H
#define OB_DYN_ANY_IMPL_H

#include <OB/Fixed.h>

namespace CORBA
{

// ----------------------------------------------------------------------
//
// The DynAny implementation class
//
// ----------------------------------------------------------------------
class DynAny_impl : virtual public DynAny
{
    //
    // Hide copy-constructor and assignment operator
    //
    DynAny_impl(const DynAny_impl&);
    void operator=(const DynAny_impl&);

protected: // internal state

    TypeCode_var type_;           // The TypeCode

protected: // constructor and destructor

    DynAny_impl(TypeCode_ptr);
    virtual ~DynAny_impl();

    // create a DynAny_impl from a TypeCode
    static DynAny_ptr create (TypeCode_ptr typecode);

    // The ORB calls the previous function
    friend class ORB;

public: // operations to access components

    virtual Any* current_any () =0;
    virtual Any* current_any_value () =0;

public: // operations to marshal and unmarshal components
        // into a complex any value

    virtual void MarshalCount(OB::CountStream&) =0;
    virtual void Marshal(OB::OutputStream&) =0;
    virtual void Unmarshal(OB::InputStream&) =0;

public: // public IDL operations

    //
    // Standard IDL to C++ Mapping
    // from DynAny interface
    //

    TypeCode_ptr type()
        throw(SystemException);

    void insert_boolean(Boolean value)
        throw(DynAny::InvalidValue,
              SystemException);
    void insert_octet(Octet value)
        throw(DynAny::InvalidValue,
              SystemException);
    void insert_char(Char value)
        throw(DynAny::InvalidValue,
              SystemException);
    void insert_wchar(WChar value)
        throw(DynAny::InvalidValue,
              SystemException);
    void insert_short(Short value)
        throw(DynAny::InvalidValue,
              SystemException);
    void insert_ushort(UShort value)
        throw(DynAny::InvalidValue,
              SystemException);
    void insert_long(Long value)
        throw(DynAny::InvalidValue,
              SystemException);
    void insert_ulong(ULong value)
        throw(DynAny::InvalidValue,
              SystemException);
    void insert_longlong(LongLong value)
        throw(DynAny::InvalidValue,
              SystemException);
    void insert_ulonglong(ULongLong value)
        throw(DynAny::InvalidValue,
              SystemException);
    void insert_float(Float value)
        throw(DynAny::InvalidValue,
              SystemException);
    void insert_double(Double value)
        throw(DynAny::InvalidValue,
              SystemException);
    void insert_longdouble(LongDouble value)
        throw(DynAny::InvalidValue,
              SystemException);
    void insert_string(const char* value)
        throw(DynAny::InvalidValue,
              SystemException);
    void insert_wstring(const wchar_t* value)
        throw(DynAny::InvalidValue,
              SystemException);
    void insert_reference(Object_ptr value)
        throw(DynAny::InvalidValue,
              SystemException);
    void insert_typecode(TypeCode_ptr value)
        throw(DynAny::InvalidValue,
              SystemException);
    void insert_any(const Any& value)
        throw(DynAny::InvalidValue,
              SystemException);
    void insert_valuetype(ValueBase* value)
        throw(DynAny::InvalidValue,
              SystemException);

    Boolean get_boolean()
        throw(DynAny::TypeMismatch,
              SystemException);
    Octet get_octet()
        throw(DynAny::TypeMismatch,
              SystemException);
    Char get_char()
        throw(DynAny::TypeMismatch,
              SystemException);
    WChar get_wchar()
        throw(DynAny::TypeMismatch,
              SystemException);
    Short get_short()
        throw(DynAny::TypeMismatch,
              SystemException);
    UShort get_ushort()
        throw(DynAny::TypeMismatch,
              SystemException);
    Long get_long()
        throw(DynAny::TypeMismatch,
              SystemException);
    ULong get_ulong()
        throw(DynAny::TypeMismatch,
              SystemException);
    LongLong get_longlong()
        throw(DynAny::TypeMismatch,
              SystemException);
    ULongLong get_ulonglong()
        throw(DynAny::TypeMismatch,
              SystemException);
    Float get_float()
        throw(DynAny::TypeMismatch,
              SystemException);
    Double get_double()
        throw(DynAny::TypeMismatch,
              SystemException);
    LongDouble get_longdouble()
        throw(DynAny::TypeMismatch,
              SystemException);
    char* get_string()
        throw(DynAny::TypeMismatch,
              SystemException);
    wchar_t* get_wstring()
        throw(DynAny::TypeMismatch,
              SystemException);
    Object_ptr get_reference()
        throw(DynAny::TypeMismatch,
              SystemException);
    TypeCode_ptr get_typecode()
        throw(DynAny::TypeMismatch,
              SystemException);
    Any* get_any()
        throw(DynAny::TypeMismatch,
              SystemException);
    ValueBase* get_valuetype()
        throw(DynAny::TypeMismatch,
              SystemException);

    virtual void destroy()
        throw(SystemException);

    // these operations are defined into subclasses
    //
    virtual void assign(DynAny_ptr dyn_any)
        throw(DynAny::Invalid,
              SystemException)=0;
    
    virtual void from_any(const Any& value)
        throw(DynAny::Invalid,
              SystemException)=0;
    
    virtual Any* to_any()
        throw(DynAny::Invalid,
              SystemException)=0;
    
    virtual DynAny_ptr copy()
        throw(SystemException) =0;
              
    virtual DynAny_ptr current_component()
        throw(SystemException) =0;
    
    virtual Boolean next()
        throw(SystemException) =0;
    
    virtual Boolean seek(Long index)
        throw(SystemException)=0;

    virtual void rewind()
        throw(SystemException) =0;
};

// ----------------------------------------------------------------------
//
// The DynBasic implementation class
//
// ----------------------------------------------------------------------
class DynBasic_impl : virtual public DynAny_impl
{
    //
    // Hide copy-constructor and assignment operator
    //
    DynBasic_impl(const DynBasic_impl&);
    void operator=(const DynBasic_impl&);

protected: // internal state

    Any_var any_; // The Any value for basic TypeCodes

protected: // constructor and destructor

    // The ORB and DynAny_impl may create DynBasic_impls
    friend class ORB;
    friend class DynAny_impl;

    DynBasic_impl(TypeCode_ptr);
    virtual ~DynBasic_impl();

protected: // operations to access components

   Any* current_any ();
   Any* current_any_value ();

protected: // operations to marshal and unmarshal components
           // into a complex any value

   void MarshalCount(OB::CountStream&);
   void Marshal(OB::OutputStream&);
   void Unmarshal(OB::InputStream&);

public: // public IDL operations

    //
    // Standard IDL to C++ Mapping
    // from DynAny interface
    //
    void assign(DynAny_ptr dyn_any)
        throw(DynAny::Invalid, SystemException);
    
    void from_any(const Any& value)
        throw(DynAny::Invalid, SystemException);
    
    Any* to_any()
        throw(DynAny::Invalid, SystemException);
    
    DynAny_ptr copy()
        throw(SystemException);

    DynAny_ptr current_component()
        throw(SystemException);
    Boolean next()
        throw(SystemException);
    Boolean seek(Long index)
        throw(SystemException);
    void rewind()
        throw(SystemException);
};

// ----------------------------------------------------------------------
//
// The DynFixed implementation class
//
// ----------------------------------------------------------------------
class DynFixed_impl : virtual public DynFixed,
		      virtual public DynAny_impl
{
    //
    // Hide copy-constructor and assignment operator
    //
    DynFixed_impl(const DynFixed_impl&);
    void operator=(const DynFixed_impl&);

protected: // internal state

    Fixed value_; // the Fixed value

protected: // constructor and destructor

    // The ORB and DynAny_impl may create DynFixed_impls
    friend class ORB;
    friend class DynAny_impl;

    DynFixed_impl(TypeCode_ptr, Fixed value = (Long)0);
    virtual ~DynFixed_impl();

protected: // operations to access components

   Any* current_any ();
   Any* current_any_value ();

protected: // operations to marshal and unmarshal components
           // into a complex any value

   void MarshalCount(OB::CountStream&);
   void Marshal(OB::OutputStream&);
   void Unmarshal(OB::InputStream&);

public: // public IDL operations

    //
    // Standard IDL to C++ Mapping
    // from DynAny interface
    //
    void assign(DynAny_ptr dyn_any)
        throw(DynAny::Invalid,
              SystemException);
    
    void from_any(const Any& value)
        throw(DynAny::Invalid, SystemException);
    
    Any* to_any()
        throw(DynAny::Invalid, SystemException);
    
    DynAny_ptr copy()
        throw(SystemException);

    DynAny_ptr current_component()
        throw(SystemException);
    
    Boolean next()
        throw(SystemException);
    
    Boolean seek(Long index)
        throw(SystemException);
    
    void rewind()
        throw(SystemException);

    //
    // Standard IDL to C++ Mapping
    // from DynFixed interface
    //
    DynAny::OctetSeq* get_value()
        throw(SystemException);
    void set_value(const DynAny::OctetSeq&)
        throw(DynAny::InvalidValue,
              SystemException);
};

// ----------------------------------------------------------------------
//
// The DynEnum implementation class
//
// ----------------------------------------------------------------------
class DynEnum_impl : virtual public DynEnum,
		     virtual public DynAny_impl
{
    //
    // Hide copy-constructor and assignment operator
    //
    DynEnum_impl(const DynEnum_impl&);
    void operator=(const DynEnum_impl&);

protected: // internal state

    ULong value_; // the enum value

protected: // constructor and destructor

    // The ORB and DynAny_impl may create DynEnum_impls
    friend class ORB;
    friend class DynAny_impl;

    DynEnum_impl(TypeCode_ptr, ULong value = 0);
    virtual ~DynEnum_impl();

protected: // operations to access components

   Any* current_any ();
   Any* current_any_value ();

protected: // operations to marshal and unmarshal components
           // into a complex any value

   void MarshalCount(OB::CountStream&);
   void Marshal(OB::OutputStream&);
   void Unmarshal(OB::InputStream&);

public: // public IDL operations

    //
    // Standard IDL to C++ Mapping
    // from DynAny interface
    //
    void assign(DynAny_ptr dyn_any)
        throw(DynAny::Invalid,
              SystemException);
    void from_any(const Any& value)
        throw(DynAny::Invalid, SystemException);    

    Any* to_any()
        throw(DynAny::Invalid, SystemException);

    DynAny_ptr copy()
        throw(SystemException);

    DynAny_ptr current_component()
        throw(SystemException);
    Boolean next()
        throw(SystemException);
    Boolean seek(Long index)
        throw(SystemException);
    void rewind()
        throw(SystemException);

    //
    // Standard IDL to C++ Mapping
    // from DynEnum interface
    //
    char* value_as_string()
        throw(SystemException);
    void value_as_string(const char*)
        throw(SystemException);
    ULong value_as_ulong()
        throw(SystemException);
    void value_as_ulong(ULong)
        throw(SystemException);
};

// ----------------------------------------------------------------------
//
// The DynComplex implementation class
//
// ----------------------------------------------------------------------
class DynComplex_impl : virtual public DynAny_impl
{
    //
    // Hide copy-constructor and assignment operator
    //
    DynComplex_impl(const DynComplex_impl&);
    void operator=(const DynComplex_impl&);

protected: // internal state

    DynAnySeq components_; // sequence of components
    ULong index_;          // current component

protected: // constructor and destructor

    DynComplex_impl(TypeCode_ptr);
    virtual ~DynComplex_impl();

protected: // operation to init components sequence

   void init_components(ULong);

protected: // operations to access components

   Any* current_any ();
   Any* current_any_value ();

protected: // operations to marshal and unmarshal components
           // into a complex any value

   void MarshalCount(OB::CountStream&);
   void Marshal(OB::OutputStream&);
   void Unmarshal(OB::InputStream&);

public: // public IDL operations

    //
    // Standard IDL to C++ Mapping
    // from DynAny interface
    //
    void assign(DynAny_ptr dyn_any)
        throw(DynAny::Invalid,
              SystemException);

    void from_any(const Any& value)
        throw(DynAny::Invalid, SystemException);

    Any* to_any()
        throw(DynAny::Invalid, SystemException);

    DynAny_ptr current_component()
        throw(SystemException);
    
    Boolean next()
        throw(SystemException);
    
    Boolean seek(Long index)
        throw(SystemException);
    
    void rewind()
        throw(SystemException);
};

// ----------------------------------------------------------------------
//
// The DynStruct implementation class
//
// ----------------------------------------------------------------------
class DynStruct_impl : virtual public DynStruct,
		       virtual public DynComplex_impl
{
    //
    // Hide copy-constructor and assignment operator
    //
    DynStruct_impl(const DynStruct_impl&);
    void operator=(const DynStruct_impl&);

protected: // internal state

protected: // constructor and destructor

    // The ORB and DynAny_impl may create DynStruct_impls
    friend class ORB;
    friend class DynAny_impl;

    DynStruct_impl(TypeCode_ptr, Boolean init = TRUE);
    virtual ~DynStruct_impl();

protected: // operations to marshal and unmarshal components
           // into a complex any value

public: // public IDL operations

    //
    // Standard IDL to C++ Mapping
    // from DynAny interface
    //
    DynAny_ptr copy()
        throw(SystemException);

    //
    // Standard IDL to C++ Mapping
    // from DynStruct interface
    //
    FieldName current_member_name()
        throw(SystemException);
    TCKind current_member_kind()
        throw(SystemException);
    NameValuePairSeq* get_members()
        throw(SystemException);
    void set_members(const NameValuePairSeq& value)
        throw(DynAny::InvalidSeq, SystemException);
};

// ----------------------------------------------------------------------
//
// The DynException implementation class
//
// ----------------------------------------------------------------------
class DynException_impl : virtual public DynStruct_impl
{
    //
    // Hide copy-constructor and assignment operator
    //
    DynException_impl(const DynException_impl&);
    void operator=(const DynException_impl&);

protected: // internal state

    const char* id_;

protected: // constructor and destructor

    // The ORB and DynAny_impl may create DynException_impls
    friend class ORB;
    friend class DynAny_impl;

    DynException_impl(TypeCode_ptr, Boolean init = TRUE);
    virtual ~DynException_impl();

protected: // operations to marshal and unmarshal components
           // into a complex any value

   void MarshalCount(OB::CountStream&);
   void Marshal(OB::OutputStream&);
   void Unmarshal(OB::InputStream&);

public: // public IDL operations

    //
    // Standard IDL to C++ Mapping
    // from DynAny interface
    //
    DynAny_ptr copy()
        throw(SystemException);
};

// ----------------------------------------------------------------------
//
// The DynUnion implementation class
//
// ----------------------------------------------------------------------
class DynUnion_impl : virtual public DynUnion,
		      virtual public DynComplex_impl
{
    //
    // Hide copy-constructor and assignment operator
    //
    DynUnion_impl(const DynUnion_impl&);
    void operator=(const DynUnion_impl&);

protected: // internal state

    OB::LongLongSeq member_labels_;
    OB::TypeCodeSeq member_types_;
    Long default_index_; // = -1 if no default or index in members

    Long current_discr_index_;

protected: // constructor and destructor

    // The ORB and DynAny_impl may create DynUnion_impls
    friend class ORB;
    friend class DynAny_impl;

    DynUnion_impl(TypeCode_ptr);
    virtual ~DynUnion_impl();

protected: // private operations

    void check_discriminator_and_set_member ();

    void set_dyn_any (DynAny_ptr dyn_any, LongLong);

    LongLong compute_any (Any* any);

    LongLong compute_default_value (TypeCode_ptr);

protected: // operations to marshal and unmarshal components
           // into a complex any value

   void Unmarshal(OB::InputStream&);

public: // public IDL operations

    //
    // Standard IDL to C++ Mapping
    // from DynAny interface
    //
    void assign(DynAny_ptr dyn_any)
        throw(DynAny::Invalid,
              SystemException);
    DynAny_ptr copy()
        throw(SystemException);

    Boolean next()
        throw(SystemException);

    //
    // Standard IDL to C++ Mapping
    // from DynUnion interface
    //
    Boolean set_as_default()
        throw(SystemException);
    
    void set_as_default(Boolean)
        throw(SystemException);
    
    DynAny_ptr discriminator()
        throw(SystemException);
    
    TCKind discriminator_kind()
        throw(SystemException);
    
    DynAny_ptr member()
        throw(SystemException);
    
    FieldName member_name()
        throw(SystemException);
    
    void member_name(const char*)
        throw(SystemException);
    
    TCKind member_kind()
        throw(SystemException);
};

// ----------------------------------------------------------------------
//
// The DynSequence implementation class
//
// ----------------------------------------------------------------------
class DynSequence_impl : virtual public DynSequence,
			 virtual public DynComplex_impl
{
    //
    // Hide copy-constructor and assignment operator
    //
    DynSequence_impl(const DynSequence_impl&);
    void operator=(const DynSequence_impl&);

protected: // internal state

protected: // constructor and destructor

    // The ORB and DynAny_impl may create DynSequence_impls
    friend class ORB;
    friend class DynAny_impl;

    DynSequence_impl(TypeCode_ptr);
    virtual ~DynSequence_impl();

protected:

   void init_components (ULong);

protected: // operations to marshal and unmarshal components
           // into a complex any value

   void MarshalCount(OB::CountStream&);
   void Marshal(OB::OutputStream&);
   void Unmarshal(OB::InputStream&);

public: // public IDL operations

    //
    // Standard IDL to C++ Mapping
    // from DynAny interface
    //
    void assign(DynAny_ptr dyn_any)
        throw(DynAny::Invalid,
              SystemException);
    DynAny_ptr copy()
        throw(SystemException);

    //
    // Standard IDL to C++ Mapping
    // from DynSequence
    //
    ULong length()
        throw(SystemException);
    void length(ULong)
        throw(SystemException);
    AnySeq* get_elements()
        throw(SystemException);
    void set_elements(const AnySeq& value)
        throw(DynAny::InvalidSeq, SystemException);
};

// ----------------------------------------------------------------------
//
// The DynArray implementation class
//
// ----------------------------------------------------------------------
class DynArray_impl : virtual public DynArray,
		      virtual public DynComplex_impl
{
    //
    // Hide copy-constructor and assignment operator
    //
    DynArray_impl(const DynArray_impl&);
    void operator=(const DynArray_impl&);

protected: // internal state

protected: // constructor and destructor

    // The ORB and DynAny_impl may create DynSequence_impls
    friend class ORB;
    friend class DynAny_impl;

    DynArray_impl(TypeCode_ptr, Boolean init = TRUE);
    virtual ~DynArray_impl();

protected: // operations to marshal and unmarshal components
           // into a complex any value

public: // public IDL operations

    //
    // Standard IDL to C++ Mapping
    // from DynAny interface
    //
    DynAny_ptr copy()
        throw(SystemException);

    //
    // Standard IDL to C++ Mapping
    // from DynArray interface
    //
    AnySeq* get_elements()
        throw(SystemException);
    void set_elements(const AnySeq& value)
        throw(DynAny::InvalidSeq,
              SystemException);
};

/*
// ----------------------------------------------------------------------
//
// The DynValue implementation class
//
// ----------------------------------------------------------------------
class DynValue_impl : virtual public DynValue,
                      virtual public DynComplex_impl
{
    DynValue_impl(const DynValue_impl&);
    void operator=(const DynValue_impl&);

    NameValueAccessSeq values_;

protected:

    DynValue_impl(TypeCode_ptr, Boolean init = TRUE);
    virtual ~DynValue_impl();

    friend class DynAny_impl;

public:

    //
    // IDL:omg.org/CORBA/DynValue/current_member_name:1.0
    //
    virtual FieldName current_member_name()
        throw(SystemException);

    //
    // IDL:omg.org/CORBA/DynValue/current_member_kind:1.0
    //
    virtual TCKind current_member_kind()
        throw(SystemException);

    //
    // IDL:omg.org/CORBA/DynValue/get_members:1.0
    //
    virtual NameValueAccessSeq* get_members()
        throw(SystemException);

    //
    // IDL:omg.org/CORBA/DynValue/set_members:1.0
    //
    virtual void set_members(const NameValueAccessSeq& values)
        throw(DynAny::InvalidSeq,
              SystemException);
};
*/

} // End of namespace CORBA

#endif

// ----------------------------------------------------------------------
// end of DynAny_impl.h file
// ----------------------------------------------------------------------
