// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Fixed.h>

#include <stdio.h>
#include <stdlib.h>

#ifdef HAVE_STRSTREAM
#   include <strstream>
#else
#   if defined(HAVE_STRSTREA_H)
#       include <strstrea.h>
#   else
#       include <strstream.h>
#   endif
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;

//
// This class implements a fixed point type with 31 digits precision.
// CORBA doesn't support more, this class might. But since it is meant
// to provide only basic support for CORBA the precision was restricted.
//

// ----------------------------------------------------------------------
// Fixed constructors and destructor
// ----------------------------------------------------------------------

Fixed::Fixed(Double val)
{
    char str[64];

    sprintf(str, "%15g", val);
    from_string(str);
}

#if SIZEOF_LONG_DOUBLE != 0
Fixed::Fixed(LongDouble val)
{
    char str[64];

    sprintf(str, "%31Lg", val);

    from_string(str);
}
#endif

Fixed::Fixed(const Fixed& val)
{
    scale_ = val.scale_;
    signum_ = val.signum_;
    memcpy(value_, val.value_, sizeof(value_));
}

// ----------------------------------------------------------------------
// Fixed public member implementation
// ----------------------------------------------------------------------

Fixed
Fixed::round(UShort n) const
{
    Fixed f = (Long)5;
    f.setScale(scale_ + 1);
    Fixed result = *this;
    result += f;

    return result.setScale(n);
}

Fixed
Fixed::truncate(UShort n) const
{
    return setScale(n);
}

UShort
Fixed::fixed_digits() const
{
    UShort digits = sizeof(value_);

    while(!value_[digits - 1] && digits > 0)
	digits--;

    return digits;
}

Short
Fixed::fixed_scale() const
{
    return scale();
}

// ----------------------------------------------------------------------
// Conversions from other types
// ----------------------------------------------------------------------

void
Fixed::from_longlong(LongLong val)
{
    if(val >= 0)
	from_ulonglong(val);
    else
	from_ulonglong(-val, -1);
}

void
Fixed::from_ulonglong(ULongLong val, Short signum)
{
    memset(value_, 0, sizeof(value_));
    scale_ = 0;
    signum_ = signum;

    Octet* v = value_;
    while(val)
    {
	*v = (Octet)(val % 10);
	v++;
	val /= 10;
    }
}

void
Fixed::from_string(const char* val)
{
    memset(value_, 0, sizeof(value_));
    scale_ = 0;
    signum_ = 1;

    //
    // This array takes the digits only temporarily
    //
    Octet dummy[sizeof(value_)];
    Octet* tmp = dummy + sizeof(value_);

    while(*val == ' ')
	val++;

    if(*val == '-')
    {
	signum_ = -1;
	val++;
    }

    if(*val == '+')
	val++;

    bool fraction = false;

    while(*val)
    {
	if(*val == '.')
	{
	    fraction = true;
	}
	else if(*val >= '0' && *val <= '9')
	{
	    tmp--;
	    *tmp = *val - '0';

	    //
	    // This digit might be placed right from the decimal point
	    //
	    if(fraction)
		scale_++;
	}
	else if(*val == 'e' || *val == 'E')
	{
	    //
	    // Exponent must be subtracted from scale
	    //
	    scale_ -= atoi(++val);

	    //
	    // Stop here
	    //
	    break;
	}
	else if(*val == 'd' || *val == 'D')
	{
	    //
	    // Trailing 'd' or 'D' is allowed.
	    // Check if it's the last character.
	    //
	    if(*(++val))
		throw DATA_CONVERSION();
	}
	else
	    throw DATA_CONVERSION();

	val++;
    }

    memcpy(value_, tmp, dummy + sizeof(value_) - tmp);
}

char*
Fixed::to_string() const
{
    Long i;
    String_var s = string_dup("");

    //
    // Digits left from the decimal point.
    //
    i = fixed_digits() - 1;
    while(i >= scale_)
    {
	s += (char)(value_[i] + '0');
	i--;
    }

    if(scale_ > 0)
	s += '.';

    //
    // Digits right from the decimal point.
    //
    while(i >= 0)
    {
	s += (char)(value_[i] + '0');
	i--;
    }

    String_var v;
    if(strlen(s))
    {
	if(signum_ == -1)
	    v = string_dup("-");
    }
    else
	s = string_dup("0");

    v += s;

    return v._retn();
}

// ----------------------------------------------------------------------
// Helpers
// ----------------------------------------------------------------------

Fixed
Fixed::abs() const
{
    if(signum_ >= 0)
	return *this;
    else
	return -(*this);
}

Short
Fixed::signum() const
{
    ULong i;
    for(i = 0 ; i < sizeof(value_) ; i++)
    {
	if(value_[i])
	    return signum_;
    }

    return 0;
}

Short
Fixed::scale() const
{
    return scale_;
}

Fixed
Fixed::setScale(Short scale) const
{
    assert(scale >= 0);

    Fixed result = *this;

    Short diff = scale - scale_;
    if(diff > 0)
    {
	memmove(result.value_ + diff, result.value_,
		sizeof(result.value_) - diff);
	memset(result.value_, 0, diff);
    }
    else if(diff < 0)
    {
	memmove(result.value_, result.value_ + diff,
		sizeof(result.value_) - diff);
	memset(result.value_ + sizeof(result.value_) - diff, 0, diff);
    }

    result.scale_ = scale;

    return result;
}

Fixed
Fixed::movePointLeft(Short n) const
{
    Fixed result;

    if(n >= 0)
    {
	result = *this;
	result.scale_ += n;
    }
    else
	result = movePointRight(-n);

    return result;
}

Fixed
Fixed::movePointRight(Short n) const
{
    Fixed result;

    if(scale_ >= n)
    {
	result = *this;
	result.scale_ -= n;
    }
    else
    {
	result = setScale(n - scale_);
	result.scale_ = 0;
    }

    return result;
}

// ----------------------------------------------------------------------
// Operators
// ----------------------------------------------------------------------

Fixed::operator LongLong() const
{
    if(fixed_digits() - scale_ > 31)
	throw DATA_CONVERSION();

    Long i;
    LongLong result = 0;

    //
    // Digits left from the decimal point.
    //
    i = sizeof(value_) - 1;
    while(i >= scale_)
    {
	result *= 10;
	result += value_[i];
	i--;
    }

    if(signum_ >= 0)
	return result;
    else
	return -result;
}

Fixed::operator LongDouble() const
{
#if SIZEOF_LONG_DOUBLE != 0

    Long i;
    LongDouble result = 0;

    //
    // Digits left from the decimal point.
    //
    i = sizeof(value_) - 1;
    while(i >= scale_)
    {
	result *= 10;
	result += value_[i];
	i--;
    }

    //
    // Digits right from the decimal point.
    //
    while(i >= 0)
    {
	result += value_[i] / (10 * i);
	i--;
    }

    if(signum_ >= 0)
	return result;
    else
	return -result;

#else
    assert(SIZEOF_LONG_DOUBLE != 0);
    return LongDouble(); // Some compilers need this
#endif
}

Fixed&
Fixed::operator=(const Fixed& val)
{
    scale_ = val.scale_;
    signum_ = val.signum_;
    memcpy(value_, val.value_, sizeof(value_));

    return *this;
}

Fixed&
Fixed::operator+=(const Fixed& val)
{
    *this = *this + val;

    return *this;
}

Fixed&
Fixed::operator-=(const Fixed& val)
{
    *this = *this - val;

    return *this;
}

Fixed&
Fixed::operator*=(const Fixed& val)
{
    *this = *this * val;

    return *this;
}

Fixed&
Fixed::operator/=(const Fixed& val)
{
    *this = *this / val;

    return *this;
}

Fixed&
Fixed::operator++()
{
    *this = *this + Fixed(Short(1));

    return *this;
}

Fixed
Fixed::operator++(int)
{
    *this = *this + Fixed(Short(1));

    return *this;
}

Fixed&
Fixed::operator--()
{
    *this = *this - Fixed(Short(1));

    return *this;
}

Fixed
Fixed::operator--(int)
{
    *this = *this - Fixed(Short(1));

    return *this;
}

Fixed
Fixed::operator+() const
{
    return *this;
}

Fixed
Fixed::operator-() const
{
    Fixed result = *this;
   
    result.signum_ *= -1;

    return result;
}

Boolean
Fixed::operator!() const
{
    ULong i;
    for(i = 0 ; i < sizeof(value_) - 1 ; i++)
	if(value_[i]) return false;

    return true;
}

Fixed
Fixed::operator+(const Fixed& val)
{
    if(signum_ == -1 && val.signum_ == 1)
	return Fixed(val) - (-*this);
    else if(val.signum_ == -1 && signum_ == 1)
	return *this - (-val);

    Fixed v1 = *this;
    Fixed v2 = val;

    //
    // Adjust scales, i. e. decimal point
    //
    if(v1.scale_ < v2.scale_)
	v1 = v1.setScale(v2.scale_);
    else if(v2.scale_ < v1.scale_)
	v2 = v2.setScale(v1.scale_);

    //
    // Add digits
    //
    Octet carry = 0;
    ULong i;
    for(i = 0 ; i < sizeof(v2.value_) - 1 ; i++)
    {
	v1.value_[i] += v2.value_[i] + carry;
	if(v1.value_[i] >= 10)
	{
	    v1.value_[i] -= 10;
	    carry = 1;
	}
	else
	    carry = 0;
    }

    return v1;
}

Fixed
Fixed::operator-(const Fixed& val)
{
    if(signum_ == -1 && val.signum_ == -1)
	return -Fixed(val) - (- *this);
    else if(signum_ == -1 && val.signum_ == 1)
	return -((*this) + val);
    else if(signum_ == 1 && val.signum_ == -1)
	return *this + (-val);

    Fixed v1;
    Fixed v2;
    bool neg;
    if(*this > val)
    {
	v1 = *this;
	v2 = val;
	neg = false;
    }
    else
    {
	v2 = *this;
	v1 = val;
	neg = true;
    }

    //
    // Adjust scales, i. e. decimal point
    //
    if(v1.scale_ < v2.scale_)
	v1 = v1.setScale(v2.scale_);
    else if(v2.scale_ < v1.scale_)
	v2 = v2.setScale(v1.scale_);

    //
    // Subtract digits
    //
    Octet carry = 0;
    ULong i;
    for(i = 0 ; i < sizeof(v1.value_) - 1 ; i++)
    {
	v1.value_[i] -= v2.value_[i] + carry;
	if(v1.value_[i] >= 10)
	{
	    v1.value_[i] += 10;
	    carry = 1;
	}
	else
	    carry = 0;
    }

    if(carry || neg)
	v1.signum_ = -1;

    return v1;
}

Fixed
Fixed::operator*(const Fixed& val)
{
    Fixed result;

    ULong i;
    for(i = 0 ; i < sizeof(value_) / 2 - 1 ; i++)
    {
	//
	// Multiply absolute value by 10^i and add to result
	//
	if(val.value_[i])
	{
	    Fixed tmp = abs();
	    memmove(&tmp.value_[i], &tmp.value_, sizeof(tmp.value_) / 2);
	    memset(tmp.value_, 0, i);

	    Octet j;
	    for(j = 0 ; j < val.value_[i] ; j++)
		result += tmp;
	}
    }

    result.signum_ = signum_ * val.signum_;
    result.scale_ = scale_ + val.scale_;

    return result;
}

Fixed
Fixed::operator/(const Fixed& val)
{
    assert(Fixed((Long)0) != val);

    //
    // Rescale to largest scale
    //
    Fixed dividend, divisor;
    if(val.scale_ > scale_)
    {
        dividend = setScale(val.scale_);
	divisor = val;
    }
    else
    {
	dividend = *this;
	divisor = val.setScale(scale_);
    }

    Fixed result;

    Fixed v1 = *this;
    Fixed v2 = val;

    v1.setScale(0);
    v2.setScale(0);

    result.signum_ = v1.signum_ * v2.signum_;

    return result;
}

Boolean
Fixed::operator>(const Fixed& val)
{
    if(signum_ != val.signum_)
	return signum_ > val.signum_;

    Fixed v1 = *this;
    Fixed v2 = val;

    //
    // Adjust scales, i. e. decimal point
    //
    if(v1.scale_ < v2.scale_)
	v1 = v1.setScale(v2.scale_);
    else if(v2.scale_ < v1.scale_)
	v2 = v2.setScale(v1.scale_);

    return memcmp(v1.value_, v2.value_, sizeof(v1.value_)) <= 0;
}

Boolean
Fixed::operator<(const Fixed& val)
{
    return Fixed(val) > *this;
}

Boolean
Fixed::operator>=(const Fixed& val)
{
    return !(*this < Fixed(val));
}

Boolean
Fixed::operator<=(const Fixed& val)
{
    return !(*this > val);
}

Boolean
Fixed::operator==(const Fixed& val)
{
    return !(*this - val);
}

Boolean
Fixed::operator!=(const Fixed& val)
{
    return !(*this == val);
}

// ----------------------------------------------------------------------
// Stream I/O
// ----------------------------------------------------------------------

istream& operator>>(istream& in, Fixed &val)
{
    char str[1024];
    in.get(str, 1023);
    assert(in.good());
    val = str;

    return in;
}

ostream& operator<<(ostream& out, const Fixed &val)
{
    String_var v = val.to_string();

    return out << v;
}
