// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/Any.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/IFR1.h>
#include <OB/OBV.h>
#include <OB/Policy.h>
#include <OB/ORB.h>
#include <OB/GIOP.h>
#include <OB/OCI.h>
#include <OB/Reactor.h>
#include <OB/Stream.h>

#include <GIOPClient.h>
#include <GIOPClientWorker.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::GIOPClient_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPClient_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// Template instantiations
// ----------------------------------------------------------------------

#ifndef HAVE_NO_EXPLICIT_TEMPLATES
template class OB::ObjVar< GIOPClient >;
template class OB::ObjForSeq< GIOPClient >;
#else
#ifdef HAVE_PRAGMA_DEFINE
#pragma define(OB::ObjVar< GIOPClient >)
#pragma define(OB::ObjForSeq< GIOPClient >)
#endif
#endif

// ----------------------------------------------------------------------
// GIOPClient private and protected member implementations
// ----------------------------------------------------------------------

void
GIOPClient::addHeaders(StubImplBase* p, const char* op, ULong reqId, bool ow,
		       Buffer& buf)
{
    OutputStream out(buf, 0);

    GIOP::MessageHeader_1_0 hdr =
    {
	{ 'G', 'I', 'O', 'P' }, { 1, 0 },
	Endian, GIOP::Request, buf.len - 12
    };
    GIOP::MessageHeader_1_0::_OB_marshal(hdr, out);

    //
    // If key is not cached, extract from IOR
    //
    if(p -> _OB_key().length() == 0)
    {
	OCI::ObjectKey_var k = connector_ -> is_usable(p -> _OB_ior());
	p -> _OB_key(k.in());
	assert(p -> _OB_key().length());
    }
    //const OCI::ObjectKey key = p -> _OB_key();

    GIOP::RequestHeader_1_1 reqH;
    reqH.request_id = reqId;
    reqH.response_expected = !ow;
#ifdef OB_CLEAR_MEM
    reqH.reserved[0] = 0;
    reqH.reserved[1] = 0;
    reqH.reserved[2] = 0;
#endif
    //reqH.object_key = key;
    ULong len = p -> _OB_key().length();
    reqH.object_key.length(len);
    memcpy(reqH.object_key.get_buffer(), p -> _OB_key().get_buffer(), len);
    reqH.operation = op;
    GIOP::RequestHeader_1_1::_OB_marshal(reqH, out);
}

GIOPClientWorker_ptr
GIOPClient::getWorker()
{
#ifdef HAVE_JTC
    JTCSynchronized synchronized(workerMutex_);
#endif

    if(is_nil(worker_))
    {
	//
	// Create new transport, using the connector
	//
	OCI::Transport_var transport = connector_ -> connect();
        
	//
	// Create new worker
	//
	switch(ORB::conc_model())
	{
	case ORB::ConcModelBlocking:
	    worker_ = new GIOPClientWorkerBlocking(this, transport);
	    break;
	    
	case ORB::ConcModelReactive:
	    worker_ = new GIOPClientWorkerReactive(this, transport);
	    break;
	    
#ifdef HAVE_JTC
	case ORB::ConcModelThreaded:
	    worker_ = new GIOPClientWorkerThreaded(this, transport);
	    break;
#endif
	}
    }

    return GIOPClientWorker::_duplicate(worker_);
}

// ----------------------------------------------------------------------
// GIOPClient constructor and destructor
// ----------------------------------------------------------------------

GIOPClient::GIOPClient(OCI::Connector_ptr connector)
    : connector_(OCI::Connector::_duplicate(connector))
{
    // cout << "GIOPClient" << endl;
}

GIOPClient::~GIOPClient()
{
    //cout << "~GIOPClient" << endl;

    //
    // If worker_ is not nil, destroy() was not called
    //
    assert(is_nil(worker_));
}

// ----------------------------------------------------------------------
// GIOPClient public member implementation
// ----------------------------------------------------------------------

void
GIOPClient::destroy()
{
    //
    // Destroy the worker. Make sure that any further destructions do
    // not block.
    //
    GIOPClientWorker_var w;

    {
#ifdef HAVE_JTC
	JTCSynchronized synchronized(workerMutex_);
#endif
	w = worker_._retn();
    }

    if(!is_nil(w))
	w -> destroy();
}

void
GIOPClient::destroyWorker(GIOPClientWorker_ptr worker)
{
    assert(!is_nil(worker));
	
    //
    // Destroy the worker. Make sure that any further destructions do
    // not block.
    //
    GIOPClientWorker_var w;

    {
#ifdef HAVE_JTC
	JTCSynchronized synchronized(workerMutex_);
#endif
	if(worker_.in() == worker)
	    w = worker_._retn();
    }

    if(!is_nil(w))
	w -> destroy();
}

OCI::Connector_ptr
GIOPClient::connector()
{
    return OCI::Connector::_duplicate(connector_);
}

OCI::Transport_ptr
GIOPClient::transport()
{
    GIOPClientWorker_var worker = getWorker();
    return worker -> transport();
}

ULong
GIOPClient::offset(StubImplBase* p, const char* op)
{
    CountStream offs(12);
    
    if(p -> _OB_key().length() == 0) // If key is not cached, extract from IOR
    {
	OCI::ObjectKey_var k = connector_ -> is_usable(p -> _OB_ior());
	p -> _OB_key(k.in());
	assert(p -> _OB_key().length());
    }

    GIOP::RequestHeader_1_1 reqH;
#ifdef OB_CLEAR_MEM
    reqH.request_id = 0;
    reqH.response_expected = true;
    reqH.reserved[0] = 0;
    reqH.reserved[1] = 0;
    reqH.reserved[2] = 0;
#endif
    ULong len = p -> _OB_key().length();
    reqH.object_key.length(len);
    memcpy(reqH.object_key.get_buffer(), p -> _OB_key().get_buffer(), len);
    reqH.operation = op;
    
    GIOP::RequestHeader_1_1::_OB_marshalCount(reqH, offs);
    
    return offs.count();
}

ULong
GIOPClient::request(StubImplBase* p, const char* op, Buffer& buf, bool& swap, 
		    bool& exc, bool& fwd, Long t, bool retry)
{
    GIOP::ReplyStatusType status;
    ULong offset;

    while(true)
    {
	//
	// Get the worker
	//
	GIOPClientWorker_var worker = getWorker();
	
	//
	// Add headers
	//
	ULong reqId = worker -> getRequestId();
	addHeaders(p, op, reqId, false, buf);
	
	try
	{
	    //
	    // Send request and receive reply, with timeout
	    //
	    offset = worker -> sendReceiveTimeout(reqId, buf, swap, status, t);
	}
	catch(const COMM_FAILURE&)
	{
	    //
	    // Retry once on communication failure if the retry
	    // parameter is set to true
	    //
	    if(retry)
	    {
		retry = false;
		continue;
	    }

	    throw;
	}
	catch(const TRANSIENT&)
	{
	    //
	    // Always retry on transient failure
	    //
	    continue;
	}
	
	//
	// Was there a timeout?
	//
	if(offset == 0)
	{
	    assert(t >= 0);
	    
	    //
	    // Mark the request ID as discarded and throw a timeout
	    // exception
	    //
	    worker -> discardRequest(reqId);
	    throw NO_RESPONSE();
	}

	break;
    }

    //
    // Evaluate reply
    //
    switch(status)
    {
    case GIOP::NO_EXCEPTION:
	exc = false;
	fwd = false;
	break;
	
    case GIOP::USER_EXCEPTION:
	exc = true;
	fwd = false;
	break;
	
    case GIOP::SYSTEM_EXCEPTION:
    {
        InputStream in(buf, offset, swap);
	SystemException::_OB_unmarshalThrow(in);
	break;
    }
	
    case GIOP::LOCATION_FORWARD:
	exc = false;
	fwd = true;
	break;
    }

    return offset;
}
    
void
GIOPClient::oneway(StubImplBase* p, const char* op, Buffer& buf, bool retry)
{
    while(true)
    {
	//
	// Get the worker
	//
	GIOPClientWorker_var worker = getWorker();
	
	//
	// Add headers
	//
	ULong reqId = worker -> getRequestId();
	addHeaders(p, op, reqId, true, buf);
	
	try
	{
	    //
	    // Send request non-blocking
	    //
	    worker -> sendNonBlocking(buf);
	}
	catch(const COMM_FAILURE&)
	{
	    //
	    // Retry once on communication failure if the retry
	    // parameter is set to true
	    //
	    if(retry)
	    {
		retry = false;
		continue;
	    }
	    
	    throw;
	}
	catch(const TRANSIENT&)
	{
	    //
	    // Always retry on transient failure
	    //
	    continue;
	}
	
	break;
    }
}

void
GIOPClient::deferred(StubImplBase* p, const char* op, Buffer& buf,
		       ULong& reqId, bool retry)
{
    while(true)
    {
	//
	// Get the worker
	//
	GIOPClientWorker_var worker = getWorker();
	
	//
	// Add headers
	//
	reqId = worker -> getRequestId();
	addHeaders(p, op, reqId, false, buf);
	
	try
	{
	    //
	    // Send request non-blocking
	    //
	    // This should really be sendTimeout, but I don't have such an
	    // operation right now.
	    //
	    worker -> sendNonBlocking(buf);
	}
	catch(const COMM_FAILURE&)
	{
	    //
	    // Retry once on communication failure if the retry
	    // parameter is set to true
	    //
	    if(retry)
	    {
		retry = false;
		continue;
	    }
	    
	    throw;
	}
	catch(const TRANSIENT&)
	{
	    //
	    // Always retry on transient failure
	    //
	    continue;
	}

	break;
    }
}

ULong
GIOPClient::response(ULong reqId, Buffer& buf, bool& swap, bool& exc,
                     bool& fwd, Long t)
{
    GIOP::ReplyStatusType status;
    ULong offset;

    while(true)
    {
        //
	// Get the worker
	//
	GIOPClientWorker_var worker = getWorker();
	
	try
	{
	    //
	    // Receive reply, with timout
	    //
	    offset = worker -> receiveTimeout(reqId, buf, swap, status, t);
	}
	catch(const TRANSIENT&)
	{
	    //
	    // Always retry on transient failure
	    //
	    continue;
	}

	//
	// Was there a timeout?
	//
	if(offset == 0)
	{
	    assert(t >= 0);
	    
	    //
	    // Mark the request ID as discarded and throw a timeout
	    // exception
	    //
	    worker -> discardRequest(reqId);
	    throw NO_RESPONSE();
	}

	break;
    }

    //
    // Evaluate reply
    //
    switch(status)
    {
    case GIOP::NO_EXCEPTION:
	exc = false;
	fwd = false;
	break;
	
    case GIOP::USER_EXCEPTION:
	exc = true;
	fwd = false;
	break;
	
    case GIOP::SYSTEM_EXCEPTION:
    {
        InputStream in(buf, offset, swap);
	SystemException::_OB_unmarshalThrow(in);
	break;
    }
	
    case GIOP::LOCATION_FORWARD:
	exc = false;
	fwd = true;
	break;
    }

    return offset;
}

ULong
GIOPClient::poll(ULong reqId, Buffer& buf, bool& swap, bool& exc, bool& fwd)
{
    GIOP::ReplyStatusType status;
    ULong offset;

    while(true)
    {
	//
	// Get the worker
	//
	GIOPClientWorker_var worker = getWorker();

	try
	{
	    //
	    // Receive reply, with timeout
	    //
	    offset = worker -> receiveTimeout(reqId, buf, swap, status, 0);
	}
	catch(const TRANSIENT&)
	{
	    //
	    // Always retry on transient failure
	    //
	    continue;
	}
	
	//
	// Was there a timeout?
	//
	if(offset == 0)
	    return 0;
	
	break;
    }

    //
    // Evaluate reply
    //
    switch(status)
    {
    case GIOP::NO_EXCEPTION:
	exc = false;
	fwd = false;
	break;
	
    case GIOP::USER_EXCEPTION:
	exc = true;
	fwd = false;
	break;
	
    case GIOP::SYSTEM_EXCEPTION:
    {
        InputStream in(buf, offset, swap);
	SystemException::_OB_unmarshalThrow(in);
	break;
    }
	
    case GIOP::LOCATION_FORWARD:
	exc = false;
	fwd = true;
	break;
    }

    return offset;
}
