// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_GIOP_CLIENT_WORKER_H
#define OB_GIOP_CLIENT_WORKER_H

namespace OB
{

class GIOPClientWorker;
typedef GIOPClientWorker* GIOPClientWorker_ptr;

class GIOPClientWorkerBlocking;
typedef GIOPClientWorkerBlocking* GIOPClientWorkerBlocking_ptr;

class GIOPClientWorkerReactive;
typedef GIOPClientWorkerReactive* GIOPClientWorkerReactive_ptr;

#ifdef HAVE_JTC

class GIOPClientWorkerThreaded;
typedef GIOPClientWorkerThreaded* GIOPClientWorkerThreaded_ptr;

#endif

} // End of namespace OB

void OBDuplicate(OB::GIOPClientWorker_ptr);
void OBRelease(OB::GIOPClientWorker_ptr);
void OBDuplicate(OB::GIOPClientWorkerBlocking_ptr);
void OBRelease(OB::GIOPClientWorkerBlocking_ptr);
void OBDuplicate(OB::GIOPClientWorkerReactive_ptr);
void OBRelease(OB::GIOPClientWorkerReactive_ptr);

#ifdef HAVE_JTC

void OBDuplicate(OB::GIOPClientWorkerThreaded_ptr);
void OBRelease(OB::GIOPClientWorkerThreaded_ptr);

#endif

namespace OB
{

//
// The GIOPClientWorker class
//
typedef ObjVar< GIOPClientWorker > GIOPClientWorker_var;

class GIOPClientWorker : public RefCount OB_COMMA_MONITOR
{
protected:

    static CORBA::ULong nextRequestId_; // Next request ID
    static ULongSeq discReqIdSeq_; // Discarded request IDs
#ifdef HAVE_JTC
    static JTCMutex requestIdMutex_; // Monitor for request ID handling
#endif

    GIOPClient_var client_; // The client
    OCI::Transport_var transport_; // The transport
    OCI::BufferSeq outSeq_; // The outgoing message buffers
    OCI::Buffer_var buf_; // The incoming message buffer
    GIOP::MessageHeader_1_0 hdr_; // The incoming message header

public: // For some compilers the following struct must be public
    struct Reply
    {
	CORBA::ULong reqId;
	Buffer buf;
	bool swap;
	GIOP::ReplyStatusType status;
	CORBA::ULong offset;
	Reply* next;
    };
protected:
    Reply* head_; // I can't use OBVarSeq<> because buf must not
                  // be copied (operator=()) for performance reasons

    //
    // Add a "message error" message
    //
    void addMessageError();

    //
    // Setup buf_ so that a new incoming message can be received
    //
    void setupBuffer();

    //
    // Extract the hdr_ from buf_
    //
    void extractHeader();

    //
    // Add outgoing message
    //
    virtual void add(OCI::Buffer_ptr);

    //
    // Execute incoming message
    //
    void execute();

    //
    // Find a reply
    // Check for a reply
    //
    CORBA::ULong findReply(CORBA::ULong, Buffer&, bool&,
			   GIOP::ReplyStatusType&);
    CORBA::Boolean checkReply(CORBA::ULong);

public:

    GIOPClientWorker(GIOPClient_ptr, OCI::Transport_ptr);
    virtual ~GIOPClientWorker();

    static inline GIOPClientWorker_ptr _duplicate(GIOPClientWorker_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline GIOPClientWorker_ptr _nil()
    { return 0; }

    //
    // Destroy the worker
    //
    virtual void destroy();

    //
    // Get the OCI transport
    //
    OCI::Transport_ptr transport();

    //
    // Get a request id
    // Discard a request
    //
    CORBA::ULong getRequestId()
    {
#ifdef HAVE_JTC
	JTCSynchronized synchronized(requestIdMutex_);
#endif
	return nextRequestId_++;
    }
    void discardRequest(CORBA::ULong reqId)
    {
#ifdef HAVE_JTC
	JTCSynchronized synchronized(requestIdMutex_);
#endif
	discReqIdSeq_.append(reqId);
    }

    //
    // Send a request and receive a reply, with timeout
    // Receive a reply, with timeout
    // Send a request, non-blocking
    //
    virtual CORBA::ULong sendReceiveTimeout(CORBA::ULong, Buffer&, bool&,
					    GIOP::ReplyStatusType&,
					    CORBA::Long) = 0;
    virtual CORBA::ULong receiveTimeout(CORBA::ULong, Buffer&, bool&,
				        GIOP::ReplyStatusType&,
                                        CORBA::Long) = 0;
    virtual void sendNonBlocking(Buffer&) = 0;
};

//
// The GIOPClientWorkerBlocking class
//
typedef ObjVar< GIOPClientWorkerBlocking > GIOPClientWorkerBlocking_var;

class GIOPClientWorkerBlocking : public GIOPClientWorker
{
public:

    GIOPClientWorkerBlocking(GIOPClient_ptr, OCI::Transport_ptr);
    virtual ~GIOPClientWorkerBlocking();

    static inline GIOPClientWorkerBlocking_ptr
    _duplicate(GIOPClientWorkerBlocking_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline GIOPClientWorkerBlocking_ptr _nil()
    { return 0; }

    //
    // Send a request and receive a reply, with timeout
    // Receive a reply, with timeout
    // Send a request, non-blocking
    //
    virtual CORBA::ULong sendReceiveTimeout(CORBA::ULong, Buffer&, bool&,
					    GIOP::ReplyStatusType&,
                                            CORBA::Long);
    virtual CORBA::ULong receiveTimeout(CORBA::ULong, Buffer&, bool&,
				        GIOP::ReplyStatusType&, CORBA::Long);
    virtual void sendNonBlocking(Buffer&);
};

//
// The GIOPClientWorkerReactive class
//
typedef ObjVar< GIOPClientWorkerReactive > GIOPClientWorkerReactive_var;

class GIOPClientWorkerReactive : public GIOPClientWorker,
				 public EventHandler
{
protected:

    bool destroy_; // True if destroy() was called
    CORBA::SystemException* exception_; // The buffered exception

    //
    // Add outgoing message
    //
    virtual void add(OCI::Buffer_ptr);

public:

    GIOPClientWorkerReactive(GIOPClient_ptr, OCI::Transport_ptr);
    ~GIOPClientWorkerReactive();

    static inline GIOPClientWorkerReactive_ptr
    _duplicate(GIOPClientWorkerReactive_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline GIOPClientWorkerReactive_ptr _nil()
    { return 0; }

    //
    // Destroy the worker
    //
    virtual void destroy();

    //
    // Send a request and receive a reply, with timeout
    // Receive a reply, with timeout
    // Send a request, non-blocking
    //
    virtual CORBA::ULong sendReceiveTimeout(CORBA::ULong, Buffer&, bool&,
					    GIOP::ReplyStatusType&,
                                            CORBA::Long);
    virtual CORBA::ULong receiveTimeout(CORBA::ULong, Buffer&, bool&,
				        GIOP::ReplyStatusType&, CORBA::Long);
    virtual void sendNonBlocking(Buffer&);

    //
    // Handle event
    //
    virtual void handleEvent(Mask);
};

#ifdef HAVE_JTC

//
// The GIOPClientWorkerThreaded class
//
typedef ObjVar< GIOPClientWorkerThreaded > GIOPClientWorkerThreaded_var;

class GIOPClientWorkerThreaded : public GIOPClientWorker
{
protected:

    bool destroy_; // True if destroy() was called
    CORBA::SystemException* exception_; // The buffered exception

    //
    // The sender thread
    //
    class SenderThread : public JTCThread
    {
	GIOPClientWorkerThreaded_var worker_;

    public:
	
	SenderThread(GIOPClientWorkerThreaded_ptr worker)
	    : worker_(GIOPClientWorkerThreaded::_duplicate(worker)) { }
	virtual ~SenderThread() { }
	virtual void run();
    };
    friend class SenderThread;
    JTCThreadHandle senderThread_;

    //
    // The receiver thread
    //
    class ReceiverThread : public JTCThread
    {
	GIOPClientWorkerThreaded_var worker_;

    public:
	
	ReceiverThread(GIOPClientWorkerThreaded_ptr worker)
	    : worker_(GIOPClientWorkerThreaded::_duplicate(worker)) { }
	virtual ~ReceiverThread() { }
	virtual void run();
    };
    friend class ReceiverThread;
    JTCThreadHandle receiverThread_;

    //
    // Add outgoing message
    // Remove outgoing message
    //
    virtual void add(OCI::Buffer_ptr);
    virtual OCI::Buffer_ptr remove();

public:

    GIOPClientWorkerThreaded(GIOPClient_ptr, OCI::Transport_ptr);
    ~GIOPClientWorkerThreaded();

    static inline GIOPClientWorkerThreaded_ptr
    _duplicate(GIOPClientWorkerThreaded_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline GIOPClientWorkerThreaded_ptr _nil()
    { return 0; }

    //
    // Destroy the worker
    //
    virtual void destroy();

    //
    // Send a request and receive a reply, with timeout
    // Receive a reply, with timeout
    // Send a request, non-blocking
    //
    virtual CORBA::ULong sendReceiveTimeout(CORBA::ULong, Buffer&, bool&,
					    GIOP::ReplyStatusType&, 
                                            CORBA::Long);
    virtual CORBA::ULong receiveTimeout(CORBA::ULong, Buffer&, bool&,
				        GIOP::ReplyStatusType&, CORBA::Long);
    virtual void sendNonBlocking(Buffer&);

    //
    // Run method for receiver and sender thread
    //
    void senderRun();
    void receiverRun();
};

#endif

} // End of namespace OB


namespace CORBA
{
// XXX Are these necessary in CORBA?

inline void
release(OB::GIOPClientWorker_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::GIOPClientWorker_ptr p)
{
    return p == 0;
}

inline void
release(OB::GIOPClientWorkerBlocking_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::GIOPClientWorkerBlocking_ptr p)
{
    return p == 0;
}

inline void
release(OB::GIOPClientWorkerReactive_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::GIOPClientWorkerReactive_ptr p)
{
    return p == 0;
}

#ifdef HAVE_JTC

inline void
release(OB::GIOPClientWorkerThreaded_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::GIOPClientWorkerThreaded_ptr p)
{
    return p == 0;
}

#endif // HAVE_JTC

} // End of namespace CORBA

#endif
