// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/TemplateI.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/OCI.h>
#include <OB/GIOP.h>
#include <OB/Reactor.h>

#include <GIOPServerStarter.h>
#include <Downcall.h>
#include <GIOPServerWorker.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::GIOPServerStarter_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPServerStarter_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

void
OBDuplicate(OB::GIOPServerStarterBlocking_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPServerStarterBlocking_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

void
OBDuplicate(OB::GIOPServerStarterReactive_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPServerStarterReactive_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

#ifdef HAVE_JTC
void
OBDuplicate(OB::GIOPServerStarterThreaded_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPServerStarterThreaded_ptr p)
{
    if(p)
	p -> _OB_decRef();
}
#endif

// ----------------------------------------------------------------------
// Template instantiations
// ----------------------------------------------------------------------

#ifndef HAVE_NO_EXPLICIT_TEMPLATES
template class OB::ObjVar< GIOPServerStarter >;
template class OB::ObjForSeq< GIOPServerStarter >;
template class OB::ObjVar< GIOPServerStarterBlocking >;
template class OB::ObjForSeq< GIOPServerStarterBlocking >;
template class OB::ObjVar< GIOPServerStarterReactive >;
template class OB::ObjForSeq< GIOPServerStarterReactive >;
#ifdef HAVE_JTC
template class OB::ObjVar< GIOPServerStarterThreaded >;
template class OB::ObjForSeq< GIOPServerStarterThreaded >;
#endif
#else
#ifdef HAVE_PRAGMA_DEFINE
#pragma define(OB::ObjVar< GIOPServerStarter >)
#pragma define(OB::ObjForSeq< GIOPServerStarter >)
#pragma define(OB::ObjVar< GIOPServerStarterBlocking >)
#pragma define(OB::ObjForSeq< GIOPServerStarterBlocking >)
#pragma define(OB::ObjVar< GIOPServerStarterReactive >)
#pragma define(OB::ObjForSeq< GIOPServerStarterReactive >)
#ifdef HAVE_JTC
#pragma define(OB::ObjVar< GIOPServerStarterThreaded >)
#pragma define(OB::ObjForSeq< GIOPServerStarterThreaded >)
#endif
#endif
#endif

// ----------------------------------------------------------------------
// GIOPServerStarter constructor and destructor
// ----------------------------------------------------------------------

GIOPServerStarter::GIOPServerStarter(OCI::Acceptor_ptr acceptor, bool enabled,
                                     Upcall* upcall)
    : acceptor_(OCI::Acceptor::_duplicate(acceptor)),
      upcall_(upcall)
{
    //cout << "GIOPServerStarter" << endl;
    //
    // Only start the acceptor listening if enabled
    //
    try
    {
        if(enabled)
            acceptor_ -> listen();
    }
    catch(const SystemException&)
    {
        _OB_setRef(0);
        throw;
    }
}

GIOPServerStarter::~GIOPServerStarter()
{
    //cout << "~GIOPServerStarter" << endl;

    //
    // If workers_ is not empty, destroy() was not called
    //
    assert(workers_.length() == 0);
}

// ----------------------------------------------------------------------
// GIOPServerStarter public member implementation
// ----------------------------------------------------------------------

void
GIOPServerStarter::destroy()
{
    //
    // Close the acceptor
    //
    acceptor_ -> close();

    //
    // Destroy all workers. Make sure that any further destructions do
    // not block.
    //
    GIOPServerWorkerSeq workers;

    {
#ifdef HAVE_JTC
	JTCSynchronized synchronized(workersMutex_);
#endif
        workers = workers_;
        workers_.length(0);
    }

    for(ULong i = 0 ; i < workers.length() ; i++)
	workers[i] -> destroy();
}

void
GIOPServerStarter::destroyWorker(GIOPServerWorker_ptr worker)
{
    assert(!is_nil(worker));

    //
    // Destroy the worker. Make sure that any further destructions do
    // not block.
    //
    GIOPServerWorker_var w;
    
    {
#ifdef HAVE_JTC
	JTCSynchronized synchronized(workersMutex_);
#endif
	    
	for(ULong i = 0 ; i < workers_.length() ; i++)
	    if(workers_[i].in() == worker)
	    {
		w = workers_[i]._retn();
		workers_.remove(i);
		break;
	    }
    }

    if(!is_nil(w))
	w -> destroy();
}

bool
GIOPServerStarter::startWorker()
{
    return false;
}

// ----------------------------------------------------------------------
// GIOPServerStarterBlocking constructor and destructor
// ----------------------------------------------------------------------

GIOPServerStarterBlocking::GIOPServerStarterBlocking(
    OCI::Acceptor_ptr acceptor, bool enabled, Upcall* upcall)
    : GIOPServerStarter(acceptor, enabled, upcall)
{
}

GIOPServerStarterBlocking::~GIOPServerStarterBlocking()
{
}

// ----------------------------------------------------------------------
// GIOPServerStarterBlocking public member implementation
// ----------------------------------------------------------------------

bool
GIOPServerStarterBlocking::startWorker()
{
    try
    {
	OCI::Transport_var transport = acceptor_ -> accept();
	assert(!is_nil(transport));

	GIOPServerWorkerBlocking_var worker =
	    new GIOPServerWorkerBlocking(this, transport, upcall_);
	workers_.append(GIOPServerWorker::_duplicate(worker));
	
	while(true)
	{
	    if(workers_.length() == 0)
		break;

	    assert(workers_.length() == 1);
	    worker -> getRequest();

	    if(workers_.length() == 0)
		break;

	    assert(workers_.length() == 1);
	    worker -> sendReply();
	}
    }
    catch(const NO_PERMISSION&)
    {
	//
	// Ignore NO_PERMISSION exceptions
	//
    }

    return true;
}

// ----------------------------------------------------------------------
// GIOPServerStarterReactive constructor and destructor
// ----------------------------------------------------------------------

GIOPServerStarterReactive::GIOPServerStarterReactive(
    OCI::Acceptor_ptr acceptor, bool enabled, Upcall* upcall)
    : GIOPServerStarter(acceptor, enabled, upcall), destroy_(false)
{
    //
    // Register as read event handler
    //
    OCI::Handle handle = acceptor_ -> handle();
    assert(handle >= 0);
    Reactor::instance() -> registerHandler(this, EventRead, handle);
}

GIOPServerStarterReactive::~GIOPServerStarterReactive()
{
    assert(destroy_);
}

// ----------------------------------------------------------------------
// GIOPServerStarterReactive public member implementation
// ----------------------------------------------------------------------

void
GIOPServerStarterReactive::destroy()
{
    //
    // WARNING: The reactive starter is not thread safe!
    //

    //
    // Don't destroy twice
    //
    if(destroy_)
	return;
    
    //
    // Set the destroy flag
    //
    destroy_ = true;
    
    //
    // Unregister event handler
    //
    Reactor::instance() -> unregisterHandler(this);

    //
    // Call destroy on the super class
    //
    GIOPServerStarter::destroy();
}

void
GIOPServerStarterReactive::handleEvent(Mask mask)
{
    //
    // Handle read events
    //
    assert(mask == EventRead);

    try
    {
	//
	// Get new transport
	//
	OCI::Transport_var transport = acceptor_ -> accept();
	assert(!is_nil(transport));

	//
	// Create new worker
	//
	{
#ifdef HAVE_JTC
	    JTCSynchronized synchronized(workersMutex_);
#endif
	    workers_.append(new GIOPServerWorkerReactive(this, transport,
                                                         upcall_));
	}
    }
    catch(const NO_PERMISSION&)
    {
	//
	// Ignore NO_PERMISSION exceptions
	//
    }
}

#ifdef HAVE_JTC

// ----------------------------------------------------------------------
// GIOPServerStarterThreaded constructor and destructor
// ----------------------------------------------------------------------

GIOPServerStarterThreaded::GIOPServerStarterThreaded(
    OCI::Acceptor_ptr acceptor, bool enabled, Upcall* upcall)
    : GIOPServerStarter(acceptor, enabled, upcall), destroy_(false)
{
    //
    // Create starter thread
    //
    starterThread_ = new StarterThread(this);

    //
    // Start the starter thread if not disabled
    //
    if(enabled)
	starterThread_ -> start();
}

GIOPServerStarterThreaded::~GIOPServerStarterThreaded()
{
    assert(destroy_);
}

// ----------------------------------------------------------------------
// GIOPServerStarterThreaded public member implementation
// ----------------------------------------------------------------------

void
GIOPServerStarterThreaded::destroy()
{
    //
    // Set the destroy flag
    //
    assert(!destroy_);
    destroy_ = true;

    //
    // Shutdown the acceptor
    //
    acceptor_ -> shutdown();

    //
    // Join with the starter thread
    //
    assert(starterThread_ -> getId() != JTCThread::currentThread() -> getId());
    while(starterThread_ -> isAlive())
    {
	try
	{
	    starterThread_ -> join();
	}
	catch(const JTCInterruptedException&)
	{
	}
    }

    //
    // Call destroy on the super class
    //
    GIOPServerStarter::destroy();
}

void
GIOPServerStarterThreaded::starterRun()
{
    while(!destroy_)
    {
	try
	{
	    //
	    // Get new transport
	    //
	    OCI::Transport_var transport = acceptor_ -> accept();
	    assert(!is_nil(transport));
	    
	    //
	    // Check if thread should be stopped and destroyed
	    //
	    if(destroy_)
	    {
		transport -> close();
		return;
	    }
	    
	    //
	    // Create new worker
	    //
	    if(!is_nil(transport))
	    {
		JTCSynchronized synchronized(workersMutex_);

		workers_.append(new GIOPServerWorkerThreaded(this, transport,
                                                             upcall_));
	    }
	}
	catch(const NO_PERMISSION&)
	{
	    //
	    // Ignore NO_PERMISSION exceptions
	    //
	}
	catch(const SystemException&)
	{
	    //
	    // Destroy the worker
	    //
            // Not sure what to do here. Really destroy? Ignore?
	    //destroy();

	    return;
	}
    }
}

// ----------------------------------------------------------------------
// GIOPServerStarterThreaded::SenderThread public member implementations
// ----------------------------------------------------------------------

void
GIOPServerStarterThreaded::StarterThread::run()
{
    starter_ -> starterRun();
    // Break cyclic object dependency
    starter_ = GIOPServerStarterThreaded::_nil();
}

#endif
