// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_GIOP_SERVER_WORKER_H
#define OB_GIOP_SERVER_WORKER_H

namespace OB
{

class Upcall;

class GIOPServerWorker;
typedef GIOPServerWorker* GIOPServerWorker_ptr;

class GIOPServerWorkerBlocking;
typedef GIOPServerWorkerBlocking* GIOPServerWorkerBlocking_ptr;

class GIOPServerWorkerReactive;
typedef GIOPServerWorkerReactive* GIOPServerWorkerReactive_ptr;

#ifdef HAVE_JTC

class GIOPServerWorkerThreaded;
typedef GIOPServerWorkerThreaded* GIOPServerWorkerThreaded_ptr;

#endif

} // End of namespace OB

void OBDuplicate(OB::GIOPServerWorker_ptr);
void OBRelease(OB::GIOPServerWorker_ptr);
void OBDuplicate(OB::GIOPServerWorkerBlocking_ptr);
void OBRelease(OB::GIOPServerWorkerBlocking_ptr);
void OBDuplicate(OB::GIOPServerWorkerReactive_ptr);
void OBRelease(OB::GIOPServerWorkerReactive_ptr);

#ifdef HAVE_JTC

void OBDuplicate(OB::GIOPServerWorkerThreaded_ptr);
void OBRelease(OB::GIOPServerWorkerThreaded_ptr);

#endif

namespace OB
{

//
// The GIOPServerWorker class
//
typedef ObjVar< GIOPServerWorker > GIOPServerWorker_var;
typedef ObjSeq< GIOPServerWorker, int > GIOPServerWorkerSeq;

class GIOPServerWorker : public Downcall
{
    //
    // Hide copy-constructor and assignment operator
    //
    GIOPServerWorker(const GIOPServerWorker&);
    void operator=(const GIOPServerWorker&);

protected:

    GIOPServerStarter_var starter_; // The server starter
    OCI::Transport_var transport_; // The transport
    OCI::BufferSeq outSeq_; // The outgoing message buffers
    OCI::Buffer_var buf_; // The incoming message buffer
    GIOP::MessageHeader_1_0 hdr_; // The incoming message header
    Upcall* upcall_; // The upcall object for requests
    
    //
    // Add a "message error" message
    // Add a "close connection" message
    //
    void addMessageError();
    void addCloseConnection();

    //
    // Setup buf_ so that a new incoming message can be received
    //
    void setupBuffer();

    //
    // Extract the hdr_ from buf_
    //
    void extractHeader();

    //
    // Add outgoing message
    //
    virtual void add(OCI::Buffer_ptr);

    //
    // Execute incoming message
    //
    void execute();

    //
    // Dispatch a request
    //
    virtual void dispatchRequest(Buffer&, bool, OCI::Transport_ptr);

    void composeReply(GIOP::ReplyHeader&, Buffer&);

public:

    GIOPServerWorker(GIOPServerStarter_ptr, OCI::Transport_ptr, Upcall*);
    virtual ~GIOPServerWorker();

    static inline GIOPServerWorker_ptr _duplicate(GIOPServerWorker_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline GIOPServerWorker_ptr _nil()
    { return 0; }

    //
    // Destroy the worker
    //
    virtual void destroy();

    //
    // Downcall methods
    //
    void response(CORBA::ULong, Buffer&);
    void systemException(CORBA::ULong, Buffer&, const CORBA::SystemException&);
    void userException(CORBA::ULong, Buffer&);
    void forward(CORBA::ULong, Buffer&, CORBA::Object_ptr);
};

//
// The GIOPServerWorkerBlocking class
//
typedef ObjVar< GIOPServerWorkerBlocking > GIOPServerWorkerBlocking_var;

class GIOPServerWorkerBlocking : public GIOPServerWorker
{
    //
    // Hide copy-constructor and assignment operator
    //
    GIOPServerWorkerBlocking(const GIOPServerWorkerBlocking&);
    void operator=(const GIOPServerWorkerBlocking&);

public:

    GIOPServerWorkerBlocking(GIOPServerStarterBlocking_ptr,
			     OCI::Transport_ptr,
                             Upcall*);
    virtual ~GIOPServerWorkerBlocking();

    static inline GIOPServerWorkerBlocking_ptr
    _duplicate(GIOPServerWorkerBlocking_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline GIOPServerWorkerBlocking_ptr _nil()
    { return 0; }

    //
    // Get a request
    // Send a reply
    //
    void getRequest();
    void sendReply();
};

//
// The GIOPServerWorkerReactive class
//
typedef ObjVar< GIOPServerWorkerReactive > GIOPServerWorkerReactive_var;

class GIOPServerWorkerReactive : public GIOPServerWorker,
				 public EventHandler
{
    //
    // Hide copy-constructor and assignment operator
    //
    GIOPServerWorkerReactive(const GIOPServerWorkerReactive&);
    void operator=(const GIOPServerWorkerReactive&);

protected:

    bool destroy_; // True if destroy() was called

    //
    // Add outgoing message
    //
    virtual void add(OCI::Buffer_ptr);

public:

    GIOPServerWorkerReactive(GIOPServerStarterReactive_ptr,
			     OCI::Transport_ptr,
                             Upcall*);
    virtual ~GIOPServerWorkerReactive();

    static inline GIOPServerWorkerReactive_ptr
    _duplicate(GIOPServerWorkerReactive_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline GIOPServerWorkerReactive_ptr _nil()
    { return 0; }

    //
    // Destroy the worker
    //
    virtual void destroy();

    //
    // Handle event
    //
    virtual void handleEvent(Mask);
};

#ifdef HAVE_JTC

//
// The GIOPServerWorkerThreaded class
//
typedef ObjVar< GIOPServerWorkerThreaded > GIOPServerWorkerThreaded_var;

class GIOPServerWorkerThreaded : public GIOPServerWorker, public JTCMonitor
{
protected:

    bool destroy_; // True if destroy() was called

    //
    // The sender thread
    //
    class SenderThread : public JTCThread
    {
	GIOPServerWorkerThreaded_var worker_;

    public:
	
	SenderThread(GIOPServerWorkerThreaded_ptr worker)
	    : worker_(GIOPServerWorkerThreaded::_duplicate(worker)) { }
	virtual ~SenderThread() { }
	virtual void run();
    };
    friend class SenderThread;
    JTCThreadHandle senderThread_;

    //
    // The receiver thread
    //
    class ReceiverThread : public JTCThread
    {
	GIOPServerWorkerThreaded_var worker_;

    public:
	
	ReceiverThread(GIOPServerWorkerThreaded_ptr worker)
	    : worker_(GIOPServerWorkerThreaded::_duplicate(worker)) { }
	virtual ~ReceiverThread() { }
	virtual void run();
    };
    friend class ReceiverThread;
    JTCThreadHandle receiverThread_;

    //
    // Helper class for dispatching a request in case of a "thread per
    // request" concurrency model
    //
    class Dispatcher : public JTCThread
    {
	GIOPServerWorkerThreaded_var worker_;
	Buffer buf_;
	bool swap_;
        OCI::Transport_var transport_;

    public:

	Dispatcher(GIOPServerWorkerThreaded_ptr w, Buffer& b, bool s,
		   OCI::Transport_ptr t)
	    : worker_(GIOPServerWorkerThreaded::_duplicate(w)),
	      swap_(s), transport_(OCI::Transport::_duplicate(t))
	{
	    buf_.consume(b);
	}

	virtual ~Dispatcher()
	{
	}
       
	virtual void run()
	{
	    worker_ -> GIOPServerWorker::dispatchRequest(buf_, swap_,
							 transport_);
	}
    };
    friend class Dispatcher;

    //
    // Helper class for dispatching a request in case of a "thread
    // pool" concurrency model
    //
public: // For some compilers the following class must be public
    class ThreadPool : public JTCMonitor
    {
public: // For some compilers the following struct must be public
	struct Request
	{
	    GIOPServerWorkerThreaded_var worker;
	    Buffer buf;
	    bool swap;
	    OCI::Transport_var transport;
	    Request* next;
	};
private:
	Request* head_; // I can't use VarSeq<> because buf must not
	                // be copied (operator=()) for performance reasons

	void get(Request&);

public: // Must be public for some compilers, otherwise the "friend
        // class ThreadPool::Dispatcher;" declaration produces warnings
        class Dispatcher : public JTCThread
	{
	    ThreadPool* threadPool_;

	public:

	    Dispatcher(ThreadPool* threadPool)
		: threadPool_(threadPool)
	    {
	    }

	    virtual void
	    run()
	    {
		while(true)
		{
		    Request req;
		    threadPool_ -> get(req);
		    req.worker ->
			GIOPServerWorker::dispatchRequest(req.buf, req.swap,
							  req.transport);
		}
	    }
	};
private:
	friend class Dispatcher;

	ThreadPool();

    public:

	virtual ~ThreadPool(); // Must be virtual because of JTCMonitor d'tor

	static ThreadPool* instance();

	void add(GIOPServerWorkerThreaded_ptr, Buffer&, bool,
		 OCI::Transport_ptr);
    };
    friend class ThreadPool::Dispatcher;
private:

    //
    // Add outgoing message
    // Remove outgoing message
    //
    virtual void add(OCI::Buffer_ptr);
    virtual OCI::Buffer_ptr remove();

    //
    // Dispatch a request
    //
    virtual void dispatchRequest(Buffer&, bool, OCI::Transport_ptr);

public:

    GIOPServerWorkerThreaded(GIOPServerStarterThreaded_ptr,
			     OCI::Transport_ptr,
                             Upcall*);
    ~GIOPServerWorkerThreaded();

    static inline GIOPServerWorkerThreaded_ptr
    _duplicate(GIOPServerWorkerThreaded_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline GIOPServerWorkerThreaded_ptr _nil()
    { return 0; }

    //
    // Destroy the worker
    //
    virtual void destroy();

    //
    // Run method for receiver and sender thread
    //
    void senderRun();
    void receiverRun();
};

#endif

} // End of namespace OB


namespace CORBA
{
// XXX Are these necessary in CORBA?

inline void
release(OB::GIOPServerWorker_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::GIOPServerWorker_ptr p)
{
    return p == 0;
}

inline void
release(OB::GIOPServerWorkerBlocking_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::GIOPServerWorkerBlocking_ptr p)
{
    return p == 0;
}

inline void
release(OB::GIOPServerWorkerReactive_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::GIOPServerWorkerReactive_ptr p)
{
    return p == 0;
}

#ifdef HAVE_JTC

inline void
release(OB::GIOPServerWorkerThreaded_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::GIOPServerWorkerThreaded_ptr p)
{
    return p == 0;
}

#endif

} // End of namespace CORBA

#endif
