// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/TemplateI.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/Any.h>
#include <OB/OBV.h>
#include <OB/Principal.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/IFR1.h>
#include <OB/Policy.h>
#include <OB/ORB.h>
#include <OB/Server.h>
#include <OB/OCI.h>
#include <OB/Reactor.h>
#include <OB/Timer.h>
#include <OB/Properties.h>

#include <OB/POAServantBase.h>
#include <OB/NativeTypes.h>
#include <OB/POAInterface.h>
#include <OB/OCI.h>

#include <OB/POAServantLocator.h> // For PortableServer::ServantLocator::Cookie
#include <OB/POACurrent.h> // For PortableServer::Current

#include <OB/Reactor.h>

#include <OCI_impl.h> // For new OCI_ConFactoryRegistry_impl
                      // and OCI_Current_impl

#include <POACurrent_impl.h>
#include <POA_impl.h>

#include <GIOPServer.h>

#include <Upcall.h>
#include <OARoot.h>

using namespace CORBA;
using namespace OB;

//
// The server concurrency model.
//
ORB::ServerConcModel OARoot::serverConcModel_ =
    ORB::ServerConcModelReactive;

//
// The number of threads in the pool.
//
ULong OARoot::serverConcModelThreadPool_ = 10;

//
// The singleton
//
OARoot* OARoot::instance_ = 0;

// ----------------------------------------------------------------------
// Template instantiations
// ----------------------------------------------------------------------

#ifndef HAVE_NO_EXPLICIT_TEMPLATES
template class OB::ObjVar< Server >;
template class OB::ObjForSeq< Server >;
template class OB::ObjSeq< Server, int >;
template class OB::SeqVar< OB::ObjSeq< Server, int > >;
#else
#ifdef HAVE_PRAGMA_DEFINE
#pragma define(OB::ObjVar< Server >)
#pragma define(OB::ObjForSeq< Server >)
#pragma define(OB::ObjSeq< Server, int >)
#pragma define(OB::SeqVar< OB::ObjSeq< Server, int > >)
#endif
#endif

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::Server_ptr p)
{
    Server::_duplicate(p);
}

void
OBRelease(OB::Server_ptr p)
{
    release(p);
}

// ----------------------------------------------------------------------
// OARoot constructor and destructor
// ----------------------------------------------------------------------

OARoot::OARoot()
    : serversInitialized_(false),
      rootPOA_(0)
{
    assert(instance_ == 0);
    registry_ = new OCI::AccRegistry_impl();
    instance_ = this;
}

OARoot::~OARoot()
{
    ULong i;

    //
    // Destroy all servers
    //
    for(i = 0; i < servers_.length(); i++)
	servers_[i] -> destroy();

    release(rootPOA_);

    instance_ = 0;
}

// ----------------------------------------------------------------------
// OARoot public member implementation
// ----------------------------------------------------------------------

OARoot*
OARoot::instance()
{
    assert(instance_ != 0);
    return instance_;
}

bool
OARoot::initServers()
{
    if(serversInitialized_)
        return false;
    
#ifdef HAVE_JTC
    JTCSynchronized synchronized(serversMonitor_);
#endif

    createNewServers();

    bool allDone = true;
    for(ULong i = 0 ; i < servers_.length() ; i++)
    {
	bool done = servers_[i] -> run();
	allDone = allDone && done;
    }
    
    return allDone;
}

void
OARoot::createNewServers()
{
#ifdef HAVE_JTC
    JTCSynchronized synchronized(serversMonitor_);
#endif
    OCI::AccRegistry_var registry =
        ORB::_OB_instance_nodup() -> get_acc_registry();
    assert(!is_nil(registry));

    //
    // This creates servers for any new acceptors. It is known that
    // acceptors are only appended to the acceptors list, so we can
    // check the current length of the servers list, and the new
    // length of the acceptors list, and create servers for any newly
    // appended acceptors.
    //
    // Another way of coding this would be to look at the tags of the
    // known acceptors, and known servers can create new server for an
    // unknown acceptor tag. Of course, this assumes that the set of
    // tags is unique (must be true anyway).
    //
    OCI::AcceptorSeq_var acceptors = registry -> get_acceptors();
    if(servers_.length() < acceptors -> length())
    {
        int save = servers_.length();
        servers_.length(acceptors -> length());
        for(ULong i = save ; i < servers_.length() ; i++)
        {
            servers_[i] = new GIOPServer(acceptors[i], this);
        }
    }
}

void
OARoot::destroyServers()
{
#ifdef HAVE_JTC
    JTCSynchronized synchronized(serversMonitor_);
#endif

    //
    // Destroy and release all servers
    //
    for(ULong i = 0 ; i < servers_.length() ; i++)
	servers_[i] -> destroy();

    servers_.length(0);
}

void
OARoot::run()
{
#ifdef HAVE_JTC
    JTCSynchronized synchronized(serversMonitor_);
#endif

    //
    // Initialize servers if not already done
    //
    if(initServers())
        return; // All done

    //
    // If reactive mode is selected (client- or server-side), then we
    // must dispatch events using the reactor. Otherwise we can wait
    // until all servers are released.
    //
    if(ORB::conc_model() == ORB::ConcModelReactive ||
       serverConcModel() == ORB::ServerConcModelReactive)
    {
	//
	// Use the reactor to dispatch requests
	//
	Reactor::instance() -> dispatch();
    }
#ifdef HAVE_JTC
    else
    {
	//
	// Wait for all servers to be released
	//
	while(servers_.length() > 0)
	{
	    try
	    {
		serversMonitor_.wait();
	    }
	    catch(const JTCInterruptedException&)
	    {
	    }
	}
    }
#endif
}

void
OARoot::shutdown(Boolean)
{
    //
    // If the ORB::shutdown operation is called, it makes a call on
    // deactivate with a TRUE etherealize_objects parameter for each
    // POA manager known in the process; the wait_for_completion
    // parameter to deactivate will be the same as the similarly named
    // parameter of ORB::shutdown.
    //
    
#ifdef HAVE_JTC
    JTCSynchronized synchronized(serversMonitor_);
#endif

    //
    // This should also destroy the POAs
    //
    
    //
    // Destroy servers
    //
    destroyServers();
    assert(servers_.length() == 0);

    //
    // If reactive mode is selected (client- or server-side), then we
    // must tell the reactor to stop dispatching events. Otherwise we
    // notify the thread that waits for the servers to be released.
    //
    if(ORB::conc_model() == ORB::ConcModelReactive ||
       serverConcModel() == ORB::ServerConcModelReactive)
    {
	Reactor::instance() -> stop();
    }
#ifdef HAVE_JTC
    else
    {
	serversMonitor_.notify();
    }
#endif
}

bool
OARoot::workPending()
{
    if(initServers())
        return false;
    
    //
    // If we're running in reactive mode then always return true and
    // use the reactor to dispatch one event.
    //
    if(ORB::conc_model() == ORB::ConcModelReactive ||
       serverConcModel() == ORB::ServerConcModelReactive)
    {
        return Reactor::instance() -> eventReady();
    }

    return false;
}

void
OARoot::performWork()
{
    if(initServers())
        return;
    
    //
    // Use the reactor to dispatch one event
    //
    if(ORB::conc_model() == ORB::ConcModelReactive ||
       serverConcModel() == ORB::ServerConcModelReactive)
    {
        Reactor::instance() -> dispatchOneEvent(0);
    }
}

void
OARoot::invoke(Downcall_ptr downcall,
               ULong requestId,
               const OctetSeq& key,
               const char* op,
               Buffer& buf,
               bool sw,
               ULong offIn,
               ULong offOut,
               OCI::TransportInfo_ptr info)
{
    assert(rootPOA_ != 0);
    rootPOA_ -> _OB_invoke(downcall, requestId,
                           key, op, buf, sw, offIn, offOut,
                           info);
}

bool
OARoot::findByKey(const OctetSeq& key)
{
    assert(rootPOA_ != 0);
    return rootPOA_ -> _OB_findByKey(key);
}

void
OARoot::serverConcModel(ORB::ServerConcModel conc)
{
    serverConcModel_ = conc;
}

ORB::ServerConcModel
OARoot::serverConcModel()
{
    return serverConcModel_;
}

void
OARoot::serverThreadPool(ULong p)
{
    serverConcModelThreadPool_ = p;
}

ULong
OARoot::serverThreadPool()
{
    return serverConcModelThreadPool_;
}

OCI::AccRegistry_ptr
OARoot::getAccRegistry()
{
    return OCI::AccRegistry::_duplicate(registry_);
}

PortableServer::POA_ptr
OARoot::getRootPOA()
{
#ifdef HAVE_JTC
    JTCSynchronized sync(poaMutex_);
#endif
    if(rootPOA_ == 0)
    {
        rootPOA_ = new PortableServer::POA_impl();
        ORB::_OB_instance_nodup() -> _OB_initializeRootPOA(rootPOA_);
    }
    return PortableServer::POA::_duplicate(rootPOA_);
}

bool
OARoot::hasRootPOA() const
{
    return rootPOA_ != 0;
}
