// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/Policy.h>
#include <OB/OBPolicies.h>
#include <OB/GIOP.h>
#include <OB/IIOP.h>
#include <OB/OCI.h>
#include <OB/OCI_IIOP.h>
#include <OB/Util.h>
#include <OB/Net.h>
#include <OB/Stream.h>

#include <OCI_IIOP_impl.h>
#include <IIOPUtil.h>

//
// Undefine minor (this is needed for SunOS 4.1 and some other Unix flavors)
//
#undef minor

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;
using namespace OCI::IIOP;
using namespace OB;

//#define OB_OCI_DEBUG

// ----------------------------------------------------------------------
// Some utility functions
// ----------------------------------------------------------------------

IOP::IOR*
OB::IIOPCreateIOR(const char* host, UShort port,
		  const char* name, const char* id)
{
    //
    // Create a new profile
    //
    ::IIOP::ProfileBody_1_1 body;
    body.iiop_version.major = 1;
    body.iiop_version.minor = 1;
    body.host = host;
    body.port = port;

    //
    // Makeup object key
    //
    // I don't use string marshal functions here because I don't
    // want to have a string length encoded. That would cause byte
    // order dependencies.
    //
    body.object_key.length(strlen(name) + 1);
    strcpy((char*)body.object_key.get_buffer(), name);

    //
    // Create new IOR with the new profile
    //
    IOP::IOR_var ior = new IOP::IOR;
    ior -> type_id = id;
    ior -> profiles.length(1);
    ior -> profiles[0].tag = IOP::TAG_INTERNET_IOP;
    CountStream count;
    count.count_boolean(); // Endian
    ::IIOP::ProfileBody_1_1::_OB_marshalCount(body, count);
    ior -> profiles[0].profile_data.length(count.count());
    Octet* oct = ior -> profiles[0].profile_data.get_buffer();
#ifdef OB_CLEAR_MEM
    memset(oct, 0, count.count());
#endif
    OutputStream out(oct, 0, count.count());
    out.write_boolean(Endian);
    ::IIOP::ProfileBody_1_1::_OB_marshal(body, out);

    return ior._retn();
}

void
OB::IIOPPrintProfileBody(ostream& out, const IOP::TaggedProfile& profile)
{
    //
    // Get the IIOP profile body
    //
    const Octet* coct = profile.profile_data.get_buffer();
    ULong len = profile.profile_data.length();
    InputStream in(coct, 0, len, false);
    in._OB_readEndian();
    ::IIOP::ProfileBody_1_0 body;
    ::IIOP::ProfileBody_1_0::_OB_unmarshal(body, in);

    //
    // Print general info
    //
    out << "iiop_version: " << int(body.iiop_version.major) << '.'
	<< int(body.iiop_version.minor) << '\n';
    out << "host: " << body.host << '\n';
    out << "port: " << body.port << '\n';
    out << "object_key:\n";
    PrintOctets(out, body.object_key.get_buffer(),
		body.object_key.length());
}

Boolean
OB::IIOPCompareBodies(const ::IIOP::ProfileBody_1_0& body1,
		      const ::IIOP::ProfileBody_1_0& body2)
{
    //
    // Compare versions
    //
    if(body1.iiop_version.major != body2.iiop_version.major ||
       body1.iiop_version.minor != body2.iiop_version.minor)
	return FALSE;
    
    //
    // Compare ports
    //
    if(body1.port != body2.port)
	return FALSE;
    
    //
    // Compare object keys
    //
    if(body1.object_key.length() != body2.object_key.length())
	return FALSE;
    
    if(memcmp(body1.object_key.get_buffer(), body2.object_key.get_buffer(),
	      body1.object_key.length()) != 0)
	return FALSE;
    
    //
    // Direct host name comparison
    //
    if(strcmp(body1.host, body2.host) != 0)
    {
	//
	// Direct host name comparision failed - must look up
	// addresses to be really sure if the hosts differ
	//
	try
	{
	    struct sockaddr_in addr1;
	    GetInAddr(body1.host, addr1);
	    
	    struct sockaddr_in addr2;
	    GetInAddr(body2.host, addr2);
	    
	    if(addr1.sin_addr.s_addr != addr2.sin_addr.s_addr)
		return FALSE;
	}
	catch(const COMM_FAILURE&)
	{
	    //
	    // Return false on hostname lookup failure
	    //
	    return FALSE;
	}
    }
    
    //
    // OK, found a match
    //
    return TRUE;
}

Boolean
OB::IIOPEquivalent(const OCI::IOR& ior1, const OCI::IOR& ior2)
{
    ULong p1, p2, b1, b2;
    ULong cnt1 = 0, cnt2 = 0;
    ::IIOP::ProfileBody_1_0** bodies1 = 0;
    ::IIOP::ProfileBody_1_0** bodies2 = 0;

    //
    // Calculate number of IIOP profiles in ior1
    //
    for(p1 = 0 ; p1 < ior1.profiles.length() ; p1++)
	if(ior1.profiles[p1].tag == IOP::TAG_INTERNET_IOP)
	    cnt1++;
    
    //
    // Calculate number of IIOP profiles in ior2
    //
    for(p2 = 0 ; p2 < ior2.profiles.length() ; p2++)
	if(ior2.profiles[p2].tag == IOP::TAG_INTERNET_IOP)
	    cnt2++;

    //
    // Return false now if the number of IIOP profile bodies do not
    // match
    //
    if(cnt1 != cnt2)
	return FALSE;

    //
    // Create an array with all IIOP profile bodies of ior1
    //
    bodies1 = new ::IIOP::ProfileBody_1_0*[cnt1];
    for(p1 = 0, b1 = 0 ; p1 < ior1.profiles.length() ; p1++)
	if(ior1.profiles[p1].tag == IOP::TAG_INTERNET_IOP)
	{
	    const Octet* coct =
		ior1.profiles[p1].profile_data.get_buffer();
            ULong len = ior1.profiles[p1].profile_data.length();
            InputStream in(coct, 0, len, false);
            in._OB_readEndian();
	    ::IIOP::ProfileBody_1_0* p = new ::IIOP::ProfileBody_1_0;
	    ::IIOP::ProfileBody_1_0::_OB_unmarshal(*p, in);
	    bodies1[b1++] = p;
	}
    
    assert(b1 == cnt1);
    
    //
    // Create an array with all IIOP profile bodies of ior2
    //
    bodies2 = new ::IIOP::ProfileBody_1_0*[cnt2];
    for(p2 = 0, b2 = 0 ; p2 < ior2.profiles.length() ; p2++)
	if(ior2.profiles[p2].tag == IOP::TAG_INTERNET_IOP)
	{
	    const Octet* coct =
		ior2.profiles[p2].profile_data.get_buffer();
	    ULong len = ior2.profiles[p2].profile_data.length();
            InputStream in(coct, 0, len, false);
            in._OB_readEndian();
	    ::IIOP::ProfileBody_1_0* p = new ::IIOP::ProfileBody_1_0;
	    ::IIOP::ProfileBody_1_0::_OB_unmarshal(*p, in);
	    bodies2[b2++] = p;
	}
    
    assert(b2 == cnt2);

    //
    // Check for profile body matches
    //
    for(b1 = 0 ; b1 < cnt1 ; b1++)
    {
	for(b2 = 0 ; b2 < cnt2 ; b2++)
	{
	    if(!bodies2[b2])
		continue;
	    
	    //
	    // Compare profile bodies
	    //
	    if(!IIOPCompareBodies(*bodies1[b1], *bodies2[b2]))
		continue;
	    
	    //
	    // OK, found a match
	    //
	    delete bodies1[b1];
	    bodies1[b1] = 0;
	    delete bodies2[b2];
	    bodies2[b2] = 0;
	}
    }

    //
    // Check whether there are any unmatched IIOP profile bodies
    //
    ULong unmatched1 = 0;
    for(b1 = 0 ; b1 < cnt1 ; b1++)
	if(bodies1[b1])
	{
	    unmatched1++;
	    delete bodies1[b1];
	}

    delete [] bodies1;

    ULong unmatched2 = 0;
    for(b2 = 0 ; b2 < cnt2 ; b2++)
	if(bodies2[b2])
	{
	    unmatched2++;
	    delete bodies2[b2];
	}

    delete [] bodies2;

    return unmatched1 == 0 && unmatched2 == 0;
}

ULong
OB::IIOPHash(const OCI::IOR& ior, ULong maximum)
{
    ULong hash = 0;

    for(ULong i = 0 ; i < ior.profiles.length() ; i++)
    {
	if(ior.profiles[i].tag == IOP::TAG_INTERNET_IOP)
	{
	    //
	    // Get the IIOP profile body
	    //
	    const Octet* coct = ior.profiles[i].profile_data.get_buffer();
            ULong len = ior.profiles[i].profile_data.length();
            InputStream in(coct, 0, len, false);
            in._OB_readEndian();
	    ::IIOP::ProfileBody_1_0 body;
	    ::IIOP::ProfileBody_1_0::_OB_unmarshal(body, in);

	    //
	    // Add port to hash
	    //
	    hash ^= body.port;

	    //
	    // Add object key to hash
	    //
	    for(ULong j = 0 ; j + 1 < body.object_key.length() ; j += 2)
		hash ^= body.object_key[j + 1] * 256 + body.object_key[j];
	}
    }

    return hash % (maximum + 1);
}

OCI::ObjectKey*
OB::IIOPExtractKey(const OCI::IOR& ior, const char* host, UShort port,
		   bool loopbackMatches)
{
    for(ULong i = 0 ; i < ior.profiles.length() ; i++)
    {
	if(ior.profiles[i].tag == IOP::TAG_INTERNET_IOP)
	{
	    //
	    // Get the IIOP profile body
	    //
	    const Octet* coct = ior.profiles[i].profile_data.get_buffer();
            ULong len = ior.profiles[i].profile_data.length();
            InputStream in(coct, 0, len, false);
            in._OB_readEndian();
	    ::IIOP::ProfileBody_1_0 body;
	    ::IIOP::ProfileBody_1_0::_OB_unmarshal(body, in);

	    //
	    // Compare ports
	    //
	    if(port != body.port)
		continue;
	    
	    //
	    // Direct host name comparison
	    //
	    if(strcmp(host, body.host) != 0)
	    {
		//
		//
		// Direct host name comparision failed - must look up
		// addresses to be really sure if the hosts differ
		//
		try
		{
		    struct sockaddr_in addr1;
		    GetInAddr(host, addr1);
		    
		    struct sockaddr_in addr2;
		    GetInAddr(body.host, addr2);
		    
		    if(addr1.sin_addr.s_addr != addr2.sin_addr.s_addr)
		    {
			//
			// Address comparison failed - shall I extract
			// the key if the profile body contains the
			// loopback address?
			//
			if(loopbackMatches)
			{
			    if(addr2.sin_addr.s_addr != htonl(INADDR_LOOPBACK))
			    {
				continue;
			    }
			}
			else
			    continue;
		    }
		}
		catch(const COMM_FAILURE&)
		{
		    //
		    // Continue on hostname lookup failure
		    //
		    continue;
		}
	    }
	    
	    //
	    // OK, found a match
	    //
	    OCI::ObjectKey_var key =
                new OCI::ObjectKey(body.object_key.length());
            key -> length(body.object_key.length());
	    memcpy(key -> get_buffer(), body.object_key.get_buffer(),
                   key -> length());
	    return key._retn();
	    //return new OCI::ObjectKey(body.object_key);
	}
    }

    //
    // No match found
    //
    return new OCI::ObjectKey();
}

// ----------------------------------------------------------------------
// Transport_impl private and protected member implementations
// ----------------------------------------------------------------------

void
Transport_impl::setBlock(Boolean block)
{
    if(block != block_)
    {
	block_ = block;

	if(block_)
	    SetBlock(fd_);
	else
	    SetNoBlock(fd_);
    }
}

// ----------------------------------------------------------------------
// Transport_impl constructor and destructor
// ----------------------------------------------------------------------

Transport_impl::Transport_impl(Connector_impl* con,
			       OCI::Handle fd,
			       Boolean block)
    : fd_(fd), block_(block),
      info_(new TransportInfo_impl(this, con))
{
#ifdef OB_OCI_DEBUG
    cout << "Transport_impl" << endl;
#endif
}

Transport_impl::Transport_impl(Acceptor_impl* acc,
			       OCI::Handle fd,
			       Boolean block)
    : fd_(fd), block_(block),
      info_(new TransportInfo_impl(this, acc))
{
#ifdef OB_OCI_DEBUG
    cout << "Transport_impl" << endl;
#endif
}

Transport_impl::~Transport_impl()
{
#ifdef OB_OCI_DEBUG
    cout << "~Transport_impl" << endl;
#endif

    if(fd_ >= 0)
	close();

    release(info_);
}

// ----------------------------------------------------------------------
// Transport_impl public member implementation
// ----------------------------------------------------------------------

void
Transport_impl::close()
    throw(SystemException)
{
#ifdef OB_TRACE
    {
	OCI::AcceptorInfo_var accInfo = info_ -> acceptor_info();
	String_var msg;
	if(!is_nil(accInfo))
	    msg = string_dup("Closing connection from ");
	else
	    msg = string_dup("Closing connection to ");
        InetAddr_var addr = info_ -> remote_addr();
        msg += (int)addr[0];
        msg += ".";
        msg += (int)addr[1];
        msg += ".";
        msg += (int)addr[2];
        msg += ".";
        msg += (int)addr[3];
        msg += ":";
        msg += info_ -> remote_port();
        MessageViewer::instance() -> trace(1, msg);
    }
#endif
    
    //
    // Call callbacks
    //
    info_ -> _OB_callCloseCB(info_);

    //
    // Destroy the info object
    //
    info_ -> _OB_destroy();

    //
    // Close the socket
    //
    int saveFd = fd_;
    fd_ = -1; // Must be set to -1 before the close
    CloseSocket(saveFd);
    block_ = TRUE;
}

void
Transport_impl::shutdown()
    throw(SystemException)
{
    //
    // Shutdown the socket
    //
    ::shutdown(fd_, 2);
}

void
Transport_impl::receive(OCI::Buffer_ptr buf, Boolean block)
    throw(SystemException)
{
    setBlock(block);

    while(!buf -> is_full())
    {
	int result = ::recv(fd_, (char*)buf -> rest(),
			    buf -> rest_length(), 0);

	if(result == 0)
	{
	    throw COMM_FAILURE("", MinorRecvZero, COMPLETED_NO);
	}
	else if(result == -1)
	{
            if(IsInterrupted())
                continue;
            
	    if(!block && WouldBlock())
		return;

	    throw COMM_FAILURE(LastError(), MinorRecv, COMPLETED_NO);
	}

	buf -> advance(result);
    }
}

Boolean
Transport_impl::receive_detect(OCI::Buffer_ptr buf, Boolean block)
    throw(SystemException)
{
    setBlock(block);

    while(!buf -> is_full())
    {
        int result = ::recv(fd_, (char*)buf -> rest(),
			    buf -> rest_length(), 0);

	if(result == 0)
	{
	    return FALSE;
	}
	else if(result == -1)
	{
            if(IsInterrupted())
                continue;
            
	    if(ConnectionDown())
		return FALSE;

	    if(!block && WouldBlock())
		return TRUE;

	    throw COMM_FAILURE(LastError(), MinorRecv, COMPLETED_NO);
	}

	buf -> advance(result);
    }

    return TRUE;
}

void
Transport_impl::receive_timeout(OCI::Buffer_ptr buf, ULong t)
    throw(SystemException)
{
    if(t == 0)
    {
	receive(buf, FALSE);
	return;
    }

    setBlock(FALSE);

    while(!buf -> is_full())
    {
	fd_set fdSet;
	FD_ZERO(&fdSet);
	FD_SET(fd_, &fdSet);
	
	struct timeval tv;
	tv.tv_sec = t / 1000;
	tv.tv_usec = (t - tv.tv_sec * 1000) * 1000;

	int resultS = ::select(fd_ + 1, &fdSet, 0, 0, &tv);
	
	if(resultS == 0)
	{
	    return;
	}
	else if(resultS == -1)
	{
            if(IsInterrupted())
                continue;
            
	    throw COMM_FAILURE(LastError(), MinorSelect, COMPLETED_NO);
	}

        int result = ::recv(fd_, (char*)buf -> rest(),
			    buf -> rest_length(), 0);

        if(result == 0)
        {
            throw COMM_FAILURE("", MinorRecvZero, COMPLETED_NO);
        }
        else if(result == -1)
        {
            if(IsInterrupted())
                continue;
            
            if(WouldBlock())
                return;

            throw COMM_FAILURE(LastError(), MinorRecv, COMPLETED_NO);
        }

        buf -> advance(result);
    }
}

void
Transport_impl::send(OCI::Buffer_ptr buf, Boolean block)
    throw(SystemException)
{
    setBlock(block);

    while(!buf -> is_full())
    {
	int result = ::send(fd_, (const char*)buf -> rest(),
			    buf -> rest_length(), 0);

	if(result == 0)
	{
	    throw COMM_FAILURE("", MinorSendZero, COMPLETED_NO);
	}
	else if(result == -1)
	{
            if(IsInterrupted())
                continue;
            
	    if(!block && WouldBlock())
		return;

	    throw COMM_FAILURE(LastError(), MinorSend, COMPLETED_NO);
	}

	buf -> advance(result);
    }
}

Boolean
Transport_impl::send_detect(OCI::Buffer_ptr buf, Boolean block)
    throw(SystemException)
{
    setBlock(block);

    while(!buf -> is_full())
    {
	int result = ::send(fd_, (const char*)buf -> rest(),
			    buf -> rest_length(), 0);

	if(result == 0)
	{
	    return FALSE;
	}
	else if(result == -1)
	{
            if(IsInterrupted())
                continue;
            
	    if(ConnectionDown())
		return FALSE;

	    if(!block && WouldBlock())
		return TRUE;

	    throw COMM_FAILURE(LastError(), MinorSend, COMPLETED_NO);
	}

	buf -> advance(result);
    }

    return TRUE;
}

void
Transport_impl::send_timeout(OCI::Buffer_ptr buf, ULong t)
    throw(SystemException)
{
    if(t == 0)
    {
	send(buf, FALSE);
	return;
    }

    setBlock(FALSE);
    
    while(!buf -> is_full())
    {
	fd_set fdSet;
	FD_ZERO(&fdSet);
	FD_SET(fd_, &fdSet);
	
	struct timeval tv;
	tv.tv_sec = t / 1000;
	tv.tv_usec = (t - tv.tv_sec * 1000) * 1000;

	int resultS = ::select(fd_ + 1, 0, &fdSet, 0, &tv);
	
	if(resultS == 0)
	{
	    return;
	}
	else if(resultS == -1)
	{
            if(IsInterrupted())
                continue;
	    
	    throw COMM_FAILURE(LastError(), MinorSelect, COMPLETED_NO);
	}

        int result = ::send(fd_, (const char*)buf -> rest(),
                            buf -> rest_length(), 0);

        if(result == 0)
        {
            throw COMM_FAILURE("", MinorSendZero, COMPLETED_NO);
        }
        else if(result == -1)
        {
            if(IsInterrupted())
                continue;
            
            if(WouldBlock())
                return;

            throw COMM_FAILURE(LastError(), MinorSend, COMPLETED_NO);
        }

        buf -> advance(result);
    }
}

OCI::TransportInfo_ptr
Transport_impl::get_info()
    throw(SystemException)
{
    return OCI::TransportInfo::_duplicate(info_);
}

// ----------------------------------------------------------------------
// Connector_impl private and protected member implementations
// ----------------------------------------------------------------------

void
Connector_impl::close()
{
    //
    // Destroy the info object
    //
    info_ -> _OB_destroy();

    //
    // Close the socket
    //
    int saveFd = fd_;
    fd_ = -1; // Must be set to -1 before the close
    CloseSocket(saveFd);
}

// ----------------------------------------------------------------------
// Connector_impl constructor and destructor
// ----------------------------------------------------------------------

Connector_impl::Connector_impl(const char* host,
			       UShort port,
			       const OCI::ConnectCBSeq& cb)
    : host_(host), port_(port), fd_(-1),
      info_(new ConnectorInfo_impl(this, cb))
{
#ifdef OB_OCI_DEBUG
    cout << "Connector_impl" << endl;
#endif

    //
    // Get address
    //
    memset(&addr_, 0, sizeof(addr_));
    addr_.sin_family = AF_INET;
    addr_.sin_port = htons(port_);
    try
    {
        GetInAddr(host_, addr_);
    }
    catch(const SystemException&)
    {
        _OB_setRef(0);
        throw;
    }
}

Connector_impl::~Connector_impl()
{
#ifdef OB_OCI_DEBUG
    cout << "~Connector_impl" << endl;
#endif

    if(fd_ >= 0)
	close();

    release(info_);
}

// ----------------------------------------------------------------------
// Connector_impl public member implementation
// ----------------------------------------------------------------------

OCI::Transport_ptr
Connector_impl::connect()
    throw(SystemException)
{
    if(fd_ >= 0)
	close();

    //
    // Create socket
    //
    if((fd_ = CreateSocket()) == -1)
    {
	fd_ = -1;
	throw COMM_FAILURE(LastError(), MinorSocket, COMPLETED_NO);
    }

    //
    // Set TCP_NODELAY option
    //
    int no = 1;
    if(setsockopt(fd_, IPPROTO_TCP, TCP_NODELAY, (char*)&no, sizeof(int))
       == -1)
    {
	int err = GetError();
	CloseSocket(fd_);
	fd_ = -1;
	SetError(err);
	throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
    }
   
    //
    // Connect
    //
#ifdef OB_TRACE
    {
        String_var msg = string_dup("New connection to ");
        msg += inet_ntoa(addr_.sin_addr);
        msg += ":";
        msg += ntohs(addr_.sin_port);
        MessageViewer::instance() -> trace(1, msg);
    }
#endif
    while(true)
    {
        if(::connect(fd_, (struct sockaddr*)&addr_, sizeof(addr_)) == -1)
        {
            if(IsInterrupted())
                continue;
	    
	    int err = GetError();
	    CloseSocket(fd_);
	    fd_ = -1;
	    SetError(err);
	    throw COMM_FAILURE(LastError(), MinorConnect, COMPLETED_NO);
        }

        break;
    }
    
    //
    // Create new transport (the last argument TRUE means blocking)
    //
    OCI::Transport_var tr = new Transport_impl(this, fd_, TRUE);
    fd_ = -1;

    //
    // Call callbacks
    //
    OCI::TransportInfo_var trInfo = tr -> get_info();
    try
    {
	info_ -> _OB_callConnectCB(trInfo);
    }
    catch(const SystemException&)
    {
	tr -> close();
	throw;
    }

    //
    // Return new transport
    //
    return tr._retn();
}

OCI::Transport_ptr
Connector_impl::connect_timeout(ULong t)
    throw(SystemException)
{
    if(fd_ >= 0)
	close();

    //
    // Create socket
    //
    if((fd_ = CreateSocket()) == -1)
    {
	fd_ = -1;
	throw COMM_FAILURE(LastError(), MinorSocket, COMPLETED_NO);
    }

    //
    // Set TCP_NODELAY option
    //
    int no = 1;
    if(setsockopt(fd_, IPPROTO_TCP, TCP_NODELAY, (char*)&no, sizeof(int))
       == -1)
    {
	int err = GetError();
	CloseSocket(fd_);
	fd_ = -1;
	SetError(err);
	throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
    }

    //
    // Set socket to non-blocking mode
    //
    SetNoBlock(fd_);
   
    //
    // Connect
    //
#ifdef OB_TRACE
    {
        String_var msg = string_dup("New connection to ");
        msg += inet_ntoa(addr_.sin_addr);
        msg += ":";
        msg += ntohs(addr_.sin_port);
        MessageViewer::instance() -> trace(1, msg);
    }
#endif
    while(true)
    {
        if(::connect(fd_, (struct sockaddr*)&addr_, sizeof(addr_)) == -1)
        {
            if(IsInterrupted())
                continue;

	    if(InProgress())
	    {
		//
		// Call select() until timeout expired or connection
		// fd_ becomes writeable.
		//
		while(true)
		{
		    fd_set writeFdSet;
		    FD_ZERO(&writeFdSet);
		    FD_SET(fd_, &writeFdSet);

		    //
		    // Under WIN32 ::connect failure notification is
		    // done through the exceptfds.
		    //
		    fd_set exceptFdSet;
		    FD_ZERO(&exceptFdSet);
		    FD_SET(fd_, &exceptFdSet);
		    
		    struct timeval tv;
		    tv.tv_sec = t / 1000;
		    tv.tv_usec = (t - tv.tv_sec * 1000) * 1000;

		    int resultS = ::select(fd_ + 1, 0, &writeFdSet,
					   &exceptFdSet, &tv);
		    
		    if(resultS == 0)
		    {
			CloseSocket(fd_);
			fd_ = -1;
			return OCI::Transport::_nil();
		    }
		    else if(resultS == -1)
		    {
			if(IsInterrupted())
			    continue;
			
			throw COMM_FAILURE(LastError(), MinorSelect,
					   COMPLETED_NO);
		    }
		    
		    //
		    // Check for connect error
		    //
#ifdef HAVE_GETSOCKNAME_WITH_SIZE_T
		    size_t optlen = sizeof(int);
#else
		    int optlen = sizeof(int);
#endif
		    int optval;
		    if(getsockopt(fd_, SOL_SOCKET, SO_ERROR, (char*)&optval,
				  &optlen) == -1)
		    {
			int err = GetError();
			CloseSocket(fd_);
			fd_ = -1;
			SetError(err);
			throw COMM_FAILURE(LastError(), MinorGetsockopt,
					   COMPLETED_NO);
		    }
		    
		    if(optval > 0)
		    {
			CloseSocket(fd_);
			fd_ = -1;
			SetError(optval);
			throw COMM_FAILURE(LastError(), MinorConnect,
					   COMPLETED_NO);
		    }
		    
#ifdef WIN32
		    //
		    // Hack for WIN32: Ensure the connection actually
		    // worked. We do this by calling getpeername(). If the
		    // call succeeds, then we're fully connected.
		    //
#ifdef HAVE_GETSOCKNAME_WITH_SIZE_T
		    size_t addrLen = sizeof(struct sockaddr_in);
#else
		    int addrLen = sizeof(struct sockaddr_in);
#endif
		    struct sockaddr_in newAddr;
		    if(getpeername(fd_, (struct sockaddr*)&newAddr, &addrLen)
		       != 0)
			continue;
#endif
		    break;
		}
	    }
	    else
	    {
		int err = GetError();
		CloseSocket(fd_);
		fd_ = -1;
		SetError(err);
		throw COMM_FAILURE(LastError(), MinorConnect, COMPLETED_NO);
	    }
        }
	
        break;
    }
    
    //
    // Create new transport (the last argument FALSE means non-blocking)
    //
    OCI::Transport_var tr = new Transport_impl(this, fd_, FALSE);
    fd_ = -1;

    //
    // Call callbacks
    //
    OCI::TransportInfo_var trInfo = tr -> get_info();
    try
    {
	info_ -> _OB_callConnectCB(trInfo);
    }
    catch(const SystemException&)
    {
	tr -> close();
	throw;
    }

    //
    // Return new transport
    //
    return tr._retn();
}

OCI::ObjectKey*
Connector_impl::is_usable(const OCI::IOR& ior)
    throw(SystemException)
{
    return IIOPExtractKey(ior, host_, port_, false);
}

OCI::ObjectKey*
Connector_impl::is_usable_with_policies(const OCI::IOR& ior,
                                        const PolicyList& policies)
    throw(SystemException)
{
    //
    // Make sure that the set of policies is met
    //
    for(ULong i = 0; i < policies.length(); i++)
    {
        if(policies[i] -> policy_type() == SSL::CONNECT_POLICY)
        {
            SSL::ConnectPolicy_var connectPolicy =
                SSL::ConnectPolicy::_narrow(policies[i]);

	    assert(!is_nil(connectPolicy));

            if(connectPolicy -> value() == SSL::ConnectSecure)
                return new OCI::ObjectKey();
        }
        else if(policies[i] -> policy_type() == OB::PROTOCOL_POLICY)
        {
            OB::ProtocolPolicy_var protocolPolicy =
                OB::ProtocolPolicy::_narrow(policies[i]);
            assert(!is_nil(protocolPolicy));
            if(protocolPolicy -> value() != tag())
                return new OCI::ObjectKey();
        }
    }

    //
    // Make sure that the IOR is usable
    //
    return is_usable(ior);

}

OCI::ConnectorInfo_ptr
Connector_impl::get_info()
    throw(SystemException)
{
    return OCI::ConnectorInfo::_duplicate(info_);
}

// ----------------------------------------------------------------------
// Acceptor_impl constructor and destructor
// ----------------------------------------------------------------------

Acceptor_impl::Acceptor_impl(const char* host, UShort port)
    : host_(host), port_(port), fd_(-1),
      info_(new AcceptorInfo_impl(this))
{
#ifdef OB_OCI_DEBUG
    cout << "OCI::IIOP::Acceptor_impl" << endl;
#endif

    int no;

    //
    // Get address
    //
    memset(&addr_, 0, sizeof(addr_));
    addr_.sin_family = AF_INET;
    addr_.sin_addr.s_addr = htonl(INADDR_ANY);

    //
    // Create socket
    //
    if((fd_ = CreateSocket()) == -1)
    {
	_OB_setRef(0);
	throw COMM_FAILURE(LastError(), MinorSocket, COMPLETED_NO);
    }

    //
    // Set TCP_NODELAY option
    //
    no = 1;
    if(setsockopt(fd_, IPPROTO_TCP, TCP_NODELAY, (char*)&no, sizeof(int))
       == -1)
    {
	int err = GetError();
	CloseSocket(fd_);
	fd_ = -1;
	_OB_setRef(0);
	SetError(err);
	throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
    }

    //
    // Bind
    //
    no = 1;
    if(setsockopt(fd_, SOL_SOCKET, SO_REUSEADDR, (char*)&no,
		  sizeof(int)) == -1)
    {
	int err = GetError();
	CloseSocket(fd_);
	fd_ = -1;
	_OB_setRef(0);
	SetError(err);
	throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
    }
	
    addr_.sin_port = htons(port_);

    if(bind(fd_, (struct sockaddr*)&addr_, sizeof(addr_)) == -1)
    {
	int err = GetError();
	CloseSocket(fd_);
	fd_ = -1;
	_OB_setRef(0);
	SetError(err);
	throw COMM_FAILURE(LastError(), MinorBind, COMPLETED_NO);
    }

#ifdef HAVE_GETSOCKNAME_WITH_SIZE_T
    size_t len = sizeof(addr_);
#else
    int len = sizeof(addr_);
#endif
    int result = getsockname(fd_, (struct sockaddr*)&addr_, &len);
    assert(result != -1);

    port_ = ntohs(addr_.sin_port);

#ifdef OB_TRACE
    {
        String_var msg = string_dup("Accepting connections on port ");
        msg += ntohs(addr_.sin_port);
        MessageViewer::instance() -> trace(1, msg);
    }
#endif
}

Acceptor_impl::~Acceptor_impl()
{
#ifdef OB_OCI_DEBUG
    cout << "~Acceptor_impl" << endl;
#endif

    if(fd_ >= 0)
	close();

    release(info_);
}

// ----------------------------------------------------------------------
// Acceptor_impl public member implementation
// ----------------------------------------------------------------------

void
Acceptor_impl::close()
    throw(SystemException)
{
    //
    // Destroy the info object
    //
    info_ -> _OB_destroy();

    //
    // Close the socket
    //
    int saveFd = fd_;
    fd_ = -1;
    CloseSocket(saveFd);
}

void
Acceptor_impl::shutdown()
    throw(SystemException)
{
    //
    // Create socket for interrupt
    //
    int intFd;
    if((intFd = CreateSocket()) == -1)
    {
	intFd = -1;
	throw COMM_FAILURE(LastError(), MinorSocket, COMPLETED_NO);
    }

    //
    // Connect to the acceptor socket
    //
    struct sockaddr_in addr;
    memset(&addr, 0, sizeof(addr));
    addr.sin_family = addr_.sin_family;
    addr.sin_port = addr_.sin_port;
    addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
    while(true)
    {
        if(::connect(intFd, (struct sockaddr*)&addr, sizeof(addr)) == -1)
        {
            if(IsInterrupted())
                continue;
	//
	// Ignore connect failures - it's possible that accept has not
	// been called
	//
/*
	int err = GetError();
	CloseSocket(intFd);
	intFd = -1;
	SetError(err);
	throw COMM_FAILURE(LastError(), MinorConnect, COMPLETED_NO);
*/
        }
        break;
    }

    //
    // Close the interrupt socket
    //
    CloseSocket(intFd);
}

void
Acceptor_impl::listen()
    throw(SystemException)
{
    //
    // Listen
    //
    while(true)
    {
	if(::listen(fd_, 5) == -1)
	{
            if(IsInterrupted())
		continue;
	    int err = GetError();
	    CloseSocket(fd_);
	    fd_ = -1;
	    SetError(err);
	    throw COMM_FAILURE(LastError(), MinorListen, COMPLETED_NO);
	}
	break;
    }
}

OCI::Transport_ptr
Acceptor_impl::accept()
    throw(SystemException)
{
    //
    // Accept
    //
    struct sockaddr_in newAddr;
#ifdef HAVE_ACCEPT_WITH_SIZE_T
    size_t len = sizeof(newAddr);
#else
    int len = sizeof(newAddr);
#endif
    int newFd;
    while(true)
    {
	if((newFd = ::accept(fd_, (struct sockaddr*)&newAddr, &len)) == -1)
        {
	    //
	    // Here we have special handling for this case, since
	    // accept can generate other loop-and-retry errors on some
	    // platforms.
	    //
            if(AcceptRetry())
                continue;
	    
            int err = GetError();
            CloseSocket(fd_);
            fd_ = -1;
            SetError(err);
            throw COMM_FAILURE(LastError(), MinorAccept, COMPLETED_NO);
        }
	
        break;
    }
	
#ifdef OB_TRACE
    {
        String_var msg = string_dup("New connection from ");
        msg += inet_ntoa(newAddr.sin_addr);
        msg += ":";
        msg += ntohs(newAddr.sin_port);
        MessageViewer::instance() -> trace(1, msg);
    }
#endif

    //
    // Set TCP_NODELAY option
    //
    int no = 1;
    if(setsockopt(newFd, IPPROTO_TCP, TCP_NODELAY, (char*)&no, sizeof(int))
       == -1)
    {
	int err = GetError();
	CloseSocket(fd_);
	CloseSocket(newFd);
	fd_ = -1;
	SetError(err);
	throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
    }
    
    //
    // Create new transport
    //
    OCI::Transport_var tr = new Transport_impl(this, newFd, TRUE);
    
    //
    // Call callbacks
    //
    OCI::TransportInfo_var trInfo = tr -> get_info();
    try
    {
	info_ -> _OB_callAcceptCB(trInfo);
    }
    catch(const SystemException&)
    {
	tr -> close();
	throw;
    }

    //
    // Return new transport
    //
    return tr._retn();
}

void
Acceptor_impl::add_profile(const OCI::ObjectKey& key, OCI::IOR& ior)
    throw(SystemException)
{
    assert(port_ > 0);

    ::IIOP::ProfileBody_1_1 body;
    body.iiop_version.major = 1;
    body.iiop_version.minor = 1;
    body.host = host_;
    body.port = port_;
    body.object_key.length(key.length());
    memcpy(body.object_key.get_buffer(), key.get_buffer(), key.length());

    ULong len = ior.profiles.length() + 1;
    ior.profiles.length(len);
    ior.profiles[len - 1].tag = tag();
    CountStream count;
    count.count_boolean(); // Endian
    ::IIOP::ProfileBody_1_1::_OB_marshalCount(body, count);
    ior.profiles[len - 1].profile_data.length(count.count());
    Octet* oct = ior.profiles[len - 1].profile_data.get_buffer();
#ifdef OB_CLEAR_MEM
    memset(oct, 0, count.count());
#endif
    OutputStream out(oct, 0, count.count());
    out.write_boolean(Endian);
    ::IIOP::ProfileBody_1_1::_OB_marshal(body, out);
}

OCI::ObjectKey*
Acceptor_impl::is_local(const OCI::IOR& ior)
    throw(SystemException)
{
    return IIOPExtractKey(ior, host_, port_, true);
}

OCI::AcceptorInfo_ptr
Acceptor_impl::get_info()
    throw(SystemException)
{
    return OCI::AcceptorInfo::_duplicate(info_);
}

// ----------------------------------------------------------------------
// ConFactory_impl constructor and destructor
// ----------------------------------------------------------------------

ConFactory_impl::ConFactory_impl()
    : info_(new ConFactoryInfo_impl)
{
#ifdef OB_OCI_DEBUG
    cout << "ConFactory_impl" << endl;
#endif
}

ConFactory_impl::~ConFactory_impl()
{
#ifdef OB_OCI_DEBUG
    cout << "~ConFactory_impl" << endl;
#endif
    release(info_);
}

// ----------------------------------------------------------------------
// ConFactory_impl public member implementation
// ----------------------------------------------------------------------

OCI::Connector_ptr
ConFactory_impl::create(const OCI::IOR& ior)
    throw(SystemException)
{
    for(ULong i = 0 ; i < ior.profiles.length() ; i++)
    {
	if(ior.profiles[i].tag == tag())
	{
	    //
	    // Get the IIOP profile body
	    //
	    const Octet* coct = ior.profiles[i].profile_data.get_buffer();
            ULong len = ior.profiles[i].profile_data.length();
            InputStream in(coct, 0, len, false);
            in._OB_readEndian();
	    ::IIOP::ProfileBody_1_0 body;
	    ::IIOP::ProfileBody_1_0::_OB_unmarshal(body, in);

	    //
	    // Create new connector
	    //
            OCI::ConnectCBSeq_var cbs = info_ -> _OB_getConnectCBSeq();
	    return new Connector_impl(body.host, body.port, cbs);
	}
    }

    return OCI::Connector::_nil();
}

OCI::Connector_ptr
ConFactory_impl::create_with_policies(const OCI::IOR& ior,
                                      const PolicyList& policies)
    throw(SystemException)
{
    if(!consider_with_policies(ior, policies))
        return OCI::Connector::_nil();

    return create(ior);
}

Boolean
ConFactory_impl::consider_with_policies(const OCI::IOR& ior,
                                        const PolicyList& policies)
    throw(SystemException)
{
    for(ULong i = 0; i < ior.profiles.length(); ++i)
    {
        if(ior.profiles[i].tag == tag())
        {
            //
            // Make sure that the policies are met
            //
            for(ULong j = 0; j < policies.length(); j++)
            {
                if(policies[j] -> policy_type() == SSL::CONNECT_POLICY)
                {
                    SSL::ConnectPolicy_var connectPolicy =
                        SSL::ConnectPolicy::_narrow(policies[j]);
		    
		    assert(!is_nil(connectPolicy));
		    
                    if(connectPolicy -> value() == SSL::ConnectSecure)
                        return FALSE;
                }
                else if(policies[j] -> policy_type() == OB::PROTOCOL_POLICY)
                {
                    OB::ProtocolPolicy_var protocolPolicy =
                        OB::ProtocolPolicy::_narrow(policies[j]);
                    assert(!is_nil(protocolPolicy));
                    if(protocolPolicy -> value() != tag())
                        return FALSE;
                }
            }

            return TRUE;
        }
    }

    return FALSE;
}

Boolean
ConFactory_impl::equivalent(const OCI::IOR& ior1, const OCI::IOR& ior2)
    throw(SystemException)
{
    return IIOPEquivalent(ior1, ior2);
}

ULong
ConFactory_impl::hash(const OCI::IOR& ior, ULong maximum)
    throw(SystemException)
{
    return IIOPHash(ior, maximum);
}

OCI::ConFactoryInfo_ptr
ConFactory_impl::get_info()
    throw(SystemException)
{
    return ConFactoryInfo::_duplicate(info_);
}

// ----------------------------------------------------------------------
// TransportInfo_impl private and protected member implementation
// ----------------------------------------------------------------------

void
TransportInfo_impl::init()
{
    //
    // I have to save the local and remote address and port, since I
    // get garbage if I calculate these values "on the fly" during the
    // close callback.
    //

#ifdef HAVE_GETSOCKNAME_WITH_SIZE_T
    size_t len = sizeof(struct sockaddr_in);
#else
    int len = sizeof(struct sockaddr_in);
#endif
    int result;

    //
    // Get the local address and port
    //
    result = getsockname(transport_ -> fd_, (struct sockaddr*)&addr_, &len);
    assert(result == 0);

    //
    // Get the remote address and port
    //
    result = getpeername(transport_ -> fd_, (struct sockaddr*)&remoteAddr_,
			 &len);
    assert(result == 0);
}

// ----------------------------------------------------------------------
// TransportInfo_impl constructor and destructor
// ----------------------------------------------------------------------

TransportInfo_impl::TransportInfo_impl(Transport_impl* transport,
                                       Connector_impl* connector)
    : transport_(transport)
{
    connectorInfo_ = connector -> get_info();

    //
    // Initialize some data members
    //
    init();
}

TransportInfo_impl::TransportInfo_impl(Transport_impl* transport,
                                       Acceptor_impl* acceptor)
    : transport_(transport)
{
    acceptorInfo_ = acceptor -> get_info();

    //
    // Initialize some data members
    //
    init();
}

TransportInfo_impl::~TransportInfo_impl()
{
}

// ----------------------------------------------------------------------
// TransportInfo_impl public member implementation
// ----------------------------------------------------------------------

OCI::ConnectorInfo_ptr
TransportInfo_impl::connector_info()
    throw(SystemException)
{
    return OCI::ConnectorInfo::_duplicate(connectorInfo_);
}

OCI::AcceptorInfo_ptr
TransportInfo_impl::acceptor_info()
    throw(SystemException)
{
    return OCI::AcceptorInfo::_duplicate(acceptorInfo_);
}

void
TransportInfo_impl::add_close_cb(OCI::CloseCB_ptr cb)
    throw(SystemException)
{
    OB_SYNCHRONIZED

    for(ULong i = 0 ; i < closeCBSeq_.length() ; i++)
	if(closeCBSeq_[i] -> _is_equivalent(cb))
	    return; // Already registered

    closeCBSeq_.append(OCI::CloseCB::_duplicate(cb)); // Add callback object
}

void
TransportInfo_impl::remove_close_cb(OCI::CloseCB_ptr cb)
    throw(SystemException)
{
    OB_SYNCHRONIZED

    for(ULong i = 0 ; i < closeCBSeq_.length() ; i++)
	if(closeCBSeq_[i] -> _is_equivalent(cb))
	{
	    closeCBSeq_.remove(i); // Remove callback object
	    i--;
	}
}

InetAddr_slice*
TransportInfo_impl::addr()
    throw(SystemException)
{
    OB_SYNCHRONIZED

    if(transport_ == 0)
	throw NO_RESOURCES();

    InetAddr_slice* addr = new InetAddr;
    memcpy(addr, (const char*)&addr_.sin_addr.s_addr, 4);
    return addr;
}

UShort
TransportInfo_impl::port()
    throw(SystemException)
{
    OB_SYNCHRONIZED

    if(transport_ == 0)
	throw NO_RESOURCES();

    return ntohs(addr_.sin_port);
}

InetAddr_slice*
TransportInfo_impl::remote_addr()
    throw(SystemException)
{
    OB_SYNCHRONIZED

    if(transport_ == 0)
	throw NO_RESOURCES();

    InetAddr_slice* addr = new InetAddr;
    memcpy(addr, (const char*)&remoteAddr_.sin_addr.s_addr, 4);
    return addr;
}

UShort
TransportInfo_impl::remote_port()
    throw(SystemException)
{
    OB_SYNCHRONIZED

    if(transport_ == 0)
	throw NO_RESOURCES();

    return ntohs(remoteAddr_.sin_port);
}

void
TransportInfo_impl::_OB_callCloseCB(OCI::TransportInfo_ptr info)
{
    OB_SYNCHRONIZED

    for(ULong i = 0 ; i < closeCBSeq_.length() ; i++)
	closeCBSeq_[i] -> close_cb(info);
}

void
TransportInfo_impl::_OB_destroy()
{
    OB_SYNCHRONIZED

    transport_ = 0;
}

// ----------------------------------------------------------------------
// ConnectorInfo_impl public member implementation
// ----------------------------------------------------------------------

void
ConnectorInfo_impl::add_connect_cb(OCI::ConnectCB_ptr cb)
    throw(SystemException)
{
    OB_SYNCHRONIZED

    for(ULong i = 0 ; i < connectCBSeq_.length() ; i++)
	if(connectCBSeq_[i] -> _is_equivalent(cb))
	    return; // Already registered

    connectCBSeq_.append(OCI::ConnectCB::_duplicate(cb)); // Add callback object
}

void
ConnectorInfo_impl::remove_connect_cb(OCI::ConnectCB_ptr cb)
    throw(SystemException)
{
    OB_SYNCHRONIZED

    for(ULong i = 0 ; i < connectCBSeq_.length() ; i++)
	if(connectCBSeq_[i] -> _is_equivalent(cb))
	{
	    connectCBSeq_.remove(i); // Remove callback object
	    i--;
	}
}

InetAddr_slice*
ConnectorInfo_impl::remote_addr()
    throw(SystemException)
{
    OB_SYNCHRONIZED

    if(connector_ == 0)
	throw NO_RESOURCES();

    InetAddr_var rc = new InetAddr;
    memcpy(rc.inout(), (const char*)&connector_ -> addr_.sin_addr.s_addr, 4);

    return rc._retn();
}

UShort
ConnectorInfo_impl::remote_port()
    throw(SystemException)
{
    OB_SYNCHRONIZED

    if(connector_ == 0)
	throw NO_RESOURCES();

    return ntohs(connector_ -> addr_.sin_port);
}

void
ConnectorInfo_impl::_OB_callConnectCB(OCI::TransportInfo_ptr info)
{
    OB_SYNCHRONIZED

    for(ULong i = 0 ; i < connectCBSeq_.length() ; i++)
	connectCBSeq_[i] -> connect_cb(info);
}

void
ConnectorInfo_impl::_OB_destroy()
{
    OB_SYNCHRONIZED

    connector_ = 0;
}

// ----------------------------------------------------------------------
// AcceptorInfo_impl public member implementation
// ----------------------------------------------------------------------

void
AcceptorInfo_impl::add_accept_cb(OCI::AcceptCB_ptr cb)
    throw(SystemException)
{
    OB_SYNCHRONIZED

    for(ULong i = 0 ; i < acceptCBSeq_.length() ; i++)
	if(acceptCBSeq_[i] -> _is_equivalent(cb))
	    return; // Already registered

    acceptCBSeq_.append(OCI::AcceptCB::_duplicate(cb)); // Add callback object
}

void
AcceptorInfo_impl::remove_accept_cb(OCI::AcceptCB_ptr cb)
    throw(SystemException)
{
    OB_SYNCHRONIZED

    for(ULong i = 0 ; i < acceptCBSeq_.length() ; i++)
	if(acceptCBSeq_[i] -> _is_equivalent(cb))
	{
	    acceptCBSeq_.remove(i); // Remove callback object
            return;
	}
}

char*
AcceptorInfo_impl::host()
    throw(SystemException)
{
    OB_SYNCHRONIZED

    if(acceptor_ == 0)
	throw NO_RESOURCES();

    return string_dup(acceptor_ -> host_);
}

InetAddr_slice*
AcceptorInfo_impl::addr()
    throw(SystemException)
{
    OB_SYNCHRONIZED

    if(acceptor_ == 0)
	throw NO_RESOURCES();

    struct sockaddr_in addr;
#ifdef HAVE_GETSOCKNAME_WITH_SIZE_T
    size_t len = sizeof(addr);
#else
    int len = sizeof(addr);
#endif
    int result = getsockname(acceptor_ -> fd_, (struct sockaddr*)&addr, &len);
    InetAddr_var rc = new InetAddr;
    if(result == 0)
        memcpy(rc.inout(), (const char*)&addr.sin_addr.s_addr, 4);

    return rc._retn();
}

UShort
AcceptorInfo_impl::port()
    throw(SystemException)
{
    OB_SYNCHRONIZED

    if(acceptor_ == 0)
	throw NO_RESOURCES();

    struct sockaddr_in addr;
#ifdef HAVE_GETSOCKNAME_WITH_SIZE_T
    size_t len = sizeof(addr);
#else
    int len = sizeof(addr);
#endif
    int result = getsockname(acceptor_ -> fd_, (struct sockaddr*)&addr, &len);

    if(result == 0)
        return ntohs(addr.sin_port);
    return 0;
}

void
AcceptorInfo_impl::_OB_callAcceptCB(OCI::TransportInfo_ptr info)
{
    OB_SYNCHRONIZED

    for(ULong i = 0 ; i < acceptCBSeq_.length() ; i++)
	acceptCBSeq_[i] -> accept_cb(info);
}

void
AcceptorInfo_impl::_OB_destroy()
{
    OB_SYNCHRONIZED

    acceptor_ = 0;
}

// ----------------------------------------------------------------------
// ConFactoryInfo_impl public member implementation
// ----------------------------------------------------------------------

void
ConFactoryInfo_impl::add_connect_cb(OCI::ConnectCB_ptr cb)
    throw(SystemException)
{
    OB_SYNCHRONIZED

    for(ULong i = 0 ; i < connectCBSeq_.length() ; i++)
	if(connectCBSeq_[i] -> _is_equivalent(cb))
	    return; // Already registered

    connectCBSeq_.append(OCI::ConnectCB::_duplicate(cb)); // Add callback object
}

void
ConFactoryInfo_impl::remove_connect_cb(OCI::ConnectCB_ptr cb)
    throw(SystemException)
{
    OB_SYNCHRONIZED

    for(ULong i = 0 ; i < connectCBSeq_.length() ; i++)
	if(connectCBSeq_[i] -> _is_equivalent(cb))
	{
	    connectCBSeq_.remove(i); // Remove callback object
	    i--;
	}
}

OCI::ConnectCBSeq*
ConFactoryInfo_impl::_OB_getConnectCBSeq()
{
    OB_SYNCHRONIZED

    return new OCI::ConnectCBSeq(connectCBSeq_);
}
