// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/POAServantBase.h>
#include <OB/OCI.h>

#include <OCI_impl.h>

using namespace CORBA;
using namespace OCI;

//
// XXX
// destroy operations
//

// ----------------------------------------------------------------------
// ConFactoryRegistry_impl public member implementation
// ----------------------------------------------------------------------

void
ConFactoryRegistry_impl::add_factory(ConFactory_ptr factory)
    throw(SystemException)
{
    factories_.insert(ConFactory::_duplicate(factory));
}

ConFactory_ptr
ConFactoryRegistry_impl::get_factory(const IOR& ior)
    throw(SystemException)
{
    PolicyList dummyPolicies;
    return get_factory_with_policies(ior, dummyPolicies);
}

ConFactory_ptr
ConFactoryRegistry_impl::get_factory_with_policies(const IOR& ior,
                                                   const PolicyList& policies)
    throw(SystemException)
{
    //
    // Need to search each profile tag, then factories to force
    // order correctness.
    //
    for(ULong j = 0 ; j < factories_.length() ; j++)
        if(factories_[j] -> consider_with_policies(ior, policies))
            return ConFactory::_duplicate(factories_[j]);
    
    return ConFactory::_nil();
}

ConFactorySeq*
ConFactoryRegistry_impl::get_factories()
    throw(SystemException)
{
    return new ConFactorySeq(factories_);
}

Boolean
ConFactoryRegistry_impl::equivalent(const IOR& ior1, const IOR& ior2)
    throw(SystemException)
{
    for(ULong i = 0 ; i < factories_.length() ; i++)
	if(!factories_[i] -> equivalent(ior1, ior2))
	    return FALSE;
    
    return TRUE;
}

ULong
ConFactoryRegistry_impl::hash(const IOR& ior, ULong maximum)
    throw(SystemException)
{
    ULong hash = 0;
    
    for(ULong i = 0 ; i < factories_.length() ; i++)
	hash ^= factories_[i] -> hash(ior, maximum);
    
    return hash % (maximum + 1);
}

// ----------------------------------------------------------------------
// AccRegistry_impl public member implementation
// ----------------------------------------------------------------------

void
AccRegistry_impl::add_acceptor(Acceptor_ptr acceptor)
    throw(SystemException)
{
    acceptors_.length(acceptors_.length() + 1);
    acceptors_[acceptors_.length() - 1] = Acceptor::_duplicate(acceptor);
}

AcceptorSeq*
AccRegistry_impl::get_acceptors()
    throw(SystemException)
{
    return new AcceptorSeq(acceptors_);
}

void
AccRegistry_impl::add_profiles(const ObjectKey& object_key, IOR& ior)
    throw(SystemException)
{
    for(ULong i = 0 ; i < acceptors_.length() ; i++)
	acceptors_[i] -> add_profile(object_key, ior);
}

ObjectKey*
AccRegistry_impl::is_local(const IOR& ior)
    throw(SystemException)
{
    for(ULong i = 0 ; i < acceptors_.length() ; i++)
    {
	ObjectKey_var key = acceptors_[i] -> is_local(ior);
	if(key -> length() > 0)
	    return key._retn();
    }
    
    return new ObjectKey();
}

// ----------------------------------------------------------------------
// CurrentState class. This holds the internal state of the OCI
// Current implementation object.
// ----------------------------------------------------------------------

namespace OCI
{

class CurrentState
{
    void clean()
    {
        info = TransportInfo::_nil();
    }

#ifdef HAVE_JTC
    static JTCMutex freeMutex_;
#endif
    static CurrentState* freeList_;
    
public:
    
    static void freeState(CurrentState* next)
    {
#ifdef HAVE_JTC
        JTCSynchronized sync(freeMutex_);
#endif
        next -> clean();
        next -> next = freeList_;
        freeList_ = next;
    }
    
    static CurrentState* newState()
    {
#ifdef HAVE_JTC
        JTCSynchronized sync(freeMutex_);
#endif
        CurrentState* next;
        if(freeList_ != 0)
        {
            next = freeList_;
            freeList_ = freeList_ -> next;
        }
        else
        {
            next = new CurrentState;
        }
        return next;
    }

    TransportInfo_var info;
    CurrentState* next;
};

} // End of namespace OCI

#ifdef HAVE_JTC
JTCMutex CurrentState::freeMutex_;
#endif
CurrentState* CurrentState::freeList_ = 0;

// ----------------------------------------------------------------------
// Current_impl constructor and destructor
// ----------------------------------------------------------------------

Current_impl::Current_impl()
{
#ifdef HAVE_JTC
    //
    // Initialize the transport TSS key
    //
    stateKey_ = JTCTSS::allocate();
#else
    state_ = 0;
#endif
}

Current_impl::~Current_impl()
{
    //
    //XXX - cleanup state objects.
    //
#ifdef HAVE_JTC
    //
    // Release the transport TSS key
    //
    JTCTSS::release(stateKey_);
#endif
}

// ----------------------------------------------------------------------
// Current_impl public member implementation
// ----------------------------------------------------------------------

TransportInfo_ptr
Current_impl::get_oci_transport_info()
    throw(SystemException)
{
#ifdef HAVE_JTC
    CurrentState* state_ = (CurrentState*)JTCTSS::get(stateKey_);
#endif
    if(state_ == 0 || is_nil(state_ -> info))
        // some exception
        ;
    return TransportInfo::_duplicate(state_ -> info);
}

AcceptorInfo_ptr
Current_impl::get_oci_acceptor_info()
    throw(SystemException)
{
    TransportInfo_var info = get_oci_transport_info();
    if(!is_nil(info))
        return info -> acceptor_info();
    else
        return AcceptorInfo::_nil();
}

void
Current_impl::_OB_preinvoke(TransportInfo_ptr info)
{
#ifdef HAVE_JTC
    CurrentState* state_ = (CurrentState*)JTCTSS::get(stateKey_);
#endif

    CurrentState* newState = CurrentState::newState();
    newState -> info = TransportInfo::_duplicate(info);

    newState -> next = state_;
    state_ = newState;

#ifdef HAVE_JTC
    JTCTSS::set(stateKey_, state_);
#endif
}

void
Current_impl::_OB_postinvoke()
{
#ifdef HAVE_JTC
    CurrentState* state_ = (CurrentState*)JTCTSS::get(stateKey_);
#endif

    CurrentState* save = state_;
    state_ = state_ -> next;
    CurrentState::freeState(save);

#ifdef HAVE_JTC
    JTCTSS::set(stateKey_, state_);
#endif
}
