// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/Any.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/IFR1.h>
#include <OB/OBV.h>
#include <OB/Policy.h>
#include <OB/ORB.h>
#include <OB/NativeTypes.h>
#include <OB/POAInterface.h>
#include <OB/POAServantLocator.h>
#include <OB/POAServantBase.h>
#include <OB/POACurrent.h>

#include <POACurrent_impl.h>

using namespace CORBA;
using namespace PortableServer;

// ----------------------------------------------------------------------
// POACurrentState class. This holds the internal state of the POA
// Current implementation object.
// ----------------------------------------------------------------------

namespace PortableServer
{

class POACurrentState
{
    void clean()
    {
        poa = POA::_nil();
        servant = 0;
	op = 0;
        key = 0;
        oid = 0;
        cookie = 0;
    }

#ifdef HAVE_JTC
    static JTCMutex freeMutex_;
#endif
    static POACurrentState* freeList_;
    
public:
    
    POACurrentState() : cookie(0) { }

    static void freeState(POACurrentState* next)
    {
#ifdef HAVE_JTC
        JTCSynchronized sync(freeMutex_);
#endif
        next -> clean();
        next -> next = freeList_;
        freeList_ = next;
    }
    
    static POACurrentState* newState()
    {
#ifdef HAVE_JTC
        JTCSynchronized sync(freeMutex_);
#endif
        POACurrentState* next;
        if(freeList_ != 0)
        {
            next = freeList_;
            freeList_ = freeList_ -> next;
        }
        else
        {
            next = new POACurrentState;
        }
        return next;
    }

    POA_var poa;
    ServantBase_var servant;
    ServantLocator::Cookie cookie;

    //
    // We can make these fields pointers since they should be valid
    // for the duration of the invocation.
    //
    const ObjectId* oid;
    const OB::OctetSeq* key;
    const char* op;

    POACurrentState* next;
};

} // End of namespace PortableServer

#ifdef HAVE_JTC
JTCMutex POACurrentState::freeMutex_;
#endif
POACurrentState* POACurrentState::freeList_ = 0;

// ----------------------------------------------------------------------
// Current_impl constructor and destructor
// ----------------------------------------------------------------------

Current_impl::Current_impl()
{
#ifdef HAVE_JTC
    stateKey_ = JTCTSS::allocate();
#else
    state_ = 0;
#endif
}

Current_impl::~Current_impl()
{
    //
    // XXX - cleanup state objects
    //
#ifdef HAVE_JTC
    JTCTSS::release(stateKey_);
#endif
}

// ----------------------------------------------------------------------
// Current_impl public member implementation
// ----------------------------------------------------------------------

POA_ptr
Current_impl::get_POA()
    throw(Current::NoContext,
          SystemException)
{
#ifdef HAVE_JTC
    POACurrentState* state_ = (POACurrentState*)JTCTSS::get(stateKey_);
#endif
    
    if(state_ == 0 || is_nil(state_ -> poa))
        throw Current::NoContext();
    return POA::_duplicate(state_ -> poa);
}


ObjectId*
Current_impl::get_object_id()
    throw(Current::NoContext,
          SystemException)
{
#ifdef HAVE_JTC
    POACurrentState* state_ = (POACurrentState*)JTCTSS::get(stateKey_);
#endif
    
    if(state_ == 0 || is_nil(state_ -> poa))
        throw Current::NoContext();
    return new ObjectId(*state_ -> oid);
}

void
Current_impl::_OB_preinvoke(
    POA_ptr poa,
    ServantBase* servant,
    const char* op,
    const ObjectId& oid,
    const OB::OctetSeq& key,
    ServantLocator::Cookie cookie)
{
#ifdef HAVE_JTC
    POACurrentState* state_ = (POACurrentState*)JTCTSS::get(stateKey_);
#endif
    
    POACurrentState* ns = POACurrentState::newState();
    ns -> poa = POA::_duplicate(poa);

    //
    // These should be valid for the duration of the invocation
    //
    ns -> oid = &oid;
    ns -> key = &key;
    ns -> cookie = cookie;
    ns -> servant = servant;
    ns -> op = op;
    if(servant != 0)
    {
        servant -> _add_ref();
    }

    ns -> next = state_;
    state_ = ns;

#ifdef HAVE_JTC
    JTCTSS::set(stateKey_, state_);
#endif
}

void
Current_impl::_OB_postinvoke()
{
#ifdef HAVE_JTC
    POACurrentState* state_ = (POACurrentState*)JTCTSS::get(stateKey_);
#endif
    POACurrentState* save = state_;
    state_ = state_ -> next;
    POACurrentState::freeState(save);
    
#ifdef HAVE_JTC
    JTCTSS::set(stateKey_, state_);
#endif
}

bool
Current_impl::_OB_inUpcall() const
{
#ifdef HAVE_JTC
    POACurrentState* state_ = (POACurrentState*)JTCTSS::get(stateKey_);
#endif
    
    return state_ != 0;
}

ServantBase*
Current_impl::_OB_getServant()
{
#ifdef HAVE_JTC
    POACurrentState* state_ = (POACurrentState*)JTCTSS::get(stateKey_);
#endif
    
    assert(state_);
    return state_ -> servant;
}

const OB::OctetSeq&
Current_impl::_OB_getKey()
{
#ifdef HAVE_JTC
    POACurrentState* state_ = (POACurrentState*)JTCTSS::get(stateKey_);
#endif
    
    assert(state_);
    return *state_ -> key;
}

ServantLocator::Cookie
Current_impl::_OB_getCookie()
{
#ifdef HAVE_JTC
    POACurrentState* state_ = (POACurrentState*)JTCTSS::get(stateKey_);
#endif

    assert(state_);
    return state_ -> cookie;
}

const ObjectId&
Current_impl::_OB_getObjectId()
{
#ifdef HAVE_JTC
    POACurrentState* state_ = (POACurrentState*)JTCTSS::get(stateKey_);
#endif

    assert(state_);
    return *state_ -> oid;
}

const char*
Current_impl::_OB_getOp()
{
#ifdef HAVE_JTC
    POACurrentState* state_ = (POACurrentState*)JTCTSS::get(stateKey_);
#endif

    assert(state_);
    return state_ -> op;
}
