// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/TemplateI.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/Any.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/IFR1.h>
#include <OB/IFR4.h>
#include <OB/OBV.h>
#include <OB/Policy.h>
#include <OB/ORB.h>
#include <OB/OCI_IIOP.h>
#include <OB/Timer.h> // for TimerList::timeNow...
#include <OB/Hashtable.h>
#include <OB/HashtableI.h>
#include <OB/NativeTypes.h>
#include <OB/PortableServer.h>
#include <OB/POAServantBase.h>
#include <OB/ValueBase.h>
#include <OB/POAInterface.h>
#include <OB/POAPolicies.h>
#include <OB/POACurrent.h>
#include <OB/POAManager.h>
#include <OB/POAServantLocator.h>
#include <OB/POAServantActivator.h>
#include <OB/Util.h>
#include <OB/OCI.h> // for OCI::impl...
#include <OB/Stream.h>

#include <POAPolicies_impl.h>
#include <POA_impl.h>
#include <POACurrent_impl.h>
#include <POAManager_impl.h>
#include <OB/Hashers.h>

#include <OCI_impl.h> // For OCI::Current_impl...

#include <Upcall.h>
#include <OARoot.h>
#include <CollocatedClient.h>
#include <Downcall.h>

#include <stdlib.h>

using namespace CORBA;
using namespace PortableServer;
using namespace OB;

//
// Synchronization notes:
//
// The mutex protecting the POA implementation must be recursive.
// Consider: find_POA. find_POA must lock the POA to operate. find_POA
// can call unknown_adapter which may call create_POA on the same POA,
// thus recursively locking the same mutex. It's also not possible to
// unlock the mutex before calling unknown_adapter since it would then
// be possible to create a race between two threads finding the same
// POA on who creates the POA.
//
// NOTE: There is a way around this. We can use a non-recursive mutex
// and a condition variable as outlined by Jonathan Biggar
// <jon@floorboard.com> on the ace users list. Not sure how this would
// translate into java.
//
// Private methods don't need to synchronize. It is assumed that all
// public methods provide appropriate synchronization.
//
// UNRESOLVED: [XXX]
// - create_POA/find_POA conundrum.
// - Lots of other stuff, no doubt!
// - The _OB_findServant method has a comment at the end saying that `we
//   don't want to actually activate the servant at this point', but
//   without checking with the servant manager, you can't really claim
//   that the object is `here', can you? (see POA. p. 11-27 for relevant
//   quote)
// - Speed up dispatch process. findPOA is most likely the largest bottle-
//   neck here. We need an optimized method of doing this.
// - see XXX throughout code for issues to resolve.
// - We probably want to use an invocation record cache
//

#ifdef HAVE_JTC
    #define OB_READ_LOCK JTCSynchronized synchronized(mutex_);
    #define OB_WRITE_LOCK JTCSynchronized synchronized(mutex_);
#else
    #define OB_READ_LOCK /**/
    #define OB_WRITE_LOCK /**/
#endif

//
// The ORB Synchronization Mutex
//
#ifdef HAVE_JTC
JTCRecursiveMutex POA_impl::orbSyncMutex_;
#endif

// ----------------------------------------------------------------------
// Template instantiations
// ----------------------------------------------------------------------

#ifndef HAVE_NO_EXPLICIT_TEMPLATES
template class OB::VarSeq< POA_impl::InvocationRecord*, int >;

template class OB::Hashtable< ObjectId, ServantBase_var >;
template class OB::Hashtable< String_var, POA_var >;
template class OB::Hashtable< Servant, ObjectId >;
template class OB::Hashtable< ObjectId, OB::FixSeq<StubImplBase*, int>* >;
template class OB::FixSeq< StubImplBase*, int >;
#else
#ifdef HAVE_PRAGMA_DEFINE
#pragma define(OB::SeqVar< OB::ObjSeq< POA > >)
//XXX - Fix template instantiation stuff
#endif
#endif

// ----------------------------------------------------------------------
// PortableServer namespace functions
// ----------------------------------------------------------------------

char*
PortableServer::ObjectId_to_string(const ObjectId& id)
{
    //
    // If we have any '\0' characters in the id then throw a BAD_PARAM
    // exception
    //
    if(memchr(id.get_buffer(), 0, id.length()) != 0)
        throw BAD_PARAM();
    
    String_var str = string_alloc(id.length());
    memcpy(str.inout(), id.get_buffer(), id.length());
    str[id.length()] = '\0';
    return str._retn();
}

WChar*
PortableServer::ObjectId_to_wstring(const ObjectId& id)
{
    //
    // If the length of the ObjectId isn't divisible by the size of a
    // wide-character then it isn't possible to construct a valid
    // WString for this object-id.
    //
    if(id.length() % sizeof(WChar) != 0)
	throw BAD_PARAM();

    //
    // If we have any L'\0' characters in the id then throw a
    // BAD_PARAM exception.
    //
    const Octet* end = id.get_buffer() + id.length();
    for(const Octet* buf = id.get_buffer(); buf < end; buf += sizeof(WChar))
	if(*((WChar*)buf) == L'\0')
	    throw BAD_PARAM();

    //
    // The length of the WString is the length of the octet sequence
    // divided by the sizeof(WChar)
    //
    CORBA::ULong len = id.length() / sizeof(WChar);

    //
    // Allocate memory
    //
    WString_var str = wstring_alloc(len);
    memcpy(str.inout(), id.get_buffer(), id.length());

    //
    // null-terminate
    //
    str[len] = L'\0';

    return str._retn();
}

ObjectId*
PortableServer::string_to_ObjectId(const char* str)
{
    size_t len = strlen(str);
    ObjectId_var oid = new ObjectId(len);
    oid -> length(len);
    memcpy(oid -> get_buffer(), str, len);
    return oid._retn();
}

ObjectId*
PortableServer::wstring_to_ObjectId(const CORBA::WChar* str)
{
    CORBA::ULong len = OB::wcslen(str) * sizeof(WChar);
    ObjectId_var oid = new ObjectId(len);
    oid -> length(len);
    memcpy(oid -> get_buffer(), str, len);
    return oid._retn();
}

// ----------------------------------------------------------------------
// Helper operations for the hashtable
// ----------------------------------------------------------------------

ULong
OBHash(ServantBase* servant)
{
    //
    // XXX - 64 bit problem? ie. should we fold the two 32 bit values
    // somehow?
    //
    return (ULong)servant;
}

bool
OBHashCompare(ServantBase* s1, ServantBase* s2)
{
    return s1 == s2;
}

namespace OB
{

//
// This class represents the set of POA policies defined on the
// current POA. This class is immutable, and needs no mutex
// protection.
//
class POAPolicies
{
    SynchronizationPolicyValue synchronizationPolicyValue_;
    LifespanPolicyValue lifespanPolicyValue_;
    IdUniquenessPolicyValue idUniquenessPolicyValue_;
    IdAssignmentPolicyValue idAssignmentPolicyValue_;
    ImplicitActivationPolicyValue implicitActivationPolicyValue_;
    ServantRetentionPolicyValue servantRetentionPolicyValue_;
    RequestProcessingPolicyValue requestProcessingPolicyValue_;

public:
    POAPolicies(const PolicyList& policies);

    SynchronizationPolicyValue synchronizationPolicy()
    {
        return synchronizationPolicyValue_;
    }
    
    LifespanPolicyValue lifespanPolicy()
    {
        return lifespanPolicyValue_;
    }
    
    IdUniquenessPolicyValue idUniquenessPolicy()
    {
        return idUniquenessPolicyValue_;
    }
    
    IdAssignmentPolicyValue idAssignmentPolicy()
    {
        return idAssignmentPolicyValue_;
    }
    
    ImplicitActivationPolicyValue implicitActivationPolicy()
    {
        return implicitActivationPolicyValue_;
    }
    
    ServantRetentionPolicyValue servantRetentionPolicy()
    {
        return servantRetentionPolicyValue_;
    }
    
    RequestProcessingPolicyValue requestProcessingPolicy()
    {
        return requestProcessingPolicyValue_;
    }
};

} // End of namespace OB

// ----------------------------------------------------------------------
// POAPolicies constructor and destructor
// ----------------------------------------------------------------------

POAPolicies::POAPolicies(const PolicyList& policies)
    : synchronizationPolicyValue_(NO_SYNCHRONIZATION),
      lifespanPolicyValue_(PortableServer::TRANSIENT),
      idUniquenessPolicyValue_(UNIQUE_ID),
      idAssignmentPolicyValue_(SYSTEM_ID),
      implicitActivationPolicyValue_(IMPLICIT_ACTIVATION),
      servantRetentionPolicyValue_(RETAIN),
      requestProcessingPolicyValue_(USE_ACTIVE_OBJECT_MAP_ONLY)
{
    for(ULong i = 0; i < policies.length(); ++i)
    {
        PolicyType policyType = policies[i] -> policy_type();
        if(policyType == THREAD_POLICY_ID)
        {
            ThreadPolicy_var policy =
                ThreadPolicy::_narrow(policies[i]);
            synchronizationPolicyValue_ =
		(policy -> value() == ORB_CTRL_MODEL)
		? NO_SYNCHRONIZATION
		: SYNCHRONIZE_ON_ORB;
        }
        else if(policyType == SYNCHRONIZATION_POLICY_ID)
        {
            SynchronizationPolicy_var policy =
                SynchronizationPolicy::_narrow(policies[i]);
            synchronizationPolicyValue_ = policy -> value();
        }
        else if(policyType ==  LIFESPAN_POLICY_ID)
        {
            LifespanPolicy_var policy =
                LifespanPolicy::_narrow(policies[i]);
            lifespanPolicyValue_ = policy -> value();
        }
        else if(policyType == ID_UNIQUENESS_POLICY_ID)
        {
            IdUniquenessPolicy_var policy =
                IdUniquenessPolicy::_narrow(policies[i]);
            idUniquenessPolicyValue_ = policy -> value();
        }
        else if(policyType == ID_ASSIGNMENT_POLICY_ID)
        {
            IdAssignmentPolicy_var policy =
                IdAssignmentPolicy::_narrow(policies[i]);
            idAssignmentPolicyValue_ = policy -> value();
        }
        else if(policyType == IMPLICIT_ACTIVATION_POLICY_ID)
        {
            ImplicitActivationPolicy_var policy =
                ImplicitActivationPolicy::_narrow(policies[i]);
            implicitActivationPolicyValue_ = policy -> value();
        }
        else if(policyType == SERVANT_RETENTION_POLICY_ID)
        {
            ServantRetentionPolicy_var policy =
                ServantRetentionPolicy::_narrow(policies[i]);
            servantRetentionPolicyValue_ = policy -> value();
        }
        else if(policyType == REQUEST_PROCESSING_POLICY_ID)
        {
            RequestProcessingPolicy_var policy =
                RequestProcessingPolicy::_narrow(policies[i]);
            requestProcessingPolicyValue_ = policy -> value();
        }
    }
}

// ----------------------------------------------------------------------
// POA_impl private member implementation
// ----------------------------------------------------------------------

//
// Create a new system generated object id
//
// XXX
// NOTE: (jon@biggar.org) You may want to use 12 octet ObjectIds
// instead.  Assign a magic number to the first 4 bytes that is
// unlikely to be picked by a user (binary, not ascii).  The next 4
// bytes should have the creation timestamp of the POA.  The final 4
// bytes have the counter value.  Your is_poa_generated_id() would
// then verify that the ObjectId is 12 bytes long, that it has the
// correct magic number, and that the timestamp is the same as the POA
// if the POA is transient or equal or less than the timestamp of the
// POA if the POA is persistent.  This maintains a reasonable size for
// the ObjectId, but still follows the letter of the specification
// better.
//
void
POA_impl::createObjectId(ObjectId*& id)
{
    String_var str = string_dup("");
    str += nextSystemID_++;
    id = string_to_ObjectId(str);
}

//
// Determine if this ID was system generated
//
bool
POA_impl::isSystemGeneratedId(const ObjectId&)
{
    return true;
}

//
// Create an object key given an object-id.
//
// The object key is: persistent[time\0]`poa \0 poa \0 poa \0\0 id'.
// NOTE: we then know that we've reached the id by the presence of the
// two \0\0.  The time field we add for the POA create time (if
// transient) is current seconds. This is a tradeoff - it's most
// likely to be good enough.
//
POA_impl::ObjectKey*
POA_impl::createObjectKey(const ObjectId& id)
{
    ObjectKey_var key = new ObjectKey;
    
    ULong i;

    //
    // Count the number of bytes for the poas.
    //
    ULong len = 0;
    for(i = 0; i < poaId_.length(); ++i)
        len += strlen(poaId_[i]) + 1;
    
    //
    // Add one for the '\0'
    //
    len += 1;

    //
    // Add one for persistent bit
    //
    len += 1;
    String_var time;

    //
    // If the POA is transient then add the time the poa was created
    //
    if(!persistent_)
    {
        time = string_dup("");

        time += poaCreateTime_.tv_sec;
        len += strlen(time) + 1;
    }
    
    //
    // Set the length
    //
    key -> length(id.length() + len);

    char* data = (char*)key -> get_buffer();

    if(persistent_)
    {
        *data++ = '0';
    }
    else
    {
        *data++ = '1';
        strcpy(data, time);
        data += strlen(time)+1;
    }
    
    //
    // Copy in the data for the poas
    //
    for(i = 0; i < poaId_.length(); ++i)
    {
        strcpy(data, poaId_[i]);
        data += strlen(poaId_[i])+1;
    }

    //
    // Add additional '\0'
    //
    *data++ = '\0';

    memcpy((Octet*)data, id.get_buffer(), id.length());
    
    return key._retn();
}

//
// Parse an object key, filling in the fields of the ObjectKeyData.
// Return false if the object-key cannot be parsed, true otherwise.
//
bool
POA_impl::parseObjectKey(const ObjectKey& key,
                         POA_impl::ObjectKeyData& keyData)
{
    //
    // Parse out the POA's. This should be a sequence of strings
    // terminated by two null values.
    //
    const Octet* data = key.get_buffer();
    const Octet* end = data + key.length();

    if(*data == '0') // persistent
    {
        keyData.persistent = true;
        ++data;
    }
    else
    {
        if(*data != '1')
            return false;
        keyData.persistent = false;
        ++data;
        //
        // Remember the start of the time stamp
        //
        const char* start = (const char*)data;
        while(data < end && *data != '\0')
            data++;
        if(data >= end)
            return false;
        
        keyData.createTime = atol(start);
        //
        // skip the null byte
        //
        ++data;
    }
    
    while(data < end)
    {
        //
        // Remember the start of the POA name
        //
        const char* start = (const char*)data;
        while(data < end && *data != '\0')
            data++;

        //
        // Ensure that we haven't gone too far...
        //
        if(data >= end)
            return false;

        //
        // Append this to the sequence of POA's.
        //
        keyData.poaID.append(start);
        
        //
        // Skip this byte, check to see if we have another '\0'
        //
        if(*++data == '\0')
        {
            //
            // Skip this byte
            //
            ++data;
            break;
        }
    }

    //
    // Verify that we haven't gone too far.
    //
    if(data >= end)
        return false;
    
    //
    // Remaining portion is the ObjectId.
    //
    ULong len = end - data;
    keyData.oid.length(len);
    memcpy(keyData.oid.get_buffer(), data, len);

    return true;
}

//
// Remove a child POA given it's name
//
void
POA_impl::removeChildPOA(const char* name)
{
    //
    // XXX - this violates private member needing concurrency
    //
    OB_WRITE_LOCK
    
    assert(children_ -> contains(name));
    POA_var child;
    children_ -> get(name, child);
    children_ -> remove(name);
}

//
// Generate a new object id, and add the servant to the active object map
//
ObjectId*
POA_impl::generateIdAddServant(ServantBase* s)
{
    ObjectId_var oid;
    createObjectId(oid);
    addServantActiveObjectMap(oid, s);

    return oid._retn();
}

//
// Add a servant to the active object map
//
void
POA_impl::addServantActiveObjectMap(const ObjectId& oid, ServantBase* servant)
{
    //
    // Increment the reference count of the servant
    //
    servant -> _add_ref();

    //
    // Insert the servant in the active object table with the provided id.
    //
    activeObjectMap_ -> put(oid, servant);

    //
    // If we're using UNIQUE_ID add the servant to the servantIdMap
    //
    if(policies_ -> idUniquenessPolicy() == UNIQUE_ID)
    {
        assert(!servantIdMap_ -> contains(servant));
        servantIdMap_ -> put(servant, oid);
    }
}

//
// Internal findPOA operation. This increments the reference count of
// the POA.
//
POA_impl*
POA_impl::findPOA(const char* adapter, Boolean activate, bool inRequest)
{
    bool check = true;
    
    //
    // If the child isn't registered, and we've been asked to activate
    // the POA and we have an AdapterActivator and the POAManager's
    // state is active then we activate the POA (see below requirement)
    //
    // From 11-18:
    //
    // In addition, when a POA manager is in the holding state, the
    // adapter activators registered with the associated POAs will not
    // get called.
    //

    if(!children_ -> contains(adapter) &&
       activate && !is_nil(adapterActivator_) &&
       (!inRequest || manager_ -> get_state() == POAManager::ACTIVE))
    {
        //
        // This can cause a create_POA - and therefore a deadlock on
        // our non-recursive mutex.
        //
        check = adapterActivator_ -> unknown_adapter(this, adapter)
	    == TRUE ? true : false;
    }

    //
    // Lookup the POA in our set of children.
    //
    POA_var child;
    if(check && children_ -> get(adapter, child))
        return dynamic_cast<POA_impl*>(child._retn());
    
   return 0;
}

//
// Determine if this POA matches the provided object key
//
bool
POA_impl::poaMatches(const POA_impl::ObjectKeyData& data)
{
    //
    // If the length is incorrect, return false
    //
    if(data.poaID.length() != poaId_.length())
	return false;

    //
    // Check each name
    //
    for(ULong i = 0; i < data.poaID.length(); ++i)
	if(strcmp(data.poaID[i], poaId_[i]) != 0)
	    return false;

    //
    // If both are persistent, then true
    //
    if(persistent_ && data.persistent)
        return true;
    
    assert(!persistent_);

    //
    // Otherwise, the time has to be the same
    //
    return data.createTime == poaCreateTime_.tv_sec;
}

//
// Validate the set of provided POA policies
//
bool
POA_impl::validatePolicies()
{
    //
    // NON_RETAIN requires either USE_DEFAULT_SERVANT or
    // USE_SERVANT_MANAGER
    //
    if(policies_ -> servantRetentionPolicy() == NON_RETAIN &&
       policies_ -> requestProcessingPolicy() != USE_SERVANT_MANAGER &&
        policies_ -> requestProcessingPolicy() != USE_DEFAULT_SERVANT)
        return false;

    //
    // USE_ACTIVE_OBJECT_MAP_ONLY requires RETAIN
    //
    if (policies_ -> requestProcessingPolicy() == USE_ACTIVE_OBJECT_MAP_ONLY &&
        policies_ -> servantRetentionPolicy() != RETAIN)
        return false;

    //
    // USE_DEFAULT_SERVANT requires MULTIPLE_ID
    //
    if (policies_ -> requestProcessingPolicy() == USE_DEFAULT_SERVANT &&
        policies_ -> idUniquenessPolicy() != MULTIPLE_ID)
        return false;
    
    //
    // IMPLICIT_ACTIVATION requires SYSTEM_ID and RETAIN
    //
    if (policies_ -> implicitActivationPolicy() == IMPLICIT_ACTIVATION &&
        policies_ -> servantRetentionPolicy() != RETAIN &&
        policies_ -> idAssignmentPolicy() != SYSTEM_ID)
        return false;
    
    return true;
}

StubImplBase*
POA_impl::createStubImplBase(const ObjectKey& key,
                             const POA_impl::ObjectKeyData& data,
                             const char* intf)
{
    //
    // If RETAIN policy then look in the active object map.
    //
    if(policies_ -> servantRetentionPolicy() == RETAIN)
    {
        ServantBase_var impl;
        if(activeObjectMap_ -> get(data.oid, impl))
        {
            if(impl -> _OB_isDynamic())
                return 0;
            return impl -> _OB_createStubImpl(this, data.oid, key);
        }
    }

    //
    // If the POA has the NON_RETAIN policy, or RETAIN but no entry in
    // the active object map
    //

    //
    // USE_ACTIVE_OBJECT_MAP_ONLY, then we have no servant
    //
    if(policies_ -> requestProcessingPolicy() == USE_ACTIVE_OBJECT_MAP_ONLY)
    {
        throw OBJECT_NOT_EXIST();
    }

    //
    // USE_DEFAULT_SERVANT
    //
    if(policies_ -> requestProcessingPolicy() == USE_DEFAULT_SERVANT)
    {
        //
        // If no servant has been associated with the POA, the POA
        // raises the OBJ_ADAPTER system exception.
        //
        if(defaultServant_.in() == 0)
            throw OBJ_ADAPTER();

        if(defaultServant_ -> _OB_isDynamic())
            return 0;
        
        return defaultServant_ -> _OB_createStubImpl(this, data.oid, key);
    }

    //
    // USE_SERVANT_MANAGER
    //
    assert(policies_ -> requestProcessingPolicy() == USE_SERVANT_MANAGER);

    //
    // If no servant has been associated with the POA, the POA raises
    // the OBJ_ADAPTER system exception.
    //
    if(is_nil(servantActivator_) &&
       is_nil(servantLocator_))
    {
        throw OBJ_ADAPTER();
    }
    
    ServantBase* servant;
    
    if(policies_ -> servantRetentionPolicy() == RETAIN)
    {
        try
        {
            servant = servantActivator_ -> incarnate(data.oid, this);
        }
        catch(const ForwardRequest& ex)
        {
            throw StubImplForward(ex.forward_reference -> _OB_ior());
        }
        
        //
        // 11-23:
        //
        // If a ServantManager returns a null Servant (or the
        // equivalent in a language mapping) as the result of an
        // incarnate() or preinvoke() operation, the POA will return
        // the OBJ_ADAPTER system exception as the result of the
        // request.
        //
        if(servant == 0)
            throw OBJ_ADAPTER();
        
        //
        // 11-25:
        //
        // If the incarnate operation returns a servant that is
        // already active for a different Object Id and if the POA
        // also has the UNIQUE_ID policy, the incarnate has violated
        // the POA policy and is considered to be in error. The POA
        // will raise an OBJ_ADAPTER system exception for the request.
        //
        if(policies_ -> idUniquenessPolicy() == UNIQUE_ID &&
           servantIdMap_ -> contains(servant))
        {
            throw OBJ_ADAPTER();
        }
        
        //
        // Enter the servant into the active object map
        //
        addServantActiveObjectMap(data.oid, servant);
        if(servant -> _OB_isDynamic())
            return 0;
        return servant -> _OB_createStubImpl(this, data.oid, key);
    }

    //
    // NO_RETAIN
    //
    return 0;
}

void
POA_impl::setupInvoke(Downcall_ptr downcall,
                      ULong requestId,
                      const ObjectKey& key,
                      const ObjectId& oid,
                      const char* op,
                      Buffer& buf,
                      bool sw,
                      ULong offIn,
                      ULong offOut,
                      OCI::TransportInfo_ptr info)
{
    for(;;)
    {
        switch(manager_ -> get_state())
        {
        case POAManager::HOLDING:
            switch(ORB::server_conc_model())
            {
            case ORB::ServerConcModelBlocking:
                //
                // XXX - deadlock
                //
                assert(false);
                return;
                
            case ORB::ServerConcModelReactive:
            {
                //
                // Place request in queue
                //
                InvocationRecord* rec = new InvocationRecord;
                rec -> downcall = Downcall::_duplicate(downcall);
                rec -> requestId = requestId;
                rec -> key = key;
                rec -> oid = oid;
                rec -> op = op;
                rec -> buf.consume(buf);
                rec -> sw = sw;
                rec -> offIn = offIn;
                rec -> offOut = offOut;
                rec -> info = OCI::TransportInfo::_duplicate(info);
                
                records_.append(rec);
                return;
            }

#ifdef HAVE_JTC
            case ORB::ServerConcModelThreaded:
            case ORB::ServerConcModelThreadPerClient:
            case ORB::ServerConcModelThreadPerRequest:
            case ORB::ServerConcModelThreadPool:
            {
                POAManager_impl* m =
                    dynamic_cast<POAManager_impl*>(manager_.in());
                m -> _OB_waitStateChange();
                break;
            }
#endif
            default:
                assert(false);
            }
            break;
            
        case POAManager::DISCARDING:
            if(!is_nil(downcall))
                downcall -> systemException(requestId, buf,
                                            CORBA::TRANSIENT());
            return;
            
        case POAManager::INACTIVE:
            if(!is_nil(downcall))
                downcall -> systemException(requestId, buf, OBJ_ADAPTER());
            return;
            
        case POAManager::ACTIVE:
            invoke(downcall, requestId, key, oid, op, buf, sw,
                   offIn, offOut, info);
            return;
            
        default:
            assert(false);
        }
    }
}

void
POA_impl::invoke(Downcall_ptr downcall,
                 ULong requestId,
                 const ObjectKey& key,
                 const ObjectId& oid,
                 const char* op,
                 Buffer& buf,
                 bool sw,
                 ULong offIn,
                 ULong offOut,
                 OCI::TransportInfo_ptr info)
{
    //
    // Locate the servant.
    //
    ServantBase_var servant;
    ServantLocator::Cookie cookie;

    try
    {
        servant = _OB_locateServant(oid, op, cookie);
    }
    catch(const ForwardRequest& ex)
    {
        if(!is_nil(downcall))
            downcall -> forward(requestId, buf, ex.forward_reference);
        return;
    }
    catch(const SystemException& ex)
    {
        if(!is_nil(downcall))
            downcall -> systemException(requestId, buf, ex);
        return;
    }
    
    if(servant.in() != 0)
    {
        _OB_preinvoke(op, oid, key, servant, cookie, info);

        //
        // Ensure servant is not destroyed for duration of invocation.
        // The reference count is decremented at the end of this
        // block, prior to calling postinvoke (as per the
        // specification).
        //
	servant -> _add_ref();
        
        bool replied = false;
        DispatchStatus status;
        try
        {
            status = servant -> _OB_dispatch(op, buf, sw, offIn, offOut);
        }
        catch(const SystemException& ex)
        {
            if(!is_nil(downcall))
                downcall -> systemException(requestId, buf, ex);
            replied = true;
        }
        catch(...)
        {
	    //
            // Print a warning message
            //
	    String_var msg =
		string_dup("Servant method raised a non-CORBA exception");
	    if(!is_nil(downcall))
		msg += "\nClient receives this exception as CORBA::UNKNOWN";

	    msg += "\noperation name=\"";
	    msg += op;
	    msg += "\" acceptor=";
	    OCI::AcceptorInfo_var acceptorInfo = info -> acceptor_info();
	    OCI::IIOP::AcceptorInfo_var iiopAcceptorInfo =
		OCI::IIOP::AcceptorInfo::_narrow(acceptorInfo);
	    if(!is_nil(iiopAcceptorInfo))
	    {
		String_var host = iiopAcceptorInfo -> host();
		UShort port = iiopAcceptorInfo -> port();
		msg += host;
		msg += ":";
		msg += port;
	    }
	    else
	    {
		msg += "not iiop";
	    }
	    MessageViewer::instance() -> warning(msg);

            if(!is_nil(downcall))
                downcall -> systemException(requestId, buf, UNKNOWN());
            replied = true;
        }
        
        if(!replied)
        {
            switch(status)
            {
            case DispatchStatusOK:
                if(!is_nil(downcall))
                    downcall -> response(requestId, buf);
                break;
            case DispatchStatusExcept:
                if(!is_nil(downcall))
                    downcall -> userException(requestId, buf);
                break;
            case DispatchStatusNoOperation:
                if(!is_nil(downcall))
                    downcall -> systemException(requestId, buf,
                                                BAD_OPERATION());
                break;
            default:
                assert(false);
            }
        }

        servant -> _remove_ref();

        _OB_postinvoke();
    }
    else
    {
	//
	// Accept _non_existent and _not_existent. (See the ORB Core
	// revision task force mailing list for a discussion why.)
	//
	if((strcmp(op, "_non_existent") == 0 ||
	    strcmp(op, "_not_existent") == 0) && !is_nil(downcall))
        {
            Boolean r = true;
            
            CountStream cnt(offOut);
            cnt.count_boolean();
            
            buf.alloc(cnt.count());
            OutputStream out(buf, offOut);
            out.write_boolean(r);
            
            downcall -> response(requestId, buf);
        }
        if(!is_nil(downcall))
            downcall -> systemException(requestId, buf, OBJECT_NOT_EXIST());
    }
}

//
// Common initialization code
//
void
POA_impl::init(const PolicyList& policies, POAManager_ptr manager)
{
    nextSystemID_ = 0;
    num_ = 0;
    destroyed_ = false;

    policies_ = new POAPolicies(policies);
    activeObjectMap_ = new ActiveObjectMap(1023);
    servantIdMap_ = new ServantIdMap(1023);
    children_ = new POAHashTable(31);
    localStubImplSeqMap_ = new LocalStubImplSeqMap(1023);
    if(is_nil(manager))
    {
        manager_ = new POAManager_impl();
    }
    else
    {
        manager_ = POAManager::_duplicate(manager);
    }

    //
    // Register with my POAManager
    //
    POAManager_impl* m = dynamic_cast<POAManager_impl*>(manager_.in());
    m -> _OB_add(this);

    persistent_ = policies_ -> lifespanPolicy() == PERSISTENT;
    if(!persistent_)
        poaCreateTime_ = TimerList::timeNow();
}


// ----------------------------------------------------------------------
// POA_impl constructor and destructor
// ----------------------------------------------------------------------

//
// Private, internal constructor
//
POA_impl::POA_impl(Current_impl* current,
                   POA_impl* parent,
                   POAManager_ptr manager,
                   const char* name,
                   const PolicyList& policies)
    : name_(string_dup(name)),
      parent_(parent),
      poaCurrent_(current)
{
    assert(parent != 0);
    poaId_ = parent -> poaId_;
    poaId_.append(name);
    init(policies, manager);
}

//
// Constructor for the "rootPOA".
//
POA_impl::POA_impl()
    : name_(string_dup("RootPOA")),
      parent_(0),
      poaCurrent_(new Current_impl)
{
    //
    // Register the POA::Current, and the OCI::Current classes with
    // the ORB.
    //
    ORB_ptr orb = ORB::_OB_instance_nodup();
    
    orb -> add_initial_reference("POACurrent", poaCurrent_);

    PolicyList pl;
    init(pl, POAManager::_nil());
    poaId_.append(name_);
}

POA_impl::~POA_impl()
{
#ifdef OB_TRACE
    {
	String_var msg = string_dup("");
	msg += "Destroy POA: ";
	msg += name_;
        MessageViewer::instance() -> trace(2, msg);
    }
#endif

    //
    // Unregister from my POAManager
    //
    POAManager_impl* m = dynamic_cast<POAManager_impl*>(manager_.in());
    m -> _OB_remove(this);

    delete policies_;
    delete activeObjectMap_;
    delete servantIdMap_;
    delete children_;

    {
	//
	// Clear each of the sequences from the map
	//
	LocalStubImplSeqMap::Enumerator keys =
	    localStubImplSeqMap_ -> keys();
	for(ULong i = 0; i < keys.length(); ++i)
	{
	    StubImplBasePtrSeq* seq;
	    localStubImplSeqMap_ -> get(keys[i], seq);
	    delete seq;
	}
	localStubImplSeqMap_ -> clear();
    }
    delete localStubImplSeqMap_;

    //
    // If this is the RootPOA then release the poaCurrent_
    //
    if(parent_ == 0)
    {
	assert(strcmp(name_, "RootPOA") == 0);
        release(poaCurrent_);
    }
}

// ----------------------------------------------------------------------
// POA_impl public member implementation
// ----------------------------------------------------------------------

POA_ptr
POA_impl::create_POA(const char* adapter,
                     POAManager_ptr manager,
                     const PolicyList& policies)
    throw(POA::AdapterAlreadyExists,
          POA::InvalidPolicy,
          SystemException)
{
    OB_WRITE_LOCK

    if(children_ -> contains(adapter))
        throw POA::AdapterAlreadyExists();

    POA_impl* child =
        new POA_impl(poaCurrent_, this, manager, adapter, policies);

    //
    // We have this crap in to work around the virtual inheritance
    // constructor bug with MSVC 5.0.
    //
    if(!child -> validatePolicies())
    {
        //
        // Trash the child
        //
        release(child);
        throw POA::InvalidPolicy();
    }
    
    children_ -> put(adapter, POA::_duplicate(child));
    
    return child;
}

POA_ptr
POA_impl::find_POA(const char* adapter, Boolean activate)
    throw(POA::AdapterNonExistent,
          SystemException)
{
    OB_WRITE_LOCK

    POA_impl* poa = findPOA(adapter, activate, false);
    if(poa == 0)
        throw POA::AdapterNonExistent();

    return poa;
}

#if 0
//
// Proposed destroy implementation
//
void
POA_impl::destroy(Boolean etherealize, Boolean wait)
    throw(SystemException)
{
    if(wait && currentThreadIsInvocationContext())
	throw BAD_INV_ORDER();
    
    inDestroy_++;

    //
    // Is this another concurrent call of destroy?
    //
    if(inDestroy_ > 1)
    {
	//
	// If this thread is waiting, then wait for all requests to
	// complete.
	//
	if(wait)
	    waitRequestsToComplete();
	return;
    }

    destroyDecendents();

    //
    // Deactivate every local stub impl - this means that new
    // invocations will have to pick up the new POA.
    //XXX
    //
    
    //
    // We need to tell the POA not to accept any new requests.
    //
    blockNewRequests();

    //
    // If we want to wait for requests to complete, wait.
    //
    if(wait)
	waitForRequestsToComplete();

    //
    // The apparent destruction of the POA occurs before the
    // etherealization is done. Once this has completed an invocation
    // of create_POA with the same POA name blocks until the POA
    // destruction has finished. We can accomplish this by marking
    // this POA name in our parent POA as blocking.
    //
    disengageFromParent();

    //
    // We need to check the destroyed_ flag in each operation, and
    // throw OBJECT_NOT_EXIST(), if true.
    //
    destroyed_ = true;

    if(etherealize && wait)
	etherealizeServants();

    //
    // At this point we actually remove the POA reference in the
    // parent POA. This will also wake any blocked callers of
    // create_POA in the parent POA.
    //
    removeFromParent();

    --inDestroy_;

    //
    // We need to release the servants in the active object map.  If
    // there's a servant activator, then all of the servants will have
    // already been etherealized (and removed from the active object
    // map), otherwise we have to invoke _remove_ref on each of them.
    // This can be accomplished just by clearing the active object
    // map.
    //
    activeObjectMap_ -> clear();
}
#endif

void
POA_impl::destroy(Boolean etherealize, Boolean wait)
    throw(SystemException)
{
    //
    // XXX
    // still multiple problems with this method.
    //
    // wait_for_completion is not implemented at all.
    //
    // Is there a problem with the `apparent destruction' of the POA?
    // Here's some notes from Mark:
    //
    //  I'm a little unclear about what exactly "apparent destruction"
    //  means wrt a POA object.  Does that mean it's removed from its
    //  parent's list of children?  Or something else?
    //
    //  If the former, I think you're doing this too early.  Currently
    //  you do this right at the start of destroy(), but the spec says
    //  (pg 11-33) that "The apparent destruction of the POA occurs
    //  only after all executing requests in the POA have completed."
    //
    //  Furthermore, if etherealize_objects is TRUE, then the apparent
    //  destruction of the POA must occur *before* etherealizing the
    //  servants.
    //
    //  So, what the hell are they talking about?! :-)
    //
    // Is this requirement met?
    //
    // Once apparent destruction has occurred, the POA behaves as if
    // its POAManager is in the holding state until destruction is
    // complete. Thus, for example, an invocation of create_POA with
    // the same name blocks until POA destruction has finished.
    //
    //
    
    //
    // Remove myself from my parent.
    //
    if(parent_ != 0)
        parent_ -> removeChildPOA(name_);

    //
    // After this point the POA cannot be located by incoming requests
    // since it is not in the list of POA's anymore. However, it's
    // possible that a method invocation has already gotten a
    // reference to this POA. Both destroy and invoke attempt to write
    // lock the POA. If destroy() write locks the POA first then the
    // invocation thread will obtain the lock after the POA has been
    // destroyed, and will check the destroyed_ flag, and attempt to
    // re-obtain the POA. Otherwise the invoke will acquire the lock
    // first and the method call will be dispatched before the
    // destroy() is completed.
    //
    // I'm not convinced that this implementation satisfies this (From
    // 11-34) since the invocation can obtain the POA, but not
    // dispatch a request on it (thus there are outstanding requests
    // to execute on the POA and this text is violated):
    //
    // After all descendant POAs have been destroyed and their
    // servants etherealized, the POA continues to process requests
    // until there are no requests executing in the POA. The apparent
    // destruction of the POA occurs only after all executing requests
    // in the POA have completed. After destruction has become
    // apparent, the POA may be re-created via either an
    // AdapterActivator or a call to create_POA.
    //

    //
    // Obtain write lock on the implementation.
    //
    OB_WRITE_LOCK
    
    //
    // Recursively destroy all children
    //
    POAHashTable::Enumerator keys = children_ -> keys();
    for(ULong i = 0; i < keys.length(); ++i)
    {
        POA_var child;
        bool rc = children_ -> get(keys[i], child);
        assert(rc);
        child -> destroy(etherealize, wait);
    }

    //
    // Etherealize all servants, if asked.
    //
    if(etherealize)
    {
        if(policies_ -> servantRetentionPolicy() == RETAIN &&
           policies_ -> requestProcessingPolicy() == USE_SERVANT_MANAGER &&
           !is_nil(servantActivator_))
        {
            ActiveObjectMap::Enumerator keys = activeObjectMap_ -> keys();
            for(ULong i = 0; i < keys.length(); ++i)
            {
                ServantBase_var servant;
                activeObjectMap_ -> get(keys[i], servant);
                activeObjectMap_ -> remove(keys[i]);
                
                bool remaining = false;
                if(policies_ -> idUniquenessPolicy() == MULTIPLE_ID)
                {
                    //
                    //XXX - optimize!
                    //
                    // Marks observation: We could add
                    // a 'containsValue(T*)' function to the hashtable,
                    // but that raises the issue of determining value
                    // equality, not to mention that this would also be
                    // a linear search, although a slightly more
                    // efficient one. (Matthew: There must be a better way
                    // than a straight linear search anyway - probably
                    // a hash table to a number of times in the active
                    // object map.
                    //
                    // Don't forget if we fix this then the same thing
                    // is done is deactivate_object.
                    //XXX
                    //
                    // If servant is still in the active object map, then
                    // we still have remaining activations.
                    //
                    for(ULong j = i; j < keys.length(); ++j)
                    {
                        ServantBase_var servant2;
                        activeObjectMap_ -> get(keys[j], servant2);
                        if(servant2.in() == servant.in())
                        {
                            remaining = true;
                            break;
                        }
                    }
                }

                //
                // etherealize consumes the servant - decrement
                // the reference count to its state prior to being
                // added to the active object map. Note that we
                // are not allowed to access the servant after
                // calling etherealize, so we can't use a
                // ServantBase_var.
                //
                servant -> _remove_ref();
                try
                {
                    servantActivator_ -> etherealize(keys[i], this,
                                                     servant._retn(),
                                                     true, remaining);
                }
                catch(const SystemException&)
                {
                    // Ignore
                }
            }
        }
    }

    //
    // We need to release the servants in the active object map.  If
    // there's a servant activator, then all of the servants will have
    // already been etherealized (and removed from the active object
    // map), otherwise we have to invoke _remove_ref on each of them.
    // This can be accomplished just by clearing the active object
    // map.
    //
    activeObjectMap_ -> clear();

    //
    // Clear internal state variables
    //
    defaultServant_ = (Servant)0;
    servantLocator_ = ServantLocator::_nil();
    servantActivator_ = ServantActivator::_nil();
    adapterActivator_ = AdapterActivator::_nil();

    //
    // We use this flag so that the method invocation dispatcher can
    // determine if it has located a POA that has been destroyed. If
    // so it is expected to re-locate the POA.
    //
    destroyed_ = true;
}

// ----------------------------------------------------------------------
// Factories for policy objects
// ----------------------------------------------------------------------

ThreadPolicy_ptr
POA_impl::create_thread_policy(ThreadPolicyValue value)
    throw(SystemException)
{
    return new ThreadPolicy_impl(value);
}

LifespanPolicy_ptr
POA_impl::create_lifespan_policy(LifespanPolicyValue value)
    throw(SystemException)
{
    return new LifespanPolicy_impl(value);
}

IdUniquenessPolicy_ptr
POA_impl::create_id_uniqueness_policy(IdUniquenessPolicyValue value)
    throw(SystemException)
{
    return new IdUniquenessPolicy_impl(value);
}

IdAssignmentPolicy_ptr
POA_impl::create_id_assignment_policy(IdAssignmentPolicyValue value)
    throw(SystemException)
{
    return new IdAssignmentPolicy_impl(value);
}

ImplicitActivationPolicy_ptr
POA_impl::create_implicit_activation_policy(
    ImplicitActivationPolicyValue value)
    throw(SystemException)
{
    return new ImplicitActivationPolicy_impl(value);
}

ServantRetentionPolicy_ptr
POA_impl::create_servant_retention_policy(ServantRetentionPolicyValue value)
    throw(SystemException)
{
    return new ServantRetentionPolicy_impl(value);
}

RequestProcessingPolicy_ptr
POA_impl::create_request_processing_policy(RequestProcessingPolicyValue value)
    throw(SystemException)
{
    return new RequestProcessingPolicy_impl(value);
}

SynchronizationPolicy_ptr
POA_impl::create_synchronization_policy(SynchronizationPolicyValue value)
    throw(SystemException)
{
    return new SynchronizationPolicy_impl(value);
}

// ----------------------------------------------------------------------
// POA attributes
// ----------------------------------------------------------------------

char*
POA_impl::the_name()
    throw(SystemException)
{
    //
    // Invariant -- doesn't need concurrency protection
    //
    return string_dup(name_);
}

POA_ptr
POA_impl::the_parent()
    throw(SystemException)
{
    //
    // Invariant -- doesn't need concurrency protection
    //
    return POA::_duplicate(parent_);
}

POAManager_ptr
POA_impl::the_POAManager()
    throw(SystemException)
{
    //
    // Invariant -- doesn't need concurrency protection
    //
    return POAManager::_duplicate(manager_);
}

AdapterActivator_ptr
POA_impl::the_activator()
    throw(SystemException)
{
    OB_READ_LOCK

    return AdapterActivator::_duplicate(adapterActivator_);
}

void
POA_impl::the_activator(AdapterActivator_ptr activator)
    throw(SystemException)
{
    OB_WRITE_LOCK

    adapterActivator_ = AdapterActivator::_duplicate(activator);
}

// ----------------------------------------------------------------------
// Servant manager registration
// ----------------------------------------------------------------------

ServantManager_ptr
POA_impl::get_servant_manager()
    throw(POA::WrongPolicy,
          SystemException)
{
    //
    // Requires the USE_SERVANT_MANAGER policy
    //
    if(policies_ -> requestProcessingPolicy() !=
       USE_SERVANT_MANAGER)
    {
        throw POA::WrongPolicy();
    }

    OB_READ_LOCK

    if(!is_nil(servantActivator_))
        return ServantManager::_duplicate(servantActivator_);
    
    if(!is_nil(servantLocator_))
        return ServantManager::_duplicate(servantLocator_);

    return ServantManager::_nil();
}

void
POA_impl::set_servant_manager(ServantManager_ptr imgr)
    throw(POA::WrongPolicy,
          SystemException)
{
    //
    // Requires the USE_SERVANT_MANAGER policy
    //
    if(policies_ -> requestProcessingPolicy() !=
       USE_SERVANT_MANAGER)
    {
        throw POA::WrongPolicy();
    }

    OB_WRITE_LOCK

    //
    // Attempting to set the servant manager after one has already
    // been set will result in the BAD_INV_ORDER exception being
    // raised.
    //
    if(!is_nil(servantActivator_) || !is_nil(servantLocator_))
        throw BAD_INV_ORDER();

    //
    // Check the policy to determine which type of locator this is.
    //
    if(policies_ -> servantRetentionPolicy() == RETAIN)
    {
        servantActivator_ = ServantActivator::_narrow(imgr);
        if(is_nil(servantActivator_))
            throw POA::WrongPolicy();
    }
    else
    {
        servantLocator_ = ServantLocator::_narrow(imgr);
        if(is_nil(servantLocator_))
            throw POA::WrongPolicy();
    }
}

Servant
POA_impl::get_servant()
    throw(POA::NoServant,
          POA::WrongPolicy,
          SystemException)
{
    if(policies_ -> requestProcessingPolicy() != USE_DEFAULT_SERVANT)
    {
        throw POA::WrongPolicy();
    }

    OB_READ_LOCK

    if(defaultServant_.in() == 0)
        throw POA::NoServant();

    defaultServant_ -> _add_ref();
    return defaultServant_;
}

void
POA_impl::set_servant(Servant p_servant)
    throw(POA::WrongPolicy,
          SystemException)
{
    if(policies_ -> requestProcessingPolicy() != USE_DEFAULT_SERVANT)
    {
        throw POA::WrongPolicy();
    }

    OB_WRITE_LOCK

    defaultServant_ = p_servant;
    defaultServant_ -> _add_ref();
}

ObjectId*
POA_impl::activate_object(Servant servant)
    throw(POA::ServantAlreadyActive,
          POA::WrongPolicy,
          SystemException)
{
    //
    // Requires SYSTEM_ID and RETAIN policy
    //
    if(policies_ -> idAssignmentPolicy() != SYSTEM_ID ||
       policies_ -> servantRetentionPolicy() != RETAIN)
    {
        throw POA::WrongPolicy();
    }

    OB_WRITE_LOCK

    //
    // If the POA has the UNIQUE_ID policy and the servant is already
    // in the active object table then raise an ServantAlreadyActive
    // exception.
    //
    if(policies_ -> idUniquenessPolicy() == UNIQUE_ID &&
       servantIdMap_ -> contains(servant))
    {
        throw POA::ServantAlreadyActive();
    }

    return generateIdAddServant(servant);
}

void
POA_impl::activate_object_with_id(
    const ObjectId& oid,
    Servant p_servant)
    throw(POA::ServantAlreadyActive,
          POA::ObjectAlreadyActive,
          POA::WrongPolicy,
          SystemException)
{
    //
    // Requires the RETAIN policy.
    //
    if(policies_ -> servantRetentionPolicy() != RETAIN)
    {
        throw POA::WrongPolicy();
    }

    //
    // If SYSTEM_ID policy then verify that the id was generated
    // by the POA.
    //
    if(policies_ -> idAssignmentPolicy() == SYSTEM_ID &&
       !isSystemGeneratedId(oid))
    {
        throw BAD_PARAM();
    }

    OB_WRITE_LOCK

    //
    // Find out whether a servant is already bound under this id.
    //
    if(activeObjectMap_ -> contains(oid))
        throw POA::ObjectAlreadyActive();

    //
    // If the POA has the UNIQUE_ID policy and the servant is already
    // in the active object table then raise a ServantAlreadyActive
    // exception.
    //
    if(policies_ -> idUniquenessPolicy() == UNIQUE_ID &&
       servantIdMap_ -> contains(p_servant))
    {
        throw POA::ServantAlreadyActive();
    }

    addServantActiveObjectMap(oid, p_servant);
}

void
POA_impl::deactivate_object(const ObjectId& oid)
    throw(POA::ObjectNotActive,
          POA::WrongPolicy,
          SystemException)
{
    //
    // XXX
    // this operation is most likely incorrect:
    //
    // An ObjectId which has been deactivated continues to process
    // requests until there are no active requests for that
    // ObjectId. A deactivated ObjectId is removed from the Active
    // Object Map when all requests executing for that ObjectId have
    // completed.
    //
    
    //
    // Requires the RETAIN policy.
    //
    if(policies_ -> servantRetentionPolicy() != RETAIN)
    {
        throw POA::WrongPolicy();
    }

    OB_WRITE_LOCK
    
    //
    // If no object in the active object table associated with this key
    // then raise an ObjectNotActive exception.
    //
    if(!activeObjectMap_ -> contains(oid))
    {
        throw POA::ObjectNotActive();
    }

    //
    // deactivate_object is required to invoke _remove_ref on the
    // servant if a servant activator is not in use.  The 'servant'
    // variable below will do this when it goes out of scope.
    //
    ServantBase_var servant;
    activeObjectMap_ -> get(oid, servant);
    
    activeObjectMap_ -> remove(oid);

    //
    // Mark each stub in the map as deactivated
    //
    StubImplBasePtrSeq* seq;
    if(localStubImplSeqMap_ -> get(oid, seq))
        for(ULong i = 0; i < seq -> length(); ++i)
            (*seq)[i] -> _OB_deactivate();

    //
    // If we're using UNIQUE_ID then remove the servant from the
    // servantIdMap.
    //
    if(policies_ -> idUniquenessPolicy() == UNIQUE_ID)
    {
        servantIdMap_ -> remove(servant);
    }

    //
    // If we have a servant activator then we have to etherealize the
    // object.
    //
    if(!is_nil(servantActivator_))
    {
        bool remaining = false;

        //
        //XXX - optimize!
        //
        // If servant is still in the active object map, then we still
        // have remaining activations.
        //
        ActiveObjectMap::Enumerator keys = activeObjectMap_ -> keys();
        for(ULong i = 0; i < keys.length(); ++i)
        {
            ServantBase_var servant2;
            activeObjectMap_ -> get(keys[i], servant2);
            if(servant2.in() == servant)
            {
                remaining = true;
                break;
            }
        }

        //
        // etherealize consumes the servant
        //
        try
        {
            servantActivator_ -> etherealize(oid, this, servant._retn(), false,
                                             remaining);
        }
        catch(const SystemException&)
        {
            // Ignore
        }
    }
}

Object_ptr
POA_impl::create_reference(const char* intf)
    throw(POA::WrongPolicy,
          SystemException)
{
    //
    // Requires the SYSTEM_ID policy.
    //
    if(policies_ -> idAssignmentPolicy() != SYSTEM_ID)
    {
        throw POA::WrongPolicy();
    }

    OB_WRITE_LOCK

    ObjectId_var oid;
    createObjectId(oid);
    ObjectKey_var key = createObjectKey(oid);

    return _OB_createReference(key, intf);
}

Object_ptr
POA_impl::create_reference_with_id(const ObjectId& oid,
                                   const char* intf)
    throw(POA::WrongPolicy,
          SystemException)
{
    //
    // If SYSTEM_ID policy is set we should check the oid to see
    // if it was not generated by the POA and raise BAD_PARAM if so.
    //
    if(policies_ -> idAssignmentPolicy() == SYSTEM_ID &&
       !isSystemGeneratedId(oid))
    {
        throw BAD_PARAM();
    }

    ObjectKey_var key = createObjectKey(oid);

    return _OB_createReference(key, intf);
}

ObjectId*
POA_impl::servant_to_id(Servant p_servant)
    throw(POA::ServantNotActive,
          POA::WrongPolicy,
          SystemException)
{
    //
    // Requires USE_DEFAULT_SERVANT policy or RETAIN policy and either
    // the UNIQUE_ID policy or the IMPLICIT_ACTIVATION policies.
    //
    if(policies_ -> requestProcessingPolicy() != USE_DEFAULT_SERVANT &&
       (policies_ -> servantRetentionPolicy() != RETAIN ||
	(policies_ -> idUniquenessPolicy() != UNIQUE_ID &&
	 policies_ -> implicitActivationPolicy() != IMPLICIT_ACTIVATION)))
    {
        throw POA::WrongPolicy();
    }

    OB_WRITE_LOCK

    //
    // If the POA has the USE_DEFAULT_SERVANT policy, the servant
    // specified is the default servant, and the operation is being
    // invoked in the context of executing a request on the default
    // servant, then the ObjectId associated with the current
    // invocation is returned.
    //
    if(policies_ -> requestProcessingPolicy() == USE_DEFAULT_SERVANT)
	if(poaCurrent_ -> _OB_inUpcall() &&
	   poaCurrent_ -> _OB_getServant() == defaultServant_.in())
	    return poaCurrent_ -> get_object_id();
        
    //
    // If the POA has the UNIQUE_ID policy and the servant is active,
    // then return the ObjectId associated with the servant.
    //
    if(policies_ -> idUniquenessPolicy() == UNIQUE_ID)
    {
        ObjectId id;
        if(servantIdMap_ -> get(p_servant, id))
            return new ObjectId(id);
    }

    //
    // If the POA doesn't have the IMPLICIT_ACTIVATION policy or
    // either the POA doesn't have the MULTIPLE_ID policy and the
    // servant is active then a ServantNotActive exception
    //
    if(policies_ -> implicitActivationPolicy() != IMPLICIT_ACTIVATION ||
       (policies_ -> idUniquenessPolicy() != MULTIPLE_ID &&
        servantIdMap_ -> contains(p_servant)))
    {
        throw POA::ServantNotActive();
    }

    //
    // Add this servant to the active object map.
    //
    return generateIdAddServant(p_servant);
}

Object_ptr
POA_impl::servant_to_reference(Servant servant)
    throw(POA::ServantNotActive,
          POA::WrongPolicy,
          SystemException)
{
    //
    // No concurrency locking, the internal methods do this.
    //

    //
    // If the operation was invoked in the context of executing
    // request on the specified servant, the reference associated with
    // the current invocation is returned.
    //
    if(poaCurrent_ -> _OB_inUpcall() &&
       poaCurrent_ -> _OB_getServant() == servant)
    {
	ObjectId_var oid = poaCurrent_ -> get_object_id();
	String_var intf = servant -> _primary_interface(oid, this);
	return _OB_createReference(poaCurrent_ -> _OB_getKey(), intf);
    }
    
    ObjectId_var oid = servant_to_id(servant);
    String_var intf = servant -> _primary_interface(oid, this);
    return create_reference_with_id(oid, intf);
}

Servant
POA_impl::reference_to_servant(Object_ptr reference)
    throw(POA::ObjectNotActive,
          POA::WrongAdapter,
          POA::WrongPolicy,
          SystemException)
{
    //
    // Requires the RETAIN policy or the USE_DEFAULT_SERVANT policy.
    //
    if(policies_ -> servantRetentionPolicy() != RETAIN &&
       policies_ -> requestProcessingPolicy() != USE_DEFAULT_SERVANT)
    {
        throw POA::WrongPolicy();
    }

    OB_READ_LOCK
    
    //
    // If the policy is RETAIN then we return the servant associated
    // with the key in the active object map.
    //
    if(policies_ -> servantRetentionPolicy() == RETAIN)
    {
        ObjectId_var oid = reference_to_id(reference);
        ServantBase_var s;

        if(activeObjectMap_ -> get(oid, s))
            return s._retn();
    }

    //
    // If the policy is not USE_DEFAULT_SERVANT or the policy is
    // USE_DEFAULT_SERVANT and there is no default servant then we
    // throw the ObjectNotActive exception
    //
    if(policies_ -> requestProcessingPolicy() != USE_DEFAULT_SERVANT ||
       defaultServant_.in() == 0)
    {
        throw POA::ObjectNotActive();
    }
        
    defaultServant_ -> _add_ref();
    return defaultServant_;
}

ObjectId*
POA_impl::reference_to_id(Object_ptr reference)
    throw(POA::WrongAdapter,
          POA::WrongPolicy,
          SystemException)
{
    //
    // Extract the object key from the IOR of the object reference.
    //
    OCI::AccRegistry_var registry = OARoot::instance() -> getAccRegistry();
    OCI::ObjectKey_var key = registry -> is_local(reference -> _OB_ior());

    //
    // Is the key local?
    //
    if(key -> length() > 0)
    {
        ObjectKeyData keyData;
        //
        // Verify the poaIds are the same, have the persistency values
        // and the same create time if transient
        //
        if(parseObjectKey(key, keyData) && poaMatches(keyData))
            return new ObjectId(keyData.oid);
    }

    throw POA::WrongAdapter();
    return 0; // Some compilers need this
}

Servant
POA_impl::id_to_servant(const ObjectId& oid)
    throw(POA::ObjectNotActive,
          POA::WrongPolicy,
          SystemException)
{
    //
    // Requires the RETAIN policy or the USE_DEFAULT_SERVANT policy.
    //
    if(policies_ -> servantRetentionPolicy() != RETAIN &&
       policies_ -> requestProcessingPolicy() != USE_DEFAULT_SERVANT)
    {
        throw POA::WrongPolicy();
    }

    OB_READ_LOCK
    
    //
    // If the policy is RETAIN then we return the servant associated
    // with the key in the active object map.
    //
    if(policies_ -> servantRetentionPolicy() == RETAIN)
    {
        ServantBase_var s;
        if(activeObjectMap_ -> get(oid, s))
            return s._retn();
    }

    //
    // If the policy is USE_DEFAULT_SERVANT then we return the default
    // servant, if it exists
    //
    if(policies_ -> requestProcessingPolicy() != USE_DEFAULT_SERVANT ||
       defaultServant_.in() == 0)
    {
        throw POA::ObjectNotActive();
    }
    
    defaultServant_ -> _add_ref();
    
    return defaultServant_;
}

Object_ptr
POA_impl::id_to_reference(const ObjectId& oid)
    throw(POA::ObjectNotActive,
          POA::WrongPolicy,
          SystemException)
{
    //
    // Requires the RETAIN policy.
    //
    if(policies_ -> servantRetentionPolicy() != RETAIN)
    {
        throw POA::WrongPolicy();
    }

    OB_WRITE_LOCK
    
    ServantBase_var servant;
    if(!activeObjectMap_ -> get(oid, servant))
        throw POA::ObjectNotActive();

    String_var intf = servant -> _primary_interface(oid, this);

    ObjectKey_var key = createObjectKey(oid);
    return _OB_createReference(key, intf);
}

void
POA_impl::_OB_preinvoke(const char* op,
			const ObjectId& oid,
                        const ObjectKey& key,
                        ServantBase* servant,
                        ServantLocator::Cookie cookie,
                        OCI::TransportInfo_ptr info)
{
    //
    // Setup the OCI::Current context
    //
    OCI::Current_impl* ociCurrent = ORB::_OB_instance_nodup() ->
        _OB_OCICurrent();
    ociCurrent -> _OB_preinvoke(info);

    //
    // Setup the PortableServer::Current context
    //
    poaCurrent_ -> _OB_preinvoke(this, servant, op, oid, key, cookie);

    //
    // Depending on the synchronization policy we have to lock a
    // variety of mutexes.
    //
#ifdef HAVE_JTC
    switch(policies_ -> synchronizationPolicy())
    {
    case NO_SYNCHRONIZATION:
	break;

    case SYNCHRONIZE_ON_POA:
	poaSyncMutex_.lock();
	break;

    case SYNCHRONIZE_ON_ORB:
	orbSyncMutex_.lock();
	break;

    case SYNCHRONIZE_ON_SERVANT:
	servant -> _OB_lock();
	break;
    }
#endif
}

void
POA_impl::_OB_postinvoke()
{
    const ObjectId& oid = poaCurrent_ -> _OB_getObjectId();
    const char* op = poaCurrent_ -> _OB_getOp();
    ServantBase* servant = poaCurrent_ -> _OB_getServant();
    ServantLocator::Cookie cookie = poaCurrent_ -> _OB_getCookie();

    //
    // If we have a servantLocator then call postinvoke
    //
    if(!is_nil(servantLocator_))
        servantLocator_ -> postinvoke(oid, this, op, cookie, servant);

    //
    // Clean up the PortableServant::Current
    //
    poaCurrent_ -> _OB_postinvoke();

    //
    // Clean up the OCI:Current
    //
    OCI::Current_impl* ociCurrent = ORB::_OB_instance_nodup() ->
        _OB_OCICurrent();
    ociCurrent -> _OB_postinvoke();

    //
    // Depending on the synchronization policy we have to lock a
    // variety of mutexes.
    //
#ifdef HAVE_JTC
    switch(policies_ -> synchronizationPolicy())
    {
    case NO_SYNCHRONIZATION:
	break;

    case SYNCHRONIZE_ON_POA:
	poaSyncMutex_.unlock();
	break;

    case SYNCHRONIZE_ON_ORB:
	orbSyncMutex_.unlock();
	break;

    case SYNCHRONIZE_ON_SERVANT:
	servant -> _OB_unlock();
	break;
    }
#endif
}

//
// Invoke a request on the POA and object-id derived from the object-key.
//
void
POA_impl::_OB_invoke(Downcall_ptr downcall,
                     ULong requestId,
                     const ObjectKey& key,
                     const char* op,
                     Buffer& buf,
                     bool sw,
                     ULong offIn,
                     ULong offOut,
                     OCI::TransportInfo_ptr info)
{
    ObjectKeyData data;
    if(parseObjectKey(key, data))
    {
        for(;;)
        {
            POA_impl* poaImpl = _OB_locatePOA(data);
            if(poaImpl != 0)
            {
		//
		// Since _OB_locatePOA increments the reference count
		// we need to ensure that the reference count is
		// decremented when we're done.
		//
                POA_var dummy  = poaImpl;
                poaImpl -> setupInvoke(downcall, requestId,
                                       key, data.oid, op, buf, sw,
                                       offIn, offOut, info);
                return;
            }
            break;
        }
    }
    
    if(!is_nil(downcall))
	downcall -> systemException(requestId, buf, OBJECT_NOT_EXIST());
}

//
// Locate a POA from the provided key data. This increments the
// reference count of the returned POA. The caller is expected to
// decrement the reference count.
//
// XXX This is a worthy method of optimization. This is called on
// every method dispatch. Currently it requires a lock for each POA
// searched. I'm sure this could be optimized to a simple hash table
// lookup, or something similar.
//
POA_impl*
POA_impl::_OB_locatePOA(const POA_impl::ObjectKeyData& data)
{
    if(data.poaID.length() <= 0)
        return 0;

    if(strcmp(data.poaID[0], name_) != 0)
        return 0;

    POA_impl* curr = dynamic_cast<POA_impl*>(POA::_duplicate(this));

    for(ULong i = 1; i < data.poaID.length() && curr != 0; ++i)
    {
        POA_var dummy = curr;
#ifdef HAVE_JTC
        //
        // Lock the POA. We are not permitted to call findPOA without
        // first having a write lock on the POA.
        //
        JTCSynchronized sync(curr -> mutex_); // Write lock
#endif
        curr = curr -> findPOA(data.poaID[i], true, true);
    }

    //
    // Ensure we have a match
    //
    if(curr != 0 && !curr -> poaMatches(data))
    {
        release(curr);
        curr = 0;
    }

    return curr;
}

//
// Find an object-id described by the provided key
//
bool
POA_impl::_OB_findByKey(const ObjectKey& key)
{
    ObjectKeyData data;
    if(parseObjectKey(key, data))
    {
        POA_impl* poaImpl = _OB_locatePOA(data);
        if(poaImpl != 0)
        {
            POA_var dummy = poaImpl;
            return poaImpl -> _OB_findServant(data.oid);
        }
    }
    return false;
}

bool
POA_impl::_OB_findServant(const ObjectId& oid)
{
    OB_READ_LOCK
    
    //
    // If RETAIN policy then look in the active object map.
    //
    if(policies_ -> servantRetentionPolicy() == RETAIN &&
       activeObjectMap_ -> contains(oid))
    {
        return true;
    }
    
    //
    // USE_ACTIVE_OBJECT_MAP_ONLY, then we have no servant
    //
    if(policies_ -> requestProcessingPolicy() == USE_ACTIVE_OBJECT_MAP_ONLY)
    {
        return false;
    }
    
    //
    // USE_DEFAULT_SERVANT
    //
    if(policies_ -> requestProcessingPolicy() == USE_DEFAULT_SERVANT)
    {
        return defaultServant_.in() != 0;
    }
    
    assert(policies_ -> requestProcessingPolicy() == USE_SERVANT_MANAGER);
    
    //
    // No servant activator or locator, then no servant
    //
    if(is_nil(servantActivator_) &&
       is_nil(servantLocator_))
    {
        return false;
    }
    
    //
    // This is optimistic, but we don't want to actually
    // activate the servant at this point.
    //
    return true;
}

//
// Locate the POA given the provided object key
//
POA_impl*
POA_impl::_OB_locatePOA(const ObjectKey& key)
{
    POA_impl* poa = 0;

    ObjectKeyData data;
    if(parseObjectKey(key, data))
        poa = _OB_locatePOA(data);
    
    return poa;
}

//
// Create an object reference given an object key and interface name
//
Object_ptr
POA_impl::_OB_createReference(const ObjectKey& key, const char* intf)
{
    //
    // Create IOR.
    //
    IOP::IOR ior;
    ior.type_id = string_dup(intf);
    OCI::AccRegistry_var registry = OARoot::instance() -> getAccRegistry();

    OCI::ObjectKey ociKey(key.length());
    ociKey.length(key.length());
    memcpy(ociKey.get_buffer(), key.get_buffer(), ociKey.length());

    registry -> add_profiles(ociKey, ior);
//    IOPAddCodeSetInformation(&ior, ORB::native_cs(), ORB::native_wcs());
    Object_var obj = ORB::_OB_instance_nodup() -> _OB_createObject(ior);

    //
    // Set the servant on the object, if it exists.
    //
    ObjectKeyData data;
    bool rc = parseObjectKey(key, data);
    assert(rc);

    return obj._retn();

}

StubImplBase*
POA_impl::_OB_createStubImplBase(const ObjectKey& key, const char* intf)
{
    OB_WRITE_LOCK;

    ObjectKeyData data;
    bool rc = parseObjectKey(key, data);
    assert(rc);

    StubImplBase* base = createStubImplBase(key, data, intf);
    if(base != 0)
    {
        StubImplBasePtrSeq* seq;
        if(!localStubImplSeqMap_ -> get(data.oid, seq))
        {
            seq = new StubImplBasePtrSeq;
            localStubImplSeqMap_ -> put(data.oid, seq);
        }
        seq -> append(base);
    }
    return base;
}

//
// Locate a servant, given it's object id and operation name. This
// method increments the reference count of the returned servant. The
// caller is expected to decrement the reference count.
//
// This method assumes that the PortableServer::Current context has
// already been setup.
//
ServantBase*
POA_impl::_OB_locateServant(const ObjectId& oid,
                            const char* op,
                            ServantLocator::Cookie& cookie)
{
    OB_WRITE_LOCK
    
    //
    // If RETAIN policy then look in the active object map.
    //
    if(policies_ -> servantRetentionPolicy() == RETAIN)
    {
        ServantBase_var impl;
        if(activeObjectMap_ -> get(oid, impl))
            return impl._retn();
    }

    //
    // If the POA has the NON_RETAIN policy, or RETAIN but no entry in
    // the active object map
    //

    //
    // USE_ACTIVE_OBJECT_MAP_ONLY, then we have no servant
    //
    if(policies_ -> requestProcessingPolicy() == USE_ACTIVE_OBJECT_MAP_ONLY)
    {
        return 0;
    }

    //
    // USE_DEFAULT_SERVANT
    //
    if(policies_ -> requestProcessingPolicy() == USE_DEFAULT_SERVANT)
    {
        //
        // If no servant has been associated with the POA, the POA
        // raises the OBJ_ADAPTER system exception.
        //
        if(defaultServant_.in() == 0)
            throw OBJ_ADAPTER();

        defaultServant_ -> _add_ref();
        return defaultServant_;
    }

    //
    // USE_SERVANT_MANAGER
    //
    if(policies_ -> requestProcessingPolicy() == USE_SERVANT_MANAGER)
    {
        //
        // If no servant has been associated with the POA, the POA
        // raises the OBJ_ADAPTER system exception.
        //
        if(is_nil(servantActivator_) &&
           is_nil(servantLocator_))
        {
            throw OBJ_ADAPTER();
        }

        ServantBase* servant;

        if(policies_ -> servantRetentionPolicy() == RETAIN)
        {
            servant = servantActivator_ -> incarnate(oid, this);

            if(servant != 0)
            {
                //
                // 11-25:
                //
                // If the incarnate operation returns a servant that
                // is already active for a different Object Id and if
                // the POA also has the UNIQUE_ID policy, the
                // incarnate has violated the POA policy and is
                // considered to be in error. The POA will raise an
                // OBJ_ADAPTER system exception for the request.
                //
                if(policies_ -> idUniquenessPolicy() == UNIQUE_ID &&
                   servantIdMap_ -> contains(servant))
                {
                    throw OBJ_ADAPTER();
                }
                
                //
                // Enter the servant into the active object map
                //
                addServantActiveObjectMap(oid, servant);
            }
        }
        else
        {
            //
            // NO_RETAIN
            //
            servant = servantLocator_ -> preinvoke(oid, this, op, cookie);
        }

        //
        // 11-23:
        //
        // If a ServantManager returns a null Servant (or the
        // equivalent in a language mapping) as the result of an
        // incarnate() or preinvoke() operation, the POA will
        // return the OBJ_ADAPTER system exception as the result
        // of the request.
        //
        if(servant == 0)
            throw OBJ_ADAPTER();

        servant -> _add_ref();

        return servant;
    }

    return 0;
}

void
POA_impl::_OB_stateChange()
{
    if(manager_ -> get_state() == POAManager::ACTIVE &&
       records_.length() > 0)
    {
        assert(ORB::server_conc_model() == ORB::ServerConcModelReactive);
        //
        // Dispatch all outstanding requests
        //
        for(ULong i = 0; i < records_.length(); ++i)
        {
            InvocationRecord* rec = records_[i];
            invoke(rec -> downcall,
                   rec -> requestId,
                   rec -> key,
                   rec -> oid,
                   rec -> op,
                   rec -> buf,
                   rec -> sw,
                   rec -> offIn,
                   rec -> offOut,
                   rec -> info);
            delete rec;
        }
        records_.length(0);
    }
}

Current_impl*   
POA_impl::_OB_POACurrent()
{
    return poaCurrent_;               
}

void
POA_impl::_OB_removeStub(const ObjectId& oid, StubImplBase* base)
{
    OB_WRITE_LOCK

    assert(localStubImplSeqMap_ -> contains(oid));
    StubImplBasePtrSeq* seq;
    if(localStubImplSeqMap_ -> get(oid, seq))
	for(ULong i = 0; i < seq -> length(); ++i)
	    if((*seq)[i] == base)
	    {
		seq -> remove(i);
		return;
	    }
    assert(false);
}
