// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Reactor.h>
#include <OB/Timer.h>
#include <OB/Net.h>

#include <SelectReactor.h>

using namespace CORBA;
using namespace OB;

Reactor* Reactor::instance_ = 0;

namespace OB
{

class ReactorDestroyer
{
public:

    ~ReactorDestroyer()
    {
	delete Reactor::instance_;
	Reactor::instance_ = 0;
    }
};

} // End of namespace OB

static ReactorDestroyer destroyer;

// ----------------------------------------------------------------------
// Reactor public member implementation
// ----------------------------------------------------------------------

Reactor*
Reactor::instance()
{
    if(!instance_)
	return SelectReactor::instance(); // SelectReactor is default

    return instance_;
}

// ----------------------------------------------------------------------
// SelectReactor protected member implementation
// ----------------------------------------------------------------------

void
SelectReactor::reapHandlerInfoList()
{
    if(handlerInfoListUsageCount_ > 0)
	return;

    HandlerInfo** p = &handlerInfoList_;
    while(*p)
    {
	HandlerInfo* info = *p;

	if(info -> handler == 0)
	{
	    *p = info -> next;
	    delete info;
	}
	else
	{
	    p = &(info -> next);
	}
    }
}

int
SelectReactor::setFdSets(fd_set& readFdSet, fd_set& writeFdSet) const
{
    FD_ZERO(&readFdSet);
    FD_ZERO(&writeFdSet);
    int nfds = 0;

    HandlerInfo* info = handlerInfoList_;
    while(info)
    {
	if(info -> handler)
	{
	    Mask mask = info -> mask;
	    Handle handle = info -> handle;
	    
	    if(mask && handle >= nfds)
		nfds = handle + 1;
	    
	    if(mask & EventRead)
		FD_SET(handle, &readFdSet);
	    
	    if(mask & EventWrite)
		FD_SET(handle, &writeFdSet);
	}

	info = info -> next;
    }

    return nfds;
}

void
SelectReactor::evalFdSets(fd_set& readFdSet, fd_set& writeFdSet)
{
    handlerInfoListUsageCount_++;
    HandlerInfo* info = handlerInfoList_;
    
    while(info)
    {
	if(info -> handler)
	{
	    Mask mask = 0;
	    Handle handle = info -> handle;
	    
	    if(FD_ISSET(handle, &readFdSet))
		mask |= EventRead;
	    
	    if(FD_ISSET(handle, &writeFdSet))
		mask |= EventWrite;
	    
	    if(mask)
	    {
		try
		{
		    info -> handler -> handleEvent(mask);
		}
		catch(...)
		{
		    handlerInfoListUsageCount_--;
		    reapHandlerInfoList();
		    throw;
		}
	    }
	}
	
	info = info -> next;
    }

    handlerInfoListUsageCount_--;
    reapHandlerInfoList();
}

// ----------------------------------------------------------------------
// SelectReactor constructor and destructor
// ----------------------------------------------------------------------

SelectReactor::SelectReactor() 
    : stop_(false), handlerInfoList_(0), handlerInfoListUsageCount_(0)
{
}

SelectReactor::~SelectReactor()
{
    //
    // Delete all handler infos
    //
    assert(handlerInfoListUsageCount_ == 0);
    while(handlerInfoList_)
    {
	HandlerInfo* next = handlerInfoList_ -> next;
	delete handlerInfoList_;
	handlerInfoList_ = next;
    }
}

// ----------------------------------------------------------------------
// SelectReactor public member implementation
// ----------------------------------------------------------------------

Reactor*
SelectReactor::instance()
{
    if(!instance_)
	instance_ = new SelectReactor;

    return instance_;
}

void
SelectReactor::registerHandler(EventHandler* handler, Mask mask, Handle handle)
{
    assert(handler);

    HandlerInfo* info = handlerInfoList_;
    while(info)
    {
	if(info -> handler == handler)
	{
	    //
	    // Handler already in list, update mask and handle
	    //
	    info -> mask = mask;
	    info -> handle = handle;
	    return;
	}
	
	info = info -> next;
    }

    //
    // Add new handler info
    //
    info = new HandlerInfo;
    info -> handler = handler;
    info -> mask = mask;
    info -> handle = handle;
    info -> next = handlerInfoList_;
    handlerInfoList_ = info;
}

void
SelectReactor::unregisterHandler(EventHandler* handler)
{
    assert(handler);

    HandlerInfo* info = handlerInfoList_;
    while(info)
    {
	if(info -> handler == handler)
	{
	    //
	    // Handler info found, invalidate the handler
	    //
	    info -> handler = 0;
	    reapHandlerInfoList();
	    return;
	}
	
	info = info -> next;
    }

    assert(false);
}

void
SelectReactor::dispatch()
{
    stop_ = false;

    TimerList* timerList = TimerList::instance();

    while(!stop_ && handlerInfoList_)
    {
	//
	// Get the current time
	//
	timeval now = TimerList::timeNow();

	// 
	// Check for expired timers
	//
	if(!timerList -> isEmpty())
	    timerList -> expire(now);
	
	//
	// The filedescriptor sets and the nfds parameter for select()
	//
	fd_set readFdSet;
	fd_set writeFdSet;
	FD_ZERO(&readFdSet);
	FD_ZERO(&writeFdSet);
	int nfds = setFdSets(readFdSet, writeFdSet);

	//
	// Return value for select()
	//
	int result;
	    
	//
	// Is there a timer left?
	//
	if(!timerList -> isEmpty())
	{
	    //
	    // Calculate timer list timeout value
	    //
	    timeval tv =
		TimerList::timeSub(timerList -> first() -> timeout(), now);

	    //
	    // Do select
	    //
	    result = select(nfds, &readFdSet, &writeFdSet, 0, &tv);
	    if(result == 0)
		continue; // Timer list timeout
	}
	else
	{
	    //
	    // Do select
	    //
	    result = select(nfds, &readFdSet, &writeFdSet, 0, 0);
	    assert(result != 0);
	}

	//
	// Select error?
	//
	if(result == -1)
	{
            if(IsInterrupted())
                continue;

            throw COMM_FAILURE(LastError(), MinorSelect, COMPLETED_NO);
	}
	
	//
	// Evaluate file descriptor sets
	//
	evalFdSets(readFdSet, writeFdSet);
    }
}

bool
SelectReactor::dispatchOneEvent(Long timeout)
{
    TimerList* timerList = TimerList::instance();

    while(!stop_ && handlerInfoList_)
    {
	//
	// Get the current time
	//
	timeval now = TimerList::timeNow();

	// 
	// Check for expired timers
	//
	if(!timerList -> isEmpty())
	    if(timerList -> expireOneTimer(now))
		return true; // A timer event is treated just like a
                             // regular event
	
	//
	// The filedescriptor sets and the nfds parameter for select()
	//
	fd_set readFdSet;
	fd_set writeFdSet;
	FD_ZERO(&readFdSet);
	FD_ZERO(&writeFdSet);
	int nfds = setFdSets(readFdSet, writeFdSet);
	    
	//
	// Return value for select()
	//
	int result;

	//
	// Is there a timer left?
	//
	if(!timerList -> isEmpty())
	{
	    //
	    // Calculate timer list timeout value
	    //
	    timeval tv =
		TimerList::timeSub(timerList -> first() -> timeout(), now);

	    if(timeout >= 0)
	    {
		//
		// Calculate regular timeout value
		//
		struct timeval rtv;
		rtv.tv_sec = timeout / 1000;
		rtv.tv_usec = (timeout - rtv.tv_sec * 1000) * 1000;

		if(TimerList::timeCmp(rtv, tv) < 0)
		{
		    //
		    // Do select
		    //
		    result = select(nfds, &readFdSet, &writeFdSet, 0, &rtv);
		    if(result == 0)
			return false; // Regular timeout
		}
		else
		{
		    //
		    // Do select
		    //
		    result = select(nfds, &readFdSet, &writeFdSet, 0, &tv);
		    if(result == 0)
			continue; // Timer list timeout
		}
	    }
	    else
	    {
		//
		// Do select
		//
		result = select(nfds, &readFdSet, &writeFdSet, 0, &tv);
		if(result == 0)
		    continue; // Timer list timeout
	    }
	}
	else
	{
	    if(timeout >= 0)
	    {
		//
		// Calculate regular timeout value
		//
		struct timeval rtv;
		rtv.tv_sec = timeout / 1000;
		rtv.tv_usec = (timeout - rtv.tv_sec * 1000) * 1000;

		//
		// Do select
		//
		result = select(nfds, &readFdSet, &writeFdSet, 0, &rtv);
		if(result == 0)
		    return false; // Regular timeout
	    }
	    else
	    {
		//
		// Do select
		//
		result = select(nfds, &readFdSet, &writeFdSet, 0, 0);
		assert(result != 0);
	    }
	}
	
	//
	// Select error?
	//
	if(result == -1)
	{
            if(IsInterrupted())
                continue;

            throw COMM_FAILURE(LastError(), MinorSelect, COMPLETED_NO);
	}

	//
	// Evaluate file descriptor sets
	//
	evalFdSets(readFdSet, writeFdSet);
	return true;
    }

    return true;
}

bool
SelectReactor::eventReady()
{
    TimerList* timerList = TimerList::instance();

    //
    // Get the current time
    //
    timeval now = TimerList::timeNow();

    // 
    // Check for expired timers
    //
    if(!timerList -> isEmpty())
        return true;

    //
    // The filedescriptor sets and the nfds parameter for select()
    //
    fd_set readFdSet;
    fd_set writeFdSet;
    FD_ZERO(&readFdSet);
    FD_ZERO(&writeFdSet);
    int nfds = setFdSets(readFdSet, writeFdSet);
	    
    //
    // Return value for select()
    //
    int result;

    //
    // Calculate regular timeout value
    //
    struct timeval tv;
    tv.tv_sec = 0;
    tv.tv_usec = 0;

    while(true)
    {
        //
        // Do select
        //
        result = select(nfds, &readFdSet, &writeFdSet, 0, &tv);
        if(result == 0)
            return false; // Regular timeout
            
        //
        // Select error?
        //
        if(result == -1)
        {
            if(IsInterrupted())
                continue;
                
            throw COMM_FAILURE(LastError(), MinorSelect, COMPLETED_NO);
        }

        //
        // An event is ready
        //
        return true;
    }
}
