// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Fixed.h>
#include <OB/Hashtable.h>
#include <OB/HashtableI.h>
#include <OB/Template.h>
#include <OB/TemplateI.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/Hashers.h>
#include <OB/Any.h>
#include <OB/OBV.h>
#include <OB/TCKind.h>
#include <OB/TypeCode.h>
#include <OB/Principal.h>
#include <OB/Environment.h>
#include <OB/Context.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/Policy.h>
#include <OB/POAServantBase.h>
#include <OB/IFR1.h>
#include <OB/IFR4.h>
#include <OB/ORB.h>
#include <OB/ValueBase.h>
#include <OB/ValueFactoryBase.h>
#include <OB/CONV_FRAME.h>
#include <OB/CodeSetDatabase.h>
#include <OB/CodeConverter.h>
#include <OB/AbstractBase.h>
#include <OB/Stream.h>
#include <stdio.h>
#include <stdlib.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Required functions for hashtable
// ----------------------------------------------------------------------

ULong
OBHash(Octet* v)
{
    return (ULong)v;
}

bool
OBHashCompare(Octet* v1, Octet* v2)
{
    return v1 == v2;
}

ULong
OBHash(TypeCode_ptr v)
{
    return (ULong)v;
}

bool
OBHashCompare(TypeCode_ptr v1, TypeCode_ptr v2)
{
    // TODO: This should really be v1 -> equal(v2). But what do I use
    // as hash then?
    return v1 == v2;
}

ULong
OBHash(ValueBase* value)
{
    return (ULong)value;
}

bool
OBHashCompare(ValueBase* v1, ValueBase* v2)
{
    return (v1 == v2);
}

ULong
OBHash(const char** value)
{
    ULong result = 0;

    ULong i = 0;
    while(value[i] != 0)
    {
        result += OBHash(value[i]);
        i++;
    }

    return result;
}

bool
OBHashCompare(const char** v1, const char** v2)
{
    bool result = true;

    ULong i = 0;
    while(result && v1[i] != 0 && v2[i] != 0)
    {
        if(strcmp(v1[i], v2[i]) != 0)
            result = false;
        i++;
    }

    //
    // Ensure both arrays are the same length
    //
    if(result && (v1[i] != 0 || v2[i] != 0))
        result = false;

    return result;
}

ULong
OBHash(const Octet* o)
{
    return (ULong)o;
}

bool
OBHashCompare(const Octet* o1, const Octet* o2)
{
    return (o1 == o2);
}

// ----------------------------------------------------------------------
// Template instantiations
// ----------------------------------------------------------------------

#ifndef HAVE_NO_EXPLICIT_TEMPLATES
template class OB::Hashtable< const Octet*, TypeCode_ptr >;
template class OB::Hashtable< TypeCode_ptr, const Octet* >;
template class OB::Hashtable< ValueBase*, Long >;
template class OB::Hashtable< const char*, Long >;
template class OB::Hashtable< const char**, Long >;
template class OB::Hashtable< const Octet*, ValueBase* >;
template class OB::Hashtable< const Octet*, OB::InputStream::ValueInfo >;
#else
#ifdef HAVE_PRAGMA_DEFINE
#pragma define(OB::Hashtable< const Octet*, TypeCode_ptr >)
#pragma define(OB::Hashtable< TypeCode_ptr, const Octet* >)
#pragma define(OB::Hashtable< ValueBase*, Long >)
#pragma define(OB::Hashtable< const char*, Long >)
#pragma define(OB::Hashtable< const char**, Long >)
#pragma define(OB::Hashtable< const Octet*, ValueBase* >)
#pragma define(OB::Hashtable< const Octet*, OB::InputStream::ValueInfo >)
#endif
#endif

// ----------------------------------------------------------------------
// Private functions for InputStream, OutputStream and CountStream
// ----------------------------------------------------------------------

Long
InputStream::readLongImpl()
{
    //
    // The chunking code needs to read a long value without entering
    // an infinite loop (i.e., checkChunk calling read_long, which
    // calls checkChunk, etc.)
    //

    Long val;

    ULong offs = 3 - (((unsigned long)pos_ + 3) % 4);

#ifdef OB_MARSHAL_CHECK
    if((unsigned long)(pos_ - start_) + offs + 4 > size_)
        throw MARSHAL(MinorReadLongOverflow, COMPLETED_NO);
#endif

    pos_ += offs;
    val = *(const Long*)pos_;
    pos_ += 4;

    if(swap_)
    {
        ULong& v = (ULong&)val;
        v = ((v << 24) | ((v & 0xff00) << 8)
             | ((v >> 8) & 0xff00) | (v >> 24));
    }

    return val;
}

void
InputStream::readValueInfo(ValueInfo& info)
{
    //
    // Special cases are handled by read_Value
    //
    assert(info.tag != 0 && info.tag != (Long)0xffffffff);

    //
    // Check if the value is chunked
    //
    if((info.tag & 0x00000008) == 8)
        info.state.chunked = true;

    //
    // Check for presence of codebase URL
    //
    if(info.tag & 0x00000001 == 1)
    {
        String_var codebase;

        //
        // Check for indirection tag
        //
        const Octet* save = pos_;
        Long indTag = read_long();
        if(indTag == (Long)0xffffffff)
        {
            Long offs = read_long();
            assert(offs < 0);
            const Octet* tmp = pos_;
            pos_ = pos_ - 4 + offs;
            assert(pos_ >= start_);
            codebase = read_string();
            pos_ = tmp;
        }
        else
        {
            pos_ = save;
            codebase = read_string();
        }
    }

    //
    // Extract repository ID information
    //
    if((info.tag & 0x00000006) == 0)
    {
        //
        // No type information was marshalled
        //
    }
    else if((info.tag & 0x00000006) == 6)
    {
        //
        // Extract a list of repository IDs, representing the
        // truncatable types for this value
        //

        //
        // Check for indirection tag
        //
        const Octet* save = pos_;
        Long indTag = read_long();
        if(indTag == (Long)0xffffffff)
        {
            Long offs = read_long();
            assert(offs < 0);
            save = pos_;
            pos_ = pos_ - 4 + offs;
            assert(pos_ >= start_);
        }
        else
            pos_ = save;

        Long count = read_long();
        info.ids.length(count);
        Long i;
        for(i = 0 ; i < count ; i++)
            info.ids[i] = read_string();

        if(indTag == (Long)0xffffffff)
            pos_ = save;
    }
    else if((info.tag & 0x00000006) == 2)
    {
        //
        // Extract a single repository ID
        //
        String_var id;

        //
        // Check for indirection tag
        //
        const Octet* save = pos_;
        Long indTag = read_long();
        if(indTag == (Long)0xffffffff)
        {
            Long offs = read_long();
            assert(offs < 0);
            save = pos_;
            pos_ = pos_ - 4 + offs;
            assert(pos_ >= start_);
            id = read_string();
            pos_ = save;
        }
        else
        {
            pos_ = save;
            id = read_string();
        }

        info.ids.length(1);
        info.ids[0] = id;
    }

    //
    // Record beginning of value data
    //
    info.dataPos = pos_;
}

ValueBase*
InputStream::createValue(const char* id, const ValueInfo& info)
{
    assert(info.tag >= 0x7fffff00 && info.tag != (Long)0xffffffff);

    //
    // Verify that we have sufficient type information
    //
    if(id == 0 && info.ids.length() == 0)
        throw MARSHAL(1, COMPLETED_NO);

    //
    // Try to find a factory to create the value
    //
    ORB_ptr orb = ORB::_OB_instance_nodup();
    ValueFactory factory = 0;

    if(info.ids.length() > 0)
    {
        //
        // Stop when we find the first factory
        //
        for(ULong i = 0 ; i < info.ids.length() ; i++)
        {
            factory = orb -> lookup_value_factory(info.ids[i]);
            if(factory != 0)
                break;
        }
    }
    else
    {
        assert(id != 0);
        factory = orb -> lookup_value_factory(id);
    }

    //
    // If we didn't find a factory, raise MARSHAL
    //
    if(factory == 0)
        throw MARSHAL(1, COMPLETED_NO);

    ValueFactoryBase_var factoryVar = factory;

    //
    // Instantiate value
    //
    ValueBase_var vb = factory -> create_for_unmarshal();
    assert(vb.in() != 0); // XXX raise exception instead?

    valueHistory_ -> put(info.headerPos, vb.in());

    return vb._retn();
}

void
InputStream::readChunk(ValueState& state)
{
    //
    // Check for a chunk size
    //
    Long size = readLongImpl();
    if(size >= 0 && size < 0x7fffff00) // chunk size
    {
        state.chunkStart = pos_;
        state.chunkSize = size;
    }
    else if(size < 0) // end tag
    {
        pos_ -= 4; // rewind
        state.chunkStart = pos_;
        state.chunkSize = 0;
    }
    else
    {
        pos_ -= 4; // rewind
        state.chunkStart = 0;
        state.chunkSize = 0;
    }
}

void
InputStream::checkChunk()
{
    assert(chunked());

    //
    // If we've reached the end of the current chunk, then check
    // for the start of a new chunk
    //
    if(valueState_.chunkStart &&
       valueState_.chunkStart + valueState_.chunkSize == pos_)
        readChunk(valueState_);
}

void
InputStream::_OB_readEndian()
{
    swap_ = (read_boolean() != Endian);
}

// ----------------------------------------------------------------------
// InputStream constructors
// ----------------------------------------------------------------------

InputStream::InputStream(CodeConverters* codeConverters,
			 const Buffer& buff, ULong offs, bool swap)
    : swap_(swap), readTCHistory_(0),
      valueHistory_(0), valueInfoHistory_(0)
{
    codeConverters_.charConverter = codeConverters -> charConverter;
    codeConverters_.wcharConverter = codeConverters -> wcharConverter;
    start_ = buff.data + offs;
    pos_ = start_;
    size_ = buff.len;
}

InputStream::InputStream(const Buffer& buff, ULong offs, bool swap)
    : swap_(swap), readTCHistory_(0), valueHistory_(0), valueInfoHistory_(0)
{
    codeConverters_.charConverter = LocalCodeConverter::instance();
    codeConverters_.wcharConverter = LocalCodeConverter::instance();
    start_ = buff.data + offs;
    pos_ = start_;
    size_ = buff.len;
}

InputStream::InputStream(CodeConverters* codeConverters,
			 const Octet* buff, ULong offs, ULong len, bool swap)
    : size_(len), swap_(swap),
      readTCHistory_(0), valueHistory_(0), valueInfoHistory_(0)
{
    codeConverters_.charConverter = codeConverters -> charConverter;
    codeConverters_.wcharConverter = codeConverters -> wcharConverter;
    start_ = buff + offs;
    pos_ = start_;
}

InputStream::InputStream(const Octet* buff, ULong offs, ULong len, bool swap)
    : size_(len), swap_(swap), readTCHistory_(0), valueHistory_(0),
      valueInfoHistory_(0)
{
    codeConverters_.charConverter = LocalCodeConverter::instance();
    codeConverters_.wcharConverter = LocalCodeConverter::instance();
    start_ = buff + offs;
    pos_ = start_;
}

InputStream::~InputStream()
{
    delete readTCHistory_;
    delete valueHistory_;
    delete valueInfoHistory_;
}

// ----------------------------------------------------------------------
// OutputStream constructors
// ----------------------------------------------------------------------

OutputStream::OutputStream(CodeConverters* codeConverters, ULong size,
			   ULong offs)
    : chunked_(false), nestingLevel_(0),
      writeTCHistory_(0), valueHistory_(0), idHistory_(0), idListHistory_(0)
{
    codeConverters_.charConverter = codeConverters -> charConverter;
    codeConverters_.wcharConverter = codeConverters -> wcharConverter;
    buf_ = new Buffer(size);
    start_ = buf_ -> data + offs;
    pos_ = start_;
    size_ = size;
    origSize_ = size;
}

OutputStream::OutputStream(ULong size, ULong offs)
    : chunked_(false), nestingLevel_(0), writeTCHistory_(0), valueHistory_(0),
      idHistory_(0), idListHistory_(0)
{
    codeConverters_.charConverter = LocalCodeConverter::instance();
    codeConverters_.wcharConverter = LocalCodeConverter::instance();
    buf_ = new Buffer(size);
    start_ = buf_ -> data + offs;
    pos_ = start_;
    size_ = size;
    origSize_ = size;
}

OutputStream::OutputStream(CodeConverters* codeConverters, Buffer& buff,
			   ULong offs)
    : buf_(0), chunked_(false),
      nestingLevel_(0), writeTCHistory_(0), valueHistory_(0), idHistory_(0),
      idListHistory_(0)
{
    codeConverters_.charConverter = codeConverters -> charConverter;
    codeConverters_.wcharConverter = codeConverters -> wcharConverter;
    start_ = buff.data + offs;
    pos_ = start_;
    size_ = buff.len;
    origSize_ = buff.len;
}

OutputStream::OutputStream(Buffer& buff, ULong offs)
    : buf_(0), chunked_(false), nestingLevel_(0), writeTCHistory_(0),
      valueHistory_(0), idHistory_(0), idListHistory_(0)
{
    codeConverters_.charConverter = LocalCodeConverter::instance();
    codeConverters_.wcharConverter = LocalCodeConverter::instance();
    start_ = buff.data + offs;
    pos_ = start_;
    size_ = buff.len;
    origSize_ = buff.len;
}

OutputStream::OutputStream(CodeConverters* codeConverters, Octet* buff,
			   ULong offs, ULong size)
    : size_(size), origSize_(size), buf_(0),
      chunked_(false), nestingLevel_(0), writeTCHistory_(0), valueHistory_(0),
      idHistory_(0), idListHistory_(0)
{
    codeConverters_.charConverter = codeConverters -> charConverter;
    codeConverters_.wcharConverter = codeConverters -> wcharConverter;
    start_ = buff + offs;
    pos_ = start_;
}

OutputStream::OutputStream(Octet* buff, ULong offs, ULong size)
    : size_(size), origSize_(size), buf_(0), chunked_(false), nestingLevel_(0),
      writeTCHistory_(0), valueHistory_(0), idHistory_(0), idListHistory_(0)
{
    codeConverters_.charConverter = LocalCodeConverter::instance();
    codeConverters_.wcharConverter = LocalCodeConverter::instance();
    start_ = buff + offs;
    pos_ = start_;
}

OutputStream::~OutputStream()
{
    delete buf_;
    delete writeTCHistory_;
    delete valueHistory_;
    delete idHistory_;
    delete idListHistory_;
}

// ----------------------------------------------------------------------
// CountStream constructors
// ----------------------------------------------------------------------

CountStream::CountStream(CodeConverters* codeConverters)
    : count_(0), chunked_(false),
      nestingLevel_(0), countTCHistory_(0), valueHistory_(0), idHistory_(0),
      idListHistory_(0)
{
    codeConverters_.charConverter = codeConverters -> charConverter;
    codeConverters_.wcharConverter = codeConverters -> wcharConverter;
}

CountStream::CountStream()
    : count_(0), chunked_(false), nestingLevel_(0), countTCHistory_(0),
      valueHistory_(0), idHistory_(0), idListHistory_(0)
{
    codeConverters_.charConverter = LocalCodeConverter::instance();
    codeConverters_.wcharConverter = LocalCodeConverter::instance();
}

CountStream::CountStream(CodeConverters* codeConverters, ULong start)
    : count_(start), chunked_(false),
      nestingLevel_(0), countTCHistory_(0), valueHistory_(0), idHistory_(0),
      idListHistory_(0)
{
    codeConverters_.charConverter = codeConverters -> charConverter;
    codeConverters_.wcharConverter = codeConverters -> wcharConverter;
}

CountStream::CountStream(ULong start)
    : count_(start), chunked_(false), nestingLevel_(0), countTCHistory_(0),
      valueHistory_(0), idHistory_(0), idListHistory_(0)
{
    codeConverters_.charConverter = LocalCodeConverter::instance();
    codeConverters_.wcharConverter = LocalCodeConverter::instance();
}

CountStream::~CountStream()
{
    delete countTCHistory_;
    delete valueHistory_;
    delete idHistory_;
    delete idListHistory_;
}

// ----------------------------------------------------------------------
// Public functions for InputStream, OutputStream and CountStream
// ----------------------------------------------------------------------

Buffer*
OutputStream::_OB_buffer(Boolean rel)
{
    Buffer* result = buf_;
    if(rel)
        buf_ = 0;
    return result;
}

void
OutputStream::_OB_addCapacity(ULong size, ULong align)
{
    assert(buf_ != 0);

    //
    // Increase size to allow for alignment
    //
    if(align > 0)
        size += (align - 1) - (((unsigned long)pos_ + (align - 1)) % align);

    //
    // Determine how much "extra" space we currently have left.  This
    // is over and above the originally-requested buffer size.
    //
    // NOTE: We can NEVER use any of the originally-requested buffer size
    // when adding capacity.
    //
    ULong count = (unsigned long)(pos_ - start_);
    ULong remaining = size_ - origSize_;
    if(count > origSize_)
        remaining -= (count - origSize_);

    //
    // If necessary, expand buffer to hold requested size
    //
    if(size > remaining)
    {
        ULong min = size - remaining; // the minimum necessary
        ULong adj = min > 2 * size_ ? min : 2 * size_;
        buf_ -> realloc(size_ + adj);
        start_ = buf_ -> data;
        pos_ = start_ + count;
        size_ = buf_ -> len;
    }
}

Char
InputStream::read_char()
{
    if(chunked())
        checkChunk();

#ifdef OB_MARSHAL_CHECK
    if((((unsigned long)(pos_ - start_))) + 1 > size_)
	throw MARSHAL(MinorReadCharOverflow, COMPLETED_NO);
#endif

    return codeConverters_.charConverter -> read_char(pos_);
}

WChar
InputStream::read_wchar()
{
    if(chunked())
        checkChunk();

#ifdef OB_MARSHAL_CHECK
    ULong len = *pos_;

    if((unsigned long)(pos_ - start_) + len + 1 > size_)
        throw MARSHAL(MinorReadWCharOverflow, COMPLETED_NO);
#endif

    if(!codeConverters_.wcharConverter)
	throw BAD_PARAM();

    return codeConverters_.wcharConverter -> read_wchar(pos_);
}

void
OutputStream::write_char(Char value)
{
#ifdef OB_MARSHAL_CHECK
    if((((unsigned long)(pos_ - start_))) + 1 > size_)
	throw MARSHAL(MinorWriteCharOverflow, COMPLETED_NO);
#endif

    codeConverters_.charConverter -> write_char(pos_, value);
}

void
OutputStream::write_wchar(WChar value)
{
#ifdef OB_MARSHAL_CHECK
    CountStream count(&codeConverters_);
    count.count_wchar(value);
    ULong len = count.count();

    if((unsigned long)(pos_ - start_) + len + 1 > size_)
        throw MARSHAL(MinorWriteWCharOverflow, COMPLETED_NO);
#endif

    if(!codeConverters_.wcharConverter)
	throw BAD_PARAM();

    codeConverters_.wcharConverter -> write_wchar(pos_, value);
}

void
CountStream::count_wchar(WChar value)
{
    if(!codeConverters_.wcharConverter)
	throw BAD_PARAM();

    count_ += codeConverters_.wcharConverter -> count_wchar(value);
}

Short
InputStream::read_short()
{
    Short val;

    if(chunked())
        checkChunk();

    ULong offs = (unsigned long)pos_ % 2;

#ifdef OB_MARSHAL_CHECK
    if((unsigned long)(pos_ - start_) + offs + 2 > size_)
        throw MARSHAL(MinorReadShortOverflow, COMPLETED_NO);
#endif

    pos_ += offs;
    val = *(const Short*)pos_;
    pos_ += 2;

    if(swap_)
    {
        UShort& v = (UShort&)val;
        v = (v << 8) | (v >> 8);
    }

    return val;
}

UShort
InputStream::read_ushort()
{
    UShort val;

    if(chunked())
        checkChunk();

    ULong offs = (unsigned long)pos_ % 2;

#ifdef OB_MARSHAL_CHECK
    if((unsigned long)(pos_ - start_) + offs + 2 > size_)
        throw MARSHAL(MinorReadUShortOverflow, COMPLETED_NO);
#endif

    pos_ += offs;
    val = *(const UShort*)pos_;
    pos_ += 2;

    if(swap_)
        val = (val << 8) | (val >> 8);

    return val;
}

Long
InputStream::read_long()
{
    if(chunked())
        checkChunk();

    return readLongImpl();
}

ULong
InputStream::read_ulong()
{
    ULong val;

    if(chunked())
        checkChunk();

    ULong offs = 3 - (((unsigned long)pos_ + 3) % 4);

#ifdef OB_MARSHAL_CHECK
    if((unsigned long)(pos_ - start_) + offs + 4 > size_)
        throw MARSHAL(MinorReadULongOverflow, COMPLETED_NO);
#endif

    pos_ += offs;
    val = *(const ULong*)pos_;
    pos_ += 4;

    if(swap_)
        val = ((val << 24) | ((val & 0xff00) << 8)
               | ((val >> 8) & 0xff00) | (val >> 24));

    return val;
}

LongLong
InputStream::read_longlong()
{
    LongLong val;

    if(chunked())
        checkChunk();

    ULong offs = 7 - (((unsigned long)pos_ + 7) % 8);

#ifdef OB_MARSHAL_CHECK
    if((unsigned long)(pos_ - start_) + offs + 8 > size_)
        throw MARSHAL(MinorReadLongLongOverflow, COMPLETED_NO);
#endif

    pos_ += offs;
    val = *(const LongLong*)pos_;
    pos_ += 8;

    if(swap_)
    {
	ULongLong& v = (ULongLong&)val;
	v = ((v << 56) | ((v & 0xff00) << 40)
	     | ((v & 0xff0000) << 24) | ((v & 0xff000000) << 8)
	     | ((v >> 8) & 0xff000000) | ((v >> 24) & 0xff0000)
	     | ((v >> 40) & 0xff00) | (v >> 56));
    }

    return val;
}

ULongLong
InputStream::read_ulonglong()
{
    ULongLong val;

    if(chunked())
        checkChunk();

    ULong offs = 7 - (((unsigned long)pos_ + 7) % 8);

#ifdef OB_MARSHAL_CHECK
    if((unsigned long)(pos_ - start_) + offs + 8 > size_)
        throw MARSHAL(MinorReadULongLongOverflow, COMPLETED_NO);
#endif

    pos_ += offs;
    val = *(const ULongLong*)pos_;
    pos_ += 8;

    if(swap_)
        val = ((val << 56) | ((val & 0xff00) << 40)
               | ((val & 0xff0000) << 24) | ((val & 0xff000000) << 8)
               | ((val >> 8) & 0xff000000) | ((val >> 24) & 0xff0000)
               | ((val >> 40) & 0xff00) | (val >> 56));

    return val;
}

Float
InputStream::read_float()
{
    Float val;

    if(chunked())
        checkChunk();

    ULong offs = 3 - (((unsigned long)pos_ + 3) % 4);

#ifdef OB_MARSHAL_CHECK
    if((unsigned long)(pos_ - start_) + offs + 4 > size_)
        throw MARSHAL(MinorReadFloatOverflow, COMPLETED_NO);
#endif

    pos_ += offs;
    val = *(const Float*)pos_;
    pos_ += 4;

    if(swap_)
    {
        ULong* v = (ULong*)&val;
        *v = ((*v << 24) | ((*v & 0xff00) << 8)
              | ((*v >> 8) & 0xff00) | (*v >> 24));
    }

    return val;
}

Double
InputStream::read_double()
{
    Double val;

    if(chunked())
        checkChunk();

    ULong offs = 7 - (((unsigned long)pos_ + 7) % 8);

#ifdef OB_MARSHAL_CHECK
    if((unsigned long)(pos_ - start_) + offs + 8 > size_)
        throw MARSHAL(MinorReadDoubleOverflow, COMPLETED_NO);
#endif

    pos_ += offs;
    val = *(const Double*)pos_;
    pos_ += 8;

    if(swap_)
    {
        ULong v0 = ((ULong*)&val)[0];
        ULong v1 = ((ULong*)&val)[1];
        v0 = ((v0 << 24) | ((v0 & 0xff00) << 8)
              | ((v0 >> 8) & 0xff00) | (v0 >> 24));
        v1 = ((v1 << 24) | ((v1 & 0xff00) << 8)
              | ((v1 >> 8) & 0xff00) | (v1 >> 24));
        ((ULong*)&val)[0] = v1;
        ((ULong*)&val)[1] = v0;
    }

    return val;
}

#if SIZEOF_LONG_DOUBLE == 0

LongDouble
InputStream::read_longdouble()
{
    throw MARSHAL(MinorLongDoubleNotSupported, COMPLETED_NO);
    return LongDouble(); // Some compilers need this
}

void
OutputStream::write_longdouble(LongDouble value)
{
    throw MARSHAL(MinorLongDoubleNotSupported, COMPLETED_NO);
}

#else

LongDouble
InputStream::read_longdouble()
{
    LongDouble val;

    if(chunked())
        checkChunk();

    ULong offs = 7 - (((unsigned long)pos_ + 7) % 8);

#ifdef OB_MARSHAL_CHECK
    if((unsigned long)(pos_ - start_) + offs + 16 > size_)
        throw MARSHAL(MinorReadLongDoubleOverflow, COMPLETED_NO);
#endif

    pos_ += offs;

#if SIZEOF_LONG_DOUBLE == 16

    val = *(const LongDouble*)pos_;
    pos_ += 16;

    if(swap_)
    {
        ULongLong v0 = ((ULongLong*)&val)[0];
        ULongLong v1 = ((ULongLong*)&val)[1];
        v0 = ((v0 << 56) | ((v0 & 0xff00) << 40)
              | ((v0 & 0xff0000) << 24) | ((v0 & 0xff000000) << 8)
              | ((v0 >> 8) & 0xff000000) | ((v0 >> 24) & 0xff0000)
              | ((v0 >> 40) & 0xff00) | (v0 >> 56));
        v1 = ((v1 << 56) | ((v1 & 0xff00) << 40)
              | ((v1 & 0xff0000) << 24) | ((v1 & 0xff000000) << 8)
              | ((v1 >> 8) & 0xff000000) | ((v1 >> 24) & 0xff0000)
              | ((v1 >> 40) & 0xff00) | (v1 >> 56));
        ((ULongLong*)&val)[0] = v1;
        ((ULongLong*)&val)[1] = v0;
    }

#endif

#if SIZEOF_LONG_DOUBLE == 12

#   ifdef WORDS_BIGENDIAN

#       error 12-bytes long doubles only supported for little endian!

#   else

    Octet tmp[18];

    if(swap_)
    {
	*(UShort*)&tmp = 0;
	memcpy(tmp + 2, pos_, 14);

        ULongLong v0 = ((ULongLong*)&tmp)[0];
        ULongLong v1 = ((ULongLong*)&tmp)[1];
        v0 = ((v0 << 56) | ((v0 & 0xff00) << 40)
              | ((v0 & 0xff0000) << 24) | ((v0 & 0xff000000) << 8)
              | ((v0 >> 8) & 0xff000000) | ((v0 >> 24) & 0xff0000)
              | ((v0 >> 40) & 0xff00) | (v0 >> 56));
        v1 = ((v1 << 56) | ((v1 & 0xff00) << 40)
              | ((v1 & 0xff0000) << 24) | ((v1 & 0xff000000) << 8)
              | ((v1 >> 8) & 0xff000000) | ((v1 >> 24) & 0xff0000)
              | ((v1 >> 40) & 0xff00) | (v1 >> 56));
        ((ULongLong*)&tmp)[0] = v1;
        ((ULongLong*)&tmp)[1] = v0;

	*(ULongLong*)(tmp + 4) >>= 1; // Adjust mantissa
	tmp[11] |= 0x80;                    // Set leading significant bit

	val = *(const LongDouble*)(tmp + 4);
    }
    else
    {
	*(UShort*)&tmp[16] = 0;
	memcpy(tmp + 2, pos_ + 2, 14);

	*(ULongLong*)(tmp + 6) >>= 1; // Adjust mantissa
	tmp[13] |= 0x80;                    // Set leading significant bit

	val = *(const LongDouble*)(tmp + 6);
    }

    pos_ += 16;

#   endif

#endif

    return val;
}

void
OutputStream::write_longdouble(LongDouble value)
{
    ULong offs = 7 - (((unsigned long)pos_ + 7) % 8);

#ifdef OB_MARSHAL_CHECK
    if((unsigned long)(pos_ - start_) + offs + 16 > size_)
        throw MARSHAL(MinorWriteLongDoubleOverflow, COMPLETED_NO);
#endif

    pos_ += offs;

#if SIZEOF_LONG_DOUBLE == 16

    *(LongDouble*)pos_ = value;

#endif

#if SIZEOF_LONG_DOUBLE == 12

#   ifdef WORDS_BIGENDIAN

#       error 12-bytes long doubles only supported for little endian!

#   else

    *(ULongLong*)pos_ = 0;
    *(LongDouble*)(pos_ + 6) = value;
    *(pos_ + 13) &= 0x7f;                // Clear leading significant bit
    *(ULongLong*)(pos_ + 6) <<= 1; // Adjust mantissa

#   endif

#endif

    pos_ += 16;
}

#endif

char*
InputStream::read_string()
{
    char* s;

    if(chunked())
        checkChunk();

    ULong len = read_ulong();

    if(len == 0)
        throw MARSHAL(MinorReadStringZeroLength, COMPLETED_NO);

#ifdef OB_MARSHAL_CHECK
    if((unsigned long)(pos_ - start_) + len > size_)
        throw MARSHAL(MinorReadStringOverflow, COMPLETED_NO);
#endif

    s = string_alloc(len - 1);

    const char* p = (const char*)pos_;
    strcpy(s, p);

    if(strlen(s) + 1 != len)
    {
        string_free(s);
        throw MARSHAL(MinorReadStringNullChar, COMPLETED_NO);
    }

    pos_ += len;

    return s;
}

void
OutputStream::write_string(const char* value)
{
    assert_nca(value, NCANullString);
    ULong len = strlen(value) + 1;
    write_ulong(len);

#ifdef OB_MARSHAL_CHECK
    if((((unsigned long)(pos_ - start_))) + len > size_)
	throw MARSHAL(MinorWriteStringOverflow, COMPLETED_NO);
#endif

    strcpy((char*)pos_, value);
    pos_ += len;
}

wchar_t*
InputStream::read_wstring()
{
    wchar_t* s;

    if(chunked())
        checkChunk();

    ULong len = read_ulong();

    if(len == 0)
        throw MARSHAL(MinorReadWStringZeroLength, COMPLETED_NO);

    WString_var ws;
    while(len > 0)
    {
        len -= (((ULong)*pos_) & 0xff) + 1;

        wchar_t v = read_wchar();
        if(!v)
            throw MARSHAL(MinorReadWStringNullWChar, COMPLETED_NO);
        ws += v;
    }

    s = wstring_alloc(wcslen(ws));
    wcscpy(s, ws);

    return s;
}

void
OutputStream::write_wstring(const wchar_t* value)
{
    assert_nca(value, NCANullString);

    CountStream len(&codeConverters_);
    ULong i;
    for(i = 0 ; i < wcslen(value) ; i++)
        len.count_wchar(value[i]);
    write_ulong(len.count());

    for(i = 0 ; i < wcslen(value) ; i++)
        write_wchar(value[i]);
}

void
CountStream::count_wstring(const wchar_t* value)
{
    assert_nca(value, NCANullString);

    count_ulong(); // 0 // length

    ULong i;
    for(i = 0 ; i < wcslen(value) ; i++)
        count_wchar(value[i]);
}

Object_ptr
InputStream::read_Object()
{
    if(chunked())
        checkChunk();

    IOP::IOR ior;
    IOP::IOR::_OB_unmarshal(ior, *this);
    ORB_ptr orb = ORB::_OB_instance_nodup();
    if(is_nil(orb))
        throw INITIALIZE("ORB not initialized");
    return orb -> _OB_createObject(ior);
}

void
OutputStream::write_Object(Object_ptr value)
{
    if(is_nil(value))
    {
        IOP::IOR ior;
        ior.type_id = string_dup("");
        IOP::IOR::_OB_marshal(ior, *this);
    }
    else
    {
        if(value -> _OB_isInternalLocal())
            throw MARSHAL(MinorWriteObjectLocal, COMPLETED_NO);
        IOP::IOR::_OB_marshal(value -> _ob_ior_, *this);
    }
}

void
CountStream::count_Object(Object_ptr value)
{
    if(is_nil(value))
    {
        IOP::IOR ior;
        ior.type_id = string_dup("");
        IOP::IOR::_OB_marshalCount(ior, *this);
    }
    else
    {
        if(value -> _OB_isInternalLocal())
            throw MARSHAL(MinorCountObjectLocal, COMPLETED_NO);
        IOP::IOR::_OB_marshalCount(value -> _ob_ior_, *this);
    }
}

TypeCode_ptr
InputStream::read_TypeCode()
{
    if(chunked())
        checkChunk();

    //
    // NOTE:
    //
    // No data with natural alignment of greater than four octets is
    // needed for TypeCode. Therefore it is not necessary to do
    // encapsulation in a seperate buffer.
    //

    if(readTCHistory_ == 0)
        readTCHistory_ = new ReadTCHistory(131);

    const Octet* oldPos = pos_;
    TCKind kind = (TCKind)read_ulong();

    TypeCode_var tc;
    if(kind == (TCKind)0xffffffff)
    {
        Long offs = read_long();
        const Octet* indirectionPos = pos_ - 4 + offs;
        TypeCode_ptr p;
        if(!readTCHistory_ -> get(indirectionPos, p))
            throw MARSHAL(MinorReadInvTypeCodeIndirection, COMPLETED_NO);
        readTCHistory_ -> put(oldPos, p);
        tc = TypeCode::_duplicate(p);
        // Don't use _OB_resolveRecTC() here. The TypeCode from the
        // history has already been resolved, or will be resolved by
        // callers of this operation.
    }
    else
    {
        switch(kind)
        {
        case tk_null:
        case tk_void:
        case tk_short:
        case tk_long:
        case tk_longlong:
        case tk_ushort:
        case tk_ulong:
        case tk_ulonglong:
        case tk_float:
        case tk_double:
        case tk_longdouble:
        case tk_fixed: // TODO: Is this really correct for fixed?
        case tk_boolean:
        case tk_char:
        case tk_wchar:
        case tk_octet:
        case tk_any:
        case tk_TypeCode:
        case tk_Principal:
        {
            tc = new TypeCode;
            readTCHistory_ -> put(oldPos, tc.in());
            tc -> kind_ = kind;
            break;
        }

        case tk_objref:
        {
            read_ulong(); // encapsulation length
            bool swap = swap_;
            _OB_readEndian();

            tc = new TypeCode;
            readTCHistory_ -> put(oldPos, tc.in());
            tc -> kind_ = kind;
            tc -> id_ = read_string();
            tc -> name_ = read_string();

            swap_ = swap;
            break;
        }

        case tk_struct:
        case tk_except:
        {
            read_ulong(); // encapsulation length
            bool swap = swap_;
            _OB_readEndian();

            tc = new TypeCode;
            readTCHistory_ -> put(oldPos, tc.in());
            tc -> kind_ = kind;
            tc -> id_ = read_string();
            tc -> name_ = read_string();
            ULong num = read_ulong();
            tc -> memberNames_.length(num);
            tc -> memberTypes_.length(num);
            for(ULong i = 0 ; i < num ; i++)
            {
                tc -> memberNames_[i] = read_string();
                tc -> memberTypes_[i] = read_TypeCode();
            }
            TypeCode::_OB_resolveRecTC(tc);

            swap_ = swap;
            break;
        }

        case tk_union:
        {
            read_ulong(); // encapsulation length
            bool swap = swap_;
            _OB_readEndian();

            tc = new TypeCode;
            readTCHistory_ -> put(oldPos, tc.in());
            tc -> kind_ = kind;
            tc -> id_ = read_string();
            tc -> name_ = read_string();
            tc -> discriminatorType_ = read_TypeCode();
            Long defaultIndex = read_long();
            ULong num = read_ulong();
            tc -> labels_.length(num);
            tc -> memberNames_.length(num);
            tc -> memberTypes_.length(num);
            for(ULong i = 0 ; i < num ; i++)
            {
                if((Long)i == defaultIndex)
                    read_any_NoTypeCode(tc -> labels_[i], _tc_octet);
                else
                    read_any_NoTypeCode(tc -> labels_[i],
                                        tc -> discriminatorType_);

                tc -> memberNames_[i] = read_string();
                tc -> memberTypes_[i] = read_TypeCode();
            }
            TypeCode::_OB_resolveRecTC(tc);

            swap_ = swap;
            break;
        }

        case tk_enum:
        {
            read_ulong(); // encapsulation length
            bool swap = swap_;
            _OB_readEndian();

            tc = new TypeCode;
            readTCHistory_ -> put(oldPos, tc.in());
            tc -> kind_ = kind;
            tc -> id_ = read_string();
            tc -> name_ = read_string();
            ULong num = read_ulong();
            tc -> memberNames_.length(num);
            for(ULong i = 0 ; i < num ; i++)
                tc -> memberNames_[i] = read_string();

            swap_ = swap;
            break;
        }

        case tk_string:
        case tk_wstring:
        {
            ULong bound = read_ulong();
            tc = new TypeCode;
            readTCHistory_ -> put(oldPos, tc.in());
            tc -> kind_ = kind;
            tc -> length_ = bound;
            break;
        }

        case tk_sequence:
        case tk_array:
        {
            read_ulong(); // encapsulation length
            bool swap = swap_;
            _OB_readEndian();

            tc = new TypeCode;
            readTCHistory_ -> put(oldPos, tc.in());
            tc -> kind_ = kind;
            tc -> contentType_ = read_TypeCode();
            tc -> length_ = read_ulong();

            swap_ = swap;
            break;
        }

        case tk_alias:
        {
            read_ulong(); // encapsulation length
            bool swap = swap_;
            _OB_readEndian();

            tc = new TypeCode;
            readTCHistory_ -> put(oldPos, tc.in());
            tc -> kind_ = kind;
            tc -> id_ = read_string();
            tc -> name_ = read_string();
            tc -> contentType_ = read_TypeCode();

            swap_ = swap;
            break;
        }

        case tk_value:
        {
            read_ulong(); // encapsulation length
            bool swap = swap_;
            _OB_readEndian();

            tc = new TypeCode;
            readTCHistory_ -> put(oldPos, tc.in());
            tc -> kind_ = kind;
            tc -> id_ = read_string();
            tc -> name_ = read_string();
            tc -> typeModifier_ = read_short();
            ULong num = read_ulong();
            tc -> memberNames_.length(num);
            tc -> memberTypes_.length(num);
            tc -> memberVisibility_.length(num);
            for(ULong i = 0 ; i < num ; i++)
            {
                tc -> memberNames_[i] = read_string();
                tc -> memberTypes_[i] = read_TypeCode();
                tc -> memberVisibility_[i] = read_short();
            }
            TypeCode::_OB_resolveRecTC(tc);

            swap_ = swap;
            break;
        }

        case tk_value_box:
        {
            read_ulong(); // encapsulation length
            bool swap = swap_;
            _OB_readEndian();

            tc = new TypeCode;
            readTCHistory_ -> put(oldPos, tc.in());
            tc -> kind_ = kind;
            tc -> id_ = read_string();
            tc -> name_ = read_string();
            tc -> contentType_ = read_TypeCode();
            TypeCode::_OB_resolveRecTC(tc);

            swap_ = swap;
            break;
        }

        case tk_abstract_interface:
        {
            read_ulong(); // encapsulation length
            bool swap = swap_;
            _OB_readEndian();

            tc = new TypeCode;
            readTCHistory_ -> put(oldPos, tc.in());
            tc -> kind_ = kind;
            tc -> id_ = read_string();
            tc -> name_ = read_string();

            swap_ = swap;
            break;
        }

        case tk_native:
        {
            read_ulong(); // encapsulation length
            bool swap = swap_;
            _OB_readEndian();

            tc = new TypeCode;
            readTCHistory_ -> put(oldPos, tc.in());
            tc -> kind_ = kind;
            tc -> id_ = read_string();
            tc -> name_ = read_string();

            swap_ = swap;
            break;
        }
        }
    }

    assert(!is_nil(tc));
    return tc._retn();
}

void
OutputStream::write_TypeCode(TypeCode_ptr tc)
{
    //
    // NOTE:
    //
    // No data with natural alignment of greater than four octets is
    // needed for TypeCode. Therefore it is not necessary to do
    // encapsulation in a seperate buffer.
    //

    if(is_nil(tc))
        throw BAD_TYPECODE();

    if(tc -> recId_.in())
    {
        assert_nca(!is_nil(tc -> recType_), NCAUnembeddedRecTC);
        write_TypeCode(tc -> recType_);
        return;
    }

    if(writeTCHistory_ == 0)
        writeTCHistory_ = new WriteTCHistory(131);

    const Octet* indirectionPos;
    if(writeTCHistory_ -> get(tc, indirectionPos))
    {
        write_ulong(0xffffffff);
        Long offs = indirectionPos - pos_;
        write_long(offs);
    }
    else
    {
        Octet* oldPos = pos_;
        write_ulong((ULong)tc -> kind());

        switch(tc -> kind())
        {
        case tk_null:
        case tk_void:
        case tk_short:
        case tk_long:
        case tk_longlong:
        case tk_ushort:
        case tk_ulong:
        case tk_ulonglong:
        case tk_float:
        case tk_double:
        case tk_longdouble:
        case tk_fixed: // TODO: Is this really correct for fixed?
        case tk_boolean:
        case tk_char:
        case tk_wchar:
        case tk_octet:
        case tk_any:
        case tk_TypeCode:
        case tk_Principal:
            break;

        case tk_objref:
        {
            writeTCHistory_ -> put(tc, oldPos);

            CountStream encapsLen;
            encapsLen.count_boolean(); // Endian
            encapsLen.count_string(tc -> id_);
            encapsLen.count_string(tc -> name_);

            write_ulong(encapsLen.count());
            write_boolean(Endian);
            write_string(tc -> id_);
            write_string(tc -> name_);

            break;
        }

        case tk_struct:
        case tk_except:
        {
            writeTCHistory_ -> put(tc, oldPos);

            ULong i;

            CountStream encapsLen;
            encapsLen.count_boolean(); // Endian
            encapsLen.count_string(tc -> id_);
            encapsLen.count_string(tc -> name_);
            encapsLen.count_ulong(); // tc -> memberNames_.length()
            for(i = 0 ; i < tc -> memberNames_.length() ; i++)
            {
                encapsLen.count_string(tc -> memberNames_[i]);
                encapsLen.count_TypeCode(tc -> memberTypes_[i]);
            }

            write_ulong(encapsLen.count());
            write_boolean(Endian);
            write_string(tc -> id_);
            write_string(tc -> name_);
            write_ulong(tc -> memberNames_.length());
            for(i = 0 ; i < tc -> memberNames_.length() ; i++)
            {
                write_string(tc -> memberNames_[i]);
                write_TypeCode(tc -> memberTypes_[i]);
            }

            break;
        }

        case tk_union:
        {
            writeTCHistory_ -> put(tc, oldPos);

            ULong i;

            CountStream encapsLen;
            encapsLen.count_boolean(); // Endian
            encapsLen.count_string(tc -> id_);
            encapsLen.count_string(tc -> name_);
            encapsLen.count_TypeCode(tc -> discriminatorType_);
            Long defaultIndex = tc -> default_index();
            encapsLen.count_long(); // defaultIndex
            encapsLen.count_ulong(); // tc -> memberNames_.length()
            for(i = 0 ; i < tc -> memberNames_.length() ; i++)
            {
                encapsLen.count_any_NoTypeCode(tc -> labels_[i]);
                encapsLen.count_string(tc -> memberNames_[i]);
                encapsLen.count_TypeCode(tc -> memberTypes_[i]);
            }

            write_ulong(encapsLen.count());
            write_boolean(Endian);
            write_string(tc -> id_);
            write_string(tc -> name_);
            write_TypeCode(tc -> discriminatorType_);
            write_long(defaultIndex);
            write_ulong(tc -> memberNames_.length());
            for(i = 0 ; i < tc -> memberNames_.length() ; i++)
            {
                write_any_NoTypeCode(tc -> labels_[i]);
                write_string(tc -> memberNames_[i]);
                write_TypeCode(tc -> memberTypes_[i]);
            }

            break;
        }

        case tk_enum:
        {
            writeTCHistory_ -> put(tc, oldPos);

            ULong i;

            CountStream encapsLen;
            encapsLen.count_boolean(); // Endian
            encapsLen.count_string(tc -> id_);
            encapsLen.count_string(tc -> name_);
            encapsLen.count_ulong(); // tc -> memberNames_.length()
            for(i = 0 ; i < tc -> memberNames_.length() ; i++)
                encapsLen.count_string(tc -> memberNames_[i]);

            write_ulong(encapsLen.count());
            write_boolean(Endian);
            write_string(tc -> id_);
            write_string(tc -> name_);
            write_ulong(tc -> memberNames_.length());
            for(i = 0 ; i < tc -> memberNames_.length() ; i++)
                write_string(tc -> memberNames_[i]);

            break;
        }

        case tk_string:
        case tk_wstring:
            write_ulong(tc -> length_);
            break;

        case tk_sequence:
        case tk_array:
        {
            writeTCHistory_ -> put(tc, oldPos);

            CountStream encapsLen;
            encapsLen.count_boolean(); // Endian
            encapsLen.count_TypeCode(tc -> contentType_);
            encapsLen.count_ulong(); // tc -> length_

            write_ulong(encapsLen.count());
            write_boolean(Endian);
            write_TypeCode(tc -> contentType_);
            write_ulong(tc -> length_);

            break;
        }

        case tk_alias:
        {
            writeTCHistory_ -> put(tc, oldPos);

            CountStream encapsLen;
            encapsLen.count_boolean(); // Endian
            encapsLen.count_string(tc -> id_);
            encapsLen.count_string(tc -> name_);
            encapsLen.count_TypeCode(tc -> contentType_);

            write_ulong(encapsLen.count());
            write_boolean(Endian);
            write_string(tc -> id_);
            write_string(tc -> name_);
            write_TypeCode(tc -> contentType_);

            break;
        }

        case tk_value:
        {
            writeTCHistory_ -> put(tc, oldPos);

            ULong i;

            CountStream encapsLen;
            encapsLen.count_boolean(); // Endian
            encapsLen.count_string(tc -> id_);
            encapsLen.count_string(tc -> name_);
            encapsLen.count_short(); // tc -> typeModifier_
            encapsLen.count_ulong(); // tc -> memberNames_.length()
            for(i = 0 ; i < tc -> memberNames_.length() ; i++)
            {
                encapsLen.count_string(tc -> memberNames_[i]);
                encapsLen.count_TypeCode(tc -> memberTypes_[i]);
                encapsLen.count_short(); // tc -> memberVisibility_[i]
            }

            write_ulong(encapsLen.count());
            write_boolean(Endian);
            write_string(tc -> id_);
            write_string(tc -> name_);
            write_short(tc -> typeModifier_);
            write_ulong(tc -> memberNames_.length());
            for(i = 0 ; i < tc -> memberNames_.length() ; i++)
            {
                write_string(tc -> memberNames_[i]);
                write_TypeCode(tc -> memberTypes_[i]);
                write_short(tc -> memberVisibility_[i]);
            }

            break;
        }

        case tk_value_box:
        {
            writeTCHistory_ -> put(tc, oldPos);

            CountStream encapsLen;
            encapsLen.count_boolean(); // Endian
            encapsLen.count_string(tc -> id_);
            encapsLen.count_string(tc -> name_);
            encapsLen.count_TypeCode(tc -> contentType_);

            write_ulong(encapsLen.count());
            write_boolean(Endian);
            write_string(tc -> id_);
            write_string(tc -> name_);
            write_TypeCode(tc -> contentType_);

            break;
        }

        case tk_abstract_interface:
        {
            writeTCHistory_ -> put(tc, oldPos);

            CountStream encapsLen;
            encapsLen.count_boolean(); // Endian
            encapsLen.count_string(tc -> id_);
            encapsLen.count_string(tc -> name_);

            write_ulong(encapsLen.count());
            write_boolean(Endian);
            write_string(tc -> id_);
            write_string(tc -> name_);

            break;
        }

        case tk_native:
        {
            writeTCHistory_ -> put(tc, oldPos);

            CountStream encapsLen;
            encapsLen.count_boolean(); // Endian
            encapsLen.count_string(tc -> id_);
            encapsLen.count_string(tc -> name_);

            write_ulong(encapsLen.count());
            write_boolean(Endian);
            write_string(tc -> id_);
            write_string(tc -> name_);

            break;
        }
        }
    }
}

void
CountStream::count_TypeCode(TypeCode_ptr tc)
{
    //
    // NOTE:
    //
    // No data with natural alignment of greater than four octets is
    // needed for TypeCode. Therefore it is not necessary to do
    // encapsulation in a seperate buffer.
    //

    if(is_nil(tc))
        throw BAD_TYPECODE();

    if(tc -> recId_.in())
    {
        assert_nca(!is_nil(tc -> recType_), NCAUnembeddedRecTC);
        count_TypeCode(tc -> recType_);
	return;
    }

    if(countTCHistory_ == 0)
        countTCHistory_ = new CountTCHistory(131);

    const Octet* dummy;
    if(countTCHistory_ -> get(tc, dummy))
    {
        count_ulong(); // 0xffffffff
        count_long();
    }
    else
    {
        count_ulong();

        switch(tc -> kind())
        {
        case tk_null:
        case tk_void:
        case tk_short:
        case tk_long:
        case tk_longlong:
        case tk_ushort:
        case tk_ulong:
        case tk_ulonglong:
        case tk_float:
        case tk_double:
        case tk_longdouble:
        case tk_fixed: // TODO: Is this really correct for fixed?
        case tk_boolean:
        case tk_char:
        case tk_wchar:
        case tk_octet:
        case tk_any:
        case tk_TypeCode:
        case tk_Principal:
            break;

        case tk_objref:
        {
            countTCHistory_ -> put(tc, 0);

            count_ulong(); // encapsulation length
            count_boolean();
            count_string(tc -> id_);
            count_string(tc -> name_);

            break;
        }

        case tk_struct:
        case tk_except:
        {
            countTCHistory_ -> put(tc, 0);

            count_ulong(); // encapsulation length
            count_boolean();
            count_string(tc -> id_);
            count_string(tc -> name_);
            count_ulong(); // tc -> memberNames_.length()
            for(ULong i = 0 ; i < tc -> memberNames_.length() ; i++)
            {
                count_string(tc -> memberNames_[i]);
                count_TypeCode(tc -> memberTypes_[i]);
            }

            break;
        }

        case tk_union:
        {
            countTCHistory_ -> put(tc, 0);

            count_ulong(); // encapsulation length
            count_boolean();
            count_string(tc -> id_);
            count_string(tc -> name_);
            count_TypeCode(tc -> discriminatorType_);
            count_long(); // tc -> default_index()
            count_ulong(); // tc -> memberNames_.length()
            for(ULong i = 0 ; i < tc -> memberNames_.length() ; i++)
            {
                count_any_NoTypeCode(tc -> labels_[i]);
                count_string(tc -> memberNames_[i]);
                count_TypeCode(tc -> memberTypes_[i]);
            }

            break;
        }

        case tk_enum:
        {
            countTCHistory_ -> put(tc, 0);

            count_ulong(); // encapsulation length
            count_boolean();
            count_string(tc -> id_);
            count_string(tc -> name_);
            count_ulong(); // tc -> memberNames_.length()
            for(ULong i = 0 ; i < tc -> memberNames_.length() ; i++)
                count_string(tc -> memberNames_[i]);

            break;
        }

        case tk_string:
        case tk_wstring:
            count_ulong(); // tc -> length_
            break;

        case tk_sequence:
        case tk_array:
            countTCHistory_ -> put(tc, 0);

            count_ulong(); // encapsulation length
            count_boolean();
            count_TypeCode(tc -> contentType_);
            count_ulong(); // tc -> length_

            break;

        case tk_alias:
            countTCHistory_ -> put(tc, 0);

            count_ulong(); // encapsulation length
            count_boolean();
            count_string(tc -> id_);
            count_string(tc -> name_);
            count_TypeCode(tc -> contentType_);

            break;

        case tk_value:
        {
            countTCHistory_ -> put(tc, 0);

            count_ulong(); // encapsulation length
            count_boolean();
            count_string(tc -> id_);
            count_string(tc -> name_);
            count_short(); // tc -> typeModifier_
            count_ulong(); // tc -> memberNames_.length()
            for(ULong i = 0 ; i < tc -> memberNames_.length() ; i++)
            {
                count_string(tc -> memberNames_[i]);
                count_TypeCode(tc -> memberTypes_[i]);
                count_short(); // tc -> memberVisibility_[i]
            }

            break;
        }

        case tk_value_box:
        {
            countTCHistory_ -> put(tc, 0);

            count_ulong(); // encapsulation length
            count_boolean();
            count_string(tc -> id_);
            count_string(tc -> name_);
            count_TypeCode(tc -> contentType_);

            break;
        }

        case tk_abstract_interface:
        {
            countTCHistory_ -> put(tc, 0);

            count_ulong(); // encapsulation length
            count_boolean();
            count_string(tc -> id_);
            count_string(tc -> name_);

            break;
        }

        case tk_native:
        {
            countTCHistory_ -> put(tc, 0);

            count_ulong(); // encapsulation length
            count_boolean();
            count_string(tc -> id_);
            count_string(tc -> name_);

            break;
        }
        }
    }
}

Principal_ptr
InputStream::read_Principal()
{
    if(chunked())
        checkChunk();

    //
    // Read an empty sequence of octets
    //
    read_ulong();
    return Principal::_nil();
}

void
OutputStream::write_Principal(Principal_ptr)
{
    //
    // Just marshal an empty sequence of octets
    //
    write_ulong(0);
}

void
CountStream::count_Principal(Principal_ptr)
{
    //
    // Count an empty sequence of octets
    //
    count_ulong(); // 0
}

Any*
InputStream::read_any()
{
    if(chunked())
        checkChunk();

    TypeCode_var tc = read_TypeCode();
    Any_var any = new Any;
    read_any_NoTypeCode(any.inout(), tc);
    return any._retn();
}

void
OutputStream::write_any(const Any& value)
{
    write_TypeCode(value.type_);
    write_any_NoTypeCode(value);
}

void
CountStream::count_any(const Any& value)
{
    count_TypeCode(value.type_);
    count_any_NoTypeCode(value);
}

void
InputStream::read_any(Any& any)
{
    if(chunked())
        checkChunk();

    TypeCode_var tc = read_TypeCode();
    read_any_NoTypeCode(any, tc);
}

Fixed
InputStream::read_fixed()
{
    if(chunked())
        checkChunk();

    String_var v;

    String_var s;
    while(true)
    {
#ifdef OB_MARSHAL_CHECK
        if((unsigned long)(pos_ - start_) + 1 > size_)
            throw MARSHAL(MinorReadFixedOverflow, COMPLETED_NO);
#endif
        Octet b = *pos_++;

        Octet hi = (b >> 4) & 0x0f;
        if(hi > 9)
            throw MARSHAL(MinorReadFixedInvalid, COMPLETED_NO);
        v += (char)(hi + '0');

        Octet lo = b & 0x0f;
        if(lo < 10)
            v += (char)(lo + '0');
        else if(lo == 0x0c || lo == 0x0d)
        {
            if(lo == 0x0d)
                s = string_dup("-");
            break;
        }
        else
            throw MARSHAL(MinorReadFixedInvalid, COMPLETED_NO);
    }

    s += v;

    return Fixed((const char*)s);
}

void
OutputStream::write_fixed(const Fixed& value)
{
    String_var v = (value.abs()).to_string();

#ifdef OB_MARSHAL_CHECK
    ULong fixLen = (value.fixed_digits() + 2) / 2;
    if((unsigned long)(pos_ - start_) + fixLen > size_)
        throw MARSHAL(MinorWriteFixedOverflow, COMPLETED_NO);
#endif

    //
    // Append coded sign to value string
    //
    if(value.signum() == -1)
        v += (char)('0' + 0x0d);
    else
        v += (char)('0' + 0x0c);

    String_var s;
    if(strlen(v) % 2 != 0)
        s = string_dup("0");

    s += v;

    ULong i;
    for(i = 0 ; i < strlen(v) ; i += 2)
        *pos_++ = ((s[i] - '0') << 4) | (s[i + 1] - '0');
}

void
CountStream::count_fixed(const CORBA::Fixed& value)
{
    count_ += (value.fixed_digits() + 2) / 2;
}


ValueBase*
InputStream::read_Value(const char* id)
{
    //
    // The `id' parameter indicates the type being unmarshalled at the
    // current "position".  This does not necessarily specify the most
    // derived type of the value.
    //

    //
    // Lazy instantiation of hashtable
    //
    if(valueHistory_ == 0)
        valueHistory_ = new ValueHistory(131);

    ValueInfo info;
    info.headerPos = pos_;
    info.tag = read_long();
    info.state = valueState_;

    //
    // Check for special cases (null values and indirections)
    //
    if(info.tag == 0)
    {
        valueHistory_ -> put(info.headerPos, 0);
        return 0;
    }
    else if(info.tag == (Long)0xffffffff)
    {
        Long offs = read_long();
        const Octet* pos = pos_ - 4 + offs;

        //
        // Check the history for a value that was seen at the specified
        // position. Generally, we should expect the value to be present.
        // However, in the case of chunking, it's possible for an
        // indirection to refer to a nested value in the truncated state
        // of an enclosing value.
        //
        ValueBase_var v;
        if(valueHistory_ -> get(pos, v.out()))
        {
            add_ref(v);
            return v._retn();
        }
        else
        {
            const Octet* save = pos_;

            //
            // Check for indirection to null value
            //
            pos_ = pos; // rewind to offset position
            if(readLongImpl() == 0)
            {
                pos_ = save;
                valueHistory_ -> put(pos, 0);
                return 0;
            }

            //
            // If it's not null and it's not in our info history, then
            // there's no hope
            //
            ValueInfo nest;
            if(valueInfoHistory_ == 0 || !valueInfoHistory_ -> get(pos, nest))
                throw MARSHAL(1, COMPLETED_NO);

            //
            // Attempt to create the value (this could raise MARSHAL)
            //
            v = createValue(id, nest);

            //
            // Now it's time to unmarshal the value members
            //
            pos_ = nest.dataPos;
            ValueState saveState(valueState_);
            valueState_ = nest.state;
            if(chunked())
                readChunk(valueState_);

            v -> _OB_unmarshalValue(*this);

            //
            // Restore state
            //
            pos_ = save;
            valueState_ = saveState;

            return v._retn();
        }
    }

    //
    // Read value header info
    //
    readValueInfo(info);
    valueState_ = info.state;

    //
    // Instantiate value
    //
    ValueBase_var vb = createValue(id, info);

    if(chunked())
    {
        readChunk(valueState_);
        valueState_.nestingLevel++;
    }

    //
    // Unmarshal value state
    //
    vb -> _OB_unmarshalValue(*this);

    if(chunked())
    {
        //
        // At this point, the unmarshalling code has finished. However,
        // we may have a truncated value, or we may have unmarshalled a
        // custom value. In either case, we can't be sure that the
        // unmarshalling code has positioned the stream at the end of
        // this value.
        //
        // Therefore we will advance, chunk by chunk, until we reach
        // the end of the value.
        //

        //
        // Skip to the end of the current chunk (if necessary)
        //
        if(valueState_.chunkStart)
        {
            pos_ = valueState_.chunkStart;
            _OB_skip(valueState_.chunkSize);
        }

        valueState_.chunkStart = 0;
        valueState_.chunkSize = 0;

        //
        // Loop until we have reached the end of this value. We may
        // have to process nested values, chunks, etc.
        //
        Long level = valueState_.nestingLevel;
        const Octet* pos = pos_;
        Long l = readLongImpl();
        while(l >= 0 || (l < 0 && l < -valueState_.nestingLevel))
        {
            if(l >= 0x7fffff00)
            {
                //
                // This indicates a nested value. We read the header
                // information and store it away, in case a subsequent
                // indirection refers to this value.
                //
                level++;
                ValueInfo nest;
                nest.tag = l;
                nest.headerPos = pos;
                nest.state.nestingLevel = level;
                readValueInfo(nest);
                if(valueInfoHistory_ == 0)
                    valueInfoHistory_ = new ValueInfoHistory(131);
                valueInfoHistory_ -> put(nest.headerPos, nest);
            }
            else if(l >= 0)
            {
                //
                // Chunk size - advance the stream past the chunk
                //
                _OB_skip(l);
            }
            else
            {
                //
                // l is less than 0, so this is an end tag for a nested
                // value
                //
                level--;
            }

            pos = pos_;
            l = readLongImpl();
        }

        //
        // If the tag is greater than our nesting level, then this
        // value coterminates with an outer value. We rewind the
        // stream so that the outer value can read this tag.
        //
        if(l > -valueState_.nestingLevel)
            pos_ -= 4;

        valueState_.nestingLevel--;

        if(valueState_.nestingLevel == 0)
            valueState_.chunked = false;
        else
        {
            //
            // We're chunked and still processing nested values, so
            // another chunk may follow
            //
            readChunk(valueState_);
        }
    }

    return vb._retn();
}

void
OutputStream::write_Value(ValueBase* value, const char* id)
{
    //
    // The `id' parameter indicates the type being marshalled at the
    // current "position".  This does not necessarily specify the most
    // derived type of the value.
    //

    //
    // Lazy instantiation of hashtables
    //
    if(valueHistory_ == 0)
    {
        valueHistory_ = new ValueHistory(131);
        idHistory_ = new IdHistory(131);
        idListHistory_ = new IdListHistory(131);
    }

    //
    // Check for null value
    //
    if(value == 0)
    {
        write_long(0);
        return;
    }

    //
    // Check to see if we've already marshalled this value
    //
    Long pos;
    if(valueHistory_ -> get(value, pos))
    {
        //
        // Write indirection
        //
        write_long((Long)0xffffffff);
        Long off = pos - ((Long)(pos_ - start_));
        write_long(off);
        return;
    }

    //
    // Remember value in history at current position
    //
    pos = (Long)(pos_ - start_);
    valueHistory_ -> put(value, pos);

    //
    // If nestingLevel_ > 0, then we are about to marshal a nested valuetype,
    // and therefore we have to fix the size of the previous chunk
    //
    if(nestingLevel_ > 0)
    {
        //
        // Compute the size of the previous chunk.  If the size is 0,
        // then we can remove the chunk.
        //
        Long size = (Long)(pos - (chunkPos_ + 4));

        if(size == 0)
            pos_ -= 4;  // rewind past the chunk size
        else
        {
            Long* pSize = (Long*)(start_ + chunkPos_);
            *pSize = size;
        }
    }

    //
    // Get the actual type of the value and determine if it matches
    // the formal type
    //
    const char* valueId = value -> _OB_typeId();
    bool typeMatch = (id != 0 && strcmp(id, valueId) == 0);

    //
    // Chunking is necessary when:
    //
    // * value is custom
    // * value is truncatable and actual type does not match formal type
    //
    Boolean isCustom = value -> _OB_isCustom();
    Boolean isTruncatable = value -> _OB_isTruncatable();
    if(isCustom || (isTruncatable && !typeMatch))
        chunked_ = true;

    if(chunked())
        nestingLevel_++;

    //
    // Value header
    //
    // Determine whether we need to transmit repository ID information
    // for this value. This is necessary when:
    //
    // * The formal type wasn't supplied or doesn't match the actual type
    // * This is a nested value and it is truncatable
    //

    if(!typeMatch || (nestingLevel_ > 1 && isTruncatable))
    {
        if(isTruncatable)
        {
            //
            // Value is truncatable, so write the tag and then
            // a sequence of strings representing the repository
            // IDs of the safe bases.
            //
            // The tag is 0x7fffff00 + 0x06 (partial type information) +
            // 0x08 (chunked).
            //
            write_long((Long)0x7fffff0e);

            //
            // TODO: We can optimize the list of identifiers by analyzing
            // the receiving context (see GIOP section 15.3.4.1).
            //

            const char** ids = value -> _OB_truncatableIds();

            //
            // Check for possible indirection of repository IDs
            //
            if(idListHistory_ -> get(ids, pos))
            {
                //
                // Write indirection
                //
                write_long((Long)0xffffffff);
                Long off = pos - ((Long)(pos_ - start_));
                write_long(off);
            }
            else
            {
                pos = (Long)(pos_ - start_);
                idListHistory_ -> put(ids, pos);
                ULong i, idCount = 0;
                for(i = 0 ; ids[i] != 0 ; i++)
                    idCount++;
                write_long((Long)idCount);
                for(i = 0 ; i < idCount ; i++)
                    write_string(ids[i]);
            }
        }
        else
        {
            //
            // No formal ID was provided, or the formal ID doesn't
            // match the actual ID, so we need to send a single
            // repository ID.
            //
            // The tag is 0x7fffff00 + 0x02 (single ID) [+ 0x08 if chunked]
            //

            Long tag = (Long)0x7fffff02;

            if(chunked())
                tag |= 0x08;

            write_long(tag);

            //
            // Check to see if we've already marshalled this repository ID,
            // and if so, we write an indirection marker.
            //
            if(idHistory_ -> get(valueId, pos))
            {
                //
                // Write indirection
                //
                write_long((Long)0xffffffff);
                Long off = pos - ((Long)(pos_ - start_));
                write_long(off);
            }
            else
            {
                //
                // Remember ID in history at current position
                //
                pos = (Long)(pos_ - start_);
                idHistory_ -> put(valueId, pos);
                write_string(valueId);
            }
        }
    }
    else
    {
        //
        // The value's dynamic type matches the type of the
        // position being marshalled, so we just need to
        // marshal the tag (no repository ID).
        //
        // The tag is 0x7fffff00 [+ 0x08 if chunked]
        //

        Long tag = (Long)0x7fffff00;

        if(chunked())
            tag |= 0x08;

        write_long(tag);
    }

    if(chunked())
    {
        //
        // Write a placeholder for the chunk size and remember its position
        //
        write_long(0);
        chunkPos_ = (ULong)(pos_ - 4 - start_);
        endTag_ = false;
    }

    //
    // Marshal the value data
    //
    value -> _OB_marshalValue(*this);

    if(chunked())
    {
        //
        // Compute the size of the chunk and the pointer to the chunk size
        //
        Long size = (Long)(pos_ - (start_ + chunkPos_ + 4));
        Long* pSize = (Long*)(start_ + chunkPos_);

        //
        // If endTag_ is true then the size follows an end tag.
        // Furthermore, if the computed size is 0, then there was no
        // data written between the last end tag and this one, so
        // this value coterminates with the nested value. Don't
        // write a new end tag, just increment the previous one.
        //
        // XXX Perhaps we should always eliminate the chunk if size == 0,
        //     regardless of the value of endTag_ (?)
        //
        if(endTag_ && size == 0)
        {
            Long* tag = (Long*)(pSize - 1);
            (*tag)++;  // increment the previous end tag
        }
        else
        {
            //
            // Fix the chunk size and write the end tag
            //
            *pSize = size;
            write_long(-nestingLevel_);
        }

        //
        // We're finished chunking if nestingLevel_ == 1
        //
        if(nestingLevel_ == 1)
        {
            chunked_ = false;
            chunkPos_ = 0;
            //
            // We are finished marshalling the outermost value.
            // If no data was written since the previous end tag,
            // we need to rewind past the chunk size that was added
            // by the previous value. (Since the size follows an end
            // tag, we don't have to worry about alignment)
            //
            if(endTag_ && size == 0)
                pos_ -= 4;
        }
        else
        {
            //
            // We've ended a nested valuetype, but we may not have finished
            // marshalling the outer valuetype(s), so we start a new chunk.
            // If this is also the end of the outer valuetype, then the
            // code above will rewind the buffer.
            //
            // We are setting endTag_ to true, which signals the code that
            // this size was written after an end tag.
            //
            write_long(0);
            chunkPos_ = (ULong)(pos_ - 4 - start_);
            endTag_ = true;
        }

        nestingLevel_--;
    }
}

void
CountStream::count_Value(ValueBase* value, const char* id)
{
    //
    // The `id' parameter indicates the type being marshalled at the
    // current "position".  This does not necessarily specify the most
    // derived type of the value.
    //

    //
    // Lazy instantiation of hashtables
    //
    if(valueHistory_ == 0)
    {
        valueHistory_ = new ValueHistory(131);
        idHistory_ = new IdHistory(131);
        idListHistory_ = new IdListHistory(131);
    }

    //
    // Check for null value
    //
    if(value == 0)
    {
        count_long();
        return;
    }

    //
    // Check to see if we've already marshalled this value
    //
    Long pos;
    if(valueHistory_ -> get(value, pos))
    {
        //
        // Write indirection
        //
        count_long(); // 0xffffffff
        count_long(); // offset
        return;
    }

    //
    // Remember value in history at current position
    //
    valueHistory_ -> put(value, 0);

    //
    // If nestingLevel_ > 0, then we are counting a nested valuetype,
    // and therefore we have to fix the size of the previous chunk
    //
    if(nestingLevel_ > 0)
    {
        //
        // Compute the size of the previous chunk.  If the size is 0,
        // then we can remove the chunk.
        //
        Long size = (Long)(count_ - (chunkPos_ + 4));
        if(size == 0)
            count_ -= 4;  // rewind past the chunk size
    }

    //
    // Get the actual type of the value and determine if it matches
    // the formal type
    //
    const char* valueId = value -> _OB_typeId();
    bool typeMatch = (id != 0 && strcmp(id, valueId) == 0);

    //
    // Chunking is necessary when:
    //
    // * value is custom
    // * value is truncatable and actual type does not match formal type
    //
    Boolean isCustom = value -> _OB_isCustom();
    Boolean isTruncatable = value -> _OB_isTruncatable();
    if(isCustom || (isTruncatable && !typeMatch))
        chunked_ = true;

    if(chunked())
        nestingLevel_++;

    //
    // Value header
    //
    // Determine whether we need to transmit repository ID information
    // for this value. This is necessary when:
    //
    // * The formal type wasn't supplied or doesn't match the actual type
    // * This is a nested value and it is truncatable
    //

    if(!typeMatch || (nestingLevel_ > 1 && isTruncatable))
    {
        if(isTruncatable)
        {
            //
            // Value is truncatable, so write the tag and then
            // a sequence of strings representing the repository
            // IDs of the safe bases.
            //
            // The tag is 0x7fffff00 + 0x06 (partial type information) +
            // 0x08 (chunked).
            //
            count_long(); // 0x7fffff0e

            //
            // TODO: We can optimize the list of identifiers by analyzing
            // the receiving context.
            //

            const char** ids = value -> _OB_truncatableIds();

            //
            // Check for possible indirection of repository IDs
            //
            if(idListHistory_ -> get(ids, pos))
            {
                //
                // Write indirection
                //
                count_long(); // 0xffffffff
                count_long(); // offset
            }
            else
            {
                //
                // The ID of the value's type is not included in the array,
                // so we need to write it first.
                //
                idListHistory_ -> put(ids, 0);
                ULong i, idCount = 0;
                for(i = 0 ; ids[i] != 0 ; i++)
                    idCount++;
                count_long(); // idCount
                for(i = 0 ; i < idCount ; i++)
                    count_string(ids[i]);
            }
        }
        else
        {
            //
            // No formal ID was provided, or the formal ID doesn't
            // match the actual ID, so we need to send a single
            // repository ID.
            //
            // The tag is 0x7fffff00 + 0x02 (single ID) [+ 0x08 if chunked]
            //
            count_long(); // tag

            //
            // Check to see if we've already marshalled this repository ID,
            // and if so, we write an indirection marker.
            //
            if(idHistory_ -> get(valueId, pos))
            {
                //
                // Write indirection
                //
                count_long(); // 0xffffffff
                count_long(); // offset
            }
            else
            {
                //
                // Remember ID in history at current position
                //
                idHistory_ -> put(valueId, 0);
                count_string(valueId);
            }
        }
    }
    else
    {
        //
        // The value's dynamic type matches the type of the
        // position being marshalled, so we just need to
        // marshal the tag (no repository ID).
        //
        // The tag is 0x7fffff00 [+ 0x08 if chunked]
        //
        count_long(); // tag
    }

    if(chunked())
    {
        //
        // Write a dummy value for the chunk size and remember its position
        //
        count_long(); // size (dummy)
        chunkPos_ = count_ - 4;
        endTag_ = false;
    }

    //
    // Marshal the value data
    //
    value -> _OB_marshalCountValue(*this);

    if(chunked())
    {
        //
        // Compute the chunk size.  If endTag_ is true then the
        // size follows an end tag.  Furthermore, if the computed
        // size is 0, then there was no data written between the
        // last end tag and this one, so we can optimize things.
        //
        ULong size = count_ - (chunkPos_ + 4);

        //
        // If endTag_ is true then the size follows an end tag.
        // Furthermore, if the computed size is 0, then there was no
        // data written between the last end tag and this one, so
        // we can share the last one (i.e., only write another end tag
        // when necessary).
        //
        if(!endTag_ || size > 0)
        {
            //
            // Write the end tag
            //
            count_long(); // -nestingLevel_
        }

        //
        // We're finished chunking if nestingLevel_ == 1
        //
        if(nestingLevel_ == 1)
        {
            chunked_ = false;
            chunkPos_ = 0;
            //
            // We are finished marshalling the outermost value.
            // If no data was written since the previous end tag,
            // we need to rewind past the chunk size that was added
            // by the previous value.  (Since the size follows an end
            // tag, we don't have to worry about alignment)
            //
            if(endTag_ && size == 0)
                count_ -= 4;
        }
        else
        {
            //
            // We've ended a nested valuetype, but we may not have finished
            // marshalling the outer valuetype(s), so we start a new chunk.
            // If this is also the end of the outer valuetype, then the
            // code above will rewind the buffer.
            //
            // We are setting endTag_ to true, which signals the code that
            // this size was written after an end tag.
            //
            count_long(); // size (dummy)
            chunkPos_ = count_ - 4;
            endTag_ = true;
        }

        nestingLevel_--;
    }
}

void
OutputStream::write_Abstract(AbstractBase_ptr value)
{
    if(value)
    {
        Object_ptr obj = value -> _OB_toObject();
        if(!is_nil(obj))
        {
            write_boolean(TRUE); // discriminator for objref
            write_Object(obj);
        }
        else
        {
            ValueBase* vb = value -> _OB_toValue();
            assert(vb != 0); // XXX exception?
            write_boolean(FALSE); // discriminator for valuetype
            write_Value(vb);
        }
    }
    else
    {
        // XXX what to do for a nil abstract interface??
        write_boolean(FALSE); // discriminator for valuetype
        write_Value(0);
    }
}

void
CountStream::count_Abstract(AbstractBase_ptr value)
{
    if(value)
    {
        Object_ptr obj = value -> _OB_toObject();
        if(!is_nil(obj))
        {
            count_boolean(); // TRUE - discriminator for objref
            count_Object(obj);
        }
        else
        {
            ValueBase* vb = value -> _OB_toValue();
            assert(vb != 0); // XXX exception?
            count_boolean(); // FALSE - discriminator for valuetype
            count_Value(vb);
        }
    }
    else
    {
        // XXX what to do for a nil abstract interface??
        count_boolean(); // FALSE - discriminator for valuetype
        count_Value(0);
    }
}

void
InputStream::read_wchar_array(WChar* value, ULong length)
{
    if(length)
    {
        assert_nca(value, NCANullValue);

        if(chunked())
            checkChunk();

        for(ULong i = 0 ; i < length ; i++)
            value[i] = read_wchar();
    }
}

void
OutputStream::write_wchar_array(const WChar* value, ULong length)
{
    if(length)
    {
        assert_nca(value, NCANullValue);
        const WChar* v = value;
        ULong i;
        for(i = 0 ; i < length ; i++)
            write_wchar(v[i]);
    }
}

void
CountStream::count_wchar_array(const WChar* value, ULong length)
{
    if(length)
    {
        assert_nca(value, NCANullValue);
        ULong i;
        for(i = 0 ; i < length ; i++)
            count_wchar(value[i]);
    }
}

void
InputStream::read_short_array(Short* value, ULong length)
{
    if(length)
    {
        assert_nca(value, NCANullValue);

        if(chunked())
            checkChunk();

        ULong offs = (unsigned long)pos_ % 2;
#ifdef OB_MARSHAL_CHECK
        if((unsigned long)(pos_ - start_) + offs + length * 2 > size_)
            throw MARSHAL(MinorReadShortArrayOverflow,
                                COMPLETED_NO);
#endif
        pos_ += offs;
        memcpy(value, pos_, length * 2);
        pos_ += length * 2;

        if(swap_)
        {
            UShort* v = (UShort*)value;
            while(length-- > 0)
            {
                *v = (*v << 8) | (*v >> 8);
                v++;
            }
        }
    }
}

void
InputStream::read_ushort_array(UShort* value, ULong length)
{
    if(length)
    {
        assert_nca(value, NCANullValue);

        if(chunked())
            checkChunk();

        ULong offs = (unsigned long)pos_ % 2;
#ifdef OB_MARSHAL_CHECK
        if((unsigned long)(pos_ - start_) + offs + length * 2 > size_)
            throw MARSHAL(MinorReadUShortArrayOverflow,
                                COMPLETED_NO);
#endif
        pos_ += offs;
        memcpy(value, pos_, length * 2);
        pos_ += length * 2;

        if(swap_)
        {
            while(length-- > 0)
            {
                *value = (*value << 8) | (*value >> 8);
                value++;
            }
        }
    }
}

void
InputStream::read_long_array(Long* value, ULong length)
{
    if(length)
    {
        assert_nca(value, NCANullValue);

        if(chunked())
            checkChunk();

        ULong offs = 3 - (((unsigned long)pos_ + 3) % 4);
#ifdef OB_MARSHAL_CHECK
        if((unsigned long)(pos_ - start_) + offs + length * 4 > size_)
            throw MARSHAL(MinorReadLongArrayOverflow,
                                COMPLETED_NO);
#endif
        pos_ += offs;
        memcpy(value, pos_, length * 4);
        pos_ += length * 4;

        if(swap_)
        {
            ULong* v = (ULong*)value;
            while(length-- > 0)
            {
                *v = ((*v << 24) | ((*v & 0xff00) << 8)
                      | ((*v >> 8) & 0xff00) | (*v >> 24));
                v++;
            }
        }
    }
}

void
InputStream::read_ulong_array(ULong* value, ULong length)
{
    if(length)
    {
        assert_nca(value, NCANullValue);

        if(chunked())
            checkChunk();

        ULong offs = 3 - (((unsigned long)pos_ + 3) % 4);
#ifdef OB_MARSHAL_CHECK
        if((unsigned long)(pos_ - start_) + offs + length * 4 > size_)
            throw MARSHAL(MinorReadULongArrayOverflow,
                                COMPLETED_NO);
#endif
        pos_ += offs;
        memcpy(value, pos_, length * 4);
        pos_ += length * 4;

        if(swap_)
        {
            while(length-- > 0)
            {
                *value = ((*value << 24) | ((*value & 0xff00) << 8)
                      | ((*value >> 8) & 0xff00) | (*value >> 24));
                value++;
            }
        }
    }
}

void
InputStream::read_longlong_array(LongLong* value, ULong length)
{
    if(length)
    {
        assert_nca(value, NCANullValue);

        if(chunked())
            checkChunk();

        ULong offs = 7 - (((unsigned long)pos_ + 7) % 8);
#ifdef OB_MARSHAL_CHECK
        if((unsigned long)(pos_ - start_) + offs + length * 8 > size_)
            throw MARSHAL(MinorReadLongLongArrayOverflow,
                                COMPLETED_NO);
#endif
        pos_ += offs;
        memcpy(value, pos_, length * 8);
        pos_ += length * 8;

        if(swap_)
        {
            ULongLong* v = (ULongLong*)value;
            while(length-- > 0)
            {
                *v = ((*v << 56) | ((*v & 0xff00) << 40)
                      | ((*v & 0xff0000) << 24) | ((*v & 0xff000000) << 8)
                      | ((*v >> 8) & 0xff000000) | ((*v >> 24) & 0xff0000)
                      | ((*v >> 40) & 0xff00) | (*v >> 56));
                v++;
            }
        }
    }
}

void
InputStream::read_ulonglong_array(ULongLong* value, ULong length)
{
    if(length)
    {
        assert_nca(value, NCANullValue);

        if(chunked())
            checkChunk();

        ULong offs = 7 - (((unsigned long)pos_ + 7) % 8);
#ifdef OB_MARSHAL_CHECK
        if((unsigned long)(pos_ - start_) + offs + length * 8 > size_)
            throw MARSHAL(MinorReadULongLongArrayOverflow,
                                COMPLETED_NO);
#endif
        pos_ += offs;
        memcpy(value, pos_, length * 8);
        pos_ += length * 8;

        if(swap_)
        {
            ULongLong* v = value;
            while(length-- > 0)
            {
                *v = ((*v << 56) | ((*v & 0xff00) << 40)
                      | ((*v & 0xff0000) << 24) | ((*v & 0xff000000) << 8)
                      | ((*v >> 8) & 0xff000000) | ((*v >> 24) & 0xff0000)
                      | ((*v >> 40) & 0xff00) | (*v >> 56));
                v++;
            }
        }
    }
}

void
InputStream::read_float_array(Float* value, ULong length)
{
    if(length)
    {
        assert_nca(value, NCANullValue);

        if(chunked())
            checkChunk();

        ULong offs = 3 - (((unsigned long)pos_ + 3) % 4);
#ifdef OB_MARSHAL_CHECK
        if((unsigned long)(pos_ - start_) + offs + length * 4 > size_)
            throw MARSHAL(MinorReadFloatArrayOverflow,
                                COMPLETED_NO);
#endif
        pos_ += offs;
        memcpy(value, pos_, length * 4);
        pos_ += length * 4;

        if(swap_)
        {
            ULong* v = (ULong*)value;
            while(length-- > 0)
            {
                *v = ((*v << 24) | ((*v & 0xff00) << 8)
                      | ((*v >> 8) & 0xff00) | (*v >> 24));
                v++;
            }
        }
    }
}

void
InputStream::read_double_array(Double* value, ULong length)
{
    if(length)
    {
        assert_nca(value, NCANullValue);

        if(chunked())
            checkChunk();

        ULong offs = 7 - (((unsigned long)pos_ + 7) % 8);
#ifdef OB_MARSHAL_CHECK
        if((unsigned long)(pos_ - start_) + offs + length * 8 > size_)
            throw MARSHAL(MinorReadDoubleArrayOverflow, COMPLETED_NO);
#endif
        pos_ += offs;
        memcpy(value, pos_, length * 8);
        pos_ += length * 8;

        if(swap_)
        {
            ULong* v = (ULong*)value;
            while(length-- > 0)
            {
                ULong v0 = v[0];
                ULong v1 = v[1];
                v0 = ((v0 << 24) | ((v0 & 0xff00) << 8)
                      | ((v0 >> 8) & 0xff00) | (v0 >> 24));
                v1 = ((v1 << 24) | ((v1 & 0xff00) << 8)
                      | ((v1 >> 8) & 0xff00) | (v1 >> 24));
                v[0] = v1;
                v[1] = v0;
                v += 2;
            }
        }
    }
}

#if SIZEOF_LONG_DOUBLE == 0

void
InputStream::read_longdouble_array(LongDouble* value, ULong length)
{
    throw MARSHAL(MinorLongDoubleNotSupported, COMPLETED_NO);
}

void
OutputStream::write_longdouble_array(const LongDouble* value, ULong length)
{
    throw MARSHAL(MinorLongDoubleNotSupported, COMPLETED_NO);
}

#else

void
InputStream::read_longdouble_array(LongDouble* value, ULong length)
{
    if(length)
    {
        assert_nca(value, NCANullValue);

        if(chunked())
            checkChunk();

#if SIZEOF_LONG_DOUBLE == 16

        ULong offs = 7 - (((unsigned long)pos_ + 7) % 8);
#ifdef OB_MARSHAL_CHECK
        if((unsigned long)(pos_ - start_) + offs + length * 16 > size_)
            throw MARSHAL(MinorReadLongDoubleArrayOverflow, COMPLETED_NO);
#endif
        pos_ += offs;
        memcpy(value, pos_, length * 16);
        pos_ += length * 16;

        if(swap_)
        {
            ULongLong* v = (ULongLong*)value;
            while(length-- > 0)
            {
                ULongLong v0 = v[0];
                ULongLong v1 = v[1];
                v0 = ((v0 << 56) | ((v0 & 0xff00) << 40)
                      | ((v0 & 0xff0000) << 24) | ((v0 & 0xff000000) << 8)
                      | ((v0 >> 8) & 0xff000000) | ((v0 >> 24) & 0xff0000)
                      | ((v0 >> 40) & 0xff00) | (v0 >> 56));
                v1 = ((v1 << 56) | ((v1 & 0xff00) << 40)
                      | ((v1 & 0xff0000) << 24) | ((v1 & 0xff000000) << 8)
                      | ((v1 >> 8) & 0xff000000) | ((v1 >> 24) & 0xff0000)
                      | ((v1 >> 40) & 0xff00) | (v1 >> 56));
                v[0] = v1;
                v[1] = v0;
                v += 2;
            }
        }

#endif

#if SIZEOF_LONG_DOUBLE == 12

        ULong i;
        for(i = 0 ; i < length ; i++)
            value[i] = read_longdouble();

#endif
    }
}

void
OutputStream::write_longdouble_array(const LongDouble* value, ULong length)
{
    if(length)
    {
#if SIZEOF_LONG_DOUBLE == 16

        ULong offs = 7 - (((unsigned long)pos_ + 7) % 8);
#ifdef OB_MARSHAL_CHECK
        if((unsigned long)(pos_ - start_) + offs + length * 16 > size_)
            throw MARSHAL(MinorWriteLongDoubleArrayOverflow, COMPLETED_NO);
#endif
        pos_ += offs;
        memcpy(pos_, value, length * 8);
        pos_ += length * 16;

#endif

#if SIZEOF_LONG_DOUBLE == 12

        ULong i;
        for(i = 0 ; i < length ; i++)
            write_longdouble(value[i]);

#endif
    }
}

#endif

void
InputStream::read_any_NoTypeCode(Any& any, TypeCode_ptr tc)
{
    void* val = 0;
    
    TypeCode_var p = GetOrigType(tc);
    
    switch(p -> kind())
    {
    case tk_null:
    case tk_void:
	break;
	
    case tk_short:
    {
	Short v = read_short();
	val = new Short(v);
	break;
    }

    case tk_long:
    {
	Long v = read_long();
	val = new Long(v);
	break;
    }

    case tk_longlong:
    {
	LongLong v = read_longlong();
	val = new LongLong(v);
	break;
    }

    case tk_ushort:
    {
	UShort v = read_ushort();
	val = new UShort(v);
	break;
    }

    case tk_ulong:
    case tk_enum:
    {
	ULong v = read_ulong();
	val = new ULong(v);
	break;
    }

    case tk_ulonglong:
    {
	ULongLong v = read_ulonglong();
	val = new ULongLong(v);
	break;
    }

    case tk_float:
    {
	Float v = read_float();
	val = new Float(v);
	break;
    }

    case tk_double:
    {
	Double v = read_double();
	val = new Double(v);
	break;
    }

    case tk_longdouble:
    {
	LongDouble v = read_longdouble();
	val = new LongDouble(v);
	break;
    }

    case tk_fixed:
    {
	Fixed v = read_fixed();;
	val = new Fixed(v);
	break;
    }

    case tk_boolean:
    {
	Boolean v = read_boolean();
	val = new Boolean(v);
	break;
    }

    case tk_char:
    {
	Char v = read_char();
	val = new Char(v);
	break;
    }

    case tk_wchar:
    {
	WChar v = read_wchar();
// XXX Codeset conversion
	val = new WChar(v);
	break;
    }

    case tk_octet:
    {
	Octet v = read_octet();
	val = new Octet(v);
	break;
    }

    case tk_any:
    {
	val = read_any();
	break;
    }

    case tk_TypeCode:
    {
	val = read_TypeCode();
	break;
    }

    case tk_Principal:
    {
	val = read_Principal();
	break;
    }
	
    case tk_objref:
    {
	val = read_Object();
	break;
    }

    case tk_struct:
    case tk_sequence:
    case tk_union:
    case tk_array:
    case tk_except:

	if(p -> _OB_isSystemException())
	{
	    val = SystemException::_OB_unmarshal(*this);
	}
	else
	{
            CountStream count;
            const Octet* save = pos_;
	    count.count_InputStream(p, *this);
            pos_ = save;
	    Buffer* bufp = new Buffer(count.count());
	    val = bufp;
            OutputStream out(*bufp, 0);
	    out.write_InputStream(p, *this);
	}
	break;

    case tk_string:
    {
	val = read_string();
	break;
    }
	
    case tk_wstring:
    {
	wchar_t* s = read_wstring();
// XXX Codeset conversion
	val = s;
	break;
    }

    case tk_value:
    case tk_value_box:
        val = read_Value(tc -> id());
        break;

    case tk_alias: // Alias not possible here
    case tk_abstract_interface: // Abstract interface not possible here
    case tk_native: // Native not possible here
        assert(false);
        break;     // this is to keep the compiler happy
    }

    any.replace(tc, val, true);
}

void
OutputStream::write_any_NoTypeCode(const Any& any)
{
    TypeCode_var p = GetOrigType(any.type_);
    
    switch(p -> kind())
    {
    case tk_null:
    case tk_void:
	    break;
	    
    case tk_short:
	write_short(*(const Short*)any.value_);
	break;
	
    case tk_long:
	write_long(*(const Long*)any.value_);
	break;

    case tk_longlong:
	write_longlong(*(const LongLong*)any.value_);
	break;
	
    case tk_ushort:
	write_ushort(*(const UShort*)any.value_);
	break;
	
    case tk_ulong:
    case tk_enum:
	write_ulong(*(const ULong*)any.value_);
	break;

    case tk_ulonglong:
	write_ulonglong(*(const ULongLong*)any.value_);
	break;

    case tk_float:
	write_float(*(const Float*)any.value_);
	break;
	
    case tk_double:
	write_double(*(const Double*)any.value_);
	break;
	
    case tk_longdouble:
	write_longdouble(*(const LongDouble*)any.value_);
	break;
	
    case tk_fixed:
	write_fixed(*(const Fixed*)any.value_);
	break;
	
    case tk_boolean:
	write_boolean(*(const Boolean*)any.value_);
	break;
	
    case tk_char:
	write_char(*(const Char*)any.value_);
	break;

    case tk_wchar:
	write_wchar(*(const WChar*)any.value_);
	break;

    case tk_octet:
	write_octet(*(const Octet*)any.value_);
	break;
	
    case tk_any:
	write_any(*(const Any*)any.value_);
	break;
	
    case tk_TypeCode:
	write_TypeCode((TypeCode_ptr)any.value_);
	break;
	
    case tk_Principal:
	write_Principal((Principal_ptr)any.value_);
	break;
	
    case tk_objref:
	write_Object((Object_ptr)any.value_);
	break;
	
    case tk_struct:
    case tk_sequence:
    case tk_union:
    case tk_array:
    case tk_except:

	if(p -> _OB_isSystemException())
	{
	    SystemException::_OB_marshal(*(SystemException*)any.value_, *this);
	}
	else if(any.info_)
	{
	    any.info_ -> marshal(any.value_, *this);
	}
	else
	{
	    const Buffer* cbufp = (const Buffer*)any.value_;
            InputStream in(*cbufp, 0, false);
	    write_InputStream(p, in);
	}
	break;
	
    case tk_string:
	write_string((const char*)any.value_);
	break;
	
    case tk_wstring:
	write_wstring((const wchar_t*)any.value_);
	break;

    case tk_value:
    case tk_value_box:
        write_Value((ValueBase*)any.value_, p -> id());
        break;

    case tk_alias: // Alias not possible here
    case tk_abstract_interface: // Abstract interface not possible here
    case tk_native: // Native not possible here
        assert(false);
        break;     // this is to keep the compiler happy
    }
}

void
CountStream::count_any_NoTypeCode(const Any& any)
{
    TypeCode_var p = GetOrigType(any.type_);

    switch(p -> kind())
    {
    case tk_null:
    case tk_void:
	break;

    case tk_short:
	count_short(); // *(const Short*)any.value_
	break;

    case tk_long:
	count_long(); // *(const Long*)any.value_
	break;

    case tk_longlong:
	count_longlong(); // *(const LongLong*)any.value_
	break;

    case tk_ushort:
	count_ushort(); // *(const UShort*)any.value_
	break;

    case tk_ulong:
    case tk_enum:
	count_ulong(); // *(const ULong*)any.value_
	break;

    case tk_ulonglong:
	count_ulonglong(); // *(const ULongLong*)any.value_
	break;

    case tk_float:
	count_float(); // *(const Float*)any.value_
	break;

    case tk_double:
	count_double(); // *(const Double*)any.value_
	break;

    case tk_longdouble:
	count_longdouble(); // *(const LongDouble*)any.value_
	break;

    case tk_fixed:
	count_fixed(*(const Fixed*)any.value_);
	break;

    case tk_boolean:
	count_boolean(); // *(const Boolean*)any.value_
	break;

    case tk_char:
	count_char(); // *(const Char*)any.value_
	break;

    case tk_wchar:
	count_wchar(*(const WChar*)any.value_);
	break;

    case tk_octet:
	count_octet(); // *(const Octet*)any.value_
	break;

    case tk_any:
	count_any(*(const Any*)any.value_);
	break;

    case tk_TypeCode:
	count_TypeCode((TypeCode_ptr)any.value_);
	break;

    case tk_Principal:
	count_Principal((Principal_ptr)any.value_);
	break;

    case tk_objref:
	count_Object((Object_ptr)any.value_);
	break;

    case tk_struct:
    case tk_sequence:
    case tk_union:
    case tk_array:
    case tk_except:

	if(p -> _OB_isSystemException())
	{
	    SystemException::_OB_marshalCount(*(SystemException*)any.value_,
                                              *this);
	}
	else if(any.info_)
	{
	    any.info_ -> marshalCount(any.value_, *this);
	}
	else
	{
	    const Buffer* cbufp = (const Buffer*)any.value_;
            InputStream in(*cbufp, 0, false);
	    count_InputStream(p, in);
	}
	break;

    case tk_string:
	count_string((const char*)any.value_);
	break;

    case tk_wstring:
	count_wstring((const wchar_t*)any.value_);
	break;

    case tk_value:
    case tk_value_box:
        count_Value((ValueBase*)any.value_, p -> id());
        break;

    case tk_alias: // Alias not possible here
    case tk_abstract_interface: // Abstract interface not possible here
    case tk_native: // Native not possible here
        assert(false);
        break;     // this is to keep the compiler happy
    }
}

void
OutputStream::write_InputStream(TypeCode_ptr tc, InputStream& in)
{
    switch(tc -> kind())
    {
    case tk_null:
    case tk_void:
	break;
	
    case tk_boolean:
	write_boolean(in.read_boolean());
	break;

    case tk_char:
	write_char(in.read_char());
	break;

    case tk_octet:
	write_octet(in.read_octet());
	break;

    case tk_wchar:
	write_wchar(in.read_wchar());
// XXX Codeset conversion
	break;

    case tk_short:
	write_short(in.read_short());
	break;

    case tk_ushort:
	write_ushort(in.read_ushort());
	break;
    
    case tk_long:
	write_long(in.read_long());
	break;
    
    case tk_ulong:
    case tk_enum:
	write_ulong(in.read_ulong());
	break;
    
    case tk_float:
	write_float(in.read_float());
	break;
    
    case tk_longlong:
	write_longlong(in.read_longlong());
	break;
    
    case tk_ulonglong:
	write_ulonglong(in.read_ulonglong());
	break;
    
    case tk_double:
	write_double(in.read_double());
	break;

    case tk_longdouble:
	write_longdouble(in.read_longdouble());
	break;

    case tk_fixed:
	write_fixed(in.read_fixed());
	break;

    case tk_any:
    {
	// Don't do this:
	//
	// Any_var any = in.read_any();
	// write_any(any);
	//
	// This is faster:

	TypeCode_var p = in.read_TypeCode();
	write_TypeCode(p);
	write_InputStream(p, in);
	break;
    }
    
    case tk_TypeCode:
    {
	// Don't do this:
	//
	// TypeCode_var p = in.read_TypeCode();
	// write_TypeCode(p);
	//
	// This is faster:
	
        const Octet* save = in._OB_pos();
	TCKind kind = (TCKind)in.read_ulong();

	if(kind == (TCKind)0xffffffff)
	{
            in._OB_pos(save);
            TypeCode_var p = in.read_TypeCode();
            write_TypeCode(p);
	}
	else
	{
	    write_ulong((ULong)kind);

	    switch(kind)
	    {
	    case tk_null:
	    case tk_void:
	    case tk_short:
	    case tk_long:
	    case tk_longlong:
	    case tk_ushort:
	    case tk_ulong:
	    case tk_ulonglong:
	    case tk_float:
	    case tk_double:
	    case tk_longdouble:
	    case tk_fixed:
	    case tk_boolean:
	    case tk_char:
	    case tk_wchar:
	    case tk_octet:
	    case tk_any:
	    case tk_TypeCode:
	    case tk_Principal:
		break;
		
	    case tk_objref:
	    case tk_struct:
	    case tk_except:
	    case tk_union:
	    case tk_enum:
	    case tk_sequence:
	    case tk_array:
	    case tk_alias:
	    case tk_value:
	    case tk_value_box:
	    case tk_abstract_interface:
	    case tk_native:
	    {
		ULong len = in.read_ulong();
		write_ulong(len);
#ifdef OB_MARSHAL_CHECK
                if((unsigned long)(pos_ - start_) + len > size_)
                    throw MARSHAL(MinorWriteOverflow, COMPLETED_NO);
#endif
		in.read_octet_array(pos_, len);
		pos_ += len;
		break;
	    }
	    
	    case tk_string:
	    case tk_wstring:
	    {
		ULong bound = in.read_ulong();
// XXX Codeset conversion
		write_ulong(bound);
		break;
	    }
       	    }
	}
	
	break;
    }
    
    case tk_Principal:
    {
	Principal_var p = in.read_Principal();
	write_Principal(p);
	break;
    }

    case tk_objref:
    {
	// Don't do this:
	//
	// Object_var p = in.read_Object();
	// write_Object(p);
	//
	// This is faster:

	IOP::IOR ior;
	IOP::IOR::_OB_unmarshal(ior, in);
	IOP::IOR::_OB_marshal(ior, *this);
	break;
    }

    case tk_struct:
    {
	for(ULong i = 0 ; i < tc -> memberTypes_.length() ; i++)
	    write_InputStream(tc -> memberTypes_[i], in);
	
	break;
    }
	
    case tk_except:
    {
	String_var id = in.read_string();
	write_string(id);

	for(ULong i = 0 ; i < tc -> memberTypes_.length() ; i++)
	    write_InputStream(tc -> memberTypes_[i], in);
	
	break;
    }
	
    case tk_union:
    {
	Long defaultIndex = tc -> default_index();
	Long memberIndex = -1;

	TypeCode_var discType = GetOrigType(tc -> discriminatorType_);

	switch(discType -> kind())
	{
	case tk_short:
	{
	    Short val = in.read_short();
	    write_short(val);
	    
	    for(ULong i = 0 ; i < tc -> labels_.length() ; i++)
		if((Long)i != defaultIndex)
		    if(val == *(Short*)tc -> labels_[i].value())
		    {
			memberIndex = i;
			break;
		    }
	    
	    break;
	}

	case tk_ushort:
	{
	    UShort val = in.read_ushort();
	    write_ushort(val);

	    for(ULong i = 0 ; i < tc -> labels_.length() ; i++)
		if((Long)i != defaultIndex)
		    if(val == *(UShort*)tc -> labels_[i].value())
		    {
			memberIndex = i;
			break;
		    }

	    break;
	}

	case tk_long:
	{
	    Long val = in.read_long();
	    write_long(val);

	    for(ULong i = 0 ; i < tc -> labels_.length() ; i++)
		if((Long)i != defaultIndex)
		    if(val == *(Long*)tc -> labels_[i].value())
		    {
			memberIndex = i;
			break;
		    }

	    break;
	}

	case tk_ulong:
	case tk_enum:
	{
	    ULong val = in.read_ulong();
	    write_ulong(val);

	    for(ULong i = 0 ; i < tc -> labels_.length() ; i++)
		if((Long)i != defaultIndex)
		    if(val == *(ULong*)tc -> labels_[i].value())
		    {
			memberIndex = i;
			break;
		    }

	    break;
	}

	case tk_longlong:
	{
	    LongLong val = in.read_longlong();
	    write_longlong(val);

	    for(ULong i = 0 ; i < tc -> labels_.length() ; i++)
		if((Long)i != defaultIndex)
		    if(val == *(LongLong*)tc -> labels_[i].value())
		    {
			memberIndex = i;
			break;
		    }

	    break;
	}

	case tk_ulonglong:
	{
	    ULongLong val = in.read_ulonglong();
	    write_ulonglong(val);

	    for(ULong i = 0 ; i < tc -> labels_.length() ; i++)
		if((ULong)i != (ULong)defaultIndex)
		    if(val == *(ULongLong*)tc -> labels_[i].value())
		    {
			memberIndex = i;
			break;
		    }

	    break;
	}

	case tk_char:
	{
	    Char val = in.read_char();
	    write_char(val);

	    for(ULong i = 0 ; i < tc -> labels_.length() ; i++)
		if((Long)i != defaultIndex)
		    if(val == *(Char*)tc -> labels_[i].value())
		    {
			memberIndex = i;
			break;
		    }

	    break;
	}

	case tk_wchar:
	{
	    WChar val = in.read_wchar();
	    write_wchar(val);

	    for(ULong i = 0 ; i < tc -> labels_.length() ; i++)
		if((Long)i != defaultIndex)
		    if(val == *(WChar*)tc -> labels_[i].value())
		    {
			memberIndex = i;
			break;
		    }

	    break;
	}

	case tk_boolean:
	{
	    Boolean val = in.read_boolean();
	    write_boolean(val);

	    for(ULong i = 0 ; i < tc -> labels_.length() ; i++)
		if((Long)i != defaultIndex)
		    if(val == *(Boolean*)tc -> labels_[i].value())
		    {
			memberIndex = i;
			break;
		    }

	    break;
	}

	default:
	    assert(false);
	    break;
	}

	if(memberIndex >= 0)
	    write_InputStream(tc -> memberTypes_[memberIndex], in);
	else if(defaultIndex >= 0)
	    write_InputStream(tc -> memberTypes_[defaultIndex], in);

	break;
    }
	
    case tk_string:
    {
	String_var s = in.read_string();
	write_string(s);
	break;
    }

    case tk_wstring:
    {
	WString_var s = in.read_wstring();
	write_wstring(s);
	break;
    }

    case tk_sequence:
    case tk_array:
    {
	ULong len;

	if(tc -> kind() == tk_sequence)
	{
	    len = in.read_ulong();
	    write_ulong(len);
	}
	else
	    len = tc -> length_;
	
	if(len > 0)
	{
	    assert(!is_nil(tc -> contentType_));

	    TypeCode_var origContentType = GetOrigType(tc -> contentType_);
	    
	    switch(origContentType -> kind())
	    {
	    case tk_null:
	    case tk_void:
		break;
		
	    case tk_short:
	    {
                ULong offs = (unsigned long)pos_ % 2;
#ifdef OB_MARSHAL_CHECK
                if((unsigned long)(pos_ - start_) + offs + len * 2 > size_)
                    throw MARSHAL(MinorWriteShortArrayOverflow,
                                        COMPLETED_NO);
#endif
                pos_ += offs;
		in.read_short_array((Short*)pos_, len);
		pos_ += len * 2;
		break;
	    }
	    
	    case tk_ushort:
	    {
                ULong offs = (unsigned long)pos_ % 2;
#ifdef OB_MARSHAL_CHECK
                if((unsigned long)(pos_ - start_) + offs + len * 2 > size_)
                    throw MARSHAL(MinorWriteUShortArrayOverflow,
                                        COMPLETED_NO);
#endif
                pos_ += offs;
		in.read_ushort_array((UShort*)pos_, len);
		pos_ += len * 2;
		break;
	    }
	    
	    case tk_long:
	    {
                ULong offs = 3 - (((unsigned long)pos_ + 3) % 4);
#ifdef OB_MARSHAL_CHECK
                if((unsigned long)(pos_ - start_) + offs + len * 4 > size_)
                    throw MARSHAL(MinorWriteLongArrayOverflow,
                                        COMPLETED_NO);
#endif
                pos_ += offs;
		in.read_long_array((Long*)pos_, len);
		pos_ += len * 4;
		break;
	    }

	    case tk_ulong:
	    {
                ULong offs = 3 - (((unsigned long)pos_ + 3) % 4);
#ifdef OB_MARSHAL_CHECK
                if((unsigned long)(pos_ - start_) + offs + len * 4 > size_)
                    throw MARSHAL(MinorWriteULongArrayOverflow,
                                        COMPLETED_NO);
#endif
                pos_ += offs;
		in.read_ulong_array((ULong*)pos_, len);
		pos_ += len * 4;
		break;
	    }

	    case tk_float:
	    {
                ULong offs = 3 - (((unsigned long)pos_ + 3) % 4);
#ifdef OB_MARSHAL_CHECK
                if((unsigned long)(pos_ - start_) + offs + len * 4 > size_)
                    throw MARSHAL(MinorWriteFloatArrayOverflow,
                                        COMPLETED_NO);
#endif
                pos_ += offs;
		in.read_float_array((Float*)pos_, len);
		pos_ += len * 4;
		break;
	    }

	    case tk_longlong:
	    {
                ULong offs = 7 - (((unsigned long)pos_ + 7) % 8);
#ifdef OB_MARSHAL_CHECK
                if((unsigned long)(pos_ - start_) + offs + len * 8 > size_)
                    throw MARSHAL(MinorWriteLongLongArrayOverflow,
                                        COMPLETED_NO);
#endif
                pos_ += offs;
		in.read_longlong_array((LongLong*)pos_, len);
		pos_ += len * 8;
		break;
	    }

	    case tk_ulonglong:
	    {
                ULong offs = 7 - (((unsigned long)pos_ + 7) % 8);
#ifdef OB_MARSHAL_CHECK
                if((unsigned long)(pos_ - start_) + offs + len * 8 > size_)
                    throw MARSHAL(MinorWriteULongLongArrayOverflow,
                                        COMPLETED_NO);
#endif
                pos_ += offs;
		in.read_ulonglong_array((ULongLong*)pos_, len);
		pos_ += len * 8;
		break;
	    }
	    
	    case tk_double:
	    {
                ULong offs = 7 - (((unsigned long)pos_ + 7) % 8);
#ifdef OB_MARSHAL_CHECK
                if((unsigned long)(pos_ - start_) + offs + len * 8 > size_)
                    throw MARSHAL(MinorWriteDoubleArrayOverflow,
                                        COMPLETED_NO);
#endif
                pos_ += offs;
		in.read_double_array((Double*)pos_, len);
		pos_ += len * 8;
		break;
	    }

	    case tk_longdouble:
	    {
                ULong offs = 7 - (((unsigned long)pos_ + 7) % 8);
#ifdef OB_MARSHAL_CHECK
                if((unsigned long)(pos_ - start_) + offs + len * 16 > size_)
                    throw MARSHAL(MinorWriteLongDoubleArrayOverflow,
                                        COMPLETED_NO);
#endif
                pos_ += offs;
		in.read_longdouble_array((LongDouble*)pos_, len);
		pos_ += len * 16;
		break;
	    }

	    case tk_boolean:
	    {
#ifdef OB_MARSHAL_CHECK
                if((unsigned long)(pos_ - start_) + len > size_)
                    throw MARSHAL(MinorWriteBooleanArrayOverflow,
                                        COMPLETED_NO);
#endif
		in.read_boolean_array(pos_, len);
		pos_ += len;
		break;
	    }

	    case tk_char:
	    {
#ifdef OB_MARSHAL_CHECK
                if((unsigned long)(pos_ - start_) + len > size_)
                    throw MARSHAL(MinorWriteCharArrayOverflow,
                                        COMPLETED_NO);
#endif
		in.read_char_array((CORBA::Char*)pos_, len);
		pos_ += len;
		break;
	    }

	    case tk_octet:
	    {
#ifdef OB_MARSHAL_CHECK
                if((unsigned long)(pos_ - start_) + len > size_)
                    throw MARSHAL(MinorWriteOctetArrayOverflow,
                                        COMPLETED_NO);
#endif
		in.read_octet_array(pos_, len);
		pos_ += len;
		break;
	    }

	    case tk_wchar:
	    {
                ULong sz = sizeof(wchar_t);
                ULong offs = (sz - 1) - (((unsigned long)pos_ +
                                   (sz - 1)) % sz);
#ifdef OB_MARSHAL_CHECK
                if((unsigned long)(pos_ - start_) + offs + len * sz > size_)
                    throw MARSHAL(MinorWriteWCharArrayOverflow,
                                        COMPLETED_NO);
#endif
                pos_ += offs;
		in.read_wchar_array((WChar*)pos_, len);
		pos_ += len * sz;
// XXX Codeset conversion
		break;
	    }
    
	    case tk_alias:
		assert(false);
		break;
		
	    default:
		for(ULong i = 0 ; i < len ; i++)
		    write_InputStream(tc -> contentType_, in);
		break;
	    }
	}

	break;
    }

    case tk_alias:
	write_InputStream(tc -> contentType_, in);
	break;

    case tk_value:
    {
        // XXX header info?
	for(ULong i = 0 ; i < tc -> memberTypes_.length() ; i++)
	    write_InputStream(tc -> memberTypes_[i], in);
	break;
    }

    case tk_value_box:
        // XXX header info?
	write_InputStream(tc -> contentType_, in);
	break;

    case tk_abstract_interface:
    case tk_native:
        assert(false);
        break;     // this is to keep the compiler happy
    }
}

void
CountStream::count_InputStream(TypeCode_ptr tc, InputStream& in)
{
    switch(tc -> kind())
    {
    case tk_null:
    case tk_void:
	break;
	
    case tk_boolean:
    case tk_char:
    case tk_octet:
    {
	in.read_boolean();
	count_boolean();
	break;
    }

    case tk_wchar:
    {
	WChar val = in.read_wchar();
// XXX Codeset conversion
	count_wchar(val);
	break;
    }

    case tk_short:
    {
	in.read_short();
	count_short();
	break;
    }
    
    case tk_ushort:
    {
	in.read_ushort();
	count_ushort();
	break;
    }
    
    case tk_long:
    {
	in.read_long();
	count_long();
	break;
    }

    case tk_enum:
    case tk_ulong:
    {
	in.read_ulong();
	count_ulong();
	break;
    }

    case tk_float:
    {
	in.read_float();
	count_float();
	break;
    }

    case tk_longlong:
    {
	in.read_longlong();
	count_longlong();
	break;
    }
    
    case tk_ulonglong:
    {
	in.read_ulonglong();
	count_ulonglong();
	break;
    }
    
    case tk_double:
    {
	in.read_double();
	count_double();
	break;
    }

    case tk_longdouble:
    {
	in.read_longdouble();
	count_longdouble();
	break;
    }

    case tk_fixed:
    {
	Fixed val = in.read_fixed();
	count_fixed(val);
	break;
    }

    case tk_any:
    {
	// Don't do this:
	//
	// Any_var any = in.read_any();
	// count_any(any);
	//
	// This is faster:

	TypeCode_var p = in.read_TypeCode();
	count_TypeCode(p);
	count_InputStream(p, in);
	break;
    }
    
    case tk_TypeCode:
    {
	// Don't do this:
	//
	// TypeCode_var p = in.read_TypeCode();
	// count_TypeCode(p);
	//
	// This is faster:
	
        const Octet* save = in._OB_pos();
	TCKind kind = (TCKind)in.read_ulong();

	if(kind == (TCKind)0xffffffff)
	{
            in._OB_pos(save);
            TypeCode_var p = in.read_TypeCode();
            count_TypeCode(p);
	}
	else
	{
	    count_ulong(); // (ULong)kind

	    switch(kind)
	    {
	    case tk_null:
	    case tk_void:
	    case tk_short:
	    case tk_long:
	    case tk_longlong:
	    case tk_ushort:
	    case tk_ulong:
	    case tk_ulonglong:
	    case tk_float:
	    case tk_double:
	    case tk_longdouble:
	    case tk_fixed:
	    case tk_boolean:
	    case tk_char:
	    case tk_wchar:
	    case tk_octet:
	    case tk_any:
	    case tk_TypeCode:
	    case tk_Principal:
		break;
		
	    case tk_objref:
	    case tk_struct:
	    case tk_except:
	    case tk_union:
	    case tk_enum:
	    case tk_sequence:
	    case tk_array:
	    case tk_alias:
	    case tk_value:
	    case tk_value_box:
	    case tk_abstract_interface:
	    case tk_native:
	    {
		ULong len = in.read_ulong();
		count_ulong(); // len
		in._OB_skip(len);
		count_ += len;
		break;
	    }
	    
	    case tk_string:
	    case tk_wstring:
	    {
		in.read_ulong();
		count_ulong(); // bound
		break;
	    }
	    }
	}
	
	break;
    }
    
    case tk_Principal:
    {
        Principal_var p = in.read_Principal();
	count_Principal(p);
	break;
    }

    case tk_objref:
    {
	// Don't do this:
	//
	// Object_var p = in.read_Object();
	// count_Object(p);
	//
	// This is faster:

	IOP::IOR ior;
	IOP::IOR::_OB_unmarshal(ior, in);
	IOP::IOR::_OB_marshalCount(ior, *this);
	break;
    }
	
    case tk_struct:
    {
	for(ULong i = 0 ; i < tc -> memberTypes_.length() ; i++)
	    count_InputStream(tc -> memberTypes_[i], in);
	
	break;
    }
	
    case tk_except:
    {
	String_var id = in.read_string();
	count_string(id);

	for(ULong i = 0 ; i < tc -> memberTypes_.length() ; i++)
	    count_InputStream(tc -> memberTypes_[i], in);
	
	break;
    }

    case tk_union:
    {
	Long defaultIndex = tc -> default_index();
	Long memberIndex = -1;
	
	TypeCode_var discType = GetOrigType(tc -> discriminatorType_);

	switch(discType -> kind())
	{
	case tk_short:
	{
	    Short val = in.read_short();
	    count_short(); // val
	    
	    ULong i;
	    for(i = 0 ; i < tc -> labels_.length() ; i++)
		if((Long)i != defaultIndex)
		    if(val == *(Short*)tc -> labels_[i].value())
		    {
			memberIndex = i;
			break;
		    }
	    
	    break;
	}

	case tk_ushort:
	{
	    UShort val = in.read_ushort();
	    count_ushort(); // val

	    for(ULong i = 0 ; i < tc -> labels_.length() ; i++)
		if((Long)i != defaultIndex)
		    if(val == *(UShort*)tc -> labels_[i].value())
		    {
			memberIndex = i;
			break;
		    }

	    break;
	}

	case tk_long:
	{
	    Long val = in.read_long();
	    count_long(); // val

	    for(ULong i = 0 ; i < tc -> labels_.length() ; i++)
		if((Long)i != defaultIndex)
		    if(val == *(Long*)tc -> labels_[i].value())
		    {
			memberIndex = i;
			break;
		    }

	    break;
	}

	case tk_ulong:
	case tk_enum:
	{
	    ULong val = in.read_ulong();
	    count_ulong(); // val

	    for(ULong i = 0 ; i < tc -> labels_.length() ; i++)
		if((Long)i != defaultIndex)
		    if(val == *(ULong*)tc -> labels_[i].value())
		    {
			memberIndex = i;
			break;
		    }

	    break;
	}

	case tk_longlong:
	{
	    LongLong val = in.read_longlong();
	    count_longlong(); // val

	    for(ULong i = 0 ; i < tc -> labels_.length() ; i++)
		if((Long)i != defaultIndex)
		    if(val == *(LongLong*)tc -> labels_[i].value())
		    {
			memberIndex = i;
			break;
		    }

	    break;
	}

	case tk_ulonglong:
	{
	    ULongLong val = in.read_ulonglong();
	    count_ulonglong(); // val

	    for(ULong i = 0 ; i < tc -> labels_.length() ; i++)
		if((ULong)i != (ULong)defaultIndex)
		    if(val == *(ULongLong*)tc -> labels_[i].value())
		    {
			memberIndex = i;
			break;
		    }

	    break;
	}

	case tk_char:
	{
	    Char val = in.read_char();
	    count_char(); // val

	    for(ULong i = 0 ; i < tc -> labels_.length() ; i++)
		if((Long)i != defaultIndex)
		    if(val == *(Char*)tc -> labels_[i].value())
		    {
			memberIndex = i;
			break;
		    }

	    break;
	}

	case tk_wchar:
	{
	    WChar val = in.read_wchar();
	    count_wchar(val);
// XXX Codeset conversion

	    for(ULong i = 0 ; i < tc -> labels_.length() ; i++)
		if((Long)i != defaultIndex)
		    if(val == *(WChar*)tc -> labels_[i].value())
		    {
			memberIndex = i;
			break;
		    }

	    break;
	}

	case tk_boolean:
	{
	    Boolean val = in.read_boolean();
	    count_boolean(); // val

	    for(ULong i = 0 ; i < tc -> labels_.length() ; i++)
		if((Long)i != defaultIndex)
		    if(val == *(Boolean*)tc -> labels_[i].value())
		    {
			memberIndex = i;
			break;
		    }

	    break;
	}

	default:
	    assert(false);
	    break;
	}

	if(memberIndex >= 0)
	    count_InputStream(tc -> memberTypes_[memberIndex], in);
	else if(defaultIndex >= 0)
	    count_InputStream(tc -> memberTypes_[defaultIndex], in);

	break;
    }
	
    case tk_string:
    {
	String_var s = in.read_string();
	count_string(s);
	break;
    }

    case tk_wstring:
    {
	WString_var s = in.read_wstring();
// XXX Codeset conversion
	count_wstring(s);
	break;
    }

    case tk_sequence:
    case tk_array:
    {
	ULong len;

	if(tc -> kind() == tk_sequence)
	{
	    len = in.read_ulong();
	    count_ulong(); // len
	}
	else
	    len = tc -> length_;

	if(len > 0)
	{
	    assert(!is_nil(tc -> contentType_));

	    TypeCode_var origContentType = GetOrigType(tc -> contentType_);
	    
	    switch(origContentType -> kind())
	    {
	    case tk_null:
	    case tk_void:
		break;
		
	    case tk_short:
	    case tk_ushort:
	    {
                in._OB_skipAlign(2);
                in._OB_skip(len * 2);
		count_ += count_ % 2;
		count_ += len * 2;
		break;
	    }
	    
	    case tk_long:
	    case tk_ulong:
	    case tk_float:
	    {
		in._OB_skipAlign(4);
		in._OB_skip(len * 4);
		count_ += 3;
		count_ -= count_ % 4;
		count_ += len * 4;
		break;
	    }
	    
	    case tk_longlong:
	    case tk_ulonglong:
	    case tk_double:
	    {
		in._OB_skipAlign(8);
		in._OB_skip(len * 8);
		count_ += 7;
		count_ -= count_ % 8;
		count_ += len * 8;
		break;
	    }

	    case tk_longdouble:
	    {
		in._OB_skipAlign(8);
		in._OB_skip(len * 16);
		count_ += 7;
		count_ -= count_ % 8;
		count_ += len * 16;
		break;
	    }
	    
	    case tk_boolean:
	    case tk_char:
	    case tk_octet:
	    {
		in._OB_skip(len);
		count_ += len;
		break;
	    }

            case tk_wchar:
            {
                in._OB_skipAlign(sizeof(wchar_t));
		in._OB_skip(len * sizeof(wchar_t));
		count_ += sizeof(wchar_t) - 1;
		count_ -= count_ % sizeof(wchar_t);
		count_ += len * sizeof(wchar_t);
		break;
	    }
	    
	    case tk_alias:
		assert(false);
		break;
		
	    default:
		for(ULong i = 0 ; i < len ; i++)
		    count_InputStream(tc -> contentType_, in);
		break;
	    }
	}
	
	break;
    }

    case tk_alias:
	count_InputStream(tc -> contentType_, in);
	break;
	
    case tk_value:
    {
        // XXX header info?
	for(ULong i = 0 ; i < tc -> memberTypes_.length() ; i++)
	    count_InputStream(tc -> memberTypes_[i], in);
	
	break;
    }
	
    case tk_value_box:
        // XXX header info?
	count_InputStream(tc -> contentType_, in);
	break;
	
    case tk_abstract_interface:
    case tk_native:
        assert(false);
        break;     // this is to keep the compiler happy
    }
}
