// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/TemplateI.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/Any.h>
#include <OB/OBV.h>
#include <OB/TCKind.h>
#include <OB/TypeCode.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/IFR4.h>
#include <OB/POAServantBase.h>
#include <OB/ValueBase.h>
#include <OB/Stream.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(TypeCode_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(TypeCode_ptr p)
{
    TypeCode::_OB_release(p);
}

// ----------------------------------------------------------------------
// Template instantiations
// ----------------------------------------------------------------------

#ifndef HAVE_NO_EXPLICIT_TEMPLATES
template class OB::ObjVar< TypeCode >;
template class OB::ObjForSeq< TypeCode >;
#else
#ifdef HAVE_PRAGMA_DEFINE
#pragma define(OB::ObjVar< TypeCode >)
#pragma define(OB::ObjForSeq< TypeCode >)
#endif
#endif

// ----------------------------------------------------------------------
// TypeCode constructor and destructor
// ----------------------------------------------------------------------

TypeCode::TypeCode()
{
}

TypeCode::~TypeCode()
{
}

// ----------------------------------------------------------------------
// TypeCode public member implementation
// ----------------------------------------------------------------------

Boolean
TypeCode::equal(TypeCode_ptr p) const
{
    if(is_nil(p))
        return false;

    if(p == this)
	return true;

    if(recType_.in() == p || p -> recType_.in() == this)
        return true;

    if(recId_.in() || p -> recId_.in())
    {
	if(!(recId_.in() && p -> recId_.in()))
	    return false;

	if(strcmp(recId_, p -> recId_) == 0)
	    return true;
	else
	    return false;
    }

    if(kind_ != p -> kind_)
	return false;

    if(kind_ == tk_objref ||
       kind_ == tk_struct ||
       kind_ == tk_union ||
       kind_ == tk_enum ||
       kind_ == tk_alias ||
       kind_ == tk_value ||
       kind_ == tk_value_box ||
       kind_ == tk_native ||
       kind_ == tk_abstract_interface ||
       kind_ == tk_except)
    {
 	if(strcmp(id_, p -> id_) != 0 ||
	   strcmp(name_, p -> name_) != 0)
	    return false;
    }
    
    if(kind_ == tk_struct ||
       kind_ == tk_union ||
       kind_ == tk_enum ||
       kind_ == tk_value ||
       kind_ == tk_except)
    {
 	if(memberNames_.length() != p -> memberNames_.length())
 	    return false;
	
 	for(ULong i = 0 ; i < memberNames_.length() ; i++)
 	    if(strcmp(memberNames_[i], p -> memberNames_[i]) != 0)
 		return false;
    }
    
    if(kind_ == tk_struct ||
       kind_ == tk_union ||
       kind_ == tk_value ||
       kind_ == tk_except)
    {
	if(memberTypes_.length() != p -> memberTypes_.length())
	    return false;
	
	for(ULong i = 0 ; i < memberTypes_.length() ; i++)
	    if(!memberTypes_[i] -> equal(p -> memberTypes_[i]))
		return false;
    }
    
    if(kind_ == tk_union)
    {
	if(labels_.length() != p -> labels_.length())
	    return false;
	
	for(ULong i = 0 ; i < labels_.length() ; i++)
	{
	    TypeCode_var ltc1 = labels_[i].type();
	    TypeCode_var ltc2 = p -> labels_[i].type();
    
	    if(!ltc1 -> equal(ltc2))
		return false;

	    const void* v1 = labels_[i].value();
	    const void* v2 = p -> labels_[i].value();

	    ltc1 = GetOrigType(ltc1);
	    switch(ltc1 -> kind())
	    {
	    case tk_short:
		if(*(Short*)v1 != *(Short*)v2)
		    return false;
		break;

	    case tk_ushort:
		if(*(UShort*)v1 != *(UShort*)v2)
		    return false;
		break;

	    case tk_long:
		if(*(Long*)v1 != *(Long*)v2)
		    return false;
		break;

	    case tk_ulong:
	    case tk_enum:
		if(*(ULong*)v1 != *(ULong*)v2)
		    return false;
		break;

	    case tk_longlong:
		if(*(LongLong*)v1 != *(LongLong*)v2)
		    return false;
		break;

	    case tk_ulonglong:
		if(*(ULongLong*)v1 != *(ULongLong*)v2)
		    return false;
		break;

	    case tk_char:
		if(*(Char*)v1 != *(Char*)v2)
		    return false;
		break;

	    case tk_wchar:
		if(*(WChar*)v1 != *(WChar*)v2)
		    return false;
		break;

	    case tk_boolean:
		if(*(Boolean*)v1 != *(Boolean*)v2)
		    return false;
		break;

	    case tk_octet:
		break;

	    default:
		assert(false);
		break;
	    }
	}
    }
    
    if(kind_ == tk_union)
    {
	if(!discriminatorType_ -> equal(p -> discriminatorType_))
	    return false;
    }
    
    if(kind_ == tk_string ||
       kind_ == tk_wstring ||
       kind_ == tk_sequence ||
       kind_ == tk_array)
    {
	if(length_ != p -> length_)
	    return false;
    }
    
    if(kind_ == tk_sequence ||
       kind_ == tk_array ||
       kind_ == tk_value_box ||
       kind_ == tk_alias)
    {
	if(!contentType_ -> equal(p -> contentType_))
	    return false;
    }

    return true;
}

Boolean
TypeCode::equivalent(TypeCode_ptr p) const
{
    if(is_nil(p))
        return false;

    if(p == this)
	return true;

    if(recType_.in() == p || p -> recType_.in() == this)
        return true;

    if(recId_.in() || p -> recId_.in())
    {
	if(!(recId_.in() && p -> recId_.in()))
	    return false;

	if(strcmp(recId_, p -> recId_) == 0)
	    return true;
	else
	    return false;
    }

    TypeCode_var tc1 = GetOrigType((TypeCode_ptr)this);
    TypeCode_var tc2 = GetOrigType(p);

    if(tc1 -> kind_ != tc2 -> kind_)
	return false;

    if(tc1 -> kind_ == tk_objref ||
       tc1 -> kind_ == tk_struct ||
       tc1 -> kind_ == tk_union ||
       tc1 -> kind_ == tk_enum ||
       tc1 -> kind_ == tk_alias ||
       tc1 -> kind_ == tk_value ||
       tc1 -> kind_ == tk_value_box ||
       tc1 -> kind_ == tk_native ||
       tc1 -> kind_ == tk_abstract_interface ||
       tc1 -> kind_ == tk_except)
    {
	if(strcmp(tc1 -> id_, "") != 0 && strcmp(tc2 -> id_, "") != 0)
	{
	    if(strcmp(tc1 -> id_, tc2 -> id_) == 0)
		return TRUE;
	    else
		return FALSE;
	}
    }
    
    // names_ and memberNames_ must be ignored
    
    if(tc1 -> kind_ == tk_struct ||
       tc1 -> kind_ == tk_union ||
       tc1 -> kind_ == tk_value ||
       tc1 -> kind_ == tk_except)
    {
	if(tc1 -> memberTypes_.length() != tc2 -> memberTypes_.length())
	    return false;
	
	for(ULong i = 0 ; i < tc1 -> memberTypes_.length() ; i++)
	    if(!(tc1 -> memberTypes_[i] ->
		 equivalent(tc2 -> memberTypes_[i])))
		return false;
    }
    
    if(tc1 -> kind_ == tk_union)
    {
	if(tc1 -> labels_.length() != tc2 -> labels_.length())
	    return false;
	
	for(ULong i = 0 ; i < tc1 -> labels_.length() ; i++)
	{
	    TypeCode_var ltc1 = tc1 -> labels_[i].type();
	    TypeCode_var ltc2 = tc2 -> labels_[i].type();
    
	    if(!ltc1 -> equivalent(ltc2))
		return false;

	    const void* v1 = tc1 -> labels_[i].value();
	    const void* v2 = tc2 -> labels_[i].value();

	    ltc1 = GetOrigType(ltc1);
	    switch(ltc1 -> kind())
	    {
	    case tk_short:
		if(*(Short*)v1 != *(Short*)v2)
		    return false;
		break;

	    case tk_ushort:
		if(*(UShort*)v1 != *(UShort*)v2)
		    return false;
		break;

	    case tk_long:
		if(*(Long*)v1 != *(Long*)v2)
		    return false;
		break;

	    case tk_ulong:
	    case tk_enum:
		if(*(ULong*)v1 != *(ULong*)v2)
		    return false;
		break;

	    case tk_longlong:
		if(*(LongLong*)v1 != *(LongLong*)v2)
		    return false;
		break;

	    case tk_ulonglong:
		if(*(ULongLong*)v1 != *(ULongLong*)v2)
		    return false;
		break;

	    case tk_char:
		if(*(Char*)v1 != *(Char*)v2)
		    return false;
		break;

	    case tk_wchar:
		if(*(WChar*)v1 != *(WChar*)v2)
		    return false;
		break;

	    case tk_boolean:
		if(*(Boolean*)v1 != *(Boolean*)v2)
		    return false;
		break;

	    case tk_octet:
		break;

	    default:
		assert(false);
		break;
	    }
	}
    }
    
    if(tc1 -> kind_ == tk_union)
    {
	if(!(tc1 -> discriminatorType_ ->
	     equivalent(tc2 -> discriminatorType_)))
	    return false;
    }
    
    if(tc1 -> kind_ == tk_string ||
       tc1 -> kind_ == tk_wstring ||
       tc1 -> kind_ == tk_sequence ||
       tc1 -> kind_ == tk_array)
    {
	if(tc1 -> length_ != tc2 -> length_)
	    return false;
    }
    
    if(tc1 -> kind_ == tk_sequence ||
       tc1 -> kind_ == tk_array ||
       tc1 -> kind_ == tk_value_box ||
       tc1 -> kind_ == tk_alias)
    {
	if(!(tc1 -> contentType_ -> equivalent(tc2 -> contentType_)))
	    return false;
    }

    return true;
}

TypeCode_ptr
TypeCode::get_compact_typecode() const
{
    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> get_compact_typecode();
    }

    assert_nca(false, NCANotImpl);
    return TypeCode::_nil();
}

TCKind
TypeCode::kind() const
{
    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> kind();
    }

    return kind_;
}

const char*
TypeCode::id() const
{
    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> id();
    }

    if(!(kind_ == tk_objref ||
	 kind_ == tk_struct ||
	 kind_ == tk_union ||
	 kind_ == tk_enum ||
	 kind_ == tk_alias ||
	 kind_ == tk_value ||
	 kind_ == tk_value_box ||
	 kind_ == tk_native ||
	 kind_ == tk_abstract_interface ||
	 kind_ == tk_except))
    {
	throw BadKind();
    }

    return id_;
}

const char*
TypeCode::name() const
{
    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> name();
    }

    if(!(kind_ == tk_objref ||
	 kind_ == tk_struct ||
	 kind_ == tk_union ||
	 kind_ == tk_enum ||
	 kind_ == tk_alias ||
	 kind_ == tk_value ||
	 kind_ == tk_value_box ||
	 kind_ == tk_native ||
	 kind_ == tk_abstract_interface ||
	 kind_ == tk_except))
    {
	throw BadKind();
    }
    
    return name_;
}

ULong
TypeCode::member_count() const
{
    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> member_count();
    }

    if(!(kind_ == tk_struct ||
	 kind_ == tk_union ||
	 kind_ == tk_enum ||
	 kind_ == tk_value ||
	 kind_ == tk_except))
    {
	throw BadKind();
    }
    
    return memberNames_.length();
}

const char*
TypeCode::member_name(ULong index) const
{
    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> member_name(index);
    }

    if(!(kind_ == tk_struct ||
	 kind_ == tk_union ||
	 kind_ == tk_enum ||
	 kind_ == tk_value ||
	 kind_ == tk_except))
    {
	throw BadKind();
    }

    if(index >= memberNames_.length())
    {
	throw Bounds();
    }

    return memberNames_[index];
}

TypeCode_ptr
TypeCode::member_type(ULong index) const
{
    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> member_type(index);
    }

    if(!(kind_ == tk_struct ||
	 kind_ == tk_union ||
	 kind_ == tk_value ||
	 kind_ == tk_except))
    {
	throw BadKind();
    }

    if(index >= memberNames_.length())
    {
	throw Bounds();
    }

    return TypeCode::_duplicate(memberTypes_[index]);
}

Any*
TypeCode::member_label(ULong index) const
{
    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> member_label(index);
    }

    if(!(kind_ == tk_union))
    {
	throw BadKind();
    }

    if(index >= memberNames_.length())
    {
	throw Bounds();
    }

    return new Any(labels_[index]);
}

TypeCode_ptr
TypeCode::discriminator_type() const
{
    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> discriminator_type();
    }

    if(!(kind_ == tk_union))
    {
	throw BadKind();
    }

    return TypeCode::_duplicate(discriminatorType_);
}

Long
TypeCode::default_index() const
{
    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> default_index();
    }

    if(!(kind_ == tk_union))
    {
	throw BadKind();
    }

    for(Long i = 0 ; i < (Long)labels_.length() ; i++)
    {
	TypeCode_var tc = labels_[i].type();
	if(tc -> kind() == tk_octet)
	    return i;
    }

    return -1;
}

ULong
TypeCode::length() const
{
    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> length();
    }

    if(!(kind_ == tk_string ||
	 kind_ == tk_wstring ||
	 kind_ == tk_sequence ||
	 kind_ == tk_array))
    {
	throw BadKind();
    }

    return length_;
}

TypeCode_ptr
TypeCode::content_type() const
{
    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> content_type();
    }

    if(!(kind_ == tk_sequence ||
	 kind_ == tk_array ||
	 kind_ == tk_value_box ||
	 kind_ == tk_alias))
    {
	throw BadKind();
    }

    return TypeCode::_duplicate(contentType_);
}

UShort
TypeCode::fixed_digits() const
{
    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> fixed_digits();
    }

    if(!(kind_ == tk_fixed))
    {
	throw BadKind();
    }

    return fixedDigits_;
}

Short
TypeCode::fixed_scale() const
{
    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> fixed_scale();
    }

    if(!(kind_ == tk_fixed))
    {
	throw BadKind();
    }

    return fixedScale_;
}

Visibility
TypeCode::member_visibility(ULong index) const
{
    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> member_visibility(index);
    }

    if(!(kind_ == tk_value))
    {
	throw BadKind();
    }

    if(index >= memberVisibility_.length())
    {
	throw Bounds();
    }

    return memberVisibility_[index];
}

ValueModifier
TypeCode::type_modifier() const
{
    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> type_modifier();
    }

    if(!(kind_ == tk_value))
    {
	throw BadKind();
    }

    return typeModifier_;
}

TypeCode_ptr
TypeCode::concrete_base_type() const
{
    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> concrete_base_type();
    }

    if(!(kind_ == tk_value))
    {
	throw BadKind();
    }

    return TypeCode::_duplicate(concreteBaseType_);
}

bool
TypeCode::_OB_isSystemException() const
{
    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> _OB_isSystemException();
    }

    if(kind_ != tk_except)
	return false;

    if(strcmp(id_, "IDL:omg.org/CORBA/UNKNOWN:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/BAD_PARAM:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/NO_MEMORY:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/IMP_LIMIT:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/COMM_FAILURE:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/INV_OBJREF:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/NO_PERMISSION:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/INTERNAL:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/MARSHAL:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/INITIALIZE:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/NO_IMPLEMENT:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/BAD_TYPECODE:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/BAD_OPERATION:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/NO_RESOURCES:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/NO_RESPONSE:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/PERSIST_STORE:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/BAD_INV_ORDER:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/TRANSIENT:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/FREE_MEM:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/INV_IDENT:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/INV_FLAG:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/INTF_REPOS:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/BAD_CONTEXT:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/OBJ_ADAPTER:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/DATA_CONVERSION:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/OBJECT_NOT_EXIST:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/TRANSACTION_REQUIRED:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/TRANSACTION_ROLLEDBACK:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/INVALID_TRANSACTION:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/INV_POLICY:1.0") == 0)
    {
	return true;
    }
    else
    {
	return false;
    }
}

// ----------------------------------------------------------------------
// Safely decrement reference count
// ----------------------------------------------------------------------

void
TypeCode::_OB_release(TypeCode_ptr p)
{
    if(p)
    {
/* TODO:
	if(p -> _OB_getRef() > 1)
	{
	    //
	    // Find out whether this is a non-reachable TypeCode,
	    // ignoring the reference that still has to be released.
	    //
	    if(_OB_checkForNonReachableTC(p, 1))
	    {
		p -> contentType_ = TypeCode::_nil();
		p -> memberTypes_.length(0);
		assert(p -> _OB_getRef() == 1);
	    }
	}
*/
	
	p -> _OB_decRef();
    }
}

// ----------------------------------------------------------------------
// Check whether a TypeCode is not reachable
// ----------------------------------------------------------------------

bool
TypeCode::_OB_checkForNonReachableTC(TypeCode_ptr tc, ULong ignore)
{
    //
    // Special handling for recursive placeholder TypeCodes
    //
    if(tc -> recId_.in())
    {
	if(tc -> _OB_getRef() - ignore > 1)
	    return false;
	else
	    return true;
    }
    
    if(tc -> kind_ == tk_struct ||
       tc -> kind_ == tk_except ||
       tc -> kind_ == tk_union ||
       tc -> kind_ == tk_value)
    {
	//
	// The number of cyclic references can never be greater than
	// the number of references
	//
	assert(tc -> _OB_getRef() - ignore >= tc -> cyclic_);

	//
	// If there are any non-cyclic references, the TypeCode is
	// reachable
	//
	if(tc -> _OB_getRef() - ignore > tc -> cyclic_)
	    return false;
    }
	
    //
    // Check content type
    //
    if(tc -> kind_ == tk_sequence ||
       tc -> kind_ == tk_value_box ||
       tc -> kind_ == tk_array ||
       tc -> kind_ == tk_alias)
    {
	if(!_OB_checkForNonReachableTC(tc -> contentType_, 0))
	    return false;
    }
    
    //
    // Check member types
    //
    if(tc -> kind_ == tk_struct ||
       tc -> kind_ == tk_union ||
       tc -> kind_ == tk_value ||
       tc -> kind_ == tk_except)
    {
	for(ULong i = 0 ; i < tc -> memberTypes_.length() ; i++)
	    if(!_OB_checkForNonReachableTC(tc -> memberTypes_[i], 0))
		return false;
    }

    //
    // TypeCode is not reachable
    //
    return true;
}

// ----------------------------------------------------------------------
// Embed recursive placeholder TypeCodes
// ----------------------------------------------------------------------

void
TypeCode::_OB_embedRecTC(TypeCode_ptr outer)
{
    //
    // Recursive placeholder TypeCodes are illegal as "outer" argument
    //
    assert(!outer -> recId_.in());

    //
    // Check for illegal recursion
    //
    if(!(outer -> kind_ == tk_struct ||
	 outer -> kind_ == tk_except ||
	 outer -> kind_ == tk_union ||
	 outer -> kind_ == tk_value))
    {
	throw BAD_TYPECODE();
    }

    outer -> cyclic_ = 0;
    _OB_embedRecTC(outer, outer);
}

void
TypeCode::_OB_embedRecTC(TypeCode_ptr outer, TypeCode_ptr inner)
{
    //
    // Embed recursive placeholder TypeCodes
    //
    if(inner -> recId_.in())
    {
	if(strcmp(inner -> recId_, outer -> id_) == 0)
	{
	    if(inner -> recType_.in())
	    {
		assert_nca(inner -> recType_.in() == outer, NCAEmbeddedRecTC);
	    }
	    else
	    {
		//
		// Embed the recursive placeholder TypeCode
		//
		inner -> recType_ = TypeCode::_duplicate(outer);

		//
		// Update cyclic reference count
		// 
		outer -> cyclic_++;
	    }
	}
    }
    else
    {
	//
	// Embed content type
	//
	if(inner -> kind_ == tk_sequence ||
	   inner -> kind_ == tk_value_box ||
	   inner -> kind_ == tk_array ||
	   inner -> kind_ == tk_alias)
	{
	    assert(inner -> contentType_.in() != outer);
	    _OB_embedRecTC(outer, inner -> contentType_);
	}
    
	//
	// Embed member types
	//
	if(inner -> kind_ == tk_struct ||
	   inner -> kind_ == tk_union ||
	   inner -> kind_ == tk_value ||
	   inner -> kind_ == tk_except)
	{
	    for(ULong i = 0 ; i < inner -> memberTypes_.length() ; i++)
	    {
		assert(inner -> memberTypes_[i].in() != outer);
		_OB_embedRecTC(outer, inner -> memberTypes_[i]);
	    }
	}
    }
}

// ----------------------------------------------------------------------
// Resolve recursive TypeCodes
// ----------------------------------------------------------------------

void
TypeCode::_OB_resolveRecTC(TypeCode_ptr outer)
{
    //
    // Recursive placeholder TypeCodes are illegal as "outer" argument
    //
    assert(!outer -> recId_.in());

    //
    // Check for illegal recursion
    //
    if(!(outer -> kind_ == tk_struct ||
	 outer -> kind_ == tk_except ||
	 outer -> kind_ == tk_union ||
	 outer -> kind_ == tk_value))
    {
	throw BAD_TYPECODE();
    }

    outer -> cyclic_ = 0;
    _OB_resolveRecTC(outer, outer);
}

void
TypeCode::_OB_resolveRecTC(TypeCode_ptr outer, TypeCode_ptr inner)
{
    //
    // Ignore recursive placeholder TypeCodes
    //
    if(inner -> recId_.in())
	return;

    //
    // Check for everything having a content type
    //
    if(inner -> kind_ == tk_sequence ||
       inner -> kind_ == tk_value_box ||
       inner -> kind_ == tk_array ||
       inner -> kind_ == tk_alias)
    {
	if(inner -> contentType_.in() == outer)
	{
	    //
	    // Check for illegal recursion
	    //
	    if(outer -> kind_ == tk_struct ||
	       outer -> kind_ == tk_except ||
	       outer -> kind_ == tk_union)
	    {
		if(inner -> kind_ != tk_sequence)
		    throw BAD_TYPECODE();
	    }
	    else
	    {
		assert(outer -> kind_ == tk_value);
		throw BAD_TYPECODE();
	    }

	    //
	    // Create and embed recursive placeholder TypeCode
	    //
	    TypeCode_ptr tc = new TypeCode;
	    tc -> recId_ = outer -> id_;
	    tc -> recType_ = TypeCode::_duplicate(outer);
	    inner -> contentType_ =  tc;
	    
	    //
	    // Update cyclic reference count
	    // 
	    outer -> cyclic_++;
	}
	else
	    _OB_resolveRecTC(outer, inner -> contentType_);
    }
    
    //
    // Check for everything having member types
    //
    if(inner -> kind_ == tk_struct ||
       inner -> kind_ == tk_union ||
       inner -> kind_ == tk_value ||
       inner -> kind_ == tk_except)
    {
	for(ULong i = 0 ; i < inner -> memberTypes_.length() ; i++)
	{
	    if(inner -> memberTypes_[i].in() == outer)
	    {
		//
		// Check for illegal recursion
		//
		if(outer -> kind_ == tk_struct ||
		   outer -> kind_ == tk_except ||
		   outer -> kind_ == tk_union)
		{
		    throw BAD_TYPECODE();
		}
		else
		{
		    assert(outer -> kind_ == tk_value);

		    if(inner -> kind_ != tk_value)
			throw BAD_TYPECODE();
		}
		
		//
		// Create and embed recursive placeholder TypeCode
		//
		TypeCode_ptr tc = new TypeCode;
		tc -> recId_ = outer -> id_;
		tc -> recType_ = TypeCode::_duplicate(outer);
		inner -> memberTypes_[i] =  tc;
		
		//
		// Update cyclic reference count
		// 
		outer -> cyclic_++;
	    }
	    else
		_OB_resolveRecTC(outer, inner -> memberTypes_[i]);
	}
    }
}

// ----------------------------------------------------------------------
// Classes for _tc_<type> constant
// ----------------------------------------------------------------------

TypeCodeSimpleConst::TypeCodeSimpleConst(TCKind k)
    : typeCode_(TypeCode::_nil()), kind_(k)
{
    assert(kind_ == tk_null ||
	   kind_ == tk_void ||
	   kind_ == tk_short ||
	   kind_ == tk_long ||
	   kind_ == tk_longlong ||
	   kind_ == tk_ushort ||
	   kind_ == tk_ulong ||
	   kind_ == tk_ulonglong ||
	   kind_ == tk_float ||
	   kind_ == tk_double ||
	   kind_ == tk_longdouble ||
	   kind_ == tk_boolean ||
	   kind_ == tk_char ||
	   kind_ == tk_wchar ||
	   kind_ == tk_octet ||
	   kind_ == tk_any ||
	   kind_ == tk_TypeCode ||
	   kind_ == tk_Principal);
};

TypeCodeSimpleConst::operator TypeCode_ptr()
{
    if(is_nil(typeCode_))
    {
	typeCode_ = new TypeCode;
	typeCode_ -> kind_ = kind_;
    }

    return typeCode_; // No _duplicate() here !
}

TypeCodeStringConst::TypeCodeStringConst(bool wide)
    : typeCode_(TypeCode::_nil()), wide_(wide)
{
};

TypeCodeStringConst::operator TypeCode_ptr()
{
    if(is_nil(typeCode_))
    {
	typeCode_ = new TypeCode;

	if(wide_)
	    typeCode_ -> kind_ = tk_wstring;
	else
	    typeCode_ -> kind_ = tk_string;

	typeCode_ -> length_ = 0;
    }

    return typeCode_; // No _duplicate() here !
}

TypeCode_ptr
TypeCodeStringConst::bounded(ULong bound)
{
    TypeCode_var tc = new TypeCode;

    if(wide_)
	tc -> kind_ = tk_wstring;
    else
	tc -> kind_ = tk_string;

    tc -> length_ = bound;

    return tc._retn();
}

TypeCodeFixedConst::TypeCodeFixedConst()
    : typeCode_(TypeCode::_nil())
{
};

TypeCodeFixedConst::operator TypeCode_ptr()
{
    if(is_nil(typeCode_))
    {
	typeCode_ = new TypeCode;

	typeCode_ -> kind_ = tk_fixed;

	typeCode_ -> fixedDigits_ = 0;
	typeCode_ -> fixedScale_ = 0;
    }

    return typeCode_; // No _duplicate() here !
}

TypeCode_ptr
TypeCodeFixedConst::fixed(UShort digits, UShort scale)
{
    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_fixed;

    tc -> fixedDigits_ = digits;
    tc -> fixedScale_ = scale;

    return tc._retn();
}

TypeCodeObjectConst::TypeCodeObjectConst()
    : typeCode_(TypeCode::_nil())
{
};

TypeCodeObjectConst::operator TypeCode_ptr()
{
    if(is_nil(typeCode_))
    {
	typeCode_ = new TypeCode;
	typeCode_ -> kind_ = tk_objref;
	typeCode_ -> id_ = string_dup("IDL:omg.org/CORBA/Object:1.0");
	typeCode_ -> name_ = string_dup("Object");
    }

    return typeCode_; // No _duplicate() here !
}

TypeCodeSysExConst::TypeCodeSysExConst(const char* name)
    : name_(name)
{
};

TypeCodeSysExConst::operator TypeCode_ptr()
{
    if(is_nil(typeCode_))
    {
	typeCode_ = new TypeCode;
	typeCode_ -> kind_ = tk_except;
	typeCode_ -> id_ = string_dup("IDL:omg.org/CORBA/");
	typeCode_ -> id_ += name_;
	typeCode_ -> id_ += ":1.0";
	typeCode_ -> name_ = string_dup(name_);
	typeCode_ -> memberNames_.length(2);
	typeCode_ -> memberNames_[0] = string_dup("minor");
	typeCode_ -> memberNames_[1] = string_dup("completed");
	typeCode_ -> memberTypes_.length(2);
	typeCode_ -> memberTypes_[0] = TypeCode::_duplicate(_tc_ulong);
	TypeCode_var et = new TypeCode;
	et -> kind_ = tk_enum;
	et -> id_ = string_dup("IDL:omg.org/CORBA/CompletionStatus:1.0");
	et -> name_ = string_dup("CompletionStatus");
	et -> memberNames_.length(3);
	et -> memberNames_[0] = string_dup("COMPLETED_YES");
	et -> memberNames_[1] = string_dup("COMPLETED_NO");
	et -> memberNames_[2] = string_dup("COMPLETED_MAYBE");
	typeCode_ -> memberTypes_[1] = et;
    }

    return typeCode_; // No _duplicate() here !
}

TypeCodeConst::operator TypeCode_ptr()
{
    if(is_nil(typeCode_))
    {
	assert(str_ != 0);
	Octet* oct = AsciiToOctets(str_);
        InputStream in(oct, 0, strlen(str_) / 2, false);
	in._OB_readEndian();
	typeCode_ = in.read_TypeCode();
	delete [] oct;
    }
    
    return typeCode_; // No _duplicate() here !
}

namespace CORBA
{

TypeCodeSimpleConst _tc_null(tk_null);
TypeCodeSimpleConst _tc_void(tk_void);
TypeCodeSimpleConst _tc_short(tk_short);
TypeCodeSimpleConst _tc_long(tk_long);
TypeCodeSimpleConst _tc_longlong(tk_longlong);
TypeCodeSimpleConst _tc_ushort(tk_ushort);
TypeCodeSimpleConst _tc_ulong(tk_ulong);
TypeCodeSimpleConst _tc_ulonglong(tk_ulonglong);
TypeCodeSimpleConst _tc_float(tk_float);
TypeCodeSimpleConst _tc_double(tk_double);
TypeCodeSimpleConst _tc_longdouble(tk_longdouble);
TypeCodeSimpleConst _tc_boolean(tk_boolean);
TypeCodeSimpleConst _tc_char(tk_char);
TypeCodeSimpleConst _tc_wchar(tk_wchar);
TypeCodeSimpleConst _tc_octet(tk_octet);
TypeCodeSimpleConst _tc_any(tk_any);
TypeCodeSimpleConst _tc_TypeCode(tk_TypeCode);
TypeCodeSimpleConst _tc_Principal(tk_Principal);

TypeCodeStringConst _tc_string(false);
TypeCodeStringConst _tc_wstring(true);

TypeCodeFixedConst _tc_fixed;

TypeCodeObjectConst _tc_Object;

TypeCodeSysExConst _tc_UNKNOWN("UNKNOWN");
TypeCodeSysExConst _tc_BAD_PARAM("BAD_PARAM");
TypeCodeSysExConst _tc_NO_MEMORY("NO_MEMORY");
TypeCodeSysExConst _tc_IMP_LIMIT("IMP_LIMIT");
TypeCodeSysExConst _tc_COMM_FAILURE("COMM_FAILURE");
TypeCodeSysExConst _tc_INV_OBJREF("INV_OBJREF");
TypeCodeSysExConst _tc_NO_PERMISSION("NO_PERMISSION");
TypeCodeSysExConst _tc_INTERNAL("INTERNAL");
TypeCodeSysExConst _tc_MARSHAL("MARSHAL");
TypeCodeSysExConst _tc_INITIALIZE("INITIALIZE");
TypeCodeSysExConst _tc_NO_IMPLEMENT("NO_IMPLEMENT");
TypeCodeSysExConst _tc_BAD_TYPECODE("BAD_TYPECODE");
TypeCodeSysExConst _tc_BAD_OPERATION("BAD_OPERATION");
TypeCodeSysExConst _tc_NO_RESOURCES("NO_RESOURCES");
TypeCodeSysExConst _tc_NO_RESPONSE("NO_RESPONSE");
TypeCodeSysExConst _tc_PERSIST_STORE("PERSIST_STORE");
TypeCodeSysExConst _tc_BAD_INV_ORDER("BAD_INV_ORDER");
TypeCodeSysExConst _tc_TRANSIENT("TRANSIENT");
TypeCodeSysExConst _tc_FREE_MEM("FREE_MEM");
TypeCodeSysExConst _tc_INV_IDENT("INV_IDENT");
TypeCodeSysExConst _tc_INV_FLAG("INV_FLAG");
TypeCodeSysExConst _tc_INTF_REPOS("INTF_REPOS");
TypeCodeSysExConst _tc_BAD_CONTEXT("BAD_CONTEXT");
TypeCodeSysExConst _tc_OBJ_ADAPTER("OBJ_ADAPTER");
TypeCodeSysExConst _tc_DATA_CONVERSION("DATA_CONVERSION");
TypeCodeSysExConst _tc_OBJECT_NOT_EXIST("OBJECT_NOT_EXIST");
TypeCodeSysExConst _tc_TRANSACTION_REQUIRED("TRANSACTION_REQUIRED");
TypeCodeSysExConst _tc_TRANSACTION_ROLLEDBACK("TRANSACTION_ROLLEDBACK");
TypeCodeSysExConst _tc_INVALID_TRANSACTION("INVALID_TRANSACTION");
TypeCodeSysExConst _tc_INV_POLICY("INV_POLICY");

} // End of namespace CORBA

// ----------------------------------------------------------------------
// Get original type, i.e. remove aliases
// ----------------------------------------------------------------------

TypeCode_ptr
OB::GetOrigType(TypeCode_ptr p)
{
    TypeCode_var tc = TypeCode::_duplicate(p);

    while(tc -> kind() == tk_alias)
	tc = tc -> content_type();

    return tc._retn();
}
