// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <X11/Intrinsic.h>

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/X11.h>
#include <OB/Reactor.h>
#include <OB/Net.h>

#include <X11Reactor.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// X11Reactor input callback procedures
// ----------------------------------------------------------------------

static void
ReadInputCB(XtPointer clientData, int*, XtInputId*)
{
    EventHandler* handler = (EventHandler*)clientData;
    handler -> handleEvent(EventRead);
}

static void
WriteInputCB(XtPointer clientData, int*, XtInputId*)
{
    EventHandler* handler = (EventHandler*)clientData;
    handler -> handleEvent(EventWrite);
}

static void
TimeoutCB(XtPointer clientData, XtIntervalId*)
{
    *(bool*)clientData = true;
}

// ----------------------------------------------------------------------
// X11Reactor constructor and destructor
// ----------------------------------------------------------------------

X11Reactor::X11Reactor(XtAppContext context)
    : stop_(false), context_(context), handlerInfoList_(0)
{
}

X11Reactor::~X11Reactor()
{
    //
    // Delete all handler infos
    //
    while(handlerInfoList_)
    {
	HandlerInfo* next = handlerInfoList_ -> next;
	delete handlerInfoList_;
	handlerInfoList_ = next;
    }
}

// ----------------------------------------------------------------------
// X11Reactor public member implementation
// ----------------------------------------------------------------------

Reactor*
X11Reactor::instance(XtAppContext context)
{
    if(!instance_)
	instance_ = new X11Reactor(context);

    return instance_;
}

void
X11Reactor::registerHandler(EventHandler* handler, Mask mask, Handle handle)
{
    assert(handler);

    assert(handler);

    HandlerInfo* info = handlerInfoList_;
    while(info)
    {
	if(info -> handler == handler)
	{
	    //
	    // Handler already in list, change inputs if necessary and
	    // update mask and handle
	    //
	    
	    if((info -> mask & EventRead) != (mask & EventRead) ||
	       info -> handle != handle)
	    {
		if(info -> mask & EventRead)
		{
		    XtRemoveInput(info -> readId);
		}

		if(mask & EventRead)
		{
		    info -> readId =
			XtAppAddInput(context_,
				      handle,
				      (XtPointer)XtInputReadMask,
				      ReadInputCB,
				      (XtPointer)info -> handler);
		}
	    }
	    
	    if((info -> mask & EventWrite) != (mask & EventWrite) ||
	       info -> handle != handle)
	    {
		if(info -> mask & EventWrite)
		{
		    XtRemoveInput(info -> writeId);
		}
		
		if(mask & EventWrite)
		{
		    info -> writeId =
			XtAppAddInput(context_,
				      handle,
				      (XtPointer)XtInputWriteMask,
				      WriteInputCB,
				      (XtPointer)info -> handler);
		}
	    }

	    info -> mask = mask;
	    info -> handle = handle;
	    return;
	}
	
	info = info -> next;
    }

    //
    // Add new handler info
    //
    info = new HandlerInfo;
    info -> handler = handler;
    info -> mask = mask;
    info -> handle = handle;
    info -> next = handlerInfoList_;
    handlerInfoList_ = info;

    //
    // Add inputs
    //
    if(mask & EventRead)
	info -> readId = XtAppAddInput(context_,
				    info -> handle,
				    (XtPointer)XtInputReadMask,
				    ReadInputCB,
				    (XtPointer)info -> handler);
    
    if(mask & EventWrite)
	info -> writeId = XtAppAddInput(context_,
				     info -> handle,
				     (XtPointer)XtInputWriteMask,
				     WriteInputCB,
				     (XtPointer)info -> handler);
}

void
X11Reactor::unregisterHandler(EventHandler* handler)
{
    assert(handler);

    HandlerInfo** p = &handlerInfoList_;
    while(*p)
    {
	HandlerInfo* info = *p;

	if(info -> handler == handler)
	{
	    //
	    // Handler info found, remove inputs if necessary, and
	    // invalidate the handler
	    //

	    if(info -> mask & EventRead)
		XtRemoveInput(info -> readId);
	    
	    if(info -> mask & EventWrite)
		XtRemoveInput(info -> writeId);
	    
	    *p = info -> next;
	    delete info;
	    return;
	}
	
	p = &(info -> next);
    }
	
    assert(false);
}

void
X11Reactor::dispatch()
{
    stop_ = false;

    while(!stop_ && handlerInfoList_)
    {
	//
	// Process one X11 event
	//
	XtAppProcessEvent(context_, XtIMAll);
    }
}

bool
X11Reactor::dispatchOneEvent(Long timeout)
{
    if(stop_ || !handlerInfoList_)
	return true;

    if(timeout <= 0)
    {
	if(timeout == 0)
	{
	    //
	    // Handle zero timeouts
	    //
	    if(XtAppPending(context_) == 0)
	    {
		return false;
	    }
	}
	    
	//
	// Process one X11 event
	//
	XtAppProcessEvent(context_, XtIMAll);

	//
	// No timeout - always return true
	//
	return true;
    }
    else
    {
	//
	// Set timeout callback
	//
	bool b = false;
	XtIntervalId id = XtAppAddTimeOut(context_, timeout, TimeoutCB, &b);

	//
	// Process one X11 event
	//
	XtAppProcessEvent(context_, XtIMAll);

	if(b)
	{
	    //
	    // There was a timeout
	    // Timeout callback has been removed automatically
	    //
	    return false;
	}
	else
	{
	    //
	    // Remove timeout callback
	    //
	    XtRemoveTimeOut(id);

	    //
	    // Return true - no timeout
	    //
	    return true;
	}
    }
}

bool
X11Reactor::eventReady()
{
    return XtAppPending(context_);
}

// ----------------------------------------------------------------------
// Initialization functions
// ----------------------------------------------------------------------

void
OB::X11Init(XtAppContext appContext)
{
    //
    // This will choose X11Reactor as reactor singleton
    //
    X11Reactor::instance(appContext);
}
