// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef GEN_CPP_H
#define GEN_CPP_H

//
// Declare classes
//
class IdlPrettyPrint;

//
// The C++ code generator class
//
class IdlCPPGenerator
{
    //
    // The header
    //
    static const char* header_;

    //
    // The program name
    //
    const char* prog_;

    //
    // The ORB
    //
    CORBA::ORB_var orb_;
    
    //
    // The repository
    //
    CORBA::Repository_var repository_;

    //
    // If partial_ is set to true, code is only generated for
    // everything that's in idSeq_;
    //
    bool partial_;
    CORBA::RepositoryIdSeq idSeq_;
    IdlStringSeq includeSeq_;

    //
    // If true, don't generate TypeCodes and Any insertion/extraction
    //
    bool noTC_;

private:
    bool local_;

    //
    // If true, don't use virtual inheritance
    //
    bool noVirtual_;

    //
    // If true, then we're not generating skeletons
    //
    bool noSkel_;
    
    //
    // If set, make include statements for header files for current
    // IDL relative to this path
    //
    CORBA::String_var thisHeaderDir_;

    //
    // If set, make include statements for header files for included
    // IDL files relative to this path
    //
    CORBA::String_var otherHeaderDir_;

    //
    // If set, put this statement in front of every symbol that needs
    // an explicit DLL import
    //
    CORBA::String_var dllImport_;

    //
    // Suffix for header and source files
    //
    CORBA::String_var hSuff_;
    CORBA::String_var cSuff_;

    //
    // Cache for absolute names and for unique type names
    //
    typedef OB::Hashtable< CORBA::String_var, CORBA::String_var > StringMap;
    StringMap* absMap_;
    StringMap* uniMap_;
    
    //
    // Print header comment
    //
    void comment(const char*, IdlPrettyPrint&);

    //
    // Precede C++ keywords by an underscore
    //
    char* fixKwd(const char*);

    //
    // Create type string
    //
    enum GetType
    {
	GetTypeNormal,  // Normal type string
	GetTypePtr,     // _ptr type string
	GetTypeRef,     // Ref (== _ptr) type string
	GetTypeVar,     // _var type string
	GetTypeStruct,  // Type string for use in structs
	GetTypeArray,   // Type string for use in arrays
	GetTypeSlice,   // Type string for use as array slice
	GetTypeForAny,  // Type string for use as array for any insert/extract
	GetTypeIn,      // Type string for use as in parameter
	GetTypeOut,     // Type string for use as out parameter
	GetTypeInOut,   // Type string for use as inout parameter
	GetTypeReturn,  // Type string for use as return value
	GetTypeConst    // Type string for use in constant declarations
    };
    
    char* getTypeString(const char*, CORBA::TypeCode_ptr,
			GetType = GetTypeNormal,
			bool = false, const char* = 0, const char* = 0);

    //
    // Get array type and total number of elements
    //
    void getArrayType(CORBA::TypeCode_ptr, CORBA::TypeCode_ptr&, CORBA::ULong&);

    //
    // Determine if storage class for TypeCode is extern
    //
    bool isTypeCodeExtern(CORBA::Container_ptr);

    //
    // Determine if content buffer is needed for a sequence or array
    // type in order to count its size
    //
    bool needDataForCount(CORBA::TypeCode_ptr);

    //
    // Write read/write/count code
    //
    enum RWC
    {
	RWCReadReturnAlloc,
	RWCReadReturnDirect,
	RWCReadOut,
	RWCReadInOut,
	RWCReadStructElem,
	RWCCountIn,
	RWCCountInOut,
	RWCWriteIn,
	RWCWriteInOut
    };

    void writeReadWriteCount(const char*, CORBA::TypeCode_ptr, const char*,
			     CORBA::ULong&, RWC, const char*, IdlPrettyPrint&);

    //
    // Write array copy code
    //
    void writeArrayCopy(const char*, CORBA::TypeCode_ptr, const char*,
                        const char*, const char*, IdlPrettyPrint&);

    //
    // Write array allocation code
    //
    void writeArrayAlloc(const char*, CORBA::TypeCode_ptr, const char*,
                         IdlPrettyPrint&);

    //
    // Write array duplication code
    //
    void writeArrayDup(const char*, CORBA::TypeCode_ptr, const char*,
                       const char*, const char*, IdlPrettyPrint&);

    //
    // Write array destruction code
    //
    void writeArrayFree(const char*, CORBA::TypeCode_ptr, const char*,
                        IdlPrettyPrint&);

    //
    // Write value member unmarshalling code
    //
    void writeValueMemberUnmarshal(const char*, const char*, const char*,
                                   const char*, CORBA::TypeCode_ptr,
                                   CORBA::ULong&, IdlPrettyPrint&);

    //
    // Check if code needs to be generated for a specific repository id
    //
    bool check(const char*);

    //
    // Get absolute name; if scope is provided, it will be removed from
    // the result
    //
    char* getAbsolute(CORBA::Contained_ptr, const char* = 0);
    char* getAbsolute(const char*, const char* = 0);

    //
    // Get unique type names (for sequence templates)
    //
    char* getUnique(CORBA::Contained_ptr);
    char* getUnique(const char*);

    //
    // Get scope
    //
    char* getScope(CORBA::Container_ptr);

    //
    // Determines if the object resides at the top level of the IR
    //
    bool isTopLevel(CORBA::Contained_ptr);

    //
    // Get prefixed name.  There are two prefixes, a top prefix
    // and an optional name prefix. The top prefix is only prepended
    // if the Contained object resides at the top level. The name
    // prefix, if provided, is always prepended to the "fixed" name
    // of the Contained object.
    //
    char* getPrefixedName(const char*, CORBA::Contained_ptr, const char* = 0);

    //
    // Get a prefixed absolute name. There are two prefixes, a top prefix
    // and an optional name prefix. The top prefix is always prepended
    // to the absolute name of the object. The name prefix, if provided,
    // is prepended to the last component of the absolute name.
    // The prefixed scope is removed if possible.
    //
    char* getPrefixedAbsolute(const char*, const char*, const char*,
                              const char* = 0);
    char* getPrefixedAbsolute(const char*, CORBA::Contained_ptr, const char*,
                              const char* = 0);

    //
    // Converts a scoped name into a flattened name, replacing all
    // instances of "::" with "_"
    //
    char* getFlattenedName(const char*);

    //
    // Print union labels
    //
    void printUnionLabels(const IdlUnionMemberInfo&, IdlPrettyPrint&,
			  const char*);

    //
    // Generate code for the header file, forward declarations
    //
    void genCPlusPlusHFwd(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate code for the header file, global function forward declarations
    //
    void genCPlusPlusHGlobalFwd(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate code for the header file, _var forward declarations
    //
    void genCPlusPlusHVarFwd(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate code for the header file, StubImpl forward declarations
    //
    void genCPlusPlusHStubImplFwd(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate code for the header file, StubImpl global functions
    //
    void genCPlusPlusHStubImplGlobalFwd(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate code for the header file, StubImpl _var declarations
    //
    void genCPlusPlusHStubImplVarFwd(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate code for the header file, general part
    //
    void genCPlusPlusH(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate code for the header file, collocated part
    //
    void genCPlusPlusHStubImpl(CORBA::Container_ptr, IdlPrettyPrint&);
    
    //
    // Generate code for the header file, global namespace part
    //
    void genCPlusPlusHGlobalNamespace(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate OBV_* classes for the header file, general part
    //
    void genCPlusPlusHOBV(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate code for a value member
    //
    void genCPlusPlusHValueMember(CORBA::Container_ptr, CORBA::ValueMember*,
				  const char*, bool, IdlPrettyPrint&);

    //
    // Generate type-specific value factory class
    //
    void genCPlusPlusHValueFactory(CORBA::ValueDef_ptr, IdlPrettyPrint&);

    //
    // Generate operation
    //
    void genCPlusPlusHOperation(CORBA::Container::Description&,
                                const char*, bool, bool, IdlPrettyPrint&);

    //
    // Generate attribute
    //
    void genCPlusPlusHAttribute(CORBA::Container::Description&,
                                const char*, bool, bool, IdlPrettyPrint&);

    //
    // Generate code for valuetype header file, general part
    //
    void genCPlusPlusHValueMembers(CORBA::Container_ptr,
                                   CORBA::Visibility,
                                   IdlPrettyPrint&);

    //
    // Generate basic value box
    //
    void genCPlusPlusHBasicBox(CORBA::ValueBoxDef_ptr, const char*,
                               const char*, CORBA::TypeDescription*,
                               CORBA::TypeCode_ptr, IdlPrettyPrint&);

    //
    // Generate struct value box
    //
    void genCPlusPlusHStructBox(CORBA::ValueBoxDef_ptr, const char*,
                                const char*, CORBA::TypeDescription*,
                                CORBA::TypeCode_ptr, IdlPrettyPrint&);

    //
    // Generate code for a struct value box member
    //
    void genCPlusPlusHStructBoxMember(const char*, const CORBA::StructMember&,
                                      const char*, IdlPrettyPrint&);

    //
    // Generate string/wstring value box
    //
    void genCPlusPlusHStringBox(CORBA::ValueBoxDef_ptr, const char*,
                                const char*, CORBA::TypeDescription*,
                                CORBA::TypeCode_ptr, IdlPrettyPrint&);

    //
    // Generate union value box
    //
    void genCPlusPlusHUnionBox(CORBA::ValueBoxDef_ptr, const char*,
                               const char*, CORBA::TypeDescription*,
                               CORBA::TypeCode_ptr, IdlPrettyPrint&);

    //
    // Generate code for a union value box member
    //
    void genCPlusPlusHUnionBoxMember(const char*, const IdlUnionMemberInfo&,
                                     const char*, IdlPrettyPrint&);

    //
    // Generate sequence value box
    //
    void genCPlusPlusHSequenceBox(CORBA::ValueBoxDef_ptr, const char*,
                                  const char*, CORBA::TypeDescription*,
                                  CORBA::TypeCode_ptr, IdlPrettyPrint&);

    //
    // Generate any value box
    //
    void genCPlusPlusHAnyBox(CORBA::ValueBoxDef_ptr, const char*,
                             const char*, CORBA::TypeDescription*,
                             CORBA::TypeCode_ptr, IdlPrettyPrint&);

    //
    // Generate fixed value box
    //
    void genCPlusPlusHFixedBox(CORBA::ValueBoxDef_ptr, const char*,
                               const char*, CORBA::TypeDescription*,
                               CORBA::TypeCode_ptr, IdlPrettyPrint&);

    //
    // Generate array value box
    //
    void genCPlusPlusHArrayBox(CORBA::ValueBoxDef_ptr, const char*,
                               const char*, CORBA::TypeDescription*,
                               CORBA::TypeCode_ptr, IdlPrettyPrint&);

    //
    // Generate code for the source file, variable definitions
    //
    void genCPlusPlusCDefs(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate code for the source file, general part
    //
    void genCPlusPlusC(CORBA::Container_ptr, IdlPrettyPrint&);
    
    //
    // Generate code for the source file, collocated part
    //
    void genCPlusPlusCStubImpl(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate OBV_* classes
    //
    void genCPlusPlusCOBV(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate accessor and mutator functions for value member
    //
    void genCPlusPlusCValueMember(CORBA::Container_ptr, CORBA::ValueMember*,
                                  const char*, IdlPrettyPrint&);

    //
    // Generate type-specific value factory implementation
    //
    void genCPlusPlusCValueFactory(CORBA::ValueDef_ptr, IdlPrettyPrint&);

    //
    // Generate code for the skeleton header file, general part
    //
    void genCPlusPlusSkelH(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate code for the skeleton header file, general part
    //
    void genCPlusPlusSkelHStubImpl(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate code for the skeleton header file, protected part
    //
    void genCPlusPlusSkelHProtected(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate code for the skeleton source file, general part
    //
    void genCPlusPlusSkelC(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate code for the skeleton header file, general part
    //
    void genCPlusPlusSkelCStubImpl(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate code for the skeleton header file, protected part
    //
    void genCPlusPlusSkelCProtected(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate code for the tie header file
    //
    void genCPlusPlusTieH(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate code for the example header file
    //
    void genCPlusPlusImplH(CORBA::Container_ptr, bool, IdlPrettyPrint&);

    //
    // Generate header code for an example attribute
    //
    void genCPlusPlusImplHAttribute(const char*,
                                    CORBA::AttributeDescription*,
                                    IdlPrettyPrint&);

    //
    // Generate header code for an example operation
    //
    void genCPlusPlusImplHOperation(const char*,
                                    CORBA::OperationDescription*,
                                    IdlPrettyPrint&);

    //
    // Generate code for the example source file
    //
    void genCPlusPlusImplC(CORBA::Container_ptr, bool, IdlPrettyPrint&);

    //
    // Generate code for an example attribute implementation
    //
    void genCPlusPlusImplCAttribute(const char*,
                                    CORBA::AttributeDescription*,
                                    IdlPrettyPrint&);

    //
    // Generate code for an example operation implementation
    //
    void genCPlusPlusImplCOperation(const char*,
                                    CORBA::OperationDescription*,
                                    IdlPrettyPrint&);

    //
    // Generate default return statement for example implementation
    //
    void genCPlusPlusImplCReturn(CORBA::TypeCode_ptr, const char*,
                                 IdlPrettyPrint&);

public:

    IdlCPPGenerator(const char*, CORBA::ORB_ptr, CORBA::Repository_ptr,
		    const CORBA::RepositoryIdSeq&, const IdlStringSeq&,
		    bool, bool, bool, bool,
		    const char*, const char*,
		    const char*, const char*, const char*);

    IdlCPPGenerator(const char*, CORBA::ORB_ptr, CORBA::Repository_ptr,
		    bool, bool, bool, bool,
		    const char*, const char*,
		    const char*, const char*, const char*);

    ~IdlCPPGenerator();

    //
    // Generate C++ code
    //
    int gen(const char*, const char*);
    int genSkel(const char*, const char*);
    int genTie(const char*, const char*);
    int genImpl(const char*, const char*, bool);
};

#endif
