// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <Types.h>
#include <GenUtil.h>
#include <GenCPP.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Generate code for the skeleton header file, general part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusSkelH(Container_ptr container,
				   IdlPrettyPrint& out)
{
    //
    // Get scope
    //
    String_var scope = getScope(container);

    //
    // Get container contents descriptions
    //
    Container::DescriptionSeq_var seq =
	container -> describe_contents(dk_all, true, -1);

    ULong i, j;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get container
	//
	Container_var cont = Container::_narrow(seq[i].contained_object);

	switch(seq[i].kind)
	{
	case dk_Module:
	{
	    //
	    // Module
	    //
	    ModuleDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    // Don't add this here!!!
	    // if(!check(desc -> id))
	    //     continue;

            String_var name =
                getPrefixedName("POA", seq[i].contained_object);

            comment(desc -> id, out);
            out << "\nnamespace " << name;
            out << "\n{";

            assert(!is_nil(cont));
	    genCPlusPlusSkelH(cont, out);

            out.sep();
            out << "\n} // End of namespace " << name;

	    break;
	}

	case dk_Interface:
	{
	    //
	    // Interface
	    //
	    InterfaceDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

            InterfaceDef_var interf =
                InterfaceDef::_narrow(seq[i].contained_object);
            assert(!is_nil(interf));

            //
            // The name of an abstract interface skeleton is
            // POA_M::OBAbstract_Foo or POA_OBAbstract_Foo
            //
            String_var name;
            if(desc -> is_abstract)
                name = getPrefixedName("POA", seq[i].contained_object,
                                       "OBAbstract");
            else
                name = getPrefixedName("POA", seq[i].contained_object);

            ScopedName_var absolute = getAbsolute(desc -> id);

	    comment(desc -> id, out);
	    
	    out.sep();
	    out << "\nclass " << name << " : ";

	    unsigned int indent = out.getIndent();
	    out.setIndent(out.getPos());

            IdlBaseInterfaceInfoSeq baseInfo;
            IdlGetBaseInterfaces(interf, baseInfo, false);

	    if(baseInfo.length() > 0)
	    {
		for(j = 0 ; j < baseInfo.length() ; j++)
		{
		    //
		    // Use fully-qualified class name
		    //
                    ScopedName_var baseAbsolute;
                    if(baseInfo[j].is_abstract)
                    {
                        baseAbsolute =
                            getPrefixedAbsolute("POA", baseInfo[j].intf, "",
                                                "OBAbstract");
                    }
                    else
                    {
                        baseAbsolute =
                            getPrefixedAbsolute("POA", baseInfo[j].id, "");
                    }

		    if(j > 0)
			out << ",\n";

		    //
		    // Abstract interfaces are always inherited as virtual
		    //
		    if(baseInfo[j].is_abstract || !noVirtual_)
			out << "virtual ";

		    out << "public " << baseAbsolute;
		}
	    }
	    else
	    {
		if(!noVirtual_)
		    out << "virtual ";

		out << "public PortableServer::ServantBase";
	    }

	    out.setIndent(indent);

	    out.start();

	    out << '\n' << name << "(const " << name << "&);";
	    out << "\nvoid operator=(const " << name << "&);";

            // XXX necessary for abstract skeleton?
            if(!desc -> is_abstract)
            {
		out.sep();
		out << '\n' << absolute << "_var stub_;";
            }

	    assert(!is_nil(cont));
	    genCPlusPlusSkelHProtected(cont, out);

	    out.sep();
	    out.dec();
	    out << "\npublic:";
	    out.inc();
	    out.sep();
	    out << '\n' << name << "() { }";

	    out.sep();
	    out << "\nvirtual CORBA::Boolean _is_a(const char*)";
	    out.inc();
	    out << "\nthrow(CORBA::SystemException);";
	    out.dec();

            if(!desc -> is_abstract)
            {
		out.sep();
		out << '\n' << absolute << "_ptr _this();";

		out.sep();
		out << "\nvirtual CORBA::RepositoryId _primary_interface("
		    << "const PortableServer::ObjectId&, "
		    << "PortableServer::POA_ptr);";

		out.sep();
		out << "\nvirtual OB::DispatchStatus _OB_dispatch(const char*,"
		    << " OB::Buffer&, bool, CORBA::ULong, CORBA::ULong);";

		out.sep();
		out << "\nvirtual OB::StubImplBase* _OB_createStubImpl(";
		indent = out.getIndent();
		out.setIndent(out.getPos());
		out << "PortableServer::POA_ptr,";
		out << "\nconst PortableServer::ObjectId&,";
		out << "\nconst OB::OctetSeq&);";
		out.setIndent(indent);
	    }
            
	    assert(!is_nil(cont));
	    genCPlusPlusSkelH(cont, out);

	    out.end();
	    out << ';';

	    break;
	}

	case dk_Value:
        {
            //
            // Valuetype
            //
            ValueDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            ValueDef_var value = ValueDef::_narrow(cont);
            assert(! is_nil(value));

            ValueDef::FullValueDescription_var fullDesc =
                value -> describe_value();

            //
            // Skip this valuetype if it doesn't support any interfaces
            // (either directly or indirectly)
            //
            if(fullDesc -> supported_interfaces.length() == 0)
                continue;

            //
            // XXX - We are generating a skeleton for an abstract valuetype
            // that supports an interface, but should we use a "special"
            // name for the skeleton class?  (e.g., POA_OBAbstract_Foo?)
            //

            String_var name = getPrefixedName("POA", seq[i].contained_object);
            ScopedName_var absolute = getAbsolute(value);

            comment(desc -> id, out);

            out.sep();
            out << "\nclass " << name << " : ";

            unsigned int indent = out.getIndent();
            out.setIndent(out.getPos());

            //
            // TODO - Generate _OB_toAbstractBase and _OB_insertAny
            // if the valuetype supports an abstract interface
            //

            //
            // If concrete base value supports an interface, then derive
            // from the skeleton of the concrete base value
            //
            ValueDef_var base = value -> base_value();
            if(!is_nil(base))
            {
                InterfaceDef_var intf = base -> supported_interface();
                if(!is_nil(intf))
                {
                    ScopedName_var skel =
                        getPrefixedAbsolute("POA", desc -> base_value, "");
                    out << "virtual public " << skel << ",\n";
                }
            }

            //
            // If any of the abstract base values supports an interface,
            // then derive from the skeleton of the abstract base value
            //
            ValueDefSeq_var abstractBases = value -> abstract_base_values();
            for(j = 0 ; j < abstractBases -> length() ; j++)
            {
                InterfaceDef_var intf =
                    abstractBases[j] -> supported_interface();
                if(!is_nil(intf))
                {
                    //
                    // XXX - We may want to use a special name for the
                    // skeleton of an abstract base value
                    //
                    ScopedName_var skel =
                        getPrefixedAbsolute("POA",
                                            desc -> abstract_base_values[j],
                                            "");
                    out << "virtual public " << skel << ",\n";
                }
            }

            //
            // If the valuetype supports an interface, then derive from
            // the interface skeleton
            //
            if(strlen(desc -> supported_interface) > 0)
            {
                InterfaceDef_var intf = value -> supported_interface();
                ScopedName_var skel;
                if(intf -> is_abstract())
                    skel = getPrefixedAbsolute("POA", intf, "", "OBAbstract");
                else
                    skel = getPrefixedAbsolute("POA",
                                               desc -> supported_interface,
                                               "");
                out << "virtual public " << skel << ",\n";
            }

#if 0
            //
            // XXX - Currently we are deriving directly from the interface
            // skeleton class.  Should we really be deriving from the
            // valuetype skeleton class instead?
            //
            for(j = 0 ; j < fullDesc -> supported_interfaces.length() ; j++)
            {
                if(j > 0)
                    out << ",\n";
                const char* id = fullDesc -> supported_interfaces[j];
                Identifier_var intfName =
                    getPrefixedAbsolute("POA", id, scope);
                out << "virtual public " << intfName;
            }
#endif

            //
            // Always derive from the valuetype class
            //
            out << "virtual public " << absolute;

            out.setIndent(indent);

            out.start();

            out << '\n' << name << "(const " << name << "&);";
            out << "\nvoid operator=(const " << name << "&);";
            //out << '\n' << absolute << "_var stub_;";

            out.sep();
            out.dec();
            out << "\npublic:";
            out.inc();
            out.sep();
            out << '\n' << name << "() { }";

            // XXX We don't really need to generate signatures for
            //     interface operations and attributes, because we
            //     inherit them from the interface's skeleton. (Right?)
            //assert(!is_nil(intf));
            //genCPlusPlusSkelH(intf, out);

            out.end();
            out << ';';

            break;
        }

	case dk_Attribute:
	{
	    //
	    // Attribute
	    //
	    AttributeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);

	    comment(desc -> id, out);

	    String_var ret = getTypeString("", desc -> type, GetTypeReturn);
	    String_var sysEx = string_dup("CORBA::SystemException");

	    out.sep();
	    out << "\nvirtual " << ret << ' ' << name << "()";
	    out.inc();
	    out << "\nthrow(" << sysEx << ") = 0;";
	    out.dec();

	    if(desc -> mode == ATTR_NORMAL)
	    {
		String_var in = getTypeString("", desc -> type, GetTypeIn);
		out << "\nvirtual void " << name << '(' << in << ')';
		out.inc();
		out << "\nthrow(" << sysEx << ") = 0;";
		out.dec();
	    }

	    break;
	}

	case dk_Operation:
	{
	    //
	    // Operation
	    //
	    OperationDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);

	    comment(desc -> id, out);

	    String_var ret = getTypeString("", desc -> result, GetTypeReturn);
	    
	    out.sep();
	    out << "\nvirtual " << ret << ' ' << name << "(";
	    
	    unsigned int indent = out.getIndent();
	    out.setIndent(out.getPos());

	    for(j = 0 ; j < desc -> parameters.length() ; j++)
	    {
		GetType tt;
		switch(desc -> parameters[j].mode)
		{
		case PARAM_IN:
		    tt = GetTypeIn;
		    break;

		case PARAM_INOUT:
		    tt = GetTypeInOut;
		    break;

		case PARAM_OUT:
		    tt = GetTypeOut;
		    break;
		}

		String_var argName = fixKwd(desc -> parameters[j].name);

		String_var s = getTypeString("", desc -> parameters[j].type,
                                             tt, false, argName);

		if(j > 0)
		    out << ",\n";

		out << s;
	    }

	    if(desc -> contexts.length() > 0)
	    {
		if(desc -> parameters.length() > 0)
		    out << ",\n";
		
		out << "CORBA::Context_ptr";
	    }

	    out << ")";

	    out.setIndent(indent);

	    //
	    // Skeletons must declare throw() clauses
	    //

	    out.inc();
	    out << "\nthrow(";

	    indent = out.getIndent();
	    out.setIndent(out.getPos());

	    for(j = 0 ; j < desc -> exceptions.length() ; j++)
	    {
		Identifier_var exName = getAbsolute(desc -> exceptions[j].id);
		out << exName << ",\n";
	    }

	    out << "CORBA::SystemException) = 0;";
	    out.setIndent(indent);
	    out.dec();

	    break;
	}

	default:
	    break;
	}
    }
}

// ----------------------------------------------------------------------
// Generate code for the header file, stub part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusSkelHStubImpl(Container_ptr container,
                                           IdlPrettyPrint& out)
{
    //
    // Get scope
    //
    String_var scope = getScope(container);

    //
    // Get container contents descriptions
    //
    Container::DescriptionSeq_var seq =
	container -> describe_contents(dk_all, true, -1);

    ULong i, j;
    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get container
	//
	Container_var cont = Container::_narrow(seq[i].contained_object);

	switch(seq[i].kind)
	{
        case dk_Module:
        {
	    //
	    // Module
	    //
	    ModuleDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    // Don't add this here!!!
	    // if(!check(desc -> id))
	    //     continue;

            String_var name =
                getPrefixedName("OBLocalStubImpl", seq[i].contained_object);

            comment(desc -> id, out);
            out << "\nnamespace " << name;
            out << "\n{";

            assert(!is_nil(cont));
	    genCPlusPlusSkelHStubImpl(cont, out);

            out.sep();
            out << "\n} // End of namespace " << name;

            break;
        }

	case dk_Interface:
	{
	    //
	    // Interface
	    //
	    InterfaceDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

            InterfaceDef_var interf =
                InterfaceDef::_narrow(seq[i].contained_object);
            assert(!is_nil(interf));

            String_var name =
                getPrefixedName("OBLocalStubImpl", seq[i].contained_object);

            ScopedName_var absolute = getAbsolute(desc -> id);

            //
            // We have to use these mangled names to avoid a bug
            // in MSVC
            //
            String_var flat = getFlattenedName(absolute);
            String_var servant = string_dup("_ob_servant_");
            servant += flat;
            servant += "_";

	    comment(desc -> id, out);

	    out.sep();
            out << "\nclass " << name << " : ";
            unsigned int indent = out.getIndent();
            out.setIndent(out.getPos());
            if(!noVirtual_)
                out << "virtual ";
            out << "public OBStubImpl_" << absolute;

            IdlBaseInterfaceInfoSeq baseInfo;
            IdlGetBaseInterfaces(interf, baseInfo, false);

	    if(baseInfo.length() > 0)
	    {
		for(j = 0 ; j < baseInfo.length() ; j++)
		{
                    //
                    // Use fully-qualified class name
                    //
		    ScopedName_var baseAbs =
                        getPrefixedAbsolute("OBLocalStubImpl", baseInfo[j].id,
                                            "");

                    out << ",\n";

		    if(!noVirtual_)
			out << "virtual ";

		    out << "public " << baseAbs;
		}
	    }
            out.setIndent(indent);
            
	    out.start();

	    out << '\n' << name << "(const " << name << "&);";
	    out << "\nvoid operator=(const " << name << "&);";

	    out.sep();
	    out << "\nPortableServer::ServantBase* _OB_setupServant();\n";

            out.dec();
            out.sep();
            out.direct("\n#ifdef HAVE_VCPLUSPLUS_BUGS");
            out << "\npublic:";
            out.direct("\n#else");
            out << "\nprotected:";
            out.direct("\n#endif");
            out.inc();
            out.sep();
            if(desc -> is_abstract)
            {
                ScopedName_var s = getPrefixedAbsolute("POA", interf, "",
                                                       "OBAbstract");
                out << '\n' << s << "* " << servant << ';';

                //
                // The class for an abstract interface has a protected
                // constructor
                //
                out.sep();
                out << '\n' << name << "() { }";
            }
            else
		out << "\nPOA_" << absolute << "* " << servant << ';';

            out.dec();
            out.sep();
            out << "\npublic:";
            out.inc();

            if(!desc -> is_abstract)
            {
		out.sep();
		out << '\n' << name << "() { }";
		out.sep();
		out << '\n' << name << "(";
		indent = out.getIndent();
		out.setIndent(out.getPos());
		out << "PortableServer::POA_ptr,";
		out << "\nconst PortableServer::ObjectId&,";
		out << "\nconst OB::OctetSeq&,";
		out << "\nPOA_" << absolute << "*);";
		out.setIndent(indent);
            }

	    //
	    // _OB_isLocal
	    //
	    out.sep();
	    out << "\nvirtual bool _OB_isLocal() { return true; }";

	    assert(!is_nil(cont));
	    genCPlusPlusSkelHStubImpl(cont, out);

	    out.end();
	    out << ';';

	    break;
	}

	case dk_Attribute:
	{
	    //
	    // Attribute
	    //
	    AttributeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);

	    comment(desc -> id, out);

	    String_var ret = getTypeString("", desc -> type, GetTypeReturn);

	    out.sep();
            out << "\nvirtual " << ret << ' ' << name << "();";

	    if(desc -> mode == ATTR_NORMAL)
	    {
		String_var in = getTypeString("", desc -> type, GetTypeIn);
                out << "\nvirtual void " << name << '(' << in << ");";
	    }

	    break;
	}

	case dk_Operation:
	{
	    //
	    // Operation
	    //
	    OperationDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);

	    comment(desc -> id, out);

	    String_var ret = getTypeString("", desc -> result, GetTypeReturn);

	    out.sep();
	    out << "\nvirtual " << ret << ' ' << name << "(";

	    unsigned int indent = out.getIndent();
	    out.setIndent(out.getPos());

	    for(j = 0 ; j < desc -> parameters.length() ; j++)
	    {
		GetType tt;
		switch(desc -> parameters[j].mode)
		{
		case PARAM_IN:
		    tt = GetTypeIn;
		    break;

		case PARAM_INOUT:
		    tt = GetTypeInOut;
		    break;

		case PARAM_OUT:
		    tt = GetTypeOut;
		    break;
		}

		String_var argName = fixKwd(desc -> parameters[j].name);

		String_var s = getTypeString("", desc -> parameters[j].type,
                                             tt, false, argName);

		if(j > 0)
		    out << ",\n";

		out << s;
	    }

	    if(desc -> contexts.length() > 0)
	    {
		if(desc -> parameters.length() > 0)
		    out << ",\n";

		out << "CORBA::Context_ptr";
	    }

            out << ");";

	    out.setIndent(indent);

	    break;
	}

	default:
	    break;
	}
    }
}

// ----------------------------------------------------------------------
// Generate code for the skeleton header file, "protected" part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusSkelHProtected(Container_ptr container,
					    IdlPrettyPrint& out)
{
    //
    // Get scope
    //
    String_var scope = getScope(container);

    //
    // Get container contents descriptions
    //
    Container::DescriptionSeq_var seq =
	container -> describe_contents(dk_all, true, -1);

    ULong i;
    bool first = true;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get container
	//
	Container_var cont = Container::_narrow(seq[i].contained_object);

	switch(seq[i].kind)
	{
	case dk_Attribute:
	{
	    if(first)
	    {
		out.sep();
		out.dec();
		out << "\nprotected:";
		out.inc();
		first = false;
	    }

	    //
	    // Attribute
	    //
	    AttributeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);

	    comment(desc -> id, out);

	    //
	    // Don't use name here instead of desc -> name. We need
	    // the name "unfixed".
	    //
	    out.sep();
	    out << "\nOB::DispatchStatus _OB_get_" << desc -> name
		<< "(OB::Buffer&, bool, CORBA::ULong, CORBA::ULong);";
	    if(desc -> mode == ATTR_NORMAL)
		out << "\nOB::DispatchStatus _OB_set_" << desc -> name
		    << "(OB::Buffer&, bool, CORBA::ULong, CORBA::ULong);";

	    break;
	}

	case dk_Operation:
	{
	    if(first)
	    {
		out.sep();
		out.dec();
		out << "\nprotected:";
		out.inc();
		first = false;
	    }

	    //
	    // Operation
	    //
	    OperationDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);

	    comment(desc -> id, out);

	    //
	    // Don't use name here instead of desc -> name. We need
	    // the name "unfixed".
	    //
	    out.sep();
	    out << "\nOB::DispatchStatus _OB_op_" << desc -> name
		<< "(OB::Buffer&, bool, CORBA::ULong, CORBA::ULong);";

	    break;
	}

	default:
	    break;
	}
    }
}
