// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Hashtable.h>
#include <Types.h>
#include <Util.h>
#include <GenUtil.h>
#include <GenCPP.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Print header comment
// ----------------------------------------------------------------------

void
IdlCPPGenerator::comment(const char* s, IdlPrettyPrint& out)
{
    out.ignoreNextSep(false);
    out.sep();
    out << "\n//";
    out << "\n// " << s;
    out << "\n//";
    out.ignoreNextSep(true);
}

// ----------------------------------------------------------------------
// Precede C++ keywords by an underscore
// ----------------------------------------------------------------------

char*
IdlCPPGenerator::fixKwd(const char* abs)
{
    assert_nca(abs, NCANullString);

    if(*abs == '\0')
	return string_dup("");

    static const char* kwds[] =
    {
	"and", "and_eq", "asm", "auto", "bitand",
	"bitor", "bool", "break", "case", "catch",
	"char", "class", "compl", "const", "const_cast",
	"continue", "default", "delete", "do", "double",
	"dynamic_cast", "else", "enum", "explicit", "extern",
	"false", "float", "for", "friend", "goto",
	"if", "inline", "int", "long", "mutable",
	"namespace", "new", "not", "not_eq", "operator",
	"or", "or_eq", "private", "protected", "public",
	"register", "reinterpret_cast", "return", "short", "signed",
	"sizeof", "static", "static_cast", "struct", "switch",
	"template", "this", "throw", "true", "try",
	"typedef", "typeid", "typename", "union", "unsigned",
	"using", "virtual", "void", "volatile", "wchar_t",
	"while", "xor", "xor_eq",
	0
    };

    String_var right = (const char*)"";
    const char* s = strchr(abs, ':');
    if(s)
    {
	assert(s[1] == ':');
	right = fixKwd(s + 2);
    }

    String_var result;
    ULong len = s ? s - abs : strlen(abs);

    ULong i;
    for(i = 0 ; kwds[i] != 0 ; i++)
	if(strlen(kwds[i]) == len && strncmp(kwds[i], abs, len) == 0)
	{
            const char* prefix = "_cxx_";
	    result = string_alloc(len + 5);
            strcpy(result.inout(), prefix);
	    strncpy(&result[5], abs, len);
	    result[len + 5] = '\0';
	    break;
	}
    
    if(kwds[i] == 0)
    {
	result = string_alloc(len);
	strncpy(result.inout(), abs, len);
	result[len] = '\0';
    }

    if(s)
    {
	result += "::";
	result += right;
    }

    return result._retn();
}

// ----------------------------------------------------------------------
// Create type string
// ----------------------------------------------------------------------

char*
IdlCPPGenerator::getTypeString(const char* scope, TypeCode_ptr type,
			       GetType gt, bool noAlias, const char* ident,
			       const char* uniqueType)
{
    //
    // Get original type
    //
    TypeCode_var origType = GetOrigType(type);

    //
    // Check if original type is fixed-sized or variable-sized
    //
    bool variable = IdlIsVariable(origType);

    //
    // If alias is wanted, get alias name
    //
    String_var alias;
    if(noAlias == false && type -> kind() == tk_alias)
    {
	alias = getAbsolute(type -> id());
	IdlRemoveScope(scope, alias.inout());
    }
    else
	alias = string_dup("");

    //
    // Get result
    //
    String_var result = string_dup("???");
    TCKind kind = origType -> kind();
    switch(kind)
    {
    case tk_void:
    case tk_short:
    case tk_long:
    case tk_longlong:
    case tk_ushort:
    case tk_ulong:
    case tk_ulonglong:
    case tk_float:
    case tk_double:
    case tk_longdouble:
    case tk_boolean:
    case tk_char:
    case tk_wchar:
    case tk_octet:
    {
	// **************************************************
	// Basic types (without string)
	// **************************************************

	if(strlen(alias))
	    result = alias;
	else
	{
	    static const char* names[] =
	    {
		"???", // null
		"void", // void
		"CORBA::Short", // short
		"CORBA::Long", // long
		"CORBA::UShort", // ushort
		"CORBA::ULong", // ulong
		"CORBA::Float", // float
		"CORBA::Double", // double
		"CORBA::Boolean", // boolean
		"CORBA::Char", // char
		"CORBA::Octet", // octet
		"???", // any
		"???", // TypeCode
		"???", // Principal
		"???", // objref
		"???", // struct
		"???", // union
		"???", // enum
		"???", // string
		"???", // sequence
		"???", // array
		"???", // alias
		"???", // except
		"CORBA::LongLong", // longlong
		"CORBA::ULongLong", // ulonglong
		"CORBA::LongDouble", // longdouble
		"CORBA::WChar", // wchar
		"???", // wstring
		"???", // fixed
		"???", // value
		"???", // value_box
		"???", // native
		"???"  // abstract_interface
	    };
	    
	    result = names[(int)kind];
	}

	switch(gt)
	{
	case GetTypeNormal:
	case GetTypeStruct:
	case GetTypeArray:
	case GetTypeIn:
	case GetTypeReturn:
	    break;

	case GetTypePtr:
	case GetTypeRef:
	case GetTypeVar:
	case GetTypeSlice:
	case GetTypeForAny:
	    result = string_dup("???");
	    break;

	case GetTypeOut:
            result += "_out";
            break;
            
	case GetTypeInOut:
	    result += '&';
	    break;

	case GetTypeConst:
	{
	    String_var tmp = string_dup("const ");
	    tmp += result;
	    result = tmp._retn();
	    break;
	}
	}

	if(ident && strlen(result))
	{
	    result += ' ';
	    result += ident;
	}

	break;
    }

    case tk_fixed:
    {
	// **************************************************
	// The fixed type
	// **************************************************

	if(strlen(alias))
	    result = alias;
	else
	    result = string_dup("CORBA::Fixed");

	switch(gt)
	{
	case GetTypeNormal:
	case GetTypeStruct:
	case GetTypeArray:
	case GetTypeIn:
	case GetTypeReturn:
	    break;

	case GetTypePtr:
	case GetTypeRef:
	case GetTypeVar:
	case GetTypeSlice:
	case GetTypeForAny:
	    result = string_dup("???");
	    break;

	case GetTypeOut:
	    if(strlen(alias))
		result += "_out";
	    else
		result += "&";
            break;
            
	case GetTypeInOut:
	    result += '&';
	    break;

	case GetTypeConst:
	{
	    String_var tmp = string_dup("const ");
	    tmp += result;
	    result = tmp._retn();
	    break;
	}
	}

	if(ident && strlen(result))
	{
	    result += ' ';
	    result += ident;
	}

	break;
    }

    case tk_string:
    {
	// **************************************************
	// Strings
	// **************************************************

	if(strlen(alias))
	    result = alias;
	else
	    result = string_dup("char*");

	switch(gt)
	{
	case GetTypeNormal:
	case GetTypeReturn:
	    break;

	case GetTypePtr:
	case GetTypeRef:
	case GetTypeSlice:
	case GetTypeForAny:
	    result = string_dup("???");
	    break;

	case GetTypeIn:
	case GetTypeConst:
	    result = string_dup("const char*");
	    break;

	case GetTypeOut:
            if(strlen(alias))
                result += "_out";
            else
                result = string_dup("CORBA::String_out");
            break;
            
	case GetTypeInOut:
	    result += '&';
	    break;

	case GetTypeVar:
	    if(strlen(alias))
		result += "_var";
	    else
		result = string_dup("CORBA::String_var");
	    break;

	case GetTypeStruct:
	case GetTypeArray:
	    result = string_dup("OB::StrForStruct");
	    break;
	}

	if(ident && strlen(result))
	{
	    result += ' ';
	    result += ident;
	}

	break;
    }

    case tk_wstring:
    {
	// **************************************************
	// Wide strings
	// **************************************************

	if(strlen(alias))
	    result = alias;
	else
	    result = string_dup("wchar_t*");

	switch(gt)
	{
	case GetTypeNormal:
	case GetTypeReturn:
	    break;

	case GetTypePtr:
	case GetTypeRef:
	case GetTypeSlice:
	case GetTypeForAny:
	    result = string_dup("???");
	    break;

	case GetTypeIn:
	case GetTypeConst:
	    result = string_dup("const wchar_t*");
	    break;

	case GetTypeOut:
            if(strlen(alias))
		result += "_out";
	    else
		result = string_dup("CORBA::WString_out");
            break;
            
	case GetTypeInOut:
	    result += '&';
	    break;

	case GetTypeVar:
            if(strlen(alias))
		result += "_var";
	    else
		result = string_dup("CORBA::WString_var");
            break;

	case GetTypeStruct:
	case GetTypeArray:
            result = string_dup("OB::WStrForStruct");
	    break;
	}

	if(ident && strlen(result))
	{
	    result += ' ';
	    result += ident;
	}

	break;
    }

    case tk_any:
    {
	// **************************************************
	// The any type
	// **************************************************

	if(strlen(alias))
	    result = alias;
	else
	    result = string_dup("CORBA::Any");

	switch(gt)
	{
	case GetTypeNormal:
	case GetTypeStruct:
	case GetTypeArray:
	    break;

	case GetTypeIn:
	{
	    String_var tmp = string_dup("const ");
	    tmp += result;
	    result = tmp._retn();
	    result += '&';
	    break;
	}

	case GetTypeReturn:
	    result += '*';
	    break;

	case GetTypePtr:
	case GetTypeRef:
	case GetTypeSlice:
	case GetTypeForAny:
	    result = string_dup("???");
	    break;

	case GetTypeVar:
	    result += "_var";
	    break;

	case GetTypeInOut:
	    result += '&';
	    break;

	case GetTypeOut:
	    result += "_out";
	    break;

	case GetTypeConst:
	    result = string_dup("???");
	    break;
	}

	if(ident && strlen(result))
	{
	    result += ' ';
	    result += ident;
	}

	break;
    }

    case tk_TypeCode:
    case tk_Principal:
    case tk_objref:
    case tk_abstract_interface:
    {
	// ******************************************************
	// Object references (including pseudo object references)
        // ******************************************************

	if(strlen(alias))
	    result = alias;
	else
	{
	    if(kind == tk_TypeCode)
	    {
		result = string_dup("CORBA::TypeCode");
	    }
	    else if(kind == tk_Principal)
	    {
		result = string_dup("CORBA::Principal");
	    }
	    else if(kind == tk_objref)
	    {
		if(strcmp(origType -> id(),
			  "IDL:omg.org/CORBA/Object:1.0") == 0)
		{
		    result = string_dup("CORBA::Object");
		}
		else
		{
		    String_var s = getAbsolute(origType -> id());
		    IdlRemoveScope(scope, s.inout());
		    result = s;
		}
	    }
	    else if(kind == tk_abstract_interface)
	    {
		String_var s = getAbsolute(origType -> id());
		IdlRemoveScope(scope, s.inout());
		result = s;
	    }
	}

	switch(gt)
	{
	case GetTypeNormal:
	    break;

	case GetTypeConst:
	case GetTypeSlice:
	case GetTypeForAny:
	    result = string_dup("???");
	    break;

	case GetTypePtr:
	case GetTypeReturn:
	case GetTypeIn:
	    result += "_ptr";
	    break;

	case GetTypeRef:
	    result += "Ref";
	    break;

	case GetTypeVar:
	case GetTypeStruct:
	case GetTypeArray:
	    result += "_var";
	    break;

	case GetTypeOut:
            result += "_out";
            break;
            
	case GetTypeInOut:
	    result += "_ptr&";
	    break;
	}

	if(ident && strlen(result))
	{
	    result += ' ';
	    result += ident;
	}

	break;
    }

    case tk_struct:
    case tk_union:
    {
	// **************************************************
	// Structs and unions
	// **************************************************

	if(strlen(alias))
	    result = alias;
	else
	{
	    String_var s = getAbsolute(origType -> id());
	    IdlRemoveScope(scope, s.inout());
	    result = s;
	}

	switch(gt)
	{
	case GetTypeNormal:
	case GetTypeStruct:
	case GetTypeArray:
	    break;

	case GetTypePtr:
	case GetTypeRef:
	case GetTypeConst:
	case GetTypeSlice:
	case GetTypeForAny:
	    result = string_dup("???");
	    break;

	case GetTypeVar:
	    result += "_var";
	    break;

	case GetTypeIn:
	{
	    String_var tmp = string_dup("const ");
	    tmp += result;
	    result = tmp._retn();
	    result += '&';
	    break;
	}

	case GetTypeInOut:
	    result += '&';
	    break;

	case GetTypeOut:
            result += "_out";
	    break;

	case GetTypeReturn:
	    if(variable)
		result += '*';
	    break;
	}

	if(ident && strlen(result))
	{
	    result += ' ';
	    result += ident;
	}

	break;
    }

    case tk_except:
    {
	// **************************************************
	// Exceptions
	// **************************************************

	if(strlen(alias))
	    result = alias;
	else
	{
	    String_var s = getAbsolute(origType -> id());
	    IdlRemoveScope(scope, s.inout());
	    result = s;
	}

	switch(gt)
	{
	case GetTypeNormal:
	case GetTypeStruct:
	case GetTypeArray:
	    break;

	case GetTypePtr:
	case GetTypeRef:
	case GetTypeVar:
	case GetTypeConst:
	case GetTypeSlice:
	case GetTypeForAny:
	case GetTypeIn:
	case GetTypeInOut:
	case GetTypeOut:
	case GetTypeReturn:
	    result = string_dup("???");
	    break;
	}

	if(ident && strlen(result))
	{
	    result += ' ';
	    result += ident;
	}

	break;
    }

    case tk_enum:
    {
	// **************************************************
	// Enums
	// **************************************************

	if(strlen(alias))
	    result = alias;
	else
	{
	    String_var s = getAbsolute(origType -> id());
	    IdlRemoveScope(scope, s.inout());
	    result = s;
	}

	switch(gt)
	{
	case GetTypeNormal:
	case GetTypeStruct:
	case GetTypeArray:
	case GetTypeIn:
	case GetTypeReturn:
	    break;

	case GetTypePtr:
	case GetTypeRef:
	case GetTypeVar:
	case GetTypeConst:
	case GetTypeSlice:
	case GetTypeForAny:
	    result = string_dup("???");
	    break;

	case GetTypeOut:
            result += "_out";
            break;
            
	case GetTypeInOut:
	    result += '&';
	    break;
	}

	if(ident && strlen(result))
	{
	    result += ' ';
	    result += ident;
	}

	break;
    }

    case tk_sequence:
    {
	// **************************************************
	// Sequences
	// **************************************************

	if(strlen(alias))
	    result = alias;
	else
	{
	    //
	    // Overwrite the unique type string if necessary
	    //
	    String_var unique = uniqueType;
	    if(type -> kind() == tk_alias)
	    {
		TypeCode_var lastAliasType =
		    TypeCode::_duplicate(type);
		
		TypeCode_var lastAliasContentType =
		    lastAliasType -> content_type();
		
		while(lastAliasContentType -> kind() == tk_alias)
		{
		    lastAliasType = lastAliasContentType;
		    lastAliasContentType = lastAliasType -> content_type();
		}
		
		if(lastAliasContentType -> kind() == tk_sequence)
		{
		    unique = getUnique(lastAliasType -> id());
		    IdlRemoveScope(scope, unique.inout());
		}
	    }

	    TypeCode_var contentType = origType -> content_type();
	    TypeCode_var origContentType = GetOrigType(contentType);

	    String_var contentTypeString;
	    if(origContentType -> kind() == tk_array)
		contentTypeString = getTypeString(scope, contentType,
						  GetTypeSlice, noAlias, 0,
						  unique);
	    else
		contentTypeString = getTypeString(scope, contentType,
						  GetTypeNormal, noAlias, 0,
						  unique);
		
	    switch(origContentType -> kind())
	    {
	    case tk_null:
	    case tk_void:
	    case tk_alias:
		assert(false);
		break;
		    
	    case tk_short:
	    case tk_long:
	    case tk_longlong:
	    case tk_ushort:
	    case tk_ulong:
	    case tk_ulonglong:
	    case tk_float:
	    case tk_double:
	    case tk_longdouble:
	    case tk_fixed:
	    case tk_boolean:
	    case tk_char:
	    case tk_wchar:
	    case tk_octet:
	    case tk_any:
	    case tk_struct:
	    case tk_union:
	    case tk_enum:
	    case tk_sequence:
	    case tk_except:
		    
		if(IdlIsVariable(contentType))
		{
		    if(origType -> length())
			result = string_dup("OB::BndVarSeq< ");
		    else
			result = string_dup("OB::VarSeq< ");
		}
		else
		{
		    if(origType -> length())
			result = string_dup("OB::BndFixSeq< ");
		    else
			result = string_dup("OB::FixSeq< ");
		}
		    
		result += contentTypeString;
		    
		if(origType -> length())
		{
		    result += ", ";
		    result += origType -> length();
		}
		    
		result += ", ";
		result += unique;
		result += " >";
		    
		break;
		    
	    case tk_array:
	    {
		if(IdlIsVariable(contentType))
		{
		    if(origType -> length())
			result = string_dup("OB::BndVarArraySeq< ");
		    else
			result = string_dup("OB::VarArraySeq< ");
		}
		else
		{
		    if(origType -> length())
			result = string_dup("OB::BndFixArraySeq< ");
		    else
			result = string_dup("OB::FixArraySeq< ");
		}
		    
		TypeCode_var contentContentType;
		ULong length;
		getArrayType(origContentType, contentContentType.out(),
			     length);
		String_var contCont =
		    getTypeString(scope, contentContentType, GetTypeArray,
				  noAlias, 0, unique);
		    
		result += contentTypeString;
		result += ", ";
		result += contCont;
		result += ", ";
		result += length;
		    
		if(origType -> length())
		{
		    result += ", ";
		    result += origType -> length();
		}
		    
		result += ", ";
		result += unique;
		result += " >";
		    
		break;
	    }
		
	    case tk_TypeCode:
	    case tk_Principal:
	    case tk_objref:
	    case tk_abstract_interface:
	    case tk_value:
	    case tk_value_box:
	    {
		if(origType -> length())
		    result = string_dup("OB::BndObjSeq< ");
		else
		    result = string_dup("OB::ObjSeq< ");
		    
		result += contentTypeString;
		    
		if(origType -> length())
		{
		    result += ", ";
		    result += origType -> length();
		}
		    
		result += ", ";
		result += unique;
		result += " >";

		break;
            }
		    
	    case tk_string:
		    
		if(origType -> length())
		{
		    result = string_dup("OB::BndStrSeq< ");
		    result += origType -> length();
		    result += ", ";
		    result += unique;
		    result += " >";
		}
		else
		{
		    result = string_dup("OB::StrSeq< ");
		    result += unique;
		    result += " >";
		}
		    
		break;
		    
	    case tk_wstring:
		    
		if(origType -> length())
		{
		    result = string_dup("OB::BndWStrSeq< ");
		    result += origType -> length();
		    result += ", ";
		    result += unique;
		    result += " >";
		}
		else
		{
		    result = string_dup("OB::WStrSeq< ");
		    result += unique;
		    result += " >";
		}
		    
		break;
	    }
	}

	switch(gt)
	{
	case GetTypeNormal:
	case GetTypeStruct:
	case GetTypeArray:
	    break;

	case GetTypePtr:
	case GetTypeRef:
	case GetTypeConst:
	case GetTypeSlice:
	case GetTypeForAny:
	    result = string_dup("???");
	    break;

	case GetTypeVar:
	{
	    if(strlen(alias))
	    {
		result += "_var";
	    }
	    else
	    {
		String_var tmp = string_dup("OB::SeqVar< ");
		tmp += result;
		result = tmp._retn();
		result += " >";
	    }
	    break;
	}

	case GetTypeIn:
	{
	    String_var tmp = string_dup("const ");
	    tmp += result;
	    result = tmp._retn();
	    result += '&';
	    break;
	}

	case GetTypeOut:
        {
            if(strlen(alias))
            {
                result += "_out";
            }
            else
            {
                String_var tmp = string_dup("OB::SeqOut< ");
                tmp += result;
                result = tmp._retn();
                result += " >";
            }
	    break;
        }

	case GetTypeInOut:
	    result += '&';
	    break;

	case GetTypeReturn:
	    result += '*';
	    break;
	}

	if(ident && strlen(result))
	{
	    result += ' ';
	    result += ident;
	}

	break;
    }

    case tk_array:
    {
	// **************************************************
	// Arrays
	// **************************************************

	if(strlen(alias))
	{
	    result = alias;

	    switch(gt)
	    {
	    case GetTypeNormal:
	    case GetTypeStruct:
	    case GetTypeArray:
	    case GetTypeInOut:
		break;
		
	    case GetTypePtr:
	    case GetTypeRef:
	    case GetTypeConst:
		result = string_dup("???");
		break;

	    case GetTypeVar:
		result += "_var";
		break;

	    case GetTypeSlice:
		result += "_slice";
		break;

	    case GetTypeForAny:
		result += "_forany";
		break;

	    case GetTypeIn:
	    {
		String_var tmp = string_dup("const ");
		tmp += result;
		result = tmp._retn();
		break;
	    }
	    
	    case GetTypeOut:
		//if(variable)
                result += "_out";
		break;

	    case GetTypeReturn:
		result += "_slice*";
		break;
	    }

	    if(ident && strlen(result))
	    {
		result += ' ';
		result += ident;
	    }
	}
	else
	{
	    OB::ULongSeq sizes;
	    
	    TypeCode_var contentType = TypeCode::_duplicate(type);
	    TypeCode_var origContentType = GetOrigType(contentType);
	    
	    while(origContentType -> kind() == tk_array)
	    {
		sizes.append(origContentType -> length());

		contentType = origContentType -> content_type();
		origContentType = GetOrigType(contentType);
	    }
	    
	    String_var array = string_dup("");
	    String_var slice = string_dup("");

	    ULong length = 1;
	    for(ULong i = 0 ; i < sizes.length() ; i++)
	    {
		length *= sizes[i];

		array += '[';
		array += sizes[i];
		array += ']';

		if(i > 0)
		{
		    slice += '[';
		    slice += sizes[i];
		    slice += ']';
		}
	    }

	    result = getTypeString(scope, contentType, GetTypeArray, noAlias,
				   0, uniqueType);

	    switch(gt)
	    {
	    case GetTypeNormal:
	    case GetTypeStruct:
	    case GetTypeArray:
	    case GetTypeInOut:

		if(ident)
		{
		    result += ' ';
		    result += ident;
		}

		result += array;
		 
		break;
		
	    case GetTypePtr:
	    case GetTypeRef:
	    case GetTypeConst:
		result = string_dup("???");
		break;

	    case GetTypeVar:
	    {
		String_var cont = result;

		if(variable)
		{
		    String_var tmp =
			string_dup("OB::VarArrayVar< ");
		    tmp += result;
		    result = tmp._retn();
		}
		else
		{
		    String_var tmp =
			string_dup("OB::FixArrayVar< ");
		    tmp += result;
		    result = tmp._retn();
		}

		result += slice;
		result += ", ";
		result += cont;
		result += ", ";
		result += length;
		result += " >";

		if(ident)
		{
		    result += ' ';
		    result += ident;
		}

		break;
	    }

	    case GetTypeForAny:
	    {
		String_var cont = result;

		if(variable)
		{
		    String_var tmp =
			string_dup("OB::VarArrayForAny< ");
		    tmp += result;
		    result = tmp._retn();
		}
		else
		{
		    String_var tmp =
			string_dup("OB::FixArrayForAny< ");
		    tmp += result;
		    result = tmp._retn();
		}

		result += slice;
		result += ", ";
		result += cont;
		result += ", ";
		result += length;
		result += " >";
		
		if(ident)
		{
		    result += ' ';
		    result += ident;
		}

		break;
	    }

	    case GetTypeSlice:

		if(ident)
		{
		    result += ' ';
		    result += ident;
		}

		result += slice;
		 
		break;
		
	    case GetTypeIn:
	    {
		String_var tmp = string_dup("const ");
		tmp += result;
		result = tmp._retn();

		if(ident)
		{
		    result += ' ';
		    result += ident;
		}

		result += array;

		break;
	    }

	    case GetTypeOut:
            {
                String_var cont = result;

		if(variable)
		{
                    String_var tmp =
                        string_dup("OB::VarArrayOut< ");
                    tmp += result;
                    result = tmp._retn();
                    
                    result += slice;
                    result += ", ";
                    result += cont;
                    result += ", ";
                    result += length;
                    result += " >";
                    if(ident)
                    {
                        result += ' ';
                        result += ident;
                    }
                }
                else
                {
                    if(ident)
                    {
                        result += ' ';
                        result += ident;
                    }
                    result += array;
                }
                break;
            }

	    case GetTypeReturn:

		if(strlen(slice) > 0)
		    result += " (";

		result += '*';
		
		if(ident)
		{
		    if(strlen(slice) == 0)
			result += ' ';

		    result += ident;
		}
		
		if(strlen(slice) > 0)
		    result += ')';

		result += slice;

		break;
	    }
	}
	
	break;
    }

    case tk_alias:
    case tk_null:
	assert(false);
	break;

    case tk_native:
    {
	// **************************************************
	// The native type
	// **************************************************

	if(strlen(alias))
	    result = alias;
	else
	{
	    String_var s = getAbsolute(origType -> id());
	    IdlRemoveScope(scope, s.inout());
	    result = s;
	}

	switch(gt)
	{
	case GetTypeNormal:
	case GetTypeReturn:
	case GetTypeIn:
	    break;

	//XXX - for _OB_op_ _ob_a? param writing
	case GetTypeStruct:
	    break;

	case GetTypePtr:
	case GetTypeRef:
	case GetTypeVar:
	case GetTypeArray:
	case GetTypeSlice:
	case GetTypeForAny:
	case GetTypeConst:
	    result = string_dup("???");
	    break;

	case GetTypeOut:
	case GetTypeInOut:
	    result += '&';
	    break;
	}

	if(ident && strlen(result))
	{
	    result += ' ';
	    result += ident;
	}

	break;
    }

    case tk_value:
    case tk_value_box:
    {
	// **************************************************
	// Valuetypes
	// **************************************************

	if(strlen(alias))
	    result = alias;
	else
	{
	    if(strcmp(origType -> id(),
                      "IDL:omg.org/CORBA/ValueBase:1.0") == 0)
	    {
		result = string_dup("CORBA::ValueBase");
	    }
	    else
	    {
		String_var s = getAbsolute(origType -> id());
		IdlRemoveScope(scope, s.inout());
		result = s;
	    }
	}

	switch(gt)
	{
	case GetTypeNormal:
	    break;

	case GetTypePtr:
	case GetTypeArray:
	    result += '*';
	    break;

	case GetTypeRef:
	case GetTypeConst:
	case GetTypeSlice:
	case GetTypeForAny:
	    result = string_dup("???");
	    break;

	case GetTypeStruct:
	case GetTypeVar:
	    result += "_var";
	    break;

	case GetTypeIn:
	    result += '*';
	    break;

	case GetTypeInOut:
	    result += "*&";
	    break;

	case GetTypeOut:
            result += "*&"; // XXX _out?
	    break;

	case GetTypeReturn:
	    result += '*';
	    break;
	}

	if(ident && strlen(result))
	{
	    result += ' ';
	    result += ident;
	}

	break;
    }
    }

    return result._retn();
}

// ----------------------------------------------------------------------
// Get array type and total number of elements
// ----------------------------------------------------------------------

void
IdlCPPGenerator::getArrayType(TypeCode_ptr type,
			      TypeCode_ptr& cont,
			      ULong& length)
{
    TypeCode_var tc = GetOrigType(type);
    assert(tc -> kind() == tk_array);
    
    length = 1;
    TypeCode_var contentType = tc;
    while(contentType -> kind() == tk_array)
    {
	length *= contentType -> length();
	contentType = contentType -> content_type();
	contentType = GetOrigType(contentType);
    }

    // TODO: cont is now always origContentType, but it should be just
    // contentType
    cont = contentType._retn();
}

// ----------------------------------------------------------------------
// Determine if storage class for TypeCode is extern
// ----------------------------------------------------------------------

bool
IdlCPPGenerator::isTypeCodeExtern(Container_ptr cont)
{
    return(cont -> def_kind() != dk_Interface &&
           cont -> def_kind() != dk_Value);
}

// ----------------------------------------------------------------------
// Determine if content buffer is needed for a sequence or array
// type in order to count its size
// ----------------------------------------------------------------------

bool
IdlCPPGenerator::needDataForCount(TypeCode_ptr type)
{
    bool result;

    TypeCode_var origType = GetOrigType(type);

    assert(origType -> kind() == tk_array ||
           origType -> kind() == tk_sequence);

    TypeCode_var contentType = origType -> content_type();
    TypeCode_var origContentType = GetOrigType(contentType);

    switch(origContentType -> kind())
    {
    case tk_null:
    case tk_void:
    case tk_alias:
	assert(false);
	break;

    case tk_short:
    case tk_long:
    case tk_ushort:
    case tk_ulong:
    case tk_float:
    case tk_double:
    case tk_boolean:
    case tk_char:
    case tk_octet:
    case tk_longlong:
    case tk_ulonglong:
    case tk_longdouble:
        result = false;
        break;

    case tk_array:
        result = needDataForCount(origContentType);
        break;

    default:
        result = true;
        break;
    }

    return result;
}

// ----------------------------------------------------------------------
// Write read/write/count code
// ----------------------------------------------------------------------

void
IdlCPPGenerator::writeReadWriteCount(const char* scope,
				     TypeCode_ptr type,
				     const char* name,
				     ULong& iter,
				     RWC rwc,
				     const char* uniqueType,
				     IdlPrettyPrint& out)
{
    static const char* ops[] =
    {
	"???", // null
	"???", // void
	"short", // short
	"long", // long
	"ushort", // ushort
	"ulong", // ulong
	"float", // float
	"double", // double
	"boolean", // boolean
	"char", // char
	"octet", // octet
	"???", // any
	"???", // TypeCode
	"???", // Principal
	"???", // objref
	"???", // struct
	"???", // union
	"???", // enum
	"???", // string
	"???", // sequence
	"???", // array
	"???", // alias
	"???", // except
	"longlong", // longlong
	"ulonglong", // ulonglong
	"longdouble", // longdouble
	"wchar", // wchar
	"???", // wstring
	"???", // fixed
	"???", // value
	"???", // value_box
	"???", // native
	"???"  // abstract_interface
    };

    static const char* opsArray[] =
    {
	"???", // null
	"???", // void
	"short", // short
	"long", // long
	"ushort", // ushort
	"ulong", // ulong
	"float", // float
	"double", // double
	"boolean", // boolean
	"char", // char
	"octet", // octet
	"???", // any
	"???", // TypeCode
	"???", // Principal
	"???", // objref
	"???", // struct
	"???", // union
	"???", // enum
	"???", // string
	"???", // sequence
	"???", // array
	"???", // alias
	"???", // except
	"longlong", // longlong
	"ulonglong", // ulonglong
	"longdouble", // longdouble
	"wchar", // wchar
	"???", // wstring
	"???", // fixed
	"???", // value
	"???", // value_box
	"???", // native
	"???"  // abstract_interface
    };

    TypeCode_var origType = GetOrigType(type);

    switch(origType -> kind())
    {
    case tk_null:
    case tk_void:
	assert(false);
	break;

    case tk_short:
    case tk_long:
    case tk_ushort:
    case tk_ulong:
    case tk_float:
    case tk_double:
    case tk_boolean:
    case tk_char:
    case tk_octet:
    case tk_longlong:
    case tk_ulonglong:
    case tk_longdouble:
    case tk_wchar:
    {
	// **************************************************
	// Basic types (without string)
	// **************************************************

	const char* op = ops[(int)origType -> kind()];

	switch(rwc)
	{
	case RWCReadReturnAlloc:
	{
	    String_var ret = getTypeString(scope, type, GetTypeReturn);
	    out << '\n' << ret << ' ' << name << " = _ob_in.read_" << op
		<< "();";
	    break;
	}

	case RWCReadReturnDirect:
	    out << "\nreturn _ob_in.read_" << op << "();";
	    break;

	case RWCReadOut:
	case RWCReadInOut:
	case RWCReadStructElem:
	    out << '\n' << name << " = _ob_in.read_" << op << "();";
	    break;

	case RWCCountIn:
	case RWCCountInOut:
	    if(origType -> kind() == tk_wchar)
		out << "\n_ob_cnt.count_wchar(" << name << ");";
	    else
		out << "\n_ob_cnt.count_" << op << "();";
	    break;

	case RWCWriteIn:
	case RWCWriteInOut:
	    out << "\n_ob_out.write_" << op << '(' << name << ");";
	    break;
	}

	break;
    }

    case tk_fixed:
    {
	// **************************************************
	// Fixed type
	// **************************************************

	switch(rwc)
	{
	case RWCReadReturnAlloc:
	{
	    String_var ret = getTypeString(scope, type, GetTypeReturn);
	    out << '\n' << ret << ' ' << name
		<< " = _ob_in.read_fixed().movePointLeft("
		<< origType -> fixed_scale() << ");";
	    break;
	}
	
	case RWCReadReturnDirect:
	    out << "\nreturn _ob_in.read_fixed().movePointLeft("
		<< origType -> fixed_scale() << ");";
	    break;
	    
	case RWCReadOut:
	case RWCReadInOut:
	case RWCReadStructElem:
	    out << '\n' << name << " = _ob_in.read_fixed().movePointLeft("
		<< origType -> fixed_scale() << ");";
	    break;

	case RWCCountIn:
	case RWCCountInOut:
	    out << "\n_ob_cnt.count_fixed(" << name << ");";
	    break;

	case RWCWriteIn:
	case RWCWriteInOut:
	    out << "\n_ob_out.write_fixed(" << name << ".movePointRight("
		<< origType -> fixed_scale() << "));";
	    break;
	}

	break;
    }

    case tk_string:
    {
	// **************************************************
	// Strings
	// **************************************************

	switch(rwc)
	{
	case RWCReadReturnAlloc:
	{
	    String_var var = getTypeString(scope, type, GetTypeVar);
	    out << '\n' << var << ' ' << name << " = _ob_in.read_string();";
	    break;
	}
	
	case RWCReadReturnDirect:
	    out << "\nreturn _ob_in.read_string();";
	    break;

	case RWCReadOut:
	case RWCReadStructElem:
	    out << '\n' << name << " = _ob_in.read_string();";
	    break;

	case RWCReadInOut:
	    out << "\nCORBA::string_free(" << name << ");";
	    out << '\n' << name << " = 0;";
	    out << '\n' << name << " = _ob_in.read_string();";
	    break;

	case RWCCountIn:
	case RWCCountInOut:
	    out << "\n_ob_cnt.count_string(" << name << ");";
	    break;

	case RWCWriteIn:
	case RWCWriteInOut:
	    out << "\n_ob_out.write_string(" << name << ");";
	    break;
	}
	
	break;
    }
    
    case tk_wstring:
    {
	// **************************************************
	// Wide strings
	// **************************************************

	switch(rwc)
	{
	case RWCReadReturnAlloc:
	{
	    String_var var = getTypeString(scope, type, GetTypeVar);
	    out << '\n' << var << ' ' << name << " = _ob_in.read_wstring();";
	    break;
	}
	
	case RWCReadReturnDirect:
	    out << "\nreturn _ob_in.read_wstring();";
	    break;

	case RWCReadOut:
	case RWCReadStructElem:
	    out << '\n' << name << " = _ob_in.read_wstring();";
	    break;

	case RWCReadInOut:
	    out << "\nCORBA::wstring_free(" << name << ");";
	    out << '\n' << name << " = 0;";
	    out << '\n' << name << " = _ob_in.read_wstring();";
	    break;

	case RWCCountIn:
	case RWCCountInOut:
	    out << "\n_ob_cnt.count_wstring(" << name << ");";
	    break;

	case RWCWriteIn:
	case RWCWriteInOut:
	    out << "\n_ob_out.write_wstring(" << name << ");";
	    break;
	}
	
	break;
    }
    
    case tk_any:
    {
	// **************************************************
	// The any type
	// **************************************************

	switch(rwc)
	{
	case RWCReadReturnAlloc:
	{
	    String_var var = getTypeString(scope, type, GetTypeVar);
	    out << '\n' << var << ' ' << name << " = _ob_in.read_any();";
	    break;
	}
	
	case RWCReadReturnDirect:
	    out << "\nreturn _ob_in.read_any();";
	    break;

	case RWCReadOut:
	    out << '\n' << name << " = _ob_in.read_any();";
	    break;

	case RWCReadInOut:
	case RWCReadStructElem:
	    out << "\n_ob_in.read_any(" << name << ");";
	    break;

	case RWCCountIn:
	case RWCCountInOut:
	    out << "\n_ob_cnt.count_any(" << name << ");";
	    break;

	case RWCWriteIn:
	case RWCWriteInOut:
	    out << "\n_ob_out.write_any(" << name << ");";
	    break;
	}
	
	break;
    }
    
    case tk_TypeCode:
    {
	// ******************************************************
	// TypeCode
        // ******************************************************

	switch(rwc)
	{
	case RWCReadReturnAlloc:
	{
	    String_var var = getTypeString(scope, type, GetTypeVar);
	    out << '\n' << var << ' ' << name << " = _ob_in.read_TypeCode()";
	    break;
	}
	
	case RWCReadReturnDirect:
	    out << "\nreturn _ob_in.read_TypeCode();";
	    break;

	case RWCReadOut:
	case RWCReadStructElem:
	    out << '\n' << name << " = _ob_in.read_TypeCode();";
	    break;

	case RWCReadInOut:
	{
	    String_var norm = getTypeString(scope, type);
	    out << "\nCORBA::release(" << name << ");";
	    out << '\n' << name << " = " << norm << "::_nil();";
	    out << '\n' << name << " = _ob_in.read_TypeCode();";
	    break;
	}

	case RWCCountIn:
	case RWCCountInOut:
	    out << "\n_ob_cnt.count_TypeCode(" << name << ");";
	    break;

	case RWCWriteIn:
	case RWCWriteInOut:
	    out << "\n_ob_out.write_TypeCode(" << name << ");";
	    break;
	}
	
	break;
    }
    
    case tk_Principal:
    {
	// ******************************************************
	// Prinicpal
        // ******************************************************

	switch(rwc)
	{
	case RWCReadReturnAlloc:
	{
	    String_var var = getTypeString(scope, type, GetTypeVar);
	    out << '\n' << var << ' ' << name << " = _ob_in.read_Principal()";
	    break;
	}
	
	case RWCReadReturnDirect:
	    out << "\nreturn _ob_in.read_Principal();";
	    break;

	case RWCReadOut:
	case RWCReadStructElem:
	    out << '\n' << name << " = _ob_in.read_Principal();";
	    break;

	case RWCReadInOut:
	{
	    String_var norm = getTypeString(scope, type);
	    out << "\nCORBA::release(" << name << ");";
	    out << '\n' << name << " = " << norm << "::_nil();";
	    out << '\n' << name << " = _ob_in.read_Principal();";
	    break;
	}

	case RWCCountIn:
	case RWCCountInOut:
	    out << "\n_ob_cnt.count_Principal(" << name << ");";
	    break;

	case RWCWriteIn:
	case RWCWriteInOut:
	    out << "\n_ob_out.write_Principal(" << name << ");";
	    break;
	}
	
	break;
    }
    
    case tk_objref:
    {
	// ******************************************************
	// Object references
        // ******************************************************

	RepositoryId_var id = type -> id();

	if(strcmp(id, "IDL:omg.org/CORBA/Object:1.0") == 0)
	{
	    switch(rwc)
	    {
	    case RWCReadReturnAlloc:
	    {
		String_var var = getTypeString(scope, type, GetTypeVar);
		out << '\n' << var << ' ' << name
                    << " = _ob_in.read_Object();";
		break;
	    }
	    
	    case RWCReadReturnDirect:
		out << "\nreturn _ob_in.read_Object();";
		break;

	    case RWCReadOut:
	    case RWCReadStructElem:
		out << '\n' << name << " = _ob_in.read_Object();";
		break;
		
	    case RWCReadInOut:
	    {
		String_var norm = getTypeString(scope, type);
		out << "\nCORBA::release(" << name << ");";
		out << '\n' << name << " = " << norm << "::_nil();";
		out << '\n' << name << " = _ob_in.read_Object();";
		break;
	    }

	    case RWCCountIn:
	    case RWCCountInOut:
		out << "\n_ob_cnt.count_Object(" << name << ");";
		break;
		
	    case RWCWriteIn:
	    case RWCWriteInOut:
		out << "\n_ob_out.write_Object(" << name << ");";
		break;
	    }
	}
	else
	{
	    String_var norm = getTypeString(scope, type);

	    switch(rwc)
	    {
	    case RWCReadReturnAlloc:
	    {
		String_var var = getTypeString(scope, type, GetTypeVar);
		out << '\n' << var << ' ' << name << ';';
		out << '\n' << norm << "::_OB_unmarshal(" << name
                    << ".inout(), _ob_in);";
		break;
	    }

	    case RWCReadReturnDirect:
	    {
		String_var ptr = getTypeString(scope, type, GetTypePtr);
		out << '\n' << ptr << ' ' << name << " = " << norm
		    << "::_nil();";
		out << '\n' << norm << "::_OB_unmarshal(" << name
                    << ", _ob_in);";
		out << "\nreturn " << name << ';';
		break;
	    }

	    case RWCReadOut:
	    {
		out << '\n' << name << " = " << norm << "::_nil();";
		out << '\n' << norm << "::_OB_unmarshal(" << name
                    << ", _ob_in);";
		break;
	    }

	    case RWCReadInOut:
		out << '\n' << norm << "::_OB_unmarshal(" << name
                    << ", _ob_in);";
		break;

	    case RWCReadStructElem:
		out << '\n' << norm << "::_OB_unmarshal(" << name
                    << ".inout(), _ob_in);";
		break;

	    case RWCCountIn:
	    case RWCCountInOut:
		out << "\n_ob_cnt.count_Object(" << name << ");";
		break;
		
	    case RWCWriteIn:
	    case RWCWriteInOut:
		out << "\n_ob_out.write_Object(" << name << ");";
		break;
	    }
	}

	break;
    }
    
    case tk_abstract_interface:
    {
        // ******************************************************
        // Abstract interface
        // ******************************************************

        RepositoryId_var id = type -> id();
	String_var norm = getTypeString(scope, type);

        switch(rwc)
        {
        case RWCReadReturnAlloc:
        {
            String_var var = getTypeString(scope, type, GetTypeVar);
            out << '\n' << var << ' ' << name << ';';
	    out << '\n' << norm << "::_OB_unmarshal(" << name
		<< ".inout(), _ob_in);";
            break;
        }

        case RWCReadReturnDirect:
        {
            String_var ptr = getTypeString(scope, type, GetTypePtr);
            out << '\n' << ptr << ' ' << name << " = " << norm << "::_nil();";
	    out << '\n' << norm << "::_OB_unmarshal(" << name << ", _ob_in);";
            out << "\nreturn " << name << ';';
            break;
        }

        case RWCReadOut:
        {
            out << '\n' << name << " = " << norm << "::_nil();";
	    out << '\n' << norm << "::_OB_unmarshal(" << name << ", _ob_in);";
            break;
        }

        case RWCReadInOut:
	    out << '\n' << norm << "::_OB_unmarshal(" << name << ", _ob_in);";
            break;

        case RWCReadStructElem:
	    out << '\n' << norm << "::_OB_unmarshal(" << name
		<< ".inout(), _ob_in);";
            break;

        case RWCCountIn:
        case RWCCountInOut:
            out << "\n_ob_cnt.count_Abstract(" << name << ");";
            break;
            
        case RWCWriteIn:
        case RWCWriteInOut:
            out << "\n_ob_out.write_Abstract(" << name << ");";
            break;
        }

        break;
    }
    
    case tk_struct:
    case tk_union:
    {
	// ******************************************************
	// Structs and unions
        // ******************************************************

	bool variable = IdlIsVariable(type);
	String_var norm = getTypeString(scope, type);

	switch(rwc)
	{
	case RWCReadReturnAlloc:
	case RWCReadReturnDirect:
	    if(variable)
	    {
		String_var var = getTypeString(scope, type, GetTypeVar);
		out << '\n' << var << ' ' << name << " = new " << norm << ';';
		out << '\n' << norm << "::_OB_unmarshal(" << name
                    << ".inout(), _ob_in);";
		if(rwc == RWCReadReturnDirect)
		    out << "\nreturn " << name << "._retn();";
	    }
	    else
	    {
		String_var ret = getTypeString(scope, type,
						     GetTypeReturn);
		out << '\n' << ret << ' ' << name << ';';
		out << '\n' << norm << "::_OB_unmarshal(" << name
                    << ", _ob_in);";
		if(rwc == RWCReadReturnDirect)
		    out << "\nreturn " << name << ';';
	    }
	    break;

	case RWCReadOut:
	    if(variable)
	    {
		out << '\n' << name << " = new " << norm << ';';
		out << '\n' << norm << "::_OB_unmarshal(*" << name
		    << ", _ob_in);";
	    }
	    else
		out << '\n' << norm << "::_OB_unmarshal(" << name
		    << ", _ob_in);";
	    break;

	case RWCReadInOut:
	case RWCReadStructElem:
	    out << '\n' << norm << "::_OB_unmarshal(" << name << ", _ob_in);";
	    break;

	case RWCCountIn:
	case RWCCountInOut:
	    out << '\n' << norm << "::_OB_marshalCount(" << name
                << ", _ob_cnt);";
	    break;

	case RWCWriteIn:
	case RWCWriteInOut:
	    out << '\n' << norm << "::_OB_marshal(" << name << ", _ob_out);";
	    break;
	}

	break;
    }

    case tk_enum:
    {
	// ******************************************************
	// Enums
        // ******************************************************

	switch(rwc)
	{
	case RWCReadReturnAlloc:
	{
	    String_var ret = getTypeString(scope, type, GetTypeReturn);
	    out << '\n' << ret << ' ' << name << " = (" << ret
		<< ")_ob_in.read_ulong();";
	    break;
	}

	case RWCReadReturnDirect:
	{
	    String_var ret = getTypeString(scope, type, GetTypeReturn);
	    out << "\nreturn (" << ret << ")_ob_in.read_ulong();";
	    break;
	}

	case RWCReadOut:
	case RWCReadInOut:
	case RWCReadStructElem:
	{
	    String_var str = getTypeString(scope, type, GetTypeArray);
	    out << '\n' << name << " = (" << str << ")_ob_in.read_ulong();";
	    break;
	}

	case RWCCountIn:
	case RWCCountInOut:
	    out << "\n_ob_cnt.count_ulong();";
	    break;

	case RWCWriteIn:
	case RWCWriteInOut:
	    out << "\n_ob_out.write_ulong((CORBA::ULong)" << name << ");";
	    break;
	}

	break;
    }

    case tk_sequence:
    {
	// ******************************************************
	// Sequences
        // ******************************************************

	TypeCode_var contType = origType -> content_type();
	TypeCode_var origContType = GetOrigType(contType);

        String_var unique;

        //
        // If a unique name wasn't supplied, then try to compute it
        //
        if(type -> kind() == tk_alias)
        {
            if(uniqueType == 0)
                unique = getUnique(type -> id());
            else
                unique = uniqueType;
        }
        else
        {
            if(uniqueType == 0)
                unique = string_dup("");
            else
                unique = uniqueType;
        }

	switch(origContType -> kind())
	{
	case tk_null:
	case tk_void:
	    assert(false);
	    break;
	    
	case tk_short:
	case tk_long:
	case tk_ushort:
	case tk_ulong:
	case tk_float:
	case tk_double:
	case tk_boolean:
	case tk_char:
	case tk_octet:
	case tk_longlong:
	case tk_ulonglong:
	case tk_longdouble:
	case tk_wchar:
	{
	    const char* op = opsArray[(int)origContType -> kind()];
	
	    switch(rwc)
	    {
	    case RWCReadReturnAlloc:
	    case RWCReadReturnDirect:
            {
		assert(type -> kind() == tk_alias); // Only for aliases
                String_var var = getTypeString(scope, type, GetTypeVar);
		String_var norm = 
		    getTypeString(scope, type, GetTypeNormal, false, 0,
				  unique);
                out << '\n' << var << ' ' << name << " = new " << norm << ';';
		out << "\nCORBA::ULong _ob_len" << iter
		    << " = _ob_in.read_ulong();";
		out << '\n' << name << " -> length(_ob_len" << iter
		    << ");";
		out << "\n_ob_in.read_" << op << "_array(" << name
		    << " -> get_buffer(), _ob_len" << iter << ");";
		iter++;
                if(rwc == RWCReadReturnDirect)
                    out << "\nreturn " << name << "._retn();";
                break;
            }

	    case RWCReadOut:
	    {
		String_var norm = 
		    getTypeString(scope, type, GetTypeNormal, false, 0,
				  unique);
		out << '\n'  << name << " = new " << norm << ';';
		out << "\nCORBA::ULong _ob_len" << iter
		    << " = _ob_in.read_ulong();";
		out << '\n' << name << " -> length(_ob_len" << iter
		    << ");";
		out << "\n_ob_in.read_" << op << "_array(" << name
		    << " -> get_buffer(), _ob_len" << iter << ");";
		iter++;
		break;
	    }

	    case RWCReadInOut:
	    case RWCReadStructElem:
		out << "\nCORBA::ULong _ob_len" << iter
		    << " = _ob_in.read_ulong();";
		out << '\n' << name << ".length(_ob_len" << iter << ");";
		out << "\n_ob_in.read_" << op << "_array(" << name
		    << ".get_buffer(), _ob_len" << iter << ");";
		iter++;
		break;

	    case RWCCountIn:
	    case RWCCountInOut:
		out << "\nCORBA::ULong _ob_len" << iter << " = " << name
		    << ".length();";
		out << "\n_ob_cnt.count_ulong();";
		if(origContType -> kind() == tk_wchar)
		    out << "\n_ob_cnt.count_wchar_array(" << name
			<< ".get_buffer(), _ob_len" << iter << ");";
		else
		    out << "\n_ob_cnt.count_" << op << "_array(_ob_len"
                        << iter << ");";
		iter++;
		break;

	    case RWCWriteIn:
	    case RWCWriteInOut:
		out << "\nCORBA::ULong _ob_len" << iter << " = " << name
		    << ".length();";
		out << "\n_ob_out.write_ulong(_ob_len" << iter << ");";
		out << "\n_ob_out.write_" << op << "_array(" << name
		    << ".get_buffer(), _ob_len" << iter << ");";
		iter++;
		break;
	    }
	
	    break;
	}
	
	case tk_enum:
	{
	    switch(rwc)
	    {
	    case RWCReadReturnAlloc:
	    case RWCReadReturnDirect:
            {
		assert(type -> kind() == tk_alias); // Only for aliases
                String_var var = getTypeString(scope, type, GetTypeVar);
		String_var norm = 
		    getTypeString(scope, type, GetTypeNormal, false, 0,
				  unique);
                out << '\n' << var << ' ' << name << " = new " << norm << ';';
		out << "\nCORBA::ULong _ob_len" << iter
		    << " = _ob_in.read_ulong();";
		out << '\n' << name << " -> length(_ob_len" << iter
		    << ");";
		out << "\n_ob_in.read_ulong_array((CORBA::ULong*)" << name
		    << " -> get_buffer(), _ob_len" << iter << ");";
		iter++;
                if(rwc == RWCReadReturnDirect)
                    out << "\nreturn " << name << "._retn();";
                break;
            }

	    case RWCReadOut:
	    {
		String_var norm = 
		    getTypeString(scope, type, GetTypeNormal, false, 0,
				  unique);
		out << '\n'  << name << " = new " << norm << ';';
		out << "\nCORBA::ULong _ob_len" << iter
		    << " = _ob_in.read_ulong();";
		out << '\n' << name << " -> length(_ob_len" << iter
		    << ");";
		out << "\n_ob_in.read_ulong_array((CORBA::ULong*)" << name
		    << " -> get_buffer(), _ob_len" << iter << ");";
		iter++;
		break;
	    }

	    case RWCReadInOut:
	    case RWCReadStructElem:
		out << "\nCORBA::ULong _ob_len" << iter
		    << " = _ob_in.read_ulong();";
		out << '\n' << name << ".length(_ob_len" << iter << ");";
		out << "\n_ob_in.read_ulong_array((CORBA::ULong*)" << name
		    << ".get_buffer(), _ob_len" << iter << ");";
		iter++;
		break;

	    case RWCCountIn:
	    case RWCCountInOut:
		out << "\nCORBA::ULong _ob_len" << iter << " = " << name
		    << ".length();";
		out << "\n_ob_cnt.count_ulong();";
		out << "\n_ob_cnt.count_ulong_array(_ob_len" << iter << ");";
		iter++;
		break;

	    case RWCWriteIn:
	    case RWCWriteInOut:
		out << "\nCORBA::ULong _ob_len" << iter << " = " << name
		    << ".length();";
		out << "\n_ob_out.write_ulong(_ob_len" << iter << ");";
		out << "\n_ob_out.write_ulong_array((const CORBA::ULong*)"
		    << name << ".get_buffer(), _ob_len" << iter << ");";
		iter++;
		break;
	    }
	
	    break;
	}
	
	case tk_any:
	case tk_TypeCode:
	case tk_Principal:
	case tk_objref:
	case tk_struct:
	case tk_union:
	case tk_string:
	case tk_sequence:
	case tk_array:
	case tk_except:
	case tk_wstring:
	case tk_fixed:
	case tk_value:
	case tk_value_box:
	case tk_abstract_interface:
	{
	    String_var obb = string_dup("_ob_buf");
	    obb += iter;
	    String_var obi = string_dup("_ob_i");
	    obi += iter;
	    String_var obl = string_dup("_ob_len");
	    obl += iter;
	    String_var elem = obb;
	    elem += '[';
	    elem += obi;
	    elem += ']';

	    String_var obbp = string_dup("* ");
	    obbp += obb;
	    String_var cobbp = string_dup("const * ");
	    cobbp += obb;

	    String_var bufDef;
	    String_var constBufDef;

	    if(origContType -> kind() == tk_TypeCode ||
	       origContType -> kind() == tk_Principal ||
	       origContType -> kind() == tk_objref ||
	       origContType -> kind() == tk_value ||
	       origContType -> kind() == tk_value_box ||
	       origContType -> kind() == tk_abstract_interface)
	    {
		bufDef = getTypeString(scope, contType, GetTypePtr,
				       false, obbp, unique);
		constBufDef = getTypeString(scope, contType, GetTypePtr,
					    false, cobbp, unique);
	    }
	    else if(origContType -> kind() == tk_string)
	    {
		bufDef = getTypeString(scope, contType, GetTypeNormal,
				       false, obbp, unique);
		constBufDef = string_dup("const char* const *");
		constBufDef += obb;
	    }
	    else if(origContType -> kind() == tk_wstring)
	    {
		bufDef = getTypeString(scope, contType, GetTypeNormal,
				       false, obbp, unique);
		constBufDef = string_dup("const wchar_t* const *");
		constBufDef += obb;
	    }
	    else if(origContType -> kind() == tk_array)
	    {
		bufDef = getTypeString(scope, contType, GetTypeReturn,
				       false, obbp, unique);
		constBufDef = getTypeString(scope, contType, GetTypeReturn,
					    false, cobbp, unique);
	    }
	    else
	    {
		bufDef = getTypeString(scope, contType, GetTypeNormal,
				       false, obbp, unique);
		constBufDef = getTypeString(scope, contType, GetTypeNormal,
					    false, cobbp, unique);
	    }

	    switch(rwc)
	    {
	    case RWCReadReturnAlloc:
	    case RWCReadReturnDirect:
	    {
		assert(type -> kind() == tk_alias); // Only for aliases
                String_var var = getTypeString(scope, type, GetTypeVar);
		String_var norm = 
		    getTypeString(scope, type, GetTypeNormal, false, 0,
				  unique);
                out << '\n' << var << ' ' << name << " = new " << norm << ';';
		out << "\nCORBA::ULong " << obl
		    << " = _ob_in.read_ulong();";
		out << '\n' << name << " -> length(" << obl << ");";
		out << '\n' << bufDef << " = " << name
		    << " -> get_buffer();";
		out << "\nfor(CORBA::ULong " << obi << " = 0 ; " << obi
		    << " < " << obl << " ; " << obi << "++)";
		out.start();
		iter++;
		writeReadWriteCount(scope, contType, elem, iter,
				    RWCReadInOut, unique, out);
		out.end();
                if(rwc == RWCReadReturnDirect)
                    out << "\nreturn " << name << "._retn();";
		break;
	    }

	    case RWCReadOut:
	    {
		String_var norm = 
		    getTypeString(scope, type, GetTypeNormal, false, 0,
				  unique);
		out << '\n'  << name << " = new " << norm << ';';
		out << "\nCORBA::ULong " << obl
		    << " = _ob_in.read_ulong();";
		out << '\n' << name << " -> length(" << obl << ");";
		out << '\n' << bufDef << " = " << name
		    << " -> get_buffer();";
		out << "\nfor(CORBA::ULong " << obi << " = 0 ; " << obi
		    << " < " << obl << " ; " << obi << "++)";
		out.start();
		iter++;
		writeReadWriteCount(scope, contType, elem, iter,
				    RWCReadInOut, unique, out);
		out.end();
		break;
	    }

	    case RWCReadInOut:
	    case RWCReadStructElem:
	    {
		out << "\nCORBA::ULong " << obl
		    << " = _ob_in.read_ulong();";
		out << '\n' << name << ".length(" << obl << ");";
		out << '\n' << bufDef << " = " << name << ".get_buffer();";
		out << "\nfor(CORBA::ULong " << obi << " = 0 ; " << obi
		    << " < " << obl << " ; " << obi << "++)";
		iter++;
		out.start();
		writeReadWriteCount(scope, contType, elem, iter,
				    RWCReadInOut, unique, out);
		out.end();
		break;
	    }

	    case RWCCountIn:
	    case RWCCountInOut:
	    {
		out << "\nCORBA::ULong " << obl << " = " << name
		    << ".length();";
		out << "\n_ob_cnt.count_ulong();";
                //
                // Determine if we actually need the buffer in order
                // to count its size - avoids compiler warnings about
                // unused variables
                //
		if(needDataForCount(type))
		    out << '\n' << constBufDef << " = " << name
			<< ".get_buffer();";
		out << "\nfor(CORBA::ULong " << obi << " = 0 ; " << obi
		    << " < " << obl << " ; " << obi << "++)";
		iter++;
		out.start();
		writeReadWriteCount(scope, contType, elem, iter,
				    RWCCountIn, unique, out);
		out.end();
		break;
	    }

	    case RWCWriteIn:
	    case RWCWriteInOut:
	    {
		out << "\nCORBA::ULong " << obl << " = " << name
		    << ".length();";
		out << "\n_ob_out.write_ulong(" << obl << ");";
		out << '\n' << constBufDef << " = " << name
		    << ".get_buffer();";
		out << "\nfor(CORBA::ULong " << obi << " = 0 ; " << obi
		    << " < " << obl << " ; " << obi << "++)";
		iter++;
		out.start();
		writeReadWriteCount(scope, contType, elem, iter,
				    RWCWriteIn, unique, out);
		out.end();
		break;
	    }
	    }
	    
	    break;
	}

	case tk_native:
	    assert(false); // TODO: Must be added
	    break;
	    
	case tk_alias:
	    assert(false);
	    break;
	}
	
	break;
    }

    case tk_array:
    {
	// ******************************************************
	// Arrays
        // ******************************************************

	TypeCode_var contType;
	ULong length;
	getArrayType(origType, contType.out(), length);
	TypeCode_var origContType = GetOrigType(contType);
	// TODO: contType is already the same as origContType

	String_var cont = getTypeString(scope, contType, GetTypeArray);

	String_var cast = string_dup("");
	String_var constCast = string_dup("");
	String_var backCast = string_dup("");
	String_var constBackCast = string_dup("");
	if(length != origType -> length())
	{
	    String_var ret = getTypeString(scope, type, GetTypeReturn,
					   false, 0, uniqueType);

	    cast += '(';
	    cast += ret;
	    cast += ')';

	    constCast += "(const ";
	    constCast += ret;
	    constCast += ')';

	    backCast += '(';
	    backCast += cont;
	    backCast += "*)";

	    constBackCast += "(const ";
	    constBackCast += cont;
	    constBackCast += "*)";
	}

	switch(origContType -> kind())
	{
	case tk_null:
	case tk_void:
	    assert(false);
	    break;

	case tk_short:
	case tk_long:
	case tk_ushort:
	case tk_ulong:
	case tk_float:
	case tk_double:
	case tk_boolean:
	case tk_char:
	case tk_octet:
	case tk_enum:
	case tk_longlong:
	case tk_ulonglong:
	case tk_longdouble:
	case tk_wchar:
	{
	    const char* op;

	    if(origContType -> kind() == tk_enum)
	    {
		backCast = string_dup("(CORBA::ULong*)");
		constBackCast = string_dup("(const CORBA::ULong*)");
		op = "ulong";
	    }
	    else
		op = opsArray[(int)origContType -> kind()];
	    
	    switch(rwc)
	    {
	    case RWCReadReturnAlloc:
	    case RWCReadReturnDirect:
	    {
		String_var var =
		    getTypeString(scope, type, GetTypeVar, false, name);
		out << '\n' << var << " = " << cast << "new " << cont
		    << '[' << length << "];";
		out << "\n_ob_in.read_" << op << "_array(" << backCast
		    << name << ".inout(), " << length << ");";
		if(rwc == RWCReadReturnDirect)
		    out << "\nreturn " << name << "._retn();";
		break;
	    }

	    case RWCReadOut:
	    case RWCReadInOut:
	    case RWCReadStructElem:
		out << "\n_ob_in.read_" << op << "_array(" << backCast
		    << name << ", " << length << ");";
		break;
		
	    case RWCCountIn:
	    case RWCCountInOut:
		if(origContType -> kind() == tk_wchar)
		    out << "\n_ob_cnt.count_wchar_array(" << constBackCast
			<< name << ", " << length << ");";
		else
		    out << "\n_ob_cnt.count_" << op << "_array("
			<< length << ");";
		break;
		
	    case RWCWriteIn:
	    case RWCWriteInOut:
		out << "\n_ob_out.write_" << op << "_array(" << constBackCast
		    << name << ", " << length << ");";
		break;
	    }
	    
	    break;
	}

	case tk_any:
	case tk_TypeCode:
	case tk_Principal:
	case tk_objref:
	case tk_struct:
	case tk_union:
	case tk_string:
	case tk_sequence:
	case tk_except:
	case tk_wstring:
	case tk_fixed:
	case tk_value:
	case tk_value_box:
	case tk_abstract_interface:
	{
	    bool variable = IdlIsVariable(type);

	    switch(rwc)
	    {
	    case RWCReadReturnAlloc:
	    case RWCReadReturnDirect:
	    case RWCReadOut:
	    case RWCReadInOut:
	    case RWCReadStructElem:
	    {
		if(rwc == RWCReadReturnAlloc || rwc == RWCReadReturnDirect)
		{
		    String_var var =
			getTypeString(scope, type, GetTypeVar, false, name);
		    out << '\n' << cont << "* _ob_ar" << iter << " = new "
			<< cont << '[' << length << "];";
		    out << '\n' << var << " = " << cast << "_ob_ar" << iter
			<< ';';
		}
		else if(rwc == RWCReadOut && variable)
		{
		    out << '\n' << cont << "* _ob_ar" << iter << " = new "
			<< cont << '[' << length << "];";
		    out << '\n' << name << " = " << cast << "_ob_ar" << iter
			<< ';';
		}
		else
		{
		    out << '\n' << cont << "* _ob_ar" << iter << " = "
			<< backCast << name << ';';
		}

		out << "\nfor(CORBA::ULong _ob_i" << iter << " = 0 ; _ob_i"
		    << iter << " < " << length << " ; _ob_i" << iter << "++)";
		String_var elem = string_dup("_ob_ar");
		elem += iter;
		elem += "[_ob_i";
		elem += iter;
		elem += ']';
		iter++;
		out.start();
		writeReadWriteCount(scope, contType, elem, iter,
				    RWCReadStructElem, uniqueType, out);
		out.end();
		if(rwc == RWCReadReturnDirect)
		    out << "\nreturn " << name << "._retn();";
		break;
	    }

	    case RWCCountIn:
	    case RWCCountInOut:
	    case RWCWriteIn:
	    case RWCWriteInOut:
	    {
		out << "\nconst " << cont << "* _ob_ar" << iter << " = "
		    << constBackCast << name << ';';
		out << "\nfor(CORBA::ULong _ob_i" << iter << " = 0 ; _ob_i"
		    << iter << " < " << length << " ; _ob_i" << iter << "++)";
		String_var elem = string_dup("_ob_ar");
		elem += iter;
		elem += "[_ob_i";
		elem += iter;
		elem += ']';
		iter++;
		out.start();
		if(rwc == RWCCountIn || rwc == RWCCountInOut)
		    writeReadWriteCount(scope, contType, elem, iter,
					RWCCountIn, uniqueType, out);
		else
		    writeReadWriteCount(scope, contType, elem, iter,
					RWCWriteIn, uniqueType, out);
		out.end();
		break;
	    }
	    }

	    break;
	}

        case tk_native:
	    assert(false); // TODO: Must be added
	    break;

	case tk_array:
	case tk_alias:
	    assert(false);
	    break;
	}

	break;
    }

    case tk_except:
    {
	// ******************************************************
	// Exceptions
        // ******************************************************

	assert(false); // This is never needed for exceptions
	break;
    }

    case tk_value:
    case tk_value_box:
    {
        // ******************************************************
        // Valuetypes and value boxes
        // ******************************************************

	String_var norm = getTypeString(scope, type);

	switch(rwc)
	{
	case RWCReadReturnAlloc:
	{
	    String_var var = getTypeString(scope, type, GetTypeVar);
	    out << '\n' << var << ' ' << name << ';';
	    out << '\n' << norm << "::_OB_unmarshal(" << name
		<< ".inout(), _ob_in);";
	    break;
	}

	case RWCReadReturnDirect:
	{
	    String_var ptr = getTypeString(scope, type, GetTypePtr);
	    out << '\n' << ptr << ' ' << name << " = 0;";
	    out << '\n' << norm << "::_OB_unmarshal(" << name
		<< ", _ob_in);";
	    out << "\nreturn " << name << ';';
	    break;
	}

	case RWCReadOut:
	    out << '\n' << name << " = 0;";
	    out << '\n' << norm << "::_OB_unmarshal(" << name
		<< ", _ob_in);";
	    break;

	case RWCReadInOut:
	    out << '\n' << norm << "::_OB_unmarshal(" << name
		<< ", _ob_in);";
	    break;

	case RWCReadStructElem:
	    out << '\n' << norm << "::_OB_unmarshal(" << name
		<< ".inout(), _ob_in);";
	    break;

	case RWCCountIn:
	case RWCCountInOut:
	    out << '\n' << norm << "::_OB_marshalCount(" << name
		<< ", _ob_cnt);";
	    break;

	case RWCWriteIn:
	case RWCWriteInOut:
	    out << '\n' << norm << "::_OB_marshal(" << name
		<< ", _ob_out);";
	    break;
	}

        break;
    }

    case tk_native:
	//
	// TODO: Marshal exception
	///
	out << "\n//";
	out << "\n// TODO: Marshal exception";
	out << "\n//";
	switch(rwc)
	{
	case RWCReadReturnAlloc:
	{
	    String_var var = getTypeString(scope, type, GetTypeNormal);
	    out << '\n' << var << ' ' << name << ";";
	    break;
	}
	
	case RWCReadReturnDirect:
	    out << "\nreturn 0;";
	    break;

	case RWCReadOut:
	case RWCReadStructElem:
	case RWCReadInOut:
	case RWCCountIn:
	case RWCCountInOut:
	case RWCWriteIn:
	case RWCWriteInOut:
	    out << "\nassert(false);";
	    break;
	}
	break;

    case tk_alias:
	assert(false);
    }
}

// ----------------------------------------------------------------------
// Write array copy code
// ----------------------------------------------------------------------

void
IdlCPPGenerator::writeArrayCopy(const char* scope,
                                TypeCode_ptr type,
                                const char* target,
                                const char* source,
                                const char* index,
                                IdlPrettyPrint& out)
{
    TypeCode_var origType = GetOrigType(type);
    assert(origType -> kind() == tk_array);

    if(type -> kind() == tk_alias)
    {
        ScopedName_var absolute = getTypeString("", type, GetTypeNormal);
        out << '\n' << absolute << "_copy(" << source << ", " << target
            << ");";
    }
    else
    {
        bool variable = IdlIsVariable(origType);

        TypeCode_var contentType;
        ULong length;
        getArrayType(origType, contentType.out(), length);

        String_var cont =
            getTypeString(scope, contentType, GetTypeArray);

        if(variable)
        {        
            out << "\nfor(CORBA::ULong " << index << " = 0 ; " << index << " < "
                << length << " ; " << index << "++)";
            out.inc();
            if(length != origType -> length())
            {
                out << "\n((" << cont << "*)"
                    << target << ")[" << index << "] = ((" << cont << "*)"
                    << source << ")[" << index << "];";
            }
            else
            {
                out << '\n' << target << '[' << index << "] = " << source
                    << '[' << index << "];";
            }
            out.dec();
        }
        else
        {
            out << "\nmemcpy(" << target << ", " << source << ", " << length
                << " * sizeof(" << cont << "));";
        }
    }
}

// ----------------------------------------------------------------------
// Write array allocation code
// ----------------------------------------------------------------------

void
IdlCPPGenerator::writeArrayAlloc(const char* scope,
                                 TypeCode_ptr type,
                                 const char* arg,
                                 IdlPrettyPrint& out)
{
    TypeCode_var origType = GetOrigType(type);
    assert(origType -> kind() == tk_array);

    if(type -> kind() == tk_alias)
    {
        ScopedName_var absolute = getTypeString("", type, GetTypeNormal);
        out << '\n' << arg << " = " << absolute << "_alloc();";
    }
    else
    {
        TypeCode_var contentType;
        ULong length;
        getArrayType(origType, contentType.out(), length);

        String_var cont =
            getTypeString(scope, contentType, GetTypeArray);
        String_var cast = getTypeString(scope, type, GetTypeReturn);

        out << '\n' << arg << " = ";
        if(length != origType -> length())
            out << '(' << cast << ") ";
        out << "new " << cont << '[' << length << "];";
    }
}

// ----------------------------------------------------------------------
// Write array duplication code
// ----------------------------------------------------------------------

void
IdlCPPGenerator::writeArrayDup(const char* scope,
                               TypeCode_ptr type,
                               const char* target,
                               const char* source,
                               const char* index,
                               IdlPrettyPrint& out)
{
    TypeCode_var origType = GetOrigType(type);
    assert(origType -> kind() == tk_array);

    if(type -> kind() == tk_alias)
    {
        ScopedName_var absolute = getTypeString("", type, GetTypeNormal);
        out << '\n' << target << " = " << absolute << "_dup(" << source
            << ");";
    }
    else
    {
        //
        // Allocate & copy
        //
        writeArrayAlloc(scope, type, target, out);
        writeArrayCopy(scope, type, target, source, index, out);
    }
}

// ----------------------------------------------------------------------
// Write array destruction code
// ----------------------------------------------------------------------

void
IdlCPPGenerator::writeArrayFree(const char* scope,
                                TypeCode_ptr type,
                                const char* arg,
                                IdlPrettyPrint& out)
{
    TypeCode_var origType = GetOrigType(type);
    assert(origType -> kind() == tk_array);

    TypeCode_var contentType;
    ULong length;
    getArrayType(origType, contentType.out(), length);

    if(type -> kind() == tk_alias)
    {
        ScopedName_var absolute = getTypeString("", type, GetTypeNormal);
        out << '\n' << absolute << "_free(" << arg << ");";
    }
    else
    {
        String_var cont =
            getTypeString(scope, contentType, GetTypeArray);
        out << "\ndelete [] (" << cont << "*)" << arg << ';';
    }
}

// ----------------------------------------------------------------------
// Check if code needs to be generated for a specific repository id
// ----------------------------------------------------------------------

bool
IdlCPPGenerator::check(const char* id)
{
    if(!partial_)
	return true;

    for(ULong i = 0 ; i < idSeq_.length() ; i++)
	if(strcmp(id, idSeq_[i]) == 0)
	    return true;

    return false;
}

// ----------------------------------------------------------------------
// Get absolute name
// ----------------------------------------------------------------------

char*
IdlCPPGenerator::getAbsolute(Contained_ptr contained, const char* scope)
{
    ScopedName_var abs = contained -> absolute_name();
    String_var name = fixKwd(abs);
    if(scope != 0)
        IdlRemoveScope(scope, name.inout());
    return name._retn();
}

char*
IdlCPPGenerator::getAbsolute(const char* id, const char* scope)
{
    String_var rc;

    if(!absMap_ -> get(id, rc))
    {
        Contained_var contained = repository_ -> lookup_id(id);
        assert(!is_nil(contained));
        
        rc = getAbsolute(contained);
        absMap_ -> put(id, string_dup(rc));
    }

    if(scope != 0)
        IdlRemoveScope(scope, rc.inout());

    return rc._retn();
}

// ----------------------------------------------------------------------
// Get unique type names (for sequence templates)
// ----------------------------------------------------------------------

char*
IdlCPPGenerator::getUnique(Contained_ptr contained)
{
    assert(contained -> def_kind() == dk_Alias);

    Identifier_var name = string_dup("OBUnique_");
    Identifier_var tmp = contained -> name();
    name += tmp;

    Container_var container = contained -> defined_in();
    Contained_var contained2 = Contained::_narrow(container);

    if(is_nil(contained2))
    {
	return string_dup(name);
    }
    else
    {
	ScopedName_var s = getAbsolute(contained2); // Not getUnique()!
	char* result = string_alloc(strlen(s) + 2 + strlen(name));
	strcpy(result, s);
	strcat(result, "::");
	strcat(result, name);
	return result;
    }
}

char*
IdlCPPGenerator::getUnique(const char* id)
{
    String_var rc;
    if(!uniMap_ -> get(id, rc))
    {
        Contained_var contained = repository_ -> lookup_id(id);
        assert(!is_nil(contained));
        
        rc = getUnique(contained);
        uniMap_ -> put(id, string_dup(rc));
    }
    return rc._retn();
}

// ----------------------------------------------------------------------
// Get scope
// ----------------------------------------------------------------------

char*
IdlCPPGenerator::getScope(Container_ptr container)
{
    String_var scope;

    Contained_var contained = Contained::_narrow(container);
    if(is_nil(contained))
    {
        scope = string_dup("");
    }
    else
    {
        RepositoryId_var id = contained -> id();
        ScopedName_var s = getAbsolute(id);
        scope = string_alloc(strlen(s) + 2);
        strcpy(scope.inout(), s);
        strcat(scope.inout(), "::");
    }

    return scope._retn();
}

// ----------------------------------------------------------------------
// Determines if the object resides at the top level
// ----------------------------------------------------------------------

bool
IdlCPPGenerator::isTopLevel(Contained_ptr c)
{
    assert(!is_nil(c));

    Container_var container = c -> defined_in();
    assert(!is_nil(container));

    //
    // If the container is not the repository, then the Contained object
    // must reside inside another scope
    //
    return(container -> def_kind() == dk_Repository);
}

// ----------------------------------------------------------------------
// Get a prefixed name
// ----------------------------------------------------------------------

char*
IdlCPPGenerator::getPrefixedName(const char* topPrefix, Contained_ptr c,
                                 const char* namePrefix)
{
    assert(!is_nil(c));

    Identifier_var n = c -> name();
    n = fixKwd(n);

    Identifier_var name;
    if(namePrefix)
    {
        name = namePrefix;
        name += "_";
        name += n;
    }
    else
        name = n;

    //
    // The top prefix is only prepended if the Contained object resides
    // at the top level
    //
    if(isTopLevel(c))
    {
        String_var s = topPrefix;
        s += "_";
        s += name;
        return s._retn();
    }
    else
    {
        return name._retn();
    }
}

// ----------------------------------------------------------------------
// Get an absolute name with a top prefix and possibly a name prefix
// ----------------------------------------------------------------------

char*
IdlCPPGenerator::getPrefixedAbsolute(const char* topPrefix, const char* id,
                                     const char* scope, const char* namePrefix)
{
    if(namePrefix)
    {
        Contained_var contained = repository_ -> lookup_id(id);
        assert(!is_nil(contained));
        return getPrefixedAbsolute(topPrefix, contained, scope, namePrefix);
    }
    else
    {
        ScopedName_var abs = getAbsolute(id);
        String_var s = topPrefix;
        s += "_";
        s += abs;
        return s._retn();
    }
}

char*
IdlCPPGenerator::getPrefixedAbsolute(const char* topPrefix, Contained_ptr c,
                                     const char* scope, const char* namePrefix)
{
    assert(!is_nil(c));

    String_var s;

    if(namePrefix)
    {
        Identifier_var name = c -> name();
        name = fixKwd(name);
        Container_var cont = c -> defined_in();
        assert(!is_nil(cont));
        ScopedName_var contScope = getScope(cont);
        ULong len = strlen(topPrefix) + strlen(contScope) + strlen(namePrefix)
                  + strlen(name) + 2;
        s = string_alloc(len);
        strcpy(s.inout(), topPrefix);
        strcat(s.inout(), "_");
        strcat(s.inout(), contScope);
        strcat(s.inout(), namePrefix);
        strcat(s.inout(), "_");
        strcat(s.inout(), name);
    }
    else
    {
        RepositoryId_var id = c -> id();
        ScopedName_var abs = getAbsolute(id);
        s = topPrefix;
        s += "_";
        s += abs;
    }

    if(scope != 0 && strlen(scope) > 0)
    {
        String_var prefScope = topPrefix;
        prefScope += "_";
        prefScope += scope;
        IdlRemoveScope(prefScope, s.inout());
    }

    return s._retn();
}

// ----------------------------------------------------------------------
// Converts a scoped name into a flattened name
// ----------------------------------------------------------------------

char*
IdlCPPGenerator::getFlattenedName(const char* name)
{
    ULong len = strlen(name);
    char* result = string_alloc(len);
    result[0] = '\0';
    const char* start = name;

    //
    // Skip leading colons
    //
    if(strncmp(start, "::", 2) == 0)
        start += 2;
    assert(*start != ':');

    ULong n = 0;
    char* cp = strchr(start, ':');
    while(cp)
    {
        assert(*(cp + 1) == ':');
        memcpy(&result[n], start, cp - start);
        n += cp - start;
        start = cp + 2;
        if(*start)
        {
            result[n] = '_';
            n++;
        }
        cp = strchr(start, ':');
    }

    strcpy(&result[n], start);
    return result;
}

// ----------------------------------------------------------------------
// Print union labels
// ----------------------------------------------------------------------

void
IdlCPPGenerator::printUnionLabels(const IdlUnionMemberInfo& info,
				  IdlPrettyPrint& out,
				  const char* cast)
{
    out.dec();

    if(is_nil(info.type))
    {
	assert(info.isDefault);
	out << "\ndefault:";
    }
    else
    {
	if(info.isDefault)
	    out << "\ndefault:";
	else
	{
	    ULong i;
	    
	    for(i = 0 ; i < info.pLabels.length() ; i++)
		out << "\ncase " << cast << info.pLabels[i] << ':';
	    
	    for(i = 0 ; i < info.nLabels.length() ; i++)
		out << "\ncase " << cast << '-' << info.nLabels[i] << ':';
	}
    }

    out.inc();
}
