// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <Types.h>
#include <Util.h>
#include <GenUtil.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#include <GenJava.h>

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Generate Java code
// ----------------------------------------------------------------------

int
IdlJavaGenerator::genJava(Container_ptr container, IdlPrettyPrint& out,
			  GenMode mode)
{
    //
    // Get container contents descriptions
    //
    Container::DescriptionSeq_var seq =
	container -> describe_contents(dk_all, true, -1);

    ULong i, j, k;

    int status = 0;
    for(i = 0 ; status == 0 && i < seq -> length() ; i++)
    {
	//
	// Operations interface holds operations and attributes,
	// signature interface holds the rest
	//
	if(mode == GenModeOperations)
	{
	    if(seq[i].kind != dk_Operation &&
	       seq[i].kind != dk_Attribute)
		continue;
	}
	else if(mode == GenModeSignatures)
	{
	    if(seq[i].kind == dk_Operation ||
	       seq[i].kind == dk_Attribute)
		continue;
	}

	//
	// Get container
	//
	Container_var cont =
	    Container::_narrow(seq[i].contained_object);

	switch(seq[i].kind)
	{
	case dk_Constant:
	{
	    //
	    // Constant
	    //
	    ConstantDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixName(desc -> name);
	    String_var package = getPackage(desc -> id);

	    //
	    // Constants declared within an interface are mapped to a
	    // static member, all other constants are mapped to a separate
	    // class.
	    //
	    if(container -> def_kind() != dk_Interface)
	    {
		PrettyPrint out(prog_, name, package, outDir_);
		if(!out.good())
		    return 1;

		comment(desc -> id, out);
		docComment(desc -> id, out);

		out.sep();
		out << "\npublic interface " << name;
		out.start();
		String_var norm = getTypeString(package, desc -> type);
		out << "\npublic static final " << norm
		    << " value = ";
		writeConstant(desc -> value, out);
		out << ';';
		out.end();
	    }
	    else
	    {
		comment(desc -> id, out);
		docComment(desc -> id, out);

		String_var norm = getTypeString(package, desc -> type);
		out.sep();
		out << "\npublic static final " << norm
		    << ' ' << name << " = ";
		writeConstant(desc -> value, out);
		out << ';';
	    }

	    break;
	}

	case dk_Module:
	{
	    //
	    // Module
	    //
	    ModuleDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    // Don't add this here!!!
	    // if(!check(desc -> id))
	    //     continue;

	    assert(!is_nil(cont));
	    status = genJava(cont, out, mode);

	    break;
	}

	case dk_Interface:
	{
	    //
	    // Interface
	    //
 	    InterfaceDef_var interf =
		InterfaceDef::_narrow(seq[i].contained_object);
 	    assert(!is_nil(interf));
 	    InterfaceDef::FullInterfaceDescription_var desc =
 		interf -> describe_interface();

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixName(desc -> name);
	    String_var package = getPackage(desc -> id);

	    unsigned int indent;

	    //
	    // Operations interface for the interface
	    //
	    String_var opName = name;
	    opName += "Operations";
	    PrettyPrint outOp(prog_, opName, package, outDir_);
	    if(!outOp.good())
		return 1;
	    
	    comment(desc -> id, outOp);
	    docComment(desc -> id, outOp);

	    outOp.sep();
	    outOp << "\npublic interface " << opName;

	    if(desc -> base_interfaces.length() > 0)
	    {
		outOp << " extends ";

		indent = outOp.getIndent();
		outOp.setIndent(outOp.getPos());

		for(j = 0 ; j < desc -> base_interfaces.length() ; j++)
		{
		    const char* id = desc -> base_interfaces[j];
		    ScopedName_var absolute = getAbsolute(id);
		    removePackage(package, absolute.inout());

		    if(j > 0)
			outOp << ",\n";

		    outOp << absolute << "Operations";
		}
	
		outOp.setIndent(indent);
	    }

	    outOp.start();
	    status = genJava(cont, outOp, GenModeOperations);
	    outOp.end();

	    //
	    // Signature interface for the interface
	    //
	    PrettyPrint outSig(prog_, name, package, outDir_);
	    if(!outSig.good())
		return 1;
	    
	    comment(desc -> id, outSig);
	    docComment(desc -> id, outSig);

	    outSig.sep();
	    outSig << "\npublic interface " << name << " extends ";

	    indent = outSig.getIndent();
	    outSig.setIndent(outSig.getPos());

	    outSig << opName;

	    if(desc -> base_interfaces.length() > 0)
	    {
		for(j = 0 ; j < desc -> base_interfaces.length() ; j++)
		{
		    const char* id = desc -> base_interfaces[j];
		    ScopedName_var absolute = getAbsolute(id);
		    removePackage(package, absolute.inout());

		    outSig << ",\n" << absolute;
		}
	    }
	    else
		outSig << ",\norg.omg.CORBA.Object";

	    outSig << ",\norg.omg.CORBA.portable.IDLEntity";

	    outSig.setIndent(indent);

	    outSig.start();
	    status = genJava(cont, outSig, GenModeSignatures);
	    outSig.end();

	    //
	    // Stub for the interface
	    //
	    if(!local_ && noStreams_)
	    {
		String_var stub = string_dup("StubFor");
		stub += name;
		PrettyPrint outStub(prog_, stub, package, outDir_);
		if(!outStub.good())
		    return 1;

		comment(desc -> id, outStub);

		outStub.sep();
		outStub << "\npublic class " << stub << ' ';
		indent = outStub.getIndent();
		outStub.setIndent(outStub.getPos());
		outStub << "extends org.omg.CORBA.portable.ObjectImpl";
		outStub << "\nimplements " << name;
		outStub.setIndent(indent);

		outStub.start();

		ContainedSeq_var attributes =
		    interf -> contents(dk_Attribute, false);

		ContainedSeq_var operations =
		    interf -> contents(dk_Operation, false);

		InterfaceDefSeq allBaseInterfaces;
		IdlGetAllBaseInterfaces(interf, allBaseInterfaces);

		outStub << "\nstatic final String[] _ob_ids_ =";
		outStub.start();
		outStub << "\n\"" << desc -> id << "\",";
		for(j = 0 ; j < allBaseInterfaces.length() ; j++)
		{
		    RepositoryId_var id =
			allBaseInterfaces[j] -> id();
		    outStub << "\n\"" << id << '"';
		    if(j + 1 < allBaseInterfaces.length())
			outStub << ',';
		}
		outStub.end();
		outStub << ';';

		outStub.sep();
		outStub << "\npublic String[]\n_ids()";
		outStub.start();
		outStub << "\nreturn _ob_ids_;";
		outStub.end();

		for(j = 0 ; j < attributes -> length() ; j++)
		{
		    Contained::Description_var contDesc =
			attributes[j] -> describe();
		    AttributeDescription* atDesc;
		    Boolean b = (contDesc -> value) >>= atDesc;
		    assert(b);

		    Identifier_var atName = fixName(atDesc -> name);

		    comment(atDesc -> id, outStub);

		    String_var result =
			getTypeString(package, atDesc -> type);
		    outStub.sep();
		    outStub << "\npublic " << result << '\n' << atName << "()";

		    outStub.start();

		    outStub << "\norg.omg.CORBA.Request _ob_req = "
			    << "_request(\"_get_" << atDesc -> name << "\");";
		    outStub << "\norg.omg.CORBA.ORB _ob_orb = _orb();";

		    outStub.sep();
		    outStub << "\n_ob_req.set_return_type(";
		    writeTypeCode("_ob_orb", package, atDesc -> type, outStub);
		    outStub << ");";

		    outStub.sep();
		    outStub << "\n_ob_req.invoke();";

		    outStub.sep();
		    outStub << "\njava.lang.Exception _ob_ex = "
			    << "_ob_req.env().exception();";
		    outStub << "\nif(_ob_ex != null)";
		    outStub.inc();
		    outStub << "\nthrow (org.omg.CORBA.SystemException)"
			    << "_ob_ex;";
		    outStub.dec();

		    outStub.sep();
		    outStub << "\norg.omg.CORBA.Any _ob_ra"
			    << " = _ob_req.return_value();";
		    outStub << "\nreturn ";
		    writeInsertExtract(package, atDesc -> type, "_ob_ra", true,
				       outStub);
		    outStub << ';';

		    outStub.end();

		    if(atDesc -> mode == ATTR_NORMAL)
		    {
			outStub.sep();
			outStub << "\npublic void\n" << atName << '('
				<< result << " _ob_a)";

			outStub.start();

			outStub << "\norg.omg.CORBA.Request _ob_req = "
				<< "_request(\"_set_"
				<< atDesc -> name << "\");";
			outStub << "\norg.omg.CORBA.ORB _ob_orb = _orb();";

			outStub.sep();
			outStub << "\norg.omg.CORBA.Any _ob_aa = "
				<< "_ob_req.add_in_arg();";
			outStub << '\n';
			writeInsertExtract(package, atDesc -> type, "_ob_aa",
					   false, outStub);
			outStub << "_ob_a";
			TypeCode_var origType =
			    GetOrigType(atDesc -> type);
			if(origType -> kind() == tk_fixed)
			{
			    outStub << ", _ob_orb.create_fixed_tc((short)"
				    << origType -> fixed_digits()
				    << ", (short)"
				    << origType -> fixed_scale() << ')';
			}
			outStub << ");";

			outStub.sep();
			outStub << "\n_ob_req.invoke();";

			outStub.sep();
			outStub << "\njava.lang.Exception _ob_ex = "
				<< "_ob_req.env().exception();";
			outStub << "\nif(_ob_ex != null)";
			outStub.inc();
			outStub << "\nthrow (org.omg.CORBA.SystemException)"
				<< "_ob_ex;";
			outStub.dec();

			outStub.end();
		    }
		}

		for(j = 0 ; j < operations -> length() ; j++)
		{
		    Contained::Description_var contDesc =
			operations[j] -> describe();
		    OperationDescription* opDesc;
		    Boolean b = (contDesc -> value) >>= opDesc;
		    assert(b);

		    Identifier_var opName = fixName(opDesc -> name);

		    comment(opDesc -> id, outStub);

		    bool haveOutOrInOut = false;

		    TypeCode_var origRet =
			GetOrigType(opDesc -> result);

		    String_var result = getTypeString(package, origRet);
		    outStub.sep();
		    outStub << "\npublic " << result << '\n' << opName << "(";

		    unsigned int indent = outStub.getIndent();
		    outStub.setIndent(outStub.getPos());

		    for(k = 0 ; k < opDesc -> parameters.length() ; k++)
		    {
			if(k > 0)
			    outStub << ",\n";

			switch(opDesc -> parameters[k].mode)
			{
			case PARAM_IN:
			{
			    String_var str =
				getTypeString(package,
					      opDesc -> parameters[k].type);
			    outStub << str << " _ob_a" << k;
			    break;
			}

			case PARAM_INOUT:
			case PARAM_OUT:
			{
			    String_var str =
				getTypeString(package,
					      opDesc -> parameters[k].type,
					      GetTypeHolder);
			    outStub << str << " _ob_ah" << k;
			    haveOutOrInOut = true;
			    break;
			}
			}
		    }

		    if(opDesc -> contexts.length() > 0)
		    {
			if(opDesc -> parameters.length() > 0)
			    outStub << ",\n";

			outStub << "org.omg.CORBA.Context _ob_ctx";
		    }

		    outStub.setIndent(indent);

		    outStub << ')';

		    if(opDesc -> exceptions.length())
		    {
			outStub.inc();
			outStub << "\nthrows ";

			unsigned int indent = outStub.getIndent();
			outStub.setIndent(outStub.getPos());

			for(k = 0 ; k < opDesc -> exceptions.length() ; k++)
			{
			    if(k)
				outStub << ",\n";

			    String_var str =
				getTypeString(package,
					      opDesc -> exceptions[k].type);

			    outStub << str;
			}

			outStub.setIndent(indent);
			outStub.dec();
		    }

		    outStub.start();

		    outStub << "\norg.omg.CORBA.Request _ob_req = _request(\""
			    << opDesc -> name << "\");";
		    outStub << "\norg.omg.CORBA.ORB _ob_orb = _orb();";

		    for(k = 0 ; k < opDesc -> exceptions.length() ; k++)
		    {
			if(k == 0)
			    outStub.sep();

			String_var str =
			    getTypeString(package,
					  opDesc -> exceptions[k].type,
					  GetTypeHelper);

			outStub << "\n_ob_req.exceptions().add(" << str
				<< ".type());";
		    }

		    for(k = 0 ; k < opDesc -> parameters.length() ; k++)
		    {
			if(k == 0)
			    outStub.sep();

			switch(opDesc -> parameters[k].mode)
			{
			case PARAM_IN:
			{
			    outStub << "\norg.omg.CORBA.Any _ob_aa" << k
				    << " = _ob_req.add_in_arg();";
			    String_var any = string_dup("_ob_aa");
			    any += k;
			    outStub << '\n';
			    writeInsertExtract(package,
					       opDesc -> parameters[k].type,
					       any, false, outStub);
			    outStub << "_ob_a" << k;
			    TypeCode_var origType =
			    GetOrigType(opDesc -> parameters[k].type);
			    if(origType -> kind() == tk_fixed)
			    {
				outStub << ", _ob_orb.create_fixed_tc((short)"
					<< origType -> fixed_digits()
					<< ", (short)"
					<< origType -> fixed_scale() << ')';
			    }
			    outStub << ");";
			    break;
			}

			case PARAM_INOUT:
			{
			    outStub << "\norg.omg.CORBA.Any _ob_aa" << k
				    << " = _ob_req.add_inout_arg();";
			    String_var any = string_dup("_ob_aa");
			    any += k;
			    outStub << '\n';
			    writeInsertExtract(package,
					       opDesc -> parameters[k].type,
					       any, false, outStub);
			    outStub << "_ob_ah" << k << ".value";
			    TypeCode_var origType =
				GetOrigType(opDesc -> parameters[k].type);
			    if(origType -> kind() == tk_fixed)
			    {
				outStub << ", _ob_orb.create_fixed_tc((short)"
					<< origType -> fixed_digits()
					<< ", (short)"
					<< origType -> fixed_scale() << ')';
			    }
			    outStub << ");";
			    break;
			}

			case PARAM_OUT:
			{
			    outStub << "\norg.omg.CORBA.Any _ob_aa" << k
				    << " = _ob_req.add_out_arg();";
			    outStub << "\n_ob_aa" << k << ".type(";
			    writeTypeCode("_ob_orb", package,
					  opDesc -> parameters[k].type,
					  outStub);
			    outStub << ");";
			    break;
			}
			}
		    }

		    if(opDesc -> contexts.length() > 0)
		    {
			outStub.sep();
			for(k = 0 ; k < opDesc -> contexts.length() ; k++)
			{
			    outStub << "\n_ob_req.contexts().add(\"";
			    outStub << opDesc -> contexts[k];
			    outStub << "\");";
			}
			outStub << "\n_ob_req.ctx(_ob_ctx);";
		    }

		    if(origRet -> kind() != tk_void)
		    {
			outStub.sep();
			outStub << "\n_ob_req.set_return_type(";
			writeTypeCode("_ob_orb", package,
				      opDesc -> result, outStub);
			outStub << ");";
		    }

		    outStub.sep();
		    if(opDesc -> mode == OP_ONEWAY)
			outStub << "\n_ob_req.send_oneway();";
		    else
			outStub << "\n_ob_req.invoke();";

		    outStub.sep();
		    outStub << "\njava.lang.Exception _ob_ex = "
			    << "_ob_req.env().exception();";
		    outStub << "\nif(_ob_ex != null)";
		    if(opDesc -> exceptions.length() > 0)
		    {
			outStub.start();
			outStub << "\norg.omg.CORBA.UnknownUserException "
				<< "_ob_uex;";
			outStub << "\ntry";
			outStub.start();
			outStub << "\n_ob_uex = "
				<< "(org.omg.CORBA.UnknownUserException)"
				<< "_ob_ex;";
			outStub.end();
			outStub << "\ncatch(ClassCastException _ob_dummy)";
			outStub.start();
			outStub << "\nthrow (org.omg.CORBA.SystemException)"
				<< "_ob_ex;";
			outStub.end();
			outStub.sep();
			for(k = 0 ; k < opDesc -> exceptions.length() ; k++)
			{
			    String_var strHelper =
				getTypeString(package,
					      opDesc -> exceptions[k].type,
					      GetTypeHelper);

			    outStub << "\ntry";
			    outStub.start();
			    outStub << "\nthrow " << strHelper
				    << ".extract(_ob_uex.except);";
			    outStub.end();
			    outStub << "\ncatch(org.omg.CORBA.BAD_OPERATION "
				    << "_ob_dummy)";
			    outStub.start();
			    outStub.end();
			}
			outStub.sep();
			outStub << "\nthrow new org.omg.CORBA.UNKNOWN();";
			outStub.end();
		    }
		    else
		    {
			outStub.inc();
			outStub << "\nthrow (org.omg.CORBA.SystemException)"
				<< "_ob_ex;";
			outStub.dec();
		    }

		    if(haveOutOrInOut)
			outStub.sep();

		    for(k = 0 ; k < opDesc -> parameters.length() ; k++)
		    {
			if(opDesc -> parameters[k].mode != PARAM_IN)
			{
			    String_var any = string_dup("_ob_aa");
			    any += k;
			    outStub << "\n_ob_ah" << k << ".value = ";
			    writeInsertExtract(package,
					       opDesc -> parameters[k].type,
					       any, true, outStub);
			    outStub << ';';
			}
		    }

		    if(origRet -> kind() != tk_void)
		    {
			outStub.sep();
			outStub << "\norg.omg.CORBA.Any _ob_ra"
				<< " = _ob_req.return_value();";
			outStub << "\nreturn ";
			writeInsertExtract(package, opDesc -> result, "_ob_ra",
					   true, outStub);
			outStub << ';';
		    }

		    outStub.end();
		}

		outStub.end();
	    }

	    //
	    // Stream-based non-local stub for the interface
	    //
//	    if(!local_)
	    if(!noStreams_)
	    {
		String_var stub = string_dup("_");
		stub += name;
		stub += "Stub";
		PrettyPrint outStreamStub(prog_, stub, package, outDir_);
		if(!outStreamStub.good())
		    return 1;

		comment(desc -> id, outStreamStub);

		outStreamStub.sep();
		outStreamStub << "\npublic class " << stub << ' ';
		indent = outStreamStub.getIndent();
		outStreamStub.setIndent(outStreamStub.getPos());
		outStreamStub << "extends org.omg.CORBA.portable.ObjectImpl";
		outStreamStub << "\nimplements " << name;
		outStreamStub.setIndent(indent);

		outStreamStub.start();

		ContainedSeq_var attributes =
		    interf -> contents(dk_Attribute, false);

		ContainedSeq_var operations =
		    interf -> contents(dk_Operation, false);

		InterfaceDefSeq allBaseInterfaces;
		IdlGetAllBaseInterfaces(interf, allBaseInterfaces);

		outStreamStub << "\nprivate static final String[] _ob_ids_ =";
		outStreamStub.start();
		outStreamStub << "\n\"" << desc -> id << "\",";
		for(j = 0 ; j < allBaseInterfaces.length() ; j++)
		{
		    RepositoryId_var id =
			allBaseInterfaces[j] -> id();
		    outStreamStub << "\n\"" << id << '"';
		    if(j + 1 < allBaseInterfaces.length())
			outStreamStub << ',';
		}
		outStreamStub.end();
		outStreamStub << ';';

		outStreamStub.sep();
		outStreamStub << "\npublic String[]\n_ids()";
		outStreamStub.start();
		outStreamStub << "\nreturn _ob_ids_;";
		outStreamStub.end();

		for(j = 0 ; j < attributes -> length() ; j++)
		{
		    Contained::Description_var contDesc =
			attributes[j] -> describe();
		    AttributeDescription* atDesc;
		    Boolean b = (contDesc -> value) >>= atDesc;
		    assert(b);

		    Identifier_var atName = fixName(atDesc -> name);

		    comment(atDesc -> id, outStreamStub);

		    String_var result =
			getTypeString(package, atDesc -> type);
		    outStreamStub.sep();
		    outStreamStub << "\npublic " << result << '\n' << atName
				  << "()";

		    outStreamStub.start();

		    outStreamStub << "\norg.omg.CORBA.portable.OutputStream "
				  << "out = null;";
		    outStreamStub << "\norg.omg.CORBA.portable.InputStream "
				  << "in = null;";

		    outStreamStub.sep();
		    outStreamStub << "\nwhile(true)";
		    outStreamStub.start();
		    outStreamStub << "\ntry";
		    outStreamStub.start();
		    outStreamStub << "\nout = _request(\"_get_" << atName
				  << "\", true);";

		    outStreamStub << "\nin = _invoke(out);";

		    String_var str = result;
		    str += " _ob_r";
		    ULong iter = 0;
		    writeReaderWriter(package, atDesc -> type, str, iter, true,
				      outStreamStub);

		    outStreamStub << "\nreturn _ob_r;";

		    outStreamStub.end();

		    outStreamStub << "\ncatch(org.omg.CORBA.portable."
				  << "RemarshalException _ob_ex)";
		    outStreamStub.start();
		    outStreamStub << "\ncontinue;";
		    outStreamStub.end();

		    outStreamStub << "\ncatch(org.omg.CORBA.portable."
				  << "ApplicationException _ob_ex)";
		    outStreamStub.start();
		    outStreamStub << "\nfinal String _ob_id = _ob_ex.getId();";
		    outStreamStub.sep();

		    outStreamStub << "\nthrow new org.omg.CORBA.UNKNOWN("
				  << "\"Unexpected User Exception: \" + "
				  << "_ob_id);";
		    outStreamStub.end();
		    outStreamStub << "\nfinally";
		    outStreamStub.start();
		    outStreamStub << "\n_releaseReply(in);";
		    outStreamStub.end();

		    outStreamStub.end();

		    outStreamStub.end();

		    if(atDesc -> mode == ATTR_NORMAL)
		    {
			outStreamStub.sep();
			outStreamStub << "\npublic void\n" << atName << '('
				<< result << " _ob_a)";

			outStreamStub.start();

			outStreamStub << "\norg.omg.CORBA.portable."
				      << "OutputStream out = null;";
			outStreamStub << "\norg.omg.CORBA.portable."
				      << "InputStream in = null;";

			outStreamStub.sep();
			outStreamStub << "\nwhile(true)";
			outStreamStub.start();
			outStreamStub << "\ntry";
			outStreamStub.start();
			outStreamStub << "\nout = _request(\"_set_" << atName 
				      << "\", true);";

			ULong iter = 0;
			writeReaderWriter(package,
					  atDesc -> type, "_ob_a", iter, false,
					  outStreamStub);

			outStreamStub << "\nin = _invoke(out);";

			outStreamStub.end();

			outStreamStub << "\ncatch(org.omg.CORBA.portable."
				      << "RemarshalException _ob_ex)";
			outStreamStub.start();
			outStreamStub << "\ncontinue;";
			outStreamStub.end();

			outStreamStub << "\ncatch(org.omg.CORBA.portable."
				      << "ApplicationException _ob_ex)";
			outStreamStub.start();
			outStreamStub << "\nfinal String _ob_id = "
				      << "_ob_ex.getId();";
			outStreamStub.sep();

			outStreamStub << "\nthrow new org.omg.CORBA.UNKNOWN("
				      << "\"Unexpected User Exception: \" + "
				      << "_ob_id);";
			outStreamStub.end();
			outStreamStub << "\nfinally";
			outStreamStub.start();
			outStreamStub << "\n_releaseReply(in);";
			outStreamStub.end();

			outStreamStub.end();

			outStreamStub.end();
		    }
		}

		for(j = 0 ; j < operations -> length() ; j++)
		{
		    Contained::Description_var contDesc =
			operations[j] -> describe();
		    OperationDescription* opDesc;
		    Boolean b = (contDesc -> value) >>= opDesc;
		    assert(b);

		    Identifier_var opName = fixName(opDesc -> name);

		    comment(opDesc -> id, outStreamStub);

		    TypeCode_var origRet =
			GetOrigType(opDesc -> result);

		    String_var result = getTypeString(package, origRet);
		    outStreamStub.sep();
		    outStreamStub << "\npublic " << result << '\n' << opName
				  << "(";

		    unsigned int indent = outStreamStub.getIndent();
		    outStreamStub.setIndent(outStreamStub.getPos());

		    for(k = 0 ; k < opDesc -> parameters.length() ; k++)
		    {
			if(k > 0)
			    outStreamStub << ",\n";

			switch(opDesc -> parameters[k].mode)
			{
			case PARAM_IN:
			{
			    String_var str =
				getTypeString(package,
					      opDesc -> parameters[k].type);
			    outStreamStub << str << " _ob_a" << k;
			    break;
			}

			case PARAM_INOUT:
			case PARAM_OUT:
			{
			    String_var str =
				getTypeString(package,
					      opDesc -> parameters[k].type,
					      GetTypeHolder);
			    outStreamStub << str << " _ob_ah" << k;
			    break;
			}
			}
		    }

		    if(opDesc -> contexts.length() > 0)
		    {
			if(opDesc -> parameters.length() > 0)
			    outStreamStub << ",\n";
		    
			outStreamStub << "org.omg.CORBA.Context _ob_ctx";
		    }

		    outStreamStub.setIndent(indent);

		    outStreamStub << ')';

		    if(opDesc -> exceptions.length())
		    {
			outStreamStub.inc();
			outStreamStub << "\nthrows ";

			unsigned int indent = outStreamStub.getIndent();
			outStreamStub.setIndent(outStreamStub.getPos());

			for(k = 0 ; k < opDesc -> exceptions.length() ; k++)
			{
			    if(k)
				outStreamStub << ",\n";

			    String_var str =
				getTypeString(package,
					      opDesc -> exceptions[k].type);

			    outStreamStub << str;
			}

			outStreamStub.setIndent(indent);
			outStreamStub.dec();
		    }

		    outStreamStub.start();

		    outStreamStub << "\norg.omg.CORBA.portable.OutputStream "
				  << "out = null;";
		    outStreamStub << "\norg.omg.CORBA.portable.InputStream "
				  << "in = null;";

		    outStreamStub.sep();
		    outStreamStub << "\nwhile(true)";
		    outStreamStub.start();
		    outStreamStub << "\ntry";
		    outStreamStub.start();

		    if(opDesc -> mode == OP_ONEWAY)
		    {
			outStreamStub << "\nout = _request(\"" << opName
				      << "\", false);";
		    }
		    else
		    {
			outStreamStub << "\nout = _request(\"" << opName
				      << "\", true);";
		    }

		    for(k = 0 ; k < opDesc -> parameters.length() ; k++)
		    {
			switch(opDesc -> parameters[k].mode)
			{
			case PARAM_IN:
			{
			    String_var str = string_dup("_ob_a");
			    str += k;
			    ULong iter = 0;
			    writeReaderWriter(package,
					      opDesc -> parameters[k].type,
					      str, iter, false, outStreamStub);
			    break;
			}

			case PARAM_INOUT:
			{
			    String_var str = string_dup("_ob_ah");
			    str += k;
			    str += ".value";
			    ULong iter = 0;
			    writeReaderWriter(package,
					      opDesc -> parameters[k].type,
					      str, iter, false, outStreamStub);
			    break;
			}

			case PARAM_OUT:
			    break;
			}
		    }

		    if(opDesc -> contexts.length() > 0)
			outStreamStub << "\nout.write_Context(_ob_ctx);";

		    outStreamStub << "\nin = _invoke(out);";
		
		    if(origRet -> kind() != tk_void)
		    {
			String_var str = result;
			str += " _ob_r";
			ULong iter = 0;
			writeReaderWriter(package, opDesc -> result, str, iter,
					  true, outStreamStub);
		    }

		    for(k = 0 ; k < opDesc -> parameters.length() ; k++)
		    {
			if(opDesc -> parameters[k].mode != PARAM_IN)
			{
			    String_var str = string_dup("_ob_ah");
			    str += k;
			    str += ".value";
			    ULong iter = 0;
			    writeReaderWriter(package,
					      opDesc -> parameters[k].type,
					      str, iter, false, outStreamStub);
			}
		    }

		    if(origRet -> kind() != tk_void)
		    {
			outStreamStub << "\nreturn _ob_r;";
		    }

		    outStreamStub.end();

		    outStreamStub << "\ncatch(org.omg.CORBA.portable."
				  << "RemarshalException _ob_ex)";
		    outStreamStub.start();
		    outStreamStub << "\ncontinue;";
		    outStreamStub.end();

		    outStreamStub << "\ncatch(org.omg.CORBA.portable."
				  << "ApplicationException _ob_ex)";
		    outStreamStub.start();
		    outStreamStub << "\nfinal String _ob_id = _ob_ex.getId();";
		    outStreamStub.sep();

		    for(k = 0 ; k < opDesc -> exceptions.length() ; k++)
		    {
			String_var str =
			    getTypeString(package,
					  opDesc -> exceptions[k].type,
					  GetTypeHelper);

			outStreamStub << "\nif(_ob_id.equals(" << str
				      << ".id()))";
			outStreamStub.start();
			outStreamStub << "\nthrow " << str << ".read("
				      << "_ob_ex.getInputStream());";
			outStreamStub.end();
		    }

		    outStreamStub << "\nthrow new org.omg.CORBA.UNKNOWN("
				  << "\"Unexpected User Exception: \" + "
				  << "_ob_id);";
		    outStreamStub.end();
		    outStreamStub << "\nfinally";
		    outStreamStub.start();
		    outStreamStub << "\n_releaseReply(in);";
		    outStreamStub.end();

		    outStreamStub.end();
		    outStreamStub.end();
		}

		outStreamStub.end();
	    }

	    //
	    // Stream-based local stub for the interface
	    //
	    if(!noStreams_)
	    {
	    String_var stub = string_dup("_");
	    stub += name;
	    stub += "LocalStub";
	    PrettyPrint outLocalStub(prog_, stub, package, outDir_);
	    if(!outLocalStub.good())
		return 1;

	    comment(desc -> id, outLocalStub);

	    outLocalStub.sep();
	    outLocalStub << "\npublic class " << stub << ' ';
	    outLocalStub << "extends _" << name << "Stub";

	    outLocalStub.start();

	    ContainedSeq_var attributes =
		interf -> contents(dk_Attribute, false);

	    ContainedSeq_var operations =
		interf -> contents(dk_Operation, false);

	    InterfaceDefSeq allBaseInterfaces;
	    IdlGetAllBaseInterfaces(interf, allBaseInterfaces);

	    outLocalStub << "\nfinal public static Class _ob_class = " << name
			 << "Operations.class;";
	    outLocalStub.sep();

	    for(j = 0 ; j < attributes -> length() ; j++)
	    {
		Contained::Description_var contDesc =
		    attributes[j] -> describe();
		AttributeDescription* atDesc;
		Boolean b = (contDesc -> value) >>= atDesc;
		assert(b);

		Identifier_var atName = fixName(atDesc -> name);

		comment(atDesc -> id, outLocalStub);

		String_var result =
		    getTypeString(package, atDesc -> type);
		outLocalStub.sep();
		outLocalStub << "\npublic " << result << '\n' << atName
				  << "()";

		outLocalStub.start();
		outLocalStub << "\norg.omg.CORBA.portable.ServantObject "
			     << "_ob_so =";
		outLocalStub.inc();
		outLocalStub << "\n_servant_preinvoke(\"_get_" << atName
			     << "\", _ob_class);";
		outLocalStub.dec();
		outLocalStub << "\nif(_ob_so == null)";
		outLocalStub.start();
		outLocalStub << "\nreturn super." << atName << "();";
		outLocalStub.end();

		outLocalStub.sep();
		outLocalStub << '\n' << name << "Operations _self = ("
			     << name << "Operations)_ob_so.servant;";
		outLocalStub << "\ntry";
		outLocalStub.start();
		outLocalStub << "\nreturn _self." << atName << "();";
		outLocalStub.end();
		outLocalStub << "\nfinally";
		outLocalStub.start();
		outLocalStub << "\n_servant_postinvoke(_ob_so);";
		outLocalStub.end();

		outLocalStub.end();

		if(atDesc -> mode == ATTR_NORMAL)
		{
		    outLocalStub.sep();
		    outLocalStub << "\npublic void\n" << atName << '('
				 << result << " _ob_a)";

		    outLocalStub.start();

		    outLocalStub << "\norg.omg.CORBA.portable.ServantObject "
				 << "_ob_so =";
		    outLocalStub.inc();
		    outLocalStub << "\n_servant_preinvoke(\"_get_" << atName
				 << "\", _ob_class);";
		    outLocalStub.dec();
		    outLocalStub << "\nif(_ob_so == null)";
		    outLocalStub.start();
		    outLocalStub << "\nsuper." << atName << "(_ob_a);";
		    outLocalStub.end();

		    outLocalStub.sep();
		    outLocalStub << '\n' << name << "Operations _self = ("
				 << name << "Operations)_ob_so.servant;";
		    outLocalStub << "\ntry";
		    outLocalStub.start();
		    outLocalStub << "\n_self." << atName << "(_ob_a);";
		    outLocalStub.end();
		    outLocalStub << "\nfinally";
		    outLocalStub.start();
		    outLocalStub << "\n_servant_postinvoke(_ob_so);";
		    outLocalStub.end();

		    outLocalStub.end();
		}
	    }

	    for(j = 0 ; j < operations -> length() ; j++)
	    {
		Contained::Description_var contDesc =
		    operations[j] -> describe();
		OperationDescription* opDesc;
		Boolean b = (contDesc -> value) >>= opDesc;
		assert(b);

		Identifier_var opName = fixName(opDesc -> name);

		comment(opDesc -> id, outLocalStub);

		TypeCode_var origRet =
		    GetOrigType(opDesc -> result);
		
		String_var result = getTypeString(package, origRet);
		outLocalStub.sep();
		outLocalStub << "\npublic " << result << '\n' << opName << "(";

		unsigned int indent = outLocalStub.getIndent();
		outLocalStub.setIndent(outLocalStub.getPos());

		for(k = 0 ; k < opDesc -> parameters.length() ; k++)
		{
		    if(k > 0)
			outLocalStub << ",\n";

		    switch(opDesc -> parameters[k].mode)
		    {
		    case PARAM_IN:
		    {
			String_var str =
			    getTypeString(package,
					  opDesc -> parameters[k].type);
			outLocalStub << str << " _ob_a" << k;
			break;
			}

		    case PARAM_INOUT:
		    case PARAM_OUT:
		    {
			String_var str =
			    getTypeString(package,
					  opDesc -> parameters[k].type,
					  GetTypeHolder);
			outLocalStub << str << " _ob_ah" << k;
			break;
		    }
		    }
		}

		if(opDesc -> contexts.length() > 0)
		{
		    if(opDesc -> parameters.length() > 0)
			outLocalStub << ",\n";
		    
		    outLocalStub << "org.omg.CORBA.Context _ob_ctx";
		}

		outLocalStub.setIndent(indent);
		
		outLocalStub << ')';

		if(opDesc -> exceptions.length())
		{
		    outLocalStub.inc();
		    outLocalStub << "\nthrows ";
		    
		    unsigned int indent = outLocalStub.getIndent();
		    outLocalStub.setIndent(outLocalStub.getPos());
		    
		    for(k = 0 ; k < opDesc -> exceptions.length() ; k++)
		    {
			if(k)
			    outLocalStub << ",\n";
			
			String_var str =
			    getTypeString(package,
					  opDesc -> exceptions[k].type);
			
			outLocalStub << str;
		    }
		    
		    outLocalStub.setIndent(indent);
		    outLocalStub.dec();
		}
		
		outLocalStub.start();
		
		outLocalStub << "\norg.omg.CORBA.portable.ServantObject "
			     << "_ob_so =";
		outLocalStub.inc();
		outLocalStub << "\n_servant_preinvoke(\"" << opName
			     << "\", _ob_class);";
		outLocalStub.dec();
		outLocalStub << "\nif(_ob_so == null)";
		outLocalStub.start();

		outLocalStub << '\n';
		if(origRet -> kind() != tk_void)
		    outLocalStub << "return ";

		outLocalStub << "super." << opName << "(";

		indent = outLocalStub.getIndent();
		outLocalStub.setIndent(outLocalStub.getPos());

		if(opDesc -> parameters.length() > 0)
		{
		    for(k = 0 ; k < opDesc -> parameters.length() ; k++)
		    {
			switch(opDesc -> parameters[k].mode)
                        {
                        case PARAM_IN:
                        {
			    if(k)
				outLocalStub << ",\n";
			    outLocalStub << "_ob_a" << k;
			    break;
			}

			case PARAM_INOUT:
			case PARAM_OUT:
			{
			    if(k)
				outLocalStub << ",\n";
			    outLocalStub << "_ob_ah" << k;
			    break;
			}
			}

		    }
		}

		if(opDesc -> contexts.length() > 0)
		{
		    if(opDesc -> parameters.length() > 0)
			outLocalStub << ",\n";

			outLocalStub << "_ob_ctx";
		}

		outLocalStub.setIndent(indent);

		outLocalStub << ");";
		outLocalStub.end();
		
		outLocalStub.sep();
		outLocalStub << '\n' << name << "Operations _self = ("
			     << name << "Operations)_ob_so.servant;";
		outLocalStub << "\ntry";
		outLocalStub.start();

		outLocalStub << '\n';
		if(origRet -> kind() != tk_void)
		    outLocalStub << "return ";

		outLocalStub << "_self." << opName << "(";

		indent = outLocalStub.getIndent();
		outLocalStub.setIndent(outLocalStub.getPos());

		if(opDesc -> parameters.length() > 0)
		{
		    for(k = 0 ; k < opDesc -> parameters.length() ; k++)
		    {
			switch(opDesc -> parameters[k].mode)
                        {
                        case PARAM_IN:
                        {
			    if(k)
				outLocalStub << ",\n";
			    outLocalStub << "_ob_a" << k;
			    break;
			}

			case PARAM_INOUT:
			case PARAM_OUT:
			{
			    if(k)
				outLocalStub << ",\n";
			    outLocalStub << "_ob_ah" << k;
			    break;
			}
			}
		    }
		}

		if(opDesc -> contexts.length() > 0)
		{
		    if(opDesc -> parameters.length() > 0)
			outLocalStub << ",\n";

			outLocalStub << "_ob_ctx";
		}

		outLocalStub.setIndent(indent);

		outLocalStub << ");";
		outLocalStub.end();

		outLocalStub << "\nfinally";
		outLocalStub.start();
		outLocalStub << "\n_servant_postinvoke(_ob_so);";
		outLocalStub.end();

		outLocalStub.end();
	    }

	    outLocalStub.end();
	    }

	    //
	    // Write the holder class
	    //
	    String_var holder = name;
	    holder += "Holder";
	    PrettyPrint outHolder(prog_, holder, package, outDir_);
	    if(!outHolder.good())
		return 1;
	    
	    comment(desc -> id, outHolder);

	    writeHolder(name, name, outHolder);

	    //
	    // Write the helper class
	    //
//	    if(!local_)
	    {
		String_var helper = name;
		helper += "Helper";
		PrettyPrint outHelper(prog_, helper, package, outDir_);
		if(!outHelper.good())
		    return 1;
		
		comment(desc -> id, outHelper);
		
		writeHelper(package, desc -> type, outHelper);
	    }

	    break;
	}

	case dk_Alias:
	{
	    //
	    // Alias
	    //
	    TypeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixName(desc -> name);
	    String_var package = getPackage(desc -> id);

	    //
	    // Write the holder class only for sequence and array typedefs
	    //
	    TypeCode_var origType = GetOrigType(desc -> type);
	    if(origType -> kind() == tk_array ||
	       origType -> kind() == tk_sequence)
	    {
		String_var holder = name;
		holder += "Holder";
		PrettyPrint outHolder(prog_, holder, package, outDir_);
		if(!outHolder.good())
		    return 1;
		
		comment(desc -> id, outHolder);
		
		String_var norm = getTypeString(package, desc -> type);
		writeHolder(norm, name, outHolder);
	    }

	    //
	    // Write the helper class
	    //
//	    if(!local_)
	    {
		String_var helper = name;
		helper += "Helper";
		PrettyPrint outHelper(prog_, helper, package, outDir_);
		if(!outHelper.good())
		return 1;
		
		comment(desc -> id, outHelper);
		
		writeHelper(package, desc -> type, outHelper);
	    }

	    break;
	}

	case dk_Enum:
	{
	    //
	    // Enum
	    //
	    TypeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixName(desc -> name);
	    String_var package = getPackage(desc -> id);

	    PrettyPrint out(prog_, name, package, outDir_);
	    if(!out.good())
		return 1;

	    comment(desc -> id, out);
	    docComment(desc -> id, out);

	    out.sep();
	    out << "\nfinal public class " << name
		<< " implements org.omg.CORBA.portable.IDLEntity";
	    out.start();

	    out << "\nprivate static " << name << " [] values_ = new "
		<< name << "[" << desc -> type -> member_count() << "];";
	    out << "\nprivate int value_;";
	    out.sep();

	    for(j = 0 ; j < desc -> type -> member_count() ; j++)
	    {
		Identifier_var memberName =
		    desc -> type -> member_name(j);
		memberName = fixName(memberName);
		out << "\npublic final static int _" << memberName
		    << " = " << j << ";";
		out << "\npublic final static " << name << " " << memberName
		    << " = new " << name << "(_" << memberName << ");";
	    }

	    out.sep();
	    out << "\nprotected\n" << name << "(int value)";
	    out.start();
	    out << "\nvalues_[value] = this;";
	    out << "\nvalue_ = value;";
	    out.end();

	    out.sep();
	    out << "\npublic int\nvalue()";
	    out.start();
	    out << "\nreturn value_;";
	    out.end();

	    out.sep();
	    out << "\npublic static " << name << "\nfrom_int(int value)";
	    out.start();
	    out << "\nreturn values_[value];";
	    out.end();

	    if(clone_)
		writeClone(seq[i].kind, name, package, desc -> type, out);

	    out.end();

	    //
	    // Write the holder class
	    //
	    String_var holder = name;
	    holder += "Holder";
	    PrettyPrint outHolder(prog_, holder, package, outDir_);
	    if(!outHolder.good())
		return 1;
	    
	    comment(desc -> id, outHolder);
	    
	    writeHolder(name, name, outHolder);

	    //
	    // Write the helper class
	    //
//	    if(!local_)
	    {
		String_var helper = name;
		helper += "Helper";
		PrettyPrint outHelper(prog_, helper, package, outDir_);
		if(!outHelper.good())
		    return 1;

		comment(desc -> id, outHelper);

		writeHelper(package, desc -> type, outHelper);
	    }

	    break;
	}

	case dk_Struct:
	case dk_Exception:
	{
	    RepositoryId_var id;
	    Identifier_var name;
	    TypeCode_var type;
	    
	    if(seq[i].kind == dk_Struct)
	    {
		//
		// Struct
		//
		TypeDescription* desc;
		Boolean b = seq[i].value >>= desc;
		assert(b);

		name = fixName(desc -> name);
		id = desc -> id;
		type = desc -> type;
	    }
	    else
	    {
		//
		// Exception
		//
		ExceptionDescription* desc;
		Boolean b = seq[i].value >>= desc;
		assert(b);

		name = fixName(desc -> name);
		id = desc -> id;
		type = desc -> type;
	    }

	    if(!check(id))
		continue;

	    String_var package = getPackage(id);

	    PrettyPrint out(prog_, name, package, outDir_);
	    if(!out.good())
		return 1;

	    comment(id, out);
	    docComment(id, out);

	    out.sep();
	    out << "\nfinal public class " << name;
	    if(seq[i].kind == dk_Exception)
		out << " extends org.omg.CORBA.UserException";
	    else
		out << " implements org.omg.CORBA.portable.IDLEntity";

	    out.start();

	    out << "\npublic\n" << name << "()";
	    out.start();
	    if(seq[i].kind == dk_Exception)
		out << "\nsuper(" << name << "Helper.id());";
	    out.end();

	    if(type -> member_count() > 0)
	    {
		out.sep();
		out << "\npublic\n" << name << "(";
		unsigned int indent = out.getIndent();
		out.setIndent(out.getPos());
		for(j = 0 ; j < type -> member_count() ; j++)
		{
		    TypeCode_var memberType = type -> member_type(j);
		    String_var norm = getTypeString(package, memberType);

		    if(j)
			out << ",\n";

		    out << norm << " _ob_a" << j;
		}
		out << ")";
		out.setIndent(indent);

		out.start();

		if(seq[i].kind == dk_Exception)
		    out << "\nsuper(" << name << "Helper.id());";

		for(j = 0 ; j < type -> member_count() ; j++)
		{
		    TypeCode_var memberType = type -> member_type(j);
		    Identifier_var memberName = type -> member_name(j);
		    memberName = fixName(memberName);

		    out << "\n" << memberName << " = _ob_a" << j << ';';
		}
		out.end();
	    }

	    if(seq[i].kind == dk_Exception)
	    {
		out.sep();
		out << "\npublic\n" << name << "(";
		unsigned int indent = out.getIndent();
		out.setIndent(out.getPos());
		out << "String reason";

		for(j = 0 ; j < type -> member_count() ; j++)
		{
		    TypeCode_var memberType = type -> member_type(j);
		    String_var norm = getTypeString(package, memberType);

		    out << ",\n";

		    out << norm << " _ob_a" << j;
		}
		out << ")";
		out.setIndent(indent);

		out.start();

		out << "\nsuper(" << name
		    << "Helper.id() + \" \" + reason);";

		for(j = 0 ; j < type -> member_count() ; j++)
		{
		    TypeCode_var memberType = type -> member_type(j);
		    Identifier_var memberName =
			type -> member_name(j);
		    memberName = fixName(memberName);

		    out << "\n" << memberName << " = _ob_a" << j << ';';
		}
		out.end();
	    }

	    if(clone_)
		writeClone(seq[i].kind, name, package, type, out);

	    for(j = 0 ; j < type -> member_count() ; j++)
	    {
		TypeCode_var memberType = type -> member_type(j);
		Identifier_var memberName = type -> member_name(j);
		memberName = fixName(memberName);

		String_var norm = getTypeString(package, memberType);

		if(j == 0)
		    out.sep();

		out << "\npublic " << norm << ' ' << memberName << ";";
	    }

	    out.end();

	    //
	    // Write the holder class
	    //
	    String_var holder = name;
	    holder += "Holder";
	    PrettyPrint outHolder(prog_, holder, package, outDir_);
	    if(!outHolder.good())
		return 1;
	    
	    comment(id, outHolder);
	    
	    writeHolder(name, name, outHolder);

	    //
	    // Write the helper class
	    //
//	    if(!local_)
	    {
		String_var helper = name;
		helper += "Helper";
		PrettyPrint outHelper(prog_, helper, package, outDir_);
		if(!outHelper.good())
		    return 1;

		comment(id, outHelper);

		writeHelper(package, type, outHelper);
	    }

	    break;
	}

	case dk_Union:
	{
	    //
	    // Union
	    //
	    TypeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixName(desc -> name);
	    String_var package = getPackage(desc -> id);

	    TypeCode_var discType = desc -> type -> discriminator_type();
	    TypeCode_var origDiscType = GetOrigType(discType);
	    String_var disc = getTypeString(package, origDiscType);
	    String_var intDisc;
	    if(origDiscType -> kind() == tk_enum ||
	       origDiscType -> kind() == tk_boolean)
		intDisc = string_dup("int");
	    else
		intDisc = disc;

	    PrettyPrint out(prog_, name, package, outDir_);
	    if(!out.good())
		return 1;

	    IdlUnionMemberInfoSeq_var infoSeq = IdlGetUnionInfo(desc -> type);
	    for(j = 0 ; j < infoSeq -> length() ; j++)
		infoSeq[j].name = fixName(infoSeq[j].name);

	    String_var discStr = string_dup("discriminator");
	    String_var defStr = string_dup("_default");
	    for(j = 0 ; j < infoSeq -> length() ; j++)
	    {
		if(strcmp(infoSeq[j].name, discStr) == 0)
		    discStr = string_dup("_discriminator");

		if(strcmp(infoSeq[j].name, defStr) == 0)
		    defStr = string_dup("__default");
	    }

	    comment(desc -> id, out);
	    docComment(desc -> id, out);

	    out.sep();
	    out << "\nfinal public class " << name
		<< " implements org.omg.CORBA.portable.IDLEntity";
	    out.start();

	    out << "\njava.lang.Object _ob_v_;";
	    out << "\nboolean _ob_i_;";
	    out << '\n' << intDisc << " _ob_d_;";

	    out.sep();
	    out << "\nstatic boolean\n_OB_check(" << intDisc << " d0, "
		<< intDisc << " d1)";
	    out.start();
	    bool needSwitch = false;
	    for(j = 0 ; j < infoSeq -> length() ; j++)
	    {
		IdlUnionMemberInfo& info = infoSeq[j];
		if(info.nLabels.length() + info.pLabels.length() > 1 ||
		   info.isDefault)
		{
		    needSwitch = true;
		    break;
		}
	    }
	    if(needSwitch)
	    {
		out << '\n' << intDisc << " d[] = new " << intDisc << "[2];";
		out << "\nd[0] = d0;";
		out << "\nd[1] = d1;";
		out.sep();
		out << "\nfor(int i = 0 ; i < 2 ; i++)";
		out.start();
		out << "\nswitch(d[i])";
		out.start();
		for(j = 0 ; j < infoSeq -> length() ; j++)
		{
		    IdlUnionMemberInfo& info = infoSeq[j];
		    IdlStringSeq_var labels =
			getUnionLabelSeq(package, origDiscType, info);
		    
		    if(j > 0)
			out.sep();
		    
		    out.dec();
		    
		    if(is_nil(info.type))
		    {
			assert(info.isDefault);
			out << "\ndefault:";
		    }
		    else
		    {
			if(info.isDefault)
			    out << "\ndefault:";
			else
			{
			    for(k = 0 ; k < labels -> length() ; k++)
				out << "\ncase " << labels[k] << ':';
			}
		    }
		    
		    out.inc();
		    
		    if(labels -> length() > 1 || info.isDefault)
			out << "\nd[i] = " << labels[0] << ';';
		    
		    out << "\nbreak;";
		}
		out.end();
		out.end();
		out.sep();
		out << "\nreturn d[0] == d[1];";
	    }
	    else
	    {
		out << "\nreturn d0 == d1;";
	    }
	    out.end();

	    out.sep();
	    out << "\npublic\n" << name << "()";
	    out.start();
	    out << "\n_ob_i_ = false;";
	    out.end();

	    out.sep();
	    out << "\npublic " << disc << '\n' << discStr << "()";
	    out.start();
	    out << "\nif(!_ob_i_)";
	    out.inc();
	    out << "\nthrow new org.omg.CORBA.BAD_OPERATION();";
	    out.dec();
	    out.sep();
	    if(origDiscType -> kind() == tk_enum)
		out << "\nreturn " << disc << ".from_int(_ob_d_);";
	    else if(origDiscType -> kind() == tk_boolean)
		out << "\nreturn _ob_d_ == 0 ? false : true;";
	    else
		out << "\nreturn _ob_d_;";
	    out.end();

	    for(j = 0 ; j < infoSeq -> length() ; j++)
	    {
		IdlUnionMemberInfo& info = infoSeq[j];
		IdlStringSeq_var labels =
		    getUnionLabelSeq(package, origDiscType, info);

		if(is_nil(info.type))
		{
		    out.sep();
		    out << "\npublic void\n" << defStr << "()";
		    out.start();
		    out << "\n_ob_i_ = true;";
		    out << "\n_ob_d_ = " << labels[0] << ';';
		    out << "\n_ob_v_ = null;";
		    out.end();

		    if(info.nLabels.length() + info.pLabels.length() > 1 ||
		       info.isDefault)
		    {
			out.sep();
			out << "\npublic void\n" << defStr << '('
			    << disc << " _ob_d)";
			out.start();
			if(origDiscType -> kind() == tk_enum)
			    out << "\nif(!_OB_check(_ob_d.value(), "
				<< labels[0] << "))";
			else if(origDiscType -> kind() == tk_boolean)
			    out << "\nif(!_OB_check(_ob_d ? 1 : 0, "
				<< labels[0] << "))";
			else
			    out << "\nif(!_OB_check(_ob_d, "
				<< labels[0] << "))";
			out.inc();
			out << "\nthrow new org.omg.CORBA.BAD_OPERATION();";
			out.dec();
			out.sep();
			out << "\n_ob_i_ = true;";
			if(origDiscType -> kind() == tk_enum)
			    out << "\n_ob_d_ = _ob_d.value();";
			else if(origDiscType -> kind() == tk_boolean)
			    out << "\n_ob_d_ = _ob_d ? 1 : 0;";
			else
			    out << "\n_ob_d_ = _ob_d;";
			out << "\n_ob_v_ = null;";
			out.end();
		    }
		}
		else
		{
		    String_var norm = getTypeString(package, info.type);
		    TypeCode_var origMemberType =
			GetOrigType(info.type);

		    String_var holder =
			getUnionMemberHolder(origMemberType);

		    out.sep();
		    out << "\npublic " << norm << '\n' << info.name << "()";
		    out.start();
		    out << "\nif(!_ob_i_)";
		    out.inc();
		    out << "\nthrow new org.omg.CORBA.BAD_OPERATION();";
		    out.dec();
		    out.sep();
		    out << "\nif(!_OB_check(_ob_d_, " << labels[0] << "))";
		    out.inc();
		    out << "\nthrow new org.omg.CORBA.BAD_OPERATION();";
		    out.dec();
		    out.sep();
		    if(strlen(holder) > 0)
			out << "\nreturn ((" << holder << ")_ob_v_).value;";
		    else
			out << "\nreturn (" << norm << ")_ob_v_;";
		    out.end();

		    out.sep();
		    out << "\npublic void\n" << info.name << '(' << norm
			<< " _ob_v)";
		    out.start();
		    out << "\n_ob_i_ = true;";
		    out << "\n_ob_d_ = " << labels[0] << ';';
		    if(strlen(holder) > 0)
			out << "\n_ob_v_ = new " << holder << "(_ob_v);";
		    else
			out << "\n_ob_v_ = _ob_v;";
		    out.end();

		    if(info.nLabels.length() + info.pLabels.length() > 1 ||
		       info.isDefault)
		    {
			out.sep();
			out << "\npublic void\n" << info.name << '('
			    << disc << " _ob_d, " << norm << " _ob_v)";
			out.start();
			if(origDiscType -> kind() == tk_enum)
			    out << "\nif(!_OB_check(_ob_d.value(), "
				<< labels[0] << "))";
			else if(origDiscType -> kind() == tk_boolean)
			    out << "\nif(!_OB_check(_ob_d ? 1 : 0, "
				<< labels[0] << "))";
			else
			    out << "\nif(!_OB_check(_ob_d, "
				<< labels[0] << "))";
			out.inc();
			out << "\nthrow new org.omg.CORBA.BAD_OPERATION();";
			out.dec();
			out.sep();
			out << "\n_ob_i_ = true;";
			if(origDiscType -> kind() == tk_enum)
			    out << "\n_ob_d_ = _ob_d.value();";
			else if(origDiscType -> kind() == tk_boolean)
			    out << "\n_ob_d_ = _ob_d ? 1 : 0;";
			else
			    out << "\n_ob_d_ = _ob_d;";
			if(strlen(holder) > 0)
			    out << "\n_ob_v_ = new " << holder << "(_ob_v);";
			else
			    out << "\n_ob_v_ = _ob_v;";
			out.end();
		    }
		}
	    }

	    if(clone_)
		writeClone(seq[i].kind, name, package, desc -> type, out);

	    out.end();

	    //
	    // Write the holder class
	    //
	    String_var holder = name;
	    holder += "Holder";
	    PrettyPrint outHolder(prog_, holder, package, outDir_);
	    if(!outHolder.good())
		return 1;
	    
	    comment(desc -> id, outHolder);
	    
	    writeHolder(name, name, outHolder);

	    //
	    // Write the helper class
	    //
//	    if(!local_)
	    {
		String_var helper = name;
		helper += "Helper";
		PrettyPrint outHelper(prog_, helper, package, outDir_);
		if(!outHelper.good())
		    return 1;

		comment(desc -> id, outHelper);

		writeHelper(package, desc -> type, outHelper);
	    }

	    break;
	}

	case dk_Attribute:
	{
	    //
	    // Attribute
	    //
	    AttributeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixName(desc -> name);
	    String_var package = getPackage(desc -> id);

	    comment(desc -> id, out);
	    docComment(desc -> id, out);

	    String_var norm = getTypeString(package, desc -> type);

	    out.sep();
	    out << "\npublic " << norm << '\n' << name << "();";
	    if(desc -> mode == ATTR_NORMAL)
	    {
		out.sep();
		out << "\npublic void\n" << name << "(" << norm << " value);";
	    }

	    break;
	}

	case dk_Operation:
	{
	    //
	    // Operation
	    //
	    OperationDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixName(desc -> name);
	    String_var package = getPackage(desc -> id);

	    comment(desc -> id, out);
	    docComment(desc -> id, out);

	    String_var norm = getTypeString(package, desc -> result);

	    out.sep();
	    out << "\npublic " << norm << '\n' << name << "(";
	    
	    unsigned int indent = out.getIndent();
	    out.setIndent(out.getPos());

	    for(j = 0 ; j < desc -> parameters.length() ; j++)
	    {
		GetType gt;
		switch(desc -> parameters[j].mode)
		{
		case PARAM_IN:
		    gt = GetTypeNormal;
		    break;

		case PARAM_INOUT:
		case PARAM_OUT:
		    gt = GetTypeHolder;
		    break;
		}

		String_var str =
		    getTypeString(package, desc -> parameters[j].type, gt);

		if(j > 0)
		    out << ",\n";

		String_var argName = fixName(desc -> parameters[j].name);

		out << str << ' ' << argName;
	    }

	    if(desc -> contexts.length() > 0)
	    {
		if(desc -> parameters.length() > 0)
		    out << ",\n";
		
		out << "org.omg.CORBA.Context ctx";
	    }

	    out.setIndent(indent);

	    out << ')';

	    if(desc -> exceptions.length())
	    {
		out.inc();
		out << "\nthrows ";

		unsigned int indent = out.getIndent();
		out.setIndent(out.getPos());

		for(j = 0 ; j < desc -> exceptions.length() ; j++)
		{
		    if(j)
			out << ",\n";

		    String_var str =
			getTypeString(package, desc -> exceptions[j].type);
    
		    out << str;
		}

		out.setIndent(indent);
		out.dec();
	    }

	    out << ';';

	    break;
	}

	case tk_value:
	case tk_value_box:
	{
	    RepositoryId_var id;
	    Identifier_var name;
	    TypeCode_var type;

	    //
	    // Value
	    //
	    ValueDef::FullValueDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    ValueDef_var value =
		ValueDef::_narrow(seq[i].contained_object);

	    name = fixName(desc -> name);
	    id = desc -> id;
	    type = desc -> type;

	    if(!check(id))
		continue;

	    String_var package = getPackage(id);

	    PrettyPrint out(prog_, name, package, outDir_);
	    if(!out.good())
		return 1;

	    comment(id, out);
	    docComment(id, out);

	    out.sep();

	    ValueDefSeq allBaseValues;
	    IdlGetAllBaseValues(value, allBaseValues);

	    out << "\nfinal public class " << name << " implements ";

	    if(allBaseValues.length() > 0)
	    {
		for(j = 0 ; j < allBaseValues.length() ; j++)
		{
		    if(j)
			out << ",\n";

		    RepositoryId_var id = allBaseValues[j] -> id();
		    out << id;
		}
	    }
	    else
		out << "java.io.Serializable";

	    out.sep();
	    out.start();

	    out << "\npublic\n" << name << "()";
	    out.start();
	    out.end();

	    if(type -> member_count() > 0)
	    {
		out.sep();
		out << "\npublic\n" << name << "(";
		unsigned int indent = out.getIndent();
		out.setIndent(out.getPos());
		for(j = 0 ; j < type -> member_count() ; j++)
		{
		    TypeCode_var memberType = type -> member_type(j);
		    String_var norm = getTypeString(package, memberType);

		    if(j)
			out << ",\n";

		    out << norm << " _ob_a" << j;
		}
		out << ")";
		out.setIndent(indent);

		out.start();

		for(j = 0 ; j < type -> member_count() ; j++)
		{
		    TypeCode_var memberType = type -> member_type(j);
		    Identifier_var memberName = type -> member_name(j);
		    memberName = fixName(memberName);

		    out << "\n" << memberName << " = _ob_a" << j << ';';
		}
		out.end();
	    }

	    if(clone_)
		writeClone(seq[i].kind, name, package, type, out);

	    for(j = 0 ; j < type -> member_count() ; j++)
	    {
		TypeCode_var memberType = type -> member_type(j);
		Identifier_var memberName = type -> member_name(j);
		memberName = fixName(memberName);

		String_var norm = getTypeString(package, memberType);

		if(j == 0)
		    out.sep();

		out << "\npublic " << norm << ' ' << memberName << ";";
	    }

	    out.end();

	    //
	    // Write the holder class
	    //
	    String_var holder = name;
	    holder += "Holder";
	    PrettyPrint outHolder(prog_, holder, package, outDir_);
	    if(!outHolder.good())
		return 1;
	    
	    comment(id, outHolder);
	    
	    writeHolder(name, name, outHolder);

	    //
	    // Write the helper class
	    //
//	    if(!local_)
	    {
		String_var helper = name;
		helper += "Helper";
		PrettyPrint outHelper(prog_, helper, package, outDir_);
		if(!outHelper.good())
		    return 1;

		comment(id, outHelper);

		writeHelper(package, type, outHelper);
	    }

	    break;
	}

	default:
	    break;
	}
    }

    return status;
}

// ----------------------------------------------------------------------
// Generate Java skeleton code
// ----------------------------------------------------------------------

int
IdlJavaGenerator::genJavaSkel(Container_ptr container,
			      IdlPrettyPrint& out)
{
    //
    // Get container contents descriptions
    //
    Container::DescriptionSeq_var seq =
	container -> describe_contents(dk_all, true, -1);

    ULong i, j, k;

    int status = 0;
    for(i = 0 ; status == 0 && i < seq -> length() ; i++)
    {
	//
	// Get container
	//
	Container_var cont =
	    Container::_narrow(seq[i].contained_object);

	switch(seq[i].kind)
	{
	case dk_Module:
	{
	    //
	    // Module
	    //
	    ModuleDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    // Don't add this here!!!
	    // if(!check(desc -> id))
	    //     continue;

	    assert(!is_nil(cont));
	    status = genJavaSkel(cont, out);

	    break;
	}

	case dk_Interface:
	    if(noStreams_)
	{
	    //
	    // Interface
	    //
 	    InterfaceDef_var interf =
 		InterfaceDef::_narrow(seq[i].contained_object);
 	    assert(!is_nil(interf));
 	    InterfaceDef::FullInterfaceDescription_var desc =
 		interf -> describe_interface();

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixName(desc -> name);
	    String_var package = getPackage(desc -> id);

	    String_var skel = string_dup("_");
	    skel += name;
	    skel += "ImplBase";

	    unsigned int indent;

	    PrettyPrint out(prog_, skel, package, outDir_);
	    if(!out.good())
		return 1;
	    
	    comment(desc -> id, out);

	    out.sep();
	    out << "\npublic abstract class " << skel << ' ';
	    indent = out.getIndent();
	    out.setIndent(out.getPos());
	    out << "extends org.omg.CORBA.DynamicImplementation";
	    out << "\nimplements " << name;
	    out.setIndent(indent);

	    out.start();

	    ContainedSeq_var operations =
		interf -> contents(dk_Operation, false);

	    ContainedSeq_var attributes =
		interf -> contents(dk_Attribute, false);

	    InterfaceDefSeq allBaseInterfaces;
	    IdlGetAllBaseInterfaces(interf, allBaseInterfaces);

	    out << "\nstatic final String[] _ob_ids_ =";
	    out.start();
	    out << "\n\"" << desc -> id << "\",";
	    for(j = 0 ; j < allBaseInterfaces.length() ; j++)
	    {
		RepositoryId_var id =
		    allBaseInterfaces[j] -> id();
		out << "\n\"" << id << '"';
		if(j + 1 < allBaseInterfaces.length())
		    out << ',';
	    }
	    out.end();
	    out << ';';

	    out.sep();
	    out << "\npublic String[]\n_ids()";
	    out.start();
	    out << "\nreturn _ob_ids_;";
	    out.end();

	    out.sep();
	    out << "\nfinal public void\ninvoke(org.omg.CORBA.ServerRequest "
		<< "_ob_req)";
	    out.start();

	    bool first = true;

	    for(j = 0 ; j < attributes -> length() ; j++)
	    {
		Contained::Description_var contDesc =
		    attributes[j] -> describe();
		AttributeDescription* atDesc;
		Boolean b = (contDesc -> value) >>= atDesc;
		assert(b);
		
		Identifier_var atName = fixName(atDesc -> name);

		if(first)
		{
		    out << "\norg.omg.CORBA.ORB _ob_orb = _orb();";
		    out << "\norg.omg.CORBA.NVList _ob_list = "
			<< "_ob_orb.create_list(0);";
		    out << "\nString _ob_op = _ob_req.operation();";

		    out.sep();
		    out << '\n';
		}
		else
		    out << "\nelse ";

		out << "if(_ob_op.equals(\"_get_" << atDesc -> name << "\"))";
		out.start();
		
		out << "\n_ob_req.arguments(_ob_list);";
		
		String_var result =
		    getTypeString(package, atDesc -> type);

		out.sep();
		out << '\n' << result << " _ob_r = " << atName << "();";

		out.sep();
		out << "\norg.omg.CORBA.Any _ob_ra = _ob_orb.create_any();";
		out << '\n';
		writeInsertExtract(package, atDesc -> type, "_ob_ra", false,
				   out);
		out << "_ob_r";
		TypeCode_var origType = GetOrigType(atDesc -> type);
		if(origType -> kind() == tk_fixed)
		{
		    out << ", _ob_orb.create_fixed_tc((short)"
			<< origType -> fixed_digits() << ", (short)"
			<< origType -> fixed_scale() << ')';
		}
		out << ");";
		out << "\n_ob_req.set_result(_ob_ra);";
		
		out.end();
		
		if(atDesc -> mode == ATTR_NORMAL)
		{
		    out << "\nelse if(_ob_op.equals(\"_set_" << atDesc -> name
			<< "\"))";
		    out.start();
		    
		    out << "\norg.omg.CORBA.Any _ob_aa = "
			<< "_ob_orb.create_any();";
		    out << "\n_ob_aa.type(";
		    writeTypeCode("_ob_orb", package, atDesc -> type, out);
		    out << ");";
		    out << "\n_ob_list.add_value(\"\", _ob_aa, "
			<< "org.omg.CORBA.ARG_IN.value);";
		    out << "\n_ob_req.arguments(_ob_list);";
		  
		    out.sep();
		    out << '\n' << result << " _ob_a = ";
		    writeInsertExtract(package, atDesc -> type, "_ob_aa", true,
				       out);
		    out << ';';

		    out.sep();
		    out << '\n' << atName << "(_ob_a);";
		    
		    out.end();
		}

		first = false;
	    }

	    for(j = 0 ; j < operations -> length() ; j++)
	    {
		Contained::Description_var contDesc =
		    operations[j] -> describe();
		OperationDescription* opDesc;
		Boolean b = (contDesc -> value) >>= opDesc;
		assert(b);
		
		Identifier_var opName = fixName(opDesc -> name);

		if(first)
		{
		    out << "\norg.omg.CORBA.ORB _ob_orb = _orb();";
		    out << "\norg.omg.CORBA.NVList _ob_list = "
			<< "_ob_orb.create_list(0);";
		    out << "\nString _ob_op = _ob_req.operation();";

		    out.sep();
		    out << '\n';
		}
		else
		    out << "\nelse ";

		bool haveOutOrInOut = false;

		out << "if(_ob_op.equals(\"" << opDesc -> name << "\"))";
		out.start();

		for(k = 0 ; k < opDesc -> parameters.length() ; k++)
		{
		    out << "\norg.omg.CORBA.Any _ob_aa" << k
			<< " = _ob_orb.create_any();";

		    switch(opDesc -> parameters[k].mode)
		    {
		    case PARAM_IN:
			out << "\n_ob_aa" << k << ".type(";
			writeTypeCode("_ob_orb", package,
				      opDesc -> parameters[k].type, out);
			out << ");";
			out << "\n_ob_list.add_value(\"\", _ob_aa" << k
			    << ", org.omg.CORBA.ARG_IN.value);";
			break;
			
		    case PARAM_INOUT:
			out << "\n_ob_aa" << k << ".type(";
			writeTypeCode("_ob_orb", package,
				      opDesc -> parameters[k].type, out);
			out << ");";
			out << "\n_ob_list.add_value(\"\", _ob_aa" << k
			    << ", org.omg.CORBA.ARG_INOUT.value);";
			haveOutOrInOut = true;
			break;
			
		    case PARAM_OUT:
			out << "\n_ob_list.add_value(\"\", _ob_aa" << k
			    << ", org.omg.CORBA.ARG_OUT.value);";
			haveOutOrInOut = true;
			break;
		    }
		}

		out << "\n_ob_req.arguments(_ob_list);";

		if(opDesc -> contexts.length() > 0)
		    out << "\norg.omg.CORBA.Context _ob_ctx = _ob_req.ctx();";

		for(k = 0 ; k < opDesc -> parameters.length() ; k++)
		{
		    if(k == 0)
			out.sep();

		    switch(opDesc -> parameters[k].mode)
		    {
		    case PARAM_IN:
		    {
			String_var str =
			    getTypeString(package,
					  opDesc -> parameters[k].type);
			out << '\n' << str << " _ob_a" << k << " = ";
			String_var any = string_dup("_ob_aa");
			any += k;
			writeInsertExtract(package,
					   opDesc -> parameters[k].type,
					   any, true, out);
			out << ';';
			break;
		    }

		    case PARAM_INOUT:
		    {
			String_var str =
			    getTypeString(package,
					  opDesc -> parameters[k].type,
					  GetTypeHolder);
			out << '\n' << str << " _ob_ah" << k << " = new "
			    << str << "();";
			out << "\n_ob_ah" << k << ".value = ";
			String_var any = string_dup("_ob_aa");
			any += k;
			writeInsertExtract(package,
					   opDesc -> parameters[k].type,
					   any, true, out);
			out << ';';
			break;
		    }

		    case PARAM_OUT:
		    {
			String_var str =
			    getTypeString(package,
					  opDesc -> parameters[k].type,
					  GetTypeHolder);
			out << '\n' << str << " _ob_ah" << k << " = new "
			    << str << "();";
			break;
		    }
		    }
		}

		TypeCode_var origRet = GetOrigType(opDesc -> result);
		
		if(origRet -> kind() != tk_void)
		{
		    String_var result =
			getTypeString(package, opDesc -> result);
		    out.sep();
		    if(opDesc -> exceptions.length() > 0)
		    {
			out << "\ntry";
			out.start();
		    }
		    out << '\n' << result << " _ob_r = ";
		}
		else
		{
		    out.sep();
		    if(opDesc -> exceptions.length() > 0)
		    {
			out << "\ntry";
			out.start();
		    }
		    out << '\n';
		}

		out << opName << '(';

		for(k = 0 ; k < opDesc -> parameters.length() ; k++)
		{
		    if(k > 0)
			out << ", ";

		    switch(opDesc -> parameters[k].mode)
		    {
		    case PARAM_IN:
			out << "_ob_a" << k;
			break;

		    case PARAM_INOUT:
		    case PARAM_OUT:
			out << "_ob_ah" << k;
			break;
		    }
		}

                if(opDesc -> contexts.length() > 0)
                {
                    if(opDesc -> parameters.length() > 0)
                        out << ", ";
 
                    out << "_ob_ctx";
                }

		out << ");";

		if(haveOutOrInOut)
		    out.sep();

		for(k = 0 ; k < opDesc -> parameters.length() ; k++)
		{
		    switch(opDesc -> parameters[k].mode)
		    {
		    case PARAM_IN:
			break;

		    case PARAM_INOUT:
		    case PARAM_OUT:
		    {
			out << '\n';
			String_var any = string_dup("_ob_aa");
			any += k;
			writeInsertExtract(package,
					   opDesc -> parameters[k].type, any,
					   false, out);
			out << "_ob_ah" << k << ".value";
			TypeCode_var origType =
			    GetOrigType(opDesc -> parameters[k].type);
			if(origType -> kind() == tk_fixed)
			{
			    out << ", _ob_orb.create_fixed_tc((short)"
				<< origType -> fixed_digits() << ", (short)"
				<< origType -> fixed_scale() << ')';
			}
			out << ");";
			break;
		    }
		    }
		}

		if(origRet -> kind() != tk_void)
		{
		    out.sep();
		    out << "\norg.omg.CORBA.Any _ob_ra = "
			<< "_ob_orb.create_any();";
		    out << '\n';
		    writeInsertExtract(package, opDesc -> result, "_ob_ra",
				       false, out);
		    out << "_ob_r";
		    TypeCode_var origType =
			GetOrigType(opDesc -> result);
		    if(origType -> kind() == tk_fixed)
		    {
			out << ", _ob_orb.create_fixed_tc((short)"
			    << origType -> fixed_digits() << ", (short)"
			    << origType -> fixed_scale() << ')';
		    }
		    out << ");";
		    out << "\n_ob_req.set_result(_ob_ra);";
		}

		if(opDesc -> exceptions.length() > 0)
		{
		    out.end();

		    for(k = 0 ; k < opDesc -> exceptions.length() ; k++)
		    {
			String_var str =
			    getTypeString(package,
					  opDesc -> exceptions[k].type);
			
			String_var strHelper =
			    getTypeString(package,
					  opDesc -> exceptions[k].type,
					  GetTypeHelper);

			out << "\ncatch(" << str << " _ob_ex)";
			out.start();
			out << "\norg.omg.CORBA.Any _ob_exa = "
			    << "_ob_orb.create_any();";
			out << '\n' << strHelper
			    << ".insert(_ob_exa, _ob_ex);";
			out << "\n_ob_req.set_exception(_ob_exa);";
			out.end();
		    }
		}

		out.end();

		first = false;
	    }

	    if(!first)
	    {
		out << "\nelse";
		out.inc();
	    }
	    
	    out << "\nthrow new org.omg.CORBA.BAD_OPERATION();";

	    if(!first)
	    {
		out.dec();
	    }

	    out.end();

	    out.end();
	}

	if(!noStreams_)
	{
	    //
	    // Interface, DSI-based skeleton
	    //
 	    InterfaceDef_var interf =
 		InterfaceDef::_narrow(seq[i].contained_object);
 	    assert(!is_nil(interf));
 	    InterfaceDef::FullInterfaceDescription_var desc =
 		interf -> describe_interface();

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixName(desc -> name);
	    String_var package = getPackage(desc -> id);

	    String_var skel = string_dup("POA_");
	    skel += name;

	    PrettyPrint out(prog_, skel, package, outDir_);
	    if(!out.good())
		return 1;

	    comment(desc -> id, out);

	    out.sep();
	    out << "\npublic abstract class " << skel;
	    out.inc();
	    out << "\nextends org.omg.PortableServer.DynamicImplementation";
	    out << "\nimplements " << name << "Operations";
	    out.dec();

	    out.start();

	    ContainedSeq_var operations =
		interf -> contents(dk_Operation, false);

	    ContainedSeq_var attributes =
		interf -> contents(dk_Attribute, false);

	    InterfaceDefSeq allBaseInterfaces;
	    IdlGetAllBaseInterfaces(interf, allBaseInterfaces);

	    out << "\nstatic final String[] _ob_ids_ =";
	    out.start();
	    out << "\n\"" << desc -> id << "\",";
	    for(j = 0 ; j < allBaseInterfaces.length() ; j++)
	    {
		RepositoryId_var id =
		    allBaseInterfaces[j] -> id();
		out << "\n\"" << id << '"';
		if(j + 1 < allBaseInterfaces.length())
		    out << ',';
	    }
	    out.end();
	    out << ';';

	    out.sep();
	    out << "\npublic String[]\n_ids()";
	    out.start();
	    out << "\nreturn _ob_ids_;";
	    out.end();

	    out.sep();
	    out << "\npublic " << name
		<< "\n_this()";
	    out.start();
	    out << "\nreturn " << name << "Helper.narrow(_this_object());";
	    out.end();

	    out.sep();
	    out << "\npublic " << name
		<< "\n_this(org.omg.CORBA.ORB orb)";
	    out.start();
	    out << "\nreturn " << name << "Helper.narrow(_this_object(orb));";
	    out.end();

	    out.sep();
	    out << "\npublic String[]"
		<< "\n_all_interfaces("
		<< "org.omg.PortableServer.POA poa, byte[] objectId)";
	    out.start();
	    out << "\nreturn _ob_ids_;";
	    out.end();

	    out.sep();
	    out << "\nfinal public void\ninvoke(org.omg.CORBA.ServerRequest "
		<< "_ob_req)";
	    out.start();

	    bool first = true;

	    for(j = 0 ; j < attributes -> length() ; j++)
	    {
		Contained::Description_var contDesc =
		    attributes[j] -> describe();
		AttributeDescription* atDesc;
		Boolean b = (contDesc -> value) >>= atDesc;
		assert(b);
		
		Identifier_var atName = fixName(atDesc -> name);

		if(first)
		{
		    out << "\norg.omg.CORBA.ORB _ob_orb = _orb();";
		    out << "\norg.omg.CORBA.NVList _ob_list = "
			<< "_ob_orb.create_list(0);";
		    out << "\nString _ob_op = _ob_req.operation();";

		    out.sep();
		    out << '\n';
		}
		else
		    out << "\nelse ";

		out << "if(_ob_op.equals(\"_get_" << atDesc -> name << "\"))";
		out.start();
		
		out << "\n_ob_req.arguments(_ob_list);";
		
		String_var result =
		    getTypeString(package, atDesc -> type);

		out.sep();
		out << '\n' << result << " _ob_r = " << atName << "();";

		out.sep();
		out << "\norg.omg.CORBA.Any _ob_ra = _ob_orb.create_any();";
		out << '\n';
		writeInsertExtract(package, atDesc -> type, "_ob_ra", false,
				   out);
		out << "_ob_r";
		TypeCode_var origType = GetOrigType(atDesc -> type);
		if(origType -> kind() == tk_fixed)
		{
		    out << ", _ob_orb.create_fixed_tc((short)"
			<< origType -> fixed_digits() << ", (short)"
			<< origType -> fixed_scale() << ')';
		}
		out << ");";
		out << "\n_ob_req.set_result(_ob_ra);";
		
		out.end();
		
		if(atDesc -> mode == ATTR_NORMAL)
		{
		    out << "\nelse if(_ob_op.equals(\"_set_" << atDesc -> name
			<< "\"))";
		    out.start();
		    
		    out << "\norg.omg.CORBA.Any _ob_aa = "
			<< "_ob_orb.create_any();";
		    out << "\n_ob_aa.type(";
		    writeTypeCode("_ob_orb", package, atDesc -> type, out);
		    out << ");";
		    out << "\n_ob_list.add_value(\"\", _ob_aa, "
			<< "org.omg.CORBA.ARG_IN.value);";
		    out << "\n_ob_req.arguments(_ob_list);";
		  
		    out.sep();
		    out << '\n' << result << " _ob_a = ";
		    writeInsertExtract(package, atDesc -> type, "_ob_aa", true,
				       out);
		    out << ';';

		    out.sep();
		    out << '\n' << atName << "(_ob_a);";
		    
		    out.end();
		}

		first = false;
	    }

	    for(j = 0 ; j < operations -> length() ; j++)
	    {
		Contained::Description_var contDesc =
		    operations[j] -> describe();
		OperationDescription* opDesc;
		Boolean b = (contDesc -> value) >>= opDesc;
		assert(b);
		
		Identifier_var opName = fixName(opDesc -> name);

		if(first)
		{
		    out << "\norg.omg.CORBA.ORB _ob_orb = _orb();";
		    out << "\norg.omg.CORBA.NVList _ob_list = "
			<< "_ob_orb.create_list(0);";
		    out << "\nString _ob_op = _ob_req.operation();";

		    out.sep();
		    out << '\n';
		}
		else
		    out << "\nelse ";

		bool haveOutOrInOut = false;

		out << "if(_ob_op.equals(\"" << opDesc -> name << "\"))";
		out.start();

		for(k = 0 ; k < opDesc -> parameters.length() ; k++)
		{
		    out << "\norg.omg.CORBA.Any _ob_aa" << k
			<< " = _ob_orb.create_any();";

		    switch(opDesc -> parameters[k].mode)
		    {
		    case PARAM_IN:
			out << "\n_ob_aa" << k << ".type(";
			writeTypeCode("_ob_orb", package,
				      opDesc -> parameters[k].type, out);
			out << ");";
			out << "\n_ob_list.add_value(\"\", _ob_aa" << k
			    << ", org.omg.CORBA.ARG_IN.value);";
			break;
			
		    case PARAM_INOUT:
			out << "\n_ob_aa" << k << ".type(";
			writeTypeCode("_ob_orb", package,
				      opDesc -> parameters[k].type, out);
			out << ");";
			out << "\n_ob_list.add_value(\"\", _ob_aa" << k
			    << ", org.omg.CORBA.ARG_INOUT.value);";
			haveOutOrInOut = true;
			break;
			
		    case PARAM_OUT:
			out << "\n_ob_list.add_value(\"\", _ob_aa" << k
			    << ", org.omg.CORBA.ARG_OUT.value);";
			haveOutOrInOut = true;
			break;
		    }
		}

		out << "\n_ob_req.arguments(_ob_list);";

		if(opDesc -> contexts.length() > 0)
		    out << "\norg.omg.CORBA.Context _ob_ctx = _ob_req.ctx();";

		for(k = 0 ; k < opDesc -> parameters.length() ; k++)
		{
		    if(k == 0)
			out.sep();

		    switch(opDesc -> parameters[k].mode)
		    {
		    case PARAM_IN:
		    {
			String_var str =
			    getTypeString(package,
					  opDesc -> parameters[k].type);
			out << '\n' << str << " _ob_a" << k << " = ";
			String_var any = string_dup("_ob_aa");
			any += k;
			writeInsertExtract(package,
					   opDesc -> parameters[k].type,
					   any, true, out);
			out << ';';
			break;
		    }

		    case PARAM_INOUT:
		    {
			String_var str =
			    getTypeString(package,
					  opDesc -> parameters[k].type,
					  GetTypeHolder);
			out << '\n' << str << " _ob_ah" << k << " = new "
			    << str << "();";
			out << "\n_ob_ah" << k << ".value = ";
			String_var any = string_dup("_ob_aa");
			any += k;
			writeInsertExtract(package,
					   opDesc -> parameters[k].type,
					   any, true, out);
			out << ';';
			break;
		    }

		    case PARAM_OUT:
		    {
			String_var str =
			    getTypeString(package,
					  opDesc -> parameters[k].type,
					  GetTypeHolder);
			out << '\n' << str << " _ob_ah" << k << " = new "
			    << str << "();";
			break;
		    }
		    }
		}

		TypeCode_var origRet = GetOrigType(opDesc -> result);
		
		if(origRet -> kind() != tk_void)
		{
		    String_var result =
			getTypeString(package, opDesc -> result);
		    out.sep();
		    if(opDesc -> exceptions.length() > 0)
		    {
			out << "\ntry";
			out.start();
		    }
		    out << '\n' << result << " _ob_r = ";
		}
		else
		{
		    out.sep();
		    if(opDesc -> exceptions.length() > 0)
		    {
			out << "\ntry";
			out.start();
		    }
		    out << '\n';
		}

		out << opName << '(';

		for(k = 0 ; k < opDesc -> parameters.length() ; k++)
		{
		    if(k > 0)
			out << ", ";

		    switch(opDesc -> parameters[k].mode)
		    {
		    case PARAM_IN:
			out << "_ob_a" << k;
			break;

		    case PARAM_INOUT:
		    case PARAM_OUT:
			out << "_ob_ah" << k;
			break;
		    }
		}

		if(opDesc -> contexts.length() > 0)
		{
		    if(opDesc -> parameters.length() > 0)
			out << ", ";

		    out << "_ob_ctx";
		}

		out << ");";

		if(haveOutOrInOut)
		    out.sep();

		for(k = 0 ; k < opDesc -> parameters.length() ; k++)
		{
		    switch(opDesc -> parameters[k].mode)
		    {
		    case PARAM_IN:
			break;

		    case PARAM_INOUT:
		    case PARAM_OUT:
		    {
			out << '\n';
			String_var any = string_dup("_ob_aa");
			any += k;
			writeInsertExtract(package,
					   opDesc -> parameters[k].type, any,
					   false, out);
			out << "_ob_ah" << k << ".value";
			TypeCode_var origType =
			    GetOrigType(opDesc -> parameters[k].type);
			if(origType -> kind() == tk_fixed)
			{
			    out << ", _ob_orb.create_fixed_tc((short)"
				<< origType -> fixed_digits()
				<< ", (short)"
				<< origType -> fixed_scale() << ')';
			}
			out << ");";
			break;
		    }
		    }
		}

		if(origRet -> kind() != tk_void)
		{
		    out.sep();
		    out << "\norg.omg.CORBA.Any _ob_ra = "
			<< "_ob_orb.create_any();";
		    out << '\n';
		    writeInsertExtract(package, opDesc -> result, "_ob_ra",
				       false, out);
		    out << "_ob_r);";
		    out << "\n_ob_req.set_result(_ob_ra);";
		}

		if(opDesc -> exceptions.length() > 0)
		{
		    out.end();

		    for(k = 0 ; k < opDesc -> exceptions.length() ; k++)
		    {
			String_var str =
			    getTypeString(package,
					  opDesc -> exceptions[k].type);
			
			String_var strHelper =
			    getTypeString(package,
					  opDesc -> exceptions[k].type,
					  GetTypeHelper);

			out << "\ncatch(" << str << " _ob_ex)";
			out.start();
			out << "\norg.omg.CORBA.Any _ob_exa = "
			    << "_ob_orb.create_any();";
			out << '\n' << strHelper
			    << ".insert(_ob_exa, _ob_ex);";
			out << "\n_ob_req.set_exception(_ob_exa);";
			out.end();
		    }
		}

		out.end();

		first = false;
	    }

	    if(!first)
	    {
		out << "\nelse";
		out.inc();
	    }
	    
	    out << "\nthrow new org.omg.CORBA.BAD_OPERATION();";

	    if(!first)
	    {
		out.dec();
	    }

	    out.end();

	    out.end();
	}

	if(!noStreams_)
	{
	    //
	    // Interface, stream-based skeleton
	    //
 	    InterfaceDef_var interf =
 		InterfaceDef::_narrow(seq[i].contained_object);
 	    assert(!is_nil(interf));
 	    InterfaceDef::FullInterfaceDescription_var desc =
 		interf -> describe_interface();

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixName(desc -> name);
	    String_var package = getPackage(desc -> id);

	    String_var skel = string_dup("POA_");
	    skel += name;

	    unsigned int indent;

	    PrettyPrint outStream(prog_, skel, package, outDir_);
	    if(!outStream.good())
		return 1;

	    comment(desc -> id, outStream);

	    outStream.sep();
	    outStream << "\npublic abstract class " << skel;
	    outStream.inc();
	    outStream << "\nextends org.omg.PortableServer.Servant"
		      << "\nimplements ";
	    indent = outStream.getIndent();
	    outStream.setIndent(outStream.getPos());
	    outStream << "org.omg.CORBA.portable.InvokeHandler,"
		      << '\n' << name << "Operations";
	    outStream.setIndent(indent);
	    outStream.dec();

	    outStream.start();

	    ContainedSeq_var operations =
		interf -> contents(dk_Operation, false);

	    ContainedSeq_var attributes =
		interf -> contents(dk_Attribute, false);

	    InterfaceDefSeq allBaseInterfaces;
	    IdlGetAllBaseInterfaces(interf, allBaseInterfaces);

	    outStream << "\nstatic final String[] _ob_ids_ =";
	    outStream.start();
	    outStream << "\n\"" << desc -> id << "\",";
	    for(j = 0 ; j < allBaseInterfaces.length() ; j++)
	    {
		RepositoryId_var id =
		    allBaseInterfaces[j] -> id();
		outStream << "\n\"" << id << '"';
		if(j + 1 < allBaseInterfaces.length())
		    outStream << ',';
	    }
	    outStream.end();
	    outStream << ';';

	    outStream.sep();
	    outStream << "\nfinal public " << name << " _this()";
	    outStream.start();
	    outStream << "\nreturn " << name
		      << "Helper.narrow(super._this_object());";
	    outStream.end();
	    outStream << "\nfinal public " << name
		      << " _this(org.omg.CORBA.ORB orb)";
	    outStream.start();
	    outStream << "\nreturn " << name
		      << "Helper.narrow(super._this_object(orb));";
	    outStream.end();

	    outStream.sep();
	    outStream << "\nfinal public String[]"
		      << "\n_all_interfaces("
		      << "org.omg.PortableServer.POA poa, byte[] objectId)";
	    outStream.start();
	    outStream << "\nreturn _ob_ids_;";
	    outStream.end();

	    outStream.sep();
	    outStream << "\nfinal public org.omg.CORBA.portable.OutputStream "
		      << "\n_invoke(";
	    indent = outStream.getIndent();
	    outStream.setIndent(outStream.getPos());
	    outStream << "String opName,";
	    outStream << "\norg.omg.CORBA.portable.InputStream in,";
	    outStream << "\norg.omg.CORBA.portable.ResponseHandler handler)";
	    outStream.setIndent(indent);

	    outStream.start();

	    if(attributes -> length() > 0 || operations -> length() > 0)
	    {
		outStream << "\norg.omg.CORBA.portable.OutputStream "
			  << "out = null;";
		outStream.sep();
	    }

	    for(j = 0 ; j < attributes -> length() ; j++)
	    {
		Contained::Description_var contDesc =
		    attributes[j] -> describe();
		AttributeDescription* atDesc;
		Boolean b = (contDesc -> value) >>= atDesc;
		assert(b);
		
		Identifier_var atName = fixName(atDesc -> name);

		String_var result =
		    getTypeString(package, atDesc -> type);

		outStream << "\nif(opName.equals(\"_get_" << atDesc -> name
			  << "\"))";
		outStream.start();
		outStream << '\n' << result << " _ob_r = " << atDesc -> name
			  << "();";
		outStream << "\nout = handler.createReply();";
		ULong iter = 0;
		writeReaderWriter(package, atDesc -> type, "_ob_r", iter,
				  false, outStream);
		outStream << "\nreturn out;";
		outStream.end();
		
		if(atDesc -> mode == ATTR_NORMAL)
		{
		    outStream << "\nelse if(opName.equals(\"_set_"
			      << atDesc -> name << "\"))";
		    outStream.start();
		    String_var str = result;
		    str += " _ob_a";
		    ULong iter = 0;
		    writeReaderWriter(package, atDesc -> type, str, iter, true,
				      outStream);
		    outStream << '\n' << atName << "(_ob_a);";
		    outStream << "\nreturn handler.createReply();";
		    outStream.end();
		}
	    }

	    for(j = 0 ; j < operations -> length() ; j++)
	    {
		Contained::Description_var contDesc =
		    operations[j] -> describe();
		OperationDescription* opDesc;
		Boolean b = (contDesc -> value) >>= opDesc;
		assert(b);
		
		Identifier_var opName = fixName(opDesc -> name);

		outStream << '\n';
		if(j || attributes -> length() > 0)
		    outStream << "else ";
		outStream << "if(opName.equals(\"" << opDesc -> name
			  << "\"))";
		outStream.start();

		if(opDesc -> exceptions.length() > 0)
		{
		    outStream << "\ntry";
		    outStream.start();
		}

		for(k = 0 ; k < opDesc -> parameters.length() ; k++)
		{
		    switch(opDesc -> parameters[k].mode)
		    {
		    case PARAM_IN:
		    {
			String_var str =
			    getTypeString(package,
					  opDesc -> parameters[k].type);
			str += " _ob_a";
			str += k;
			ULong iter = 0;
			writeReaderWriter(package,
					  opDesc -> parameters[k].type, str,
					  iter, true, outStream);
			break;
		    }

		    case PARAM_INOUT:
		    {
			String_var str =
			    getTypeString(package,
					  opDesc -> parameters[k].type,
					  GetTypeHolder);
			outStream << '\n' << str << " _ob_ah" << k << " = new "
				  << str << "();";
			str = string_dup("_ob_ah");
			str += k;
			str += ".value";
			ULong iter = 0;
			writeReaderWriter(package,
					  opDesc -> parameters[k].type, str,
					  iter, true, outStream);
			break;
		    }

		    case PARAM_OUT:
		    {
			String_var str =
			    getTypeString(package,
					  opDesc -> parameters[k].type,
					  GetTypeHolder);
			outStream << '\n' << str << " _ob_ah" << k << " = new "
				  << str << "();";
			break;
		    }
		    }
		}

		TypeCode_var origRet = GetOrigType(opDesc -> result);
		
		if(origRet -> kind() != tk_void)
		{
		    String_var result =
			getTypeString(package, opDesc -> result);
		    outStream << '\n' << result << " _ob_r = ";
		}
		else
		    outStream << '\n';

		outStream << opName << '(';

		for(k = 0 ; k < opDesc -> parameters.length() ; k++)
		{
		    if(k > 0)
			outStream << ", ";

		    switch(opDesc -> parameters[k].mode)
		    {
		    case PARAM_IN:
			outStream << "_ob_a" << k;
			break;

		    case PARAM_INOUT:
		    case PARAM_OUT:
			outStream << "_ob_ah" << k;
			break;
		    }
		}

		if(opDesc -> contexts.length() > 0)
		{
		    if(opDesc -> parameters.length() > 0)
			outStream << ", in.read_Context()";
		}

		outStream << ");";

		outStream << "\nout = handler.createReply();";

		for(k = 0 ; k < opDesc -> parameters.length() ; k++)
		{
		    switch(opDesc -> parameters[k].mode)
		    {
		    case PARAM_IN:
			break;

		    case PARAM_INOUT:
		    case PARAM_OUT:
		    {
			String_var str = string_dup("_ob_ah");
			str += k;
			str += ".value";
			ULong iter = 0;
			writeReaderWriter(package,
					  opDesc -> parameters[k].type, str,
					  iter, false, outStream);
			break;
		    }
		    }
		}

		if(origRet -> kind() != tk_void)
		{
		    ULong iter = 0;
		    writeReaderWriter(package,
				      opDesc -> result, "_ob_r", iter, false,
				      outStream);
		}

		if(opDesc -> exceptions.length() > 0)
		{
		    outStream.end();

		    for(k = 0 ; k < opDesc -> exceptions.length() ; k++)
		    {
			String_var str =
			    getTypeString(package,
					  opDesc -> exceptions[k].type);
			
			String_var strHelper =
			    getTypeString(package,
					  opDesc -> exceptions[k].type,
					  GetTypeHelper);

			outStream << "\ncatch(" << str << " _ob_ex)";
			outStream.start();
			outStream << "\nout = handler."
				  << "createExceptionReply();";
			outStream << '\n' << strHelper
			    << ".write(out, _ob_ex);";
			outStream.end();
		    }
		}

		outStream << "\nreturn out;";

		outStream.end();
	    }

	    if(attributes -> length() > 0 || operations -> length() > 0)
	    {
		outStream << "\nelse";
		outStream.inc();
		outStream << "\nthrow new org.omg.CORBA.BAD_OPERATION();";
		outStream.dec();
	    }
	    else
		outStream << "\nthrow new org.omg.CORBA.BAD_OPERATION();";

	    outStream.end();

	    outStream.end();

	    break;
	}

	default:
	    break;
	}
    }

    return status;
}

// ----------------------------------------------------------------------
// Generate Java TIE classes
// ----------------------------------------------------------------------

int
IdlJavaGenerator::genJavaTie(Container_ptr container,
			     IdlPrettyPrint& out)
{
    //
    // Get container contents descriptions
    //
    Container::DescriptionSeq_var seq =
	container -> describe_contents(dk_all, true, -1);

    ULong i, j, k;

    int status = 0;
    for(i = 0 ; status == 0 && i < seq -> length() ; i++)
    {
	//
	// Get container
	//
	Container_var cont =
	    Container::_narrow(seq[i].contained_object);

	switch(seq[i].kind)
	{
	case dk_Module:
	{
	    //
	    // Module
	    //
	    ModuleDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    // Don't add this here!!!
	    // if(!check(desc -> id))
	    //     continue;

	    assert(!is_nil(cont));
	    status = genJavaTie(cont, out);

	    break;
	}

	case dk_Interface:
	{
	if(noStreams_)
	{
	    //
	    // Interface
	    //
 	    InterfaceDef_var interf =
 		InterfaceDef::_narrow(seq[i].contained_object);
 	    assert(!is_nil(interf));
 	    InterfaceDef::FullInterfaceDescription_var desc =
 		interf -> describe_interface();

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixName(desc -> name);
	    String_var package = getPackage(desc -> id);

	    //
	    // Name of Operations Interface
	    //
	    String_var op = string_dup(name);
	    op += "Operations";

	    //
	    // TIE class for the interface
	    //
	    String_var tie = string_dup("_");
	    tie += name;
	    tie += "ImplBase_tie";
	    PrettyPrint outTie(prog_, tie, package, outDir_);
	    if(!outTie.good())
		return 1;

	    comment(desc -> id, outTie);

	    outTie.sep();
	    outTie << "\npublic class " << tie << " extends _" << name
		   << "ImplBase";

	    outTie.start();

	    outTie << "\nprivate " << op << " delegate_;";
	    outTie.sep();

	    outTie << "\npublic" << '\n' << tie << "(" << op << " op)";
	    outTie.start();
	    outTie << "\ndelegate_ = op;";
	    outTie.end();

	    outTie.sep();

	    outTie << "\npublic " << op << '\n' << "_delegate()";
	    outTie.start();
	    outTie << "\nreturn delegate_;";
	    outTie.end();

	    outTie.sep();

	    outTie << "\npublic void" << '\n' << "_delegate(" << op
		   << " op)";
	    outTie.start();
	    outTie << "\ndelegate_ = op;";
	    outTie.end();

	    ContainedSeq_var attributes =
		interf -> contents(dk_Attribute, false);

	    ContainedSeq_var operations =
		interf -> contents(dk_Operation, false);

	    for(j = 0 ; j < attributes -> length() ; j++)
	    {
		Contained::Description_var contDesc =
		    attributes[j] -> describe();
		AttributeDescription* atDesc;
		Boolean b = (contDesc -> value) >>= atDesc;
		assert(b);

		Identifier_var atName = fixName(atDesc -> name);

		comment(atDesc -> id, outTie);

		String_var result =
		    getTypeString(package, atDesc -> type);
		outTie.sep();
		outTie << "\npublic " << result << '\n' << atName << "()";
		outTie.start();
		outTie << "\nreturn delegate_." << atName << "();";
		outTie.end();

		if(atDesc -> mode == ATTR_NORMAL)
		{
		    outTie.sep();
		    outTie << "\npublic void\n" << atName << '('
			<< result << " _ob_a)";
		    outTie.start();
		    outTie << "\ndelegate_." << atName << "(_ob_a);";
		    outTie.end();
		}
	    }

	    for(j = 0 ; j < operations -> length() ; j++)
	    {
		Contained::Description_var contDesc =
		    operations[j] -> describe();
		OperationDescription* opDesc;
		Boolean b = (contDesc -> value) >>= opDesc;
		assert(b);

		Identifier_var opName = fixName(opDesc -> name);

		comment(opDesc -> id, outTie);

		TypeCode_var origRet =
		    GetOrigType(opDesc -> result);

		String_var result = getTypeString(package, origRet);
		outTie.sep();
		outTie << "\npublic " << result << '\n' << opName << "(";

		unsigned int indent = outTie.getIndent();
		outTie.setIndent(outTie.getPos());

		for(k = 0 ; k < opDesc -> parameters.length() ; k++)
		{
		    if(k > 0)
			outTie << ",\n";

		    switch(opDesc -> parameters[k].mode)
		    {
		    case PARAM_IN:
		    {
			String_var str =
			    getTypeString(package,
					  opDesc -> parameters[k].type);
			outTie << str << " _ob_a" << k;
			break;
		    }

		    case PARAM_INOUT:
		    case PARAM_OUT:
		    {
			String_var str =
			    getTypeString(package,
					  opDesc -> parameters[k].type,
					  GetTypeHolder);
			outTie << str << " _ob_ah" << k;
			break;
		    }
		    }
		}

		if(opDesc -> contexts.length() > 0)
		{
		    if(opDesc -> parameters.length() > 0)
			outTie << ",\n";

		    outTie << "org.omg.CORBA.Context _ob_ctx";
		}

		outTie.setIndent(indent);

		outTie << ')';

		if(opDesc -> exceptions.length())
		{
		    outTie.inc();
		    outTie << "\nthrows ";

		    unsigned int indent = outTie.getIndent();
		    outTie.setIndent(outTie.getPos());

		    for(k = 0 ; k < opDesc -> exceptions.length() ; k++)
		    {
			if(k)
			    outTie << ",\n";

			String_var str =
			    getTypeString(package,
					  opDesc -> exceptions[k].type);

			outTie << str;
		    }

		    outTie.setIndent(indent);
		    outTie.dec();
		}

		outTie.start();

		outTie << '\n';

		if(origRet -> kind() != tk_void)
		    outTie << "return ";

		outTie << "delegate_." << opName << "(";
		indent = outTie.getIndent();
		outTie.setIndent(outTie.getPos());

		for(k = 0 ; k < opDesc -> parameters.length() ; k++)
		{
		    if(k > 0)
			outTie << ",\n";

		    switch(opDesc -> parameters[k].mode)
		    {
		    case PARAM_IN:
		    {
			outTie << "_ob_a" << k;
			break;
		    }

		    case PARAM_INOUT:
		    case PARAM_OUT:
		    {
			outTie << "_ob_ah" << k;
			break;
		    }
		    }
		}

		if(opDesc -> contexts.length() > 0)
		{
		    if(opDesc -> parameters.length() > 0)
			outTie << ",\n";

		    outTie << "_ob_ctx";
		}

		outTie << ");";
		outTie.setIndent(indent);
		outTie.end();
	    }

	    outTie.end();
	}

	if(!noStreams_)
	    {
	    //
	    // Interface
	    //
		InterfaceDef_var interf =
 		InterfaceDef::_narrow(seq[i].contained_object);
 	    assert(!is_nil(interf));
 	    InterfaceDef::FullInterfaceDescription_var desc =
 		interf -> describe_interface();

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixName(desc -> name);
	    String_var package = getPackage(desc -> id);

	    //
	    // Name of Operations Interface
	    //
	    String_var op = string_dup(name);
	    op += "Operations";

	    //
	    // TIE class for the interface
	    //
	    String_var tie = string_dup("POA_");
	    tie += name;
	    tie += "_tie";
	    PrettyPrint outTie(prog_, tie, package, outDir_);
	    if(!outTie.good())
		return 1;

	    comment(desc -> id, outTie);

	    outTie.sep();
	    outTie << "\npublic class " << tie << " extends POA_" << name;

	    outTie.start();

	    outTie << "\nprivate " << op << " delegate_;";
	    outTie << "\nprivate org.omg.PortableServer.POA poa_;";
	    outTie.sep();

	    outTie << "\npublic" << '\n' << tie << "(" << op
		   << " delegate)";
	    outTie.start();
	    outTie << "\ndelegate_ = delegate;";
	    outTie.end();

	    outTie.sep();

	    outTie << "\npublic" << '\n' << tie << "(" << op
		   << " delegate, org.omg.PortableServer.POA poa)";
	    outTie.start();
	    outTie << "\ndelegate_ = delegate;";
	    outTie << "\npoa_ = poa;";
	    outTie.end();

	    outTie.sep();

	    outTie << "\npublic " << op << '\n' << "_delegate()";
	    outTie.start();
	    outTie << "\nreturn delegate_;";
	    outTie.end();

	    outTie.sep();

	    outTie << "\npublic void" << '\n' << "_delegate(" << op
		   << " delegate)";
	    outTie.start();
	    outTie << "\ndelegate_ = delegate;";
	    outTie.end();

	    outTie.sep();

	    outTie << "\npublic org.omg.PortableServer.POA"
		   << "\n_default_POA()";
	    outTie.start();
	    outTie << "\nif(poa_ != null)";
	    outTie.inc();
	    outTie << "\nreturn poa_;";
	    outTie.dec();
	    outTie << "\nelse";
	    outTie.inc();
	    outTie << "\nreturn super._default_POA();";
	    outTie.dec();
	    outTie.end();

	    ContainedSeq_var attributes =
		interf -> contents(dk_Attribute, false);

	    ContainedSeq_var operations =
		interf -> contents(dk_Operation, false);

	    for(j = 0 ; j < attributes -> length() ; j++)
	    {
		Contained::Description_var contDesc =
		    attributes[j] -> describe();
		AttributeDescription* atDesc;
		Boolean b = (contDesc -> value) >>= atDesc;
		assert(b);

		Identifier_var atName = fixName(atDesc -> name);

		comment(atDesc -> id, outTie);

		String_var result =
		    getTypeString(package, atDesc -> type);
		outTie.sep();
		outTie << "\npublic " << result << '\n' << atName << "()";
		outTie.start();
		outTie << "\nreturn delegate_." << atName << "();";
		outTie.end();

		if(atDesc -> mode == ATTR_NORMAL)
		{
		    outTie.sep();
		    outTie << "\npublic void\n" << atName << '('
			<< result << " _ob_a)";
		    outTie.start();
		    outTie << "\ndelegate_." << atName << "(_ob_a);";
		    outTie.end();
		}
	    }

	    for(j = 0 ; j < operations -> length() ; j++)
	    {
		Contained::Description_var contDesc =
		    operations[j] -> describe();
		OperationDescription* opDesc;
		Boolean b = (contDesc -> value) >>= opDesc;
		assert(b);

		Identifier_var opName = fixName(opDesc -> name);

		comment(opDesc -> id, outTie);

		TypeCode_var origRet =
		    GetOrigType(opDesc -> result);

		String_var result = getTypeString(package, origRet);
		outTie.sep();
		outTie << "\npublic " << result << '\n' << opName << "(";

		unsigned int indent = outTie.getIndent();
		outTie.setIndent(outTie.getPos());

		for(k = 0 ; k < opDesc -> parameters.length() ; k++)
		{
		    if(k > 0)
			outTie << ",\n";

		    switch(opDesc -> parameters[k].mode)
		    {
		    case PARAM_IN:
		    {
			String_var str =
			    getTypeString(package,
					  opDesc -> parameters[k].type);
			outTie << str << " _ob_a" << k;
			break;
		    }

		    case PARAM_INOUT:
		    case PARAM_OUT:
		    {
			String_var str =
			    getTypeString(package,
					  opDesc -> parameters[k].type,
					  GetTypeHolder);
			outTie << str << " _ob_ah" << k;
			break;
		    }
		    }
		}

		if(opDesc -> contexts.length() > 0)
		{
		    if(opDesc -> parameters.length() > 0)
			outTie << ",\n";

		    outTie << "org.omg.CORBA.Context _ob_ctx";
		}

		outTie.setIndent(indent);

		outTie << ')';

		if(opDesc -> exceptions.length())
		{
		    outTie.inc();
		    outTie << "\nthrows ";

		    unsigned int indent = outTie.getIndent();
		    outTie.setIndent(outTie.getPos());

		    for(k = 0 ; k < opDesc -> exceptions.length() ; k++)
		    {
			if(k)
			    outTie << ",\n";

			String_var str =
			    getTypeString(package,
					  opDesc -> exceptions[k].type);

			outTie << str;
		    }

		    outTie.setIndent(indent);
		    outTie.dec();
		}

		outTie.start();

		outTie << '\n';

		if(origRet -> kind() != tk_void)
		    outTie << "return ";

		outTie << "delegate_." << opName << "(";
		indent = outTie.getIndent();
		outTie.setIndent(outTie.getPos());

		for(k = 0 ; k < opDesc -> parameters.length() ; k++)
		{
		    if(k > 0)
			outTie << ",\n";

		    switch(opDesc -> parameters[k].mode)
		    {
		    case PARAM_IN:
		    {
			outTie << "_ob_a" << k;
			break;
		    }

		    case PARAM_INOUT:
		    case PARAM_OUT:
		    {
			outTie << "_ob_ah" << k;
			break;
		    }
		    }
		}

		if(opDesc -> contexts.length() > 0)
		{
		    if(opDesc -> parameters.length() > 0)
			outTie << ",\n";

		    outTie << "_ob_ctx";
		}

		outTie << ");";
		outTie.setIndent(indent);
		outTie.end();
	    }

	    outTie.end();
	    }
	}

	    break;

	case dk_Attribute:
	{
	    //
	    // Attribute
	    //
	    AttributeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixName(desc -> name);
	    String_var package = getPackage(desc -> id);

	    comment(desc -> id, out);
	    docComment(desc -> id, out);

	    String_var norm = getTypeString(package, desc -> type);

	    out.sep();
	    out << "\npublic " << norm << '\n' << name << "();";
	    if(desc -> mode == ATTR_NORMAL)
	    {
		out.sep();
		out << "\npublic void\n" << name << "(" << norm << " value);";
	    }

	    break;
	}

	case dk_Operation:
	{
	    //
	    // Operation
	    //
	    OperationDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixName(desc -> name);
	    String_var package = getPackage(desc -> id);

	    comment(desc -> id, out);
	    docComment(desc -> id, out);

	    String_var norm = getTypeString(package, desc -> result);

	    out.sep();
	    out << "\npublic " << norm << '\n' << name << "(";
	    
	    unsigned int indent = out.getIndent();
	    out.setIndent(out.getPos());

	    for(j = 0 ; j < desc -> parameters.length() ; j++)
	    {
		GetType gt;
		switch(desc -> parameters[j].mode)
		{
		case PARAM_IN:
		    gt = GetTypeNormal;
		    break;

		case PARAM_INOUT:
		case PARAM_OUT:
		    gt = GetTypeHolder;
		    break;
		}

		String_var str =
		    getTypeString(package, desc -> parameters[j].type, gt);

		if(j > 0)
		    out << ",\n";

		String_var argName = fixName(desc -> parameters[j].name);

		out << str << ' ' << argName;
	    }

	    if(desc -> contexts.length() > 0)
	    {
		if(desc -> parameters.length() > 0)
		    out << ",\n";
		
		out << "org.omg.CORBA.Context ctx";
	    }

	    out.setIndent(indent);

	    out << ')';

	    if(desc -> exceptions.length())
	    {
		out.inc();
		out << "\nthrows ";

		unsigned int indent = out.getIndent();
		out.setIndent(out.getPos());

		for(j = 0 ; j < desc -> exceptions.length() ; j++)
		{
		    if(j)
			out << ",\n";

		    String_var str =
			getTypeString(package, desc -> exceptions[j].type);
    
		    out << str;
		}

		out.setIndent(indent);
		out.dec();
	    }

	    out << ';';

	    break;
	}

	default:
	    break;
	}
    }

    return status;
}

// ----------------------------------------------------------------------
// Write constants
// ----------------------------------------------------------------------

void
IdlJavaGenerator::writeConstant(const Any& value, IdlPrettyPrint& out)
{
    TypeCode_var tc = value.type();
    tc = GetOrigType(tc);

    switch(tc -> kind())
    {
    case tk_short:
    {
	Short v;
	Boolean b = value >>= v;
	assert(b);
	out << "(short)(" << v << "L)";
	break;
    }

    case tk_ushort:
    {
	UShort v;
	Boolean b = value >>= v;
	assert(b);
	out << "(short)(" << v << "L)";
	break;
    }

    case tk_long:
    {
	Long v;
	Boolean b = value >>= v;
	assert(b);
	out << "(int)(" << v << "L)";
	break;
    }

    case tk_ulong:
    {
	ULong v;
	Boolean b = value >>= v;
	assert(b);
	out << "(int)(" << v << "L)";
	break;
    }

    case tk_float:
    {
	Float v;
	Boolean b = value >>= v;
	assert(b);
	out << "(float)(" << v << "D)";
	break;
    }

    case tk_double:
    {
	Double v;
	Boolean b = value >>= v;
	assert(b);
	out << "(double)(" << v << "D)";
	break;
    }

    case tk_string:
    {
	char* v;
	ULong l = tc -> length();
	Any::to_string to(v, l);
	Boolean b = value >>= to;
	assert(b);
	String_var s = string_dup(v);
	addEscapes(s.inout(), true);
	out << s;
	break;
    }

    case tk_wstring:
    {
	wchar_t* v;
	Boolean b = value >>= v;
	assert(b);
	WString_var s = wstring_dup(v);
	addEscapes(s.inout(), true);
	out << s;
	break;
    }

    case tk_char:
    {
	Char v;
	Boolean b = value >>= Any::to_char(v);
	assert(b);
	String_var s;
	if(v == '\0')
	    s = string_dup("\\0");
	else
	{
	    s = string_alloc(1);
	    s[0] = v;
	    s[1] = '\0';
	    addEscapes(s.inout(), false);
	}
	out << "'" << s << "'";
	break;
    }

    case tk_wchar:
    {
	WChar v;
	Boolean b = value >>= Any::to_wchar(v);
	assert(b);
	WString_var s;
	if(v == L'\0')
	    s = wstring_dup(L"\\0");
	else
	{
	    s = wstring_alloc(1);
	    s[0] = v;
	    s[1] = '\0';
	    addEscapes(s.inout(), false);
	}
	out << "'" << s << "'";
	break;
    }

    case tk_boolean:
    {
	Boolean v;
	Boolean b = value >>= Any::to_boolean(v);
	assert(b);
	out << (v == TRUE ? "true" : "false");
	break;
    }

    default:
	assert(false);
	break;
    }
}

// ----------------------------------------------------------------------
// Write the holder class
// ----------------------------------------------------------------------

void
IdlJavaGenerator::writeHolder(const char* norm, const char* name,
			      IdlPrettyPrint& out)
{
    out.sep();
    out << "\nfinal public class " << name << "Holder"
	<< " implements org.omg.CORBA.portable.Streamable";
    out.start();

    out << "\npublic " << norm << " value;";

    out.sep();
    out << "\npublic\n" << name << "Holder" << "()";
    out.start();
    out.end();

    out.sep();
    out << "\npublic\n" << name << "Holder" << "(" << norm << " initial)";
    out.start();
    out << "\nvalue = initial;";
    out.end();

    out.sep();
    out << "\npublic void\n_read(org.omg.CORBA.portable.InputStream in)";
    out.start();
    if(local_)
	out << "\nthrow new org.omg.CORBA.MARSHAL();";
    else
	out << "\nvalue = " << name << "Helper.read(in);";
    out.end();

    out.sep();
    out << "\npublic void\n_write(org.omg.CORBA.portable.OutputStream out)";
    out.start();
    if(local_)
	out << "\nthrow new org.omg.CORBA.MARSHAL();";
    else
	out << "\n" << name << "Helper.write(out, value);";
    out.end();

    out.sep();
    out << "\npublic org.omg.CORBA.TypeCode\n_type()";
    out.start();
    if(local_)
	out << "\nthrow new org.omg.CORBA.BAD_OPERATION();";
    else
	out << "\nreturn " << name << "Helper.type();";
    out.end();

    out.end();
}

// ----------------------------------------------------------------------
// Write the helper class
// ----------------------------------------------------------------------

void
IdlJavaGenerator::writeHelper(const char* package, TypeCode_ptr tc,
			      IdlPrettyPrint& out)
{
    RepositoryId_var id = tc -> id();
    Identifier_var name = fixName(tc -> name());
    String_var norm = getTypeString(package, tc);

    TypeCode_var origType = GetOrigType(tc);

    out.sep();
    out << "\nfinal public class " << name << "Helper implements ";

    if(origType -> kind() == tk_value ||
       origType -> kind() == tk_value_box)
	out << "org.omg.CORBA.portable.ValueHelper";
    else
	out << "org.omg.CORBA.portable.Helper";
    out.start();

    //
    // Insertion to any
    //
    out << "\npublic static void\ninsert(org.omg.CORBA.Any any, " << norm
	<< " val)";
    out.start();
    if(local_)
    {
	out << "\nthrow new org.omg.CORBA.MARSHAL();";
    }
    else
    {
	out << "\norg.omg.CORBA.ORB orb = org.omg.CORBA.ORB.init();";

	switch(origType -> kind())
	{
	case tk_null:
	case tk_void:
	    assert(false);
	    break;
	    
	case tk_short:
	    out << "\nany.insert_short(val);";
	    break;
	    
	case tk_long:
	    out << "\nany.insert_long(val);";
	    break;
	    
	case tk_ushort:
	    out << "\nany.insert_ushort(val);";
	    break;
	    
	case tk_ulong:
	    out << "\nany.insert_ulong(val);";
	    break;

	case tk_longlong:
	    out << "\nany.insert_longlong(val);";
	    break;
	    
	case tk_ulonglong:
	    out << "\nany.insert_ulonglong(val);";
	    break;
	    
	case tk_float:
	    out << "\nany.insert_float(val);";
	    break;
	    
	case tk_double:
	    out << "\nany.insert_double(val);";
	    break;

	case tk_fixed:
	    out << "\nany.insert_fixed(val, orb.create_fixed_tc((short)"
		<< origType -> fixed_digits() << ", (short)"
		<< origType -> fixed_scale() << "));";
	    break;
	    
	case tk_boolean:
	    out << "\nany.insert_boolean(val);";
	    break;
	    
	case tk_char:
	    out << "\nany.insert_char(val);";
	    break;
	    
	case tk_wchar:
	    out << "\nany.insert_wchar(val);";
	    break;
	    
	case tk_octet:
	    out << "\nany.insert_octet(val);";
	    break;
	    
	case tk_string:
	    out << "\nany.insert_string(val);";
	    break;
	    
	case tk_wstring:
	    out << "\nany.insert_wstring(val);";
	    break;
	    
	case tk_any:
	    out << "\nany.insert_any(val);";
	    break;
	    
	case tk_TypeCode:
	    out << "\nany.insert_TypeCode(val);";
	    break;
	    
	case tk_Principal:
	    out << "\nany.insert_Principal(val);";
	    break;
	    
	case tk_objref:
	    if(strcmp(origType -> id(), "IDL:omg.org/CORBA/Object:1.0") == 0)
	    {
		out << "\nany.insert_Object(val);";
		break;
	    }
	    // No break here!

	case tk_struct:
	case tk_union:
	case tk_except:
	case tk_enum:
	case tk_sequence:
	case tk_array:
	    out << "\norg.omg.CORBA.portable.OutputStream out = "
		<< "any.create_output_stream();";
	    out << "\nwrite(out, val);";
	    out << "\nany.read_value(out.create_input_stream(), type());";
	    break;
	    
	case tk_value:
	case tk_value_box:
	    out << "\nany.insert_Value(val);";
	    break;

	case tk_alias:
	    assert(false);
	    break;
	}
    }
    out.end();

    //
    // Extraction from any
    //
    out.sep();
    out << "\npublic static " << norm << "\nextract(org.omg.CORBA.Any any)";
    out.start();
    if(local_)
    {
	out << "\nthrow new org.omg.CORBA.MARSHAL();";
    }
    else
    {
	switch(origType -> kind())
	{
	case tk_null:
	case tk_void:
	    assert(false);
	    break;

	case tk_short:
	    out << "\nreturn any.extract_short();";
	    break;

	case tk_long:
	    out << "\nreturn any.extract_long();";
	    break;

	case tk_longlong:
	    out << "\nreturn any.extract_longlong();";
	    break;

	case tk_ushort:
	    out << "\nreturn any.extract_ushort();";
	    break;

	case tk_ulong:
	    out << "\nreturn any.extract_ulong();";
	    break;

	case tk_ulonglong:
	    out << "\nreturn any.extract_ulonglong();";
	    break;

	case tk_float:
	    out << "\nreturn any.extract_float();";
	    break;

	case tk_double:
	    out << "\nreturn any.extract_double();";
	    break;

	case tk_fixed:
	    out << "\nreturn any.extract_fixed();";
	    break;

	case tk_boolean:
	    out << "\nreturn any.extract_boolean();";
	    break;

	case tk_char:
	    out << "\nreturn any.extract_char();";
	    break;

	case tk_wchar:
	    out << "\nreturn any.extract_wchar();";
	    break;

	case tk_octet:
	    out << "\nreturn any.extract_octet();";
	    break;

	case tk_string:
	    out << "\nreturn any.extract_string();";
	    break;

	case tk_wstring:
	    out << "\nreturn any.extract_wstring();";
	    break;

	case tk_any:
	    out << "\nreturn any.extract_any();";
	    break;

	case tk_TypeCode:
	    out << "\nreturn any.extract_TypeCode();";
	    break;

	case tk_Principal:
	    out << "\nreturn any.extract_Principal();";
	    break;

	case tk_objref:
	    if(strcmp(origType -> id(), "IDL:omg.org/CORBA/Object:1.0") == 0)
	    {
		out << "\nreturn any.extract_Object();";
		break;
	    }
	    // No break here!

	case tk_struct:
	case tk_union:
	case tk_except:
	case tk_enum:
	case tk_sequence:
	case tk_array:
	    out << "\nif(any.type().equivalent(type()))";
	    out.inc();
	    out << "\nreturn read(any.create_input_stream());";
	    out.dec();
	    out << "\nelse";
	    out.inc();
	    out << "\nthrow new org.omg.CORBA.BAD_OPERATION();";
	    out.dec();
	    break;

	case tk_value:
	case tk_value_box:
	    out << "\nreturn any.extract_Value();";
	    break;

	case tk_alias:
	    assert(false);
	    break;
	}
    }
    out.end();

    //
    // Return the TypeCode
    //
    if(!local_)
    {
	out.sep();
	out << "\nprivate static org.omg.CORBA.TypeCode typeCode_;";
    }
    out.sep();
    out << "\npublic static org.omg.CORBA.TypeCode\ntype()";
    out.start();
    if(local_)
    {
	out << "\nthrow new org.omg.CORBA.BAD_OPERATION();";
    }
    else
    {
	out << "\nif(typeCode_ == null)";
	out.start();
	out << "\norg.omg.CORBA.ORB orb = org.omg.CORBA.ORB.init();";
	switch(tc -> kind()) // Don't use origType here!
	{
	case tk_null:
	case tk_void:
	    assert(false);
	    break;

	case tk_short:
	case tk_long:
	case tk_ushort:
	case tk_ulong:
	case tk_longlong:
	case tk_ulonglong:
	case tk_float:
	case tk_double:
	case tk_fixed:
	case tk_boolean:
	case tk_char:
	case tk_wchar:
	case tk_octet:
	case tk_string:
	case tk_wstring:
	case tk_any:
	case tk_TypeCode:
	case tk_Principal:
	case tk_sequence:
	case tk_array:
	    out << "\ntypeCode_ = ";
	    writeTypeCode("orb", package, tc, out);
	    out << ';';
	    break;

	case tk_objref:
	{
	    RepositoryId_var id = tc -> id();
	    Identifier_var name = tc -> name();
	    out << "\ntypeCode_ = orb.create_interface_tc(id(), \""
		<< name << "\");";
	    break;
	}

	case tk_struct:
	case tk_except:
	{
	    out << "\norg.omg.CORBA.StructMember[] members = "
		<< "new org.omg.CORBA.StructMember["
		<< tc -> member_count() << "];";

	    for(ULong i = 0 ; i < tc -> member_count() ; i++)
	    {
		TypeCode_var memberType = tc -> member_type(i);
		Identifier_var memberName = tc -> member_name(i);
		memberName = fixName(memberName);

		out.sep();
		out << "\nmembers[" << i << "] = "
		    << "new org.omg.CORBA.StructMember();";
		out << "\nmembers[" << i << "].name = \"" << memberName
		    << "\";";
		out << "\nmembers[" << i << "].type = ";
		writeTypeCode("orb", package, memberType, out);
		out << ';';
	    }

	    out.sep();
	    if(tc -> kind() == tk_struct)
		out << "\ntypeCode_ = orb.create_struct_tc(id(), \"";
	    else
		out << "\ntypeCode_ = orb.create_exception_tc(id(), \"";

	    out << name << "\", members);";

	    break;
	}

	case tk_union:
	{
	    out << "\norg.omg.CORBA.UnionMember[] members = "
		<< "new org.omg.CORBA.UnionMember["
		<< tc -> member_count() << "];";

	    for(ULong i = 0 ; i < tc -> member_count() ; i++)
	    {
		TypeCode_var memberType = tc -> member_type(i);
		Identifier_var memberName = tc -> member_name(i);
		memberName = fixName(memberName);
		Any_var memberLabel = tc -> member_label(i); 

		out.sep();
		out << "\nmembers[" << i
		    << "] = new org.omg.CORBA.UnionMember();";
		out << "\nmembers[" << i << "].name = \"" << memberName
		    << "\";";
		out << "\nmembers[" << i << "].type = ";
		writeTypeCode("orb", package, memberType, out);
		out << ';';
		out << "\nmembers[" << i << "].label = orb.create_any();";

		TypeCode_var origMemberLabelType = memberLabel -> type();
		origMemberLabelType = GetOrigType(origMemberLabelType);
		switch(origMemberLabelType -> kind())
		{
		case tk_short:
		    out << "\nmembers[" << i << "].label."
			<< "insert_short(";
		    writeConstant(memberLabel, out);
		    out << ");";
		    break;

		case tk_ushort:
		    out << "\nmembers[" << i << "].label."
			<< "insert_ushort(";
		    writeConstant(memberLabel, out);
		    out << ");";
		    break;

		case tk_long:
		    out << "\nmembers[" << i << "].label."
			<< "insert_long(";
		    writeConstant(memberLabel, out);
		    out << ");";
		    break;

		case tk_ulong:
		    out << "\nmembers[" << i << "].label."
			<< "insert_ulong(";
		    writeConstant(memberLabel, out);
		    out << ");";
		    break;

		case tk_char:
		    out << "\nmembers[" << i << "].label."
			<< "insert_char(";
		    writeConstant(memberLabel, out);
		    out << ");";
		    break;

		case tk_wchar:
		    out << "\nmembers[" << i << "].label."
			<< "insert_wchar(";
		    writeConstant(memberLabel, out);
		    out << ");";
		    break;

		case tk_boolean:
		    out << "\nmembers[" << i << "].label."
			<< "insert_boolean(";
		    writeConstant(memberLabel, out);
		    out << ");";
		    break;

		case tk_enum:
		{
		    String_var memberLabelNorm =
			getTypeString(package, origMemberLabelType);
		    String_var memberLabelHelper =
			getTypeString(package, origMemberLabelType,
				      GetTypeHelper);
		    ULong val = *(ULong*)(memberLabel -> value());
		    out << '\n' << memberLabelHelper << ".insert(members[" << i
			<< "].label, " << memberLabelNorm << ".from_int("
			<< val << "));";
		    break;
		}

		case tk_octet:
		{
#ifndef NDEBUG
		    Octet octet;
		    assert(memberLabel >>= Any::to_octet(octet));
		    assert(octet == 0);
#endif
		    out << "\nmembers[" << i
			<< "].label.insert_octet((byte)0);";
		    break;
		}

		default:
		    assert(false);
		    break;
		}
	    }

	    out.sep();
	    TypeCode_var discType = tc -> discriminator_type();
	    out << "\norg.omg.CORBA.TypeCode discType = ";
	    writeTypeCode("orb", package, discType, out);
	    out << ';';

	    out << "\ntypeCode_ = orb.create_union_tc(id(), \"" << name
		<< "\", discType, members);";

	    break;
	}

	case tk_enum:
	{
	    out << "\nString[] members = new String["
		<< tc -> member_count() << "];";

	    for(ULong i = 0 ; i < tc -> member_count() ; i++)
	    {
		Identifier_var memberName = tc -> member_name(i);
		memberName = fixName(memberName);
		out << "\nmembers[" << i << "] = \"" << memberName << "\";";
	    }

	    out << "\ntypeCode_ = orb.create_enum_tc(id(), \"" << name
		<< "\", members);";

	    break;
	}

	case tk_alias:
	{
	    TypeCode_var contentType = tc -> content_type();
	    out << "\ntypeCode_ = orb.create_alias_tc(id(), \"" << name
		<< "\", ";
	    writeTypeCode("orb", package, contentType, out);
	    out << ");";
	    break;
	}
	}
	out.end();

	out.sep();
	out << "\nreturn typeCode_;";
    }
    out.end();

    //
    // Return the repository ID
    //
    out.sep();
    out << "\npublic static String\nid()";
    out.start();
    out << "\nreturn \"" << id << "\";";
    out.end();
    
    //
    // Read from input stream
    //
    out.sep();
    out << "\npublic static " << norm
	<< "\nread(org.omg.CORBA.portable.InputStream in)";
    out.start();
    if(local_)
    {
	out << "\nthrow new org.omg.CORBA.MARSHAL();";
    }
    else
    {
	switch(tc -> kind())
	{
	case tk_null:
	case tk_void:
	    assert(false);
	    break;

	case tk_short:
	case tk_long:
	case tk_ushort:
	case tk_ulong:
	case tk_longlong:
	case tk_ulonglong:
	case tk_float:
	case tk_double:
	case tk_fixed:
	case tk_boolean:
	case tk_char:
	case tk_wchar:
	case tk_octet:
	case tk_string:
	case tk_wstring:
	case tk_any:
	case tk_TypeCode:
	case tk_Principal:
	case tk_array:
	case tk_sequence:
	{
	    out << '\n' << norm << " val;";
	    ULong iter = 0;
	    writeReaderWriter(package, tc, "val", iter, true, out);
	    out << "\nreturn val;";
	    break;
	}

	case tk_objref:
	{
	    out << "\norg.omg.CORBA.Object obj = in.read_Object();";
	    out << "\nif(obj == null)";
	    out.inc();
	    out << "\nreturn null;";
	    out.dec();
	    out.sep();
	    out << "\ntry";
	    out.start();
	    out << "\nreturn (" << norm << ")obj;";
	    out.end();
	    out << "\ncatch(ClassCastException ex)";
	    out.start();
	    out.end();
	    out.sep();
	    out << "\norg.omg.CORBA.portable.ObjectImpl impl;";
	    out << "\nimpl = (org.omg.CORBA.portable.ObjectImpl)obj;";
	    out << "\nStubFor" << norm << " stub = new StubFor" << norm
		<< "();";
	    out << "\nstub._set_delegate(impl._get_delegate());";
	    out << "\nreturn stub;";
	    break;
	}

	case tk_struct:
	case tk_except:
	{
	    if(tc -> kind() == tk_except)
	    {
		out << "\nif(!id().equals(in.read_string()))";
		out.inc();
		out << "\nthrow new org.omg.CORBA.MARSHAL();";
		out.dec();
		out.sep();
	    }

	    out << '\n' << norm << " val = new " << name << "();";
	    ULong iter = 0;
	    for(ULong i = 0 ; i < tc -> member_count() ; i++)
	    {
		TypeCode_var memberType = tc -> member_type(i);
		Identifier_var memberName = tc -> member_name(i);
		memberName = fixName(memberName);
		String_var fullName = string_dup("val.");
		fullName += memberName;
		writeReaderWriter(package, memberType, fullName, iter, true,
				  out);
	    }
	    out << "\nreturn val;";
	    break;
	}

	case tk_union:
	{
	    ULong i, j;

	    TypeCode_var discType = tc -> discriminator_type();
	    TypeCode_var origDiscType = GetOrigType(discType);
	    String_var disc = getTypeString(package, discType);
	    String_var intDisc;
	    if(origDiscType -> kind() == tk_enum ||
	       origDiscType -> kind() == tk_boolean)
		intDisc = string_dup("int");
	    else
		intDisc = disc;

	    IdlUnionMemberInfoSeq_var infoSeq = IdlGetUnionInfo(tc);
	    for(i = 0 ; i < infoSeq -> length() ; i++)
		infoSeq[i].name = fixName(infoSeq[i].name);

	    String_var defStr = string_dup("_default");
	    for(j = 0 ; j < infoSeq -> length() ; j++)
		if(strcmp(infoSeq[j].name, defStr) == 0)
		    defStr = string_dup("__default");

	    ULong iter = 0;

	    out << '\n' << norm << " val = new " << name << "();";
	    out << '\n' << disc << " d;";
	    writeReaderWriter(package, discType, "d", iter, true, out);

	    out.sep();
	    if(origDiscType -> kind() == tk_enum)
		out << "\nswitch(d.value())";
	    else if(origDiscType -> kind() == tk_boolean)
		out << "\nswitch(d ? 1 : 0)";
	    else
		out << "\nswitch(d)";
	    out.start();
	    for(i = 0 ; i < infoSeq -> length() ; i++)
	    {
		IdlUnionMemberInfo& info = infoSeq[i];
		IdlStringSeq_var labels =
		    getUnionLabelSeq(package, origDiscType, info);

		if(i > 0)
		    out.sep();

		out.dec();

		if(is_nil(info.type))
		{
		    assert(info.isDefault);
		    out << "\ndefault:";
		    out.inc();
		    if(info.nLabels.length() + info.pLabels.length() > 1 ||
		       info.isDefault)
			out << "\nval." << defStr << "(d);";
		    else
			out << "\nval." << defStr << "();";
		    out << "\nbreak;";
		    out.dec();
		}
		else
		{
		    if(info.isDefault)
			out << "\ndefault:";
		    else
		    {
			for(j = 0 ; j < labels -> length() ; j++)
			    out << "\ncase " << labels[j] << ':';
		    }

		    out.start();
		    String_var norm = getTypeString(package, info.type);
		    out << '\n' << norm << " v;";
		    writeReaderWriter(package, info.type, "v", iter, true,
				      out);
		    if(info.nLabels.length() + info.pLabels.length() > 1 ||
		       info.isDefault)
			out << "\nval." << info.name << "(d, v);";
		    else
			out << "\nval." << info.name << "(v);";
		    out << "\nbreak;";
		    out.end();
		}

		out.inc();

	    }
	    out.end();

	    out << "\nreturn val;";

	    break;
	}

	case tk_enum:
	    out << '\n' << norm << " val;";
	    out << "\nval = " << name << ".from_int(in.read_ulong());";
	    out << "\nreturn val;";
	    break;

	case tk_alias:
	{
	    TypeCode_var contentType = tc -> content_type();
	    out << '\n' << norm << " val;";
	    ULong iter = 0;
	    writeReaderWriter(package, contentType, "val", iter, true, out);
	    out << "\nreturn val;";
	    break;
	}
	}
    }
    out.end();

    //
    // Write to output stream
    //
    out.sep();
    out << "\npublic static void\nwrite("
	<< "org.omg.CORBA.portable.OutputStream out, "
	<< norm << " val)";
    out.start();
    if(local_)
    {
	out << "\nthrow new org.omg.CORBA.MARSHAL();";
    }
    else
    {
	switch(tc -> kind())
	{
	case tk_null:
	case tk_void:
	    assert(false);
	    break;

	case tk_short:
	case tk_long:
	case tk_ushort:
	case tk_ulong:
	case tk_longlong:
	case tk_ulonglong:
	case tk_float:
	case tk_double:
	case tk_fixed:
	case tk_boolean:
	case tk_char:
	case tk_wchar:
	case tk_octet:
	case tk_string:
	case tk_wstring:
	case tk_any:
	case tk_TypeCode:
	case tk_Principal:
	case tk_array:
	case tk_sequence:
	{
	    ULong iter = 0;
	    writeReaderWriter(package, tc, "val", iter, false, out);
	    break;
	}

	case tk_objref:
	    out << "\nout.write_Object(val);";
	    break;

	case tk_struct:
	case tk_except:
	{
	    if(tc -> kind() == tk_except)
		out << "\nout.write_string(id());";

	    ULong iter = 0;
	    for(ULong i = 0 ; i < tc -> member_count() ; i++)
	    {
		TypeCode_var memberType = tc -> member_type(i);
		Identifier_var memberName = tc -> member_name(i);
		memberName = fixName(memberName);
		String_var fullName = string_dup("val.");
		fullName += memberName;
		writeReaderWriter(package, memberType, fullName,
				  iter, false, out);
	    }
	    break;
	}

	case tk_union:
	{
	    ULong i, j;

	    TypeCode_var discType = tc -> discriminator_type();
	    TypeCode_var origDiscType = GetOrigType(discType);
	    String_var disc = getTypeString(package, discType);
	    String_var intDisc;
	    if(origDiscType -> kind() == tk_enum ||
	       origDiscType -> kind() == tk_boolean)
		intDisc = string_dup("int");
	    else
		intDisc = disc;

	    IdlUnionMemberInfoSeq_var infoSeq = IdlGetUnionInfo(tc);
	    for(i = 0 ; i < infoSeq -> length() ; i++)
		infoSeq[i].name = fixName(infoSeq[i].name);

	    String_var discStr = string_dup("discriminator");
	    for(i = 0 ; i < infoSeq -> length() ; i++)
		if(strcmp(infoSeq[i].name, discStr) == 0)
		    discStr = string_dup("_discriminator");

	    ULong iter = 0;

	    out << '\n' << disc << " d = val." << discStr << "();";
	    writeReaderWriter(package, discType, "d", iter, false, out);

	    out.sep();
	    if(origDiscType -> kind() == tk_enum)
		out << "\nswitch(d.value())";
	    else if(origDiscType -> kind() == tk_boolean)
		out << "\nswitch(d ? 1 : 0)";
	    else
		out << "\nswitch(d)";
	    out.start();
	    for(i = 0 ; i < infoSeq -> length() ; i++)
	    {
		IdlUnionMemberInfo& info = infoSeq[i];
		IdlStringSeq_var labels =
		    getUnionLabelSeq(package, origDiscType, info);

		if(i > 0)
		    out.sep();

		out.dec();

		if(is_nil(info.type))
		{
		    assert(info.isDefault);
		    out << "\ndefault:";
		    out.inc();
		    out << "\nbreak;";
		    out.dec();
		}
		else
		{
		    if(info.isDefault)
			out << "\ndefault:";
		    else
		    {
			for(j = 0 ; j < labels -> length() ; j++)
			    out << "\ncase " << labels[j] << ':';
		    }

		    out.start();
		    String_var norm = getTypeString(package, info.type);
		    out << '\n' << norm << " v = val." << info.name << "();";
		    writeReaderWriter(package, info.type, "v", iter, false,
				      out);
		    out << "\nbreak;";
		    out.end();
		}

		out.inc();

	    }
	    out.end();

	    break;
	}

	case tk_enum:
	    out << "\nout.write_ulong(val.value());";
	    break;

	case tk_alias:
	{
	    TypeCode_var contentType = tc -> content_type();
	    ULong iter = 0;
	    writeReaderWriter(package, contentType, "val", iter, false, out);
	    break;
	}
	}
    }
    out.end();

    //
    // Read object
    //
    out.sep();
    out << "\npublic org.omg.CORBA.Object\nread_Object("
	<< "org.omg.CORBA.portable.InputStream in)";
    out.start();
    out << "\nreturn in.read_Object();";
    out.end();

    //
    // Write object
    //
    out.sep();
    out << "\npublic void\nwrite_Object("
	<< "org.omg.CORBA.portable.OutputStream out, "
	<< "org.omg.CORBA.Object obj)";
    out.start();
    out << "\nout.write_Object(obj);";
    out.end();

    //
    // Get ID
    //
    out.sep();
    out << "\npublic String\nget_id()";
    out.start();
    out << "\nreturn id();";
    out.end();

    //
    // Get type
    //
    out.sep();
    out << "\npublic org.omg.CORBA.TypeCode\nget_type()";
    out.start();
    out << "\nreturn type();";
    out.end();

    //
    // Narrow, only for interfaces
    //
    if(origType -> kind() == tk_objref)
    {
	out.sep();
	out << "\npublic static " << norm
	    << "\nnarrow(org.omg.CORBA.Object obj)";
	out.start();
	
	if(strcmp(origType -> id(), "IDL:omg.org/CORBA/Object:1.0") == 0)
	{
	    out << "\nreturn obj;";
	}
	else
	{    
	    out << "\nif(obj != null)";
	    out.start();
	    
	    out << "\ntry";
	    out.start();
	    out << "\nreturn (" << norm << ")obj;";
	    out.end();
	    out << "\ncatch(ClassCastException ex)";
	    out.start();
	    out.end();

	    if(!local_)
	    {
		String_var stub =
		    getTypeString(package, tc, GetTypeStub);

		out.sep();
		out << "\nif(obj._is_a(id()))";
		out.start();
		out << "\norg.omg.CORBA.portable.ObjectImpl impl;";
		out << '\n' << stub << " stub;";
		out << "\nimpl = (org.omg.CORBA.portable.ObjectImpl)obj;";

		if(!noStreams_)
		{
		    String_var localStub =
			getTypeString(package, tc, GetTypeLocalStub);

		    out << "\nif(impl._get_delegate().is_local(obj))";
		    out.inc();
		    out << "\nstub = new " << localStub << "();";
		    out.dec();
		    out << "\nelse";
		    out.inc();
		    out << "\nstub = new " << stub << "();";
		    out.dec();
		}
		else
		    out << "\nstub = new " << stub << "();";

		out << "\nstub._set_delegate(impl._get_delegate());";
		out << "\nreturn stub;";
		out.end();
	    }

	    out.end();
	    out.sep();
	    out << "\nreturn null;";
	}
	    
	out.end();
    }

    //
    // Operations for value types
    //
    if(origType -> kind() == tk_value ||
       origType -> kind() == tk_value_box)
    {
	out.sep();
	out << "private static ValueHelper instance_;";

	out.sep();
	out << "\npublic static ValueHelper"
	    << "\nget_instance()";
	out.start();
	out << "\nif(instance_ == null)";
	out.inc();
	out << "\ninstance_ = new " << name << "Helper();";
	out.dec();
	out << "\nreturn instance_;";
	out.end();

	out.sep();
	out << "\npublic java.io.Serializable"
	    << "\nread_Value(org.omg.CORBA.portable.InputStream in)";
	out.start();
	out << "\nreturn in.read_Value();";
	out.end();

	out.sep();
	out << "\npublic void"
	    << "\nwrite_Value(org.omg.CORBA.portable.OutputStream out, "
	    << "java.io.Serializable obj)";
	out.start();
	out << "\nout.write_Value(obj);";
	out.end();

	out.sep();
	out << "\npublic Class"
	    << "\nget_class()";
	out.start();
	out << "\nreturn forName(\"" << name << "\");";
	out.end();

	out.sep();
	out << "\npublic String[]"
	    << "\nget_truncatable_base_ids()";
	out.start();
	out << "\nthrow new RuntimeException();";
	out.end();
    }

    out.end();
}

// ----------------------------------------------------------------------
// Write TypeCode
// ----------------------------------------------------------------------

void
IdlJavaGenerator::writeTypeCode(const char* orb, const char* package,
				TypeCode_ptr tc, IdlPrettyPrint& out)
{
    switch(tc -> kind())
    {
    case tk_null:
    case tk_void:
	assert(false);
	break;

    case tk_short:
	out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_short)";
	break;

    case tk_long:
	out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_long)";
	break;

    case tk_ushort:
	out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_ushort)";
	break;

    case tk_ulong:
	out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_ulong)";
	break;

    case tk_longlong:
	out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_longlong)";
	break;

    case tk_ulonglong:
	out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_ulonglong)";
	break;

    case tk_float:
	out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_float)";
	break;
	
    case tk_double:
	out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_double)";
	break;

    case tk_fixed:
	out << orb << ".create_fixed_tc((short)" << tc -> fixed_digits()
	    << ", (short)"  << tc -> fixed_scale() << ')';
	break;

    case tk_boolean:
	out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_boolean)";
	break;

    case tk_char:
	out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_char)";
	break;

    case tk_wchar:
	out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_wchar)";
	break;

    case tk_octet:
	out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_octet)";
	break;

    case tk_string:
    {
	if(tc -> length() == 0)
	    out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_string)";
	else
	    out << orb << ".create_string_tc(" << tc -> length() << ")";
	break;
    }

    case tk_wstring:
    {
	if(tc -> length() == 0)
	    out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_wstring)";
	else
	    out << orb << ".create_wstring_tc(" << tc -> length() << ")";
	break;
    }

    case tk_any:
	out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_any)";
	break;

    case tk_TypeCode:
	out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_TypeCode)";
	break;

    case tk_Principal:
	out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_Principal)";
	break;

    case tk_objref:
    {
	RepositoryId_var id = tc -> id();
	if(strcmp(id, "IDL:omg.org/CORBA/Object:1.0") == 0)
	{
	    Identifier_var name = tc -> name();
	    out << orb << ".create_interface_tc(\"" << id << "\", \"" << name
		<< "\")";
	}
	else
	{
	    String_var helper =
		getTypeString(package, tc, GetTypeHelper);
	    out << helper << ".type()";
	}
	break;
    }

    case tk_struct:
    case tk_except:
    case tk_union:
    case tk_enum:
    case tk_alias:
    {
	String_var helper = getTypeString(package, tc, GetTypeHelper);
	out << helper << ".type()";
	break;
    }

    case tk_sequence:
    {
/*
	if(tc -> offset() == 0)
	{
*/
	    TypeCode_var contentType = tc -> content_type();

	    out << orb << ".create_sequence_tc(" << tc -> length() << ", ";
	    writeTypeCode(orb, package, contentType, out);
	    out << ')';
/*
	}
	else
	{
	    out << orb << ".create_recursive_sequence_tc("
		<< tc -> length() << ", " << tc -> offset() << ')';
	}
*/

	break;
    }

    case tk_array:
    {
	TypeCode_var contentType = tc -> content_type();

	out << orb << ".create_array_tc(" << tc -> length() << ", ";
	writeTypeCode(orb, package, contentType, out);
	out << ')';

	break;
    }
    }
}

// ----------------------------------------------------------------------
// Generate code for reading/writing from streams
// ----------------------------------------------------------------------

void
IdlJavaGenerator::writeReaderWriter(const char* package,
				    TypeCode_ptr tc,
				    const char* name, ULong& iter,
				    bool r, IdlPrettyPrint& out)
{
    switch(tc -> kind())
    {
    case tk_null:
    case tk_void:
	assert(false);
	break;

    case tk_short:
	if(r)
	    out << '\n' << name << " = in.read_short();";
	else
	    out << "\nout.write_short(" << name << ");";
	break;

    case tk_long:
	if(r)
	    out << '\n' << name << " = in.read_long();";
	else
	    out << "\nout.write_long(" << name << ");";
	break;

    case tk_ushort:
	if(r)
	    out << '\n' << name << " = in.read_ushort();";
	else
	    out << "\nout.write_ushort(" << name << ");";
	break;

    case tk_ulong:
	if(r)
	    out << '\n' << name << " = in.read_ulong();";
	else
	    out << "\nout.write_ulong(" << name << ");";
	break;

    case tk_longlong:
	if(r)
	    out << '\n' << name << " = in.read_longlong();";
	else
	    out << "\nout.write_longlong(" << name << ");";
	break;

    case tk_ulonglong:
	if(r)
	    out << '\n' << name << " = in.read_ulonglong();";
	else
	    out << "\nout.write_ulonglong(" << name << ");";
	break;

     case tk_float:
	if(r)
	    out << '\n' << name << " = in.read_float();";
	else
	    out << "\nout.write_float(" << name << ");";
	break;

    case tk_double:
	if(r)
	    out << '\n' << name << " = in.read_double();";
	else
	    out << "\nout.write_double(" << name << ");";
	break;

    case tk_fixed:
	if(r)
	    out << '\n' << name << " = in.read_fixed();";
	else
	    out << "\nout.write_fixed(" << name << ");";
	break;

    case tk_boolean:
	if(r)
	    out << '\n' << name << " = in.read_boolean();";
	else
	    out << "\nout.write_boolean(" << name << ");";
	break;

    case tk_char:
	if(r)
	    out << '\n' << name << " = in.read_char();";
	else
	    out << "\nout.write_char(" << name << ");";
	break;

    case tk_wchar:
	if(r)
	    out << '\n' << name << " = in.read_wchar();";
	else
	    out << "\nout.write_wchar(" << name << ");";
	break;

    case tk_octet:
	if(r)
	    out << '\n' << name << " = in.read_octet();";
	else
	    out << "\nout.write_octet(" << name << ");";
	break;

    case tk_string:
	if(r)
	    out << '\n' << name << " = in.read_string();";
	else
	    out << "\nout.write_string(" << name << ");";
	break;

    case tk_wstring:
	if(r)
	    out << '\n' << name << " = in.read_wstring();";
	else
	    out << "\nout.write_wstring(" << name << ");";
	break;

    case tk_any:
	if(r)
	    out << '\n' << name << " = in.read_any();";
	else
	    out << "\nout.write_any(" << name << ");";
	break;

    case tk_TypeCode:
	if(r)
	    out << '\n' << name << " = in.read_TypeCode();";
	else
	    out << "\nout.write_TypeCode(" << name << ");";
	break;

    case tk_Principal:
	if(r)
	    out << '\n' << name << " = in.read_Principal();";
	else
	    out << "\nout.write_Principal(" << name << ");";
	break;

    case tk_objref:
    {
	RepositoryId_var id = tc -> id();
	if(strcmp(id, "IDL:omg.org/CORBA/Object:1.0") == 0)
	{
	    if(r)
		out << '\n' << name << " = in.read_Object();";
	    else
		out << "\nout.write_Object(" << name << ");";
	}
	else
	{
	    String_var helper =
		getTypeString(package, tc, GetTypeHelper);
	    if(r)
		out << '\n' << name << " = " << helper << ".read(in);";
	    else
		out << '\n' << helper << ".write(out, " << name << ");";
	}
	break;
    }
    
    case tk_struct:
    case tk_except:
    case tk_union:
    case tk_enum:
    case tk_alias:
    {
	String_var helper = getTypeString(package, tc, GetTypeHelper);
	if(r)
	    out << '\n' << name << " = " << helper << ".read(in);";
	else
	    out << '\n' << helper << ".write(out, " << name << ");";
	break;
    }

    case tk_value:
    case tk_value_box:
	if(r)
	    out << '\n' << name << " = in.read_Value();";
	else
	    out << "\nout.write_Value(" << name << ");";
	break;


    case tk_sequence:
    case tk_array:
    {
	if(r)
	{
	    if(tc -> kind() == tk_sequence)
		out << "\nint len" << iter << " = in.read_ulong();";
	    else
		out << "\nint len" << iter << " = " << tc -> length() << ';';

	    TypeCode_var t = TypeCode::_duplicate(tc);
	    String_var s = string_dup("");
	    while(t -> kind() == tk_sequence ||
		  t -> kind() == tk_array)
	    {
		if(strlen(s) > 0)
		    s += "[]";
		else
		{
		    s += "[len";
		    s += iter;
		    s += ']';
		}
		
		t = t -> content_type();
		t = GetOrigType(t);
	    }
	    String_var n = getTypeString(package, t);
	    out << '\n' << name << " = new " << n << s << ';';
	}
	else
	{
	    out << "\nint len" << iter << " = " << name << ".length;";

	    //
	    // Check array bounds
	    //
	    if(tc -> kind() == tk_array)
	    {
		out << "\nif(len" << iter << " != " << tc -> length() << ')';
		out.inc();
		out << "\n throw new org.omg.CORBA.MARSHAL();";
		out.dec();
	    }

	    if(tc -> kind() == tk_sequence)
	    {
		//
		// Check bounds for boundes sequences
		//
		if(tc -> length())
		{
		    out << "\nif(len" << iter << " > "
			<< tc -> length() << ')';
		    out.inc();
		    out << "\n throw new org.omg.CORBA.MARSHAL();";
		    out.dec();
		}
		
		out << "\nout.write_ulong(len" << iter << ");";
	    }
	}

	TypeCode_var contentType = tc -> content_type();
	switch(contentType -> kind())
	{
	case tk_null:
	case tk_void:
	    assert(false);
	    break;

	case tk_short:
	    if(r)
		out << "\nin.read_short_array(" << name << ", 0, len"
		    << iter << ");";
	    else
		out << "\nout.write_short_array(" << name << ", 0, len"
		    << iter << ");";
	    iter++;
	    break;

	case tk_long:
	    if(r)
	    out << "\nin.read_long_array(" << name << ", 0, len"
		<< iter << ");";
	    else
		out << "\nout.write_long_array(" << name << ", 0, len"
		    << iter << ");";
	    iter++;
	    break;

	case tk_longlong:
	    if(r)
	    out << "\nin.read_longlong_array(" << name << ", 0, len"
		<< iter << ");";
	    else
		out << "\nout.write_longlong_array(" << name << ", 0, len"
		    << iter << ");";
	    iter++;
	    break;

	case tk_ushort:
	    if(r)
		out << "\nin.read_ushort_array(" << name << ", 0, len"
		    << iter << ");";
	    else
		out << "\nout.write_ushort_array(" << name << ", 0, len"
		    << iter << ");";
	    iter++;
	    break;

	case tk_ulong:
	    if(r)
		out << "\nin.read_ulong_array(" << name << ", 0, len"
		    << iter << ");";
	    else
		out << "\nout.write_ulong_array(" << name << ", 0, len"
		    << iter << ");";
	    iter++;
	    break;

	case tk_ulonglong:
	    if(r)
		out << "\nin.read_ulonglong_array(" << name << ", 0, len"
		    << iter << ");";
	    else
		out << "\nout.write_ulonglong_array(" << name << ", 0, len"
		    << iter << ");";
	    iter++;
	    break;

	case tk_float:
	    if(r)
		out << "\nin.read_float_array(" << name << ", 0, len"
		    << iter << ");";
	    else
		out << "\nout.write_float_array(" << name << ", 0, len"
		    << iter << ");";
	    iter++;
	    break;

	case tk_double:
	    if(r)
		out << "\nin.read_double_array(" << name << ", 0, len"
		    << iter << ");";
	    else
		out << "\nout.write_double_array(" << name << ", 0, len"
		    << iter << ");";
	    iter++;
	    break;

	case tk_boolean:
	    if(r)
		out << "\nin.read_boolean_array(" << name << ", 0, len"
		    << iter << ");";
	    else
		out << "\nout.write_boolean_array(" << name << ", 0, len"
		    << iter << ");";
	    iter++;
	    break;

	case tk_char:
	    if(r)
		out << "\nin.read_char_array(" << name << ", 0, len"
		    << iter << ");";
	    else
		out << "\nout.write_char_array(" << name << ", 0, len"
		    << iter << ");";
	    iter++;
	    break;

	case tk_wchar:
	    if(r)
		out << "\nin.read_wchar_array(" << name << ", 0, len"
		    << iter << ");";
	    else
		out << "\nout.write_wchar_array(" << name << ", 0, len"
		    << iter << ");";
	    iter++;
	    break;

	case tk_octet:
	    if(r)
		out << "\nin.read_octet_array(" << name << ", 0, len"
		    << iter << ");";
	    else
		out << "\nout.write_octet_array(" << name << ", 0, len"
		    << iter << ");";
	    iter++;
	    break;

	case tk_string:
	{
	    out << "\nfor(int i" << iter << " = 0 ; i" << iter << " < len"
		<< iter << " ; i" << iter << "++)";
	    out.inc();
	    if(r)
		out << '\n' << name << "[i" << iter
		    << "] = in.read_string();";
	    else
		out << "\nout.write_string(" << name << "[i" << iter << "]);";
	    out.dec();
	    iter++;
	    break;
	}

	case tk_wstring:
	{
	    out << "\nfor(int i" << iter << " = 0 ; i" << iter << " < len"
		<< iter << " ; i" << iter << "++)";
	    out.inc();
	    if(r)
		out << '\n' << name << "[i" << iter
		    << "] = in.read_wstring();";
	    else
		out << "\nout.write_wstring(" << name << "[i" << iter << "]);";
	    out.dec();
	    iter++;
	    break;
	}

	case tk_any:
	{
	    out << "\nfor(int i" << iter << " = 0 ; i" << iter << " < len"
		<< iter << " ; i" << iter << "++)";
	    out.inc();
	    if(r)
		out << '\n' << name << "[i" << iter << "] = in.read_any();";
	    else
		out << "\nout.write_any(" << name << "[i" << iter << "]);";
	    out.dec();
	    iter++;
	    break;
	}

	case tk_TypeCode:
	{
	    out << "\nfor(int i" << iter << " = 0 ; i" << iter << " < len"
		<< iter << " ; i" << iter << "++)";
	    out.inc();
	    if(r)
		out << '\n' << name << "[i" << iter
		    << "] = in.read_TypeCode();";
	    else
		out << "\nout.write_TypeCode(" << name
		    << "[i" << iter << "]);";
	    out.dec();
	    iter++;
	    break;
	}

	case tk_Principal:
	{
	    out << "\nfor(int i" << iter << " = 0 ; i" << iter << " < len"
		<< iter << " ; i" << iter << "++)";
	    out.inc();
	    if(r)
		out << '\n' << name << "[i" << iter
		    << "] = in.read_Principal();";
	    else
		out << "\nout.write_Principal(" << name
		    << "[i" << iter << "]);";
	    out.dec();
	    iter++;
	    break;
	}

	case tk_objref:
	{
	    RepositoryId_var id = contentType -> id();
	    if(strcmp(id, "IDL:omg.org/CORBA/Object:1.0") == 0)	
	    {
		out << "\nfor(int i" << iter << " = 0 ; i" << iter << " < len"
		    << iter << " ; i" << iter << "++)";
		out.inc();
		if(r)
		    out << '\n' << name << "[i" << iter
			<< "] = in.read_Object();";
		else
		    out << "\nout.write_Object(" << name
			<< "[i" << iter << "]);";
		out.dec();
		iter++;
	    }
	    else
	    {
		out << "\nfor(int i" << iter << " = 0 ; i" << iter << " < len"
		    << iter << " ; i" << iter << "++)";
		String_var helper = getTypeString(package, contentType,
							GetTypeHelper);
		out.inc();
		if(r)
		    out << '\n' << name << "[i" << iter << "] = " << helper
			<< ".read(in);";
		else
		    out << '\n' << helper << ".write(out, " << name
			<< "[i" << iter << "]);";
		out.dec();
		iter++;
	    }
	    break;
	}

	case tk_struct:
	case tk_except:
	case tk_union:
	case tk_enum:
	case tk_fixed:
	case tk_alias:
	{
	    out << "\nfor(int i" << iter << " = 0 ; i" << iter << " < len"
		<< iter << " ; i" << iter << "++)";
	    String_var helper = getTypeString(package, contentType,
						    GetTypeHelper);
	    out.inc();
	    if(r)
		out << '\n' << name << "[i" << iter << "] = " << helper
		    << ".read(in);";
	    else
		out << '\n' << helper << ".write(out, " << name
		    << "[i" << iter << "]);";
	    out.dec();
	    iter++;
	    break;
	}

	case tk_value:
	case tk_value_box:
	{
	    out << "\nfor(int i" << iter << " = 0 ; i" << iter << " < len"
		<< iter << " ; i" << iter << "++)";
	    out.inc();
	    if(r)
		out << '\n' << name << "[i" << iter
		    << "] = in.read_Value();";
	    else
		out << "\nout.write_Value(" << name
		    << "[i" << iter << "]);";
	    out.dec();
	    iter++;
	    break;
	}

	case tk_array:
	case tk_sequence:
	{
	    out << "\nfor(int i" << iter << " = 0 ; i" << iter << " < len"
		<< iter << " ; i" << iter << "++)";
	    out.start();
	    String_var str = name;
	    str += "[i";
	    str += iter;
	    str += ']';
	    iter++;
	    writeReaderWriter(package, contentType, str, iter, r, out);
	    out.end();
	    
	    break;
	}
	}

	break;
    }
    }
}

// ----------------------------------------------------------------------
// Write any insertion/extraction for stubs/skeletons
// ----------------------------------------------------------------------

void
IdlJavaGenerator::writeInsertExtract(const char* package,
				     TypeCode_ptr tc,
				     const char* any, bool ex,
				     IdlPrettyPrint& out)
{
    TypeCode_var origType = GetOrigType(tc);

    switch(origType -> kind())
    {
    case tk_null:
    case tk_void:
	assert(false);
	break;

    case tk_short:
	if(ex)
	    out << any << ".extract_short()";
	else
	    out << any << ".insert_short(";
	break;

    case tk_long:
	if(ex)
	    out << any << ".extract_long()";
	else
	    out << any << ".insert_long(";
	break;

    case tk_ushort:
	if(ex)
	    out << any << ".extract_ushort()";
	else
	    out << any << ".insert_ushort(";
	break;

    case tk_ulong:
	if(ex)
	    out << any << ".extract_ulong()";
	else
	    out << any << ".insert_ulong(";
	break;

    case tk_longlong:
	if(ex)
	    out << any << ".extract_longlong()";
	else
	    out << any << ".insert_longlong(";
	break;

    case tk_ulonglong:
	if(ex)
	    out << any << ".extract_ulonglong()";
	else
	    out << any << ".insert_ulonglong(";
	break;

    case tk_float:
	if(ex)
	    out << any << ".extract_float()";
	else
	    out << any << ".insert_float(";
	break;

    case tk_double:
	if(ex)
	    out << any << ".extract_double()";
	else
	    out << any << ".insert_double(";
	break;

    case tk_fixed:
	if(ex)
	    out << any << ".extract_fixed()";
	else
	    out << any << ".insert_fixed(";
	break;

    case tk_boolean:
	if(ex)
	    out << any << ".extract_boolean()";
	else
	    out << any << ".insert_boolean(";
	break;

    case tk_char:
	if(ex)
	    out << any << ".extract_char()";
	else
	    out << any << ".insert_char(";
	break;

    case tk_wchar:
	if(ex)
	    out << any << ".extract_wchar()";
	else
	    out << any << ".insert_wchar(";
	break;

    case tk_octet:
	if(ex)
	    out << any << ".extract_octet()";
	else
	    out << any << ".insert_octet(";
	break;

    case tk_string:
	if(ex)
	    out << any << ".extract_string()";
	else
	    out << any << ".insert_string(";
	break;

    case tk_wstring:
	if(ex)
	    out << any << ".extract_wstring()";
	else
	    out << any << ".insert_wstring(";
	break;

    case tk_any:
	if(ex)
	    out << any << ".extract_any()";
	else
	    out << any << ".insert_any(";
	break;

    case tk_TypeCode:
	if(ex)
	    out << any << ".extract_TypeCode()";
	else
	    out << any << ".insert_TypeCode(";
	break;

    case tk_Principal:
	if(ex)
	    out << any << ".extract_Principal()";
	else
	    out << any << ".insert_Principal(";
	break;

    case tk_objref:
	if(strcmp(origType -> id(), "IDL:omg.org/CORBA/Object:1.0") == 0)
	{
	    if(ex)
		out << any << ".extract_Object()";
	    else
		out << any << ".insert_Object(";

	    break;
	}
	// No break here!

    case tk_struct:
    case tk_union:
    case tk_except:
    case tk_enum:
    case tk_sequence:
    case tk_array:
    {	
        String_var helper = getTypeString(package, tc, GetTypeHelper);

        if(ex)
            out << helper << ".extract(" << any << ')';
        else
            out << helper << ".insert(" << any << ", ";

        break;
    }

    case tk_alias:
	assert(false);
	break;
    }
}

// ----------------------------------------------------------------------
// Write clone() method
// ----------------------------------------------------------------------

void
IdlJavaGenerator::writeClone(DefinitionKind kind, const char *type,
			     const char* package, TypeCode_ptr tc,
			     IdlPrettyPrint& out)
{
    out.sep();
    out << "\npublic java.lang.Object";
    out << "\nclone()";
    out.start();

    switch(kind)
    {
    case dk_Enum:
	out << "\nreturn new " << type << "(value_);";
	break;

    case dk_Struct:
    case dk_Exception:
    {
	out << '\n' << type << " _ob_r = new " << type << "();";
	out.sep();

	for(ULong i = 0 ; i < tc -> member_count() ; i++)
	{
	    TypeCode_var memberType = tc -> member_type(i);
	    String_var memberTypeName =
		getTypeString(package, memberType);
	    Identifier_var memberName = tc -> member_name(i);
	    memberName = fixName(memberName);
	    String_var target = string_dup("_ob_r.");
	    target += memberName;

	    //
	    // If member is not primitive, then we need to test if
	    // the member is null
	    //
	    TypeCode_var origMemberType = GetOrigType(memberType);
	    if(! isPrimitiveMemberType(origMemberType))
	    {
		out << "\nif(" << memberName << " != null)";
		out.start();
		writeMemberClone(target, memberName, package, memberType,
				 memberTypeName, out, 0);
		out.end();
	    }
	    else
		writeMemberClone(target, memberName, package, memberType,
				 memberTypeName, out, 0);
	}

	out.sep();
	out << "\nreturn _ob_r;";
	break;
    }

    case dk_Union:
    {
	ULong i, j;

	out << '\n' << type << " _ob_r = new " << type << "();";

	out.sep();

	out << "\n_ob_r._ob_i_ = _ob_i_;";
	out << "\n_ob_r._ob_d_ = _ob_d_;";

	out.sep();

	TypeCode_var discType = tc -> discriminator_type();
	TypeCode_var origDiscType = GetOrigType(discType);

	IdlUnionMemberInfoSeq_var infoSeq = IdlGetUnionInfo(tc);
	for(i = 0 ; i < infoSeq -> length() ; i++)
	    infoSeq[i].name = fixName(infoSeq[i].name);

	//
	// Only attempt to clone member if the union has been
	// initialized and the value is not null
	//
	out << "\nif(_ob_i_ && _ob_v_ != null)";
	out.start();

	out << "\nswitch(_ob_d_)";
	out.start();
	for(i = 0 ; i < infoSeq -> length() ; i++)
	{
	    IdlUnionMemberInfo& info = infoSeq[i];
	    IdlStringSeq_var labels =
		getUnionLabelSeq(package, origDiscType, info);

	    //
	    // If the typecode is nil, we don't need to do anything
	    //
	    if(is_nil(info.type))
	    {
		assert(info.isDefault);
		continue;
	    }

	    if(i > 0)
		out.sep();

	    out.dec();

	    if(info.isDefault)
		out << "\ndefault:";
	    else
	    {
		for(j = 0 ; j < labels -> length() ; j++)
		    out << "\ncase " << labels[j] << ':';
	    }

	    out.start();

	    String_var norm = getTypeString(package, info.type);
	    TypeCode_var origMemberType = GetOrigType(info.type);

	    String_var holder = getUnionMemberHolder(origMemberType);

	    if(strlen(holder) > 0)
	    {
		//out << "\nif(_ob_v_ != null)";
		//out.inc();
		out << "\n_ob_r._ob_v_ = new " << holder << "((("
		    << holder << ")_ob_v_).value);";
		//out.dec();
	    }
	    else
	    {
		switch(origMemberType -> kind())
		{
		case tk_string:
		case tk_wstring:
		case tk_TypeCode:
		case tk_Principal:
		case tk_objref:
		{
		    writeMemberClone("_ob_r._ob_v_", "_ob_v_", package,
				     origMemberType, norm, out, 0);
		    break;
		}

		case tk_any:
		{
		    out << "\norg.omg.CORBA.Any _ob_any_v = "
			<< "(org.omg.CORBA.Any)_ob_v_;";
		    out << "\norg.omg.CORBA.Any _ob_any;";
		    writeMemberClone("_ob_any", "_ob_any_v", package,
				     origMemberType, norm, out, 0);
		    out << "\n_ob_r._ob_v_ = _ob_any;";
		    break;
		}

		case tk_struct:
		case tk_except:
		case tk_union:
		case tk_enum:
		{
		    out << "\n_ob_r._ob_v_ = ((" << norm
			<< ")_ob_v_).clone();";
		    break;
		}

		case tk_sequence:
		case tk_array:
		{
		    out << '\n' << norm << " _ob_arr_v = (" << norm << ")"
			<< "_ob_v_;";
		    out << '\n' << norm << " _ob_arr;";
		    writeMemberClone("_ob_arr", "_ob_arr_v", package,
				     origMemberType, norm, out, 0);
		    out << "\n_ob_r._ob_v_ = _ob_arr;";
		    break;
		}

		default:
		    assert(false);
		    break;
		}
	    }

	    out << "\nbreak;";

	    out.end();  // case stmt
	    out.inc();
	}

	out.end();  // switch stmt
	out.end();  // if stmt

	out.sep();
	out << "\nreturn _ob_r;";
	break;
    }

    default:
	assert(false);
	break;
    }

    out.end();
}

// ----------------------------------------------------------------------
// Write clone() assignment code for struct/union/exception member
// ----------------------------------------------------------------------

void
IdlJavaGenerator::writeMemberClone(const char* target, const char *source,
				   const char* package, TypeCode_ptr tc,
				   const char* type, IdlPrettyPrint& out,
				   int iter)
{
    switch(tc -> kind())
    {
    case tk_null:
    case tk_void:
	assert(false);
	break;

    //
    // All of these can be copied directly (strings, TypeCodes, Principals
    // and Objects are all immutable)
    //
    case tk_short:
    case tk_long:
    case tk_ushort:
    case tk_ulong:
    case tk_longlong:
    case tk_ulonglong:
    case tk_float:
    case tk_double:
    case tk_fixed:
    case tk_boolean:
    case tk_char:
    case tk_wchar:
    case tk_octet:
    case tk_string:
    case tk_wstring:
    case tk_TypeCode:
    case tk_Principal:
    case tk_objref:
	out << '\n' << target << " = " << source << ";";
	break;

    //
    // Make a deep copy of the any
    //
    case tk_any:
    {
	out << '\n' << target
	    << " = org.omg.CORBA.ORB.init().create_any();";
	out << '\n' << target << ".read_value(" << source
	    << ".create_input_stream(), " << source << ".type());";
	break;
    }

    case tk_alias:
    {
	TypeCode_var contentType = tc -> content_type();
	String_var contentTypeName = getTypeString(package, contentType);
	writeMemberClone(target, source, package, contentType, contentTypeName,
			 out, iter);
	break;
    }

    case tk_struct:
    case tk_except:
    case tk_union:
    case tk_enum:
    {
	out << '\n' << target << " = (" << type << ")"
	    << source << ".clone();";
	break;
    }

    case tk_sequence:
    case tk_array:
    {
	TypeCode_var contentType = tc -> content_type();
	String_var contentTypeName = getTypeString(package, contentType);
	TypeCode_var origContentType = GetOrigType(contentType);

	//
	// If type is primitive (or immutable), then we can just allocate
	// an array and use System.arraycopy()
	//
	if(isPrimitiveMemberType(origContentType))
	{
	    String_var orig = getTypeString(package, origContentType);
	    out << '\n' << target << " = new " << orig << '[' << source
		<< ".length];";
	    out << "\nSystem.arraycopy(" << source << ", 0, " << target
		<< ", 0, " << source << ".length);";
	}
	else
	{
	    String_var index = string_dup("_ob_i");
	    index += iter;
	    String_var targetElement = string_dup(target);
	    targetElement += "[";
	    targetElement += index;
	    targetElement += "]";
	    String_var sourceElement = string_dup(source);
	    sourceElement += "[";
	    sourceElement += index;
	    sourceElement += "]";

	    out << '\n' << target << " = (" << type << ")"
		<< source << ".clone();";

	    out << "\nfor(int " << index << " = 0 ; " << index << " < "
	        << source << ".length ; " << index << "++)";
	    out.start();
	    writeMemberClone(targetElement, sourceElement, package,
			     contentType, contentTypeName, out, iter + 1);
	    out.end();
	}
	break;
    }
    }
}

// ----------------------------------------------------------------------
// Java code generator constructor and destructor
// ----------------------------------------------------------------------

IdlJavaGenerator::IdlJavaGenerator(const char* prog,
				   Repository_ptr rep,
				   const IdlCommentSeq& commentSeq,
				   const IdlPrefixPackageSeq& prefixPackageSeq,
				   const char* package,
				   bool autoPackage,
				   bool noStreams,
				   bool noComments,
				   bool local,
				   bool clone,
				   const char* outDir)
    : prog_(prog),
      repository_(Repository::_duplicate(rep)),
      partial_(false),
      commentSeq_(commentSeq),
      prefixPackageSeq_(prefixPackageSeq),
      package_(package),
      autoPackage_(autoPackage),
      noStreams_(noStreams),
      noComments_(noComments),
      local_(local),
      clone_(clone),
      outDir_(outDir)
{
}

IdlJavaGenerator::IdlJavaGenerator(const char* prog,
				   Repository_ptr rep,
				   const RepositoryIdSeq& idSeq,
				   const IdlCommentSeq& commentSeq,
				   const IdlPrefixPackageSeq& prefixPackageSeq,
				   const char* package,
				   bool autoPackage,
				   bool noStreams,
				   bool noComments,
				   bool local,
				   bool clone,
				   const char* outDir)
    : prog_(prog),
      repository_(Repository::_duplicate(rep)),
      partial_(true),
      idSeq_(idSeq),
      commentSeq_(commentSeq),
      prefixPackageSeq_(prefixPackageSeq),
      package_(package),
      autoPackage_(autoPackage),
      noStreams_(noStreams),
      noComments_(noComments),
      local_(local),
      clone_(clone),
      outDir_(outDir)
{
}

// ----------------------------------------------------------------------
// Generate Java code
// ----------------------------------------------------------------------

int
IdlJavaGenerator::gen()
{
    IdlPrettyPrint out(cerr); // Nothing should be printed on this out
    return genJava(repository_, out);
}

int
IdlJavaGenerator::genSkel(bool)
{
    if(local_)
	return 0;

    IdlPrettyPrint out(cerr); // Nothing should be printed on this out
    return genJavaSkel(repository_, out);
}

int
IdlJavaGenerator::genTie(bool)
{
    if(local_)
	return 0;

    IdlPrettyPrint out(cerr); // Nothing should be printed on this out
    return genJavaTie(repository_, out);
}
