// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <Types.h>
#include <Util.h>
#include <GenUtil.h>

#ifdef WIN32
#   include <direct.h>
#endif

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#include <GenJava.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Special Java "Pretty-Print" class member implementation
// ----------------------------------------------------------------------

const char* IdlJavaGenerator::PrettyPrint::header_ =
"// **********************************************************************\n"
"//\n"
"// Generated by the ORBacus IDL to Java Translator\n"
"//\n"
"// Copyright (c) 1999\n"
"// Object-Oriented Concepts, Inc.\n"
"// Billerica, MA, USA\n"
"//\n"
"// Copyright (c) 1999\n"
"// Object-Oriented Concepts GmbH\n"
"// Ettlingen, Germany\n"
"//\n"
"// All Rights Reserved\n"
"//\n"
"// **********************************************************************";

IdlJavaGenerator::PrettyPrint::PrettyPrint(const char* prog,
					   const char* name,
					   const char* package,
					   const char* dir,
					   unsigned int num)
    : IdlPrettyPrint(out_, num)
{
    //
    // Create directories and construct the directory path
    //
    path_ = string_dup(dir);
    String_var str = string_dup(package);
    char* s = str.inout();
    char* tok = 0;
    const char* sep = ".";
    while((tok = strtok(s, sep)))
    {
	s = 0;

	if(strlen(tok))
	{
	    if(strlen(path_) > 0)
		path_ += '/';
	    
	    path_ += tok;

#ifdef WIN32
	    _mkdir((const char*)path_);
#else
	    mkdir((const char*)path_, S_IRWXU | S_IRWXG | S_IRWXO);
#endif
	}
    }

    //
    // Create complete file name path
    //
    if(strlen(path_) > 0)
	path_ += '/';
    path_ += name;
    path_ += ".java";

    //
    // Create the file
    //
    out_.open(path_);

    //
    // Write header and package
    //
    if(good())
    {
	*this << header_;
	*this << '\n';
	*this << "\n// Version: " << Version;
	*this << "\n// License: " << License;

	if(package && strlen(package) > 0)
	{
	    this -> sep();
	    *this << "\npackage " << package << ';';
	}
    }
    else
    {
	cerr << prog << ": can't open \"" << path_
	     << "\": " << strerror(errno) << endl;
    }
}

// ----------------------------------------------------------------------
// Print header comment
// ----------------------------------------------------------------------

void
IdlJavaGenerator::comment(const char* s, IdlPrettyPrint& out)
{
    out.ignoreNextSep(false);
    out.sep();
    out << "\n//";
    out << "\n// " << s;
    out << "\n//";
    out.ignoreNextSep(true);
}

// ----------------------------------------------------------------------
// Print documentation comment
// ----------------------------------------------------------------------

void
IdlJavaGenerator::docComment(const char* id, IdlPrettyPrint& out)
{
    if(!noComments_)
    {
	for(ULong i = 0 ; i < commentSeq_.length() ; i++)
	{
	    if(strcmp(id, commentSeq_[i].id) == 0)
	    {
		out << "\n/**";

		const char* s = commentSeq_[i].comment;

		while(*s)
		{
		    static const char* ex = "@exception";
		    if(strncmp(s, ex, strlen(ex)) == 0)
		    {
			out << ex;
			s += strlen(ex);
			
			while(*s == ' ' || *s == '\t')
			{
			    out << *s;
			    s++;
			}

			const char* s2 = s;

			while(*s2 != ' ' && *s2 != '\t' &&
			      *s2 != '\n' && *s2 != '\0')
			{
			    s2++;
			}

			if((*s2 == ' ' || *s2 == '\t') && (s2 > s))
			{
			    String_var name = string_alloc(s2 - s);
			    strncpy(name.inout(), s, s2 - s);
			    name[s2 - s] = '\0';

			    Contained_var contained =
				repository_ -> lookup_id(id);

			    if(!is_nil(contained))
			    {
				Container_var container =
				    contained -> defined_in();

				if(!is_nil(container))
				{
				    do
				    {
					try
					{
					    contained =
						container -> lookup(name);
					}
					catch(const INTF_REPOS&)
					{
					    contained =
						Contained::_nil();

					    break;
					}

					if(is_nil(contained))
					{
					    Contained_var c =
						Contained::
						_narrow(container);
					    
					    if(is_nil(c))
						break;
					    
					    container = c -> defined_in();
					}
				    }
				    while(is_nil(contained));

				    if(!is_nil(contained))
					name = getAbsolute(contained);
				}
			    }

			    out << name;

			    out << *s2;
			    s = s2 + 1;
			}

			continue;
		    }

		    if(*s == '\n')
			out << "\n *";
		    else
			out << *s;
			
		    s++;
		}

		out << "*/";

		break;
	    }
	}
    }
}

// ----------------------------------------------------------------------
// Precede Java keywords and reserved names by an underscore
// ----------------------------------------------------------------------

char*
IdlJavaGenerator::fixName(const char* abs)
{
    assert_nca(abs, NCANullString);

    if(*abs == '\0')
	return string_dup("");

    static const char* kwds[] =
    {
	"abstract", "boolean", "break", "byte",
	"case", "catch", "char", "class", "const",
	"continue", "default", "do", "double", "else",
	"extends", "final", "finally", "float", "for",
	"goto", "if", "implements", "import", "instanceof",
	"int", "interface", "long", "native", "new",
	"package", "private", "protected", "public",
	"return", "short", "static", "super", "switch",
	"synchronized", "this", "throw", "throws",
	"transient", "try", "void", "volatile", "while",
	"null", "false", "true",
	"clone", "equals", "finalize", "getClass", "hashCode",
	"notify", "notifyAll", "toString", "wait",
	0
    };

    static const char* reserved[] =
    {
	"Helper", "Holder", "Package", "Operations",
	0
    };

    String_var right = (const char*)"";
    const char* s = strchr(abs, ':');
    if(s)
    {
	assert(s[1] == ':');
	right = fixName(s + 2);
    }

    String_var result;
    ULong len = s ? s - abs : strlen(abs);

    ULong i;
    for(i = 0 ; kwds[i] != 0 ; i++)
	if(strlen(kwds[i]) == len && strncmp(kwds[i], abs, len) == 0)
	{
	    result = string_alloc(len + 1);
	    result[0] = '_';
	    strncpy(&result[1], abs, len);
	    result[len + 1] = '\0';
	    break;
	}

    ULong j;
    for(j = 0 ; reserved[j] != 0 ; j++)
    {
	ULong rlen = strlen(reserved[j]);
	if(len > rlen && strncmp(reserved[j], abs + len - rlen, rlen) == 0)
	{
	    result = string_alloc(len + 1);
	    result[0] = '_';
	    strncpy(&result[1], abs, len);
	    result[len + 1] = '\0';
	    break;
	}
    }

    if(kwds[i] == 0 && reserved[j] == 0)
    {
	result = string_alloc(len);
	strncpy(result.inout(), abs, len);
	result[len] = '\0';
    }

    if(s)
    {
	result += ".";
	result += right;
    }

    return result._retn();
}

// ----------------------------------------------------------------------
// Create type string
// ----------------------------------------------------------------------

char*
IdlJavaGenerator::getTypeString(const char* package, TypeCode_ptr type,
				GetType gt)
{
    //
    // Get result
    //
    String_var result = string_dup("???");
    if(gt == GetTypeHelper && type -> kind() == tk_alias)
    {
	//
	// Helper classes are available for all aliases
	//
	result = getAbsolute(type -> id());
	result += "Helper";
	removePackage(package, result.inout());
	return result._retn();
    }
    else
    {
	TypeCode_var origType = GetOrigType(type);
	TCKind kind = origType -> kind();
	switch(kind)
	{
	case tk_void:
	case tk_short:
	case tk_long:
	case tk_ushort:
	case tk_ulong:
	case tk_longlong:
	case tk_ulonglong:
	case tk_float:
	case tk_double:
	case tk_fixed:
	case tk_boolean:
	case tk_char:
	case tk_wchar:
	case tk_octet:
	case tk_string:
	case tk_wstring:
	{
	    // **************************************************
	    // Basic types
	    // **************************************************
	    
	    switch(gt)
	    {
	    case GetTypeHolder:
	    {
		static const char* names[] =
		{
		    "???", "???", // null, void
		    "org.omg.CORBA.ShortHolder", "org.omg.CORBA.IntHolder",
		    "org.omg.CORBA.ShortHolder", "org.omg.CORBA.IntHolder",
		    "org.omg.CORBA.FloatHolder", "org.omg.CORBA.DoubleHolder",
		    "org.omg.CORBA.BooleanHolder", "org.omg.CORBA.CharHolder",
		    "org.omg.CORBA.ByteHolder",
		    "???", "???", "???", "???", "???", "???", "???",
		    "org.omg.CORBA.StringHolder",
		    "???", "???", "???", "???",
		    "org.omg.CORBA.LongHolder", "org.omg.CORBA.LongHolder",
		    "???", "org.omg.CORBA.CharHolder",
		    "org.omg.CORBA.StringHolder", "org.omg.CORBA.FixedHolder",
		    "???", "???", // value, value_box
		    "???", "???" // native, abstract_interface
		};
		result = names[(int)kind];
		break;
	    }
	    
	    case GetTypeHelper:
	    case GetTypeStub:
	    case GetTypeLocalStub:
		result = string_dup("???");
		break;

	    case GetTypeNormal:
	    {
		static const char* names[] =
		{
		    "???", // null
		    "void",
		    "short", "int",
		    "short", "int",
		    "float", "double",
		    "boolean",
		    "char",
		    "byte",
		    "???", "???", "???", "???", "???", "???", "???",
		    "String",
		    "???", "???", "???", "???",
		    "long", "long",
		    "???",
		    "char",
		    "String",
		    "java.math.BigDecimal",
		    "java.io.Serializable",
		    "java.io.Serializable",
		    "???", "???" // native, abstract_interface
		};
		result = names[(int)kind];
		break;
	    }
	    }
	    
	    break;
	}
	
	case tk_any:
	{
	    // **************************************************
	    // The any type
	    // **************************************************

	    switch(gt)
	    {
	    case GetTypeHolder:
		result = string_dup("org.omg.CORBA.AnyHolder");
		break;
		
	    case GetTypeHelper:
	    case GetTypeStub:
	    case GetTypeLocalStub:
		result = string_dup("???");
		break;

	    case GetTypeNormal:
		result = string_dup("org.omg.CORBA.Any");
		break;
	    }
	    
	    break;
	}
	
	case tk_TypeCode:
	case tk_Principal:
	case tk_objref:
	{
	    // **************************************************
	    // Object references (including pseudo object refereces)
	    // **************************************************
	    
	    if(kind == tk_TypeCode)
	    {
		switch(gt)
		{
		case GetTypeHelper:
		case GetTypeStub:
		case GetTypeLocalStub:
		    result = string_dup("???");
		    break;

		case GetTypeNormal:
		case GetTypeHolder:
		    result = string_dup("org.omg.CORBA.TypeCode");
		    break;
		}
	    }
	    else if(kind == tk_Principal)
	    {
		switch(gt)
		{
		case GetTypeHelper:
		case GetTypeStub:
		case GetTypeLocalStub:
		    result = string_dup("???");
		    break;

		case GetTypeNormal:
		case GetTypeHolder:
		    result = string_dup("org.omg.CORBA.Principal");
		    break;
		}
	    }
	    else if(kind == tk_objref)
	    {
		if(strcmp(origType -> id(), "IDL:omg.org/CORBA/Object:1.0")
		   == 0)
		{
		    switch(gt)
		    {
		    case GetTypeHelper:
		    case GetTypeStub:
		    case GetTypeLocalStub:
			result = string_dup("???");
			break;
			
		    case GetTypeNormal:
		    case GetTypeHolder:
			result = string_dup("org.omg.CORBA.Object");
			break;
		    }
		}
		else
		{
		    String_var s = getAbsolute(origType -> id());
		    removePackage(package, s.inout());
		    
		    switch(gt)
		    {
		    case GetTypeHelper:
			result = s;
			result += "Helper";
			break;

		    case GetTypeStub:
		    {
			const char* p = strrchr(s.inout(), '.');

			if(p)
			{
			    p++;
			    assert(*p);
			    result = string_alloc(p - s.in());
			    strncpy(result.inout(), s, p - s.in());
			    result[p - s.in()] = '\0';

			    if(noStreams_)
			    {
				result += "StubFor";
				result += p;
			    }
			    else
			    {
				result += '_';
				result += p;
				result += "Stub";
			    }
			}
			else
			{
			    if(noStreams_)
			    {
				result = string_dup("StubFor");
				result += s;
			    }
			    else
			    {
				result = string_dup("_");
				result += s;
				result += "Stub";
			    }
			}

			break;
		    }
			
		    case GetTypeLocalStub:
		    {
			const char* p = strrchr(s.inout(), '.');

			if(p)
			{
			    p++;
			    assert(*p);
			    result = string_alloc(p - s.in());
			    strncpy(result.inout(), s, p - s.in());
			    result[p - s.in()] = '\0';

			    result += '_';
			    result += p;
			    result += "LocalStub";
			}
			else
			{
			    result = string_dup("_");
			    result += s;
			    result += "LocalStub";
			}

			break;
		    }
			
		    case GetTypeNormal:
		    case GetTypeHolder:
			result = s;
			break;
		    }
		}
	    }
	    
	    if(gt == GetTypeHolder)
		result += "Holder";
	    
	    break;
	}
	
	case tk_struct:
	case tk_union:
	case tk_except:
	case tk_enum:
	{
	    // **************************************************
	    // Structs, unions, exceptions, enums
	    // **************************************************
	    
	    String_var s = getAbsolute(origType -> id());
	    removePackage(package, s.inout());
	    result = s;
	    
	    switch(gt)
	    {
	    case GetTypeHelper:
		result += "Helper";
		break;

	    case GetTypeHolder:
		result += "Holder";
		break;

	    case GetTypeStub:
	    case GetTypeLocalStub:
		result = string_dup("???");
		break;
			
	    case GetTypeNormal:
		break;
	    }

	    break;
	}
	
	case tk_sequence:
	case tk_array:
	{
	    // **************************************************
	    // Sequences and arrays
	    // **************************************************
	    
	    switch(gt)
	    {
	    case GetTypeHelper:
		if(type -> kind() == tk_alias)
		{
		    String_var s = getAbsolute(type -> id());
		    removePackage(package, s.inout());
		    s += "Helper";
		    result = s;
		}
		break;

	    case GetTypeHolder:
		if(type -> kind() == tk_alias)
		{
		    String_var s = getAbsolute(type -> id());
		    removePackage(package, s.inout());
		    s += "Holder";
		    result = s;
		}
		break;

	    case GetTypeStub:
	    case GetTypeLocalStub:
		result = string_dup("???");
		break;
			
	    case GetTypeNormal:
	    {
		TypeCode_var contentType = origType -> content_type();
		result = getTypeString(package, contentType, gt);
		result += "[]";
		break;
	    }
	    }
	    
	    break;
	}
	
	case tk_alias:
	case tk_null:
	    assert(false);
	    break;
	}
    }

    return result._retn();
}

// ----------------------------------------------------------------------
// Check if code needs to be generated for a specific repository id
// ----------------------------------------------------------------------

bool
IdlJavaGenerator::check(const char* id)
{
    if(!partial_)
	return true;

    for(ULong i = 0 ; i < idSeq_.length() ; i++)
	if(strcmp(id, idSeq_[i]) == 0)
	    return true;

    return false;
}

// ----------------------------------------------------------------------
// Get package from package options
// ----------------------------------------------------------------------

char*
IdlJavaGenerator::getPackageFromOptions(const char* id)
{
    String_var package = string_dup("");

    //
    // First handle "--prefix-package"
    //
    for(ULong i = 0 ; i < prefixPackageSeq_.length() ; i++)
    {
	if(strncmp(prefixPackageSeq_[i].prefix, id,
		   strlen(prefixPackageSeq_[i].prefix)) == 0)
	{
	    package = prefixPackageSeq_[i].package;
	    break;
	}
    }

    //
    // Now handle "--package"
    //
    if(package.in() == 0 || strlen(package) == 0)
	if(strlen(package_))
	    package = package_;
    
    //
    // Handle automatic package generation
    //
    if(package.in() == 0 || strlen(package) == 0)
	if(autoPackage_)
	    package = getPackageFromPrefix(id);

    return package._retn();
}

// ----------------------------------------------------------------------
// Get absolute name
// ----------------------------------------------------------------------

char*
IdlJavaGenerator::getAbsolute(const char* id)
{
    for(ULong i = 0 ; i < absSeq_.length() ; i += 2)
    {
	if(strcmp(absSeq_[i], id) == 0)
	    return string_dup(absSeq_[i + 1]);
    }
    
    Contained_var contained = repository_ -> lookup_id(id);
    assert(!is_nil(contained));
    
    ULong len = absSeq_.length();
    absSeq_.length(len + 2);
    absSeq_[len] = id;
    absSeq_[len + 1] = getAbsolute(contained);
    return string_dup(absSeq_[len + 1]);
}

char*
IdlJavaGenerator::getAbsolute(Contained_ptr contained)
{
    RepositoryId_var id = contained -> id();
    String_var package = getPackageFromOptions(id);
    return getAbsolutePackage(contained, package);
}

char*
IdlJavaGenerator::getAbsolutePackage(Contained_ptr contained,
				     const char* package)
{
    Identifier_var name = contained -> name();
    name = fixName(name);

    Container_var container = contained -> defined_in();
    Contained_var contained2 = Contained::_narrow(container);

    if(is_nil(contained2))
    {
 	if(package && strlen(package))
	{
	    char* result = string_alloc(strlen(package) + 1 +
					      strlen(name));
	    strcpy(result, package);
	    strcat(result, ".");
	    strcat(result, name);
	    return result;
	}
	else
	{
	    if(contained -> def_kind() == dk_Module)
	    {
		return string_dup(name);
	    }
	    else
	    {
		RepositoryId_var id = contained -> id();
		cerr << prog_ << ": warning: no package for `" << id
		     << "' (using `defaultpkg')" << endl;

		static const char* s = "defaultpkg.";
		char* result = string_alloc(strlen(s) + strlen(name));
		strcpy(result, s);
		strcat(result, name);
		return result;
	    }
	}
    }
    else
    {
	String_var s = getAbsolutePackage(contained2, package);

	if(contained2 -> def_kind() == dk_Interface)
	{
	    DefinitionKind k = contained -> def_kind();
	    if(k == dk_Alias || k == dk_Enum ||
	       k == dk_Struct || k == dk_Exception ||
	       k == dk_Union)
	    {
		s += "Package";
	    }
	}
	
	char* result = string_alloc(strlen(s) + 1 + strlen(name));
	strcpy(result, s);
	strcat(result, ".");
	strcat(result, name);
	return result;
    }
}

// ----------------------------------------------------------------------
// Get package name
// ----------------------------------------------------------------------

char*
IdlJavaGenerator::getPackage(const char* id)
{
    for(ULong i = 0 ; i < pkgSeq_.length() ; i += 2)
    {
	if(strcmp(pkgSeq_[i], id) == 0)
	    return string_dup(pkgSeq_[i + 1]);
    }
    
    Contained_var contained = repository_ -> lookup_id(id);
    assert(!is_nil(contained));
    
    ULong len = pkgSeq_.length();
    pkgSeq_.length(len + 2);
    pkgSeq_[len] = id;
    pkgSeq_[len + 1] = getPackage(contained);
    return string_dup(pkgSeq_[len + 1]);
}

char*
IdlJavaGenerator::getPackage(Contained_ptr contained)
{
    RepositoryId_var id = contained -> id();
    String_var package = getPackageFromOptions(id);
    return getPackagePackage(contained, package);
}

char*
IdlJavaGenerator::getPackagePackage(Contained_ptr contained,
				    const char* package)
{
    Container_var container = contained -> defined_in();
    Contained_var contained2 = Contained::_narrow(container);

    if(is_nil(contained2))
    {
 	if(package && strlen(package))
	{
	    return string_dup(package);
	}
	else
	{
	    if(contained -> def_kind() == dk_Module)
	    {
		return string_dup("");
	    }
	    else
	    {
		RepositoryId_var id = contained -> id();
		cerr << prog_ << ": warning: no package for `" << id
		     << "' (using `defaultpkg')" << endl;

		return string_dup("defaultpkg");
	    }
	}
    }
    else
    {
	if(contained2 -> def_kind() == dk_Module)
	{
	    return getAbsolutePackage(contained2, package);
	}
	else if(contained2 -> def_kind() == dk_Interface)
	{
	    DefinitionKind k = contained -> def_kind();
	    if(k == dk_Alias || k == dk_Enum ||
	       k == dk_Struct || k == dk_Exception ||
	       k == dk_Union)
	    {
		String_var s = getAbsolutePackage(contained2, package);
		s += "Package";
		return s._retn();
	    }
	    else
	    {
		return getPackagePackage(contained2, package);
	    }
	}
	else
	{
	    return getPackagePackage(contained2, package);
	}
    }
}

// ----------------------------------------------------------------------
// Build package name from prefix in repository id
// ----------------------------------------------------------------------

char*
IdlJavaGenerator::getPackageFromPrefix(const char* id)
{
    String_var package = string_dup("");

    if(strchr(id, '/') && strncmp(id, "IDL:", 4) == 0)
    {
	String_var str = string_dup(id + 4);
	*strchr(str.inout(), '/') = '\0';
	char* s = str.inout();
	const char* tok;

	while((tok = strtok(s, ".")))
	{
	    s = 0;
	    if(strlen(package) > 0)
	    {
		String_var newPackage =
		    string_alloc(strlen(tok) + 1 + strlen(package));
		strcpy(newPackage.inout(), tok);
		strcat(newPackage.inout(), ".");
		strcat(newPackage.inout(), package);
		package = newPackage;
	    }
	    else
		package = string_dup(tok);
	}
    }

    return package._retn();
}

// ----------------------------------------------------------------------
// Convert strings with escape sequences
// ----------------------------------------------------------------------

void
IdlJavaGenerator::addEscapes(char*& str, bool addQuotes)
{
    char* s = str;
    String_var result;

    if(addQuotes)
	result += '"';

    while(*s)
    {
	switch(*s)
	{
	case '\n':
	    result += "\\n";
	    break;

	case '\t':
	    result += "\\t";
	    break;

	case '\b':
	    result += "\\b";
	    break;

	case '\r':
	    result += "\\r";
	    break;

	case '\f':
	    result += "\\f";
	    break;

	case '\\':
	    result += "\\\\";
	    break;

	case '\'':
	    result += "\\'";
	    break;

	case '"':
	    result += "\\\"";
	    break;

	default:
	    if(isprint(*s))
		result += *s;
	    else
	    {
		Char c = *s;
		char str[5];
		sprintf(str, "\\%.3o", c);
		result += str;
	    }
	    break;
	}
	
	s++;
    }

    if(addQuotes)
	result += '"';

    string_free(str);
    str = result._retn();
}
void
IdlJavaGenerator::addEscapes(wchar_t*& str, bool addQuotes)
{
    wchar_t* s = str;
    WString_var result;

    if(addQuotes)
	result += L'"';

    while(*s)
    {
	switch(*s)
	{
	case L'\n':
	    result += L"\\n";
	    break;

	case L'\t':
	    result += L"\\t";
	    break;

	case L'\b':
	    result += L"\\b";
	    break;

	case L'\r':
	    result += L"\\r";
	    break;

	case L'\f':
	    result += L"\\f";
	    break;

	case L'\\':
	    result += L"\\\\";
	    break;

	case L'\'':
	    result += L"\\'";
	    break;

	case L'"':
	    result += L"\\\"";
	    break;

	default:
//	    if(isprint(*s))
	    result += *s;
	    break;
	}
	
	s++;
    }

    if(addQuotes)
	result += L'"';

    wstring_free(str);
    str = result._retn();
}

// ----------------------------------------------------------------------
// Remove package from absolute name
// ----------------------------------------------------------------------

void
IdlJavaGenerator::removePackage(const char* package, char*& absolute)
{
    assert_nca(package, NCANullString);
    assert_nca(absolute, NCANullString);

/*
    if(strlen(absolute) > strlen(package) && strlen(package) > 0)
    {
	assert(package[strlen(package) - 1] != '.');

	if(strncmp(package, absolute, strlen(package)) == 0 &&
	   absolute[strlen(package)] == '.')
	{
	    char* a = string_dup(absolute + strlen(package) + 1);
	    string_free(absolute);
	    absolute = a;
	}
    }
*/
    if(strlen(absolute) > strlen(package) && strlen(package) > 0)
    {
	assert(package[strlen(package) - 1] != '.');

	const char* s = strrchr(absolute, '.');
	if(s)
	{
	    unsigned int len = s - absolute;
	    if(strlen(package) == len && strncmp(package, absolute, len) == 0)
	    {
		char* a = string_dup(s + 1);
		string_free(absolute);
		absolute = a;
	    }
	}
    }
}

// ----------------------------------------------------------------------
// Get union label sequence
// ----------------------------------------------------------------------

IdlStringSeq*
IdlJavaGenerator::getUnionLabelSeq(const char* package,
				   TypeCode_ptr discType,
				   const IdlUnionMemberInfo& info)
{
    TypeCode_var origDiscType = GetOrigType(discType);
    IdlStringSeq_var seq = new IdlStringSeq;
    ULong i;

    switch(origDiscType -> kind())
    {
    case tk_short:
    case tk_ushort:

	for(i = 0 ; i < info.pLabels.length() ; i++)
	{
	    String_var label = string_dup("(short)");
	    label += info.pLabels[i];
	    seq -> append(label);
	}

	for(i = 0 ; i < info.nLabels.length() ; i++)
	{
	    String_var label = string_dup("(short)");
	    label += '-';
	    label += info.nLabels[i];
	    seq -> append(label);
	}

	break;
	
    case tk_long:
    case tk_ulong:

	for(i = 0 ; i < info.pLabels.length() ; i++)
	{
	    String_var label = string_dup("");
	    label += info.pLabels[i];
	    seq -> append(label);
	}

	for(i = 0 ; i < info.nLabels.length() ; i++)
	{
	    String_var label = string_dup("-");
	    label += info.nLabels[i];
	    seq -> append(label);
	}

	break;
	
    case tk_char:

	for(i = 0 ; i < info.pLabels.length() ; i++)
	{
	    String_var s;
	    if((char)info.pLabels[i] == '\0')
		s = string_dup("\\0");
	    else
	    {
		s = string_alloc(1);
		s[0] = (char)info.pLabels[i];
		s[1] = '\0';
		addEscapes(s.inout(), false);
	    }

	    String_var label = string_dup("'");
	    label += s;
	    label += '\'';
	    seq -> append(label);
	}

	assert(info.nLabels.length() == 0);
	
	break;
	
    case tk_boolean:

	for(i = 0 ; i < info.pLabels.length() ; i++)
	{
	    String_var label = string_dup("");
	    label += info.pLabels[i];
	    seq -> append(label);
	}

	assert(info.nLabels.length() == 0);
	
	break;
	
    case tk_enum:
    {
	String_var disc = getTypeString(package, origDiscType);

	for(i = 0 ; i < info.pLabels.length() ; i++)
	{
	    Identifier_var memberName =
		origDiscType -> member_name(info.pLabels[i]);
	    memberName = fixName(memberName);

	    String_var label = disc;
	    label += "._";
	    label += memberName;
	    seq -> append(label);
	}

	assert(info.nLabels.length() == 0);

	break;
    }

    default:
	assert(false);
	break;
    }

    return seq._retn();
}

// ----------------------------------------------------------------------
// Get holder class for a union member
// ----------------------------------------------------------------------

char *
IdlJavaGenerator::getUnionMemberHolder(TypeCode_ptr tc)
{
    String_var result;

    switch(tc -> kind())
    {
    case tk_null:
    case tk_void:
	assert(false);
	break;

    case tk_short:
    case tk_ushort:
	result = string_dup("org.omg.CORBA.ShortHolder");
	break;

    case tk_long:
    case tk_ulong:
	result = string_dup("org.omg.CORBA.IntHolder");
	break;

    case tk_longlong:
    case tk_ulonglong:
	result = string_dup("org.omg.CORBA.LongHolder");
	break;

    case tk_float:
	result = string_dup("org.omg.CORBA.FloatHolder");
	break;

    case tk_double:
	result = string_dup("org.omg.CORBA.DoubleHolder");
	break;

    case tk_fixed:
	result = string_dup("org.omg.CORBA.FixedHolder");
	break;

    case tk_boolean:
	result = string_dup("org.omg.CORBA.BooleanHolder");
	break;

    case tk_char:
    case tk_wchar:
	result = string_dup("org.omg.CORBA.CharHolder");
	break;

    case tk_octet:
	result = string_dup("org.omg.CORBA.ByteHolder");
	break;

    case tk_alias:
	assert(false);
	break;

    default:
	result = string_dup("");
	break;
    }

    return result._retn();
}

// ----------------------------------------------------------------------
// Determines if the typecode is a primitive (or immutable) type
// ----------------------------------------------------------------------

bool
IdlJavaGenerator::isPrimitiveMemberType(TypeCode_ptr tc)
{
    bool result;

    switch(tc -> kind())
    {
    case tk_null:
    case tk_void:
	assert(false);
	break;

    case tk_short:
    case tk_ushort:
    case tk_long:
    case tk_ulong:
    case tk_longlong:
    case tk_ulonglong:
    case tk_float:
    case tk_double:
    case tk_fixed:
    case tk_boolean:
    case tk_char:
    case tk_wchar:
    case tk_octet:
    case tk_string:
    case tk_wstring:
    case tk_TypeCode:
    case tk_Principal:
    case tk_objref:
	result = true;
	break;

    default:
	result = false;
	break;
    }

    return result;
}

