// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <Types.h>
#include <Util.h>
#include <GenUtil.h>
#include <GenTextBase.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#include <GenRTF.h>

using namespace CORBA;
using namespace OB;

//
// The top-level file name (without ".rtf")
//
const char* IdlRTFGenerator::top_ = "_top_";

const char* IdlRTFGenerator::boldon_ = "\\b ";
const char* IdlRTFGenerator::boldoff_ = "\\b0 ";

// ----------------------------------------------------------------------
// Generate code for the header file, general part
// ----------------------------------------------------------------------

int
IdlRTFGenerator::genRTF(Container_ptr container, IdlPrettyPrint& out)
{
    //
    // Get scope
    //
    String_var scope;
    getScope(container, scope.out());

    //
    // Get container contents descriptions
    //
    Container::DescriptionSeq_var allSeq =
	container -> describe_contents(dk_all, true, -1);

    ULong i, j, k;

    //
    // Create name and description list
    //
    IdlStringSeq nameSeq;
    Container::DescriptionSeq seq;
    for(i = 0; i < allSeq -> length() ; i++)
    {
	const char* name;

	switch(allSeq[i].kind)
	{
	case dk_Attribute:
	{
	    AttributeDescription* desc;
	    Boolean b = allSeq[i].value >>= desc;
	    assert(b);
	    name = desc -> name;
	    break;
	}

	case dk_Constant:
	{
	    ConstantDescription* desc;
	    Boolean b = allSeq[i].value >>= desc;
	    assert(b);
	    name = desc -> name;
	    break;
	}

	case dk_Exception:
	{
	    ExceptionDescription* desc;
	    Boolean b = allSeq[i].value >>= desc;
	    assert(b);
	    name = desc -> name;
	    break;
	}

	case dk_Interface:
	{
	    InterfaceDescription* desc;
	    Boolean b = allSeq[i].value >>= desc;
	    assert(b);
	    name = desc -> name;
	    break;
	}

	case dk_Module:
	{
	    ModuleDescription* desc;
	    Boolean b = allSeq[i].value >>= desc;
	    assert(b);
	    name = desc -> name;
	    break;
	}

	case dk_Operation:
	{
	    OperationDescription* desc;
	    Boolean b = allSeq[i].value >>= desc;
	    assert(b);
	    name = desc -> name;
	    break;
	}

	case dk_Typedef:
	case dk_Alias:
	case dk_Struct:
	case dk_Union:
	case dk_Enum:
	{
	    TypeDescription* desc;
	    Boolean b = allSeq[i].value >>= desc;
	    assert(b);
	    name = desc -> name;
	    break;
	}

	default:
	    continue;
	}

	seq.append(allSeq[i]);
	nameSeq.append(name);
    }

    //
    // Sort name and description lists (bubblesort)
    //
    if(!noSort_)
    {
	for(i = 0 ; i < nameSeq.length() ; i++)
	{
	    for(j = i ; j < nameSeq.length() ; j++)
	    {
		if(!ignoreCase_)
		{
		    if(strcmp(nameSeq[i], nameSeq[j]) > 0)
		    {
			String_var name = nameSeq[i]._retn();
			nameSeq[i] = nameSeq[j]._retn();
			nameSeq[j] = name._retn();

			Container::Description desc = seq[i];
			seq[i] = seq[j];
			seq[j] = desc;
		    }
		}
		else
		{
#ifdef WIN32
		    if(stricmp(nameSeq[i], nameSeq[j]) > 0)
#else
		    if(strcasecmp(nameSeq[i], nameSeq[j]) > 0)
#endif
		    {
			String_var name = nameSeq[i]._retn();
			nameSeq[i] = nameSeq[j]._retn();
			nameSeq[j] = name._retn();
			
			Container::Description desc = seq[i];
			seq[i] = seq[j];
			seq[j] = desc;
		    }
		}
	    }
	}
    }

    bool first;

    // ------------------------------------------------------------
    // Pass 1: Index generation
    // ------------------------------------------------------------

    if(withIndex_)
    {
	out << '\n';

	//
	// Alias index
	//
	first = true;
	for(i = 0 ; i < seq.length() ; i++)
	{
	    if(seq[i].kind == dk_Alias)
	    {
		TypeDescription* desc;
		Boolean b = seq[i].value >>= desc;
		assert(b);

		if(first)
		{
		    out << '\n' << heading_ << "Alias Index" << body_ << '\n';
		    first = false;
		}

		intro(desc -> id, desc -> name, out);
	    }
	}
	if(!first)
	    out << '\n';

	//
	// Constant index
	//
	first = true;
	for(i = 0 ; i < seq.length() ; i++)
	{
	    if(seq[i].kind == dk_Constant)
	    {
		ConstantDescription* desc;
		Boolean b = seq[i].value >>= desc;
		assert(b);

		if(first)
		{
		    out << '\n' << heading_ << "Constant Index" << body_
			<< '\n';
		    first = false;
		}

		intro(desc -> id, desc -> name, out);
	    }
	}
	if(!first)
	    out << '\n';

	//
	// Exception index
	//
	first = true;
	for(i = 0 ; i < seq.length() ; i++)
	{
	    if(seq[i].kind == dk_Exception)
	    {
		ExceptionDescription* desc;
		Boolean b = seq[i].value >>= desc;
		assert(b);
	    
		if(first)
		{
		    out << '\n' << heading_ << "Exception Index" << body_
			<< '\n';
		    first = false;
		}

		intro(desc -> id, desc -> name, out);
	    }
	}
	if(!first)
	    out << '\n';

	//
	// Struct index
	//
	first = true;
	for(i = 0 ; i < seq.length() ; i++)
	{
	    if(seq[i].kind == dk_Struct)
	    {
		TypeDescription* desc;
		Boolean b = seq[i].value >>= desc;
		assert(b);

		if(first)
		{
		    out << '\n' << heading_ << "Struct Index" << body_ << '\n';
		    first = false;
		}
		
		intro(desc -> id, desc -> name, out);
	    }
	}
	if(!first)
	    out << '\n';

	//
	// Union index
	//
	first = true;
	for(i = 0 ; i < seq.length() ; i++)
	{
	    if(seq[i].kind == dk_Union)
	    {
		TypeDescription* desc;
		Boolean b = seq[i].value >>= desc;
		assert(b);

		if(first)
		{
		    out << '\n' << heading_ << "Union Index" << body_ << '\n';
		    first = false;
		}

		intro(desc -> id, desc -> name, out);
	    }
	}
	if(!first)
	    out << '\n';

	//
	// Enum index
	//
	first = true;
	for(i = 0 ; i < seq.length() ; i++)
	{
	    if(seq[i].kind == dk_Enum)
	    {
		TypeDescription* desc;
		Boolean b = seq[i].value >>= desc;
		assert(b);

		if(first)
		{
		    out << '\n' << heading_ << "Enum Index" << body_ << '\n';
		    first = false;
		}

		intro(desc -> id, desc -> name, out);
	    }
	}
	if(!first)
	    out << '\n';

	//
	// Attribute index
	//
	first = true;
	for(i = 0 ; i < seq.length() ; i++)
	{
	    if(seq[i].kind == dk_Attribute)
	    {
		AttributeDescription* desc;
		Boolean b = seq[i].value >>= desc;
		assert(b);

		if(first)
		{
		    out << '\n' << heading_ << "Attribute Index" << body_
			<< '\n';
		    first = false;
		}

		intro(desc -> id, desc -> name, out);
	    }
	}
	if(!first)
	    out << '\n';

	//
	// Operation index
	//
	first = true;
	for(i = 0 ; i < seq.length() ; i++)
	{
	    if(seq[i].kind == dk_Operation)
	    {
		OperationDescription* desc;
		Boolean b = seq[i].value >>= desc;
		assert(b);
	    
		if(first)
		{
		    out << '\n' << heading_ << "Operation Index" << body_
			<< '\n';
		    first = false;
		}

		intro(desc -> id, desc -> name, out);
	    }
	}
	if(!first)
	    out << '\n';
	//
	// Module index
	//
	first = true;
	for(i = 0 ; i < seq.length() ; i++)
	{
	    if(seq[i].kind == dk_Module)
	    {
		ModuleDescription* desc;
		Boolean b = seq[i].value >>= desc;
		assert(b);

		if(first)
		{
		    out << '\n' << heading_ << "Module Index" << body_ << '\n';
		    first = false;
		}

		intro(desc -> id, desc -> name, out);
	    }
	}
	if(!first)
	    out << '\n';

	//
	// Interface index
	//
	first = true;
	for(i = 0 ; i < seq.length() ; i++)
	{
	    if(seq[i].kind == dk_Interface)
	    {
		InterfaceDescription* desc;
		Boolean b = seq[i].value >>= desc;
		assert(b);

		if(first)
		{
		    out << '\n' << heading_ << "Interface Index" << body_
			<< '\n';
		    first = false;
		}

		intro(desc -> id, desc -> name, out);
	    }
	}
	if(!first)
	    out << '\n';
    }

    // ------------------------------------------------------------
    // Pass 2: Description generation
    // ------------------------------------------------------------
     
    int status = 0;

    //
    // Alias description
    //
    first = true;
    for(i = 0 ; i < seq.length() && status == 0 ; i++)
    {
	if(seq[i].kind == dk_Alias)
	{
	    TypeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(first)
	    {
		out << '\n' << heading_ << "Aliases" << body_ << '\n';
		first = false;
	    }

	    TypeCode_var origType = desc -> type -> content_type();
	    TypeCode_var origOrigType = GetOrigType(origType);

	    String_var norm = getTypeString(scope, origType, GetTypeNormal,
					    desc -> name);

	    out << '\n' << boldon_ << desc -> name << boldoff_;
	    out.inc();
	    out << '\n' << literal_;

	    out << "typedef " << norm << ';';

	    out << body_;

	    out << '\n';

	    comment(desc -> id, out);

	    out.dec();
	}
    }

    //
    // Constant description
    //
    first = true;
    for(i = 0 ; i < seq.length() && status == 0 ; i++)
    {
	if(seq[i].kind == dk_Constant)
	{
	    ConstantDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(first)
	    {
		out << '\n' << heading_ << "Constants" << body_ << '\n';
		first = false;
	    }

	    ScopedName_var absolute = getAbsolute(desc -> id);
	    
	    out << '\n' << boldon_ << desc -> name << boldoff_;
	    out.inc();
	    out << '\n' << literal_;

	    String_var s = getTypeString(scope, desc -> type, GetTypeNormal,
					 desc -> name);

	    out << "const " << s << " = ";

	    TypeCode_var tc = desc -> value.type();
	    tc = GetOrigType(tc);

	    switch(tc -> kind())
	    {
	    case tk_short:
	    {
		Short v;
		Boolean b = desc -> value >>= v;
		assert(b);
		out << v;
		break;
	    }

	    case tk_ushort:
	    {
		UShort v;
		Boolean b = desc -> value >>= v;
		assert(b);
		out << v;
		break;
	    }

	    case tk_long:
	    {
		Long v;
		Boolean b = desc -> value >>= v;
		assert(b);
		out << v;
		break;
	    }

	    case tk_ulong:
	    {
		ULong v;
		Boolean b = desc -> value >>= v;
		assert(b);
		out << v;
		break;
	    }

	    case tk_longlong:
	    {
		LongLong v;
		Boolean b = desc -> value >>= v;
		assert(b);
		out << v;
		break;
	    }

	    case tk_ulonglong:
	    {
		ULongLong v;
		Boolean b = desc -> value >>= v;
		assert(b);
		out << v;
		break;
	    }

	    case tk_float:
	    {
		Float v;
		Boolean b = desc -> value >>= v;
		assert(b);
		out << v;
		break;
	    }

	    case tk_double:
	    {
		Double v;
		Boolean b = desc -> value >>= v;
		assert(b);
		out << v;
		break;
	    }

	    case tk_longdouble:
	    {
		LongDouble v;
		Boolean b = desc -> value >>= v;
		assert(b);
		out << v;
		break;
	    }

	    case tk_string:
	    {
		char* v;
		ULong l = tc -> length();
		Any::to_string to(v, l);
		Boolean b = desc -> value >>= to;
		assert(b);
		String_var s = string_dup(v);
		IdlAddEscapes(s.inout(), true);
		out << s;
		break;
	    }

	    case tk_char:
	    {
		Char v;
		Boolean b = desc -> value >>= Any::to_char(v);
		assert(b);
		String_var s;
		if(v == '\0')
		    s = string_dup("\\0");
		else
		{
		    s = string_alloc(1);
		    s[0] = v;
		    s[1] = '\0';
		    IdlAddEscapes(s.inout(), false);
		}
		out << "'" << s << "'";
		break;
	    }

	    case tk_boolean:
	    {
		Boolean v;
		Boolean b = desc -> value >>= Any::to_boolean(v);
		assert(b);
		out << (v ? "TRUE" : "FALSE");
		break;
	    }

	    default:
		assert(false);
		break;
	    }

	    out << ';';

            out << body_;

	    out << '\n';

	    comment(desc -> id, out);

	    out.dec();
	}
    }

    //
    // Struct description
    //
    first = true;
    for(i = 0 ; i < seq.length() && status == 0 ; i++)
    {
	if(seq[i].kind == dk_Struct)
	{
	    TypeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(first)
	    {
		out << '\n' << heading_ << "Structs" << body_ << '\n';
		first = false;
	    }

	    out << '\n'<< boldon_ << desc -> name << boldoff_;
	    out.inc();
	    out << '\n' << literal_;

	    out << "struct " << desc -> name;

	    out.start();

	    for(j = 0 ; j < desc -> type -> member_count() ; j++)
	    {
		TypeCode_var memberType =
		    desc -> type -> member_type(j);
		Identifier_var memberName =
		    desc -> type -> member_name(j);

		String_var str = getTypeString(scope, memberType,
					       GetTypeNormal, memberName);
		out << '\n' << str << ';';
	    }

	    out.end();
	    out << ';';
		
            out << body_;

	    out << '\n';

	    comment(desc -> id, out);

	    out.dec();
	}
    }

    //
    // Exception description
    //
    first = true;
    for(i = 0 ; i < seq.length() && status == 0 ; i++)
    {
	if(seq[i].kind == dk_Exception)
	{
	    ExceptionDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(first)
	    {
		out << '\n' << heading_ << "Exceptions" << body_ << '\n';
		first = false;
	    }

	    out << '\n' << boldon_ << desc -> name << boldoff_;
	    out.inc();
	    out << '\n' << literal_;

	    out << "exception " << desc -> name;

	    out.start();

	    for(j = 0 ; j < desc -> type -> member_count() ; j++)
	    {
		TypeCode_var memberType =
		    desc -> type -> member_type(j);
		Identifier_var memberName =
		    desc -> type -> member_name(j);

		String_var str = getTypeString(scope, memberType,
					       GetTypeNormal, memberName);
		out << '\n' << str << ';';
	    }

	    out.end();
	    out << ';';
		
            out << body_;

	    out << '\n';

	    comment(desc -> id, out);

	    out.dec();
	}
    }

    //
    // Union description
    //
    first = true;
    for(i = 0 ; i < seq.length() && status == 0 ; i++)
    {
	if(seq[i].kind == dk_Union)
	{
	    TypeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(first)
	    {
		out << '\n' << heading_ << "Unions" << body_ << '\n';
		first = false;
	    }

	    TypeCode_var discType =
		desc -> type -> discriminator_type();
	    String_var disc = getTypeString(scope, discType);

	    IdlUnionMemberInfoSeq_var infoSeq = IdlGetUnionInfo(desc -> type);

	    out << '\n' << boldon_ << desc -> name << boldoff_;
	    out.inc();
	    out << '\n' << literal_;

	    out << "union " << desc -> name;
	    out << " switch(" << disc << ')';
	    out.start();
	    for(j = 0 ; j < infoSeq -> length() ; j++)
	    {
		IdlUnionMemberInfo& info = infoSeq[j];

		printUnionLabels(info, out, discType);

		if(!is_nil(info.type))
		{
		    String_var norm =
			getTypeString(scope, info.type, GetTypeNormal,
				      info.name);

		    out << ' ' << norm << ';';
		}
	    }

	    out.end();
	    out << ';';

	    out << body_;

	    out << '\n';

	    comment(desc -> id, out);

	    out.dec();
	}
    }

    //
    // Enum description
    //
    first = true;
    for(i = 0 ; i < seq.length() && status == 0 ; i++)
    {
	if(seq[i].kind == dk_Enum)
	{
	    TypeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(first)
	    {
		out << '\n' << heading_ << "Enums" << body_ << '\n';
		first = false;
	    }

	    out << '\n' << boldon_ << desc -> name << boldoff_;
	    out.inc();
	    out << '\n' << literal_;

	    out << "enum " << desc -> name;
	    out.start();

	    for(j = 0 ; j < desc -> type -> member_count() ; j++)
	    {
		Identifier_var memberName =
		    desc -> type -> member_name(j);
		out << '\n' << memberName;
		if(j < desc -> type -> member_count() - 1)
		    out << ',';
	    }
	    
	    out.end();
	    out << ';';

	    out << body_;

	    out << '\n';

	    comment(desc -> id, out);

	    out.dec();
	}
    }

    //
    // Attribute Description
    //
    first = true;
    for(i = 0 ; i < seq.length() && status == 0 ; i++)
    {
	if(seq[i].kind == dk_Attribute)
	{
	    AttributeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(first)
	    {
		out << '\n' << heading_ << "Attributes" << body_ << '\n';
		first = false;
	    }

	    out << '\n' << boldon_ << desc -> name << boldoff_;
	    out.inc();
	    out << '\n' << literal_;

	    if(desc -> mode == ATTR_READONLY)
		out << "readonly ";

	    String_var ret = getTypeString(scope, desc -> type);

	    out << "attribute " << ret << ' ' << desc -> name << ';';

            out << body_;

	    out << '\n';

	    comment(desc -> id, out);

	    out.dec();
	}
    }

    //
    // Operation description
    //
    first = true;
    for(i = 0 ; i < seq.length() && status == 0 ; i++)
    {
	if(seq[i].kind == dk_Operation)
	{
	    OperationDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);
	    
	    if(first)
	    {
		out << '\n' << heading_ << "Operations" << body_ << '\n';
		first = false;
	    }

	    out << '\n' << boldon_ << desc -> name << boldoff_;
	    out.inc();
	    out << '\n' << literal_;

	    if(desc -> mode == OP_ONEWAY)
		out << "oneway ";

	    String_var ret = getTypeString(scope, desc -> result);
	    
	    out << ret << ' ' << desc -> name << "(";
	    
	    unsigned int pos = out.getPos();

	    for(j = 0 ; j < desc -> parameters.length() ; j++)
	    {
		GetType tt;
		switch(desc -> parameters[j].mode)
		{
		case PARAM_IN:
		    tt = GetTypeIn;
		    break;

		case PARAM_INOUT:
		    tt = GetTypeInOut;
		    break;

		case PARAM_OUT:
		    tt = GetTypeOut;
		    break;
		}

		String_var s =
		    getTypeString(scope, desc -> parameters[j].type, tt,
				  desc -> parameters[j].name);

		if(j > 0)
		{
		    out << ",\n";

		    //
		    // Indentation has to be done manually here
		    //
		    for(k = 0 ; k < pos; k++)
			out << ' ';
		}

		out << s;
	    }

	    out << ')';

	    if(desc -> contexts.length() > 0)
	    {
		out.inc();
		out << "\ncontext(";

		for(j = 0 ; j < desc -> contexts.length() ; j++)
		{
		    if(j > 0)
			out << ", ";

		    out << "\"" << desc -> contexts[j] << "\"";
		}

		out << ')';
		out.dec();
	    }

	    if(desc -> exceptions.length())
	    {
		out.inc();
		out << "\nraises(";

		for(j = 0 ; j < desc -> exceptions.length() ; j++)
		{
		    if(j)
			out << ",\n       ";

		    String_var str =
			getTypeString(scope, desc -> exceptions[j].type);

		    out << str;
		}
		
		out << ')';

		out.dec();
	    }

	    out << ';';

            out << body_;

	    out << '\n';

	    comment(desc -> id, out);

	    out.dec();
	}
    }

    //
    // Page break for single-file mode
    //
    if(strlen(singleFile_) > 0)
	out << "\\pard \\pagebb";
    else
	out << '\n';
 
    //
    // Interface description
    //
    // (Module descriptions and interface descriptions must be second
    // last for single-file mode)
    //
    for(i = 0 ; i < seq.length() && status == 0 ; i++)
    {
	if(seq[i].kind == dk_Interface)
	{
	    InterfaceDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    //
	    // Get container
	    //
	    Container_var cont =
		Container::_narrow(seq[i].contained_object);
	    assert(!is_nil(cont));

	    ScopedName_var absolute = getAbsolute(desc -> id);

	    if(!strlen(singleFile_))
	    {
		//
		// RTF file for the interface
		//
		String_var path = fixAbsolute(absolute);
		PrettyPrint out(prog_, path, outDir_, bodyFont_, literalFont_,
				titleFont_, headingFont_);
		if(!out.good())
		    return 1;

		//
		// Generate code
		//
		out << title_ << "Interface " << absolute << body_
		    << "\n\ninterface " << boldon_ << desc -> name << boldoff_;
 
		if(desc -> base_interfaces.length() > 0)
		{
		    out << "\ninherits from ";

		    for(j = 0 ; j < desc -> base_interfaces.length() ; j++)
		    {
			ScopedName_var baseAbsolute =
			    getAbsolute(desc -> base_interfaces[j]);

			if(j > 0)
			    out << ", ";
 
			out << baseAbsolute;
		    }
		}

		out << '\n';

		comment(desc -> id, out);
    
		status = genRTF(cont, out);

		out << '\n';
	    }
	    else
	    {
		//
		// Generate code
		//
		out << title_ << "Interface " << absolute << body_
		    << "\n\ninterface " << boldon_ << desc -> name << boldoff_;
 
		if(desc -> base_interfaces.length() > 0)
		{
		    out << "\ninherits from ";

		    for(j = 0 ; j < desc -> base_interfaces.length() ; j++)
		    {
			ScopedName_var baseAbsolute =
			    getAbsolute(desc -> base_interfaces[j]);

			if(j > 0)
			    out << ", ";
 
			out << baseAbsolute;
		    }
		}

		out << '\n';

		comment(desc -> id, out);
    
		status = genRTF(cont, out);
	    }
 	}
    }

    //
    // Module description
    //
    // (Module descriptions and interface descriptions must be last for
    // single-file mode)
    //
    for(i = 0 ; i < seq.length() && status == 0 ; i++)
    {
	if(seq[i].kind == dk_Module)
	{
	    ModuleDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    //
	    // Get container
	    //
	    Container_var cont =
		Container::_narrow(seq[i].contained_object);
	    assert(!is_nil(cont));

	    ScopedName_var absolute = getAbsolute(desc -> id);

	    if(!strlen(singleFile_))
	    {
		//
		// RTF file for the interface
		//
		String_var path = fixAbsolute(absolute);
		PrettyPrint out(prog_, path, outDir_, bodyFont_, literalFont_,
				titleFont_, headingFont_);
		if(!out.good())
		    return 1;

		//
		// Generate code
		//
		out << title_ << "Module " << absolute << body_ << '\n';

		comment(desc -> id, out);

		status = genRTF(cont, out);

		out << '\n';
	    }
	    else
	    {
		//
		// Generate code
		//
		out << title_ << "Module " << absolute << body_ << '\n';

		comment(desc -> id, out);

		status = genRTF(cont, out);
	    }
	}
    }

    return status;
}

// ----------------------------------------------------------------------
// IDL code generator constructor and destructor
// ----------------------------------------------------------------------

IdlRTFGenerator::IdlRTFGenerator(const char* prog,
				 Repository_ptr rep,
				 const IdlCommentSeq& commentSeq,
				 const char* outDir,
				 const char* singleFile,
				 bool noSort,
				 bool ignoreCase,
				 bool withIndex,
				 const char* bodyFont,
				 const char* literalFont,
				 const char* titleFont,
				 const char* headingFont,
				 unsigned int bodySize,
				 unsigned int literalSize,
				 unsigned int titleSize,
				 unsigned int headingSize)
    : GenTextBase(rep), prog_(prog),
      commentSeq_(commentSeq),
      noSort_(noSort),
      ignoreCase_(ignoreCase),
      withIndex_(withIndex),
      outDir_(outDir),
      singleFile_(singleFile),
      bodyFont_(bodyFont),
      literalFont_(literalFont),
      titleFont_(titleFont),
      headingFont_(headingFont)
{
    //
    // Define fonts and sizes
    //
    body_ = string_dup("\\f0\\fs");
    body_ += bodySize;
    body_ += ' ';
    literal_ = string_dup("\\f1\\fs");
    literal_ += literalSize;
    literal_ += ' ';
    title_ = string_dup("\\f2\\fs");
    title_ += titleSize;
    title_ += ' ';
    heading_ = string_dup("\\f3\\fs");
    heading_ += headingSize;
    heading_ += ' ';
}

// ----------------------------------------------------------------------
// Generate RTF code
// ----------------------------------------------------------------------

int
IdlRTFGenerator::gen()
{
    if(!strlen(singleFile_))
    {
	PrettyPrint out(prog_, top_, outDir_, bodyFont_, literalFont_,
			titleFont_, headingFont_);

	out << title_ << "Top-Level Module" << body_ << '\n';
	return genRTF(repository_, out);
    }
    else
    {
	PrettyPrint out(prog_, singleFile_, outDir_, bodyFont_, literalFont_,
			titleFont_, headingFont_);

	out << title_ << "Top-Level Module" << body_ << '\n';
	return genRTF(repository_, out);
    }
}
