// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <Types.h>
#include <Util.h>
#include <GenUtil.h>
#include <GenTextBase.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

using namespace CORBA;

//
// The keywords
//
const char* GenTextBase::keyException_ = "exception ";
const char* GenTextBase::keyMember_ = "member ";
const char* GenTextBase::keyParam_ = "param ";
const char* GenTextBase::keyReturn_ = "return ";
const char* GenTextBase::keySee_ = "see ";
const char* GenTextBase::keySince_ = "since ";
const char* GenTextBase::keyAuthor_ = "author ";
const char* GenTextBase::keyVersion_ = "version ";

// ----------------------------------------------------------------------
// Get absolute name
// ----------------------------------------------------------------------

char*
GenTextBase::getAbsolute(Contained_ptr contained)
{
    return contained -> absolute_name();
}

char*
GenTextBase::getAbsolute(const char* id)
{
    for(ULong i = 0 ; i < absSeq_.length() ; i += 2)
    {
	if(strcmp(absSeq_[i], id) == 0)
	    return string_dup(absSeq_[i + 1]);
    }
    
    Contained_var contained = repository_ -> lookup_id(id);
    assert(!is_nil(contained));
    
    ULong len = absSeq_.length();
    absSeq_.length(len + 2);
    absSeq_[len] = id;
    absSeq_[len + 1] = getAbsolute(contained);
    return string_dup(absSeq_[len + 1]);
}

// ----------------------------------------------------------------------
// Fix absolute path name, i.e. replace "::" by "."
// ----------------------------------------------------------------------

char*
GenTextBase::fixAbsolute(const char* abs)
{
    String_var str = abs;
    String_var result;

    char* start = str.inout();
    char* end;

    while(true)
    {
	end = strchr(start, ':');

	if(end)
	{
	    *end = '\0';
	    result += start;
	    result += '.';
	    start = end + 2;
	}
	else
	{
	    result += start;
	    break;
	}
    }

    return result._retn();
}

// ----------------------------------------------------------------------
// Get scope
// ----------------------------------------------------------------------

void
GenTextBase::getScope(Container_ptr container, char*& scope)
{
    Contained_var contained = Contained::_narrow(container);
    if(is_nil(contained))
	scope = string_dup("");
    else
    {
	ScopedName_var s = contained -> absolute_name();
	scope = string_alloc(strlen(s) + 2);
	strcpy(scope, s);
	strcat(scope, "::");
    }
}

// ----------------------------------------------------------------------
// Print union labels
// ----------------------------------------------------------------------

void
GenTextBase::printUnionLabels(const IdlUnionMemberInfo& info,
			      IdlPrettyPrint& out, TypeCode_ptr type)
{
    if(!is_nil(info.type))
    {
	out.dec();

	if(info.isDefault)
	    out << "\ndefault:";
	else
	{
	    ULong i;
	    
	    for(i = 0 ; i < info.pLabels.length() ; i++)
	    {
		out << "\ncase ";

		switch(type -> kind())
		{
		case tk_char:
		{
		    char label = (char)info.pLabels[i];
		    String_var s = string_dup(&label);
		    IdlAddEscapes(s.inout(), false);

		    out << '\'' << s << '\'';

		    break;
		}

		case tk_boolean:
		{
		    if(info.pLabels[i])
			out << "TRUE";
		    else
			out << "FALSE";

		    break;
		}

		case tk_enum:
		{
		    Identifier_var memberName =
			type -> member_name(info.pLabels[i]);
		    out << memberName;

		    break;
		}

		default:
		    out << info.pLabels[i];
		}

		out << ':';
	    }
	    
	    for(i = 0 ; i < info.nLabels.length() ; i++)
		out << "\ncase -" << info.nLabels[i] << ':';
	}

	out.inc();
    }
}

// ----------------------------------------------------------------------
// Tokenize
// ----------------------------------------------------------------------

char*
GenTextBase::tokenize(char* s1)
{
    static char* str;

    if(s1)
	str = strchr(s1, '@');

    while(str)
    {
	char* ret = str + 1;

	//
	// Create a new token for reserved keywords only
	//
	for(ULong i = 0 ; i < keywordSeq_.length() ; i++)
	{
	    if(isKeyword(ret))
	    {
		str = strchr(ret, '@');
		while(str)
		{
		    if(isKeyword(str + 1))
		    {
			*str = 0;
			return ret;
		    } 

		    str = strchr(str + 1, '@');
		}

		return ret;
	    }
	}

	str = strchr(ret, '@');
    }

    return 0;
}

bool
GenTextBase::isKeyword(const char* str)
{
    for(ULong i = 0 ; i < keywordSeq_.length() ; i++)
    {
	if(strncmp(str, keywordSeq_[i], strlen(keywordSeq_[i])) == 0)
	    return true;
    }

    return false;
}

// ----------------------------------------------------------------------
// Constructor and destructor
// ----------------------------------------------------------------------

GenTextBase::GenTextBase(Repository_ptr rep)
    : repository_(Repository::_duplicate(rep))
{
    keywordSeq_.append(keyException_);
    keywordSeq_.append(keyMember_);
    keywordSeq_.append(keyParam_);
    keywordSeq_.append(keyReturn_);
    keywordSeq_.append(keySee_);
    keywordSeq_.append(keySince_);
    keywordSeq_.append(keyAuthor_);
    keywordSeq_.append(keyVersion_);
}
