// **********************************************************************
//
// Copyright (c) 1997
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <Types.h>
#include <GenUtil.h>

#include <stdio.h>
#include <ctype.h>

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Pretty-Print private and protected members
// ----------------------------------------------------------------------

void
IdlPrettyPrint::nl()
{
    out_ << '\n';
    
    for(unsigned int i = 0 ; i < indent_ ; i++)
	out_ << ' ';
    
    pos_ = indent_;
}

// ----------------------------------------------------------------------
// Pretty-Print public members
// ----------------------------------------------------------------------

void
IdlPrettyPrint::start()
{
    *this << "\n{";
    inc();
}

void
IdlPrettyPrint::end()
{
    dec();
    *this << "\n}";
}

void
IdlPrettyPrint::sep()
{
    if(ignoreNextSep_)
	ignoreNextSep_ = false;
    else
	out_ << '\n';
}

void
IdlPrettyPrint::ignoreNextSep(bool ignore)
{
    ignoreNextSep_ = ignore;
}

void
IdlPrettyPrint::printChar(char c)
{
    if(c == '\r')
    {
	// Do nothing
    }
    else if(c == '\n')
    {
	nl();
    }
    else if(c == '\t')
    {
	while(pos_ % 8)
	{
	    out_ << ' ';
	    pos_++;
	}
    }
    else
    {
	out_ << c;
	pos_++;
    }
}

void
IdlPrettyPrint::printWChar(wchar_t c)
{
    if(c == '\r')
    {
	// Do nothing
    }
    else if(c == '\n')
    {
	nl();
    }
    else if(c == '\t')
    {
	while(pos_ % 8)
	{
	    out_ << ' ';
	    pos_++;
	}
    }
    else
    {
	out_ << c;
	pos_++;
    }
}

void
IdlPrettyPrint::printString(const char* s)
{
    if(!s)
	return;

    while(*s)
    {
	printChar(*s);
	s++;
    }
}

void
IdlPrettyPrint::printWString(const wchar_t* s)
{
    if(!s)
	return;

    while(*s)
    {
	printWChar(*s);
	s++;
    }
}

// ----------------------------------------------------------------------
// Pretty-Print output operators
// ----------------------------------------------------------------------

IdlPrettyPrint&
operator<<(IdlPrettyPrint& pp, char c)
{
    pp.printChar(c);
    return pp;
}

#ifndef HAVE_WIN32_WCHAR_TYPE
IdlPrettyPrint&
operator<<(IdlPrettyPrint& pp, wchar_t c)
{
    pp.printWChar(c);
    return pp;
}
#endif

IdlPrettyPrint&
operator<<(IdlPrettyPrint& pp, const char* s)
{
    pp.printString(s);
    return pp;
}

IdlPrettyPrint&
operator<<(IdlPrettyPrint& pp, const wchar_t* s)
{
    pp.printWString(s);
    return pp;
}

IdlPrettyPrint&
operator<<(IdlPrettyPrint& pp, short val)
{
    return pp << (long)val;
}

IdlPrettyPrint&
operator<<(IdlPrettyPrint& pp, int val)
{
    return pp << (long)val;
}

IdlPrettyPrint&
operator<<(IdlPrettyPrint& pp, long val)
{
    char s[1024];
    sprintf(s, "%ld", val);
    pp.printString(s);
    return pp;
}

IdlPrettyPrint&
operator<<(IdlPrettyPrint& pp, LongLong val)
{
    char s[1024];
    sprintf(s, "%lld", val);
    pp.printString(s);
    return pp;
}

IdlPrettyPrint&
operator<<(IdlPrettyPrint& pp, unsigned short val)
{
    return pp << (unsigned long)val;
}

IdlPrettyPrint&
operator<<(IdlPrettyPrint& pp, unsigned int val)
{
    return pp << (unsigned long)val;
}

IdlPrettyPrint&
operator<<(IdlPrettyPrint& pp, unsigned long val)
{
    char s[1024];
    sprintf(s, "%ld", val);
    pp.printString(s);
    return pp;
}

IdlPrettyPrint&
operator<<(IdlPrettyPrint& pp, ULongLong val)
{
    char s[1024];
    sprintf(s, "%lld", val);
    pp.printString(s);
    return pp;
}

IdlPrettyPrint&
operator<<(IdlPrettyPrint& pp, float val)
{
    return pp << (double)val;
}

IdlPrettyPrint&
operator<<(IdlPrettyPrint& pp, double val)
{
    char s[1024];
    sprintf(s, "%.100g", val);
    pp.printString(s);
    return pp;
}

IdlPrettyPrint&
operator<<(IdlPrettyPrint& pp, LongDouble val)
{
#if SIZEOF_LONG_DOUBLE == 0
    assert(SIZEOF_LONG_DOUBLE != 0);
#else
    char s[1024];
    sprintf(s, "%.100Lg", val);
    pp.printString(s);
#endif
    return pp;
}

// ----------------------------------------------------------------------
// Remove scope from absolute name
// ----------------------------------------------------------------------

void
IdlRemoveScope(const char* scope, char*& absolute)
{
    assert_nca(scope, NCANullString);
    assert_nca(absolute, NCANullString);

    if(strlen(scope) > 0)
    {
	assert(scope[strlen(scope) - 1] == ':');
	assert(scope[strlen(scope) - 2] == ':');
	
	if(strncmp(scope, absolute, strlen(scope)) == 0)
	{
	    char* a = string_dup(absolute + strlen(scope));
	    string_free(absolute);
	    absolute = a;
	}
	else if(strchr(scope, ':') != scope + strlen(scope) - 2)
	{
	    String_var s = scope;
	    s[strlen(s) - 2] = '\0';
	    char* s2 = (char*)strrchr(s, ':');
	    if(s2)
	    {
		assert(*(s2 - 1) == ':');
		*(s2 + 1) = '\0';
		IdlRemoveScope(s, absolute);
	    }
	}
    }
}

// ----------------------------------------------------------------------
// Get original type, i.e. translate aliases to the real type
// ----------------------------------------------------------------------

IDLType_ptr
IdlGetOrigIDLType(IDLType_ptr idlType)
{
    IDLType_var origType = IDLType::_duplicate(idlType);
    
    while(true)
    {
	AliasDef_var alias = AliasDef::_narrow(origType);
	
	if(is_nil(alias))
	    break;
	else
	    origType = alias -> original_type_def();
    }
    
    return IDLType::_duplicate(origType);
}

// ----------------------------------------------------------------------
// Check for variable size type
// ----------------------------------------------------------------------

bool
IdlIsVariable(TypeCode_ptr t)
{
    TypeCode_var type = GetOrigType(t);

    switch(type -> kind())
    {
    case tk_null:
    case tk_void:
    case tk_short:
    case tk_long:
    case tk_ushort:
    case tk_ulong:
    case tk_longlong:
    case tk_ulonglong:
    case tk_float:
    case tk_double:
    case tk_longdouble:
    case tk_fixed:
    case tk_boolean:
    case tk_char:
    case tk_wchar:
    case tk_octet:
    case tk_enum:
	return false;

    case tk_any:
    case tk_TypeCode:
    case tk_Principal:
    case tk_objref:
    case tk_string:
    case tk_wstring:
    case tk_sequence:
    case tk_except:
    case tk_value:
    case tk_value_box:
    case tk_abstract_interface:
	return true;

    case tk_struct:
    case tk_union:
    {
	for(ULong i = 0 ; i < type -> member_count() ; i++)
	{
	    TypeCode_var memberType = type -> member_type(i);
	    if(IdlIsVariable(memberType))
		return true;
	}
	
	return false;
    }

    case tk_array:
    {
	TypeCode_var contentType = type -> content_type();

	if(IdlIsVariable(contentType))
	    return true;
	else
	    return false;
    }

    case tk_alias: // No alias possible here
    case tk_native: // No native possible here
	break;     // this is to keep the compiler happy
    }

    return false; // This is to keep the compiler happy
}

// ----------------------------------------------------------------------
// Get TypeCode converted to ASCII
// Output TypeCode converted to ASCII
// ----------------------------------------------------------------------

char*
IdlGetAsciiTypeCode(TypeCode_ptr tc)
{
    CountStream count;
    count.count_boolean(); // Endian
    count.count_TypeCode(tc);

    Buffer buf(count.count());
    OutputStream out(buf, 0);
    out.write_boolean(Endian);
    out.write_TypeCode(tc);

    char* typeCodeStr = OctetsToAscii(buf.data, buf.len);

    return typeCodeStr;
}

void
IdlOutputAsciiTypeCode(TypeCode_ptr tc, IdlPrettyPrint& out)
{
    String_var typeCodeStr = IdlGetAsciiTypeCode(tc);

    unsigned int indent = out.getIndent();
    out.setIndent(0);

    ULong j = 0;

    do
    {
	out << "\n\"";
	for(ULong i = 0 ; i < 77 && typeCodeStr[j] != '\0' ; i++)
	    out << typeCodeStr[j++];
	out << '"';
    }
    while(typeCodeStr[j] != '\0');

    out << '\n';
    
    out.setIndent(indent);
}

// ----------------------------------------------------------------------
// Convert path name to identifier
// ----------------------------------------------------------------------

char*
IdlPathToIdent(const char* s)
{
    String_var ident;
    
    for(ULong i = 0 ; i < strlen(s) ; i++)
    {
	char c = s[i];
	
	if(c == '.' || c == '/')
	    c = '_';
	
	if(isalnum(c) || c == '_')
	    ident += c;
    }
    
    return ident._retn();
}

// ----------------------------------------------------------------------
// Check if union has an implicit default
// ----------------------------------------------------------------------

bool
IdlImplicitUnionDefault(TypeCode_ptr type)
{
    TypeCode_var origType = GetOrigType(type);
    assert(origType -> kind() == tk_union);

    bool result = false;

    if(origType -> default_index() == -1)
    {
	result = true;
	
	TypeCode_var tc = origType -> discriminator_type();
	tc = GetOrigType(tc);
	
	if(tc -> kind() == tk_char)
	    if(origType -> member_count() == 256)
		result = false;
	
	if(tc -> kind() == tk_boolean)
	    if(origType -> member_count() == 2)
		result = false;
	
	if(tc -> kind() == tk_enum)
	    if(origType -> member_count() == tc -> member_count())
		result = false;
    }

    return result;
}

// ----------------------------------------------------------------------
// Get union info
// ----------------------------------------------------------------------

IdlUnionMemberInfoSeq*
IdlGetUnionInfo(TypeCode_ptr type)
{
    TypeCode_var origType = GetOrigType(type);
    assert(origType -> kind() == tk_union);
    
    TypeCode_var discType = origType -> discriminator_type();
    TypeCode_var origDiscType = GetOrigType(discType);

    IdlUnionMemberInfoSeq_var seq =
	new IdlUnionMemberInfoSeq(origType -> member_count());
    
    bool hasDefault = false;
    ULong defaultIndex;

    ULong i, j, k;
    
    for(i = 0 ; i < origType -> member_count() ; i++)
    {
	Identifier_var name = origType -> member_name(i);
	
	if(i > 0)
	{
	    Identifier_var s = origType -> member_name(i - 1);
	    
	    if(strcmp(s, name) != 0)
	    {
		ULong len = seq -> length();
		seq -> length(len + 1);
		seq[len].name = name;
		seq[len].type = origType -> member_type(i);
	    }
	    else
	    {
		assert(seq[seq -> length() - 1].nLabels.length() > 0 ||
		       seq[seq -> length() - 1].pLabels.length() > 0);
	    }
	}
	else
	{
	    ULong len = seq -> length();
	    seq -> length(len + 1);
	    seq[len].name = name;
	    seq[len].type = origType -> member_type(i);
	}

	IdlUnionMemberInfo& info = seq[seq -> length() - 1];

	Any_var label = origType -> member_label(i);
	TypeCode_var labelType = label -> type();
	TypeCode_var origLabelType = GetOrigType(labelType);

	switch(origLabelType -> kind())
	{
	case tk_short:
	{
	    Short val;
	    Boolean b = label >>= val;
	    assert(b);
	    if(val >= 0)
		info.pLabels.append((ULong)val);
	    else
		info.nLabels.append((ULong)-val);
	    info.isDefault = false;
	    break;
	}

	case tk_ushort:
	{
	    UShort val;
	    Boolean b = label >>= val;
	    assert(b);
	    info.pLabels.append((ULong)val);
	    info.isDefault = false;
	    break;
	}

	case tk_long:
	{
	    Long val;
	    Boolean b = label >>= val;
	    assert(b);
	    if(val >= 0)
		info.pLabels.append((ULong)val);
	    else
		info.nLabels.append((ULong)-val);
	    info.isDefault = false;
	    break;
	}

	case tk_ulong:
	{
	    ULong val;
	    Boolean b = label >>= val;
	    assert(b);
	    info.pLabels.append((ULong)val);
	    info.isDefault = false;
	    break;
	}

	case tk_enum:
	{
	    ULong val = *(ULong*)(label -> value());
	    info.pLabels.append((ULong)val);
	    info.isDefault = false;
	    break;
	}

	case tk_char:
	{
	    Char val;
	    Boolean b = label >>= Any::to_char(val);
	    assert(b);
	    info.pLabels.append((ULong)val);
	    info.isDefault = false;
	    break;
	}

	case tk_wchar:
	{
	    WChar val;
	    Boolean b = label >>= Any::to_wchar(val);
	    assert(b);
	    info.pLabels.append((ULong)val);
	    info.isDefault = false;
	    break;
	}

	case tk_boolean:
	{
	    Boolean val;
	    Boolean b = label >>= Any::to_boolean(val);
	    assert(b);
	    info.pLabels.append((ULong)val);
	    info.isDefault = false;
	    break;
	}

	case tk_octet:
	    assert(hasDefault == false);
	    hasDefault = true;
	    defaultIndex = seq -> length() - 1;
	    info.isDefault = true;
	    break;

	default:
	    assert(false);
	    break;
	}
    }

    ULong min, max;
    
    switch(origDiscType -> kind())
    {
    case tk_short:
	min = 32768;
	max = 32767;
	break;
	
    case tk_ushort:
	min = 0;
	max = 65535;
	break;
	
    case tk_long:
	min = 2147483648U;
	max = 2147483647U;
	break;
	
    case tk_ulong:
	min = 0;
	max = 4294967294U;
	//max = 4294967295U;
	break;
	
    case tk_enum:
	min = 0;
	max = origDiscType -> member_count() - 1;
	break;
	
    case tk_char:
	min = 0;
	max = 255;
	break;
	
    case tk_boolean:
	min = 0;
	max = 1;
	break;
	
    default:
	assert(false);
	break;
    }

    bool hasPDefault = false;
    bool hasNDefault = false;
    ULong defPLabel;
    ULong defNLabel;
    
    for(i = 0 ; i <= max ; i++)
    {
	bool ok = true;
	
	for(j = 0 ; ok && j < seq -> length() ; j++)
	    for(k = 0 ; ok && k < seq[j].pLabels.length() ; k++)
		if(seq[j].pLabels[k] == i)
		    ok = false;
	
	if(ok)
	{
	    hasPDefault = true;
	    defPLabel = i;
	    break;
	}
    }
    
    for(i = 1 ; i <= min ; i++)
    {
	bool ok = true;
	
	for(j = 0 ; ok && j < seq -> length() ; j++)
	    for(k = 0 ; ok && k < seq[j].nLabels.length() ; k++)
		if(seq[j].nLabels[k] == i)
		    ok = false;
	
	if(ok)
	{
	    hasNDefault = true;
	    defNLabel = i;
	    break;
	}
    }

    if(hasPDefault || hasNDefault)
    {
	if(!hasDefault)
	{
	    ULong len = seq -> length();
	    seq -> length(len + 1);
	    seq[len].name = string_dup("");
	    seq[len].isDefault = true;

	    hasDefault = true;
	    defaultIndex = len;
	}

	// This is legal:
	//
	//     union uni switch(short)
        //     {
	//     case 1 :
        //     default : long value;
	//     };
	//
        // Therefore the following asserts are wrong:
	//
	//assert(seq[defaultIndex].pLabels.length() == 0);
	//assert(seq[defaultIndex].nLabels.length() == 0);
	assert(seq[defaultIndex].isDefault);
	
	if(hasPDefault)
	    seq[defaultIndex].pLabels.append(defPLabel);
	
	if(hasNDefault)
	    seq[defaultIndex].nLabels.append(defNLabel);
    }
    else
    {
	if(hasDefault)
	{
	    seq -> remove(defaultIndex);
	    hasDefault = false;
	};
    };

    return seq._retn();
};

// ----------------------------------------------------------------------
// Get all base interfaces
// ----------------------------------------------------------------------

void
IdlGetAllBaseInterfaces(InterfaceDef_ptr interf,
			InterfaceDefSeq& seq)
{
    ULong i, j;

    InterfaceDefSeq_var baseSeq = interf -> base_interfaces();
    for(i = 0 ; i < baseSeq -> length() ; i++)
    {
	RepositoryId_var baseId = baseSeq[i] -> id(); 
	for(j = 0 ; j < seq.length() ; j++)
	{
	    RepositoryId_var id = seq[j] -> id(); 
	    if(strcmp(baseId, id) == 0)
		break;
	}

	if(j == seq.length())
	{
	    seq.length(seq.length() + 1);
	    seq[seq.length() - 1] = InterfaceDef::_duplicate(baseSeq[i]);
	    IdlGetAllBaseInterfaces(baseSeq[i], seq);
	}
    }
}

// ----------------------------------------------------------------------
// Get information about base interfaces
// ----------------------------------------------------------------------

void
IdlGetBaseInterfaces(InterfaceDef_ptr interf,
		     IdlBaseInterfaceInfoSeq& seq, bool all)
{
    ULong i, j;

    InterfaceDefSeq_var baseSeq = interf -> base_interfaces();
    for(i = 0 ; i < baseSeq -> length() ; i++)
    {
	RepositoryId_var baseId = baseSeq[i] -> id(); 
	for(j = 0 ; j < seq.length() ; j++)
	{
	    if(strcmp(baseId, seq[j].id) == 0)
		break;
	}

	if(j == seq.length())
	{
	    seq.length(seq.length() + 1);
	    seq[seq.length() - 1].intf =
                InterfaceDef::_duplicate(baseSeq[i]);
	    seq[seq.length() - 1].id = baseId;
	    seq[seq.length() - 1].is_abstract = baseSeq[i] -> is_abstract();
	    if(all)
		IdlGetBaseInterfaces(baseSeq[i], seq, all);
	}
    }
}

// ----------------------------------------------------------------------
// Get all base values
// ----------------------------------------------------------------------

void
IdlGetAllBaseValues(ValueDef_ptr value, ValueDefSeq& seq)
{
    ULong i, j;

    ValueDefSeq_var baseSeq = value -> abstract_base_values();
    for(i = 0 ; i < baseSeq -> length() ; i++)
    {
	RepositoryId_var baseId = baseSeq[i] -> id(); 
	for(j = 0 ; j < seq.length() ; j++)
	{
	    RepositoryId_var id = seq[j] -> id(); 
	    if(strcmp(baseId, id) == 0)
		break;
	}

	if(j == seq.length())
	{
	    seq.length(seq.length() + 1);
	    seq[seq.length() - 1] = ValueDef::_duplicate(baseSeq[i]);
	    IdlGetAllBaseValues(baseSeq[i], seq);
	}
    }
}

// ----------------------------------------------------------------------
// Get all truncatable base values
// ----------------------------------------------------------------------

void
IdlGetAllTruncatableBaseValues(ValueDef_ptr value,
                               ValueDefSeq& seq)
{
    if(value -> is_truncatable())
    {
        ValueDef_var base = value -> base_value();
        assert(!is_nil(base));
        seq.length(seq.length() + 1);
        seq[seq.length() - 1] = ValueDef::_duplicate(base);
        IdlGetAllTruncatableBaseValues(base, seq);
    }
}

// ----------------------------------------------------------------------
// Get all value members
// ----------------------------------------------------------------------

void
IdlGetAllValueMembers(CORBA::ValueDef_ptr value, CORBA::ValueMemberSeq& seq)
{
    ValueDef_var base = value -> base_value();
    if(!is_nil(base))
        IdlGetAllValueMembers(base, seq);

    Container::DescriptionSeq_var contents =
        value -> describe_contents(dk_ValueMember, TRUE, -1);

    for(ULong i = 0 ; i < contents -> length() ; i++)
    {
        ValueMember* vm;
        Boolean b = contents[i].value >>= vm;
        assert(b);
        seq.append(*vm);
    }
}

// ----------------------------------------------------------------------
// Get all method names
// ----------------------------------------------------------------------

IdlStringSeq*
IdlGetMethodNames(const ContainedSeq& attributes,
		  const ContainedSeq& operations)
{
    IdlStringSeq_var names = new IdlStringSeq;
    ULong i, j;

    //
    // Add each attribute to the method table
    //
    for(i = 0 ; i < attributes.length() ; i++)
    {
        Contained::Description_var contDesc =
	    attributes[i] -> describe();
	AttributeDescription* desc;
	Boolean b = (contDesc -> value) >>= desc;
	assert(b);

	String_var name = string_dup("_get_");
        name += desc -> name;
	names -> append(name._retn());

        if(desc -> mode == ATTR_NORMAL)
        {
	    String_var name = string_dup("_set_");
	    name += desc -> name;
	    names -> append(name._retn());
        }
    }

    //
    // Add each operation to the method table
    //
    for(i = 0 ; i < operations.length() ; i++)
    {
        Contained::Description_var contDesc =
	    operations[i] -> describe();
	OperationDescription* desc;
	Boolean b = (contDesc -> value) >>= desc;
	assert(b);

	names -> append(desc -> name);
    }

    //
    // Sort names (bubblesort)
    //
    for(i = 0 ; i < names -> length() ; i++)
    {
	for(j = i ; j < names -> length() ; j++)
	{
	    if(strcmp(names[i], names[j]) > 0)
	    {
		String_var name = names[i]._retn();
		names[i] = names[j]._retn();
		names[j] = name._retn();
	    }
	}
    }

    return names._retn();
}

