// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>

#include <Util.h>
#include <Types.h>
#include <IFR2_impl.h>
#include <IFR3_impl.h>
#include <IFR4_impl.h>

#include <ctype.h>

using namespace CORBA;
using namespace OB;
using namespace PortableServer;

// ----------------------------------------------------------------------
// Contained_impl member implementations
// ----------------------------------------------------------------------

Contained_impl::Contained_impl(ORB_ptr orb, Container_ptr p)
    : IRObject_impl(orb),
      definedIn_(Container::_duplicate(p))
{
}

Contained_impl::~Contained_impl()
{
}

void
Contained_impl::releaseInternal()
{
    definedIn_ = Container::_nil();
    IRObject_impl::releaseInternal();
}

void
Contained_impl::destroy()
    throw(SystemException)
{
    Contained_var self = _this();

    POA_var poa = _default_POA();
    ServantBase_var servant = poa -> reference_to_servant(definedIn_);
    Container_impl* container = dynamic_cast<Container_impl*>(servant.in());
    assert(container);
    container -> remove(self);

    IRObject_impl::destroy();
}

char*
Contained_impl::id()
    throw(SystemException)
{
    return string_dup(id_);
}

void
Contained_impl::id(const char* s)
    throw(SystemException)
{
    if(id_.in() && strcmp(id_, s) == 0)
	return;

    Repository_var rep = containing_repository();
    Contained_var con = rep -> lookup_id(s);

    if(!is_nil(con))
    {
	//
        // ID already exists
	//
	throw BAD_PARAM(2, COMPLETED_NO);
    }
    
    id_ = s;
}

char*
Contained_impl::name()
    throw(SystemException)
{
    return string_dup(name_);
}

void
Contained_impl::name(const char* s)
    throw(SystemException)
{
    if(name_.in() && strcmp(name_, s) == 0)
	return;

    //
    // Find out whether name already exists. Using case-insensitive
    // comparison.
    //
    ULong i, j;

    // Exclude inherited names
    ContainedSeq_var seq = definedIn_ -> contents(dk_all, true);

    String_var s1 = s;
    for(j = 0 ; j < strlen(s1) ; j++)
	if(isupper(s1[j]))
	    s1[j] = tolower(s1[j]);

    for(i = 0 ; i < seq -> length() ; i++)
    {
	String_var s2 = seq[i] -> name();
	for(j = 0 ; j < strlen(s2) ; j++)
	    if(isupper(s2[j]))
		s2[j] = tolower(s2[j]);

	if(strcmp(s1, s2) == 0)
	{
	    //
	    // Name already exists
	    //
	    throw BAD_PARAM(3, COMPLETED_NO);
	}
    }
    
    name_ = s;
}

char*
Contained_impl::version()
    throw(SystemException)
{
    return string_dup(version_);
}

void
Contained_impl::version(const char* s)
    throw(SystemException)
{
    version_ = s;
}

Container_ptr
Contained_impl::defined_in()
    throw(SystemException)
{
    return Container::_duplicate(definedIn_);
}

char*
Contained_impl::absolute_name()
    throw(SystemException)
{
    Contained_var contained = CORBA::Contained::_narrow(definedIn_);

    if(!is_nil(contained))
    {
	String_var s = contained -> absolute_name();
	char* result = string_alloc(strlen(s) + 2 + strlen(name_));
	strcpy(result, s);
	strcat(result, "::");
	strcat(result, name_);
	return result;
    }
    else
	return string_dup(name_);
}

Repository_ptr
Contained_impl::containing_repository()
    throw(SystemException)
{
    Repository_var repository = CORBA::Repository::_narrow(definedIn_);
    
    if(is_nil(repository))
    {
	Contained_var contained = CORBA::Contained::_narrow(definedIn_);
	assert(!is_nil(contained));
	
	return contained -> containing_repository();
    }
    else
	return Repository::_duplicate(repository);
}

void
Contained_impl::move(Container_ptr new_container,
		     const char* identifier,
		     const char* new_version)
    throw(SystemException)
{
    //
    // Get self-reference
    //
    Contained_var self = _this();

    //
    // Remove from old container
    //
    POA_var poa = _default_POA();
    ServantBase_var servant = poa -> reference_to_servant(definedIn_);
    Container_impl* oldCont = dynamic_cast<Container_impl*>(servant.in());
    assert(oldCont);
    oldCont -> remove(self);

    //
    // Add to new container
    //
    RepositoryId_var oldId = id();
    Identifier_var oldName = name();
    VersionSpec_var oldVersion = version();
    try
    {
	definedIn_ = Container::_duplicate(new_container);
	name(identifier);
	version(new_version);

        POA_var poa = _default_POA();
        ServantBase_var servant = poa -> reference_to_servant(new_container);
        Container_impl* newCont = dynamic_cast<Container_impl*>(servant.in());
	assert(newCont);
	newCont -> add(self);
    }
    catch(...)
    {
	id(oldId);
	name(oldName);
	version(oldVersion);
	oldCont -> add(self);

	throw;
    }
}

// ----------------------------------------------------------------------
// Container_impl member implementations
// ----------------------------------------------------------------------

void
Container_impl::add(Contained_ptr p)
{
    contents_.append(Contained::_duplicate(p));
}

void
Container_impl::remove(Contained_ptr p)
{
    Identifier_var n1 = p -> name();
    Long i;
    for(i = contents_.length() - 1 ; i >= 0 ; i--)
    {
	Identifier_var n2 = contents_[i] -> name();
	if(strcmp(n1, n2) == 0)
	{
	    contents_.remove(i);
	    break;
	}
    }

    assert(i > -1);
}

/*static*/ void
Container_impl::merge(ContainedSeq* p, const ContainedSeq& seq)
{
    ULong i, j;

    for(i = 0 ; i < seq.length() ; i++)
    {
        RepositoryId_var id2 = seq[i] -> id();

        for(j = 0 ; j < p -> length() ; j++)
        {
            RepositoryId_var id = (*p)[j] -> id();

            if(strcmp(id, id2) == 0)
                break;
        }

        if(j == p -> length())
            p -> append(seq[i]);
    }
}

Container_impl::Container_impl(ORB_ptr orb)
    : IRObject_impl(orb)
{
}

Container_impl::~Container_impl()
{
}

void
Container_impl::releaseInternal()
{
    for(ULong i = 0 ; i < contents_.length() ; i++)
    {
        POA_var poa = _default_POA();
        ServantBase_var servant = poa -> reference_to_servant(contents_[i]);
        Contained_impl* p = dynamic_cast<Contained_impl*>(servant.in());
	if(p)
	    p -> releaseInternal();
    }
    
    contents_.length(0);
    IRObject_impl::releaseInternal();
}

Contained_ptr
Container_impl::lookup(const char* search_name)
    throw(SystemException)
{
    if(strncmp(search_name, "::", 2) == 0) // Absolute name, begining with "::"
    {
	Container_var self = _this();
	Contained_var contained = CORBA::Contained::_narrow(self);

	//
        // Must be contained, since it cannot
	// be a repository (the repository overwrites lookup())
	//
	assert(!is_nil(contained));

	Repository_var repository = contained -> containing_repository();
	return repository -> lookup(search_name);
    }
    else // Relative name, not begining with "::"
    {
	const char* s2 = strstr(search_name, "::");
	String_var s1;

	if(s2)
	{
	    ULong len = s2 - search_name;
	    s1 = string_alloc(len);
	    strncpy(s1.inout(), search_name, len);
	    s1[len] = '\0';
	    s2 += 2;
	}
	else
	    s1 = search_name;

	for(ULong i = 0 ; i < contents_.length(); i++)
	{
	    String_var n = contents_[i] -> name();
	    if(strcmp(n, s1) == 0)
	    {
		if(s2)
		{
		    Container_var container =
			CORBA::Container::_narrow(contents_[i]);
		    if(is_nil(container))
			return Contained::_nil();
		    else
			return container -> lookup(s2);
		}
		else
		    return Contained::_duplicate(contents_[i]);
	    }
	}

	return Contained::_nil();
    }
}

ContainedSeq*
Container_impl::contents(DefinitionKind limit_type,
			 Boolean /*exclude_inherited*/)
    throw(SystemException)
{
    ContainedSeq* s = new ContainedSeq;

    for(ULong i = 0 ; i < contents_.length() ; i++)
	if(limit_type == dk_all ||
	   limit_type == contents_[i] -> def_kind())
	    s -> append(contents_[i]);
    
    return s;
}

ContainedSeq*
Container_impl::lookup_name(const char* search_name,
			    Long levels_to_search,
			    DefinitionKind limit_type,
			    Boolean exclude_inherited)
    throw(SystemException)
{
    ULong i, j, k;

    ContainedSeq* s = new ContainedSeq;

    if(levels_to_search != 0)
    {
	levels_to_search--;

	for(i = 0 ; i < contents_.length() ; i++)
	{
	    if(limit_type == dk_all ||
	       limit_type == contents_[i] -> def_kind())
	    {
		Identifier_var name = contents_[i] -> name();
		if(strcmp(search_name, name) == 0)
		    s -> append(contents_[i]);

		if(levels_to_search != 0)
		{
		    Container_var cont =
                        CORBA::Container::_narrow(contents_[i]);
		    if(!is_nil(cont))
		    {
			ContainedSeq_var s2 =
			    cont -> lookup_name(search_name,
						levels_to_search,
						limit_type,
						exclude_inherited);
			
			for(j = 0 ; j < s2 -> length() ; j++)
			{
			    RepositoryId_var id2 = s2[j] -> id(); 
			    
			    for(k = 0 ; k < s -> length() ; k++)
			    {
				RepositoryId_var id = (*s)[k] -> id(); 
				
				if(strcmp(id, id2) == 0)
				    break;
			    }
			    
			    if(k == s -> length())
				s -> append(s2[j]);
			}
		    }
		}
	    }
	}
    }
		
    return s;
}

CORBA::Container::DescriptionSeq*
Container_impl::describe_contents(DefinitionKind limit_type,
				  Boolean exclude_inherited,
				  Long max_returned_objs)
    throw(SystemException)
{
    ContainedSeq_var cs = contents(limit_type, exclude_inherited);

    ULong len = cs -> length();

    if(max_returned_objs >= 0 && len > (ULong)max_returned_objs)
	len = (ULong)max_returned_objs;

    CORBA::Container::DescriptionSeq* s =
        new CORBA::Container::DescriptionSeq(len);

    for(ULong i = 0 ; i < len ; i++)
    {
	if(limit_type == dk_all ||
	   limit_type == cs[i] -> def_kind())
	{
	    CORBA::Contained::Description_var description =
                cs[i] -> describe();
	    
	    CORBA::Container::Description desc;
	    desc.contained_object = cs[i];
	    desc.kind = description -> kind;
	    desc.value = description -> value;
	    
	    s -> append(desc);
	}
    }
    
    return s;
}

ModuleDef_ptr
Container_impl::create_module(const char* id,
			      const char* name,
			      const char* version)
    throw(SystemException)
{
    Container_var self = _this();
    ModuleDef_impl* p = new ModuleDef_impl(orb_, self);

    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
    }
    catch(...)
    {
        p -> _remove_ref();
        throw;
    }

    contents_.append(p -> _this());
    return p -> _this();
}

ConstantDef_ptr
Container_impl::create_constant(const char* id,
			        const char* name,
			        const char* version,
			        IDLType_ptr type,
			        const Any& value)
    throw(SystemException)
{
    Container_var self = _this();
    ConstantDef_impl* p = new ConstantDef_impl(orb_, self);
    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
        p -> type_def(type);
        p -> value(value);
    }
    catch(...)
    {
        p -> _remove_ref();
        throw;
    }
    contents_.append(p -> _this());
    return p -> _this();
}

StructDef_ptr
Container_impl::create_struct(const char* id,
			      const char* name,
			      const char* version,
			      const StructMemberSeq& members)
    throw(SystemException)
{
    Container_var self = _this();
    StructDef_impl* p = new StructDef_impl(orb_, self);
    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
        p -> members(members);
    }
    catch(...)
    {
        p -> _remove_ref();
        throw;
    }
    contents_.append(p -> _this());
    return p -> _this();
}

UnionDef_ptr
Container_impl::create_union(const char* id,
			     const char* name,
			     const char* version,
			     IDLType_ptr discriminator_type,
			     const UnionMemberSeq& members)
    throw(SystemException)
{
    Container_var self = _this();
    UnionDef_impl* p = new UnionDef_impl(orb_, self);

    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
        p -> discriminator_type_def(discriminator_type);
        p -> members(members);
    }
    catch(...)
    {
        p -> _remove_ref();
        throw;
    }
        
    contents_.append(p -> _this());
    return p -> _this();
}

EnumDef_ptr
Container_impl::create_enum(const char* id,
			    const char* name,
			    const char* version,
			    const EnumMemberSeq& members)
    throw(SystemException)
{
    Container_var self = _this();
    
    EnumDef_impl* p = new EnumDef_impl(orb_, self);
    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
        p -> members(members);
    }
    catch(...)
    {
        p -> _remove_ref();
        throw;
    }
        
    contents_.append(p -> _this());
    return p -> _this();
}

AliasDef_ptr
Container_impl::create_alias(const char* id,
			     const char* name,
			     const char* version,
			     IDLType_ptr original_type)
    throw(SystemException)
{
    Container_var self = _this();
    AliasDef_impl* p = new AliasDef_impl(orb_, self);

    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
        p -> original_type_def(original_type);
    }
    catch(...)
    {
        p -> _remove_ref();
        throw;
    }
    
    contents_.append(p -> _this());
    return p -> _this();
}

InterfaceDef_ptr
Container_impl::create_interface(const char* id,
				 const char* name,
				 const char* version,
				 const InterfaceDefSeq&
				 base_interfaces,
				 Boolean is_abstract)
    throw(SystemException)
{
    Container_var self = _this();
    InterfaceDef_impl* p = new InterfaceDef_impl(orb_, self);

    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
        p -> base_interfaces(base_interfaces);
        p -> is_abstract(is_abstract);
    }
    catch(...)
    {
        p -> _remove_ref();
        throw;
    }
    contents_.append(p -> _this());
    return p -> _this();
}

ValueDef_ptr
Container_impl::create_value(const char* id,
			     const char* name,
			     const char* version,
			     Boolean is_custom,
			     Boolean is_abstract,
			     ValueDef_ptr base_value,
			     Boolean is_truncatable,
			     const ValueDefSeq&
			     abstract_base_values,
			     InterfaceDef_ptr supported_interface,
			     const InitializerSeq& initializers)
    throw(SystemException)
{
    Container_var self = _this();
    ValueDef_impl* p = new ValueDef_impl(orb_, self);

    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
        p -> is_abstract(is_abstract);
        p -> is_custom(is_custom);
        p -> base_value(base_value);
        p -> abstract_base_values(abstract_base_values);
        p -> supported_interface(supported_interface);
        p -> is_truncatable(is_truncatable);
        p -> initializers(initializers);
    }
    catch(...)
    {
        p -> _remove_ref();
        throw;
    }
    contents_.append(p -> _this());
    return p -> _this();
}

ValueBoxDef_ptr
Container_impl::create_value_box(const char* id,
				 const char* name,
				 const char* version,
				 IDLType_ptr original_type_def)
    throw(SystemException)
{
    Container_var self = _this();
    ValueBoxDef_impl* p =
        new ValueBoxDef_impl(orb_, self);

    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
        p -> original_type_def(original_type_def);
    }
    catch(...)
    {
        p -> _remove_ref();
        throw;
    }
    contents_.append(p -> _this());
    return p -> _this();
}

ExceptionDef_ptr
Container_impl::create_exception(const char* id,
				 const char* name,
				 const char* version,
				 const StructMemberSeq& members)
    throw(SystemException)
{
    Container_var self = _this();
    ExceptionDef_impl* p = new ExceptionDef_impl(orb_, self);

    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
        p -> members(members);
    }
    catch(...)
    {
        p -> _remove_ref();
        throw;
    }
    contents_.append(p -> _this());
    return p -> _this();
}

NativeDef_ptr
Container_impl::create_native(const char* id,
			      const char* name,
			      const char* version)
    throw(SystemException)
{
    Container_var self = _this();
    NativeDef_impl* p = new NativeDef_impl(orb_, self);
    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
    }
    catch(...)
    {
        p -> _remove_ref();
        throw;
    }
    contents_.append(p -> _this());
    return p -> _this();
}

// ----------------------------------------------------------------------
// ExceptionDef_impl member implementations
// ----------------------------------------------------------------------

ExceptionDef_impl::ExceptionDef_impl(ORB_ptr orb, Container_ptr p)
    : IRObject_impl(orb),
      Container_impl(orb),
      Contained_impl(orb, p)
{
}

ExceptionDef_impl::~ExceptionDef_impl()
{
}

void
ExceptionDef_impl::releaseInternal()
{
    members_.length(0);
    Contained_impl::releaseInternal();
}

CORBA::Contained::Description*
ExceptionDef_impl::describe()
    throw(SystemException)
{
    ExceptionDescription* desc = new ExceptionDescription;

    desc -> name = name();
    desc -> id = id();

    Contained_var contained = CORBA::Contained::_narrow(definedIn_);
    if(!is_nil(contained))
	desc -> defined_in = contained -> id();
    else
	desc -> defined_in = string_dup("");

    desc -> version = version();
    desc -> type = type();

    CORBA::Contained::Description* description =
        new CORBA::Contained::Description;
    description -> kind = def_kind();
    description -> value <<= desc;

    return description;
}

TypeCode_ptr
ExceptionDef_impl::type()
    throw(SystemException)
{
    ContainedSeq seq;
    TypeCode_var tc = typeRecursive(seq);
    assert(seq.length() == 0);
    return tc._retn();
}

TypeCode_ptr
ExceptionDef_impl::typeRecursive(ContainedSeq& seq)
    throw(SystemException)
{
    seq.length(seq.length() + 1);
    seq[seq.length() - 1] = _this();
    
    POA_var poa = _default_POA();
    for(ULong i = 0 ; i < members_.length() ; i++)
    {
	ServantBase_var servant =
	    poa -> reference_to_servant(members_[i].type_def);
	IDLType_impl* idlTypeImpl = dynamic_cast<IDLType_impl*>(servant.in());
	assert(idlTypeImpl);
	members_[i].type = idlTypeImpl -> typeRecursive(seq);
    }

    seq.length(seq.length() - 1);
    return orb_ -> create_exception_tc(id_, name_, members_);
}

StructMemberSeq*
ExceptionDef_impl::members()
    throw(SystemException)
{
    //
    // Recalculate member TypeCodes, as ids may have changed, and also
    // in order to calculate recursive TypeCodes correctly
    //
    POA_var poa = _default_POA();
    for(ULong i = 0 ; i < members_.length() ; i++)
    {
	ServantBase_var servant =
	    poa -> reference_to_servant(members_[i].type_def);
	IDLType_impl* idlTypeImpl = dynamic_cast<IDLType_impl*>(servant.in());
	assert(idlTypeImpl);
	ContainedSeq seq;
	members_[i].type = idlTypeImpl -> typeRecursive(seq);
	assert(seq.length() == 0);
    }

    return new StructMemberSeq(members_);
}

void
ExceptionDef_impl::members(const StructMemberSeq& members)
    throw(SystemException)
{
    members_ = members;
}

// ----------------------------------------------------------------------
// AttributeDef_impl member implementations
// ----------------------------------------------------------------------

AttributeDef_impl::AttributeDef_impl(ORB_ptr orb, Container_ptr p)
    : IRObject_impl(orb),
      Contained_impl(orb, p)
{
}

AttributeDef_impl::~AttributeDef_impl()
{
}

void
AttributeDef_impl::releaseInternal()
{
    idlType_ = IDLType::_nil();
    Contained_impl::releaseInternal();
}

void
AttributeDef_impl::name(const char* s)
    throw(SystemException)
{
    if(name_.in() && strcmp(name_, s) == 0)
	return;

    //
    // Find out whether attribute already exists in a base
    // interface. Using case-insensitive comparison.
    //
    ULong i, j;

    // Do NOT exclude inherited names
    ContainedSeq_var seq = definedIn_ -> contents(dk_Attribute, false);

    String_var s1 = s;
    for(j = 0 ; j < strlen(s1) ; j++)
	if(isupper(s1[j]))
	    s1[j] = tolower(s1[j]);

    for(i = 0 ; i < seq -> length() ; i++)
    {
	String_var s2 = seq[i] -> name();
	for(j = 0 ; j < strlen(s2) ; j++)
	    if(isupper(s2[j]))
		s2[j] = tolower(s2[j]);

	if(strcmp(s1, s2) == 0)
	{
	    //
	    // Attribute already exists
	    //
	    throw BAD_PARAM(5, COMPLETED_NO);
	}
    }
    
    Contained_impl::name(s);
}

CORBA::Contained::Description*
AttributeDef_impl::describe()
    throw(SystemException)
{
    AttributeDescription* desc = new AttributeDescription;

    desc -> name = name();
    desc -> id = id();

    Contained_var contained = CORBA::Contained::_narrow(definedIn_);
    if(!is_nil(contained))
	desc -> defined_in = contained -> id();
    else
	desc -> defined_in = string_dup("");

    desc -> version = version();
    desc -> type = type();
    desc -> mode = mode();

    CORBA::Contained::Description* description =
	new CORBA::Contained::Description;
    description -> kind = def_kind();
    description -> value <<= desc;

    return description;
}

TypeCode_ptr
AttributeDef_impl::type()
    throw(SystemException)
{
    return idlType_ -> type();
}

IDLType_ptr
AttributeDef_impl::type_def()
    throw(SystemException)
{
    return IDLType::_duplicate(idlType_);
}

void
AttributeDef_impl::type_def(IDLType_ptr type)
    throw(SystemException)
{
    idlType_ = IDLType::_duplicate(type);
}

AttributeMode
AttributeDef_impl::mode()
    throw(SystemException)
{
    return attributeMode_;
}

void
AttributeDef_impl::mode(AttributeMode mode)
    throw(SystemException)
{
    attributeMode_ = mode;
}

// ----------------------------------------------------------------------
// OperationDef_impl member implementations
// ----------------------------------------------------------------------

OperationDef_impl::OperationDef_impl(ORB_ptr orb, Container_ptr p)
    : IRObject_impl(orb),
      Contained_impl(orb, p)
{
}

OperationDef_impl::~OperationDef_impl()
{
}

void
OperationDef_impl::releaseInternal()
{
    params_.length(0);
    result_ = IDLType::_nil();
    Contained_impl::releaseInternal();
}

void
OperationDef_impl::name(const char* s)
    throw(SystemException)
{
    if(name_.in() && strcmp(name_, s) == 0)
	return;

    //
    // Find out whether operation already exists in a base
    // interface. Using case-insensitive comparison.
    //
    ULong i, j;

    // Do NOT exclude inherited names
    ContainedSeq_var seq = definedIn_ -> contents(dk_Operation, false);

    String_var s1 = s;
    for(j = 0 ; j < strlen(s1) ; j++)
	if(isupper(s1[j]))
	    s1[j] = tolower(s1[j]);

    for(i = 0 ; i < seq -> length() ; i++)
    {
	String_var s2 = seq[i] -> name();
	for(j = 0 ; j < strlen(s2) ; j++)
	    if(isupper(s2[j]))
		s2[j] = tolower(s2[j]);

	if(strcmp(s1, s2) == 0)
	{
	    //
	    // Operation already exists
	    //
	    throw BAD_PARAM(5, COMPLETED_NO);
	}
    }
   
    Contained_impl::name(s);
}

CORBA::Contained::Description*
OperationDef_impl::describe()
    throw(SystemException)
{
    ULong i;
    OperationDescription* desc = new OperationDescription;

    desc -> name = name();
    desc -> id = id();

    Contained_var contained = CORBA::Contained::_narrow(definedIn_);
    if(!is_nil(contained))
	desc -> defined_in = contained -> id();
    else
	desc -> defined_in = string_dup("");

    desc -> version = version();
    desc -> result = result();
    desc -> mode = mode();
    desc -> contexts = contexts_;

    //
    // Recalculate member TypeCodes, as ids may have changed, and also
    // in order to calculate recursive TypeCodes correctly
    //
    POA_var poa = _default_POA();
    for(i = 0 ; i < params_.length() ; i++)
    {
	ServantBase_var servant =
	    poa -> reference_to_servant(params_[i].type_def);
	IDLType_impl* idlTypeImpl =
	    dynamic_cast<IDLType_impl*>(servant.in());
	assert(idlTypeImpl);
	ContainedSeq seq;
	params_[i].type = idlTypeImpl -> typeRecursive(seq);
	assert(seq.length() == 0);
    }

    desc -> parameters = params_;

    for(i = 0 ; i < exceptions_.length() ; i++)
    {
	CORBA::Contained::Description_var d = exceptions_[i] -> describe();
	ExceptionDescription* excDesc;
	Boolean b = d -> value >>= excDesc;
	assert(b);

	desc -> exceptions.append(*excDesc);
    }

    CORBA::Contained::Description* description =
	new CORBA::Contained::Description;
    description -> kind = def_kind();
    description -> value <<= desc;

    return description;
}

TypeCode_ptr
OperationDef_impl::result()
    throw(SystemException)
{
    return result_ -> type();
}

IDLType_ptr
OperationDef_impl::result_def()
    throw(SystemException)
{
    return IDLType::_duplicate(result_);
}

void
OperationDef_impl::result_def(IDLType_ptr res)
    throw(SystemException)
{
    result_ = IDLType::_duplicate(res);
}

ParDescriptionSeq*
OperationDef_impl::params()
    throw(SystemException)
{
    //
    // Recalculate member TypeCodes, as ids may have changed, and also
    // in order to calculate recursive TypeCodes correctly
    //
    POA_var poa = _default_POA();
    for(ULong i = 0 ; i < params_.length() ; i++)
    {
	ServantBase_var servant =
	    poa -> reference_to_servant(params_[i].type_def);
	IDLType_impl* idlTypeImpl = dynamic_cast<IDLType_impl*>(servant.in());
	assert(idlTypeImpl);
	ContainedSeq seq;
	params_[i].type = idlTypeImpl -> typeRecursive(seq);
	assert(seq.length() == 0);
    }

    return new ParDescriptionSeq(params_);
}

void
OperationDef_impl::params(const ParDescriptionSeq& p)
    throw(SystemException)
{
    //
    // Ensure we don't have duplicate parameter names
    //
    for(ULong i = 0 ; i < p.length() ; i++)
    {
	String_var s1 = p[i].name;
	for(ULong k = 0 ; k < strlen(s1) ; k++)
	    if(isupper(s1[k]))
		s1[k] = tolower(s1[k]);

	for(ULong j = i + 1 ; j < p.length() ; j++)
	{
	    String_var s2 = p[j].name;
	    for(ULong k = 0 ; k < strlen(s2) ; k++)
		if(isupper(s2[k]))
		    s2[k] = tolower(s2[k]);
	    
	    if(strcmp(s1, s2) == 0)
		throw BAD_PARAM(MinorDuplicateFormalParam, COMPLETED_NO);
	}
    }

    params_ = p;
}

OperationMode
OperationDef_impl::mode()
    throw(SystemException)
{
    return mode_;
}

void
OperationDef_impl::mode(OperationMode m)
    throw(SystemException)
{
    mode_ = m;
}

ContextIdSeq*
OperationDef_impl::contexts()
    throw(SystemException)
{
    return new ContextIdSeq(contexts_);
}

void
OperationDef_impl::contexts(const ContextIdSeq& c)
    throw(SystemException)
{
    contexts_ = c;
}

ExceptionDefSeq*
OperationDef_impl::exceptions()
    throw(SystemException)
{
    return new ExceptionDefSeq(exceptions_);
}

void
OperationDef_impl::exceptions(const ExceptionDefSeq& e)
    throw(SystemException)
{
    exceptions_ = e;
}

// ----------------------------------------------------------------------
// InterfaceDef_impl member implementations
// ----------------------------------------------------------------------

InterfaceDef_impl::InterfaceDef_impl(ORB_ptr orb, Container_ptr p)
    : IRObject_impl(orb),
      Container_impl(orb),
      Contained_impl(orb, p),
      IDLType_impl(orb),
      isAbstract_(FALSE)
{
}

InterfaceDef_impl::~InterfaceDef_impl()
{
}

void
InterfaceDef_impl::releaseInternal()
{
    baseInterfaces_.length(0);
    Container_impl::releaseInternal();
    Contained_impl::releaseInternal();
    IDLType_impl::releaseInternal();
}

TypeCode_ptr
InterfaceDef_impl::type()
    throw(SystemException)
{
    if(isAbstract_)
        return orb_ -> create_abstract_interface_tc(id_, name_);
    else
        return orb_ -> create_interface_tc(id_, name_);
}

CORBA::Contained::Description*
InterfaceDef_impl::describe()
    throw(SystemException)
{
    InterfaceDescription* desc = new InterfaceDescription;

    desc -> name = name();
    desc -> id = id();

    Contained_var contained = CORBA::Contained::_narrow(definedIn_);
    if(!is_nil(contained))
	desc -> defined_in = contained -> id();
    else
	desc -> defined_in = string_dup("");

    desc -> version = version();

    for(ULong i = 0 ; i < baseInterfaces_.length() ; i++)
    {
	Contained_var baseContained =
	    CORBA::Contained::_narrow(baseInterfaces_[i]);
	if(!is_nil(baseContained))
	    desc -> base_interfaces.append(baseContained -> id());
    }

    desc -> is_abstract = is_abstract();

    CORBA::Contained::Description* description =
	new CORBA::Contained::Description;
    description -> kind = def_kind();
    description -> value <<= desc;

    return description;
}

Contained_ptr
InterfaceDef_impl::lookup(const char* search_name)
    throw(SystemException)
{
    Contained_var contained = Container_impl::lookup(search_name);

    if(is_nil(contained))
    {
	for(ULong i = 0 ; i < baseInterfaces_.length() ; i++)
	{
	    Contained_var contained2 =
		baseInterfaces_[i] -> lookup(search_name);
	    
	    if(!is_nil(contained2))
	    {
		if(!is_nil(contained))
		{
		    RepositoryId_var id = contained -> id();
		    RepositoryId_var id2 = contained2 -> id();
		    if(strcmp(id, id2) != 0)
		    {
			//
			// Name is ambiguous
			//
			throw INTF_REPOS(MinorLookupAmbiguous, COMPLETED_NO);
		    }
		}

		contained = contained2;
	    }
	}
    }

    return contained._retn();
}

ContainedSeq*
InterfaceDef_impl::contents(DefinitionKind limit_type,
			    Boolean exclude_inherited)
    throw(SystemException)
{
    ContainedSeq* s = Container_impl::contents(limit_type, exclude_inherited);
    
    if(!exclude_inherited)
    {
	for(ULong i = 0 ; i < baseInterfaces_.length() ; i++)
	{
	    ContainedSeq_var s2 =
		baseInterfaces_[i] -> contents(limit_type, false);
	    merge(s, s2);
	}
    }
    
    return s;
}

ContainedSeq*
InterfaceDef_impl::lookup_name(const char* search_name,
			       Long levels_to_search,
			       DefinitionKind limit_type,
			       Boolean exclude_inherited)
    throw(SystemException)
{
    ContainedSeq* s =
	Container_impl::lookup_name(search_name, levels_to_search,
				    limit_type, exclude_inherited);

    if(!exclude_inherited)
    {
	for(ULong i = 0 ; i < baseInterfaces_.length() ; i++)
	{
	    ContainedSeq_var s2 =
		baseInterfaces_[i] -> lookup_name(search_name,
						  levels_to_search,
						  limit_type,
						  false);
	    merge(s, s2);
	}
    }
		
    return s;
}

InterfaceDefSeq*
InterfaceDef_impl::base_interfaces()
    throw(SystemException)
{
    return new InterfaceDefSeq(baseInterfaces_);
}

void
InterfaceDef_impl::base_interfaces(const InterfaceDefSeq& bi)
    throw(SystemException)
{
    ContainedSeq conSeq;

    ULong i;

    //
    // If abstract, make sure all base interfaces are also abstract
    //
    /*
    if(is_abstract())
    {
        for(ULong i = 0 ; i < bi.length() ; i++)
        {
            if(! bi[i] -> is_abstract())
                // XXX exception isn't specified for this situation!
                throw BAD_PARAM(5, COMPLETED_NO);
        }
    }
    */

    for(i = 0 ; i < bi.length() ; i++)
    {
	ContainedSeq_var seq = bi[i] -> contents(dk_Operation, false);

	for(ULong j = 0 ; j < seq -> length() ; j++)
	{
	    String_var name1 = seq[j] -> name();
	    String_var id1 = seq[j] -> id();

	    for(ULong k = 0 ; k < conSeq.length() ; k++)
	    {
		String_var name2 = conSeq[k] -> name();
		String_var id2 = conSeq[k] -> id();

		if(strcmp(id1, id2) != 0 && strcmp(name1, name2) == 0)
		{
		    //
		    // Operation defined in multiple base interfaces
		    //
		    throw BAD_PARAM(5, COMPLETED_NO);
		}
	    }

	    conSeq.append(seq[j]);
	}
    }

    conSeq.length(0);

    for(i = 0 ; i < bi.length() ; i++)
    {
	ContainedSeq_var seq =
	    bi[i] -> contents(dk_Attribute, false);

	for(ULong j = 0 ; j < seq -> length() ; j++)
	{
	    String_var name1 = seq[j] -> name();
	    String_var id1 = seq[j] -> id();

	    for(ULong k = 0 ; k < conSeq.length() ; k++)
	    {
		String_var name2 = conSeq[k] -> name();
		String_var id2 = conSeq[k] -> id();

		if(strcmp(id1, id2) != 0 && strcmp(name1, name2) == 0)
		{
		    //
		    // Attribute defined in multiple base interfaces
		    //
		    throw BAD_PARAM(5, COMPLETED_NO);
		}
	    }

	    conSeq.append(seq[j]);
	}
    }

    baseInterfaces_ = bi;
}

Boolean
InterfaceDef_impl::is_abstract()
    throw(SystemException)
{
    return isAbstract_;
}

void
InterfaceDef_impl::is_abstract(Boolean b)
    throw(SystemException)
{
    //
    // If abstract, make sure all base interfaces are also abstract
    //
    if(b)
    {
        for(ULong i = 0 ; i < baseInterfaces_.length() ; i++)
        {
            if(! baseInterfaces_[i] -> is_abstract())
                // XXX exception isn't specified for this situation!
                throw BAD_PARAM(5, COMPLETED_NO);
        }
    }

    isAbstract_ = b;
}

Boolean
InterfaceDef_impl::is_a(const char* interface_id)
    throw(SystemException)
{
    if(strcmp(id_, interface_id) == 0)
	return true;

    for(ULong i = 0 ; i < baseInterfaces_.length() ; i++)
	if(baseInterfaces_[i] -> is_a(interface_id))
	    return true;

    return false;
}

CORBA::InterfaceDef::FullInterfaceDescription*
InterfaceDef_impl::describe_interface()
    throw(SystemException)
{
    CORBA::InterfaceDef::FullInterfaceDescription* desc =
        new CORBA::InterfaceDef::FullInterfaceDescription;

    desc -> name = name();
    desc -> id = id();

    Contained_var contained = CORBA::Contained::_narrow(definedIn_);
    if(!is_nil(contained))
	desc -> defined_in = contained -> id();
    else
	desc -> defined_in = string_dup("");

    desc -> version = version();

    ULong i;

    for(i = 0 ; i < contents_.length() ; i++)
    {
	Contained_ptr contained = contents_[i];
	
	OperationDef_var operation = CORBA::OperationDef::_narrow(contained);
	AttributeDef_var attribute = CORBA::AttributeDef::_narrow(contained);

	if(!is_nil(operation))
	{
	    CORBA::Contained::Description_var d = operation -> describe();
	    OperationDescription* operationDesc;
	    Boolean b = d -> value >>= operationDesc;
	    assert(b);

	    desc -> operations.append(*operationDesc);
	}
	else if(!is_nil(attribute))
	{
	    CORBA::Contained::Description_var d = attribute -> describe();
	    AttributeDescription* attributeDesc;
	    Boolean b = d -> value >>= attributeDesc;
	    assert(b);

	    desc -> attributes.append(*attributeDesc);
	}
    }

    for(i = 0 ; i < baseInterfaces_.length() ; i++)
    {
	Contained_var baseContained =
            CORBA::Contained::_narrow(baseInterfaces_[i]);
	if(!is_nil(baseContained))
	    desc -> base_interfaces.append(baseContained -> id());
    }

    desc -> type = type();

    desc -> is_abstract = is_abstract();

    return desc;
}

AttributeDef_ptr
InterfaceDef_impl::create_attribute(const char* id,
				    const char* name,
				    const char* version,
				    IDLType_ptr type,
				    AttributeMode mode)
    throw(SystemException)
{
    InterfaceDef_var self = _this();
    AttributeDef_impl* p = new AttributeDef_impl(orb_, self);
    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
        p -> type_def(type);
        p -> mode(mode);
    }
    catch(...)
    {
        p -> _remove_ref();
        throw;
    }
    contents_.append(p -> _this());
    return p -> _this();
}

OperationDef_ptr
InterfaceDef_impl::
create_operation(const char* id,
		 const char* name,
		 const char* version,
		 IDLType_ptr result,
		 OperationMode mode,
		 const ParDescriptionSeq& params,
		 const ExceptionDefSeq& exceptions,
		 const ContextIdSeq& contexts)
    throw(SystemException)
{
    InterfaceDef_var self = _this();
    OperationDef_impl* p = new OperationDef_impl(orb_, self);
    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
        p -> result_def(result);
        p -> mode(mode);
        p -> params(params);
        p -> exceptions(exceptions);
        p -> contexts(contexts);
    }
    catch(...)
    {
        p -> _remove_ref();
        throw;
    }
    contents_.append(p -> _this());
    return p -> _this();
}
