// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>

#include <Types.h>
#include <IFR3_impl.h>

#include <stdlib.h>
#include <stdio.h>

#ifdef WIN32
#   define popen _popen
#   define pclose _pclose
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;
using namespace OB;
using namespace PortableServer;

static void
usage(const char* progName)
{
    cerr << "Usage:\n";
    cerr << progName << " [options] idl-files ...\n"
	"\n"
	"Options:\n"
	"-h, --help             Show this message.\n"
	"-v, --version          Show ORBacus version.\n"
	"-e, --cpp NAME         Use NAME as preprocessor.\n"
	"-d, --debug            Print diagnostic messages.\n"
	"-i, --ior              Print repository IOR on standard output.\n"
	"-DNAME                 Define NAME as 1.\n"
	"-DNAME=DEF             Define NAME as DEF.\n"
	"-UNAME                 Remove any definition for NAME.\n"
	"-IDIR                  Put DIR in the include file search path.\n"
        "--ignore-undefined     Don't complain about interfaces that have\n"
        "                       been declared but not defined.\n"
	;
}

int
run(int argc, char* argv[])
{
    //
    // Set program name
    //
    const char* progName = argv[0];

    //
    // Create the ORB
    //
    ORB_var orb = ORB_init(argc, argv);

    //
    // Resolve the service object, if available as an initial reference.
    //
    const char* service = "InterfaceRepository";
    Object_var serviceObject;
    try
    {
        serviceObject = orb -> resolve_initial_references(service);
    }
    catch(ORB::InvalidName&)
    {
    }

    //
    // Set the IIOP port based on the port defined in the IIOP profile
    // of the service object (if it exists). Then create the BOA.
    //
    ORB::set_iiop_port_from_object(serviceObject);
    
    Object_var poaObj;

    try
    {
        poaObj = orb -> resolve_initial_references("RootPOA");
    }
    catch(const ORB::InvalidName&)
    {
        cerr << "Error: can't resolve `RootPOA'" << endl;
        exit(1);
    }
    
    if(is_nil(poaObj))
    {
        cerr << "Error: `RootPOA' is a nil object reference" << endl;
        exit(1);
    }
    
    POA_var root = POA::_narrow(poaObj);
    
    if(is_nil(root))
    {
        cerr << "Error: `RootPOA' is not a POA object reference" << endl;
        exit(1);
    }

    POAManager_var manager = root -> the_POAManager();

    PolicyList policies(1);
    policies.length(1);
    policies[0] = root ->
    create_id_assignment_policy(USER_ID);

    POA_var childPOA =
        root -> create_POA("child", manager, policies);
    
    //
    // Get the C preprocessor
    //
    String_var cppCmd = string_dup(progName);
#ifdef WIN32
    char* p1 = strrchr(cppCmd.inout(), '\\');
    char* p2 = strrchr(cppCmd.inout(), '/');
    char* p = p1 > p2 ? p1 : p2;
#else
    char* p = strrchr(cppCmd.inout(), '/');
#endif
    if(p)
    {
	p++;
	*p = '\0';
	cppCmd += "idlcpp";
    }
    else
	cppCmd = string_dup("idlcpp");

    //
    // Get options
    //
    String_var cppArgs;
    bool ignoreUndefined = false;
    bool ior = false;

    Long i;
    for(i = 1 ; i < argc && *argv[i] == '-' ; i++)
    {
	if(strcmp(argv[i], "--help") == 0 ||
	   strcmp(argv[i], "-h") == 0)
	{
	    usage(progName);
	    return 0;
	}
	else if(strcmp(argv[i], "--version") == 0 ||
		strcmp(argv[i], "-v") == 0)
	{
	    cerr << "ORBacus " << Version << endl;
	    return 0;
	}
	else if(strcmp(argv[i], "--cpp") == 0 ||
		strcmp(argv[i], "-e") == 0)
	{
	    if(++i < argc)
		cppCmd = string_dup(argv[i]);
	    else
	    {
		cerr << progName << ": argument expected for `"
		     << argv[i - 1] << "'" << endl;
		return 1;
	    }
	}
	else if(strcmp(argv[i], "--debug") == 0 ||
		strcmp(argv[i], "-d") == 0)
	{
	    extern int yydebug;
	    yydebug = 1;
	}
	else if(strcmp(argv[i], "--ior") == 0 ||
		strcmp(argv[i], "-i") == 0)
	{
	    ior = true;
	}
	else if(strncmp(argv[i], "-D", 2) == 0 ||
		strncmp(argv[i], "-U", 2) == 0 ||
		strncmp(argv[i], "-I", 2) == 0)
	{
	    if(strlen(argv[i]) == 2)
	    {
		cerr << progName << ": argument expected for `"
		     << argv[i] << "'" << endl; // not argv[i - 1] !
		return 1;
	    }

	    cppArgs += ' ';
	    cppArgs += argv[i];
	}
	else if(strcmp(argv[i], "--ignore-undefined") == 0)
	{
	    ignoreUndefined = true;
	}
	else
	{
	    cerr << progName << ": unknown option `"
		 << argv[i] << "'\n" << endl;
	    usage(progName);
	    return 1;
	}
    }
    
    //
    // Create implementation object
    //
    ObjectId_var repositoryID = string_to_ObjectId("DefaultRepository");
    Repository_impl* impl = new Repository_impl(orb);
    childPOA -> activate_object_with_id(repositoryID, impl);

    Repository_var repository = impl -> _this();

    int status = 0;
    while(status == 0 && i < argc)
    {

#ifdef WIN32
	//
	// At least for WIN32 we can check for idlcpp being present
	//
	{
	    char buf[MAX_PATH];
	    char* tmp;

	    String_var exeCmd = cppCmd;
	    exeCmd += ".exe";
	    if(!SearchPath(0, exeCmd, 0, MAX_PATH, buf, &tmp))
	    {
		cerr << progName << ": couldn't find " << cppCmd << endl;
		return 1;
	    }
	}
#endif

	String_var file = string_dup(argv[i++]);
	
	//
	// Open input file
	//
	extern FILE* yyin;
	String_var cmd = cppCmd;
	cmd += cppArgs;
	cmd += ' ';
	cmd += file;

	if((yyin = popen(cmd, "r")) == NULL)
	    return 1;

	//
	// Do parsing
	//
	int IdlParse(ORB_ptr, Repository_ptr, bool,
		     RepositoryIdSeq*&, IdlStringSeq*&, IdlCommentSeq*&);
	RepositoryIdSeq_var idSeq; // This is an out parameter
	IdlStringSeq_var includeSeq; // This is an out paramter
	IdlCommentSeq_var commentSeq; // This is an out paramter
	status = IdlParse(orb, repository, ignoreUndefined,
			  idSeq.out(), includeSeq.out(), commentSeq.out());
	
	//
	// Close input file
	//
	pclose(yyin);
    }

    //
    // Print IOR on standard output
    //
    if(status == 0 && ior)
    {
	String_var s = orb -> object_to_string(repository);
	cout << s << endl;
    }

    manager -> activate();

    //
    // Run implementation
    //

    if(status == 0)
        orb -> run();
    
    //
    // Release internal object references to resolve
    // cyclic dependencies
    //
    impl -> releaseInternal();

    return status;
}

int
main(int argc, char* argv[], char*[])
{
    int status;

    try
    {
	//
	// Generate code
	//
	status = run(argc, argv);
    }
    catch(SystemException& ex)
    {
	PrintException(ex);
	status = 1;
    }

    return status;
}
