// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>

#include <Test_skel.h>
#include <Test_impl.h>
#include <TestUtil.h>

#include <stdlib.h>
#include <errno.h>

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;
using namespace PortableServer;

class TestActivator_impl : public POA_PortableServer::ServantActivator
{
    ObjectId oid_;
    POA_ptr poa_;
    Servant servant_;
    Boolean remaining_;
    bool valid_;

public:
    TestActivator_impl()
    {
        valid_ = false;
    }

    void expect(const ObjectId& oid,
                POA_ptr poa,
                Servant servant,
                Boolean remaining)
    {
        oid_ = oid;
        poa_ = poa;
        servant_ = servant;
        remaining_ = remaining;
        valid_ = false;
    }

    bool isValid() { return valid_; }

    Servant incarnate(const ObjectId& /*oid*/, POA_ptr /*poa*/)
        throw(ForwardRequest, SystemException)
    {
        return 0;
    }

    void etherealize(const ObjectId& oid,
                     POA_ptr poa,
                     Servant servant,
                     Boolean /*cleanup*/,
                     Boolean remaining)
        throw(SystemException)
    {
        assert(TestUtil_Compare(oid_, oid));
        assert(poa_ -> _is_equivalent(poa));
        assert(servant_ == servant);
        assert(remaining_ == remaining);
        valid_ = true;
    }
};


void
TestServantActivator(ORB_ptr orb, POA_ptr root)
{
    Object_var obj;
    POA_var system, nonretain, multiple, ether;
    ObjectId_var id1, id2, id3;
    PolicyList policies;
    Test_impl* servant1;
    Test_impl* servant2;
    ServantBase_var tmpserv;

    POAManager_var manager = root -> the_POAManager();

    //
    // Create POAs
    //

    policies.length(3);
    policies[0] =
        root -> create_id_assignment_policy(PortableServer::SYSTEM_ID);
    policies[1] =
        root -> create_id_uniqueness_policy(PortableServer::UNIQUE_ID);
    policies[2] =
        root -> create_servant_retention_policy(PortableServer::RETAIN);
    system = root -> create_POA("system_id", manager, policies);

    policies.length(5);
    policies[0] =
        root -> create_id_assignment_policy(PortableServer::USER_ID);
    policies[1] =
        root -> create_id_uniqueness_policy(PortableServer::MULTIPLE_ID);
    policies[2] =
        root -> create_servant_retention_policy(PortableServer::NON_RETAIN);
    policies[3] = root ->
        create_request_processing_policy(PortableServer::USE_DEFAULT_SERVANT);
    policies[4] = root ->
     create_implicit_activation_policy(PortableServer::NO_IMPLICIT_ACTIVATION);
    nonretain = root -> create_POA("nonretain", manager, policies);

    policies.length(3);
    policies[0] =
        root -> create_id_assignment_policy(PortableServer::SYSTEM_ID);
    policies[1] =
        root -> create_id_uniqueness_policy(PortableServer::MULTIPLE_ID);
    policies[2] =
        root -> create_servant_retention_policy(PortableServer::RETAIN);
    multiple = root -> create_POA("multiple_id", manager, policies);

    policies.length(3);
    policies[0] =
        root -> create_id_assignment_policy(PortableServer::SYSTEM_ID);
    policies[1] =
        root -> create_id_uniqueness_policy(PortableServer::MULTIPLE_ID);
    policies[2] = root ->
        create_request_processing_policy(PortableServer::USE_SERVANT_MANAGER);
    ether = root -> create_POA("ether", manager, policies);
    TestActivator_impl activatorImpl;
    PortableServer::ServantActivator_var activator = activatorImpl._this();
    ether -> set_servant_manager(activator);

    //
    // Start tests
    //

    servant1 = new Test_impl(orb, "obj1");
    servant2 = new Test_impl(orb, "obj2");

    //
    // Test: activate_object w/ SYSTEM_ID POA
    //

    id1 = system -> activate_object(servant1);
    id2 = system -> activate_object(servant2);
    assert(!TestUtil_Compare(id1, id2));
    tmpserv = system -> id_to_servant(id1);
    assert(tmpserv.in() == servant1);
    tmpserv = system -> id_to_servant(id2);
    assert(tmpserv.in() == servant2);

    //
    // Test: ServantAlreadyActive exception
    //
    try
    {
        system -> activate_object(servant1);
        assert("activate_object should not have succeeded" == 0);
    }
    catch(const POA::ServantAlreadyActive&)
    {
        // expected
    }

    try
    {
        system -> activate_object(servant2);
        assert("activate_object should not have succeeded" == 0);
    }
    catch(const POA::ServantAlreadyActive&)
    {
        // expected
    }

    //
    // Test: deactivate_object
    //
    system -> deactivate_object(id2);
    system -> deactivate_object(id1);

    //
    // Test: ObjectNotActive exception
    //
    try
    {
        system -> deactivate_object(id1);
        assert("deactivate_object should not have succeeded" == 0);
    }
    catch(const POA::ObjectNotActive&)
    {
        // expected
    }

    try
    {
        system -> deactivate_object(id2);
        assert("deactivate_object should not have succeeded" == 0);
    }
    catch(const POA::ObjectNotActive&)
    {
        // expected
    }

    //
    // Test: WrongPolicy exception
    //
    try
    {
        nonretain -> activate_object(servant1);
        assert("activate_object should not have succeeded" == 0);
    }
    catch(const POA::WrongPolicy&)
    {
        // expected
    }

    try
    {
        ObjectId_var id = string_to_ObjectId("XXX");
        nonretain -> activate_object_with_id(id, servant1);
        assert("activate_object_with_id should not have succeeded" == 0);
    }
    catch(const POA::WrongPolicy&)
    {
        // expected
    }

    try
    {
        ObjectId_var id = string_to_ObjectId("XXX");
        nonretain -> deactivate_object(id);
        assert("deactivate_object_with_id should not have succeeded" == 0);
    }
    catch(const POA::WrongPolicy&)
    {
        // expected
    }

    //
    // Test: activate_object w/ MULTIPLE_ID POA
    //

    id1 = multiple -> activate_object(servant1);
    id2 = multiple -> activate_object(servant1);
    assert(!TestUtil_Compare(id1, id2));
    tmpserv = multiple -> id_to_servant(id1);
    assert(tmpserv.in() == servant1);
    tmpserv = multiple -> id_to_servant(id2);
    assert(tmpserv.in() == servant1);

    //
    // Test: confirm servant1 is no longer active
    //
    multiple -> deactivate_object(id1);
    multiple -> deactivate_object(id2);
    try
    {
        multiple -> id_to_servant(id1);
    }
    catch(const POA::ObjectNotActive&)
    {
        // expected
    }

    try
    {
        multiple -> id_to_servant(id2);
    }
    catch(const POA::ObjectNotActive&)
    {
        // expected
    }

    //
    // Test: confirm ServantActivator::etherealize is invoked on deactivate
    //
    id1 = ether -> activate_object(servant1);
    id2 = ether -> activate_object(servant1);
    id3 = ether -> activate_object(servant2);
    activatorImpl.expect(id1, ether, servant1, TRUE);
    ether -> deactivate_object(id1);
    assert(activatorImpl.isValid());
    activatorImpl.expect(id2, ether, servant1, FALSE);
    ether -> deactivate_object(id2);
    assert(activatorImpl.isValid());
    activatorImpl.expect(id3, ether, servant2, FALSE);
    ether -> deactivate_object(id3);
    assert(activatorImpl.isValid());


    system -> destroy(TRUE, TRUE);
    nonretain -> destroy(TRUE, TRUE);
    multiple -> destroy(TRUE, TRUE);
    ether -> destroy(TRUE, TRUE);

    //
    // Since activatorImpl is a stack-based servant, we need to deactivate
    // it before it goes out of scope
    //
    ObjectId_var id = root -> servant_to_id(&activatorImpl);
    root -> deactivate_object(id);

    tmpserv = 0;
    delete servant1;
    delete servant2;
}

int
main(int argc, char* argv[], char*[])
{
    int status;

    try
    {
        //
        // Create ORB
        //
        ORB_var orb = ORB_init(argc, argv);

        POA_var root = TestUtil_GetRootPOA(orb);

        cout << "Testing servant activator... " << flush;
        TestServantActivator(orb, root);
        cout << "OK!" << endl;
    }
    catch(const UserException&)
    {
        cerr << argv[0] << ": Unexpected user exception was thrown" << endl;
        status = 1;
    }
    catch(const SystemException& ex)
    {
        OB::PrintException(ex);
        status = 1;
    }

    return status;
}
