// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>

#include <TestUtil.h>

#include <assert.h>
#include <string.h>

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;
using namespace PortableServer;

class TestAdapterActivator_impl
    : public virtual POA_PortableServer::AdapterActivator
{
    String_var expectedName_;
    bool create_;
    bool invoked_;

public:
    TestAdapterActivator_impl()
        : create_(false), invoked_(false)
    {
    }

    void reset(const char* name, bool create)
    {
        expectedName_ = string_dup(name);
        create_ = create;
        invoked_ = false;
    }

    bool invoked() const
    {
        return invoked_;
    }

    virtual Boolean unknown_adapter(POA_ptr parent, const char* name)
        throw(SystemException)
    {
        assert(strcmp(name, expectedName_) == 0);
        invoked_ = true;
        if(create_)
        {
            String_var s = parent -> the_name();
            PolicyList policies;
            POAManager_var mgr = parent -> the_POAManager();
            POA_var poa = parent -> create_POA(name, mgr, policies);
            return true;
        }
        else
            return false;
    }
};


void
TestAdapterActivator(ORB_ptr orb, POA_ptr root)
{
    Object_var obj;
    PolicyList policies;
    POA_var poa, parent;
    POAManager_var mgr;
    String_var str;

    POAManager_var rootMgr = root -> the_POAManager();
    assert(!is_nil(rootMgr));

    TestAdapterActivator_impl activatorImpl;
    AdapterActivator_var activator = activatorImpl._this();
    root -> the_activator(activator);

    //
    // Test: Activator and successful creation
    //
    activatorImpl.reset("poa1", true);
    poa = root -> find_POA("poa1", TRUE);
    assert(!is_nil(poa));
    assert(activatorImpl.invoked());
    str = poa -> the_name();
    assert(strcmp(str, "poa1") == 0);
    parent = poa -> the_parent();
    assert(!is_nil(parent));
    assert(parent -> _is_equivalent(root));

    //
    // Test: Activator and unsuccessful creation
    //
    activatorImpl.reset("poa2", false);
    try
    {
        root -> find_POA("poa2", TRUE);
    }
    catch(const POA::AdapterNonExistent&)
    {
        // expected
    }
    assert(activatorImpl.invoked());

    //
    // Test: Make sure activator isn't called when POA already exists
    //
    activatorImpl.reset("poa1", true);
    poa = root -> find_POA("poa1", TRUE);
    assert(!activatorImpl.invoked());

    //
    // Test: Disable adapter activator and make sure it isn't invoked
    //
    root -> the_activator(AdapterActivator::_nil());
    activatorImpl.reset("poa2", false);
    try
    {
        root -> find_POA("poa2", TRUE);
    }
    catch(const POA::AdapterNonExistent&)
    {
        // expected
    }
    assert(!activatorImpl.invoked());

    //
    // Since activatorImpl is a stack-based servant, we need to deactivate
    // it before it goes out of scope
    //
    ObjectId_var id = root -> servant_to_id(&activatorImpl);
    root -> deactivate_object(id);
}

int
main(int argc, char* argv[], char*[])
{
    int status;

    try
    {
        //
        // Create ORB
        //
        ORB_var orb = ORB_init(argc, argv);

        POA_var root = TestUtil_GetRootPOA(orb);

        cout << "Testing adapter activator... " << flush;
        TestAdapterActivator(orb, root);
        cout << "OK!" << endl;
    }
    catch(const UserException&)
    {
        cerr << argv[0] << ": Unexpected user exception was thrown" << endl;
        status = 1;
    }
    catch(const SystemException& ex)
    {
        OB::PrintException(ex);
        status = 1;
    }

    return status;
}
