// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/CosProperty_impl.h>

#include <Config.h> // Make sure that the compiler complains if the
                    // configuration doesn't match

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;
using namespace CosPropertyService;

// ----------------------------------------------------------------------
// PropertySetDef_impl private member implementation
// ----------------------------------------------------------------------

//
// Check name, type and property
//
void
PropertySetDef_impl::check(const char* property_name,
			   const Any& property_value,
			   PropertyModeType property_mode)
{
    ULong i;

    //
    // Check name
    //
    if(strlen(property_name) == 0)
	throw InvalidPropertyName();

    //
    // Check if property is allowed
    //
    if(allowedProperties_.length())
    {
	bool propertyIsAllowed = false;
	for(i = 0 ; i < allowedProperties_.length() ; i++)
	{
            if(strcmp(allowedProperties_[i].property_name, property_name) == 0)
	    {
		//
		// Check if the type matches
		//
		TypeCode_var tc1 = allowedProperties_[i].property_value.type();
		TypeCode_var tc2 = property_value.type();
		
		if(!tc1 -> equal(tc2))
		    throw UnsupportedTypeCode();
 
		//
		// Mode is only relevant if not undefined
		//
		if(property_mode != undefined)
		{
		    if(allowedProperties_[i].property_mode != property_mode)
			continue;
		}

		propertyIsAllowed = true;
		break;
	    }
	}

	if(!propertyIsAllowed)
	    throw UnsupportedProperty();
    }

    //
    // Check if TypeCode is allowed
    //
    if(allowedTypes_.length())
    {
	bool typeIsAllowed = false;
	for(i = 0 ; i < allowedTypes_.length() ; i++)
	{
            TypeCode_var tc = property_value.type();
            if(allowedTypes_[i] -> equal(tc))
	    {
		typeIsAllowed = true;
		break;
	    }
	}

	if(!typeIsAllowed)
	    throw UnsupportedTypeCode();
    }
}

//
// Find property by name
//
PropertyDef*
PropertySetDef_impl::get(const char* property_name, ULong* pos)
{
//    OB_SYNCHRONIZED

    for(ULong i = 0 ; i < propertySeq_.length() ; i++)
    {
	if(strcmp(propertySeq_[i].property_name, property_name) == 0)
	{
	    if(pos)
		*pos = i;
	    return &propertySeq_[i];
	}
    }

    return 0;
}

// ----------------------------------------------------------------------
// PropertySetDef_impl constructors
// ----------------------------------------------------------------------

PropertySetDef_impl::PropertySetDef_impl(PortableServer::POA_ptr poa)
    : poa_(PortableServer::POA::_duplicate(poa))
{
}

PropertySetDef_impl::PropertySetDef_impl(PortableServer::POA_ptr poa,
    const PropertyTypes& allowed_property_types,
    const PropertyDefs& allowed_property_defs)
    : poa_(PortableServer::POA::_duplicate(poa)),
      allowedTypes_(allowed_property_types),
      allowedProperties_(allowed_property_defs)
{
}

PropertySetDef_impl::PropertySetDef_impl(PortableServer::POA_ptr poa,
    const PropertyDefs& initial_property_defs)
    : poa_(PortableServer::POA::_duplicate(poa)),
      propertySeq_(initial_property_defs)
{
}

// ----------------------------------------------------------------------
// PropertySetDef_impl public member implementation
// ----------------------------------------------------------------------

void
PropertySetDef_impl::define_property(const char* property_name,
				     const Any& property_value)
    throw(InvalidPropertyName, ConflictingProperty, UnsupportedTypeCode,
	  UnsupportedProperty, ReadOnlyProperty, SystemException)
{
    try
    {
	define_property_with_mode(property_name, property_value, undefined);
    }
    catch(const UnsupportedMode&)
    {
	// This exception can't be thrown in this context
	assert(false);
    }
}

void
PropertySetDef_impl::define_properties(const Properties& nproperties)
    throw(MultipleExceptions, SystemException)
{
    ULong len = nproperties.length();
    PropertyDefs p(len);
    p.length(len);

    for(ULong i = 0 ; i < len ; i++)
    {
	p[i].property_name = nproperties[i].property_name;
	p[i].property_value = nproperties[i].property_value;
	p[i].property_mode = undefined;
    }

    define_properties_with_modes(p);
}

ULong
PropertySetDef_impl::get_number_of_properties()
    throw(SystemException)
{
//    OB_SYNCHRONIZED

    return propertySeq_.length();
}

void
PropertySetDef_impl::get_all_property_names(ULong how_many,
					    PropertyNames_out property_names,
					    PropertyNamesIterator_out rest)
    throw(SystemException)
{
//    OB_SYNCHRONIZED

    ULong i;
 
    ULong len = propertySeq_.length();
    ULong num = len < how_many ? len : how_many;
 
    property_names = new PropertyNames(num);
    property_names -> length(num);
 
    for(i = 0 ; i < num ; i++)
	(*property_names)[i] = propertySeq_[i].property_name;

    if(how_many < len)
    {
	PropertyNames rem(len - how_many);
        rem.length(len - how_many);
 
        for(i = 0 ; i < len - how_many ; i++)
	    rem[i] = propertySeq_[i + how_many].property_name;

	PortableServer::POA_var poa = _default_POA();
	PropertyNamesIterator_impl* iteratorImpl =
	    new PropertyNamesIterator_impl(poa, rem);
        rest = iteratorImpl -> _this();
    }
    else
        rest = PropertyNamesIterator::_nil();
}

Any*
PropertySetDef_impl::get_property_value(const char* property_name)
    throw(PropertyNotFound, InvalidPropertyName, SystemException)
{
//    OB_SYNCHRONIZED

    if(strlen(property_name) == 0)
	throw InvalidPropertyName();

    PropertyDef* p = get(property_name);
    if(!p)
	throw PropertyNotFound();

    return new Any(p -> property_value);
}

Boolean
PropertySetDef_impl::get_properties(const PropertyNames& property_names,
				    Properties_out nproperties)
    throw(SystemException)
{
//    OB_SYNCHRONIZED

    Boolean allNamesAreValid = true;

    ULong len = property_names.length();
    nproperties = new Properties(len);
    nproperties -> length(len);

    for(ULong i = 0 ; i < len ; i++)
    {
	(*nproperties)[i].property_name = property_names[i];

	PropertyDef* p = get(property_names[i]);
	if(p)
	    (*nproperties)[i].property_value = p -> property_value;
	else
	{
	    //
	    // Create property with value tk_void for unknown name
	    //
	    (*nproperties)[i].property_value.replace(_tc_void, 0);
	    allNamesAreValid = false;
	}
    }

    return allNamesAreValid;
}

void
PropertySetDef_impl::get_all_properties(ULong how_many,
					Properties_out nproperties,
					PropertiesIterator_out rest)
    throw(SystemException)
{
//    OB_SYNCHRONIZED

    ULong i;
 
    ULong len = propertySeq_.length();
    ULong num = len < how_many ? len : how_many;
 
    nproperties = new Properties(num);
    nproperties -> length(num);
 
    for(i = 0 ; i < num ; i++)
    {
	(*nproperties)[i].property_name = propertySeq_[i].property_name;
	(*nproperties)[i].property_value = propertySeq_[i].property_value;
    }

    if(how_many < len)
    {
	PropertyDefs rem(len - how_many);
        rem.length(len - how_many);
 
        for(i = 0 ; i < len - how_many ; i++)
	    rem[i] = propertySeq_[i + how_many];
 
	PortableServer::POA_var poa = _default_POA();
        PropertiesIterator_impl* propertiesIteratorImpl =
	    new PropertiesIterator_impl(poa, rem);
	rest = propertiesIteratorImpl -> _this();
    }
    else
        rest = PropertiesIterator::_nil();
}

void
PropertySetDef_impl::delete_property(const char* property_name)
    throw(PropertyNotFound, InvalidPropertyName, FixedProperty,
	  SystemException)
{
//    OB_SYNCHRONIZED

    if(strlen(property_name) == 0)
	throw InvalidPropertyName();

    ULong pos;
    PropertyDef* p = get(property_name, &pos);
    if(!p)
	throw PropertyNotFound();

    if(p -> property_mode == fixed_normal ||
       p -> property_mode == fixed_readonly)
	throw FixedProperty();

    propertySeq_.remove(pos);
}

void
PropertySetDef_impl::delete_properties(const PropertyNames& property_names)
    throw(MultipleExceptions, SystemException)
{
//    OB_SYNCHRONIZED

    //
    // Sequence with exceptions
    //
    MultipleExceptions mex;

    ULong len = property_names.length();
    for(ULong i = 0 ; i < len ; i++)
    {
	if(!len)
	{
	    PropertyException ex;
	    ex.reason = invalid_property_name;
	    ex.failing_property_name = property_names[i];
	    mex.exceptions.append(ex);
	    continue;
	}

	ULong pos;
	PropertyDef* p = get(property_names[i], &pos);
	if(!p)
	{
	    PropertyException ex;
	    ex.reason = property_not_found;
	    ex.failing_property_name = property_names[i];
	    mex.exceptions.append(ex);
	    continue;
	}

	if(p -> property_mode == fixed_normal ||
	   p -> property_mode == fixed_readonly)
	{
	    PropertyException ex;
	    ex.reason = fixed_property;
	    ex.failing_property_name = property_names[i];
	    mex.exceptions.append(ex);
	    continue;
	}

	propertySeq_.remove(pos);
    }

    //
    // Check for exceptions
    //
    if(mex.exceptions.length())
	throw mex;
}

Boolean
PropertySetDef_impl::delete_all_properties()
    throw(SystemException)
{
//    OB_SYNCHRONIZED

    //
    // Sequence with remaining (fixed) properties
    //
    PropertyDefs fixedSeq;

    ULong len = propertySeq_.length();
    for(ULong i = 0 ; i < len ; i++)
    {
	PropertyDef p = propertySeq_[i];
	if(p.property_mode == fixed_normal &&
	   p.property_mode == fixed_readonly)
	    fixedSeq.append(p);
    }

    propertySeq_ = fixedSeq;

    return propertySeq_.length() == 0;
}

Boolean
PropertySetDef_impl::is_property_defined(const char* property_name)
    throw(InvalidPropertyName, SystemException)
{
//    OB_SYNCHRONIZED

    return get(property_name) != 0;
}

void
PropertySetDef_impl::get_allowed_property_types(
    PropertyTypes_out property_types)
    throw(SystemException)
{
    property_types = new PropertyTypes(allowedTypes_);
}

void
PropertySetDef_impl::get_allowed_properties(PropertyDefs_out property_defs)
    throw(SystemException)
{
    property_defs = new PropertyDefs(allowedProperties_);
}

void
PropertySetDef_impl::define_property_with_mode(const char* property_name,
					       const Any& property_value,
					       PropertyModeType property_mode)
    throw(InvalidPropertyName, ConflictingProperty, UnsupportedTypeCode,
	  UnsupportedProperty, UnsupportedMode, ReadOnlyProperty,
	  SystemException)
{
//    OB_SYNCHRONIZED

    //
    // Check name, type and property
    //
    check(property_name, property_value, property_mode);

    PropertyDef* p = get(property_name);

    //
    // Overwrite existing property if allowed
    //
    if(p)
    {
	if(p -> property_mode == read_only ||
	   p -> property_mode == fixed_readonly)
	    throw ReadOnlyProperty();

	//
	// Check type
	//
        TypeCode_var tc1 = p -> property_value.type();
        TypeCode_var tc2 = property_value.type();
        if(!tc1 -> equal(tc2))
	    throw ConflictingProperty();

	//
	// Set new value
	//
	p -> property_value = property_value;
    }
    else
    {
	PropertyDef prop;
	prop.property_name = property_name;
	prop.property_value = property_value;
	prop.property_mode = property_mode;
	propertySeq_.append(prop);
    }
}

void
PropertySetDef_impl::define_properties_with_modes(
    const PropertyDefs& property_defs)
    throw(MultipleExceptions, SystemException)
{
//    OB_SYNCHRONIZED

    //
    // Sequence with exceptions
    //
    MultipleExceptions mex;

    ULong len = property_defs.length();
    for(ULong i = 0 ; i < len ; i++)
    {
	try
	{
	    check(property_defs[i].property_name,
		  property_defs[i].property_value,
		  property_defs[i].property_mode);
	}
	catch(const InvalidPropertyName&)
	{
	    PropertyException ex;
	    ex.reason = invalid_property_name;
	    ex.failing_property_name = property_defs[i].property_name;
	    mex.exceptions.append(ex);
	    continue;
	}
	catch(const UnsupportedTypeCode&)
	{
	    PropertyException ex;
	    ex.reason = unsupported_type_code;
	    ex.failing_property_name = property_defs[i].property_name;
	    mex.exceptions.append(ex);
	    continue;
	}
	catch(const UnsupportedProperty&)
	{
	    PropertyException ex;
	    ex.reason = unsupported_property;
	    ex.failing_property_name = property_defs[i].property_name;
	    mex.exceptions.append(ex);
	    continue;
	}

	PropertyDef* p = get(property_defs[i].property_name);

	//
	// Overwrite existing property if allowed
	//
	if(p)
	{
	    if(p -> property_mode == read_only ||
	       p -> property_mode == fixed_readonly)
	    {
		PropertyException ex;
		ex.reason = read_only_property;
		ex.failing_property_name = property_defs[i].property_name;
		mex.exceptions.append(ex);
		continue;
	    }

	    //
	    // Check type
	    //
            TypeCode_var tc1 = p -> property_value.type();
            TypeCode_var tc2 = property_defs[i].property_value.type();
            if(!tc1 -> equal(tc2))
	    {
		PropertyException ex;
		ex.reason = conflicting_property;
		ex.failing_property_name = property_defs[i].property_name;
		mex.exceptions.append(ex);
		continue;
	    }

	    //
	    // Set new value
	    //
	    p -> property_value = property_defs[i].property_value;
	}
	else
	    propertySeq_.append(property_defs[i]);
    }

    //
    // Check for exceptions
    //
    if(mex.exceptions.length())
	throw mex;
}

PropertyModeType
PropertySetDef_impl::get_property_mode(const char* property_name)
    throw(PropertyNotFound, InvalidPropertyName, SystemException)
{
//    OB_SYNCHRONIZED

    if(strlen(property_name) == 0)
	throw InvalidPropertyName();

    PropertyDef* p = get(property_name);
    if(!p)
	throw PropertyNotFound();

    return p -> property_mode;
}

Boolean
PropertySetDef_impl::get_property_modes(const PropertyNames& property_names,
					PropertyModes_out property_modes)
    throw(SystemException)
{
//    OB_SYNCHRONIZED

    bool allDefined = true;

    property_modes = new PropertyModes;

    for(ULong i = 0 ; i < property_names.length() ; i++)
    {
	PropertyMode m;
	m.property_name = property_names[i];

	PropertyDef* p = get(property_names[i]);
	if(p)
	    m.property_mode = p -> property_mode;
	else
	{
	    m.property_mode = undefined;
	    allDefined = false;
	}

	property_modes -> append(m);
    }

    return allDefined;
}

void
PropertySetDef_impl::set_property_mode(const char* property_name,
				       PropertyModeType property_mode)
    throw(InvalidPropertyName, PropertyNotFound, UnsupportedMode,
	  SystemException)
{
//    OB_SYNCHRONIZED

    if(strlen(property_name) == 0)
	throw InvalidPropertyName();

    PropertyDef* p = get(property_name);
    if(!p)
	throw PropertyNotFound();

    if(property_mode == undefined)
	throw UnsupportedMode();

    p -> property_mode = property_mode;
}

void
PropertySetDef_impl::set_property_modes(const PropertyModes& property_modes)
    throw(MultipleExceptions, SystemException)
{
//    OB_SYNCHRONIZED

    //
    // Sequence with exceptions
    //
    MultipleExceptions mex;

    for(ULong i = 0 ; i < property_modes.length() ; i++)
    {
	if(strlen(property_modes[i].property_name) == 0)
	{
	    PropertyException ex;
	    ex.reason = invalid_property_name;
	    ex.failing_property_name = property_modes[i].property_name;
	    mex.exceptions.append(ex);
	    continue;
	}

	PropertyDef* p = get(property_modes[i].property_name);
	if(!p)
	{
	    PropertyException ex;
	    ex.reason = property_not_found;
	    ex.failing_property_name = property_modes[i].property_name;
	    mex.exceptions.append(ex);
	    continue;
	}

	if(property_modes[i].property_mode == undefined)
	{
	    PropertyException ex;
	    ex.reason = unsupported_mode;
	    ex.failing_property_name = property_modes[i].property_name;
	    mex.exceptions.append(ex);
	    continue;
	}

	p -> property_mode = property_modes[i].property_mode;
    }

    //
    // Check for exceptions
    //
    if(mex.exceptions.length())
	throw mex;
}

// ----------------------------------------------------------------------
// PropertyNamesIterator_impl constructor
// ----------------------------------------------------------------------

PropertyNamesIterator_impl::PropertyNamesIterator_impl(
    PortableServer::POA_ptr poa, const PropertyNames& names)
    : poa_(PortableServer::POA::_duplicate(poa)), idx_(0), names_(names)
{
}

// ----------------------------------------------------------------------
// PropertyNamesIterator_impl public member implementation
// ----------------------------------------------------------------------

void
PropertyNamesIterator_impl::reset()
    throw(SystemException)
{
//    OB_SYNCHRONIZED

    idx_ = 0;
}

Boolean
PropertyNamesIterator_impl::next_one(PropertyName_out property_name)
    throw(SystemException)
{
//    OB_SYNCHRONIZED

    if(idx_ < names_.length())
    {
        property_name = string_dup(names_[idx_++]);
        return true;
    }
    else
        return false;
}

Boolean
PropertyNamesIterator_impl::next_n(ULong how_many,
				   PropertyNames_out property_names)
    throw(SystemException)
{
//    OB_SYNCHRONIZED

    ULong len = names_.length();
    ULong num = len - idx_ < how_many ? len - idx_ : how_many;

    property_names = new PropertyNames(num);

    for(ULong i = 0 ; i < num ; i++)
        (*property_names)[i] = string_dup(names_[idx_++]);
 
    return how_many <= num;
}

void
PropertyNamesIterator_impl::destroy()
    throw(SystemException)
{
//    OB_SYNCHRONIZED

    PortableServer::POA_var poa = _default_POA();
    PortableServer::ObjectId_var oid = poa -> servant_to_id(this);
    poa -> deactivate_object(oid);
}

// ----------------------------------------------------------------------
// PropertiesIterator_impl constructor
// ----------------------------------------------------------------------

PropertiesIterator_impl::PropertiesIterator_impl(
    PortableServer::POA_ptr poa,
    const PropertyDefs& properties)
    : poa_(PortableServer::POA::_duplicate(poa)), idx_(0),
      properties_(properties)
{
}

// ----------------------------------------------------------------------
// PropertiesIterator_impl public member implementation
// ----------------------------------------------------------------------

void
PropertiesIterator_impl::reset()
    throw(SystemException)
{
//    OB_SYNCHRONIZED

    idx_ = 0;
}

Boolean
PropertiesIterator_impl::next_one(Property_out aproperty)
    throw(SystemException)
{
//    OB_SYNCHRONIZED

    if(idx_ < properties_.length())
    {
        aproperty = new Property;
	aproperty -> property_name = properties_[idx_].property_name;
	aproperty -> property_value = properties_[idx_].property_value;
	idx_++;
        return true;
    }
    else
        return false;
}

Boolean
PropertiesIterator_impl::next_n(ULong how_many, Properties_out nproperties)
    throw(SystemException)
{
//    OB_SYNCHRONIZED

    ULong len = properties_.length();
    ULong num = len - idx_ < how_many ? len - idx_ : how_many;

    nproperties = new Properties(num);
    nproperties -> length(num);

    for(ULong i = 0 ; i < num ; i++)
    {
	(*nproperties)[i].property_name = properties_[idx_].property_name;
	(*nproperties)[i].property_value = properties_[idx_].property_value;
	idx_++;
    }
 
    return how_many <= num;
}

void
PropertiesIterator_impl::destroy()
    throw(SystemException)
{
//    OB_SYNCHRONIZED

    PortableServer::POA_var poa = _default_POA();
    PortableServer::ObjectId_var oid = poa -> servant_to_id(this);
    poa -> deactivate_object(oid);
}

// ----------------------------------------------------------------------
// PropertySetFactory _impl public member implementation
// ----------------------------------------------------------------------

PropertySet_ptr
PropertySetFactory_impl::create_propertyset()
    throw(SystemException)
{
    PortableServer::POA_var poa = _default_POA();
    PropertySetDef_impl* propertySetDefImpl = new PropertySetDef_impl(poa);
    return propertySetDefImpl -> _this();
}

PropertySet_ptr
PropertySetFactory_impl::create_constrained_propertyset(
    const PropertyTypes& allowed_property_types,
    const Properties& allowed_properties)
    throw(SystemException)
{
    PropertyDefs p;
    ULong len = allowed_properties.length();
    p.length(len);

    for(ULong i = 0 ; i < len ; i++)
    {
	p[i].property_name = allowed_properties[i].property_name;
	p[i].property_value = allowed_properties[i].property_value;
	p[i].property_mode = normal;
    }

    PortableServer::POA_var poa = _default_POA();
    PropertySetDef_impl* propertySetDefImpl =
	new PropertySetDef_impl(poa, allowed_property_types, p);
    return propertySetDefImpl -> _this();
}

PropertySet_ptr
PropertySetFactory_impl::create_initial_propertyset(
    const Properties& initial_properties)
    throw(SystemException)
{
    //
    // Sequence with exceptions
    //
    MultipleExceptions mex;

    PropertyDefs p;
    ULong len = initial_properties.length();
    p.length(len);

    for(ULong i = 0 ; i < len ; i++)
    {
	if(strlen(initial_properties[i].property_name) == 0)
	{
	    PropertyException ex;
	    ex.reason = invalid_property_name;
	    ex.failing_property_name = initial_properties[i].property_name;
	    mex.exceptions.append(ex);
	    continue;
	}

	p[i].property_name = initial_properties[i].property_name;
	p[i].property_value = initial_properties[i].property_value;
	p[i].property_mode = normal;
    }

    //
    // Check for exceptions
    //
    if(mex.exceptions.length())
	throw mex;
 
    PortableServer::POA_var poa = _default_POA();
    PropertySetDef_impl* propertySetDefImpl = new PropertySetDef_impl(poa, p);
    return propertySetDefImpl -> _this();
}

// ----------------------------------------------------------------------

PropertySetDef_ptr
PropertySetDefFactory_impl::create_propertysetdef()
    throw(SystemException)
{
    PortableServer::POA_var poa = _default_POA();
    PropertySetDef_impl* propertySetDefImpl = new PropertySetDef_impl(poa);
    return propertySetDefImpl -> _this();
}

PropertySetDef_ptr
PropertySetDefFactory_impl::create_constrained_propertysetdef(
    const PropertyTypes& allowed_property_types,
    const PropertyDefs& allowed_property_defs)
    throw(SystemException)
{
    PortableServer::POA_var poa = _default_POA();
    PropertySetDef_impl* propertySetDefImpl =
	new PropertySetDef_impl(poa, allowed_property_types,
				allowed_property_defs);
    return propertySetDefImpl -> _this();
}

PropertySetDef_ptr
PropertySetDefFactory_impl::create_initial_propertysetdef(
    const PropertyDefs& initial_property_defs)
    throw(SystemException)
{
    //
    // Sequence with exceptions
    //
    MultipleExceptions mex;

    ULong len = initial_property_defs.length();
    for(ULong i = 0 ; i < len ; i++)
    {
	if(strlen(initial_property_defs[i].property_name) == 0)
	{
	    PropertyException ex;
	    ex.reason = invalid_property_name;
	    ex.failing_property_name = initial_property_defs[i].property_name;
	    mex.exceptions.append(ex);
	    continue;
	}
    }

    //
    // Check for exceptions
    //
    if(mex.exceptions.length())
	throw mex;

    PortableServer::POA_var poa = _default_POA();
    PropertySetDef_impl* propertySetDefImpl =
	new PropertySetDef_impl(poa, initial_property_defs);
    return propertySetDefImpl -> _this();
}
