// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>
#include <OB/CosProperty_impl.h>

#include <Config.h> // Make sure that the compiler complains if the
                    // configuration doesn't match

#include <stdlib.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;

static void
usage(const char* progName)
{
    cerr << "Usage:\n";
    cerr << progName << " [options]\n"
	"\n"
	"Options:\n"
	"-h, --help             Show this message.\n"
	"-v, --version          Show version number.\n"
	"-i, --ior              Print IOR on standard output.\n"
	;
}

int
main(int argc, char* argv[], char*[])
{
    //
    // Set program name
    //
    const char* progName = argv[0];

    try
    {
        //
        // Create ORB
        //
        ORB_var orb = ORB_init(argc, argv);
 
        //
        // Resolve Root POA
        //
        Object_var poaObj = orb -> resolve_initial_references("RootPOA");
        PortableServer::POA_var rootPOA = PortableServer::POA::_narrow(poaObj);
 
        //
        // Get a reference to the POA manager
        //
        PortableServer::POAManager_var manager = rootPOA -> the_POAManager();

        CORBA::PolicyList policies(1);
        policies.length(1);
        policies[0] =
            rootPOA -> create_id_assignment_policy(PortableServer::USER_ID);
        PortableServer::POA_var poa =
            rootPOA -> create_POA("factory", manager, policies);

	//
	// Resolve the service object if available as an initial reference.
	//
	const char* service = "PropertyService";
	Object_var serviceObject;
	try
	{
	    serviceObject = orb -> resolve_initial_references(service);
	}
	catch(ORB::InvalidName&)
	{
	}

	//
	// Set the IIOP port based on the port defined in the IIOP profile
	// of the service object (if it exists).
	//
	ORB::set_iiop_port_from_object(serviceObject);

	//
	// Get options
	//
	String_var cppArgs;
	bool ior = false;

	Long i;
	for(i = 1 ; i < argc && *argv[i] == '-' ; i++)
	{
	    if(strcmp(argv[i], "--help") == 0 ||
	       strcmp(argv[i], "-h") == 0)
	    {
		usage(progName);
		return 0;
	    }
	    else if(strcmp(argv[i], "--version") == 0 ||
		    strcmp(argv[i], "-v") == 0)
	    {
		cerr << "Property Service " << OB::Version << endl;
		return 0;
	    }
	    else if(strcmp(argv[i], "--ior") == 0 ||
		    strcmp(argv[i], "-i") == 0)
	    {
		ior = true;
	    }
	    else
	    {
		cerr << progName << ": unknown option `"
		     << argv[i] << "'\n" << endl;
		usage(progName);
		return 1;
	    }
	}

	//
	// Create factory
	//
	CosPropertyService::PropertySetDefFactory_impl defFactory(poa);
        PortableServer::ObjectId_var oid =
	    PortableServer::string_to_ObjectId("DefaultPropertySetDefFactory");
        poa -> activate_object_with_id(oid.in(), &defFactory);

	//
	// Print IOR on standard output
	//
	if(ior)
	{
	    Object_var obj = poa -> id_to_reference(oid);
	    String_var s = orb -> object_to_string(obj);
	    cout << s << endl;
	}

	//
	// Run implementation
	//
        manager -> activate();
        orb -> run();
    }
    catch(SystemException& ex)
    {
        OB::PrintException(ex);
        return 1;
    }
 
    return 0;
}

