// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>
#include <OB/CosProperty_impl.h>

void TestUnconstrained(CosPropertyService::PropertySetDef_ptr);
void TestConstrained(CosPropertyService::PropertySetDef_ptr);
void TestInitial(CosPropertyService::PropertySetDef_ptr);

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

int
main(int argc, char* argv[], char*[])
{
    try
    {
        //
        // Create ORB
        //
        CORBA::ORB_var orb = CORBA::ORB_init(argc, argv);
 
        //
        // Resolve Root POA
        //
        CORBA::Object_var poaObj =
	    orb -> resolve_initial_references("RootPOA");
        PortableServer::POA_var rootPOA = PortableServer::POA::_narrow(poaObj);

	CosPropertyService::PropertySetDefFactory_impl factory(rootPOA);

	CosPropertyService::PropertySetDef_var set;

	//
        // Run tests with unconstrained PropertySetDef
        //
	cout << "Running tests with unconstrained PropertySet" << endl;
	set = factory.create_propertysetdef();
	TestUnconstrained(set);

        //
        // Set up PropertySetDef constraints, only tk_string is allowed
        //
	CosPropertyService::PropertyTypes allowedTypes;
	allowedTypes.length(1);
	allowedTypes[0] = CORBA::TypeCode::_duplicate(CORBA::_tc_string);

	CORBA::Any any;
	any <<= CORBA::string_dup("property_value");

	CosPropertyService::PropertyDefs allowedProperties;
	allowedProperties.length(3);
	CosPropertyService::PropertyDef p;

	p.property_name = CORBA::string_dup("read_only");
	p.property_value = any;
	p.property_mode = CosPropertyService::read_only;
	allowedProperties[0] = p;

	p.property_name = CORBA::string_dup("fixed_normal");
	p.property_value = any;
	p.property_mode = CosPropertyService::fixed_normal;
	allowedProperties[1] = p;

	p.property_name = CORBA::string_dup("fixed_readonly");
	p.property_value = any;
	p.property_mode = CosPropertyService::fixed_readonly;
	allowedProperties[2] = p;

	cout << "Running tests with constrained PropertySet" << endl;
	set = factory.create_constrained_propertysetdef(allowedTypes,
							allowedProperties);
	TestConstrained(set);

	//
        // Set up initial PropertySetDef
        //
	any <<= CORBA::string_dup("initial_property_value");
	CosPropertyService::PropertyDefs initialProperties;
	p.property_name = CORBA::string_dup("initial_property");
	p.property_value = any;
	p.property_mode = CosPropertyService::normal;
	initialProperties.length(1);
	initialProperties[0] = p;

	cout << "Running tests with initial PropertySet" << endl;
	set = factory.create_initial_propertysetdef(initialProperties);
	TestInitial(set);
    }
    catch(CORBA::SystemException& ex)
    {
        OB::PrintException(ex);
	return 1;
    }
 
    return 0;
}

//
// Tests for unconstrained PropertySetDef
//
void
TestUnconstrained(CosPropertyService::PropertySetDef_ptr set)
{
    CORBA::ULong i, j;

    {
	cout << "  Testing define_property()... ";
	cout.flush();

	for(i = 0 ; i < 10 ; i++)
	{
	    CORBA::Any any;
	    any <<= i;

	    try
	    {
		CORBA::String_var s = CORBA::string_dup("property_");
		s += i;
		set -> define_property(s, any);
	    }
	    catch(const CosPropertyService::ReadOnlyProperty&)
	    {
		assert(false);
	    }
	    catch(const CosPropertyService::ConflictingProperty&)
	    {
		assert(false);
	    }
	    catch(const CosPropertyService::UnsupportedProperty&)
	    {
		assert(false);
	    }
	    catch(const CosPropertyService::UnsupportedTypeCode&)
	    {
		assert(false);
	    }
	    catch(const CosPropertyService::InvalidPropertyName&)
	    {
		assert(false);
	    }
	}

	cout << "OK!" << endl;
    }

    {
	cout << "  Testing define_properties()... ";
	cout.flush();

	CosPropertyService::Properties nproperties;
	nproperties.length(10);

	for(i = 10 ; i < 20 ; i++)
	{
	    CORBA::Any any;
	    any <<= i;
	    CORBA::String_var s = CORBA::string_dup("property_");
	    s += i;
	    CosPropertyService::Property property;
	    property.property_name = s;
	    property.property_value = any;
	    nproperties[i - 10] = property;
	}

	try
	{
	    set -> define_properties(nproperties);
	}
	catch(const CosPropertyService::MultipleExceptions&)
	{
	    assert(false);
	}

	cout << "OK!" << endl;
    }

    {
	cout << "  Testing get_all_property_names() without iterator... ";
	cout.flush();

	CosPropertyService::PropertyNames_var names;
	CosPropertyService::PropertyNamesIterator_var iterator;

	CORBA::ULong n = set -> get_number_of_properties();
	set -> get_all_property_names(n, names.out(), iterator.out());

	assert(names -> length() == n);

	CosPropertyService::PropertyNames v;
	v.length(n);

	for(i = 0 ; i < n ; i++)
	{
	    CORBA::String_var s = CORBA::string_dup("property_");
	    s += i;
	    v[i] = s;
	}

	//
	// Check if each property name exists only once
	//
	for(i = 0 ; i < n ; i++)
	{
	    for(j = 0 ; j < v.length() ; j++)
	    {
		if(!strcmp(names[i], v[j]))
		    break;

		assert(false);
	    }

	    v.remove(j);
	}

	cout << "OK!" << endl;
    }

    {
	cout << "  Testing get_all_property_names() with iterator... ";
	cout.flush();

	CosPropertyService::PropertyNames_var names;
	CosPropertyService::PropertyNamesIterator_var iterator;

	//
	// Get one name without iterator
	//
	set -> get_all_property_names(1, names.out(), iterator.out());
	assert(names -> length() == 1);

	CORBA::ULong n = set -> get_number_of_properties();

	names -> length(n);

	//
	// Get remaining names with iterator
	//
	CosPropertyService::PropertyName_var name;
	i = 1;
	while(iterator -> next_one(name.out()))
	    names[i++] = name;

	CosPropertyService::PropertyNames v;
	v.length(n);

	for(i = 0 ; i < n ; i++)
	{
	    CORBA::String_var s = CORBA::string_dup("property_");
	    s += i;
	    v[i] = s;
	}

	//
	// Check if each property name exists only once
	//
	for(i = 0 ; i < n ; i++)
	{
	    for(j = 0 ; j < v.length() ; j++)
	    {
		if(!strcmp(names[i], v[j]))
		    break;

		assert(false);
	    }

	    v.remove(j);
	}

	iterator -> destroy();

	cout << "OK!" << endl;
    }

    {
	cout << "  Testing get_property_value()... ";
	cout.flush();

	CORBA::ULong n = set -> get_number_of_properties();
	for(i = 0 ; i < n ; i++)
	{
	    CORBA::Any_var any;

	    try
	    {
		CORBA::String_var s = CORBA::string_dup("property_");
		s += i;
		any = set -> get_property_value(s);
	    }
	    catch(const CosPropertyService::PropertyNotFound&)
	    {
		assert(false);
	    }

	    CORBA::ULong value;
	    any >>= value;
	    assert(value == i);
	}

	cout << "OK!" << endl;
	}

    {
	cout << "  Testing get_properties()... ";
	cout.flush();

	CosPropertyService::PropertyNames names;
	CORBA::ULong n = set -> get_number_of_properties();
	names.length(n);
	for(i = 0 ; i < n ; i++)
	{
	    CORBA::String_var s = CORBA::string_dup("property_");
	    s += i;
	    names[i] = s;
	}

	CosPropertyService::Properties_var nproperties;
	set -> get_properties(names, nproperties.out());

	CosPropertyService::PropertyNames vs;
	vs.length(n);
	CORBA::ULong vi[20];
	for(i = 0 ; i < n ; i++)
	{
	    CORBA::String_var s = CORBA::string_dup("property_");
	    s += i;
	    vs[i] = s;
	    vi[i] = i;
	}

	//
	// Check if each property name exists only once
	//
	CORBA::ULong match = 0;
	for(i = 0 ; i < n ; i++)
	{
	    for(j = 0 ; j < n ; j++)
	    {
		CORBA::String_var s = nproperties[i].property_name;
		if(strcmp(s, vs[j]))
		    continue;

		CORBA::Any any = nproperties[i].property_value;
		CORBA::ULong value;
		any >>= value;
		if(value != vi[j])
		    continue;

		//
		// Mark current element as invalid, it has passed the test
		//
		vs[j] = CORBA::string_dup("");
		vi[j] = 9999;

		match++;
	    }
	}

	assert(match == n);

	cout << "OK!" << endl;
    }

    {
	cout << "  Testing get_all_properties() without iterator... ";
	cout.flush();

	CosPropertyService::Properties_var properties;
	CosPropertyService::PropertiesIterator_var iterator;
	CORBA::ULong n = set -> get_number_of_properties();
	set -> get_all_properties(n, properties.out(), iterator.out());

	assert(properties -> length() == n);

	CosPropertyService::PropertyNames vs;
	vs.length(n);
	CORBA::ULong vi[20];

	for(i = 0 ; i < n ; i++)
	{
	    CORBA::String_var s = CORBA::string_dup("property_");
	    s += i;
	    vs[i] = s;
	    vi[i] = i;
	}

	//
	// Check if each property exists only once
	//
	CORBA::ULong match = 0;
	for(i = 0 ; i < n ; i++)
	{
	    for(j = 0 ; j < n ; j++)
	    {
		CORBA::String_var s = properties[i].property_name;
		if(strcmp(s, vs[j]))
		    continue;

		CORBA::Any any = properties[i].property_value;
		CORBA::ULong value;
		any >>= value;
		if(value != vi[j])
		    continue;

		//
		// Mark current element as invalid, it has passed the test
		//
		vs[j] = CORBA::string_dup("");
		vi[j] = 9999;

		match++;
	    }
	}

	assert(match == n);

	cout << "OK!" << endl;
    }

    {
	cout << "  Testing get_all_properties() with iterator... ";
	cout.flush();

	CosPropertyService::Properties_var properties;
	CosPropertyService::PropertiesIterator_var iterator;

	//
	// Get one property without iterator
	//
	set -> get_all_properties(1, properties.out(), iterator.out());
	assert(properties -> length() == 1);

	CORBA::ULong n = set -> get_number_of_properties();
	properties -> length(n);

	//
	// Get remaining properties with iterator
	//
	CosPropertyService::Property_var name;
	i = 1;
	while(iterator -> next_one(name.out()))
	    properties[i++] = name;

	CosPropertyService::PropertyNames vs;
	vs.length(n);
	CORBA::ULong vi[20];
	for(i = 0 ; i < n ; i++)
	{
	    CORBA::String_var s = CORBA::string_dup("property_");
	    s += i;
	    vs[i] = s;
	    vi[i] = i;
	}

	//
	// Check if each property exists only once
	//
	CORBA::ULong match = 0;
	for(i = 0 ; i < n ; i++)
	{
	    for(j = 0 ; j < n ; j++)
	    {
		CORBA::String_var s = properties[i].property_name;
		if(strcmp(s, vs[j]))
		    continue;

		CORBA::Any any = properties[i].property_value;
		CORBA::ULong value;
		any >>= value;
		if(value != vi[j])
		    continue;

		//
		// Mark current element as invalid, it has passed the test
		//
		vs[j] = CORBA::string_dup("");
		vi[j] = 9999;

		match++;
	    }
	}

	assert(match == n);

	iterator -> destroy();

	cout << "OK!" << endl;
    }

    {
	cout << "  Testing is_property_defined()... ";
	cout.flush();

	CORBA::ULong n = set -> get_number_of_properties();
	for(i = 0 ; i < n ; i++)
	{
	    CORBA::String_var name = CORBA::string_dup("property_");
	    name += i;

	    try
	    {
		assert(set -> is_property_defined(name));
	    }
	    catch(const CosPropertyService::InvalidPropertyName&)
	    {
		assert(false);
	    }
	}

	cout << "OK!" << endl;
    }

    {
	cout << "  Testing delete_property()... ";
	cout.flush();

	//
	// Delete first 5 properties, 15 remain
	//
	for(i = 0 ; i < 5 ; i++)
	{
	    CORBA::String_var name = CORBA::string_dup("property_");
	    name += i;

	    try
	    {
		set -> delete_property(name);
	    }
	    catch(const CosPropertyService::InvalidPropertyName&)
	    {
		assert(false);
	    }
	    catch(const CosPropertyService::PropertyNotFound&)
	    {
		assert(false);
	    }
	    catch(const CosPropertyService::FixedProperty&)
	    {
		assert(false);
	    }
	}

	cout << "OK!" << endl;
    }

    {
	cout << "  Testing delete_properties()... ";
	cout.flush();

	//
	// Delete next 10 properties, 5 remain
	//
	CosPropertyService::PropertyNames names;
	names.length(10);

	for(i = 5 ; i < 15 ; i++)
	{
	    CORBA::String_var s = CORBA::string_dup("property_");
	    s += i;

	    names[i - 5] = s;
	}

	try
	{
	    set -> delete_properties(names);
	}
	catch(const CosPropertyService::MultipleExceptions&)
	{
	    assert(false);
	}

	cout << "OK!" << endl;
    }

    {
	cout << "  Testing delete_all_properties()... ";
	cout.flush();

	//
	// 10 properties remain and are deleted now
	//
	assert(set -> delete_all_properties());

	assert(!set -> get_number_of_properties());

	cout << "OK!" << endl;
    }
}

//
// Tests for constrained PropertySetDef
//
void
TestConstrained(CosPropertyService::PropertySetDef_ptr set)
{
    CORBA::Any any;
    bool except;

    {
	cout << "  Testing get_allowed_property_types()... ";
	cout.flush();

	CosPropertyService::PropertyTypes_var types;
	set -> get_allowed_property_types(types.out());

	assert(types -> length() == 1);

	assert(types[0] -> equal(CORBA::_tc_string));

	cout << "OK!" << endl;
    }

    {
	cout << "  Testing get_allowed_properties()... ";
	cout.flush();

	CosPropertyService::PropertyDefs_var properties;
	set -> get_allowed_properties(properties);

	assert(properties -> length() == 3);

	cout << "OK!" << endl;
    }

    {
	cout << "  Testing define_property_with_mode()... ";
	cout.flush();

	//
	// Try to add allowed properties
	//
	any <<= CORBA::string_dup("property_value");
	try
	{
	    set -> define_property_with_mode("fixed_normal", any,
					     CosPropertyService::fixed_normal);
	    set -> define_property_with_mode("read_only", any,
					     CosPropertyService::read_only);
	    set -> define_property_with_mode(
		"fixed_readonly", any, CosPropertyService::fixed_readonly);
	}
	catch(const CosPropertyService::ReadOnlyProperty&)
	{
	    assert(false);
	}
	catch(const CosPropertyService::ConflictingProperty&)
	{
	    assert(false);
	}
	catch(const CosPropertyService::UnsupportedProperty&)
	{
	    assert(false);
	}
	catch(const CosPropertyService::UnsupportedTypeCode&)
	{
	    assert(false);
	}
	catch(const CosPropertyService::InvalidPropertyName&)
	{
	    assert(false);
	}
	catch(const CosPropertyService::UnsupportedMode&)
	{
	    assert(false);
	}

	//
	// Try to change readonly property
	//
	any <<= CORBA::string_dup("read_only");
	except = false;
	try
	{
	    set -> define_property("read_only", any);
	}
	catch(const CosPropertyService::ReadOnlyProperty&)
	{
	    // This exception must be thrown
	    except = true;
	}
	catch(const CosPropertyService::ConflictingProperty&)
	{
	    assert(false);
	}
	catch(const CosPropertyService::UnsupportedProperty&)
	{
	    assert(false);
	}
	catch(const CosPropertyService::UnsupportedTypeCode&)
	{
	    assert(false);
	}
	catch(const CosPropertyService::InvalidPropertyName&)
	{
	    assert(false);
	}

	assert(except);

	//
	// Try to add illegal TypeCode
	//
	any <<= (CORBA::ULong)1;
	except = false;
	try
	{
	    set -> define_property("fixed_normal", any);
	}
	catch(const CosPropertyService::ReadOnlyProperty&)
	{
	    assert(false);
	}
	catch(const CosPropertyService::ConflictingProperty&)
	{
	    assert(false);
	}
	catch(const CosPropertyService::UnsupportedProperty&)
	{
	    // This exception must be thrown
	    except = true;
	}
	catch(const CosPropertyService::UnsupportedTypeCode&)
	{
	    // This exception must be thrown
	    except = true;
	}
	catch(const CosPropertyService::InvalidPropertyName&)
	{
	    assert(false);
	}
      
	assert(except);

	cout << "OK!" << endl;
    }

    {
	cout << "  Testing get_property_mode()... ";
	cout.flush();

	CosPropertyService::PropertyModeType fixed_normal;
	CosPropertyService::PropertyModeType read_only;
	CosPropertyService::PropertyModeType fixed_readonly;
	try
	{
	    fixed_normal = set -> get_property_mode("fixed_normal");
	    read_only = set -> get_property_mode("read_only");
	    fixed_readonly = set -> get_property_mode("fixed_readonly");
	}
	catch(const CosPropertyService::InvalidPropertyName&)
	{
	    assert(false);
	}
	catch(const CosPropertyService::PropertyNotFound&)
	{
	    assert(false);
	}

	assert(fixed_normal == CosPropertyService::fixed_normal ||
               read_only == CosPropertyService::read_only ||
               fixed_readonly == CosPropertyService::fixed_readonly);

	cout << "OK!" << endl;
    }

    {
	cout << "  Testing delete_property()... ";

	except = false;
	//
	// Try to delete fixed_normal property;
	//
	try
	{
	    set -> delete_property("fixed_normal");
	}
	catch(const CosPropertyService::FixedProperty&)
	{
	    // This exception must be thrown
	    except = true;
	}
	catch(const CosPropertyService::InvalidPropertyName&)
	{
	    assert(false);
	}
	catch(const CosPropertyService::PropertyNotFound&)
	{
	    assert(false);
	}

	assert(except);

	cout << "OK!" << endl;
    }
}

//
// Tests for initial PropertySetDef
//
void
TestInitial(CosPropertyService::PropertySetDef_ptr set)
{
    {
	cout << "  Testing get_number_of_properties()... ";
	cout.flush();

	assert(set -> get_number_of_properties() == 1);

	cout << "OK!" << endl;
    }

    {
	cout << "  Testing is_property_defined()... ";
	cout.flush();

	try
	{
	    assert(set -> is_property_defined("initial_property"));
	}
	catch(const CosPropertyService::InvalidPropertyName&)
	{
	    assert(false);
	}

	try
	{
	    assert(!set -> is_property_defined("illegal_property"));
	}
	catch(const CosPropertyService::InvalidPropertyName&)
	{
	    assert(false);
	}

	cout << "OK!" << endl;
    }

    {
	cout << "  Testing get_property_value()... ";
	cout.flush();

	CORBA::Any_var any;

	try
	{
	    any = set -> get_property_value("initial_property");
	}
	catch(const CosPropertyService::InvalidPropertyName&)
	{
	    assert(false);
	}
	catch(const CosPropertyService::PropertyNotFound&)
	{
	    assert(false);
	}

	char* value;
	any >>= value;
	assert(!strcmp(value, "initial_property_value"));

	cout << "OK!" << endl;
    }
}
